/* Generated by Cython 3.0.0a11 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#define CYTHON_ABI "3_0_0a11"
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030000AB
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(CYTHON_LIMITED_API)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS 1
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(p))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE(obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyprotect__protected
#define __PYX_HAVE_API__pyprotect__protected
/* Early includes */
#include <string.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "python_visible.pxi",
  "ProtectionData.pxi",
  "Proxy.pxi",
  "Wrapped_Frozen.pxi",
  "PrivacyDict_FrozenPrivacyDict.pxi",
  "global_c_functions.pxi",
  "<stringsource>",
  "Private_FrozenPrivate.pxi",
  "Protected_FrozenProtected.pxi",
  "HiddenPartial.pxi",
  "protected.pyx",
  "type.pxd",
  "imports.pxi",
  "global_cdefs.pxi",
};
/* #### Code section: utility_code_proto_before_types ### */
/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_9pyprotect_9protected___ProtectionData;
struct __pyx_obj_9pyprotect_9protected_Proxy;
struct __pyx_obj_9pyprotect_9protected_Wrapped;
struct __pyx_obj_9pyprotect_9protected_Frozen;
struct __pyx_obj_9pyprotect_9protected_PrivacyDict;
struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict;
struct __pyx_obj_9pyprotect_9protected_Private;
struct __pyx_obj_9pyprotect_9protected_FrozenPrivate;
struct __pyx_obj_9pyprotect_9protected_Protected;
struct __pyx_obj_9pyprotect_9protected_FrozenProtected;
struct __pyx_obj_9pyprotect_9protected___HiddenPartial;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems;
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op;
struct __pyx_opt_args_9pyprotect_9protected_privatedict;
struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op;
struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_visible;
struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_writeable;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;

/* "global_c_functions.pxi":261
 * 
 * 
 * cdef privatedict(o, cn, frozen=False, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *     '''
 *     o-->Mapping (to be wrapped)
 */
struct __pyx_opt_args_9pyprotect_9protected_privatedict {
  int __pyx_n;
  PyObject *frozen;
  PyObject *oldstyle_class;
};

/* "Protected_FrozenProtected.pxi":72
 *             continue
 * 
 *     cdef check_1_op(self, a, op, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */
struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op {
  int __pyx_n;
  PyObject *use_cache;
};

/* "Protected_FrozenProtected.pxi":152
 *         return True
 * 
 *     cdef protected_visible(self, a, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */
struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_visible {
  int __pyx_n;
  PyObject *use_cache;
};

/* "Protected_FrozenProtected.pxi":167
 *         return self.check_1_op(a=a, op='r', use_cache=use_cache)
 * 
 *     cdef protected_writeable(self, a, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */
struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_writeable {
  int __pyx_n;
  PyObject *use_cache;
};
struct __pyx_defaults {
  PyObject *__pyx_arg_ro;
  PyObject *__pyx_arg_rw;
  PyObject *__pyx_arg_hide;
};

/* "ProtectionData.pxi":4
 * @cython.final
 * @cython.internal
 * cdef class __ProtectionData(object):             # <<<<<<<<<<<<<<
 *     '''
 *     Attributes:
 */
struct __pyx_obj_9pyprotect_9protected___ProtectionData {
  PyObject_HEAD
  PyObject *id;
  PyObject *id_class;
  PyObject *hash;
  PyObject *isinstance;
  PyObject *issubclass;
  PyObject *instanceof;
  PyObject *subclassof;
  PyObject *help;
  PyObject *help_str;
  PyObject *testop;
  PyObject *rules;
  PyObject *freeze;
  PyObject *private;
  PyObject *protect;
  PyObject *multiwrapped;
  PyObject *attributes_map;
};


/* "Proxy.pxi":3
 * 
 * # @cython.internal
 * cdef class Proxy(object):             # <<<<<<<<<<<<<<
 *     '''
 *     This is an object wrapper / proxy that implements the Python
 */
struct __pyx_obj_9pyprotect_9protected_Proxy {
  PyObject_HEAD
  PyObject *pvt_o;
  int frozen;
};


/* "Wrapped_Frozen.pxi":3
 * 
 * # @cython.internal
 * cdef class Wrapped(Proxy):             # <<<<<<<<<<<<<<
 *     '''
 *     This is an object wrapper / proxy that adds the 'frozen' parameter
 */
struct __pyx_obj_9pyprotect_9protected_Wrapped {
  struct __pyx_obj_9pyprotect_9protected_Proxy __pyx_base;
  struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *__pyx_vtab;
  struct __pyx_obj_9pyprotect_9protected___ProtectionData *protected_attribute;
  PyObject *cn;
  PyObject *rules;
  int oldstyle_class;
  PyObject *hidden_private_attr;
};


/* "Wrapped_Frozen.pxi":414
 * # @cython.internal
 * @cython.final
 * cdef class Frozen(Wrapped):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Wrapped that is automatically frozen
 */
struct __pyx_obj_9pyprotect_9protected_Frozen {
  struct __pyx_obj_9pyprotect_9protected_Wrapped __pyx_base;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":3
 * 
 * # @cython.internal
 * cdef class PrivacyDict(Wrapped):             # <<<<<<<<<<<<<<
 *     '''
 *     Like types.MappingProxyType - with following additional functionality:
 */
struct __pyx_obj_9pyprotect_9protected_PrivacyDict {
  struct __pyx_obj_9pyprotect_9protected_Wrapped __pyx_base;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":204
 * # @cython.internal
 * @cython.final
 * cdef class FrozenPrivacyDict(PrivacyDict):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of PrivacyDict that is automatically frozen
 */
struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict {
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict __pyx_base;
};


/* "Private_FrozenPrivate.pxi":3
 * 
 * # @cython.internal
 * cdef class Private(Wrapped):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Wrapped with following additional functionality:
 */
struct __pyx_obj_9pyprotect_9protected_Private {
  struct __pyx_obj_9pyprotect_9protected_Wrapped __pyx_base;
};


/* "Private_FrozenPrivate.pxi":144
 * # @cython.internal
 * @cython.final
 * cdef class FrozenPrivate(Private):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Private that is automatically frozen
 */
struct __pyx_obj_9pyprotect_9protected_FrozenPrivate {
  struct __pyx_obj_9pyprotect_9protected_Private __pyx_base;
};


/* "Protected_FrozenProtected.pxi":3
 * 
 * # @cython.internal
 * cdef class Protected(Private):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Private that further restriction of:
 */
struct __pyx_obj_9pyprotect_9protected_Protected {
  struct __pyx_obj_9pyprotect_9protected_Private __pyx_base;
  PyObject *acl_cache;
  PyObject *dir_out;
};


/* "Protected_FrozenProtected.pxi":279
 * 
 * # @cython.internal
 * cdef class FrozenProtected(Protected):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Protected that is automatically frozen
 */
struct __pyx_obj_9pyprotect_9protected_FrozenProtected {
  struct __pyx_obj_9pyprotect_9protected_Protected __pyx_base;
};


/* "HiddenPartial.pxi":3
 * 
 * # @cython.internal
 * cdef class __HiddenPartial(object):             # <<<<<<<<<<<<<<
 *     '''
 *     This is an object wrapper / proxy for functools.partial
 */
struct __pyx_obj_9pyprotect_9protected___HiddenPartial {
  PyObject_HEAD
  struct __pyx_vtabstruct_9pyprotect_9protected___HiddenPartial *__pyx_vtab;
  PyObject *args;
  PyObject *kwargs;
};


/* "Proxy.pxi":49
 *         return x
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         for x in iter(self.pvt_o):
 *             if self.frozen:
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ {
  PyObject_HEAD
  struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self;
  PyObject *__pyx_v_x;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Wrapped_Frozen.pxi":221
 *         return dict()
 * 
 *     cdef comparator(self, other, op):             # <<<<<<<<<<<<<<
 *         '''
 *         Operations:
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator {
  PyObject_HEAD
  PyObject *__pyx_v_op;
  PyObject *__pyx_v_other;
  struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":115
 *     # --------------------------------------------------------------------
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys {
  PyObject_HEAD
  PyObject *__pyx_v_k;
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":121
 *             yield k
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items {
  PyObject_HEAD
  PyObject *__pyx_v_k;
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":126
 *             yield self.fif((k, v))
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values {
  PyObject_HEAD
  PyObject *__pyx_v_k;
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":160
 *         return self.fif(ret)
 * 
 *     def iterkeys(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.pvt_o.keys():
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys {
  PyObject_HEAD
  PyObject *__pyx_v_k;
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":167
 *             yield k
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems {
  PyObject_HEAD
  PyObject *__pyx_v_k;
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "PrivacyDict_FrozenPrivacyDict.pxi":173
 *             yield self.fif((k, v))
 * 
 *     def itervalues(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */
struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues {
  PyObject_HEAD
  PyObject *__pyx_v_k;
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self")
 * cdef object __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(object (*f)(Wrapped) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self):
 *         """wrap(self: 'Wrapped')"""
 */
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self {
  PyObject_HEAD
  PyObject *(*__pyx_v_f)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c {
  PyObject_HEAD
  PyObject *(*__pyx_v_f)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op {
  PyObject_HEAD
  PyObject *(*__pyx_v_f)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *);
};



/* "Wrapped_Frozen.pxi":3
 * 
 * # @cython.internal
 * cdef class Wrapped(Proxy):             # <<<<<<<<<<<<<<
 *     '''
 *     This is an object wrapper / proxy that adds the 'frozen' parameter
 */

struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped {
  PyObject *(*attr_hidden)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*fif)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*freeze)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
  PyObject *(*multiwrapped)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
  PyObject *(*id_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
  PyObject *(*hash_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
  PyObject *(*isinstance_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*issubclass_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*instanceof_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*subclassof_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*help_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
  PyObject *(*help_str_protected)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
  PyObject *(*visible)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*writeable)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*testop)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *);
  PyObject *(*get_rules)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
  PyObject *(*comparator)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *);
  PyObject *(*wrapped_getattr)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*wrapped_check_setattr)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *);
  PyObject *(*wrapped_check_delattr)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *);
  PyObject *(*wrapped_dir)(struct __pyx_obj_9pyprotect_9protected_Wrapped *);
};
static struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *__pyx_vtabptr_9pyprotect_9protected_Wrapped;


/* "Wrapped_Frozen.pxi":414
 * # @cython.internal
 * @cython.final
 * cdef class Frozen(Wrapped):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Wrapped that is automatically frozen
 */

struct __pyx_vtabstruct_9pyprotect_9protected_Frozen {
  struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped __pyx_base;
};
static struct __pyx_vtabstruct_9pyprotect_9protected_Frozen *__pyx_vtabptr_9pyprotect_9protected_Frozen;


/* "PrivacyDict_FrozenPrivacyDict.pxi":3
 * 
 * # @cython.internal
 * cdef class PrivacyDict(Wrapped):             # <<<<<<<<<<<<<<
 *     '''
 *     Like types.MappingProxyType - with following additional functionality:
 */

struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict {
  struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped __pyx_base;
};
static struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *__pyx_vtabptr_9pyprotect_9protected_PrivacyDict;


/* "PrivacyDict_FrozenPrivacyDict.pxi":204
 * # @cython.internal
 * @cython.final
 * cdef class FrozenPrivacyDict(PrivacyDict):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of PrivacyDict that is automatically frozen
 */

struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivacyDict {
  struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict __pyx_base;
};
static struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivacyDict *__pyx_vtabptr_9pyprotect_9protected_FrozenPrivacyDict;


/* "Private_FrozenPrivate.pxi":3
 * 
 * # @cython.internal
 * cdef class Private(Wrapped):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Wrapped with following additional functionality:
 */

struct __pyx_vtabstruct_9pyprotect_9protected_Private {
  struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped __pyx_base;
  PyObject *(*private_visible)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *);
  PyObject *(*private_writeable)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *);
  PyObject *(*private_getattr)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *);
  PyObject *(*private_check_setattr)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *, PyObject *);
  PyObject *(*private_check_delattr)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *);
  PyObject *(*private_dir)(struct __pyx_obj_9pyprotect_9protected_Private *);
};
static struct __pyx_vtabstruct_9pyprotect_9protected_Private *__pyx_vtabptr_9pyprotect_9protected_Private;


/* "Private_FrozenPrivate.pxi":144
 * # @cython.internal
 * @cython.final
 * cdef class FrozenPrivate(Private):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Private that is automatically frozen
 */

struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivate {
  struct __pyx_vtabstruct_9pyprotect_9protected_Private __pyx_base;
};
static struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivate *__pyx_vtabptr_9pyprotect_9protected_FrozenPrivate;


/* "Protected_FrozenProtected.pxi":3
 * 
 * # @cython.internal
 * cdef class Protected(Private):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Private that further restriction of:
 */

struct __pyx_vtabstruct_9pyprotect_9protected_Protected {
  struct __pyx_vtabstruct_9pyprotect_9protected_Private __pyx_base;
  PyObject *(*process_rules)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *);
  PyObject *(*build_cache)(struct __pyx_obj_9pyprotect_9protected_Protected *);
  PyObject *(*check_1_op)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, PyObject *, struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op *__pyx_optional_args);
  PyObject *(*protected_visible)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_visible *__pyx_optional_args);
  PyObject *(*protected_writeable)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_writeable *__pyx_optional_args);
  PyObject *(*aclcheck)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, PyObject *);
  PyObject *(*protected_getattr)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *);
  PyObject *(*protected_check_setattr)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, PyObject *);
  PyObject *(*protected_check_delattr)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *);
  PyObject *(*protected_dir)(struct __pyx_obj_9pyprotect_9protected_Protected *);
};
static struct __pyx_vtabstruct_9pyprotect_9protected_Protected *__pyx_vtabptr_9pyprotect_9protected_Protected;


/* "Protected_FrozenProtected.pxi":279
 * 
 * # @cython.internal
 * cdef class FrozenProtected(Protected):             # <<<<<<<<<<<<<<
 *     '''
 *     Subclass of Protected that is automatically frozen
 */

struct __pyx_vtabstruct_9pyprotect_9protected_FrozenProtected {
  struct __pyx_vtabstruct_9pyprotect_9protected_Protected __pyx_base;
};
static struct __pyx_vtabstruct_9pyprotect_9protected_FrozenProtected *__pyx_vtabptr_9pyprotect_9protected_FrozenProtected;


/* "HiddenPartial.pxi":3
 * 
 * # @cython.internal
 * cdef class __HiddenPartial(object):             # <<<<<<<<<<<<<<
 *     '''
 *     This is an object wrapper / proxy for functools.partial
 */

struct __pyx_vtabstruct_9pyprotect_9protected___HiddenPartial {
  PyObject *(*wrapped_getattr)(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *, PyObject *);
};
static struct __pyx_vtabstruct_9pyprotect_9protected___HiddenPartial *__pyx_vtabptr_9pyprotect_9protected___HiddenPartial;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#if !CYTHON_USE_MODULE_STATE
static PyTypeObject *__pyx_CyFunctionType = 0;
#endif
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_IsCyOrPyCFunction(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   (Py_TYPE(obj)->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* RaiseClosureNameError.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_Occurred(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
static CYTHON_UNUSED int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
#if __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if __STDC_VERSION__ >= 201112L || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str);
#else
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);
#endif

/* #### Code section: module_declarations ### */
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_attr_hidden(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_attr); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_fif(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_o); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_freeze(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_multiwrapped(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_id_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_hash_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_isinstance_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_issubclass_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_instanceof_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_subclassof_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_help_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_help_str_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_visible(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_writeable(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_testop(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_get_rules(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_comparator(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_getattr(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_check_setattr(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, CYTHON_UNUSED PyObject *__pyx_v_val); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_check_delattr(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_dir(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_visible(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_writeable(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_visible(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_writeable(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_getattr(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_check_setattr(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_check_delattr(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_dir(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_get_rules(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_process_rules(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_rules); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_build_cache(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_check_1_op(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op, struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_visible(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_visible *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_writeable(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_writeable *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_aclcheck(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_getattr(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_check_setattr(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_check_delattr(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_dir(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_visible(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_9Protected_writeable(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/
static PyObject *__pyx_f_9pyprotect_9protected_15__HiddenPartial_wrapped_getattr(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a); /* proto*/

/* Module declarations from "cython" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "libc.string" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "libc.stdio" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "__builtin__" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "cpython.type" */
#if !CYTHON_USE_MODULE_STATE
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;
#endif

/* Module declarations from "cpython" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "cpython.object" */
#if !CYTHON_USE_MODULE_STATE
#endif

/* Module declarations from "pyprotect.protected" */
#if !CYTHON_USE_MODULE_STATE
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___ProtectionData = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_Proxy = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_Wrapped = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_Frozen = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_PrivacyDict = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_Private = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_FrozenPrivate = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_Protected = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected_FrozenProtected = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___HiddenPartial = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems = 0;
static PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op = 0;
#endif
static int __pyx_v_9pyprotect_9protected_PY2;
static PyObject *__pyx_v_9pyprotect_9protected_builtin_module = 0;
static PyObject *__pyx_v_9pyprotect_9protected_immutable_types_set = 0;
static PyObject *__pyx_v_9pyprotect_9protected_builtins_ids = 0;
static PyObject *__pyx_v_9pyprotect_9protected_builtin_module_immutable_attributes = 0;
static PyObject *__pyx_v_9pyprotect_9protected_PROT_ATTR_NAME = 0;
static PyObject *__pyx_v_9pyprotect_9protected_frozen_error = 0;
static PyObject *__pyx_v_9pyprotect_9protected_overridden_always = 0;
static PyObject *__pyx_v_9pyprotect_9protected_pickle_attributes = 0;
static PyObject *__pyx_v_9pyprotect_9protected_special_attributes = 0;
static PyObject *__pyx_v_9pyprotect_9protected_always_delegated = 0;
static PyObject *__pyx_v_9pyprotect_9protected_always_frozen = 0;
static PyObject *__pyx_v_9pyprotect_9protected_attr_identifier = 0;
static PyObject *__pyx_v_9pyprotect_9protected_ro_private_attr = 0;
static PyObject *__pyx_v_9pyprotect_9protected_unmangled_private_attr = 0;
static PyObject *__pyx_v_9pyprotect_9protected_mangled_private_attr_classname_regex = 0;
static PyObject *__pyx_v_9pyprotect_9protected_mangled_private_attr_regex_fmt = 0;
static PyObject *__pyx_v_9pyprotect_9protected_m_block = 0;
static PyObject *__pyx_v_9pyprotect_9protected_m_numeric = 0;
static PyObject *__pyx_v_9pyprotect_9protected_m_compare = 0;
static PyObject *__pyx_v_9pyprotect_9protected_m_safe = 0;
static PyObject *__pyx_f_9pyprotect_9protected_get_protected_attr_name(void); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected_get_builtin_obj(PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected_get_immutables(void); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected_protected_rules_from_kwargs(PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected_protected_merge_kwargs(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected_privatedict(PyObject *, PyObject *, struct __pyx_opt_args_9pyprotect_9protected_privatedict *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle___ProtectionData__set_state(struct __pyx_obj_9pyprotect_9protected___ProtectionData *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Proxy__set_state(struct __pyx_obj_9pyprotect_9protected_Proxy *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Wrapped__set_state(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Frozen__set_state(struct __pyx_obj_9pyprotect_9protected_Frozen *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_PrivacyDict__set_state(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivacyDict__set_state(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Private__set_state(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivate__set_state(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Protected__set_state(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenProtected__set_state(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *, PyObject *); /*proto*/
static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle___HiddenPartial__set_state(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *, PyObject *); /*proto*/
static PyObject *__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *)); /*proto*/
static PyObject *__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *)); /*proto*/
static PyObject *__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *)); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyprotect.protected"
extern int __pyx_module_is_main_pyprotect__protected;
int __pyx_module_is_main_pyprotect__protected = 0;

/* Implementation of "pyprotect.protected" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_help;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_object;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = "_____";
static const char __pyx_k_C[] = "C";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_o[] = "o";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "^%s$";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k__2[] = "_";
static const char __pyx_k__3[] = "";
static const char __pyx_k__4[] = "|";
static const char __pyx_k_cn[] = "cn";
static const char __pyx_k_eq[] = "__eq__";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_ge[] = "__ge__";
static const char __pyx_k_gt[] = "__gt__";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_le[] = "__le__";
static const char __pyx_k_lt[] = "__lt__";
static const char __pyx_k_ne[] = "__ne__";
static const char __pyx_k_op[] = "op";
static const char __pyx_k_or[] = "__or__";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_rl[] = "_rl";
static const char __pyx_k_ro[] = "ro";
static const char __pyx_k_rw[] = "rw";
static const char __pyx_k_tb[] = "tb";
static const char __pyx_k_0_1[] = "^__[^_].*?[^_][_]{0,1}$";
static const char __pyx_k_Set[] = "Set";
static const char __pyx_k__17[] = "\n";
static const char __pyx_k__33[] = ".";
static const char __pyx_k__44[] = "*";
static const char __pyx_k__88[] = "^_[^_].*?(?<!_)$";
static const char __pyx_k_abc[] = "abc";
static const char __pyx_k_abs[] = "__abs__";
static const char __pyx_k_add[] = "__add__";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_and[] = "__and__";
static const char __pyx_k_cmp[] = "__cmp__";
static const char __pyx_k_dir[] = "__dir__";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_ior[] = "__ior__";
static const char __pyx_k_kw1[] = "kw1";
static const char __pyx_k_kw2[] = "kw2";
static const char __pyx_k_len[] = "__len__";
static const char __pyx_k_mod[] = "__mod__";
static const char __pyx_k_mul[] = "__mul__";
static const char __pyx_k_neg[] = "__neg__";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_pos[] = "__pos__";
static const char __pyx_k_pow[] = "__pow__";
static const char __pyx_k_ret[] = "_ret";
static const char __pyx_k_ror[] = "__ror__";
static const char __pyx_k_set[] = "set";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_sub[] = "__sub__";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_x_2[] = "_x";
static const char __pyx_k_xor[] = "__xor__";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_PYPY[] = "PYPY";
static const char __pyx_k_PyPy[] = "PyPy";
static const char __pyx_k__182[] = "?";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_call[] = "__call__";
static const char __pyx_k_ceil[] = "ceil";
static const char __pyx_k_code[] = "__code__";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "dict";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_file[] = "__file__";
static const char __pyx_k_hash[] = "hash";
static const char __pyx_k_help[] = "help";
static const char __pyx_k_hide[] = "hide";
static const char __pyx_k_iadd[] = "__iadd__";
static const char __pyx_k_iand[] = "__iand__";
static const char __pyx_k_imod[] = "__imod__";
static const char __pyx_k_imul[] = "__imul__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_inst[] = "inst";
static const char __pyx_k_ipow[] = "__ipow__";
static const char __pyx_k_isub[] = "__isub__";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_ixor[] = "__ixor__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_list[] = "list";
static const char __pyx_k_long[] = "long";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_radd[] = "__radd__";
static const char __pyx_k_rand[] = "__rand__";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_rmod[] = "__rmod__";
static const char __pyx_k_rmul[] = "__rmul__";
static const char __pyx_k_rpow[] = "__rpow__";
static const char __pyx_k_rsub[] = "__rsub__";
static const char __pyx_k_rxor[] = "__rxor__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_wrap[] = "wrap";
static const char __pyx_k_Proxy[] = "Proxy";
static const char __pyx_k_add_2[] = "add";
static const char __pyx_k_aexit[] = "__aexit__";
static const char __pyx_k_alist[] = "alist";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_floor[] = "floor";
static const char __pyx_k_index[] = "__index__";
static const char __pyx_k_int_2[] = "__int__";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_major[] = "major";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_minor[] = "minor";
static const char __pyx_k_pydoc[] = "pydoc";
static const char __pyx_k_ret_2[] = "ret";
static const char __pyx_k_round[] = "__round__";
static const char __pyx_k_rules[] = "rules";
static const char __pyx_k_set_2[] = "__set__";
static const char __pyx_k_slots[] = "__slots__";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_str_2[] = "__str__";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_trunc[] = "__trunc__";
static const char __pyx_k_tuple[] = "tuple";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_union[] = "union";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_Frozen[] = "Frozen";
static const char __pyx_k_aenter[] = "__aenter__";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_bool_2[] = "__bool__";
static const char __pyx_k_ceil_2[] = "__ceil__";
static const char __pyx_k_delete[] = "__delete__";
static const char __pyx_k_dict_2[] = "__dict__";
static const char __pyx_k_dict_3[] = "_dict";
static const char __pyx_k_divmod[] = "__divmod__";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_format[] = "__format__";
static const char __pyx_k_freeze[] = "freeze";
static const char __pyx_k_frozen[] = "frozen";
static const char __pyx_k_hash_2[] = "__hash__";
static const char __pyx_k_id_val[] = "id_val";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_invert[] = "__invert__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_lshift[] = "__lshift__";
static const char __pyx_k_lstrip[] = "lstrip";
static const char __pyx_k_matmul[] = "__matmul__";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_rshift[] = "__rshift__";
static const char __pyx_k_rstrip[] = "rstrip";
static const char __pyx_k_testop[] = "testop";
static const char __pyx_k_unichr[] = "unichr";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_Mapping[] = "Mapping";
static const char __pyx_k_Private[] = "Private";
static const char __pyx_k_Wrapped[] = "Wrapped";
static const char __pyx_k_builtin[] = "__builtin__";
static const char __pyx_k_bytes_2[] = "__bytes__";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_complex[] = "complex";
static const char __pyx_k_delattr[] = "__delattr__";
static const char __pyx_k_delitem[] = "__delitem__";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_discard[] = "discard";
static const char __pyx_k_dynamic[] = "dynamic";
static const char __pyx_k_environ[] = "environ";
static const char __pyx_k_float_2[] = "__float__";
static const char __pyx_k_floor_2[] = "__floor__";
static const char __pyx_k_getitem[] = "__getitem__";
static const char __pyx_k_getsate[] = "__getsate__";
static const char __pyx_k_ilshift[] = "__ilshift__";
static const char __pyx_k_imatmul[] = "__imatmul__";
static const char __pyx_k_irshift[] = "__irshift__";
static const char __pyx_k_modules[] = "modules";
static const char __pyx_k_package[] = "__package__";
static const char __pyx_k_partial[] = "partial";
static const char __pyx_k_pattern[] = "pattern";
static const char __pyx_k_popitem[] = "popitem";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_private[] = "private";
static const char __pyx_k_protect[] = "protect";
static const char __pyx_k_rdivmod[] = "__rdivmod__";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_rlshift[] = "__rlshift__";
static const char __pyx_k_rmatmul[] = "__rmatmul__";
static const char __pyx_k_ro_data[] = "ro_data";
static const char __pyx_k_rrshift[] = "__rrshift__";
static const char __pyx_k_setattr[] = "__setattr__";
static const char __pyx_k_setitem[] = "__setitem__";
static const char __pyx_k_truediv[] = "__truediv__";
static const char __pyx_k_unicode[] = "unicode";
static const char __pyx_k_weakref[] = "__weakref__";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_Sequence[] = "Sequence";
static const char __pyx_k_builtins[] = "builtins";
static const char __pyx_k_contains[] = "__contains__";
static const char __pyx_k_defaults[] = "__defaults__";
static const char __pyx_k_endswith[] = "endswith";
static const char __pyx_k_exc_type[] = "exc_type";
static const char __pyx_k_floordiv[] = "__floordiv__";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_hash_val[] = "hash_val";
static const char __pyx_k_help_str[] = "help_str";
static const char __pyx_k_help_val[] = "help_val";
static const char __pyx_k_id_class[] = "id_class";
static const char __pyx_k_isfrozen[] = "isfrozen";
static const char __pyx_k_iterkeys[] = "iterkeys";
static const char __pyx_k_itruediv[] = "__itruediv__";
static const char __pyx_k_keys_py2[] = "keys_py2";
static const char __pyx_k_platform[] = "platform";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_ro_regex[] = "ro_regex";
static const char __pyx_k_rtruediv[] = "__rtruediv__";
static const char __pyx_k_rw_regex[] = "rw_regex";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_subclass[] = "subclass";
static const char __pyx_k_viewkeys[] = "viewkeys";
static const char __pyx_k_FrameType[] = "FrameType";
static const char __pyx_k_Protected[] = "_Protected_____";
static const char __pyx_k_Proxy_add[] = "Proxy.add";
static const char __pyx_k_Proxy_pop[] = "Proxy.pop";
static const char __pyx_k_Proxy_pxi[] = "Proxy.pxi";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_bytearray[] = "bytearray";
static const char __pyx_k_complex_2[] = "__complex__";
static const char __pyx_k_exc_value[] = "exc_value";
static const char __pyx_k_frozenset[] = "frozenset";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_ifloordiv[] = "__ifloordiv__";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_isprivate[] = "isprivate";
static const char __pyx_k_isvisible[] = "isvisible";
static const char __pyx_k_iswrapped[] = "iswrapped";
static const char __pyx_k_items_py2[] = "items_py2";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_rfloordiv[] = "__rfloordiv__";
static const char __pyx_k_ro_method[] = "ro_method";
static const char __pyx_k_viewitems[] = "viewitems";
static const char __pyx_k_ModuleType[] = "ModuleType";
static const char __pyx_k_MutableSet[] = "MutableSet";
static const char __pyx_k_Proxy_send[] = "Proxy.send";
static const char __pyx_k_Proxy_sort[] = "Proxy.sort";
static const char __pyx_k_basestring[] = "basestring";
static const char __pyx_k_contains_2[] = "contains";
static const char __pyx_k_difference[] = "difference";
static const char __pyx_k_hide_regex[] = "hide_regex";
static const char __pyx_k_instanceof[] = "instanceof";
static const char __pyx_k_isinstance[] = "isinstance";
static const char __pyx_k_isreadonly[] = "isreadonly";
static const char __pyx_k_issubclass[] = "issubclass";
static const char __pyx_k_itervalues[] = "itervalues";
static const char __pyx_k_match_args[] = "__match_args__";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_render_doc[] = "render_doc";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_splitlines[] = "splitlines";
static const char __pyx_k_startswith[] = "startswith";
static const char __pyx_k_subclassof[] = "subclassof";
static const char __pyx_k_values_py2[] = "values_py2";
static const char __pyx_k_viewvalues[] = "viewvalues";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_PrivacyDict[] = "PrivacyDict";
static const char __pyx_k_Protected_2[] = "Protected";
static const char __pyx_k_Proxy__exit[] = "_Proxy__exit";
static const char __pyx_k_Proxy_clear[] = "Proxy.clear";
static const char __pyx_k_Proxy_throw[] = "Proxy.throw";
static const char __pyx_k_build_regex[] = "_build_regex";
static const char __pyx_k_cfunc_to_py[] = "cfunc.to_py";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_isimmutable[] = "isimmutable";
static const char __pyx_k_isprotected[] = "isprotected";
static const char __pyx_k_length_hint[] = "__length_hint__";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_want_frozen[] = "want_frozen";
static const char __pyx_k_Proxy___ceil[] = "Proxy.__ceil__";
static const char __pyx_k_Proxy___exit[] = "Proxy.__exit__";
static const char __pyx_k_Proxy___iter[] = "Proxy.__iter__";
static const char __pyx_k_Proxy__aexit[] = "_Proxy__aexit";
static const char __pyx_k_Proxy_append[] = "Proxy.append";
static const char __pyx_k_Proxy_extend[] = "Proxy.extend";
static const char __pyx_k_Proxy_insert[] = "Proxy.insert";
static const char __pyx_k_Proxy_remove[] = "Proxy.remove";
static const char __pyx_k_Proxy_update[] = "Proxy.update";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_getattribute[] = "__getattribute__";
static const char __pyx_k_hide_private[] = "hide_private";
static const char __pyx_k_id_protected[] = "id_protected";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_intersection[] = "intersection";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_multiwrapped[] = "multiwrapped";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_version_info[] = "version_info";
static const char __pyx_k_FrozenPrivate[] = "FrozenPrivate";
static const char __pyx_k_HiddenPartial[] = "__HiddenPartial";
static const char __pyx_k_Private___dir[] = "Private.__dir__";
static const char __pyx_k_Proxy___aexit[] = "Proxy.__aexit__";
static const char __pyx_k_Proxy___bytes[] = "Proxy.__bytes__";
static const char __pyx_k_Proxy___enter[] = "Proxy.__enter__";
static const char __pyx_k_Proxy___floor[] = "Proxy.__floor__";
static const char __pyx_k_Proxy___round[] = "Proxy.__round__";
static const char __pyx_k_Proxy___trunc[] = "Proxy.__trunc__";
static const char __pyx_k_Proxy_discard[] = "Proxy.discard";
static const char __pyx_k_Proxy_popitem[] = "Proxy.popitem";
static const char __pyx_k_Proxy_reverse[] = "Proxy.reverse";
static const char __pyx_k_Wrapped___dir[] = "Wrapped.__dir__";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_instancecheck[] = "__instancecheck__";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_s_________0_1[] = "^_%s__[^_](.*?[^_]|)[_]{0,1}$";
static const char __pyx_k_subclasscheck[] = "__subclasscheck__";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_CollectionsABC[] = "CollectionsABC";
static const char __pyx_k_Double_wrapped[] = "Double-wrapped!";
static const char __pyx_k_MutableMapping[] = "MutableMapping";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_ProtectionData[] = "__ProtectionData";
static const char __pyx_k_Proxy___aenter[] = "Proxy.__aenter__";
static const char __pyx_k_Proxy___format[] = "Proxy.__format__";
static const char __pyx_k_RecursionError[] = "RecursionError";
static const char __pyx_k_hash_protected[] = "hash_protected";
static const char __pyx_k_help_protected[] = "help_protected";
static const char __pyx_k_isinstance_val[] = "isinstance_val";
static const char __pyx_k_issubclass_val[] = "issubclass_val";
static const char __pyx_k_oldstyle_class[] = "oldstyle_class";
static const char __pyx_k_FrozenProtected[] = "FrozenProtected";
static const char __pyx_k_MutableSequence[] = "MutableSequence";
static const char __pyx_k_Protected___dir[] = "Protected.__dir__";
static const char __pyx_k_ProtectionError[] = "ProtectionError";
static const char __pyx_k_Proxy___complex[] = "Proxy.__complex__";
static const char __pyx_k_attr_type_check[] = "attr_type_check";
static const char __pyx_k_collections_abc[] = "collections.abc";
static const char __pyx_k_init___locals_C[] = "__init__.<locals>.C";
static const char __pyx_k_never_writeable[] = "never_writeable";
static const char __pyx_k_pass_to_wrapped[] = "pass_to_wrapped";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_PrivacyDict_copy[] = "PrivacyDict.copy";
static const char __pyx_k_PrivacyDict_keys[] = "PrivacyDict.keys";
static const char __pyx_k_Proxy_setdefault[] = "Proxy.setdefault";
static const char __pyx_k_a_zA_Z_a_zA_Z0_9[] = "^[_a-zA-Z][a-zA-Z0-9_]*$";
static const char __pyx_k_o_Invalid_type_s[] = "o: Invalid type: %s";
static const char __pyx_k_FrozenPrivacyDict[] = "FrozenPrivacyDict";
static const char __pyx_k_HiddenPartial_pxi[] = "HiddenPartial.pxi";
static const char __pyx_k_PrivacyDict_items[] = "PrivacyDict.items";
static const char __pyx_k_PrivacyDict_values[] = "PrivacyDict.values";
static const char __pyx_k_ProtectionData_pxi[] = "ProtectionData.pxi";
static const char __pyx_k_Proxy___match_args[] = "Proxy.__match_args__";
static const char __pyx_k_Wrapped_Frozen_pxi[] = "Wrapped_Frozen.pxi";
static const char __pyx_k_a_zA_Z_a_zA_Z0_9_2[] = "[a-zA-Z][a-zA-Z0-9]*";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_python_visible_pxi[] = "python_visible.pxi";
static const char __pyx_k_pyx_unpickle_Proxy[] = "__pyx_unpickle_Proxy";
static const char __pyx_k_HiddenPartial___dir[] = "__HiddenPartial.__dir__";
static const char __pyx_k_Object_is_read_only[] = "Object is read-only";
static const char __pyx_k_Proxy___length_hint[] = "Proxy.__length_hint__";
static const char __pyx_k_attribute_protected[] = "attribute_protected";
static const char __pyx_k_pyprotect_protected[] = "pyprotect.protected";
static const char __pyx_k_pyx_unpickle_Frozen[] = "__pyx_unpickle_Frozen";
static const char __pyx_k_PrivacyDict_iterkeys[] = "PrivacyDict.iterkeys";
static const char __pyx_k_PrivacyDict_keys_py2[] = "PrivacyDict.keys_py2";
static const char __pyx_k_PrivacyDict_viewkeys[] = "PrivacyDict.viewkeys";
static const char __pyx_k_ProtectionData___dir[] = "__ProtectionData.__dir__";
static const char __pyx_k_isinstance_protected[] = "isinstance_protected";
static const char __pyx_k_issubclass_protected[] = "issubclass_protected";
static const char __pyx_k_pyx_unpickle_Private[] = "__pyx_unpickle_Private";
static const char __pyx_k_pyx_unpickle_Wrapped[] = "__pyx_unpickle_Wrapped";
static const char __pyx_k_same_class_protected[] = "same_class_protected";
static const char __pyx_k_HiddenPartial___bytes[] = "__HiddenPartial.__bytes__";
static const char __pyx_k_PrivacyDict_items_py2[] = "PrivacyDict.items_py2";
static const char __pyx_k_PrivacyDict_iteritems[] = "PrivacyDict.iteritems";
static const char __pyx_k_PrivacyDict_viewitems[] = "PrivacyDict.viewitems";
static const char __pyx_k_Proxy___instancecheck[] = "Proxy.__instancecheck__";
static const char __pyx_k_Proxy___reduce_cython[] = "Proxy.__reduce_cython__";
static const char __pyx_k_Proxy___subclasscheck[] = "Proxy.__subclasscheck__";
static const char __pyx_k_Read_only_attribute_s[] = "Read only attribute: %s";
static const char __pyx_k_instance_of_protected[] = "instance_of_protected";
static const char __pyx_k_py2_function_attrs_rw[] = "py2_function_attrs_rw";
static const char __pyx_k_python_implementation[] = "python_implementation";
static const char __pyx_k_subclass_of_protected[] = "subclass_of_protected";
static const char __pyx_k_Frozen___reduce_cython[] = "Frozen.__reduce_cython__";
static const char __pyx_k_PrivacyDict_itervalues[] = "PrivacyDict.itervalues";
static const char __pyx_k_PrivacyDict_values_py2[] = "PrivacyDict.values_py2";
static const char __pyx_k_PrivacyDict_viewvalues[] = "PrivacyDict.viewvalues";
static const char __pyx_k_Unknown_OldStyle_Class[] = "Unknown_OldStyle_Class";
static const char __pyx_k_global_c_functions_pxi[] = "global_c_functions.pxi";
static const char __pyx_k_pyx_unpickle_Protected[] = "__pyx_unpickle_Protected";
static const char __pyx_k_Private___reduce_cython[] = "Private.__reduce_cython__";
static const char __pyx_k_Proxy___setstate_cython[] = "Proxy.__setstate_cython__";
static const char __pyx_k_Wrapped___reduce_cython[] = "Wrapped.__reduce_cython__";
static const char __pyx_k_never_writeable_private[] = "never_writeable_private";
static const char __pyx_k_Cannot_add_attribute_s_s[] = "Cannot add attribute: %s.%s";
static const char __pyx_k_Cannot_set_attribute_s_s[] = "Cannot set attribute: %s.%s";
static const char __pyx_k_Frozen___setstate_cython[] = "Frozen.__setstate_cython__";
static const char __pyx_k_hidden_pickle_attributes[] = "hidden_pickle_attributes";
static const char __pyx_k_pyx_unpickle_PrivacyDict[] = "__pyx_unpickle_PrivacyDict";
static const char __pyx_k_Cannot_delete_attribute_s[] = "Cannot delete attribute: %s";
static const char __pyx_k_Cannot_modify_attribute_s[] = "Cannot modify attribute: %s";
static const char __pyx_k_Private_FrozenPrivate_pxi[] = "Private_FrozenPrivate.pxi";
static const char __pyx_k_Private___setstate_cython[] = "Private.__setstate_cython__";
static const char __pyx_k_Protected___reduce_cython[] = "Protected.__reduce_cython__";
static const char __pyx_k_Wrapped___setstate_cython[] = "Wrapped.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_FrozenPrivate[] = "__pyx_unpickle_FrozenPrivate";
static const char __pyx_k_Cannot_delete_attribute_s_s[] = "Cannot delete attribute: %s.%s";
static const char __pyx_k_Object_s_has_no_attribute_s[] = "Object '%s' has no attribute '%s'";
static const char __pyx_k_PrivacyDict___reduce_cython[] = "PrivacyDict.__reduce_cython__";
static const char __pyx_k_Protected___setstate_cython[] = "Protected.__setstate_cython__";
static const char __pyx_k_always_delegated_attributes[] = "always_delegated_attributes";
static const char __pyx_k_immutable_builtin_attributes[] = "immutable_builtin_attributes";
static const char __pyx_k_pyx_unpickle_FrozenProtected[] = "__pyx_unpickle_FrozenProtected";
static const char __pyx_k_pyx_unpickle___HiddenPartial[] = "__pyx_unpickle___HiddenPartial";
static const char __pyx_k_FrozenPrivate___reduce_cython[] = "FrozenPrivate.__reduce_cython__";
static const char __pyx_k_HiddenPartial___reduce_cython[] = "__HiddenPartial.__reduce_cython__";
static const char __pyx_k_PrivacyDict___setstate_cython[] = "PrivacyDict.__setstate_cython__";
static const char __pyx_k_Protected_FrozenProtected_pxi[] = "Protected_FrozenProtected.pxi";
static const char __pyx_k_pyx_unpickle___ProtectionData[] = "__pyx_unpickle___ProtectionData";
static const char __pyx_k_HiddenPartial___setstate_cytho[] = "__HiddenPartial.__setstate_cython__";
static const char __pyx_k_Module_with_methods_to_wrap_an[] = "\nModule with methods to wrap an object and additionally restrict\nvisibility and mutability of attributes\n\nVISIBILITY or READABILITY: Whether the attribute VALUE can be read\n\n- Objects wrapped with private / protect do not allow following\n  special methods to be set or deleted:\n    __getattribute__\n    __setattr__\n    __delattr__\n\nMUTABILITY or WRITEABILITY: Ability to CHANGE or DELETE an attribute\n\n- Protected object will not allow CHANGING OR DELETING an attribute\n  that is not VISIBLE\n- Objects wrapped with private / protect do not allow modification\n  of __class__, __dict__ or __slots attributes\n- When using protect(o, **kwargs), writeability depends on kwargs\n\nClasses\n=======\n\nThese classes are not directly exported by the module so as to not\nclutter the pydoc documentation for the module.\n\n                                 Proxy\n                                   \342\224\202\n                                   \342\224\202\n                                Wrapped\n                                   \342\224\202\n                                   \342\224\202\n    \342\224\214\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\264\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\220\n    \342\224\202                                          \342\224\202\n    Frozen                                  Private\n                                               \342\224\202\n                                               \342\224\202\n         \342\224\214\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342""\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\254\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\264\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\220\n         \342\224\202                        \342\224\202                            \342\224\202\n    PrivacyDict                   \342\224\202                        Protected\n         \342\224\202                        \342\224\202                            \342\224\202\n         \342\224\202                        \342\224\202                            \342\224\202\n    FrozenPrivacyDict         FrozenPrivate            FrozenProtected\n\n\n    Wrapped:\n        - Visibility: No restrictions\n        - Mutability: No restrictions\n\n    Frozen: subclass of Wrapped\n        - Visibility: No restrictions\n        - Mutability: NO ATTRIBUTES can be changed or deleted\n\n    Private: subclass of Wrapped\n        - Visibility:\n            - Cannot access traditionally 'private' mangled python attributes\n            - Cannot access any unmangled double '_' attributes\n            - Cannot access any attribute not exported by dir(o)\n        - Mutability:\n            - Cannot modify traditionally private attributes (form '_var')\n            - Cannot modify __class__ of wrapped object\n            - Cannot modify __dict__ of wrapped object\n            - Cannot modify __slots__ of wrapped object\n            - Cannot add or delete attributes\n\n    FrozenPrivate: subclass of Private\n        - Created by calling private(o, frozen=True) on an object 'o'\n        - Also created by calling freeze(private(o, froze""n=False))\n          on an object 'o'\n        - Features of Private PLUS prevents modification of ANY attribute\n        - Visibility: Same as Private\n        - Mutability: NO ATTRIBUTES can be changed or deleted\n\n    Protected: subclass of Private\n        - Created by calling protect(o, frozen=False) on an object 'o'\n        - Features of Private PLUS additional restrictions on:\n            - ADDITIONAL attributes that are NOT visible\n            - ADDITIONAL attributes that are NOT writeable\n\n    FrozenProtected: subclass of Protected\n        - Created by calling protect(o, frozen=True) on an object 'o'\n        - Also created by calling freeze(protect(o, frozen=False))\n          on an object 'o'\n        - Features of Protected PLUS prevents modification of ANY attribute\n        - Mutability: NO ATTRIBUTES can be changed or deleted\n\n    PrivacyDict: subclass of Private\n        - Not created directly\n\n    FrozenPrivacyDict: subclass of Private\n        - Created internally when accessing 'dict' attribute of a\n          Private object\n\nKey methods in the module API:\n=============================\n\nwrap(o: object) -> Wrapped:\n\nfreeze(o: object) -> object:\n    - If 'o' is immutable (e.g. int , string), returns 'o' UNCHANGED\n    - If 'o' is Wrapped, returns 'o' UNCHANGED if object WRAPPPED INSIDE\n      'o' is immutable, returns Frozen otherwise\n    - If 'o' is Frozen, returns 'o UNCHANGED\n    - If 'o' is FrozenPrivate, FrozenProtected or FrozenPrivacyDict,\n      returns 'o' UNCHANGED\n    - If 'o' is Private, returns FrozenPrivate\n    - If 'o' is Protected, returns FrozenProtected\n    - Otherwise, returns Frozen\n\n    Object returned prevents modification of ANY attribute\n\nprivate(o: object, frozen: bool = False) -> object:\n    - If 'frozen' is False:\n        - If 'o' is an instance of Private, returns 'o' UNCHANGED\n        - If 'o' is an instance of Protected, returns 'o' UNCHANGED\n    - If 'frozen' is True:\n        - If 'o' i""s an instance of Private, returns freeze(o) --> FrozenPrivate\n        - If 'o' is an instance of Protected, returns freeze(o) --> FrozenProtected\n    - Otherwise:\n        If frozen is True, returns FrozenPrivate; returns Private otherwise\n\nprotect(\n    o: object,\n    frozen: bool = False, dynamic: bool = True,\n    hide_private: bool = False,\n    ro_data: bool = False, ro_method: bool = True,\n    ro=[], rw=[], hide=[],\n):\n    o: object to be wrapped\n    frozen: bool: No attribute can be modified\n        PLUS: if 'o' is NOT a module, results returned by methods,\n        including __call__ will be frozen\n    dynamic: bool: Attribute additions, deletions, type changes in wrapped\n        object are automatically considered by hide_private, ro_data,\n        ro_method, ro, rw, hide\n        If dynamic is False, it is a pledge that attributes of wrapped\n        object will not change, and visibility and mutability rules of\n        WRAPPING object use a cache to make them faster.\n        Rules imposed by Private() are always dynamic\n    hide_private: bool: Private vars (_var) will be hidden\n    ro_data: bool: Data attributes cannot be deleted or assigned to\n    ro_method: bool: Method attributes cannot be deleted or assigned to\n    ro: list of str: attributes that will be read-only\n    rw: list of str: attributes that will be read-write\n        Overrides 'ro_*'\n    hide: list of str: attributes that will be hidden\n\n    Returns-->Instance of FrozenProtected if frozen; Protected otherwise\n\n    Default settings:\n    Features of Private:\n    PLUS:\n        - Methods are readonly - cannot be deleted or assigned to\n\n    If protect() is called on an object 'o' that is an instance of\n    Protected:\n        protect() will merge the protect() rules, enforcing the most restrictive\n        combination among the two sets of protect() options:\n         - 'hide' and 'hide_private' are OR-ed\n         - 'ro_method', 'ro_data' and 'ro' are OR-ed\n     ""    - 'rw' is AND-ed, but 'rw' of second protect overrides 'ro_*' of SECOND protect\n           but not the first protect.\n\n        In short, by calling protect() a second time (or multiple times):\n            - Additoinal attributes can be hidden\n            - Additional attributes can be made read-only\n        but:\n            - No previously hidden attribute will become visible\n            - No previously read-only attribute will become mutable\n\n\nCalling wrap operations multiple times\n======================================\n\nIn the table below, the left-most column shows starting state.\nThe top row shows operation applied to the starting state.\nThe intersecting cell shows the result.\n\n\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\244\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\nOperation  \360\237\241\206   \342\224\202 wrap        freeze      private     private     protect     protect\n\360\237\241\207  with        \342\224\202   ""                                  + frozen                + frozen\n\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\252\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\nWrapped        \342\224\202 UNCH        Frozen      Private     Frozen      Protected   FrozenProtected\n               \342\224\202 [2]         [2]                     Private\n\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\274\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224""\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\nFrozen         \342\224\202 Wrapped     UNCH        Frozen      Frozen      Frozen      Frozen\n               \342\224\202 [2]         [2]         Private     Private     Protected   Protected\n\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\274\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\nPrivate        \342\224\202 UNCH        Frozen      UNCH        Frozen      Protected   Frozen\n               \342\224\202             Private                 Priva""te                 Protected\n\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\274\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\nFrozenPrivate  \342\224\202 UNCH        UNCH        UNCH        UNCH        Frozen      FrozenProtected\n               \342\224\202                                                 Protected\n\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\274\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224""\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\nProtected      \342\224\202 UNCH        Frozen      UNCH        Frozen      Protected   FrozenProtected\n               \342\224\202             Protected               Protected   [1]         [1]\n\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\274\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\342\224\200\nFrozenProtected\342\224\202 UNCH        UNCH        UNCH        UNCH        Frozen      FrozenProtected\n               \342\224\202                                                 Prote""cted   [1]\n               \342\224\202                                                 [1]\n\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\247\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\342\225\220\n\n[1]: protect applied twice, will merge the protect() rules, enforcing the most restrictive\n     combination among the two sets of protect() options:\n     - 'hide' and 'hide_private' are OR-ed\n     - 'ro_method', 'ro_data' and 'ro' are OR-ed\n     - 'rw' is AND-ed, but 'rw' of second protect overrides 'ro_*' of SECOND protect\n       but not the first protect.\n\n    In short, by calling protect() a second time (or multiple times):\n        - Additoinal attributes can be hidden\n        - Additional attributes can be made read-only\n    but:\n        - No previously hidden attribute will become visible\n        - No previously read-only attribute will become mutable\n\n[2]: If 'x' is an immutable object (e.g. int, str ...) having isimmutable(x) is True,\n     fre""eze(x) returns x and iswrapped(freeze(x)) will be False.\n\n     For all other objects 'x', having isimmutable(x) == False, freeze(x) will return\n     a Frozen object having iswrapped(freeze(x)) == True\n\n    For all other wrapped objects 'w', created with private(x) or protect(x), freeze(w)\n    will always return a Wrapped object with iswrapped(w) == True\n\nChecking whether an object is wrapped:\n=====================================\n\niswrapped(w) -> bool: True IFF 'w' was was wrapped using\n    wrap(), freeze(), private() or protect()\n    See Note for output of freeze()\n\nisfrozen(w) -> bool: True IFF 'w' is an instance of Frozen,\nFrozenPrivate, ProzenPrivacyDict or FrozenProtected\n\nisprivate(w) -> bool: True IFF 'w' is an instance of Private,\nFrozenPrivate, Protected or FrozenProtected\n\nisprotected(w) -> bool: True IFF 'w' is an instance of Protected,\nFrozenProtected\n\n\nWhat kind of python objects can be wrapped?\n==========================================\n\n- Any object that supports getattr, setattr, delattr and __class__\n- Pickling / unpickling of wrapped objects is not supported\n    Even if / when enabled, after a pickle-unpickle cycle,\n    - Frozen objects will no longer be frozen\n    - Private objects will no longer have visibility / mutability\n      restrictions\n    - Protected objects will no longer have custom protections\n\nCan I wrap an object from a python C extension?\nYES. See answer to 'What kind of python objects can be wrapped?'\n\nWill wrapper detect attributes deleted, added or changed at RUN-TIME?\n====================================================================\nwrap / freeze / private: YES !\n\nprotect:\n    If 'dynamic' is True (default): YES !\n\n    If 'dynamic' is False, dir(wrapped_object) will not\n    accurately reflect attributes added or deleted at run-time\n\n    Note that the above caveats are UNAFFECTED by 'frozen'\n    'frozen' only controls whether object can be modified from OUTSIDE\n    the wrapped"" object\n\nWill I need to change the code for my object / class?\n====================================================\nONLY in the following cases fnd ONLY if wrapped using private / protect:\n\n- If your object DEPENDS on external visibility of traditionally\n  'private' mangled object attributes, you will need to change\n  the names of those attributes - this is a basic objective of\n  private / protect\n- If your object DEPENDS on external writeability of traditionally\n  'private' attributes of the form '_var', you will need to change\n  the names of those attributes - this is a basic objective of\n  private / protect\n- If your object DEPENDS on EXTERNAL modifability of __class__,\n  __dict__ or __slots__, you will need to change the behavior\n  of your object (change the code) - since this contradicts the\n  basic objective of private / protect.\n\nCode changes required when USING a wrapped object:\n=================================================\n\nPickling / unpickling of wrapped objects is not supported\n\nIf 'o' is your original object, and 'w' is the wrapped object:\nOne difference across wrap / freeze / private / protect:\ndir(w) will necessarily be different from dir(o):\n  Additional attributes in 'w': '_Protected_____'\n  'private':\n      Traditionally 'private' mangled attributes will not appear\n  'protect':\n      Traditionally 'private' mangled attributes will not appear\n      Further differences depending on keyword arguments to 'protect'\n\nFollowing applies only to wrapping with wrap / private / protect:\n- Change calls to w.__getattribute__(a) to getattr(w, a)\n- Change calls to w.__delattr__ to delattr(w, a)\n- Change calls to w.__setattr(a, val) to setattr(w, a, val)\n- Change isinstance(w, Mytypes) to isinstance_protected(w, MyTypes)\n    isinstance_protected can also be used transparently on objects\n    that have NOT been wrapped\n    Can also (even) alias isinstance to isinstance_protected\n- Change id(w) to id_protected(w). id_prot""ected can also be used\n    transparently on objects that have NOT been wrapped\n    Can also (even) alias id to id_protected\n- Change 'w is x' to id_protected(w) == id_protected(x)\n- Change type(w) to w.__class__ if you want to use the CLASS of w\n    but safely - not allowing class modifications\n- Getting interactive help on an object\n    Instead of help(o), use help_protected(o)\n    Can also (even) alias help to help_protected\n\nObject equality:\nTwo objects returned by wrap / freeze / private / protect are equal\nIF AND ONLY IF all the following conditions are met:\n- They wrap the SAME object - id(o1) == id(o2)\n- They were wrapped using the same method\n- For private: both were wrapped with the same value for 'frozen'\n- For protect: the EFFECTIVE visibility and writeability implied\n  by keyword arguments provided to 'protect' for the two objects\n  is identical\n\n\nChecking at run-time whether an attribute is visible:\n====================================================\n\nAssuming 'o' is the object, whether wrapped or not and 'a is attribute:\nJust use hasattr(o, a).  Works on any object, wrapped or not.\nCan also use isvisible(w, a) if 'w' is a wrapped object and 'a' is an attribute.\n'isvisible' return value (ONLY) represents whether type of wrapping imposes\nspecific visibility rules (i.e. hides visibility). \n\nChecking at run-time whether an attribute is writeable:\n======================================================\n\nAssuming 'o' is the object, whether wrapped or not and you want to set\nattribute 'a' to value 'val':\nCan use isreadonly(w, a) if 'w' is a wrapped object and 'a' is an attribute.\n'isreadonly' return value (ONLY) represents whether type of wrapping imposes\nspecific mutability rules (i.e. limits mutabiity).\n\nChecking at run-time whether an attribute can be deleted:\n========================================================\n\nAssuming 'o' is the object, whether wrapped or not and you want to delete\nattribute 'a':\nCan use ""isreadonly(w, a) if 'w' is a wrapped object and 'a' is an attribute.\n'isreadonly' return value (ONLY) represents whether type of wrapping imposes\nspecific mutability rules (i.e. limits mutabiity).\n\n\nViewing help for the classes:\n============================\nYou can see the help for each of the classes below - EXCEPT\nPrivacyDict as follows:\n\n    Wrapped         : help(type(wrap(None)))\n    Frozen          : help(type(freeze([])))\n    Private         : help(type(private(None)))\n    Protected       : help(type(protect(None)))\n    FrozenPrivate   : help(type(private(None, frozen=True)))\n    FrozenProtected : help(type(protect(None, frozen=True)))\n\nTo see help for FrozenPrivacyDict:\n    class C(object):\n        pass\n\n    help(type(private(C()).__dict__))\n\nProxy and PrivacyDict are not exposed directly.\n";
static const char __pyx_k_ProtectionData___reduce_cython[] = "__ProtectionData.__reduce_cython__";
static const char __pyx_k_ProtectionData___setstate_cyth[] = "__ProtectionData.__setstate_cython__";
static const char __pyx_k_Pyx_CFunc_object__lParenWrappe[] = "__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self.<locals>.wrap";
static const char __pyx_k_pyx_unpickle_FrozenPrivacyDict[] = "__pyx_unpickle_FrozenPrivacyDict";
static const char __pyx_k_Cannot_delete_private_attribute[] = "Cannot delete private attribute: %s.%s";
static const char __pyx_k_FrozenPrivate___setstate_cython[] = "FrozenPrivate.__setstate_cython__";
static const char __pyx_k_FrozenProtected___reduce_cython[] = "FrozenProtected.__reduce_cython__";
static const char __pyx_k_Cannot_set_private_attribute_s_s[] = "Cannot set private attribute: %s.%s";
static const char __pyx_k_FrozenPrivacyDict___reduce_cytho[] = "FrozenPrivacyDict.__reduce_cython__";
static const char __pyx_k_FrozenPrivacyDict___setstate_cyt[] = "FrozenPrivacyDict.__setstate_cython__";
static const char __pyx_k_FrozenProtected___setstate_cytho[] = "FrozenProtected.__setstate_cython__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))";
static const char __pyx_k_Object_Private_s_has_no_attribut[] = "Object Private('%s') has no attribute '%s'";
static const char __pyx_k_Object_Protected_s_has_no_attrib[] = "Object Protected('%s') has no attribute '%s'";
static const char __pyx_k_Object_Wrapped_s_has_no_attribut[] = "Object Wrapped('%s') has no attribute '%s'";
static const char __pyx_k_Object___HiddenPartial_has_no_at[] = "Object __HiddenPartial has no attribute '%s'";
static const char __pyx_k_PrivacyDict_FrozenPrivacyDict_px[] = "PrivacyDict_FrozenPrivacyDict.pxi";
static const char __pyx_k_Pyx_CFunc_object__lParenWrappe_2[] = "__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_object__lParenWrappe_3[] = "__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op.<locals>.wrap";
static const char __pyx_k_Wrapped_comparator_locals_pass_t[] = "Wrapped.comparator.<locals>.pass_to_wrapped";
static const char __pyx_k_Wrapped_object_cannot_be_pickled[] = "Wrapped object cannot be pickled";
static const char __pyx_k_protected_rules_from_kwargs_loca[] = "protected_rules_from_kwargs.<locals>._build_regex";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_5[] = "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))";
#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_n_s_;
static PyObject *__pyx_kp_s_0_1;
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_kp_s_Cannot_add_attribute_s_s;
static PyObject *__pyx_kp_s_Cannot_delete_attribute_s;
static PyObject *__pyx_kp_s_Cannot_delete_attribute_s_s;
static PyObject *__pyx_kp_s_Cannot_delete_private_attribute;
static PyObject *__pyx_kp_s_Cannot_modify_attribute_s;
static PyObject *__pyx_kp_s_Cannot_set_attribute_s_s;
static PyObject *__pyx_kp_s_Cannot_set_private_attribute_s_s;
static PyObject *__pyx_n_s_CollectionsABC;
static PyObject *__pyx_kp_s_Double_wrapped;
static PyObject *__pyx_n_s_FrameType;
static PyObject *__pyx_n_s_Frozen;
static PyObject *__pyx_n_s_FrozenPrivacyDict;
static PyObject *__pyx_n_s_FrozenPrivacyDict___reduce_cytho;
static PyObject *__pyx_n_s_FrozenPrivacyDict___setstate_cyt;
static PyObject *__pyx_n_s_FrozenPrivate;
static PyObject *__pyx_n_s_FrozenPrivate___reduce_cython;
static PyObject *__pyx_n_s_FrozenPrivate___setstate_cython;
static PyObject *__pyx_n_s_FrozenProtected;
static PyObject *__pyx_n_s_FrozenProtected___reduce_cython;
static PyObject *__pyx_n_s_FrozenProtected___setstate_cytho;
static PyObject *__pyx_n_s_Frozen___reduce_cython;
static PyObject *__pyx_n_s_Frozen___setstate_cython;
static PyObject *__pyx_n_s_HiddenPartial;
static PyObject *__pyx_n_s_HiddenPartial___bytes;
static PyObject *__pyx_n_s_HiddenPartial___dir;
static PyObject *__pyx_n_s_HiddenPartial___reduce_cython;
static PyObject *__pyx_n_s_HiddenPartial___setstate_cytho;
static PyObject *__pyx_kp_s_HiddenPartial_pxi;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_Mapping;
static PyObject *__pyx_n_s_ModuleType;
static PyObject *__pyx_n_s_MutableMapping;
static PyObject *__pyx_n_s_MutableSequence;
static PyObject *__pyx_n_s_MutableSet;
static PyObject *__pyx_n_s_None;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_kp_s_Object_Private_s_has_no_attribut;
static PyObject *__pyx_kp_s_Object_Protected_s_has_no_attrib;
static PyObject *__pyx_kp_s_Object_Wrapped_s_has_no_attribut;
static PyObject *__pyx_kp_s_Object___HiddenPartial_has_no_at;
static PyObject *__pyx_kp_s_Object_is_read_only;
static PyObject *__pyx_kp_s_Object_s_has_no_attribute_s;
static PyObject *__pyx_n_s_PYPY;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_PrivacyDict;
static PyObject *__pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px;
static PyObject *__pyx_n_s_PrivacyDict___reduce_cython;
static PyObject *__pyx_n_s_PrivacyDict___setstate_cython;
static PyObject *__pyx_n_s_PrivacyDict_copy;
static PyObject *__pyx_n_s_PrivacyDict_items;
static PyObject *__pyx_n_s_PrivacyDict_items_py2;
static PyObject *__pyx_n_s_PrivacyDict_iteritems;
static PyObject *__pyx_n_s_PrivacyDict_iterkeys;
static PyObject *__pyx_n_s_PrivacyDict_itervalues;
static PyObject *__pyx_n_s_PrivacyDict_keys;
static PyObject *__pyx_n_s_PrivacyDict_keys_py2;
static PyObject *__pyx_n_s_PrivacyDict_values;
static PyObject *__pyx_n_s_PrivacyDict_values_py2;
static PyObject *__pyx_n_s_PrivacyDict_viewitems;
static PyObject *__pyx_n_s_PrivacyDict_viewkeys;
static PyObject *__pyx_n_s_PrivacyDict_viewvalues;
static PyObject *__pyx_n_s_Private;
static PyObject *__pyx_kp_s_Private_FrozenPrivate_pxi;
static PyObject *__pyx_n_s_Private___dir;
static PyObject *__pyx_n_s_Private___reduce_cython;
static PyObject *__pyx_n_s_Private___setstate_cython;
static PyObject *__pyx_n_s_Protected;
static PyObject *__pyx_n_s_Protected_2;
static PyObject *__pyx_kp_s_Protected_FrozenProtected_pxi;
static PyObject *__pyx_n_s_Protected___dir;
static PyObject *__pyx_n_s_Protected___reduce_cython;
static PyObject *__pyx_n_s_Protected___setstate_cython;
static PyObject *__pyx_n_s_ProtectionData;
static PyObject *__pyx_n_s_ProtectionData___dir;
static PyObject *__pyx_n_s_ProtectionData___reduce_cython;
static PyObject *__pyx_n_s_ProtectionData___setstate_cyth;
static PyObject *__pyx_kp_s_ProtectionData_pxi;
static PyObject *__pyx_n_s_ProtectionError;
static PyObject *__pyx_n_s_Proxy;
static PyObject *__pyx_n_s_Proxy___aenter;
static PyObject *__pyx_n_s_Proxy___aexit;
static PyObject *__pyx_n_s_Proxy___bytes;
static PyObject *__pyx_n_s_Proxy___ceil;
static PyObject *__pyx_n_s_Proxy___complex;
static PyObject *__pyx_n_s_Proxy___enter;
static PyObject *__pyx_n_s_Proxy___exit;
static PyObject *__pyx_n_s_Proxy___floor;
static PyObject *__pyx_n_s_Proxy___format;
static PyObject *__pyx_n_s_Proxy___instancecheck;
static PyObject *__pyx_n_s_Proxy___iter;
static PyObject *__pyx_n_s_Proxy___length_hint;
static PyObject *__pyx_n_s_Proxy___match_args;
static PyObject *__pyx_n_s_Proxy___reduce_cython;
static PyObject *__pyx_n_s_Proxy___round;
static PyObject *__pyx_n_s_Proxy___setstate_cython;
static PyObject *__pyx_n_s_Proxy___subclasscheck;
static PyObject *__pyx_n_s_Proxy___trunc;
static PyObject *__pyx_n_s_Proxy__aexit;
static PyObject *__pyx_n_s_Proxy__exit;
static PyObject *__pyx_n_s_Proxy_add;
static PyObject *__pyx_n_s_Proxy_append;
static PyObject *__pyx_n_s_Proxy_clear;
static PyObject *__pyx_n_s_Proxy_discard;
static PyObject *__pyx_n_s_Proxy_extend;
static PyObject *__pyx_n_s_Proxy_insert;
static PyObject *__pyx_n_s_Proxy_pop;
static PyObject *__pyx_n_s_Proxy_popitem;
static PyObject *__pyx_kp_s_Proxy_pxi;
static PyObject *__pyx_n_s_Proxy_remove;
static PyObject *__pyx_n_s_Proxy_reverse;
static PyObject *__pyx_n_s_Proxy_send;
static PyObject *__pyx_n_s_Proxy_setdefault;
static PyObject *__pyx_n_s_Proxy_sort;
static PyObject *__pyx_n_s_Proxy_throw;
static PyObject *__pyx_n_s_Proxy_update;
static PyObject *__pyx_n_s_PyPy;
static PyObject *__pyx_n_s_Pyx_CFunc_object__lParenWrappe;
static PyObject *__pyx_n_s_Pyx_CFunc_object__lParenWrappe_2;
static PyObject *__pyx_n_s_Pyx_CFunc_object__lParenWrappe_3;
static PyObject *__pyx_kp_s_Read_only_attribute_s;
static PyObject *__pyx_n_s_RecursionError;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_Sequence;
static PyObject *__pyx_n_s_Set;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_Unknown_OldStyle_Class;
static PyObject *__pyx_n_s_Wrapped;
static PyObject *__pyx_kp_s_Wrapped_Frozen_pxi;
static PyObject *__pyx_n_s_Wrapped___dir;
static PyObject *__pyx_n_s_Wrapped___reduce_cython;
static PyObject *__pyx_n_s_Wrapped___setstate_cython;
static PyObject *__pyx_n_s_Wrapped_comparator_locals_pass_t;
static PyObject *__pyx_kp_s_Wrapped_object_cannot_be_pickled;
static PyObject *__pyx_kp_s__17;
static PyObject *__pyx_n_s__182;
static PyObject *__pyx_n_s__2;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_u__33;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_n_s__44;
static PyObject *__pyx_kp_s__88;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_kp_s_a_zA_Z_a_zA_Z0_9;
static PyObject *__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2;
static PyObject *__pyx_n_s_abc;
static PyObject *__pyx_n_s_abs;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_add_2;
static PyObject *__pyx_n_s_aenter;
static PyObject *__pyx_n_s_aexit;
static PyObject *__pyx_n_s_alist;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_always_delegated_attributes;
static PyObject *__pyx_n_s_and;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_asyncio_coroutines;
static PyObject *__pyx_n_s_attr_type_check;
static PyObject *__pyx_n_s_attribute_protected;
static PyObject *__pyx_n_s_basestring;
static PyObject *__pyx_n_s_bool;
static PyObject *__pyx_n_s_bool_2;
static PyObject *__pyx_n_s_build_regex;
static PyObject *__pyx_n_s_builtin;
static PyObject *__pyx_n_s_builtins;
static PyObject *__pyx_n_s_bytearray;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_n_s_bytes_2;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_call;
static PyObject *__pyx_n_s_ceil;
static PyObject *__pyx_n_s_ceil_2;
static PyObject *__pyx_n_s_cfunc_to_py;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_class_getitem;
static PyObject *__pyx_n_s_clear;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cmp;
static PyObject *__pyx_n_s_cn;
static PyObject *__pyx_n_s_code;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_collections_abc;
static PyObject *__pyx_n_s_compile;
static PyObject *__pyx_n_s_complex;
static PyObject *__pyx_n_s_complex_2;
static PyObject *__pyx_n_s_contains;
static PyObject *__pyx_n_s_contains_2;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_delattr;
static PyObject *__pyx_n_s_delete;
static PyObject *__pyx_n_s_delitem;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dict_2;
static PyObject *__pyx_n_s_dict_3;
static PyObject *__pyx_n_s_difference;
static PyObject *__pyx_n_s_dir;
static PyObject *__pyx_kp_u_disable;
static PyObject *__pyx_n_s_discard;
static PyObject *__pyx_n_s_divmod;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dynamic;
static PyObject *__pyx_kp_u_enable;
static PyObject *__pyx_n_s_endswith;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_environ;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_exc_type;
static PyObject *__pyx_n_s_exc_value;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_float_2;
static PyObject *__pyx_n_s_floor;
static PyObject *__pyx_n_s_floor_2;
static PyObject *__pyx_n_s_floordiv;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_freeze;
static PyObject *__pyx_n_s_frozen;
static PyObject *__pyx_n_s_frozenset;
static PyObject *__pyx_n_s_functools;
static PyObject *__pyx_kp_u_gc;
static PyObject *__pyx_n_s_ge;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getattribute;
static PyObject *__pyx_n_s_getitem;
static PyObject *__pyx_n_s_getsate;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_global_c_functions_pxi;
static PyObject *__pyx_n_s_gt;
static PyObject *__pyx_n_s_hash;
static PyObject *__pyx_n_s_hash_2;
static PyObject *__pyx_n_s_hash_protected;
static PyObject *__pyx_n_s_hash_val;
static PyObject *__pyx_n_s_help;
static PyObject *__pyx_n_s_help_protected;
static PyObject *__pyx_n_s_help_str;
static PyObject *__pyx_n_s_help_val;
static PyObject *__pyx_n_s_hidden_pickle_attributes;
static PyObject *__pyx_n_s_hide;
static PyObject *__pyx_n_s_hide_private;
static PyObject *__pyx_n_s_hide_regex;
static PyObject *__pyx_n_s_iadd;
static PyObject *__pyx_n_s_iand;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_id_class;
static PyObject *__pyx_n_s_id_protected;
static PyObject *__pyx_n_s_id_val;
static PyObject *__pyx_n_s_ifloordiv;
static PyObject *__pyx_n_s_ilshift;
static PyObject *__pyx_n_s_imatmul;
static PyObject *__pyx_n_s_immutable_builtin_attributes;
static PyObject *__pyx_n_s_imod;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_imul;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_init___locals_C;
static PyObject *__pyx_n_s_init_subclass;
static PyObject *__pyx_n_s_initializing;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_inst;
static PyObject *__pyx_n_s_instance_of_protected;
static PyObject *__pyx_n_s_instancecheck;
static PyObject *__pyx_n_s_instanceof;
static PyObject *__pyx_n_s_int;
static PyObject *__pyx_n_s_int_2;
static PyObject *__pyx_n_s_intersection;
static PyObject *__pyx_n_s_invert;
static PyObject *__pyx_n_s_ior;
static PyObject *__pyx_n_s_ipow;
static PyObject *__pyx_n_s_irshift;
static PyObject *__pyx_n_s_is_coroutine;
static PyObject *__pyx_kp_u_isenabled;
static PyObject *__pyx_n_s_isfrozen;
static PyObject *__pyx_n_s_isimmutable;
static PyObject *__pyx_n_s_isinstance;
static PyObject *__pyx_n_s_isinstance_protected;
static PyObject *__pyx_n_s_isinstance_val;
static PyObject *__pyx_n_s_isprivate;
static PyObject *__pyx_n_s_isprotected;
static PyObject *__pyx_n_s_isreadonly;
static PyObject *__pyx_n_s_issubclass;
static PyObject *__pyx_n_s_issubclass_protected;
static PyObject *__pyx_n_s_issubclass_val;
static PyObject *__pyx_n_s_isub;
static PyObject *__pyx_n_s_isvisible;
static PyObject *__pyx_n_s_iswrapped;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_items_py2;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_iteritems;
static PyObject *__pyx_n_s_iterkeys;
static PyObject *__pyx_n_s_itervalues;
static PyObject *__pyx_n_s_itruediv;
static PyObject *__pyx_n_s_ixor;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_keys_py2;
static PyObject *__pyx_n_s_kw1;
static PyObject *__pyx_n_s_kw2;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_le;
static PyObject *__pyx_n_s_len;
static PyObject *__pyx_n_s_length_hint;
static PyObject *__pyx_n_s_list;
static PyObject *__pyx_n_s_long;
static PyObject *__pyx_n_s_lshift;
static PyObject *__pyx_n_s_lstrip;
static PyObject *__pyx_n_s_lt;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_major;
static PyObject *__pyx_n_s_match;
static PyObject *__pyx_n_s_match_args;
static PyObject *__pyx_n_s_math;
static PyObject *__pyx_n_s_matmul;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_minor;
static PyObject *__pyx_n_s_mod;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_modules;
static PyObject *__pyx_n_s_mro_entries;
static PyObject *__pyx_n_s_mul;
static PyObject *__pyx_n_s_multiwrapped;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_ne;
static PyObject *__pyx_n_s_neg;
static PyObject *__pyx_n_s_never_writeable;
static PyObject *__pyx_n_s_never_writeable_private;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_o;
static PyObject *__pyx_kp_s_o_Invalid_type_s;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_oldstyle_class;
static PyObject *__pyx_n_s_op;
static PyObject *__pyx_n_s_or;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_package;
static PyObject *__pyx_n_s_partial;
static PyObject *__pyx_n_s_pass_to_wrapped;
static PyObject *__pyx_n_s_pattern;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_platform;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_popitem;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_pow;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_private;
static PyObject *__pyx_n_s_protect;
static PyObject *__pyx_n_s_protected_rules_from_kwargs_loca;
static PyObject *__pyx_n_s_py2_function_attrs_rw;
static PyObject *__pyx_n_s_pydoc;
static PyObject *__pyx_n_s_pyprotect_protected;
static PyObject *__pyx_n_s_python_implementation;
static PyObject *__pyx_kp_s_python_visible_pxi;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Frozen;
static PyObject *__pyx_n_s_pyx_unpickle_FrozenPrivacyDict;
static PyObject *__pyx_n_s_pyx_unpickle_FrozenPrivate;
static PyObject *__pyx_n_s_pyx_unpickle_FrozenProtected;
static PyObject *__pyx_n_s_pyx_unpickle_PrivacyDict;
static PyObject *__pyx_n_s_pyx_unpickle_Private;
static PyObject *__pyx_n_s_pyx_unpickle_Protected;
static PyObject *__pyx_n_s_pyx_unpickle_Proxy;
static PyObject *__pyx_n_s_pyx_unpickle_Wrapped;
static PyObject *__pyx_n_s_pyx_unpickle___HiddenPartial;
static PyObject *__pyx_n_s_pyx_unpickle___ProtectionData;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_s_radd;
static PyObject *__pyx_n_s_rand;
static PyObject *__pyx_n_s_rdivmod;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_render_doc;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_ret;
static PyObject *__pyx_n_s_ret_2;
static PyObject *__pyx_n_s_return;
static PyObject *__pyx_n_s_reverse;
static PyObject *__pyx_n_s_rfloordiv;
static PyObject *__pyx_n_s_rl;
static PyObject *__pyx_n_s_rlshift;
static PyObject *__pyx_n_s_rmatmul;
static PyObject *__pyx_n_s_rmod;
static PyObject *__pyx_n_s_rmul;
static PyObject *__pyx_n_s_ro;
static PyObject *__pyx_n_s_ro_data;
static PyObject *__pyx_n_s_ro_method;
static PyObject *__pyx_n_s_ro_regex;
static PyObject *__pyx_n_s_ror;
static PyObject *__pyx_n_s_round;
static PyObject *__pyx_n_s_rpow;
static PyObject *__pyx_n_s_rrshift;
static PyObject *__pyx_n_s_rshift;
static PyObject *__pyx_n_s_rstrip;
static PyObject *__pyx_n_s_rsub;
static PyObject *__pyx_n_s_rtruediv;
static PyObject *__pyx_n_s_rules;
static PyObject *__pyx_n_s_rw;
static PyObject *__pyx_n_s_rw_regex;
static PyObject *__pyx_n_s_rxor;
static PyObject *__pyx_kp_s_s;
static PyObject *__pyx_kp_s_s_________0_1;
static PyObject *__pyx_n_s_same_class_protected;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_set;
static PyObject *__pyx_n_s_set_2;
static PyObject *__pyx_n_s_set_name;
static PyObject *__pyx_n_s_setattr;
static PyObject *__pyx_n_s_setdefault;
static PyObject *__pyx_n_s_setitem;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_slots;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_spec;
static PyObject *__pyx_n_s_splitlines;
static PyObject *__pyx_n_s_startswith;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_str_2;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_sub;
static PyObject *__pyx_n_s_subclass;
static PyObject *__pyx_n_s_subclass_of_protected;
static PyObject *__pyx_n_s_subclasscheck;
static PyObject *__pyx_n_s_subclassof;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_tb;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_testop;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_truediv;
static PyObject *__pyx_n_s_trunc;
static PyObject *__pyx_n_s_tuple;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_types;
static PyObject *__pyx_n_s_unichr;
static PyObject *__pyx_n_s_unicode;
static PyObject *__pyx_n_s_union;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_use_setstate;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_values_py2;
static PyObject *__pyx_n_s_version_info;
static PyObject *__pyx_n_s_viewitems;
static PyObject *__pyx_n_s_viewkeys;
static PyObject *__pyx_n_s_viewvalues;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_want_frozen;
static PyObject *__pyx_n_s_weakref;
static PyObject *__pyx_n_s_wrap;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_x_2;
static PyObject *__pyx_n_s_xor;
#endif
/* #### Code section: decls ### */
static PyObject *__pyx_pf_9pyprotect_9protected_attribute_protected(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_2id_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_4hash_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_6isinstance_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyTypeObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_8issubclass_protected(CYTHON_UNUSED PyObject *__pyx_self, PyTypeObject *__pyx_v_o, PyTypeObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_10instance_of_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_w); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_12subclass_of_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_w); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_14same_class_protected(CYTHON_UNUSED PyObject *__pyx_self, PyTypeObject *__pyx_v_c, PyObject *__pyx_v_w); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16help_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_18contains(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_20isimmutable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_22iswrapped(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_24isfrozen(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_26isprivate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_28isprotected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_30isreadonly(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_32isvisible(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_34wrap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_36freeze(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_38private(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_76__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_40protect(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen, PyObject *__pyx_v_dynamic, PyObject *__pyx_v_hide_private, PyObject *__pyx_v_ro_data, PyObject *__pyx_v_ro_method, PyObject *__pyx_v_ro, PyObject *__pyx_v_rw, PyObject *__pyx_v_hide); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_42never_writeable(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_44never_writeable_private(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_46hidden_pickle_attributes(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_48always_delegated_attributes(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_50immutable_builtin_attributes(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_52__dir__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_27protected_rules_from_kwargs__build_regex(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_alist); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData___init__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_id_val, PyObject *__pyx_v_id_class, PyObject *__pyx_v_hash_val, PyObject *__pyx_v_isinstance_val, PyObject *__pyx_v_issubclass_val, PyObject *__pyx_v_instanceof, PyObject *__pyx_v_subclassof, PyObject *__pyx_v_help_val, PyObject *__pyx_v_help_str, PyObject *__pyx_v_testop, PyObject *__pyx_v_rules, PyObject *__pyx_v_freeze, PyObject *__pyx_v_private, PyObject *__pyx_v_protect, PyObject *__pyx_v_multiwrapped); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_2__getattribute__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4__setattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_a, CYTHON_UNUSED PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6__delattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_8__dir__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_2id___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_2id_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_2id_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_4help___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4help_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4help_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_7private___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7private_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7private_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10__reduce_cython__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_12__setstate_cython__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_5Proxy___init__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_2__repr__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_4__str__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_6__bytes__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_8__call__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_10__iter__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_13__format__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_5Proxy_15__bool__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_17__getitem__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_9pyprotect_9protected_5Proxy_19__contains__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static Py_ssize_t __pyx_pf_9pyprotect_9protected_5Proxy_21__len__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_23__length_hint__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_25__neg__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_27__pos__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_29__abs__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_31__invert__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_33__complex__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_35__int__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_37__float__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_39__index__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_41__round__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_43__trunc__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_45__floor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_47__ceil__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_49__add__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_51__mul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_53__sub__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_55__matmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
#endif
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_57__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_59__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_61__mod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_63__divmod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_65__pow__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_67__lshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_69__rshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_71__and__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_73__or__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_75__xor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_77__radd__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_79__rmul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_81__rsub__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_83__rmatmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
#endif
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_85__rtruediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_87__rfloordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_89__rmod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_91__rdivmod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_93__rpow__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_95__rlshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_97__rrshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_99__rand__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_101__ror__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_103__rxor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_105__iadd__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_107__imul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_109__isub__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_111__imod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_113__ilshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_115__irshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_117__iand__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_119__ior__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_121__ixor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_123__ipow__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_125__itruediv__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_127__ifloordiv__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_129__imatmul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
#endif
static int __pyx_pf_9pyprotect_9protected_5Proxy_131__setitem__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_5Proxy_133__delitem__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_135__instancecheck__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_inst); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_137__subclasscheck__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_subclass); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_139__enter__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_141__exit__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_exc_type, PyObject *__pyx_v_exc_value, PyObject *__pyx_v_tb); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_143__aenter__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_145__aexit__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_exc_type, PyObject *__pyx_v_exc_value, PyObject *__pyx_v_tb); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_147__match_args__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_149send(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_151throw(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9pyprotect_9protected_5Proxy_153__set__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_5Proxy_155__delete__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_inst); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_157clear(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_159setdefault(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_161pop(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_163popitem(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_165update(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_167append(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_169extend(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_171insert(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_173sort(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_175add(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_177discard(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_179remove(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_181reverse(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_183__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_185__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_7Wrapped___init__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen, PyObject *__pyx_v_rules, PyObject *__pyx_v_oldstyle_class); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_10comparator_pass_to_wrapped(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static int __pyx_pf_9pyprotect_9protected_7Wrapped_4__setattr__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_7Wrapped_6__delattr__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_8__dir__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_10__richcmp__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_7Wrapped_12__hash__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_14__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_16__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_6Frozen___init__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self, PyObject *__pyx_v_o); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_6Frozen_2__hash__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_6Frozen_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_6Frozen_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_6Frozen_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_11PrivacyDict___init__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_cn, PyObject *__pyx_v_frozen, PyObject *__pyx_v_oldstyle_class); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_4__getitem__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_9pyprotect_9protected_11PrivacyDict_6__delitem__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_9pyprotect_9protected_11PrivacyDict_8__setitem__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_10__repr__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_12__str__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_14copy(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_16keys(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_19items(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_22values(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_25keys_py2(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_27items_py2(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_29values_py2(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_31iterkeys(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_34iteritems(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_37itervalues(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_40viewitems(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_42viewvalues(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_44viewkeys(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_11PrivacyDict_46__hash__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_48__richcmp__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_50__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_52__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict___init__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_cn, PyObject *__pyx_v_oldstyle_class); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_2__hash__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_7Private___init__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen, PyObject *__pyx_v_rules); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Private_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static int __pyx_pf_9pyprotect_9protected_7Private_4__setattr__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_7Private_6__delattr__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Private_8__dir__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_7Private_10__hash__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Private_12__richcmp__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Private_14__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_7Private_16__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_13FrozenPrivate___init__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self, PyObject *__pyx_v_o); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_13FrozenPrivate_2__hash__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_13FrozenPrivate_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_13FrozenPrivate_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_13FrozenPrivate_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_9Protected___init__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_rules); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static int __pyx_pf_9pyprotect_9protected_9Protected_4__setattr__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_9Protected_6__delattr__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_8__dir__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_9Protected_10__hash__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_12__richcmp__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_14__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_16__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_15FrozenProtected___init__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_rules); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_15FrozenProtected_2__hash__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15FrozenProtected_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15FrozenProtected_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15FrozenProtected_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9pyprotect_9protected_15__HiddenPartial___init__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_2__getattribute__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static int __pyx_pf_9pyprotect_9protected_15__HiddenPartial_4__setattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a, CYTHON_UNUSED PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_9pyprotect_9protected_15__HiddenPartial_6__delattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_8__dir__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_9pyprotect_9protected_15__HiddenPartial_10__hash__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_12__repr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_14__str__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_16__bytes__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_18__call__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_20__reduce_cython__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_22__setstate_cython__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_54__pyx_unpickle___ProtectionData(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_56__pyx_unpickle_Proxy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_58__pyx_unpickle_Wrapped(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_60__pyx_unpickle_Frozen(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_62__pyx_unpickle_PrivacyDict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_64__pyx_unpickle_FrozenPrivacyDict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_66__pyx_unpickle_Private(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_68__pyx_unpickle_FrozenPrivate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_70__pyx_unpickle_Protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_72__pyx_unpickle_FrozenProtected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9pyprotect_9protected_74__pyx_unpickle___HiddenPartial(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_wrap(PyObject *__pyx_self, struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_wrap(PyObject *__pyx_self, struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_wrap(PyObject *__pyx_self, struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_tp_new_9pyprotect_9protected___ProtectionData(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_Proxy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_Wrapped(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_Frozen(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_PrivacyDict(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_FrozenPrivacyDict(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_Private(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_FrozenPrivate(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_Protected(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected_FrozenProtected(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___HiddenPartial(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_1_comparator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_2_keys(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_3_items(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_4_values(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_5_iterkeys(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_6_iteritems(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_7_itervalues(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_update = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_intersection = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_union = {0, 0, 0, 0, 0};
#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_6017409;
static PyObject *__pyx_int_19817578;
static PyObject *__pyx_int_50167005;
static PyObject *__pyx_int_78131249;
static PyObject *__pyx_int_115090883;
static PyObject *__pyx_int_152356376;
static PyObject *__pyx_int_155231502;
static PyObject *__pyx_int_161740782;
static PyObject *__pyx_int_180490543;
static PyObject *__pyx_int_209109265;
static PyObject *__pyx_int_210565405;
static PyObject *__pyx_int_218598457;
static PyObject *__pyx_int_247595846;
static PyObject *__pyx_int_252507329;
static PyObject *__pyx_int_262487005;
#endif
#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__18;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_codeobj__6;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__119;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__144;
static PyObject *__pyx_tuple__146;
static PyObject *__pyx_tuple__170;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__65;
static PyObject *__pyx_codeobj__67;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__69;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__75;
static PyObject *__pyx_codeobj__76;
static PyObject *__pyx_codeobj__77;
static PyObject *__pyx_codeobj__78;
static PyObject *__pyx_codeobj__79;
static PyObject *__pyx_codeobj__80;
static PyObject *__pyx_codeobj__81;
static PyObject *__pyx_codeobj__92;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__96;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__99;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__101;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__103;
static PyObject *__pyx_codeobj__104;
static PyObject *__pyx_codeobj__105;
static PyObject *__pyx_codeobj__107;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__110;
static PyObject *__pyx_codeobj__112;
static PyObject *__pyx_codeobj__113;
static PyObject *__pyx_codeobj__114;
static PyObject *__pyx_codeobj__115;
static PyObject *__pyx_codeobj__117;
static PyObject *__pyx_codeobj__118;
static PyObject *__pyx_codeobj__120;
static PyObject *__pyx_codeobj__121;
static PyObject *__pyx_codeobj__122;
static PyObject *__pyx_codeobj__123;
static PyObject *__pyx_codeobj__124;
static PyObject *__pyx_codeobj__125;
static PyObject *__pyx_codeobj__126;
static PyObject *__pyx_codeobj__127;
static PyObject *__pyx_codeobj__128;
static PyObject *__pyx_codeobj__129;
static PyObject *__pyx_codeobj__130;
static PyObject *__pyx_codeobj__131;
static PyObject *__pyx_codeobj__132;
static PyObject *__pyx_codeobj__133;
static PyObject *__pyx_codeobj__134;
static PyObject *__pyx_codeobj__135;
static PyObject *__pyx_codeobj__136;
static PyObject *__pyx_codeobj__137;
static PyObject *__pyx_codeobj__138;
static PyObject *__pyx_codeobj__139;
static PyObject *__pyx_codeobj__141;
static PyObject *__pyx_codeobj__145;
static PyObject *__pyx_codeobj__147;
static PyObject *__pyx_codeobj__148;
static PyObject *__pyx_codeobj__149;
static PyObject *__pyx_codeobj__150;
static PyObject *__pyx_codeobj__151;
static PyObject *__pyx_codeobj__152;
static PyObject *__pyx_codeobj__153;
static PyObject *__pyx_codeobj__154;
static PyObject *__pyx_codeobj__155;
static PyObject *__pyx_codeobj__156;
static PyObject *__pyx_codeobj__157;
static PyObject *__pyx_codeobj__158;
static PyObject *__pyx_codeobj__159;
static PyObject *__pyx_codeobj__160;
static PyObject *__pyx_codeobj__161;
static PyObject *__pyx_codeobj__162;
static PyObject *__pyx_codeobj__163;
static PyObject *__pyx_codeobj__164;
static PyObject *__pyx_codeobj__165;
static PyObject *__pyx_codeobj__166;
static PyObject *__pyx_codeobj__167;
static PyObject *__pyx_codeobj__168;
static PyObject *__pyx_codeobj__169;
static PyObject *__pyx_codeobj__171;
static PyObject *__pyx_codeobj__172;
static PyObject *__pyx_codeobj__173;
static PyObject *__pyx_codeobj__174;
static PyObject *__pyx_codeobj__175;
static PyObject *__pyx_codeobj__176;
static PyObject *__pyx_codeobj__177;
static PyObject *__pyx_codeobj__178;
static PyObject *__pyx_codeobj__179;
static PyObject *__pyx_codeobj__180;
static PyObject *__pyx_codeobj__181;
#endif
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
#if CYTHON_USE_MODULE_STATE
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___ProtectionData;
  PyObject *__pyx_type_9pyprotect_9protected___ProtectionData;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_Proxy;
  PyObject *__pyx_type_9pyprotect_9protected_Proxy;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_Wrapped;
  PyObject *__pyx_type_9pyprotect_9protected_Wrapped;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_Frozen;
  PyObject *__pyx_type_9pyprotect_9protected_Frozen;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_PrivacyDict;
  PyObject *__pyx_type_9pyprotect_9protected_PrivacyDict;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict;
  PyObject *__pyx_type_9pyprotect_9protected_FrozenPrivacyDict;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_Private;
  PyObject *__pyx_type_9pyprotect_9protected_Private;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_FrozenPrivate;
  PyObject *__pyx_type_9pyprotect_9protected_FrozenPrivate;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_Protected;
  PyObject *__pyx_type_9pyprotect_9protected_Protected;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected_FrozenProtected;
  PyObject *__pyx_type_9pyprotect_9protected_FrozenProtected;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___HiddenPartial;
  PyObject *__pyx_type_9pyprotect_9protected___HiddenPartial;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct____iter__;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems;
  PyTypeObject *__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues;
  PyObject *__pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues;
  PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self;
  PyObject *__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self;
  PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c;
  PyObject *__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c;
  PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op;
  PyObject *__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op;
  PyObject *__pyx_n_s_;
  PyObject *__pyx_kp_s_0_1;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_AttributeError;
  PyObject *__pyx_n_s_C;
  PyObject *__pyx_kp_s_Cannot_add_attribute_s_s;
  PyObject *__pyx_kp_s_Cannot_delete_attribute_s;
  PyObject *__pyx_kp_s_Cannot_delete_attribute_s_s;
  PyObject *__pyx_kp_s_Cannot_delete_private_attribute;
  PyObject *__pyx_kp_s_Cannot_modify_attribute_s;
  PyObject *__pyx_kp_s_Cannot_set_attribute_s_s;
  PyObject *__pyx_kp_s_Cannot_set_private_attribute_s_s;
  PyObject *__pyx_n_s_CollectionsABC;
  PyObject *__pyx_kp_s_Double_wrapped;
  PyObject *__pyx_n_s_FrameType;
  PyObject *__pyx_n_s_Frozen;
  PyObject *__pyx_n_s_FrozenPrivacyDict;
  PyObject *__pyx_n_s_FrozenPrivacyDict___reduce_cytho;
  PyObject *__pyx_n_s_FrozenPrivacyDict___setstate_cyt;
  PyObject *__pyx_n_s_FrozenPrivate;
  PyObject *__pyx_n_s_FrozenPrivate___reduce_cython;
  PyObject *__pyx_n_s_FrozenPrivate___setstate_cython;
  PyObject *__pyx_n_s_FrozenProtected;
  PyObject *__pyx_n_s_FrozenProtected___reduce_cython;
  PyObject *__pyx_n_s_FrozenProtected___setstate_cytho;
  PyObject *__pyx_n_s_Frozen___reduce_cython;
  PyObject *__pyx_n_s_Frozen___setstate_cython;
  PyObject *__pyx_n_s_HiddenPartial;
  PyObject *__pyx_n_s_HiddenPartial___bytes;
  PyObject *__pyx_n_s_HiddenPartial___dir;
  PyObject *__pyx_n_s_HiddenPartial___reduce_cython;
  PyObject *__pyx_n_s_HiddenPartial___setstate_cytho;
  PyObject *__pyx_kp_s_HiddenPartial_pxi;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_Mapping;
  PyObject *__pyx_n_s_ModuleType;
  PyObject *__pyx_n_s_MutableMapping;
  PyObject *__pyx_n_s_MutableSequence;
  PyObject *__pyx_n_s_MutableSet;
  PyObject *__pyx_n_s_None;
  PyObject *__pyx_n_s_NotImplemented;
  PyObject *__pyx_kp_s_Object_Private_s_has_no_attribut;
  PyObject *__pyx_kp_s_Object_Protected_s_has_no_attrib;
  PyObject *__pyx_kp_s_Object_Wrapped_s_has_no_attribut;
  PyObject *__pyx_kp_s_Object___HiddenPartial_has_no_at;
  PyObject *__pyx_kp_s_Object_is_read_only;
  PyObject *__pyx_kp_s_Object_s_has_no_attribute_s;
  PyObject *__pyx_n_s_PYPY;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_n_s_PrivacyDict;
  PyObject *__pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px;
  PyObject *__pyx_n_s_PrivacyDict___reduce_cython;
  PyObject *__pyx_n_s_PrivacyDict___setstate_cython;
  PyObject *__pyx_n_s_PrivacyDict_copy;
  PyObject *__pyx_n_s_PrivacyDict_items;
  PyObject *__pyx_n_s_PrivacyDict_items_py2;
  PyObject *__pyx_n_s_PrivacyDict_iteritems;
  PyObject *__pyx_n_s_PrivacyDict_iterkeys;
  PyObject *__pyx_n_s_PrivacyDict_itervalues;
  PyObject *__pyx_n_s_PrivacyDict_keys;
  PyObject *__pyx_n_s_PrivacyDict_keys_py2;
  PyObject *__pyx_n_s_PrivacyDict_values;
  PyObject *__pyx_n_s_PrivacyDict_values_py2;
  PyObject *__pyx_n_s_PrivacyDict_viewitems;
  PyObject *__pyx_n_s_PrivacyDict_viewkeys;
  PyObject *__pyx_n_s_PrivacyDict_viewvalues;
  PyObject *__pyx_n_s_Private;
  PyObject *__pyx_kp_s_Private_FrozenPrivate_pxi;
  PyObject *__pyx_n_s_Private___dir;
  PyObject *__pyx_n_s_Private___reduce_cython;
  PyObject *__pyx_n_s_Private___setstate_cython;
  PyObject *__pyx_n_s_Protected;
  PyObject *__pyx_n_s_Protected_2;
  PyObject *__pyx_kp_s_Protected_FrozenProtected_pxi;
  PyObject *__pyx_n_s_Protected___dir;
  PyObject *__pyx_n_s_Protected___reduce_cython;
  PyObject *__pyx_n_s_Protected___setstate_cython;
  PyObject *__pyx_n_s_ProtectionData;
  PyObject *__pyx_n_s_ProtectionData___dir;
  PyObject *__pyx_n_s_ProtectionData___reduce_cython;
  PyObject *__pyx_n_s_ProtectionData___setstate_cyth;
  PyObject *__pyx_kp_s_ProtectionData_pxi;
  PyObject *__pyx_n_s_ProtectionError;
  PyObject *__pyx_n_s_Proxy;
  PyObject *__pyx_n_s_Proxy___aenter;
  PyObject *__pyx_n_s_Proxy___aexit;
  PyObject *__pyx_n_s_Proxy___bytes;
  PyObject *__pyx_n_s_Proxy___ceil;
  PyObject *__pyx_n_s_Proxy___complex;
  PyObject *__pyx_n_s_Proxy___enter;
  PyObject *__pyx_n_s_Proxy___exit;
  PyObject *__pyx_n_s_Proxy___floor;
  PyObject *__pyx_n_s_Proxy___format;
  PyObject *__pyx_n_s_Proxy___instancecheck;
  PyObject *__pyx_n_s_Proxy___iter;
  PyObject *__pyx_n_s_Proxy___length_hint;
  PyObject *__pyx_n_s_Proxy___match_args;
  PyObject *__pyx_n_s_Proxy___reduce_cython;
  PyObject *__pyx_n_s_Proxy___round;
  PyObject *__pyx_n_s_Proxy___setstate_cython;
  PyObject *__pyx_n_s_Proxy___subclasscheck;
  PyObject *__pyx_n_s_Proxy___trunc;
  PyObject *__pyx_n_s_Proxy__aexit;
  PyObject *__pyx_n_s_Proxy__exit;
  PyObject *__pyx_n_s_Proxy_add;
  PyObject *__pyx_n_s_Proxy_append;
  PyObject *__pyx_n_s_Proxy_clear;
  PyObject *__pyx_n_s_Proxy_discard;
  PyObject *__pyx_n_s_Proxy_extend;
  PyObject *__pyx_n_s_Proxy_insert;
  PyObject *__pyx_n_s_Proxy_pop;
  PyObject *__pyx_n_s_Proxy_popitem;
  PyObject *__pyx_kp_s_Proxy_pxi;
  PyObject *__pyx_n_s_Proxy_remove;
  PyObject *__pyx_n_s_Proxy_reverse;
  PyObject *__pyx_n_s_Proxy_send;
  PyObject *__pyx_n_s_Proxy_setdefault;
  PyObject *__pyx_n_s_Proxy_sort;
  PyObject *__pyx_n_s_Proxy_throw;
  PyObject *__pyx_n_s_Proxy_update;
  PyObject *__pyx_n_s_PyPy;
  PyObject *__pyx_n_s_Pyx_CFunc_object__lParenWrappe;
  PyObject *__pyx_n_s_Pyx_CFunc_object__lParenWrappe_2;
  PyObject *__pyx_n_s_Pyx_CFunc_object__lParenWrappe_3;
  PyObject *__pyx_kp_s_Read_only_attribute_s;
  PyObject *__pyx_n_s_RecursionError;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_n_s_Sequence;
  PyObject *__pyx_n_s_Set;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_Unknown_OldStyle_Class;
  PyObject *__pyx_n_s_Wrapped;
  PyObject *__pyx_kp_s_Wrapped_Frozen_pxi;
  PyObject *__pyx_n_s_Wrapped___dir;
  PyObject *__pyx_n_s_Wrapped___reduce_cython;
  PyObject *__pyx_n_s_Wrapped___setstate_cython;
  PyObject *__pyx_n_s_Wrapped_comparator_locals_pass_t;
  PyObject *__pyx_kp_s_Wrapped_object_cannot_be_pickled;
  PyObject *__pyx_kp_s__17;
  PyObject *__pyx_n_s__182;
  PyObject *__pyx_n_s__2;
  PyObject *__pyx_kp_s__3;
  PyObject *__pyx_kp_u__33;
  PyObject *__pyx_kp_s__4;
  PyObject *__pyx_n_s__44;
  PyObject *__pyx_kp_s__88;
  PyObject *__pyx_n_s_a;
  PyObject *__pyx_kp_s_a_zA_Z_a_zA_Z0_9;
  PyObject *__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2;
  PyObject *__pyx_n_s_abc;
  PyObject *__pyx_n_s_abs;
  PyObject *__pyx_n_s_add;
  PyObject *__pyx_n_s_add_2;
  PyObject *__pyx_n_s_aenter;
  PyObject *__pyx_n_s_aexit;
  PyObject *__pyx_n_s_alist;
  PyObject *__pyx_n_s_all;
  PyObject *__pyx_n_s_always_delegated_attributes;
  PyObject *__pyx_n_s_and;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_attr_type_check;
  PyObject *__pyx_n_s_attribute_protected;
  PyObject *__pyx_n_s_basestring;
  PyObject *__pyx_n_s_bool;
  PyObject *__pyx_n_s_bool_2;
  PyObject *__pyx_n_s_build_regex;
  PyObject *__pyx_n_s_builtin;
  PyObject *__pyx_n_s_builtins;
  PyObject *__pyx_n_s_bytearray;
  PyObject *__pyx_n_s_bytes;
  PyObject *__pyx_n_s_bytes_2;
  PyObject *__pyx_n_s_c;
  PyObject *__pyx_n_s_call;
  PyObject *__pyx_n_s_ceil;
  PyObject *__pyx_n_s_ceil_2;
  PyObject *__pyx_n_s_cfunc_to_py;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_clear;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_cmp;
  PyObject *__pyx_n_s_cn;
  PyObject *__pyx_n_s_code;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_s_collections_abc;
  PyObject *__pyx_n_s_compile;
  PyObject *__pyx_n_s_complex;
  PyObject *__pyx_n_s_complex_2;
  PyObject *__pyx_n_s_contains;
  PyObject *__pyx_n_s_contains_2;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_d;
  PyObject *__pyx_n_s_defaults;
  PyObject *__pyx_n_s_delattr;
  PyObject *__pyx_n_s_delete;
  PyObject *__pyx_n_s_delitem;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_n_s_dict_3;
  PyObject *__pyx_n_s_difference;
  PyObject *__pyx_n_s_dir;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_discard;
  PyObject *__pyx_n_s_divmod;
  PyObject *__pyx_n_s_doc;
  PyObject *__pyx_n_s_dynamic;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_endswith;
  PyObject *__pyx_n_s_enter;
  PyObject *__pyx_n_s_environ;
  PyObject *__pyx_n_s_eq;
  PyObject *__pyx_n_s_exc_type;
  PyObject *__pyx_n_s_exc_value;
  PyObject *__pyx_n_s_exit;
  PyObject *__pyx_n_s_extend;
  PyObject *__pyx_n_s_file;
  PyObject *__pyx_n_s_float;
  PyObject *__pyx_n_s_float_2;
  PyObject *__pyx_n_s_floor;
  PyObject *__pyx_n_s_floor_2;
  PyObject *__pyx_n_s_floordiv;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_freeze;
  PyObject *__pyx_n_s_frozen;
  PyObject *__pyx_n_s_frozenset;
  PyObject *__pyx_n_s_functools;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_ge;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_getattribute;
  PyObject *__pyx_n_s_getitem;
  PyObject *__pyx_n_s_getsate;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_kp_s_global_c_functions_pxi;
  PyObject *__pyx_n_s_gt;
  PyObject *__pyx_n_s_hash;
  PyObject *__pyx_n_s_hash_2;
  PyObject *__pyx_n_s_hash_protected;
  PyObject *__pyx_n_s_hash_val;
  PyObject *__pyx_n_s_help;
  PyObject *__pyx_n_s_help_protected;
  PyObject *__pyx_n_s_help_str;
  PyObject *__pyx_n_s_help_val;
  PyObject *__pyx_n_s_hidden_pickle_attributes;
  PyObject *__pyx_n_s_hide;
  PyObject *__pyx_n_s_hide_private;
  PyObject *__pyx_n_s_hide_regex;
  PyObject *__pyx_n_s_iadd;
  PyObject *__pyx_n_s_iand;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_s_id_class;
  PyObject *__pyx_n_s_id_protected;
  PyObject *__pyx_n_s_id_val;
  PyObject *__pyx_n_s_ifloordiv;
  PyObject *__pyx_n_s_ilshift;
  PyObject *__pyx_n_s_imatmul;
  PyObject *__pyx_n_s_immutable_builtin_attributes;
  PyObject *__pyx_n_s_imod;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_imul;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init___locals_C;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_insert;
  PyObject *__pyx_n_s_inst;
  PyObject *__pyx_n_s_instance_of_protected;
  PyObject *__pyx_n_s_instancecheck;
  PyObject *__pyx_n_s_instanceof;
  PyObject *__pyx_n_s_int;
  PyObject *__pyx_n_s_int_2;
  PyObject *__pyx_n_s_intersection;
  PyObject *__pyx_n_s_invert;
  PyObject *__pyx_n_s_ior;
  PyObject *__pyx_n_s_ipow;
  PyObject *__pyx_n_s_irshift;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_isfrozen;
  PyObject *__pyx_n_s_isimmutable;
  PyObject *__pyx_n_s_isinstance;
  PyObject *__pyx_n_s_isinstance_protected;
  PyObject *__pyx_n_s_isinstance_val;
  PyObject *__pyx_n_s_isprivate;
  PyObject *__pyx_n_s_isprotected;
  PyObject *__pyx_n_s_isreadonly;
  PyObject *__pyx_n_s_issubclass;
  PyObject *__pyx_n_s_issubclass_protected;
  PyObject *__pyx_n_s_issubclass_val;
  PyObject *__pyx_n_s_isub;
  PyObject *__pyx_n_s_isvisible;
  PyObject *__pyx_n_s_iswrapped;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_items_py2;
  PyObject *__pyx_n_s_iter;
  PyObject *__pyx_n_s_iteritems;
  PyObject *__pyx_n_s_iterkeys;
  PyObject *__pyx_n_s_itervalues;
  PyObject *__pyx_n_s_itruediv;
  PyObject *__pyx_n_s_ixor;
  PyObject *__pyx_n_s_join;
  PyObject *__pyx_n_s_k;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_keys_py2;
  PyObject *__pyx_n_s_kw1;
  PyObject *__pyx_n_s_kw2;
  PyObject *__pyx_n_s_kwargs;
  PyObject *__pyx_n_s_le;
  PyObject *__pyx_n_s_len;
  PyObject *__pyx_n_s_length_hint;
  PyObject *__pyx_n_s_list;
  PyObject *__pyx_n_s_long;
  PyObject *__pyx_n_s_lshift;
  PyObject *__pyx_n_s_lstrip;
  PyObject *__pyx_n_s_lt;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_major;
  PyObject *__pyx_n_s_match;
  PyObject *__pyx_n_s_match_args;
  PyObject *__pyx_n_s_math;
  PyObject *__pyx_n_s_matmul;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_minor;
  PyObject *__pyx_n_s_mod;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_modules;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_mul;
  PyObject *__pyx_n_s_multiwrapped;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_ne;
  PyObject *__pyx_n_s_neg;
  PyObject *__pyx_n_s_never_writeable;
  PyObject *__pyx_n_s_never_writeable_private;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_o;
  PyObject *__pyx_kp_s_o_Invalid_type_s;
  PyObject *__pyx_n_s_object;
  PyObject *__pyx_n_s_oldstyle_class;
  PyObject *__pyx_n_s_op;
  PyObject *__pyx_n_s_or;
  PyObject *__pyx_n_s_os;
  PyObject *__pyx_n_s_p;
  PyObject *__pyx_n_s_package;
  PyObject *__pyx_n_s_partial;
  PyObject *__pyx_n_s_pass_to_wrapped;
  PyObject *__pyx_n_s_pattern;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_platform;
  PyObject *__pyx_n_s_pop;
  PyObject *__pyx_n_s_popitem;
  PyObject *__pyx_n_s_pos;
  PyObject *__pyx_n_s_pow;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_private;
  PyObject *__pyx_n_s_protect;
  PyObject *__pyx_n_s_protected_rules_from_kwargs_loca;
  PyObject *__pyx_n_s_py2_function_attrs_rw;
  PyObject *__pyx_n_s_pydoc;
  PyObject *__pyx_n_s_pyprotect_protected;
  PyObject *__pyx_n_s_python_implementation;
  PyObject *__pyx_kp_s_python_visible_pxi;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_Frozen;
  PyObject *__pyx_n_s_pyx_unpickle_FrozenPrivacyDict;
  PyObject *__pyx_n_s_pyx_unpickle_FrozenPrivate;
  PyObject *__pyx_n_s_pyx_unpickle_FrozenProtected;
  PyObject *__pyx_n_s_pyx_unpickle_PrivacyDict;
  PyObject *__pyx_n_s_pyx_unpickle_Private;
  PyObject *__pyx_n_s_pyx_unpickle_Protected;
  PyObject *__pyx_n_s_pyx_unpickle_Proxy;
  PyObject *__pyx_n_s_pyx_unpickle_Wrapped;
  PyObject *__pyx_n_s_pyx_unpickle___HiddenPartial;
  PyObject *__pyx_n_s_pyx_unpickle___ProtectionData;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_s_r;
  PyObject *__pyx_n_s_radd;
  PyObject *__pyx_n_s_rand;
  PyObject *__pyx_n_s_rdivmod;
  PyObject *__pyx_n_s_re;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_remove;
  PyObject *__pyx_n_s_render_doc;
  PyObject *__pyx_n_s_repr;
  PyObject *__pyx_n_s_ret;
  PyObject *__pyx_n_s_ret_2;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_reverse;
  PyObject *__pyx_n_s_rfloordiv;
  PyObject *__pyx_n_s_rl;
  PyObject *__pyx_n_s_rlshift;
  PyObject *__pyx_n_s_rmatmul;
  PyObject *__pyx_n_s_rmod;
  PyObject *__pyx_n_s_rmul;
  PyObject *__pyx_n_s_ro;
  PyObject *__pyx_n_s_ro_data;
  PyObject *__pyx_n_s_ro_method;
  PyObject *__pyx_n_s_ro_regex;
  PyObject *__pyx_n_s_ror;
  PyObject *__pyx_n_s_round;
  PyObject *__pyx_n_s_rpow;
  PyObject *__pyx_n_s_rrshift;
  PyObject *__pyx_n_s_rshift;
  PyObject *__pyx_n_s_rstrip;
  PyObject *__pyx_n_s_rsub;
  PyObject *__pyx_n_s_rtruediv;
  PyObject *__pyx_n_s_rules;
  PyObject *__pyx_n_s_rw;
  PyObject *__pyx_n_s_rw_regex;
  PyObject *__pyx_n_s_rxor;
  PyObject *__pyx_kp_s_s;
  PyObject *__pyx_kp_s_s_________0_1;
  PyObject *__pyx_n_s_same_class_protected;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set;
  PyObject *__pyx_n_s_set_2;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_setattr;
  PyObject *__pyx_n_s_setdefault;
  PyObject *__pyx_n_s_setitem;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_slots;
  PyObject *__pyx_n_s_sort;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_splitlines;
  PyObject *__pyx_n_s_startswith;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_n_s_str_2;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_sub;
  PyObject *__pyx_n_s_subclass;
  PyObject *__pyx_n_s_subclass_of_protected;
  PyObject *__pyx_n_s_subclasscheck;
  PyObject *__pyx_n_s_subclassof;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_sys;
  PyObject *__pyx_n_s_tb;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_testop;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_truediv;
  PyObject *__pyx_n_s_trunc;
  PyObject *__pyx_n_s_tuple;
  PyObject *__pyx_n_s_type;
  PyObject *__pyx_n_s_types;
  PyObject *__pyx_n_s_unichr;
  PyObject *__pyx_n_s_unicode;
  PyObject *__pyx_n_s_union;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_v;
  PyObject *__pyx_n_s_val;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_n_s_values_py2;
  PyObject *__pyx_n_s_version_info;
  PyObject *__pyx_n_s_viewitems;
  PyObject *__pyx_n_s_viewkeys;
  PyObject *__pyx_n_s_viewvalues;
  PyObject *__pyx_n_s_w;
  PyObject *__pyx_n_s_want_frozen;
  PyObject *__pyx_n_s_weakref;
  PyObject *__pyx_n_s_wrap;
  PyObject *__pyx_n_s_x;
  PyObject *__pyx_n_s_x_2;
  PyObject *__pyx_n_s_xor;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_7;
  PyObject *__pyx_int_6017409;
  PyObject *__pyx_int_19817578;
  PyObject *__pyx_int_50167005;
  PyObject *__pyx_int_78131249;
  PyObject *__pyx_int_115090883;
  PyObject *__pyx_int_152356376;
  PyObject *__pyx_int_155231502;
  PyObject *__pyx_int_161740782;
  PyObject *__pyx_int_180490543;
  PyObject *__pyx_int_209109265;
  PyObject *__pyx_int_210565405;
  PyObject *__pyx_int_218598457;
  PyObject *__pyx_int_247595846;
  PyObject *__pyx_int_252507329;
  PyObject *__pyx_int_262487005;
  PyObject *__pyx_tuple__5;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_slice__18;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__40;
  PyObject *__pyx_tuple__42;
  PyObject *__pyx_tuple__45;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__56;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__71;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__74;
  PyObject *__pyx_tuple__82;
  PyObject *__pyx_tuple__83;
  PyObject *__pyx_tuple__84;
  PyObject *__pyx_tuple__85;
  PyObject *__pyx_tuple__86;
  PyObject *__pyx_tuple__87;
  PyObject *__pyx_tuple__89;
  PyObject *__pyx_tuple__90;
  PyObject *__pyx_tuple__91;
  PyObject *__pyx_tuple__93;
  PyObject *__pyx_tuple__95;
  PyObject *__pyx_tuple__98;
  PyObject *__pyx_codeobj__6;
  PyObject *__pyx_tuple__106;
  PyObject *__pyx_tuple__108;
  PyObject *__pyx_tuple__111;
  PyObject *__pyx_tuple__116;
  PyObject *__pyx_tuple__119;
  PyObject *__pyx_tuple__140;
  PyObject *__pyx_tuple__142;
  PyObject *__pyx_tuple__143;
  PyObject *__pyx_tuple__144;
  PyObject *__pyx_tuple__146;
  PyObject *__pyx_tuple__170;
  PyObject *__pyx_codeobj__19;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__24;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__26;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__28;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__55;
  PyObject *__pyx_codeobj__57;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__60;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__65;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__68;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__78;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__80;
  PyObject *__pyx_codeobj__81;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__94;
  PyObject *__pyx_codeobj__96;
  PyObject *__pyx_codeobj__97;
  PyObject *__pyx_codeobj__99;
  PyObject *__pyx_codeobj__100;
  PyObject *__pyx_codeobj__101;
  PyObject *__pyx_codeobj__102;
  PyObject *__pyx_codeobj__103;
  PyObject *__pyx_codeobj__104;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__107;
  PyObject *__pyx_codeobj__109;
  PyObject *__pyx_codeobj__110;
  PyObject *__pyx_codeobj__112;
  PyObject *__pyx_codeobj__113;
  PyObject *__pyx_codeobj__114;
  PyObject *__pyx_codeobj__115;
  PyObject *__pyx_codeobj__117;
  PyObject *__pyx_codeobj__118;
  PyObject *__pyx_codeobj__120;
  PyObject *__pyx_codeobj__121;
  PyObject *__pyx_codeobj__122;
  PyObject *__pyx_codeobj__123;
  PyObject *__pyx_codeobj__124;
  PyObject *__pyx_codeobj__125;
  PyObject *__pyx_codeobj__126;
  PyObject *__pyx_codeobj__127;
  PyObject *__pyx_codeobj__128;
  PyObject *__pyx_codeobj__129;
  PyObject *__pyx_codeobj__130;
  PyObject *__pyx_codeobj__131;
  PyObject *__pyx_codeobj__132;
  PyObject *__pyx_codeobj__133;
  PyObject *__pyx_codeobj__134;
  PyObject *__pyx_codeobj__135;
  PyObject *__pyx_codeobj__136;
  PyObject *__pyx_codeobj__137;
  PyObject *__pyx_codeobj__138;
  PyObject *__pyx_codeobj__139;
  PyObject *__pyx_codeobj__141;
  PyObject *__pyx_codeobj__145;
  PyObject *__pyx_codeobj__147;
  PyObject *__pyx_codeobj__148;
  PyObject *__pyx_codeobj__149;
  PyObject *__pyx_codeobj__150;
  PyObject *__pyx_codeobj__151;
  PyObject *__pyx_codeobj__152;
  PyObject *__pyx_codeobj__153;
  PyObject *__pyx_codeobj__154;
  PyObject *__pyx_codeobj__155;
  PyObject *__pyx_codeobj__156;
  PyObject *__pyx_codeobj__157;
  PyObject *__pyx_codeobj__158;
  PyObject *__pyx_codeobj__159;
  PyObject *__pyx_codeobj__160;
  PyObject *__pyx_codeobj__161;
  PyObject *__pyx_codeobj__162;
  PyObject *__pyx_codeobj__163;
  PyObject *__pyx_codeobj__164;
  PyObject *__pyx_codeobj__165;
  PyObject *__pyx_codeobj__166;
  PyObject *__pyx_codeobj__167;
  PyObject *__pyx_codeobj__168;
  PyObject *__pyx_codeobj__169;
  PyObject *__pyx_codeobj__171;
  PyObject *__pyx_codeobj__172;
  PyObject *__pyx_codeobj__173;
  PyObject *__pyx_codeobj__174;
  PyObject *__pyx_codeobj__175;
  PyObject *__pyx_codeobj__176;
  PyObject *__pyx_codeobj__177;
  PyObject *__pyx_codeobj__178;
  PyObject *__pyx_codeobj__179;
  PyObject *__pyx_codeobj__180;
  PyObject *__pyx_codeobj__181;
} __pyx_mstate;

#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___ProtectionData);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___ProtectionData);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_Proxy);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_Proxy);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_Wrapped);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_Wrapped);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_Frozen);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_Frozen);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_PrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_PrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_FrozenPrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_Private);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_Private);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_FrozenPrivate);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_FrozenPrivate);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_Protected);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_Protected);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected_FrozenProtected);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected_FrozenProtected);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___HiddenPartial);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___HiddenPartial);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems);
  Py_CLEAR(clear_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues);
  Py_CLEAR(clear_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues);
  Py_CLEAR(clear_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self);
  Py_CLEAR(clear_module_state->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self);
  Py_CLEAR(clear_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c);
  Py_CLEAR(clear_module_state->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c);
  Py_CLEAR(clear_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op);
  Py_CLEAR(clear_module_state->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op);
  Py_CLEAR(clear_module_state->__pyx_n_s_);
  Py_CLEAR(clear_module_state->__pyx_kp_s_0_1);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_AttributeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_C);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_add_attribute_s_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_delete_attribute_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_delete_attribute_s_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_delete_private_attribute);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_modify_attribute_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_set_attribute_s_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_set_private_attribute_s_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_CollectionsABC);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Double_wrapped);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrameType);
  Py_CLEAR(clear_module_state->__pyx_n_s_Frozen);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenPrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenPrivacyDict___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenPrivacyDict___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenPrivate);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenPrivate___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenPrivate___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenProtected);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenProtected___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FrozenProtected___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_Frozen___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Frozen___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_HiddenPartial);
  Py_CLEAR(clear_module_state->__pyx_n_s_HiddenPartial___bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_HiddenPartial___dir);
  Py_CLEAR(clear_module_state->__pyx_n_s_HiddenPartial___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_HiddenPartial___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_kp_s_HiddenPartial_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Mapping);
  Py_CLEAR(clear_module_state->__pyx_n_s_ModuleType);
  Py_CLEAR(clear_module_state->__pyx_n_s_MutableMapping);
  Py_CLEAR(clear_module_state->__pyx_n_s_MutableSequence);
  Py_CLEAR(clear_module_state->__pyx_n_s_MutableSet);
  Py_CLEAR(clear_module_state->__pyx_n_s_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplemented);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Object_Private_s_has_no_attribut);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Object_Protected_s_has_no_attrib);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Object_Wrapped_s_has_no_attribut);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Object___HiddenPartial_has_no_at);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Object_is_read_only);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Object_s_has_no_attribute_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_PYPY);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_items_py2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_iteritems);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_iterkeys);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_itervalues);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_keys_py2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_values_py2);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_viewitems);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_viewkeys);
  Py_CLEAR(clear_module_state->__pyx_n_s_PrivacyDict_viewvalues);
  Py_CLEAR(clear_module_state->__pyx_n_s_Private);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Private_FrozenPrivate_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_Private___dir);
  Py_CLEAR(clear_module_state->__pyx_n_s_Private___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Private___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_Protected_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Protected_FrozenProtected_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_Protected___dir);
  Py_CLEAR(clear_module_state->__pyx_n_s_Protected___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Protected___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProtectionData);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProtectionData___dir);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProtectionData___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProtectionData___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_kp_s_ProtectionData_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProtectionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___aenter);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___aexit);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___complex);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___enter);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___format);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___instancecheck);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___length_hint);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___match_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___round);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___subclasscheck);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy___trunc);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy__aexit);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy__exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_discard);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_extend);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_insert);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_popitem);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Proxy_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_reverse);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_setdefault);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_sort);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proxy_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyPy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_CFunc_object__lParenWrappe);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_CFunc_object__lParenWrappe_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_CFunc_object__lParenWrappe_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Read_only_attribute_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecursionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sequence);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Unknown_OldStyle_Class);
  Py_CLEAR(clear_module_state->__pyx_n_s_Wrapped);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Wrapped_Frozen_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_Wrapped___dir);
  Py_CLEAR(clear_module_state->__pyx_n_s_Wrapped___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Wrapped___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Wrapped_comparator_locals_pass_t);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Wrapped_object_cannot_be_pickled);
  Py_CLEAR(clear_module_state->__pyx_kp_s__17);
  Py_CLEAR(clear_module_state->__pyx_n_s__182);
  Py_CLEAR(clear_module_state->__pyx_n_s__2);
  Py_CLEAR(clear_module_state->__pyx_kp_s__3);
  Py_CLEAR(clear_module_state->__pyx_kp_u__33);
  Py_CLEAR(clear_module_state->__pyx_kp_s__4);
  Py_CLEAR(clear_module_state->__pyx_n_s__44);
  Py_CLEAR(clear_module_state->__pyx_kp_s__88);
  Py_CLEAR(clear_module_state->__pyx_n_s_a);
  Py_CLEAR(clear_module_state->__pyx_kp_s_a_zA_Z_a_zA_Z0_9);
  Py_CLEAR(clear_module_state->__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_abc);
  Py_CLEAR(clear_module_state->__pyx_n_s_abs);
  Py_CLEAR(clear_module_state->__pyx_n_s_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_aenter);
  Py_CLEAR(clear_module_state->__pyx_n_s_aexit);
  Py_CLEAR(clear_module_state->__pyx_n_s_alist);
  Py_CLEAR(clear_module_state->__pyx_n_s_all);
  Py_CLEAR(clear_module_state->__pyx_n_s_always_delegated_attributes);
  Py_CLEAR(clear_module_state->__pyx_n_s_and);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_attr_type_check);
  Py_CLEAR(clear_module_state->__pyx_n_s_attribute_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_basestring);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_build_regex);
  Py_CLEAR(clear_module_state->__pyx_n_s_builtin);
  Py_CLEAR(clear_module_state->__pyx_n_s_builtins);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytearray);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_call);
  Py_CLEAR(clear_module_state->__pyx_n_s_ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_ceil_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_cfunc_to_py);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_cmp);
  Py_CLEAR(clear_module_state->__pyx_n_s_cn);
  Py_CLEAR(clear_module_state->__pyx_n_s_code);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections_abc);
  Py_CLEAR(clear_module_state->__pyx_n_s_compile);
  Py_CLEAR(clear_module_state->__pyx_n_s_complex);
  Py_CLEAR(clear_module_state->__pyx_n_s_complex_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_contains);
  Py_CLEAR(clear_module_state->__pyx_n_s_contains_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_defaults);
  Py_CLEAR(clear_module_state->__pyx_n_s_delattr);
  Py_CLEAR(clear_module_state->__pyx_n_s_delete);
  Py_CLEAR(clear_module_state->__pyx_n_s_delitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_difference);
  Py_CLEAR(clear_module_state->__pyx_n_s_dir);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_discard);
  Py_CLEAR(clear_module_state->__pyx_n_s_divmod);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_dynamic);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_endswith);
  Py_CLEAR(clear_module_state->__pyx_n_s_enter);
  Py_CLEAR(clear_module_state->__pyx_n_s_environ);
  Py_CLEAR(clear_module_state->__pyx_n_s_eq);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_extend);
  Py_CLEAR(clear_module_state->__pyx_n_s_file);
  Py_CLEAR(clear_module_state->__pyx_n_s_float);
  Py_CLEAR(clear_module_state->__pyx_n_s_float_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_floor_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_floordiv);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_freeze);
  Py_CLEAR(clear_module_state->__pyx_n_s_frozen);
  Py_CLEAR(clear_module_state->__pyx_n_s_frozenset);
  Py_CLEAR(clear_module_state->__pyx_n_s_functools);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_ge);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_getattribute);
  Py_CLEAR(clear_module_state->__pyx_n_s_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_getsate);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_kp_s_global_c_functions_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_gt);
  Py_CLEAR(clear_module_state->__pyx_n_s_hash);
  Py_CLEAR(clear_module_state->__pyx_n_s_hash_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_hash_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_hash_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_help);
  Py_CLEAR(clear_module_state->__pyx_n_s_help_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_help_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_help_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_hidden_pickle_attributes);
  Py_CLEAR(clear_module_state->__pyx_n_s_hide);
  Py_CLEAR(clear_module_state->__pyx_n_s_hide_private);
  Py_CLEAR(clear_module_state->__pyx_n_s_hide_regex);
  Py_CLEAR(clear_module_state->__pyx_n_s_iadd);
  Py_CLEAR(clear_module_state->__pyx_n_s_iand);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_id_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_id_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_id_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_ifloordiv);
  Py_CLEAR(clear_module_state->__pyx_n_s_ilshift);
  Py_CLEAR(clear_module_state->__pyx_n_s_imatmul);
  Py_CLEAR(clear_module_state->__pyx_n_s_immutable_builtin_attributes);
  Py_CLEAR(clear_module_state->__pyx_n_s_imod);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_imul);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init___locals_C);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_insert);
  Py_CLEAR(clear_module_state->__pyx_n_s_inst);
  Py_CLEAR(clear_module_state->__pyx_n_s_instance_of_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_instancecheck);
  Py_CLEAR(clear_module_state->__pyx_n_s_instanceof);
  Py_CLEAR(clear_module_state->__pyx_n_s_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_int_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_intersection);
  Py_CLEAR(clear_module_state->__pyx_n_s_invert);
  Py_CLEAR(clear_module_state->__pyx_n_s_ior);
  Py_CLEAR(clear_module_state->__pyx_n_s_ipow);
  Py_CLEAR(clear_module_state->__pyx_n_s_irshift);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_isfrozen);
  Py_CLEAR(clear_module_state->__pyx_n_s_isimmutable);
  Py_CLEAR(clear_module_state->__pyx_n_s_isinstance);
  Py_CLEAR(clear_module_state->__pyx_n_s_isinstance_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_isinstance_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_isprivate);
  Py_CLEAR(clear_module_state->__pyx_n_s_isprotected);
  Py_CLEAR(clear_module_state->__pyx_n_s_isreadonly);
  Py_CLEAR(clear_module_state->__pyx_n_s_issubclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_issubclass_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_issubclass_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_isub);
  Py_CLEAR(clear_module_state->__pyx_n_s_isvisible);
  Py_CLEAR(clear_module_state->__pyx_n_s_iswrapped);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_items_py2);
  Py_CLEAR(clear_module_state->__pyx_n_s_iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_iteritems);
  Py_CLEAR(clear_module_state->__pyx_n_s_iterkeys);
  Py_CLEAR(clear_module_state->__pyx_n_s_itervalues);
  Py_CLEAR(clear_module_state->__pyx_n_s_itruediv);
  Py_CLEAR(clear_module_state->__pyx_n_s_ixor);
  Py_CLEAR(clear_module_state->__pyx_n_s_join);
  Py_CLEAR(clear_module_state->__pyx_n_s_k);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys_py2);
  Py_CLEAR(clear_module_state->__pyx_n_s_kw1);
  Py_CLEAR(clear_module_state->__pyx_n_s_kw2);
  Py_CLEAR(clear_module_state->__pyx_n_s_kwargs);
  Py_CLEAR(clear_module_state->__pyx_n_s_le);
  Py_CLEAR(clear_module_state->__pyx_n_s_len);
  Py_CLEAR(clear_module_state->__pyx_n_s_length_hint);
  Py_CLEAR(clear_module_state->__pyx_n_s_list);
  Py_CLEAR(clear_module_state->__pyx_n_s_long);
  Py_CLEAR(clear_module_state->__pyx_n_s_lshift);
  Py_CLEAR(clear_module_state->__pyx_n_s_lstrip);
  Py_CLEAR(clear_module_state->__pyx_n_s_lt);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_major);
  Py_CLEAR(clear_module_state->__pyx_n_s_match);
  Py_CLEAR(clear_module_state->__pyx_n_s_match_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_math);
  Py_CLEAR(clear_module_state->__pyx_n_s_matmul);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_minor);
  Py_CLEAR(clear_module_state->__pyx_n_s_mod);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_modules);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_mul);
  Py_CLEAR(clear_module_state->__pyx_n_s_multiwrapped);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_ne);
  Py_CLEAR(clear_module_state->__pyx_n_s_neg);
  Py_CLEAR(clear_module_state->__pyx_n_s_never_writeable);
  Py_CLEAR(clear_module_state->__pyx_n_s_never_writeable_private);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_o);
  Py_CLEAR(clear_module_state->__pyx_kp_s_o_Invalid_type_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_oldstyle_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_op);
  Py_CLEAR(clear_module_state->__pyx_n_s_or);
  Py_CLEAR(clear_module_state->__pyx_n_s_os);
  Py_CLEAR(clear_module_state->__pyx_n_s_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_package);
  Py_CLEAR(clear_module_state->__pyx_n_s_partial);
  Py_CLEAR(clear_module_state->__pyx_n_s_pass_to_wrapped);
  Py_CLEAR(clear_module_state->__pyx_n_s_pattern);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_platform);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_popitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_pos);
  Py_CLEAR(clear_module_state->__pyx_n_s_pow);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_private);
  Py_CLEAR(clear_module_state->__pyx_n_s_protect);
  Py_CLEAR(clear_module_state->__pyx_n_s_protected_rules_from_kwargs_loca);
  Py_CLEAR(clear_module_state->__pyx_n_s_py2_function_attrs_rw);
  Py_CLEAR(clear_module_state->__pyx_n_s_pydoc);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyprotect_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_python_implementation);
  Py_CLEAR(clear_module_state->__pyx_kp_s_python_visible_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Frozen);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FrozenPrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FrozenPrivate);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FrozenProtected);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PrivacyDict);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Private);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Proxy);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Wrapped);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle___HiddenPartial);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle___ProtectionData);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_s_r);
  Py_CLEAR(clear_module_state->__pyx_n_s_radd);
  Py_CLEAR(clear_module_state->__pyx_n_s_rand);
  Py_CLEAR(clear_module_state->__pyx_n_s_rdivmod);
  Py_CLEAR(clear_module_state->__pyx_n_s_re);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_render_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_repr);
  Py_CLEAR(clear_module_state->__pyx_n_s_ret);
  Py_CLEAR(clear_module_state->__pyx_n_s_ret_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_reverse);
  Py_CLEAR(clear_module_state->__pyx_n_s_rfloordiv);
  Py_CLEAR(clear_module_state->__pyx_n_s_rl);
  Py_CLEAR(clear_module_state->__pyx_n_s_rlshift);
  Py_CLEAR(clear_module_state->__pyx_n_s_rmatmul);
  Py_CLEAR(clear_module_state->__pyx_n_s_rmod);
  Py_CLEAR(clear_module_state->__pyx_n_s_rmul);
  Py_CLEAR(clear_module_state->__pyx_n_s_ro);
  Py_CLEAR(clear_module_state->__pyx_n_s_ro_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_ro_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_ro_regex);
  Py_CLEAR(clear_module_state->__pyx_n_s_ror);
  Py_CLEAR(clear_module_state->__pyx_n_s_round);
  Py_CLEAR(clear_module_state->__pyx_n_s_rpow);
  Py_CLEAR(clear_module_state->__pyx_n_s_rrshift);
  Py_CLEAR(clear_module_state->__pyx_n_s_rshift);
  Py_CLEAR(clear_module_state->__pyx_n_s_rstrip);
  Py_CLEAR(clear_module_state->__pyx_n_s_rsub);
  Py_CLEAR(clear_module_state->__pyx_n_s_rtruediv);
  Py_CLEAR(clear_module_state->__pyx_n_s_rules);
  Py_CLEAR(clear_module_state->__pyx_n_s_rw);
  Py_CLEAR(clear_module_state->__pyx_n_s_rw_regex);
  Py_CLEAR(clear_module_state->__pyx_n_s_rxor);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s_________0_1);
  Py_CLEAR(clear_module_state->__pyx_n_s_same_class_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_setattr);
  Py_CLEAR(clear_module_state->__pyx_n_s_setdefault);
  Py_CLEAR(clear_module_state->__pyx_n_s_setitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_slots);
  Py_CLEAR(clear_module_state->__pyx_n_s_sort);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_splitlines);
  Py_CLEAR(clear_module_state->__pyx_n_s_startswith);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_str_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_sub);
  Py_CLEAR(clear_module_state->__pyx_n_s_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_subclass_of_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_subclasscheck);
  Py_CLEAR(clear_module_state->__pyx_n_s_subclassof);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_sys);
  Py_CLEAR(clear_module_state->__pyx_n_s_tb);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_testop);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_truediv);
  Py_CLEAR(clear_module_state->__pyx_n_s_trunc);
  Py_CLEAR(clear_module_state->__pyx_n_s_tuple);
  Py_CLEAR(clear_module_state->__pyx_n_s_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_unichr);
  Py_CLEAR(clear_module_state->__pyx_n_s_unicode);
  Py_CLEAR(clear_module_state->__pyx_n_s_union);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_v);
  Py_CLEAR(clear_module_state->__pyx_n_s_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_values_py2);
  Py_CLEAR(clear_module_state->__pyx_n_s_version_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_viewitems);
  Py_CLEAR(clear_module_state->__pyx_n_s_viewkeys);
  Py_CLEAR(clear_module_state->__pyx_n_s_viewvalues);
  Py_CLEAR(clear_module_state->__pyx_n_s_w);
  Py_CLEAR(clear_module_state->__pyx_n_s_want_frozen);
  Py_CLEAR(clear_module_state->__pyx_n_s_weakref);
  Py_CLEAR(clear_module_state->__pyx_n_s_wrap);
  Py_CLEAR(clear_module_state->__pyx_n_s_x);
  Py_CLEAR(clear_module_state->__pyx_n_s_x_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_xor);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_7);
  Py_CLEAR(clear_module_state->__pyx_int_6017409);
  Py_CLEAR(clear_module_state->__pyx_int_19817578);
  Py_CLEAR(clear_module_state->__pyx_int_50167005);
  Py_CLEAR(clear_module_state->__pyx_int_78131249);
  Py_CLEAR(clear_module_state->__pyx_int_115090883);
  Py_CLEAR(clear_module_state->__pyx_int_152356376);
  Py_CLEAR(clear_module_state->__pyx_int_155231502);
  Py_CLEAR(clear_module_state->__pyx_int_161740782);
  Py_CLEAR(clear_module_state->__pyx_int_180490543);
  Py_CLEAR(clear_module_state->__pyx_int_209109265);
  Py_CLEAR(clear_module_state->__pyx_int_210565405);
  Py_CLEAR(clear_module_state->__pyx_int_218598457);
  Py_CLEAR(clear_module_state->__pyx_int_247595846);
  Py_CLEAR(clear_module_state->__pyx_int_252507329);
  Py_CLEAR(clear_module_state->__pyx_int_262487005);
  Py_CLEAR(clear_module_state->__pyx_tuple__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_slice__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__40);
  Py_CLEAR(clear_module_state->__pyx_tuple__42);
  Py_CLEAR(clear_module_state->__pyx_tuple__45);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__56);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__71);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__74);
  Py_CLEAR(clear_module_state->__pyx_tuple__82);
  Py_CLEAR(clear_module_state->__pyx_tuple__83);
  Py_CLEAR(clear_module_state->__pyx_tuple__84);
  Py_CLEAR(clear_module_state->__pyx_tuple__85);
  Py_CLEAR(clear_module_state->__pyx_tuple__86);
  Py_CLEAR(clear_module_state->__pyx_tuple__87);
  Py_CLEAR(clear_module_state->__pyx_tuple__89);
  Py_CLEAR(clear_module_state->__pyx_tuple__90);
  Py_CLEAR(clear_module_state->__pyx_tuple__91);
  Py_CLEAR(clear_module_state->__pyx_tuple__93);
  Py_CLEAR(clear_module_state->__pyx_tuple__95);
  Py_CLEAR(clear_module_state->__pyx_tuple__98);
  Py_CLEAR(clear_module_state->__pyx_codeobj__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__106);
  Py_CLEAR(clear_module_state->__pyx_tuple__108);
  Py_CLEAR(clear_module_state->__pyx_tuple__111);
  Py_CLEAR(clear_module_state->__pyx_tuple__116);
  Py_CLEAR(clear_module_state->__pyx_tuple__119);
  Py_CLEAR(clear_module_state->__pyx_tuple__140);
  Py_CLEAR(clear_module_state->__pyx_tuple__142);
  Py_CLEAR(clear_module_state->__pyx_tuple__143);
  Py_CLEAR(clear_module_state->__pyx_tuple__144);
  Py_CLEAR(clear_module_state->__pyx_tuple__146);
  Py_CLEAR(clear_module_state->__pyx_tuple__170);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__24);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__26);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__28);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__55);
  Py_CLEAR(clear_module_state->__pyx_codeobj__57);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__60);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__65);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__81);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__94);
  Py_CLEAR(clear_module_state->__pyx_codeobj__96);
  Py_CLEAR(clear_module_state->__pyx_codeobj__97);
  Py_CLEAR(clear_module_state->__pyx_codeobj__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__100);
  Py_CLEAR(clear_module_state->__pyx_codeobj__101);
  Py_CLEAR(clear_module_state->__pyx_codeobj__102);
  Py_CLEAR(clear_module_state->__pyx_codeobj__103);
  Py_CLEAR(clear_module_state->__pyx_codeobj__104);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__107);
  Py_CLEAR(clear_module_state->__pyx_codeobj__109);
  Py_CLEAR(clear_module_state->__pyx_codeobj__110);
  Py_CLEAR(clear_module_state->__pyx_codeobj__112);
  Py_CLEAR(clear_module_state->__pyx_codeobj__113);
  Py_CLEAR(clear_module_state->__pyx_codeobj__114);
  Py_CLEAR(clear_module_state->__pyx_codeobj__115);
  Py_CLEAR(clear_module_state->__pyx_codeobj__117);
  Py_CLEAR(clear_module_state->__pyx_codeobj__118);
  Py_CLEAR(clear_module_state->__pyx_codeobj__120);
  Py_CLEAR(clear_module_state->__pyx_codeobj__121);
  Py_CLEAR(clear_module_state->__pyx_codeobj__122);
  Py_CLEAR(clear_module_state->__pyx_codeobj__123);
  Py_CLEAR(clear_module_state->__pyx_codeobj__124);
  Py_CLEAR(clear_module_state->__pyx_codeobj__125);
  Py_CLEAR(clear_module_state->__pyx_codeobj__126);
  Py_CLEAR(clear_module_state->__pyx_codeobj__127);
  Py_CLEAR(clear_module_state->__pyx_codeobj__128);
  Py_CLEAR(clear_module_state->__pyx_codeobj__129);
  Py_CLEAR(clear_module_state->__pyx_codeobj__130);
  Py_CLEAR(clear_module_state->__pyx_codeobj__131);
  Py_CLEAR(clear_module_state->__pyx_codeobj__132);
  Py_CLEAR(clear_module_state->__pyx_codeobj__133);
  Py_CLEAR(clear_module_state->__pyx_codeobj__134);
  Py_CLEAR(clear_module_state->__pyx_codeobj__135);
  Py_CLEAR(clear_module_state->__pyx_codeobj__136);
  Py_CLEAR(clear_module_state->__pyx_codeobj__137);
  Py_CLEAR(clear_module_state->__pyx_codeobj__138);
  Py_CLEAR(clear_module_state->__pyx_codeobj__139);
  Py_CLEAR(clear_module_state->__pyx_codeobj__141);
  Py_CLEAR(clear_module_state->__pyx_codeobj__145);
  Py_CLEAR(clear_module_state->__pyx_codeobj__147);
  Py_CLEAR(clear_module_state->__pyx_codeobj__148);
  Py_CLEAR(clear_module_state->__pyx_codeobj__149);
  Py_CLEAR(clear_module_state->__pyx_codeobj__150);
  Py_CLEAR(clear_module_state->__pyx_codeobj__151);
  Py_CLEAR(clear_module_state->__pyx_codeobj__152);
  Py_CLEAR(clear_module_state->__pyx_codeobj__153);
  Py_CLEAR(clear_module_state->__pyx_codeobj__154);
  Py_CLEAR(clear_module_state->__pyx_codeobj__155);
  Py_CLEAR(clear_module_state->__pyx_codeobj__156);
  Py_CLEAR(clear_module_state->__pyx_codeobj__157);
  Py_CLEAR(clear_module_state->__pyx_codeobj__158);
  Py_CLEAR(clear_module_state->__pyx_codeobj__159);
  Py_CLEAR(clear_module_state->__pyx_codeobj__160);
  Py_CLEAR(clear_module_state->__pyx_codeobj__161);
  Py_CLEAR(clear_module_state->__pyx_codeobj__162);
  Py_CLEAR(clear_module_state->__pyx_codeobj__163);
  Py_CLEAR(clear_module_state->__pyx_codeobj__164);
  Py_CLEAR(clear_module_state->__pyx_codeobj__165);
  Py_CLEAR(clear_module_state->__pyx_codeobj__166);
  Py_CLEAR(clear_module_state->__pyx_codeobj__167);
  Py_CLEAR(clear_module_state->__pyx_codeobj__168);
  Py_CLEAR(clear_module_state->__pyx_codeobj__169);
  Py_CLEAR(clear_module_state->__pyx_codeobj__171);
  Py_CLEAR(clear_module_state->__pyx_codeobj__172);
  Py_CLEAR(clear_module_state->__pyx_codeobj__173);
  Py_CLEAR(clear_module_state->__pyx_codeobj__174);
  Py_CLEAR(clear_module_state->__pyx_codeobj__175);
  Py_CLEAR(clear_module_state->__pyx_codeobj__176);
  Py_CLEAR(clear_module_state->__pyx_codeobj__177);
  Py_CLEAR(clear_module_state->__pyx_codeobj__178);
  Py_CLEAR(clear_module_state->__pyx_codeobj__179);
  Py_CLEAR(clear_module_state->__pyx_codeobj__180);
  Py_CLEAR(clear_module_state->__pyx_codeobj__181);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___ProtectionData);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___ProtectionData);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_Proxy);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_Proxy);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_Wrapped);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_Wrapped);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_Frozen);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_Frozen);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_PrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_PrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_FrozenPrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_Private);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_Private);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_FrozenPrivate);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_FrozenPrivate);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_Protected);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_Protected);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected_FrozenProtected);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected_FrozenProtected);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___HiddenPartial);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___HiddenPartial);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems);
  Py_VISIT(traverse_module_state->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues);
  Py_VISIT(traverse_module_state->__pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues);
  Py_VISIT(traverse_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self);
  Py_VISIT(traverse_module_state->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self);
  Py_VISIT(traverse_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c);
  Py_VISIT(traverse_module_state->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c);
  Py_VISIT(traverse_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op);
  Py_VISIT(traverse_module_state->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op);
  Py_VISIT(traverse_module_state->__pyx_n_s_);
  Py_VISIT(traverse_module_state->__pyx_kp_s_0_1);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_AttributeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_C);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_add_attribute_s_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_delete_attribute_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_delete_attribute_s_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_delete_private_attribute);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_modify_attribute_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_set_attribute_s_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_set_private_attribute_s_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_CollectionsABC);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Double_wrapped);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrameType);
  Py_VISIT(traverse_module_state->__pyx_n_s_Frozen);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenPrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenPrivacyDict___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenPrivacyDict___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenPrivate);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenPrivate___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenPrivate___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenProtected);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenProtected___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FrozenProtected___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_Frozen___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Frozen___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_HiddenPartial);
  Py_VISIT(traverse_module_state->__pyx_n_s_HiddenPartial___bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_HiddenPartial___dir);
  Py_VISIT(traverse_module_state->__pyx_n_s_HiddenPartial___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_HiddenPartial___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_kp_s_HiddenPartial_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Mapping);
  Py_VISIT(traverse_module_state->__pyx_n_s_ModuleType);
  Py_VISIT(traverse_module_state->__pyx_n_s_MutableMapping);
  Py_VISIT(traverse_module_state->__pyx_n_s_MutableSequence);
  Py_VISIT(traverse_module_state->__pyx_n_s_MutableSet);
  Py_VISIT(traverse_module_state->__pyx_n_s_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplemented);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Object_Private_s_has_no_attribut);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Object_Protected_s_has_no_attrib);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Object_Wrapped_s_has_no_attribut);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Object___HiddenPartial_has_no_at);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Object_is_read_only);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Object_s_has_no_attribute_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_PYPY);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_items_py2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_iteritems);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_iterkeys);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_itervalues);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_keys_py2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_values_py2);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_viewitems);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_viewkeys);
  Py_VISIT(traverse_module_state->__pyx_n_s_PrivacyDict_viewvalues);
  Py_VISIT(traverse_module_state->__pyx_n_s_Private);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Private_FrozenPrivate_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_Private___dir);
  Py_VISIT(traverse_module_state->__pyx_n_s_Private___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Private___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_Protected_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Protected_FrozenProtected_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_Protected___dir);
  Py_VISIT(traverse_module_state->__pyx_n_s_Protected___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Protected___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProtectionData);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProtectionData___dir);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProtectionData___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProtectionData___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_kp_s_ProtectionData_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProtectionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___aenter);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___aexit);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___complex);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___enter);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___format);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___instancecheck);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___length_hint);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___match_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___round);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___subclasscheck);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy___trunc);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy__aexit);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy__exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_discard);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_extend);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_insert);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_popitem);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Proxy_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_reverse);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_setdefault);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_sort);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proxy_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyPy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_CFunc_object__lParenWrappe);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_CFunc_object__lParenWrappe_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_CFunc_object__lParenWrappe_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Read_only_attribute_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecursionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sequence);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Unknown_OldStyle_Class);
  Py_VISIT(traverse_module_state->__pyx_n_s_Wrapped);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Wrapped_Frozen_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_Wrapped___dir);
  Py_VISIT(traverse_module_state->__pyx_n_s_Wrapped___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Wrapped___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Wrapped_comparator_locals_pass_t);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Wrapped_object_cannot_be_pickled);
  Py_VISIT(traverse_module_state->__pyx_kp_s__17);
  Py_VISIT(traverse_module_state->__pyx_n_s__182);
  Py_VISIT(traverse_module_state->__pyx_n_s__2);
  Py_VISIT(traverse_module_state->__pyx_kp_s__3);
  Py_VISIT(traverse_module_state->__pyx_kp_u__33);
  Py_VISIT(traverse_module_state->__pyx_kp_s__4);
  Py_VISIT(traverse_module_state->__pyx_n_s__44);
  Py_VISIT(traverse_module_state->__pyx_kp_s__88);
  Py_VISIT(traverse_module_state->__pyx_n_s_a);
  Py_VISIT(traverse_module_state->__pyx_kp_s_a_zA_Z_a_zA_Z0_9);
  Py_VISIT(traverse_module_state->__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_abc);
  Py_VISIT(traverse_module_state->__pyx_n_s_abs);
  Py_VISIT(traverse_module_state->__pyx_n_s_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_aenter);
  Py_VISIT(traverse_module_state->__pyx_n_s_aexit);
  Py_VISIT(traverse_module_state->__pyx_n_s_alist);
  Py_VISIT(traverse_module_state->__pyx_n_s_all);
  Py_VISIT(traverse_module_state->__pyx_n_s_always_delegated_attributes);
  Py_VISIT(traverse_module_state->__pyx_n_s_and);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_attr_type_check);
  Py_VISIT(traverse_module_state->__pyx_n_s_attribute_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_basestring);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_build_regex);
  Py_VISIT(traverse_module_state->__pyx_n_s_builtin);
  Py_VISIT(traverse_module_state->__pyx_n_s_builtins);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytearray);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_call);
  Py_VISIT(traverse_module_state->__pyx_n_s_ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_ceil_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_cfunc_to_py);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_cmp);
  Py_VISIT(traverse_module_state->__pyx_n_s_cn);
  Py_VISIT(traverse_module_state->__pyx_n_s_code);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections_abc);
  Py_VISIT(traverse_module_state->__pyx_n_s_compile);
  Py_VISIT(traverse_module_state->__pyx_n_s_complex);
  Py_VISIT(traverse_module_state->__pyx_n_s_complex_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_contains);
  Py_VISIT(traverse_module_state->__pyx_n_s_contains_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_defaults);
  Py_VISIT(traverse_module_state->__pyx_n_s_delattr);
  Py_VISIT(traverse_module_state->__pyx_n_s_delete);
  Py_VISIT(traverse_module_state->__pyx_n_s_delitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_difference);
  Py_VISIT(traverse_module_state->__pyx_n_s_dir);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_discard);
  Py_VISIT(traverse_module_state->__pyx_n_s_divmod);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_dynamic);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_endswith);
  Py_VISIT(traverse_module_state->__pyx_n_s_enter);
  Py_VISIT(traverse_module_state->__pyx_n_s_environ);
  Py_VISIT(traverse_module_state->__pyx_n_s_eq);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_extend);
  Py_VISIT(traverse_module_state->__pyx_n_s_file);
  Py_VISIT(traverse_module_state->__pyx_n_s_float);
  Py_VISIT(traverse_module_state->__pyx_n_s_float_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_floor_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_floordiv);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_freeze);
  Py_VISIT(traverse_module_state->__pyx_n_s_frozen);
  Py_VISIT(traverse_module_state->__pyx_n_s_frozenset);
  Py_VISIT(traverse_module_state->__pyx_n_s_functools);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_ge);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_getattribute);
  Py_VISIT(traverse_module_state->__pyx_n_s_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_getsate);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_kp_s_global_c_functions_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_gt);
  Py_VISIT(traverse_module_state->__pyx_n_s_hash);
  Py_VISIT(traverse_module_state->__pyx_n_s_hash_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_hash_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_hash_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_help);
  Py_VISIT(traverse_module_state->__pyx_n_s_help_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_help_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_help_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_hidden_pickle_attributes);
  Py_VISIT(traverse_module_state->__pyx_n_s_hide);
  Py_VISIT(traverse_module_state->__pyx_n_s_hide_private);
  Py_VISIT(traverse_module_state->__pyx_n_s_hide_regex);
  Py_VISIT(traverse_module_state->__pyx_n_s_iadd);
  Py_VISIT(traverse_module_state->__pyx_n_s_iand);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_id_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_id_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_id_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_ifloordiv);
  Py_VISIT(traverse_module_state->__pyx_n_s_ilshift);
  Py_VISIT(traverse_module_state->__pyx_n_s_imatmul);
  Py_VISIT(traverse_module_state->__pyx_n_s_immutable_builtin_attributes);
  Py_VISIT(traverse_module_state->__pyx_n_s_imod);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_imul);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init___locals_C);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_insert);
  Py_VISIT(traverse_module_state->__pyx_n_s_inst);
  Py_VISIT(traverse_module_state->__pyx_n_s_instance_of_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_instancecheck);
  Py_VISIT(traverse_module_state->__pyx_n_s_instanceof);
  Py_VISIT(traverse_module_state->__pyx_n_s_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_int_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_intersection);
  Py_VISIT(traverse_module_state->__pyx_n_s_invert);
  Py_VISIT(traverse_module_state->__pyx_n_s_ior);
  Py_VISIT(traverse_module_state->__pyx_n_s_ipow);
  Py_VISIT(traverse_module_state->__pyx_n_s_irshift);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_isfrozen);
  Py_VISIT(traverse_module_state->__pyx_n_s_isimmutable);
  Py_VISIT(traverse_module_state->__pyx_n_s_isinstance);
  Py_VISIT(traverse_module_state->__pyx_n_s_isinstance_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_isinstance_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_isprivate);
  Py_VISIT(traverse_module_state->__pyx_n_s_isprotected);
  Py_VISIT(traverse_module_state->__pyx_n_s_isreadonly);
  Py_VISIT(traverse_module_state->__pyx_n_s_issubclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_issubclass_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_issubclass_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_isub);
  Py_VISIT(traverse_module_state->__pyx_n_s_isvisible);
  Py_VISIT(traverse_module_state->__pyx_n_s_iswrapped);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_items_py2);
  Py_VISIT(traverse_module_state->__pyx_n_s_iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_iteritems);
  Py_VISIT(traverse_module_state->__pyx_n_s_iterkeys);
  Py_VISIT(traverse_module_state->__pyx_n_s_itervalues);
  Py_VISIT(traverse_module_state->__pyx_n_s_itruediv);
  Py_VISIT(traverse_module_state->__pyx_n_s_ixor);
  Py_VISIT(traverse_module_state->__pyx_n_s_join);
  Py_VISIT(traverse_module_state->__pyx_n_s_k);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys_py2);
  Py_VISIT(traverse_module_state->__pyx_n_s_kw1);
  Py_VISIT(traverse_module_state->__pyx_n_s_kw2);
  Py_VISIT(traverse_module_state->__pyx_n_s_kwargs);
  Py_VISIT(traverse_module_state->__pyx_n_s_le);
  Py_VISIT(traverse_module_state->__pyx_n_s_len);
  Py_VISIT(traverse_module_state->__pyx_n_s_length_hint);
  Py_VISIT(traverse_module_state->__pyx_n_s_list);
  Py_VISIT(traverse_module_state->__pyx_n_s_long);
  Py_VISIT(traverse_module_state->__pyx_n_s_lshift);
  Py_VISIT(traverse_module_state->__pyx_n_s_lstrip);
  Py_VISIT(traverse_module_state->__pyx_n_s_lt);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_major);
  Py_VISIT(traverse_module_state->__pyx_n_s_match);
  Py_VISIT(traverse_module_state->__pyx_n_s_match_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_math);
  Py_VISIT(traverse_module_state->__pyx_n_s_matmul);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_minor);
  Py_VISIT(traverse_module_state->__pyx_n_s_mod);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_modules);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_mul);
  Py_VISIT(traverse_module_state->__pyx_n_s_multiwrapped);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_ne);
  Py_VISIT(traverse_module_state->__pyx_n_s_neg);
  Py_VISIT(traverse_module_state->__pyx_n_s_never_writeable);
  Py_VISIT(traverse_module_state->__pyx_n_s_never_writeable_private);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_o);
  Py_VISIT(traverse_module_state->__pyx_kp_s_o_Invalid_type_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_oldstyle_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_op);
  Py_VISIT(traverse_module_state->__pyx_n_s_or);
  Py_VISIT(traverse_module_state->__pyx_n_s_os);
  Py_VISIT(traverse_module_state->__pyx_n_s_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_package);
  Py_VISIT(traverse_module_state->__pyx_n_s_partial);
  Py_VISIT(traverse_module_state->__pyx_n_s_pass_to_wrapped);
  Py_VISIT(traverse_module_state->__pyx_n_s_pattern);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_platform);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_popitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_pos);
  Py_VISIT(traverse_module_state->__pyx_n_s_pow);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_private);
  Py_VISIT(traverse_module_state->__pyx_n_s_protect);
  Py_VISIT(traverse_module_state->__pyx_n_s_protected_rules_from_kwargs_loca);
  Py_VISIT(traverse_module_state->__pyx_n_s_py2_function_attrs_rw);
  Py_VISIT(traverse_module_state->__pyx_n_s_pydoc);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyprotect_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_python_implementation);
  Py_VISIT(traverse_module_state->__pyx_kp_s_python_visible_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Frozen);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FrozenPrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FrozenPrivate);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FrozenProtected);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PrivacyDict);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Private);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Proxy);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Wrapped);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle___HiddenPartial);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle___ProtectionData);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_s_r);
  Py_VISIT(traverse_module_state->__pyx_n_s_radd);
  Py_VISIT(traverse_module_state->__pyx_n_s_rand);
  Py_VISIT(traverse_module_state->__pyx_n_s_rdivmod);
  Py_VISIT(traverse_module_state->__pyx_n_s_re);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_render_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_repr);
  Py_VISIT(traverse_module_state->__pyx_n_s_ret);
  Py_VISIT(traverse_module_state->__pyx_n_s_ret_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_reverse);
  Py_VISIT(traverse_module_state->__pyx_n_s_rfloordiv);
  Py_VISIT(traverse_module_state->__pyx_n_s_rl);
  Py_VISIT(traverse_module_state->__pyx_n_s_rlshift);
  Py_VISIT(traverse_module_state->__pyx_n_s_rmatmul);
  Py_VISIT(traverse_module_state->__pyx_n_s_rmod);
  Py_VISIT(traverse_module_state->__pyx_n_s_rmul);
  Py_VISIT(traverse_module_state->__pyx_n_s_ro);
  Py_VISIT(traverse_module_state->__pyx_n_s_ro_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_ro_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_ro_regex);
  Py_VISIT(traverse_module_state->__pyx_n_s_ror);
  Py_VISIT(traverse_module_state->__pyx_n_s_round);
  Py_VISIT(traverse_module_state->__pyx_n_s_rpow);
  Py_VISIT(traverse_module_state->__pyx_n_s_rrshift);
  Py_VISIT(traverse_module_state->__pyx_n_s_rshift);
  Py_VISIT(traverse_module_state->__pyx_n_s_rstrip);
  Py_VISIT(traverse_module_state->__pyx_n_s_rsub);
  Py_VISIT(traverse_module_state->__pyx_n_s_rtruediv);
  Py_VISIT(traverse_module_state->__pyx_n_s_rules);
  Py_VISIT(traverse_module_state->__pyx_n_s_rw);
  Py_VISIT(traverse_module_state->__pyx_n_s_rw_regex);
  Py_VISIT(traverse_module_state->__pyx_n_s_rxor);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s_________0_1);
  Py_VISIT(traverse_module_state->__pyx_n_s_same_class_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_setattr);
  Py_VISIT(traverse_module_state->__pyx_n_s_setdefault);
  Py_VISIT(traverse_module_state->__pyx_n_s_setitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_slots);
  Py_VISIT(traverse_module_state->__pyx_n_s_sort);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_splitlines);
  Py_VISIT(traverse_module_state->__pyx_n_s_startswith);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_str_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_sub);
  Py_VISIT(traverse_module_state->__pyx_n_s_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_subclass_of_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_subclasscheck);
  Py_VISIT(traverse_module_state->__pyx_n_s_subclassof);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_sys);
  Py_VISIT(traverse_module_state->__pyx_n_s_tb);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_testop);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_truediv);
  Py_VISIT(traverse_module_state->__pyx_n_s_trunc);
  Py_VISIT(traverse_module_state->__pyx_n_s_tuple);
  Py_VISIT(traverse_module_state->__pyx_n_s_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_unichr);
  Py_VISIT(traverse_module_state->__pyx_n_s_unicode);
  Py_VISIT(traverse_module_state->__pyx_n_s_union);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_v);
  Py_VISIT(traverse_module_state->__pyx_n_s_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_values_py2);
  Py_VISIT(traverse_module_state->__pyx_n_s_version_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_viewitems);
  Py_VISIT(traverse_module_state->__pyx_n_s_viewkeys);
  Py_VISIT(traverse_module_state->__pyx_n_s_viewvalues);
  Py_VISIT(traverse_module_state->__pyx_n_s_w);
  Py_VISIT(traverse_module_state->__pyx_n_s_want_frozen);
  Py_VISIT(traverse_module_state->__pyx_n_s_weakref);
  Py_VISIT(traverse_module_state->__pyx_n_s_wrap);
  Py_VISIT(traverse_module_state->__pyx_n_s_x);
  Py_VISIT(traverse_module_state->__pyx_n_s_x_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_xor);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_7);
  Py_VISIT(traverse_module_state->__pyx_int_6017409);
  Py_VISIT(traverse_module_state->__pyx_int_19817578);
  Py_VISIT(traverse_module_state->__pyx_int_50167005);
  Py_VISIT(traverse_module_state->__pyx_int_78131249);
  Py_VISIT(traverse_module_state->__pyx_int_115090883);
  Py_VISIT(traverse_module_state->__pyx_int_152356376);
  Py_VISIT(traverse_module_state->__pyx_int_155231502);
  Py_VISIT(traverse_module_state->__pyx_int_161740782);
  Py_VISIT(traverse_module_state->__pyx_int_180490543);
  Py_VISIT(traverse_module_state->__pyx_int_209109265);
  Py_VISIT(traverse_module_state->__pyx_int_210565405);
  Py_VISIT(traverse_module_state->__pyx_int_218598457);
  Py_VISIT(traverse_module_state->__pyx_int_247595846);
  Py_VISIT(traverse_module_state->__pyx_int_252507329);
  Py_VISIT(traverse_module_state->__pyx_int_262487005);
  Py_VISIT(traverse_module_state->__pyx_tuple__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_slice__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__40);
  Py_VISIT(traverse_module_state->__pyx_tuple__42);
  Py_VISIT(traverse_module_state->__pyx_tuple__45);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__56);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__71);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__74);
  Py_VISIT(traverse_module_state->__pyx_tuple__82);
  Py_VISIT(traverse_module_state->__pyx_tuple__83);
  Py_VISIT(traverse_module_state->__pyx_tuple__84);
  Py_VISIT(traverse_module_state->__pyx_tuple__85);
  Py_VISIT(traverse_module_state->__pyx_tuple__86);
  Py_VISIT(traverse_module_state->__pyx_tuple__87);
  Py_VISIT(traverse_module_state->__pyx_tuple__89);
  Py_VISIT(traverse_module_state->__pyx_tuple__90);
  Py_VISIT(traverse_module_state->__pyx_tuple__91);
  Py_VISIT(traverse_module_state->__pyx_tuple__93);
  Py_VISIT(traverse_module_state->__pyx_tuple__95);
  Py_VISIT(traverse_module_state->__pyx_tuple__98);
  Py_VISIT(traverse_module_state->__pyx_codeobj__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__106);
  Py_VISIT(traverse_module_state->__pyx_tuple__108);
  Py_VISIT(traverse_module_state->__pyx_tuple__111);
  Py_VISIT(traverse_module_state->__pyx_tuple__116);
  Py_VISIT(traverse_module_state->__pyx_tuple__119);
  Py_VISIT(traverse_module_state->__pyx_tuple__140);
  Py_VISIT(traverse_module_state->__pyx_tuple__142);
  Py_VISIT(traverse_module_state->__pyx_tuple__143);
  Py_VISIT(traverse_module_state->__pyx_tuple__144);
  Py_VISIT(traverse_module_state->__pyx_tuple__146);
  Py_VISIT(traverse_module_state->__pyx_tuple__170);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__24);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__26);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__28);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__55);
  Py_VISIT(traverse_module_state->__pyx_codeobj__57);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__60);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__65);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__81);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__94);
  Py_VISIT(traverse_module_state->__pyx_codeobj__96);
  Py_VISIT(traverse_module_state->__pyx_codeobj__97);
  Py_VISIT(traverse_module_state->__pyx_codeobj__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__100);
  Py_VISIT(traverse_module_state->__pyx_codeobj__101);
  Py_VISIT(traverse_module_state->__pyx_codeobj__102);
  Py_VISIT(traverse_module_state->__pyx_codeobj__103);
  Py_VISIT(traverse_module_state->__pyx_codeobj__104);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__107);
  Py_VISIT(traverse_module_state->__pyx_codeobj__109);
  Py_VISIT(traverse_module_state->__pyx_codeobj__110);
  Py_VISIT(traverse_module_state->__pyx_codeobj__112);
  Py_VISIT(traverse_module_state->__pyx_codeobj__113);
  Py_VISIT(traverse_module_state->__pyx_codeobj__114);
  Py_VISIT(traverse_module_state->__pyx_codeobj__115);
  Py_VISIT(traverse_module_state->__pyx_codeobj__117);
  Py_VISIT(traverse_module_state->__pyx_codeobj__118);
  Py_VISIT(traverse_module_state->__pyx_codeobj__120);
  Py_VISIT(traverse_module_state->__pyx_codeobj__121);
  Py_VISIT(traverse_module_state->__pyx_codeobj__122);
  Py_VISIT(traverse_module_state->__pyx_codeobj__123);
  Py_VISIT(traverse_module_state->__pyx_codeobj__124);
  Py_VISIT(traverse_module_state->__pyx_codeobj__125);
  Py_VISIT(traverse_module_state->__pyx_codeobj__126);
  Py_VISIT(traverse_module_state->__pyx_codeobj__127);
  Py_VISIT(traverse_module_state->__pyx_codeobj__128);
  Py_VISIT(traverse_module_state->__pyx_codeobj__129);
  Py_VISIT(traverse_module_state->__pyx_codeobj__130);
  Py_VISIT(traverse_module_state->__pyx_codeobj__131);
  Py_VISIT(traverse_module_state->__pyx_codeobj__132);
  Py_VISIT(traverse_module_state->__pyx_codeobj__133);
  Py_VISIT(traverse_module_state->__pyx_codeobj__134);
  Py_VISIT(traverse_module_state->__pyx_codeobj__135);
  Py_VISIT(traverse_module_state->__pyx_codeobj__136);
  Py_VISIT(traverse_module_state->__pyx_codeobj__137);
  Py_VISIT(traverse_module_state->__pyx_codeobj__138);
  Py_VISIT(traverse_module_state->__pyx_codeobj__139);
  Py_VISIT(traverse_module_state->__pyx_codeobj__141);
  Py_VISIT(traverse_module_state->__pyx_codeobj__145);
  Py_VISIT(traverse_module_state->__pyx_codeobj__147);
  Py_VISIT(traverse_module_state->__pyx_codeobj__148);
  Py_VISIT(traverse_module_state->__pyx_codeobj__149);
  Py_VISIT(traverse_module_state->__pyx_codeobj__150);
  Py_VISIT(traverse_module_state->__pyx_codeobj__151);
  Py_VISIT(traverse_module_state->__pyx_codeobj__152);
  Py_VISIT(traverse_module_state->__pyx_codeobj__153);
  Py_VISIT(traverse_module_state->__pyx_codeobj__154);
  Py_VISIT(traverse_module_state->__pyx_codeobj__155);
  Py_VISIT(traverse_module_state->__pyx_codeobj__156);
  Py_VISIT(traverse_module_state->__pyx_codeobj__157);
  Py_VISIT(traverse_module_state->__pyx_codeobj__158);
  Py_VISIT(traverse_module_state->__pyx_codeobj__159);
  Py_VISIT(traverse_module_state->__pyx_codeobj__160);
  Py_VISIT(traverse_module_state->__pyx_codeobj__161);
  Py_VISIT(traverse_module_state->__pyx_codeobj__162);
  Py_VISIT(traverse_module_state->__pyx_codeobj__163);
  Py_VISIT(traverse_module_state->__pyx_codeobj__164);
  Py_VISIT(traverse_module_state->__pyx_codeobj__165);
  Py_VISIT(traverse_module_state->__pyx_codeobj__166);
  Py_VISIT(traverse_module_state->__pyx_codeobj__167);
  Py_VISIT(traverse_module_state->__pyx_codeobj__168);
  Py_VISIT(traverse_module_state->__pyx_codeobj__169);
  Py_VISIT(traverse_module_state->__pyx_codeobj__171);
  Py_VISIT(traverse_module_state->__pyx_codeobj__172);
  Py_VISIT(traverse_module_state->__pyx_codeobj__173);
  Py_VISIT(traverse_module_state->__pyx_codeobj__174);
  Py_VISIT(traverse_module_state->__pyx_codeobj__175);
  Py_VISIT(traverse_module_state->__pyx_codeobj__176);
  Py_VISIT(traverse_module_state->__pyx_codeobj__177);
  Py_VISIT(traverse_module_state->__pyx_codeobj__178);
  Py_VISIT(traverse_module_state->__pyx_codeobj__179);
  Py_VISIT(traverse_module_state->__pyx_codeobj__180);
  Py_VISIT(traverse_module_state->__pyx_codeobj__181);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#if CYTHON_USE_MODULE_STATE
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#define __pyx_ptype_9pyprotect_9protected___ProtectionData __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___ProtectionData
#define __pyx_type_9pyprotect_9protected___ProtectionData __pyx_mstate_global->__pyx_type_9pyprotect_9protected___ProtectionData
#define __pyx_ptype_9pyprotect_9protected_Proxy __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_Proxy
#define __pyx_type_9pyprotect_9protected_Proxy __pyx_mstate_global->__pyx_type_9pyprotect_9protected_Proxy
#define __pyx_ptype_9pyprotect_9protected_Wrapped __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_Wrapped
#define __pyx_type_9pyprotect_9protected_Wrapped __pyx_mstate_global->__pyx_type_9pyprotect_9protected_Wrapped
#define __pyx_ptype_9pyprotect_9protected_Frozen __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_Frozen
#define __pyx_type_9pyprotect_9protected_Frozen __pyx_mstate_global->__pyx_type_9pyprotect_9protected_Frozen
#define __pyx_ptype_9pyprotect_9protected_PrivacyDict __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_PrivacyDict
#define __pyx_type_9pyprotect_9protected_PrivacyDict __pyx_mstate_global->__pyx_type_9pyprotect_9protected_PrivacyDict
#define __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict
#define __pyx_type_9pyprotect_9protected_FrozenPrivacyDict __pyx_mstate_global->__pyx_type_9pyprotect_9protected_FrozenPrivacyDict
#define __pyx_ptype_9pyprotect_9protected_Private __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_Private
#define __pyx_type_9pyprotect_9protected_Private __pyx_mstate_global->__pyx_type_9pyprotect_9protected_Private
#define __pyx_ptype_9pyprotect_9protected_FrozenPrivate __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_FrozenPrivate
#define __pyx_type_9pyprotect_9protected_FrozenPrivate __pyx_mstate_global->__pyx_type_9pyprotect_9protected_FrozenPrivate
#define __pyx_ptype_9pyprotect_9protected_Protected __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_Protected
#define __pyx_type_9pyprotect_9protected_Protected __pyx_mstate_global->__pyx_type_9pyprotect_9protected_Protected
#define __pyx_ptype_9pyprotect_9protected_FrozenProtected __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected_FrozenProtected
#define __pyx_type_9pyprotect_9protected_FrozenProtected __pyx_mstate_global->__pyx_type_9pyprotect_9protected_FrozenProtected
#define __pyx_ptype_9pyprotect_9protected___HiddenPartial __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___HiddenPartial
#define __pyx_type_9pyprotect_9protected___HiddenPartial __pyx_mstate_global->__pyx_type_9pyprotect_9protected___HiddenPartial
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__ __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct____iter__ __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct____iter__
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems
#define __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues __pyx_mstate_global->__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues
#define __pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues __pyx_mstate_global->__pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues
#define __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self __pyx_mstate_global->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self
#define __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self __pyx_mstate_global->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self
#define __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c __pyx_mstate_global->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c
#define __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c __pyx_mstate_global->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c
#define __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op __pyx_mstate_global->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op
#define __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op __pyx_mstate_global->__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op
#define __pyx_n_s_ __pyx_mstate_global->__pyx_n_s_
#define __pyx_kp_s_0_1 __pyx_mstate_global->__pyx_kp_s_0_1
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_AttributeError __pyx_mstate_global->__pyx_n_s_AttributeError
#define __pyx_n_s_C __pyx_mstate_global->__pyx_n_s_C
#define __pyx_kp_s_Cannot_add_attribute_s_s __pyx_mstate_global->__pyx_kp_s_Cannot_add_attribute_s_s
#define __pyx_kp_s_Cannot_delete_attribute_s __pyx_mstate_global->__pyx_kp_s_Cannot_delete_attribute_s
#define __pyx_kp_s_Cannot_delete_attribute_s_s __pyx_mstate_global->__pyx_kp_s_Cannot_delete_attribute_s_s
#define __pyx_kp_s_Cannot_delete_private_attribute __pyx_mstate_global->__pyx_kp_s_Cannot_delete_private_attribute
#define __pyx_kp_s_Cannot_modify_attribute_s __pyx_mstate_global->__pyx_kp_s_Cannot_modify_attribute_s
#define __pyx_kp_s_Cannot_set_attribute_s_s __pyx_mstate_global->__pyx_kp_s_Cannot_set_attribute_s_s
#define __pyx_kp_s_Cannot_set_private_attribute_s_s __pyx_mstate_global->__pyx_kp_s_Cannot_set_private_attribute_s_s
#define __pyx_n_s_CollectionsABC __pyx_mstate_global->__pyx_n_s_CollectionsABC
#define __pyx_kp_s_Double_wrapped __pyx_mstate_global->__pyx_kp_s_Double_wrapped
#define __pyx_n_s_FrameType __pyx_mstate_global->__pyx_n_s_FrameType
#define __pyx_n_s_Frozen __pyx_mstate_global->__pyx_n_s_Frozen
#define __pyx_n_s_FrozenPrivacyDict __pyx_mstate_global->__pyx_n_s_FrozenPrivacyDict
#define __pyx_n_s_FrozenPrivacyDict___reduce_cytho __pyx_mstate_global->__pyx_n_s_FrozenPrivacyDict___reduce_cytho
#define __pyx_n_s_FrozenPrivacyDict___setstate_cyt __pyx_mstate_global->__pyx_n_s_FrozenPrivacyDict___setstate_cyt
#define __pyx_n_s_FrozenPrivate __pyx_mstate_global->__pyx_n_s_FrozenPrivate
#define __pyx_n_s_FrozenPrivate___reduce_cython __pyx_mstate_global->__pyx_n_s_FrozenPrivate___reduce_cython
#define __pyx_n_s_FrozenPrivate___setstate_cython __pyx_mstate_global->__pyx_n_s_FrozenPrivate___setstate_cython
#define __pyx_n_s_FrozenProtected __pyx_mstate_global->__pyx_n_s_FrozenProtected
#define __pyx_n_s_FrozenProtected___reduce_cython __pyx_mstate_global->__pyx_n_s_FrozenProtected___reduce_cython
#define __pyx_n_s_FrozenProtected___setstate_cytho __pyx_mstate_global->__pyx_n_s_FrozenProtected___setstate_cytho
#define __pyx_n_s_Frozen___reduce_cython __pyx_mstate_global->__pyx_n_s_Frozen___reduce_cython
#define __pyx_n_s_Frozen___setstate_cython __pyx_mstate_global->__pyx_n_s_Frozen___setstate_cython
#define __pyx_n_s_HiddenPartial __pyx_mstate_global->__pyx_n_s_HiddenPartial
#define __pyx_n_s_HiddenPartial___bytes __pyx_mstate_global->__pyx_n_s_HiddenPartial___bytes
#define __pyx_n_s_HiddenPartial___dir __pyx_mstate_global->__pyx_n_s_HiddenPartial___dir
#define __pyx_n_s_HiddenPartial___reduce_cython __pyx_mstate_global->__pyx_n_s_HiddenPartial___reduce_cython
#define __pyx_n_s_HiddenPartial___setstate_cytho __pyx_mstate_global->__pyx_n_s_HiddenPartial___setstate_cytho
#define __pyx_kp_s_HiddenPartial_pxi __pyx_mstate_global->__pyx_kp_s_HiddenPartial_pxi
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_Mapping __pyx_mstate_global->__pyx_n_s_Mapping
#define __pyx_n_s_ModuleType __pyx_mstate_global->__pyx_n_s_ModuleType
#define __pyx_n_s_MutableMapping __pyx_mstate_global->__pyx_n_s_MutableMapping
#define __pyx_n_s_MutableSequence __pyx_mstate_global->__pyx_n_s_MutableSequence
#define __pyx_n_s_MutableSet __pyx_mstate_global->__pyx_n_s_MutableSet
#define __pyx_n_s_None __pyx_mstate_global->__pyx_n_s_None
#define __pyx_n_s_NotImplemented __pyx_mstate_global->__pyx_n_s_NotImplemented
#define __pyx_kp_s_Object_Private_s_has_no_attribut __pyx_mstate_global->__pyx_kp_s_Object_Private_s_has_no_attribut
#define __pyx_kp_s_Object_Protected_s_has_no_attrib __pyx_mstate_global->__pyx_kp_s_Object_Protected_s_has_no_attrib
#define __pyx_kp_s_Object_Wrapped_s_has_no_attribut __pyx_mstate_global->__pyx_kp_s_Object_Wrapped_s_has_no_attribut
#define __pyx_kp_s_Object___HiddenPartial_has_no_at __pyx_mstate_global->__pyx_kp_s_Object___HiddenPartial_has_no_at
#define __pyx_kp_s_Object_is_read_only __pyx_mstate_global->__pyx_kp_s_Object_is_read_only
#define __pyx_kp_s_Object_s_has_no_attribute_s __pyx_mstate_global->__pyx_kp_s_Object_s_has_no_attribute_s
#define __pyx_n_s_PYPY __pyx_mstate_global->__pyx_n_s_PYPY
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_n_s_PrivacyDict __pyx_mstate_global->__pyx_n_s_PrivacyDict
#define __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px __pyx_mstate_global->__pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px
#define __pyx_n_s_PrivacyDict___reduce_cython __pyx_mstate_global->__pyx_n_s_PrivacyDict___reduce_cython
#define __pyx_n_s_PrivacyDict___setstate_cython __pyx_mstate_global->__pyx_n_s_PrivacyDict___setstate_cython
#define __pyx_n_s_PrivacyDict_copy __pyx_mstate_global->__pyx_n_s_PrivacyDict_copy
#define __pyx_n_s_PrivacyDict_items __pyx_mstate_global->__pyx_n_s_PrivacyDict_items
#define __pyx_n_s_PrivacyDict_items_py2 __pyx_mstate_global->__pyx_n_s_PrivacyDict_items_py2
#define __pyx_n_s_PrivacyDict_iteritems __pyx_mstate_global->__pyx_n_s_PrivacyDict_iteritems
#define __pyx_n_s_PrivacyDict_iterkeys __pyx_mstate_global->__pyx_n_s_PrivacyDict_iterkeys
#define __pyx_n_s_PrivacyDict_itervalues __pyx_mstate_global->__pyx_n_s_PrivacyDict_itervalues
#define __pyx_n_s_PrivacyDict_keys __pyx_mstate_global->__pyx_n_s_PrivacyDict_keys
#define __pyx_n_s_PrivacyDict_keys_py2 __pyx_mstate_global->__pyx_n_s_PrivacyDict_keys_py2
#define __pyx_n_s_PrivacyDict_values __pyx_mstate_global->__pyx_n_s_PrivacyDict_values
#define __pyx_n_s_PrivacyDict_values_py2 __pyx_mstate_global->__pyx_n_s_PrivacyDict_values_py2
#define __pyx_n_s_PrivacyDict_viewitems __pyx_mstate_global->__pyx_n_s_PrivacyDict_viewitems
#define __pyx_n_s_PrivacyDict_viewkeys __pyx_mstate_global->__pyx_n_s_PrivacyDict_viewkeys
#define __pyx_n_s_PrivacyDict_viewvalues __pyx_mstate_global->__pyx_n_s_PrivacyDict_viewvalues
#define __pyx_n_s_Private __pyx_mstate_global->__pyx_n_s_Private
#define __pyx_kp_s_Private_FrozenPrivate_pxi __pyx_mstate_global->__pyx_kp_s_Private_FrozenPrivate_pxi
#define __pyx_n_s_Private___dir __pyx_mstate_global->__pyx_n_s_Private___dir
#define __pyx_n_s_Private___reduce_cython __pyx_mstate_global->__pyx_n_s_Private___reduce_cython
#define __pyx_n_s_Private___setstate_cython __pyx_mstate_global->__pyx_n_s_Private___setstate_cython
#define __pyx_n_s_Protected __pyx_mstate_global->__pyx_n_s_Protected
#define __pyx_n_s_Protected_2 __pyx_mstate_global->__pyx_n_s_Protected_2
#define __pyx_kp_s_Protected_FrozenProtected_pxi __pyx_mstate_global->__pyx_kp_s_Protected_FrozenProtected_pxi
#define __pyx_n_s_Protected___dir __pyx_mstate_global->__pyx_n_s_Protected___dir
#define __pyx_n_s_Protected___reduce_cython __pyx_mstate_global->__pyx_n_s_Protected___reduce_cython
#define __pyx_n_s_Protected___setstate_cython __pyx_mstate_global->__pyx_n_s_Protected___setstate_cython
#define __pyx_n_s_ProtectionData __pyx_mstate_global->__pyx_n_s_ProtectionData
#define __pyx_n_s_ProtectionData___dir __pyx_mstate_global->__pyx_n_s_ProtectionData___dir
#define __pyx_n_s_ProtectionData___reduce_cython __pyx_mstate_global->__pyx_n_s_ProtectionData___reduce_cython
#define __pyx_n_s_ProtectionData___setstate_cyth __pyx_mstate_global->__pyx_n_s_ProtectionData___setstate_cyth
#define __pyx_kp_s_ProtectionData_pxi __pyx_mstate_global->__pyx_kp_s_ProtectionData_pxi
#define __pyx_n_s_ProtectionError __pyx_mstate_global->__pyx_n_s_ProtectionError
#define __pyx_n_s_Proxy __pyx_mstate_global->__pyx_n_s_Proxy
#define __pyx_n_s_Proxy___aenter __pyx_mstate_global->__pyx_n_s_Proxy___aenter
#define __pyx_n_s_Proxy___aexit __pyx_mstate_global->__pyx_n_s_Proxy___aexit
#define __pyx_n_s_Proxy___bytes __pyx_mstate_global->__pyx_n_s_Proxy___bytes
#define __pyx_n_s_Proxy___ceil __pyx_mstate_global->__pyx_n_s_Proxy___ceil
#define __pyx_n_s_Proxy___complex __pyx_mstate_global->__pyx_n_s_Proxy___complex
#define __pyx_n_s_Proxy___enter __pyx_mstate_global->__pyx_n_s_Proxy___enter
#define __pyx_n_s_Proxy___exit __pyx_mstate_global->__pyx_n_s_Proxy___exit
#define __pyx_n_s_Proxy___floor __pyx_mstate_global->__pyx_n_s_Proxy___floor
#define __pyx_n_s_Proxy___format __pyx_mstate_global->__pyx_n_s_Proxy___format
#define __pyx_n_s_Proxy___instancecheck __pyx_mstate_global->__pyx_n_s_Proxy___instancecheck
#define __pyx_n_s_Proxy___iter __pyx_mstate_global->__pyx_n_s_Proxy___iter
#define __pyx_n_s_Proxy___length_hint __pyx_mstate_global->__pyx_n_s_Proxy___length_hint
#define __pyx_n_s_Proxy___match_args __pyx_mstate_global->__pyx_n_s_Proxy___match_args
#define __pyx_n_s_Proxy___reduce_cython __pyx_mstate_global->__pyx_n_s_Proxy___reduce_cython
#define __pyx_n_s_Proxy___round __pyx_mstate_global->__pyx_n_s_Proxy___round
#define __pyx_n_s_Proxy___setstate_cython __pyx_mstate_global->__pyx_n_s_Proxy___setstate_cython
#define __pyx_n_s_Proxy___subclasscheck __pyx_mstate_global->__pyx_n_s_Proxy___subclasscheck
#define __pyx_n_s_Proxy___trunc __pyx_mstate_global->__pyx_n_s_Proxy___trunc
#define __pyx_n_s_Proxy__aexit __pyx_mstate_global->__pyx_n_s_Proxy__aexit
#define __pyx_n_s_Proxy__exit __pyx_mstate_global->__pyx_n_s_Proxy__exit
#define __pyx_n_s_Proxy_add __pyx_mstate_global->__pyx_n_s_Proxy_add
#define __pyx_n_s_Proxy_append __pyx_mstate_global->__pyx_n_s_Proxy_append
#define __pyx_n_s_Proxy_clear __pyx_mstate_global->__pyx_n_s_Proxy_clear
#define __pyx_n_s_Proxy_discard __pyx_mstate_global->__pyx_n_s_Proxy_discard
#define __pyx_n_s_Proxy_extend __pyx_mstate_global->__pyx_n_s_Proxy_extend
#define __pyx_n_s_Proxy_insert __pyx_mstate_global->__pyx_n_s_Proxy_insert
#define __pyx_n_s_Proxy_pop __pyx_mstate_global->__pyx_n_s_Proxy_pop
#define __pyx_n_s_Proxy_popitem __pyx_mstate_global->__pyx_n_s_Proxy_popitem
#define __pyx_kp_s_Proxy_pxi __pyx_mstate_global->__pyx_kp_s_Proxy_pxi
#define __pyx_n_s_Proxy_remove __pyx_mstate_global->__pyx_n_s_Proxy_remove
#define __pyx_n_s_Proxy_reverse __pyx_mstate_global->__pyx_n_s_Proxy_reverse
#define __pyx_n_s_Proxy_send __pyx_mstate_global->__pyx_n_s_Proxy_send
#define __pyx_n_s_Proxy_setdefault __pyx_mstate_global->__pyx_n_s_Proxy_setdefault
#define __pyx_n_s_Proxy_sort __pyx_mstate_global->__pyx_n_s_Proxy_sort
#define __pyx_n_s_Proxy_throw __pyx_mstate_global->__pyx_n_s_Proxy_throw
#define __pyx_n_s_Proxy_update __pyx_mstate_global->__pyx_n_s_Proxy_update
#define __pyx_n_s_PyPy __pyx_mstate_global->__pyx_n_s_PyPy
#define __pyx_n_s_Pyx_CFunc_object__lParenWrappe __pyx_mstate_global->__pyx_n_s_Pyx_CFunc_object__lParenWrappe
#define __pyx_n_s_Pyx_CFunc_object__lParenWrappe_2 __pyx_mstate_global->__pyx_n_s_Pyx_CFunc_object__lParenWrappe_2
#define __pyx_n_s_Pyx_CFunc_object__lParenWrappe_3 __pyx_mstate_global->__pyx_n_s_Pyx_CFunc_object__lParenWrappe_3
#define __pyx_kp_s_Read_only_attribute_s __pyx_mstate_global->__pyx_kp_s_Read_only_attribute_s
#define __pyx_n_s_RecursionError __pyx_mstate_global->__pyx_n_s_RecursionError
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_n_s_Sequence __pyx_mstate_global->__pyx_n_s_Sequence
#define __pyx_n_s_Set __pyx_mstate_global->__pyx_n_s_Set
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_Unknown_OldStyle_Class __pyx_mstate_global->__pyx_n_s_Unknown_OldStyle_Class
#define __pyx_n_s_Wrapped __pyx_mstate_global->__pyx_n_s_Wrapped
#define __pyx_kp_s_Wrapped_Frozen_pxi __pyx_mstate_global->__pyx_kp_s_Wrapped_Frozen_pxi
#define __pyx_n_s_Wrapped___dir __pyx_mstate_global->__pyx_n_s_Wrapped___dir
#define __pyx_n_s_Wrapped___reduce_cython __pyx_mstate_global->__pyx_n_s_Wrapped___reduce_cython
#define __pyx_n_s_Wrapped___setstate_cython __pyx_mstate_global->__pyx_n_s_Wrapped___setstate_cython
#define __pyx_n_s_Wrapped_comparator_locals_pass_t __pyx_mstate_global->__pyx_n_s_Wrapped_comparator_locals_pass_t
#define __pyx_kp_s_Wrapped_object_cannot_be_pickled __pyx_mstate_global->__pyx_kp_s_Wrapped_object_cannot_be_pickled
#define __pyx_kp_s__17 __pyx_mstate_global->__pyx_kp_s__17
#define __pyx_n_s__182 __pyx_mstate_global->__pyx_n_s__182
#define __pyx_n_s__2 __pyx_mstate_global->__pyx_n_s__2
#define __pyx_kp_s__3 __pyx_mstate_global->__pyx_kp_s__3
#define __pyx_kp_u__33 __pyx_mstate_global->__pyx_kp_u__33
#define __pyx_kp_s__4 __pyx_mstate_global->__pyx_kp_s__4
#define __pyx_n_s__44 __pyx_mstate_global->__pyx_n_s__44
#define __pyx_kp_s__88 __pyx_mstate_global->__pyx_kp_s__88
#define __pyx_n_s_a __pyx_mstate_global->__pyx_n_s_a
#define __pyx_kp_s_a_zA_Z_a_zA_Z0_9 __pyx_mstate_global->__pyx_kp_s_a_zA_Z_a_zA_Z0_9
#define __pyx_kp_s_a_zA_Z_a_zA_Z0_9_2 __pyx_mstate_global->__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2
#define __pyx_n_s_abc __pyx_mstate_global->__pyx_n_s_abc
#define __pyx_n_s_abs __pyx_mstate_global->__pyx_n_s_abs
#define __pyx_n_s_add __pyx_mstate_global->__pyx_n_s_add
#define __pyx_n_s_add_2 __pyx_mstate_global->__pyx_n_s_add_2
#define __pyx_n_s_aenter __pyx_mstate_global->__pyx_n_s_aenter
#define __pyx_n_s_aexit __pyx_mstate_global->__pyx_n_s_aexit
#define __pyx_n_s_alist __pyx_mstate_global->__pyx_n_s_alist
#define __pyx_n_s_all __pyx_mstate_global->__pyx_n_s_all
#define __pyx_n_s_always_delegated_attributes __pyx_mstate_global->__pyx_n_s_always_delegated_attributes
#define __pyx_n_s_and __pyx_mstate_global->__pyx_n_s_and
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_attr_type_check __pyx_mstate_global->__pyx_n_s_attr_type_check
#define __pyx_n_s_attribute_protected __pyx_mstate_global->__pyx_n_s_attribute_protected
#define __pyx_n_s_basestring __pyx_mstate_global->__pyx_n_s_basestring
#define __pyx_n_s_bool __pyx_mstate_global->__pyx_n_s_bool
#define __pyx_n_s_bool_2 __pyx_mstate_global->__pyx_n_s_bool_2
#define __pyx_n_s_build_regex __pyx_mstate_global->__pyx_n_s_build_regex
#define __pyx_n_s_builtin __pyx_mstate_global->__pyx_n_s_builtin
#define __pyx_n_s_builtins __pyx_mstate_global->__pyx_n_s_builtins
#define __pyx_n_s_bytearray __pyx_mstate_global->__pyx_n_s_bytearray
#define __pyx_n_s_bytes __pyx_mstate_global->__pyx_n_s_bytes
#define __pyx_n_s_bytes_2 __pyx_mstate_global->__pyx_n_s_bytes_2
#define __pyx_n_s_c __pyx_mstate_global->__pyx_n_s_c
#define __pyx_n_s_call __pyx_mstate_global->__pyx_n_s_call
#define __pyx_n_s_ceil __pyx_mstate_global->__pyx_n_s_ceil
#define __pyx_n_s_ceil_2 __pyx_mstate_global->__pyx_n_s_ceil_2
#define __pyx_n_s_cfunc_to_py __pyx_mstate_global->__pyx_n_s_cfunc_to_py
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_clear __pyx_mstate_global->__pyx_n_s_clear
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_cmp __pyx_mstate_global->__pyx_n_s_cmp
#define __pyx_n_s_cn __pyx_mstate_global->__pyx_n_s_cn
#define __pyx_n_s_code __pyx_mstate_global->__pyx_n_s_code
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_s_collections_abc __pyx_mstate_global->__pyx_n_s_collections_abc
#define __pyx_n_s_compile __pyx_mstate_global->__pyx_n_s_compile
#define __pyx_n_s_complex __pyx_mstate_global->__pyx_n_s_complex
#define __pyx_n_s_complex_2 __pyx_mstate_global->__pyx_n_s_complex_2
#define __pyx_n_s_contains __pyx_mstate_global->__pyx_n_s_contains
#define __pyx_n_s_contains_2 __pyx_mstate_global->__pyx_n_s_contains_2
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_d __pyx_mstate_global->__pyx_n_s_d
#define __pyx_n_s_defaults __pyx_mstate_global->__pyx_n_s_defaults
#define __pyx_n_s_delattr __pyx_mstate_global->__pyx_n_s_delattr
#define __pyx_n_s_delete __pyx_mstate_global->__pyx_n_s_delete
#define __pyx_n_s_delitem __pyx_mstate_global->__pyx_n_s_delitem
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_n_s_dict_3 __pyx_mstate_global->__pyx_n_s_dict_3
#define __pyx_n_s_difference __pyx_mstate_global->__pyx_n_s_difference
#define __pyx_n_s_dir __pyx_mstate_global->__pyx_n_s_dir
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_discard __pyx_mstate_global->__pyx_n_s_discard
#define __pyx_n_s_divmod __pyx_mstate_global->__pyx_n_s_divmod
#define __pyx_n_s_doc __pyx_mstate_global->__pyx_n_s_doc
#define __pyx_n_s_dynamic __pyx_mstate_global->__pyx_n_s_dynamic
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_endswith __pyx_mstate_global->__pyx_n_s_endswith
#define __pyx_n_s_enter __pyx_mstate_global->__pyx_n_s_enter
#define __pyx_n_s_environ __pyx_mstate_global->__pyx_n_s_environ
#define __pyx_n_s_eq __pyx_mstate_global->__pyx_n_s_eq
#define __pyx_n_s_exc_type __pyx_mstate_global->__pyx_n_s_exc_type
#define __pyx_n_s_exc_value __pyx_mstate_global->__pyx_n_s_exc_value
#define __pyx_n_s_exit __pyx_mstate_global->__pyx_n_s_exit
#define __pyx_n_s_extend __pyx_mstate_global->__pyx_n_s_extend
#define __pyx_n_s_file __pyx_mstate_global->__pyx_n_s_file
#define __pyx_n_s_float __pyx_mstate_global->__pyx_n_s_float
#define __pyx_n_s_float_2 __pyx_mstate_global->__pyx_n_s_float_2
#define __pyx_n_s_floor __pyx_mstate_global->__pyx_n_s_floor
#define __pyx_n_s_floor_2 __pyx_mstate_global->__pyx_n_s_floor_2
#define __pyx_n_s_floordiv __pyx_mstate_global->__pyx_n_s_floordiv
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_freeze __pyx_mstate_global->__pyx_n_s_freeze
#define __pyx_n_s_frozen __pyx_mstate_global->__pyx_n_s_frozen
#define __pyx_n_s_frozenset __pyx_mstate_global->__pyx_n_s_frozenset
#define __pyx_n_s_functools __pyx_mstate_global->__pyx_n_s_functools
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_ge __pyx_mstate_global->__pyx_n_s_ge
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_getattribute __pyx_mstate_global->__pyx_n_s_getattribute
#define __pyx_n_s_getitem __pyx_mstate_global->__pyx_n_s_getitem
#define __pyx_n_s_getsate __pyx_mstate_global->__pyx_n_s_getsate
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_kp_s_global_c_functions_pxi __pyx_mstate_global->__pyx_kp_s_global_c_functions_pxi
#define __pyx_n_s_gt __pyx_mstate_global->__pyx_n_s_gt
#define __pyx_n_s_hash __pyx_mstate_global->__pyx_n_s_hash
#define __pyx_n_s_hash_2 __pyx_mstate_global->__pyx_n_s_hash_2
#define __pyx_n_s_hash_protected __pyx_mstate_global->__pyx_n_s_hash_protected
#define __pyx_n_s_hash_val __pyx_mstate_global->__pyx_n_s_hash_val
#define __pyx_n_s_help __pyx_mstate_global->__pyx_n_s_help
#define __pyx_n_s_help_protected __pyx_mstate_global->__pyx_n_s_help_protected
#define __pyx_n_s_help_str __pyx_mstate_global->__pyx_n_s_help_str
#define __pyx_n_s_help_val __pyx_mstate_global->__pyx_n_s_help_val
#define __pyx_n_s_hidden_pickle_attributes __pyx_mstate_global->__pyx_n_s_hidden_pickle_attributes
#define __pyx_n_s_hide __pyx_mstate_global->__pyx_n_s_hide
#define __pyx_n_s_hide_private __pyx_mstate_global->__pyx_n_s_hide_private
#define __pyx_n_s_hide_regex __pyx_mstate_global->__pyx_n_s_hide_regex
#define __pyx_n_s_iadd __pyx_mstate_global->__pyx_n_s_iadd
#define __pyx_n_s_iand __pyx_mstate_global->__pyx_n_s_iand
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_s_id_class __pyx_mstate_global->__pyx_n_s_id_class
#define __pyx_n_s_id_protected __pyx_mstate_global->__pyx_n_s_id_protected
#define __pyx_n_s_id_val __pyx_mstate_global->__pyx_n_s_id_val
#define __pyx_n_s_ifloordiv __pyx_mstate_global->__pyx_n_s_ifloordiv
#define __pyx_n_s_ilshift __pyx_mstate_global->__pyx_n_s_ilshift
#define __pyx_n_s_imatmul __pyx_mstate_global->__pyx_n_s_imatmul
#define __pyx_n_s_immutable_builtin_attributes __pyx_mstate_global->__pyx_n_s_immutable_builtin_attributes
#define __pyx_n_s_imod __pyx_mstate_global->__pyx_n_s_imod
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_imul __pyx_mstate_global->__pyx_n_s_imul
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init___locals_C __pyx_mstate_global->__pyx_n_s_init___locals_C
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_insert __pyx_mstate_global->__pyx_n_s_insert
#define __pyx_n_s_inst __pyx_mstate_global->__pyx_n_s_inst
#define __pyx_n_s_instance_of_protected __pyx_mstate_global->__pyx_n_s_instance_of_protected
#define __pyx_n_s_instancecheck __pyx_mstate_global->__pyx_n_s_instancecheck
#define __pyx_n_s_instanceof __pyx_mstate_global->__pyx_n_s_instanceof
#define __pyx_n_s_int __pyx_mstate_global->__pyx_n_s_int
#define __pyx_n_s_int_2 __pyx_mstate_global->__pyx_n_s_int_2
#define __pyx_n_s_intersection __pyx_mstate_global->__pyx_n_s_intersection
#define __pyx_n_s_invert __pyx_mstate_global->__pyx_n_s_invert
#define __pyx_n_s_ior __pyx_mstate_global->__pyx_n_s_ior
#define __pyx_n_s_ipow __pyx_mstate_global->__pyx_n_s_ipow
#define __pyx_n_s_irshift __pyx_mstate_global->__pyx_n_s_irshift
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_isfrozen __pyx_mstate_global->__pyx_n_s_isfrozen
#define __pyx_n_s_isimmutable __pyx_mstate_global->__pyx_n_s_isimmutable
#define __pyx_n_s_isinstance __pyx_mstate_global->__pyx_n_s_isinstance
#define __pyx_n_s_isinstance_protected __pyx_mstate_global->__pyx_n_s_isinstance_protected
#define __pyx_n_s_isinstance_val __pyx_mstate_global->__pyx_n_s_isinstance_val
#define __pyx_n_s_isprivate __pyx_mstate_global->__pyx_n_s_isprivate
#define __pyx_n_s_isprotected __pyx_mstate_global->__pyx_n_s_isprotected
#define __pyx_n_s_isreadonly __pyx_mstate_global->__pyx_n_s_isreadonly
#define __pyx_n_s_issubclass __pyx_mstate_global->__pyx_n_s_issubclass
#define __pyx_n_s_issubclass_protected __pyx_mstate_global->__pyx_n_s_issubclass_protected
#define __pyx_n_s_issubclass_val __pyx_mstate_global->__pyx_n_s_issubclass_val
#define __pyx_n_s_isub __pyx_mstate_global->__pyx_n_s_isub
#define __pyx_n_s_isvisible __pyx_mstate_global->__pyx_n_s_isvisible
#define __pyx_n_s_iswrapped __pyx_mstate_global->__pyx_n_s_iswrapped
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_items_py2 __pyx_mstate_global->__pyx_n_s_items_py2
#define __pyx_n_s_iter __pyx_mstate_global->__pyx_n_s_iter
#define __pyx_n_s_iteritems __pyx_mstate_global->__pyx_n_s_iteritems
#define __pyx_n_s_iterkeys __pyx_mstate_global->__pyx_n_s_iterkeys
#define __pyx_n_s_itervalues __pyx_mstate_global->__pyx_n_s_itervalues
#define __pyx_n_s_itruediv __pyx_mstate_global->__pyx_n_s_itruediv
#define __pyx_n_s_ixor __pyx_mstate_global->__pyx_n_s_ixor
#define __pyx_n_s_join __pyx_mstate_global->__pyx_n_s_join
#define __pyx_n_s_k __pyx_mstate_global->__pyx_n_s_k
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_keys_py2 __pyx_mstate_global->__pyx_n_s_keys_py2
#define __pyx_n_s_kw1 __pyx_mstate_global->__pyx_n_s_kw1
#define __pyx_n_s_kw2 __pyx_mstate_global->__pyx_n_s_kw2
#define __pyx_n_s_kwargs __pyx_mstate_global->__pyx_n_s_kwargs
#define __pyx_n_s_le __pyx_mstate_global->__pyx_n_s_le
#define __pyx_n_s_len __pyx_mstate_global->__pyx_n_s_len
#define __pyx_n_s_length_hint __pyx_mstate_global->__pyx_n_s_length_hint
#define __pyx_n_s_list __pyx_mstate_global->__pyx_n_s_list
#define __pyx_n_s_long __pyx_mstate_global->__pyx_n_s_long
#define __pyx_n_s_lshift __pyx_mstate_global->__pyx_n_s_lshift
#define __pyx_n_s_lstrip __pyx_mstate_global->__pyx_n_s_lstrip
#define __pyx_n_s_lt __pyx_mstate_global->__pyx_n_s_lt
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_major __pyx_mstate_global->__pyx_n_s_major
#define __pyx_n_s_match __pyx_mstate_global->__pyx_n_s_match
#define __pyx_n_s_match_args __pyx_mstate_global->__pyx_n_s_match_args
#define __pyx_n_s_math __pyx_mstate_global->__pyx_n_s_math
#define __pyx_n_s_matmul __pyx_mstate_global->__pyx_n_s_matmul
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_minor __pyx_mstate_global->__pyx_n_s_minor
#define __pyx_n_s_mod __pyx_mstate_global->__pyx_n_s_mod
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_modules __pyx_mstate_global->__pyx_n_s_modules
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_mul __pyx_mstate_global->__pyx_n_s_mul
#define __pyx_n_s_multiwrapped __pyx_mstate_global->__pyx_n_s_multiwrapped
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_ne __pyx_mstate_global->__pyx_n_s_ne
#define __pyx_n_s_neg __pyx_mstate_global->__pyx_n_s_neg
#define __pyx_n_s_never_writeable __pyx_mstate_global->__pyx_n_s_never_writeable
#define __pyx_n_s_never_writeable_private __pyx_mstate_global->__pyx_n_s_never_writeable_private
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_o __pyx_mstate_global->__pyx_n_s_o
#define __pyx_kp_s_o_Invalid_type_s __pyx_mstate_global->__pyx_kp_s_o_Invalid_type_s
#define __pyx_n_s_object __pyx_mstate_global->__pyx_n_s_object
#define __pyx_n_s_oldstyle_class __pyx_mstate_global->__pyx_n_s_oldstyle_class
#define __pyx_n_s_op __pyx_mstate_global->__pyx_n_s_op
#define __pyx_n_s_or __pyx_mstate_global->__pyx_n_s_or
#define __pyx_n_s_os __pyx_mstate_global->__pyx_n_s_os
#define __pyx_n_s_p __pyx_mstate_global->__pyx_n_s_p
#define __pyx_n_s_package __pyx_mstate_global->__pyx_n_s_package
#define __pyx_n_s_partial __pyx_mstate_global->__pyx_n_s_partial
#define __pyx_n_s_pass_to_wrapped __pyx_mstate_global->__pyx_n_s_pass_to_wrapped
#define __pyx_n_s_pattern __pyx_mstate_global->__pyx_n_s_pattern
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_platform __pyx_mstate_global->__pyx_n_s_platform
#define __pyx_n_s_pop __pyx_mstate_global->__pyx_n_s_pop
#define __pyx_n_s_popitem __pyx_mstate_global->__pyx_n_s_popitem
#define __pyx_n_s_pos __pyx_mstate_global->__pyx_n_s_pos
#define __pyx_n_s_pow __pyx_mstate_global->__pyx_n_s_pow
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_private __pyx_mstate_global->__pyx_n_s_private
#define __pyx_n_s_protect __pyx_mstate_global->__pyx_n_s_protect
#define __pyx_n_s_protected_rules_from_kwargs_loca __pyx_mstate_global->__pyx_n_s_protected_rules_from_kwargs_loca
#define __pyx_n_s_py2_function_attrs_rw __pyx_mstate_global->__pyx_n_s_py2_function_attrs_rw
#define __pyx_n_s_pydoc __pyx_mstate_global->__pyx_n_s_pydoc
#define __pyx_n_s_pyprotect_protected __pyx_mstate_global->__pyx_n_s_pyprotect_protected
#define __pyx_n_s_python_implementation __pyx_mstate_global->__pyx_n_s_python_implementation
#define __pyx_kp_s_python_visible_pxi __pyx_mstate_global->__pyx_kp_s_python_visible_pxi
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_Frozen __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Frozen
#define __pyx_n_s_pyx_unpickle_FrozenPrivacyDict __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FrozenPrivacyDict
#define __pyx_n_s_pyx_unpickle_FrozenPrivate __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FrozenPrivate
#define __pyx_n_s_pyx_unpickle_FrozenProtected __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FrozenProtected
#define __pyx_n_s_pyx_unpickle_PrivacyDict __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PrivacyDict
#define __pyx_n_s_pyx_unpickle_Private __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Private
#define __pyx_n_s_pyx_unpickle_Protected __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Protected
#define __pyx_n_s_pyx_unpickle_Proxy __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Proxy
#define __pyx_n_s_pyx_unpickle_Wrapped __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Wrapped
#define __pyx_n_s_pyx_unpickle___HiddenPartial __pyx_mstate_global->__pyx_n_s_pyx_unpickle___HiddenPartial
#define __pyx_n_s_pyx_unpickle___ProtectionData __pyx_mstate_global->__pyx_n_s_pyx_unpickle___ProtectionData
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_s_r __pyx_mstate_global->__pyx_n_s_r
#define __pyx_n_s_radd __pyx_mstate_global->__pyx_n_s_radd
#define __pyx_n_s_rand __pyx_mstate_global->__pyx_n_s_rand
#define __pyx_n_s_rdivmod __pyx_mstate_global->__pyx_n_s_rdivmod
#define __pyx_n_s_re __pyx_mstate_global->__pyx_n_s_re
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_remove __pyx_mstate_global->__pyx_n_s_remove
#define __pyx_n_s_render_doc __pyx_mstate_global->__pyx_n_s_render_doc
#define __pyx_n_s_repr __pyx_mstate_global->__pyx_n_s_repr
#define __pyx_n_s_ret __pyx_mstate_global->__pyx_n_s_ret
#define __pyx_n_s_ret_2 __pyx_mstate_global->__pyx_n_s_ret_2
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_reverse __pyx_mstate_global->__pyx_n_s_reverse
#define __pyx_n_s_rfloordiv __pyx_mstate_global->__pyx_n_s_rfloordiv
#define __pyx_n_s_rl __pyx_mstate_global->__pyx_n_s_rl
#define __pyx_n_s_rlshift __pyx_mstate_global->__pyx_n_s_rlshift
#define __pyx_n_s_rmatmul __pyx_mstate_global->__pyx_n_s_rmatmul
#define __pyx_n_s_rmod __pyx_mstate_global->__pyx_n_s_rmod
#define __pyx_n_s_rmul __pyx_mstate_global->__pyx_n_s_rmul
#define __pyx_n_s_ro __pyx_mstate_global->__pyx_n_s_ro
#define __pyx_n_s_ro_data __pyx_mstate_global->__pyx_n_s_ro_data
#define __pyx_n_s_ro_method __pyx_mstate_global->__pyx_n_s_ro_method
#define __pyx_n_s_ro_regex __pyx_mstate_global->__pyx_n_s_ro_regex
#define __pyx_n_s_ror __pyx_mstate_global->__pyx_n_s_ror
#define __pyx_n_s_round __pyx_mstate_global->__pyx_n_s_round
#define __pyx_n_s_rpow __pyx_mstate_global->__pyx_n_s_rpow
#define __pyx_n_s_rrshift __pyx_mstate_global->__pyx_n_s_rrshift
#define __pyx_n_s_rshift __pyx_mstate_global->__pyx_n_s_rshift
#define __pyx_n_s_rstrip __pyx_mstate_global->__pyx_n_s_rstrip
#define __pyx_n_s_rsub __pyx_mstate_global->__pyx_n_s_rsub
#define __pyx_n_s_rtruediv __pyx_mstate_global->__pyx_n_s_rtruediv
#define __pyx_n_s_rules __pyx_mstate_global->__pyx_n_s_rules
#define __pyx_n_s_rw __pyx_mstate_global->__pyx_n_s_rw
#define __pyx_n_s_rw_regex __pyx_mstate_global->__pyx_n_s_rw_regex
#define __pyx_n_s_rxor __pyx_mstate_global->__pyx_n_s_rxor
#define __pyx_kp_s_s __pyx_mstate_global->__pyx_kp_s_s
#define __pyx_kp_s_s_________0_1 __pyx_mstate_global->__pyx_kp_s_s_________0_1
#define __pyx_n_s_same_class_protected __pyx_mstate_global->__pyx_n_s_same_class_protected
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set __pyx_mstate_global->__pyx_n_s_set
#define __pyx_n_s_set_2 __pyx_mstate_global->__pyx_n_s_set_2
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_setattr __pyx_mstate_global->__pyx_n_s_setattr
#define __pyx_n_s_setdefault __pyx_mstate_global->__pyx_n_s_setdefault
#define __pyx_n_s_setitem __pyx_mstate_global->__pyx_n_s_setitem
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_slots __pyx_mstate_global->__pyx_n_s_slots
#define __pyx_n_s_sort __pyx_mstate_global->__pyx_n_s_sort
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_splitlines __pyx_mstate_global->__pyx_n_s_splitlines
#define __pyx_n_s_startswith __pyx_mstate_global->__pyx_n_s_startswith
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_n_s_str_2 __pyx_mstate_global->__pyx_n_s_str_2
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_sub __pyx_mstate_global->__pyx_n_s_sub
#define __pyx_n_s_subclass __pyx_mstate_global->__pyx_n_s_subclass
#define __pyx_n_s_subclass_of_protected __pyx_mstate_global->__pyx_n_s_subclass_of_protected
#define __pyx_n_s_subclasscheck __pyx_mstate_global->__pyx_n_s_subclasscheck
#define __pyx_n_s_subclassof __pyx_mstate_global->__pyx_n_s_subclassof
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_sys __pyx_mstate_global->__pyx_n_s_sys
#define __pyx_n_s_tb __pyx_mstate_global->__pyx_n_s_tb
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_testop __pyx_mstate_global->__pyx_n_s_testop
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_truediv __pyx_mstate_global->__pyx_n_s_truediv
#define __pyx_n_s_trunc __pyx_mstate_global->__pyx_n_s_trunc
#define __pyx_n_s_tuple __pyx_mstate_global->__pyx_n_s_tuple
#define __pyx_n_s_type __pyx_mstate_global->__pyx_n_s_type
#define __pyx_n_s_types __pyx_mstate_global->__pyx_n_s_types
#define __pyx_n_s_unichr __pyx_mstate_global->__pyx_n_s_unichr
#define __pyx_n_s_unicode __pyx_mstate_global->__pyx_n_s_unicode
#define __pyx_n_s_union __pyx_mstate_global->__pyx_n_s_union
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_v __pyx_mstate_global->__pyx_n_s_v
#define __pyx_n_s_val __pyx_mstate_global->__pyx_n_s_val
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_n_s_values_py2 __pyx_mstate_global->__pyx_n_s_values_py2
#define __pyx_n_s_version_info __pyx_mstate_global->__pyx_n_s_version_info
#define __pyx_n_s_viewitems __pyx_mstate_global->__pyx_n_s_viewitems
#define __pyx_n_s_viewkeys __pyx_mstate_global->__pyx_n_s_viewkeys
#define __pyx_n_s_viewvalues __pyx_mstate_global->__pyx_n_s_viewvalues
#define __pyx_n_s_w __pyx_mstate_global->__pyx_n_s_w
#define __pyx_n_s_want_frozen __pyx_mstate_global->__pyx_n_s_want_frozen
#define __pyx_n_s_weakref __pyx_mstate_global->__pyx_n_s_weakref
#define __pyx_n_s_wrap __pyx_mstate_global->__pyx_n_s_wrap
#define __pyx_n_s_x __pyx_mstate_global->__pyx_n_s_x
#define __pyx_n_s_x_2 __pyx_mstate_global->__pyx_n_s_x_2
#define __pyx_n_s_xor __pyx_mstate_global->__pyx_n_s_xor
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_7 __pyx_mstate_global->__pyx_int_7
#define __pyx_int_6017409 __pyx_mstate_global->__pyx_int_6017409
#define __pyx_int_19817578 __pyx_mstate_global->__pyx_int_19817578
#define __pyx_int_50167005 __pyx_mstate_global->__pyx_int_50167005
#define __pyx_int_78131249 __pyx_mstate_global->__pyx_int_78131249
#define __pyx_int_115090883 __pyx_mstate_global->__pyx_int_115090883
#define __pyx_int_152356376 __pyx_mstate_global->__pyx_int_152356376
#define __pyx_int_155231502 __pyx_mstate_global->__pyx_int_155231502
#define __pyx_int_161740782 __pyx_mstate_global->__pyx_int_161740782
#define __pyx_int_180490543 __pyx_mstate_global->__pyx_int_180490543
#define __pyx_int_209109265 __pyx_mstate_global->__pyx_int_209109265
#define __pyx_int_210565405 __pyx_mstate_global->__pyx_int_210565405
#define __pyx_int_218598457 __pyx_mstate_global->__pyx_int_218598457
#define __pyx_int_247595846 __pyx_mstate_global->__pyx_int_247595846
#define __pyx_int_252507329 __pyx_mstate_global->__pyx_int_252507329
#define __pyx_int_262487005 __pyx_mstate_global->__pyx_int_262487005
#define __pyx_tuple__5 __pyx_mstate_global->__pyx_tuple__5
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_slice__18 __pyx_mstate_global->__pyx_slice__18
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__40 __pyx_mstate_global->__pyx_tuple__40
#define __pyx_tuple__42 __pyx_mstate_global->__pyx_tuple__42
#define __pyx_tuple__45 __pyx_mstate_global->__pyx_tuple__45
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__56 __pyx_mstate_global->__pyx_tuple__56
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__71 __pyx_mstate_global->__pyx_tuple__71
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__74 __pyx_mstate_global->__pyx_tuple__74
#define __pyx_tuple__82 __pyx_mstate_global->__pyx_tuple__82
#define __pyx_tuple__83 __pyx_mstate_global->__pyx_tuple__83
#define __pyx_tuple__84 __pyx_mstate_global->__pyx_tuple__84
#define __pyx_tuple__85 __pyx_mstate_global->__pyx_tuple__85
#define __pyx_tuple__86 __pyx_mstate_global->__pyx_tuple__86
#define __pyx_tuple__87 __pyx_mstate_global->__pyx_tuple__87
#define __pyx_tuple__89 __pyx_mstate_global->__pyx_tuple__89
#define __pyx_tuple__90 __pyx_mstate_global->__pyx_tuple__90
#define __pyx_tuple__91 __pyx_mstate_global->__pyx_tuple__91
#define __pyx_tuple__93 __pyx_mstate_global->__pyx_tuple__93
#define __pyx_tuple__95 __pyx_mstate_global->__pyx_tuple__95
#define __pyx_tuple__98 __pyx_mstate_global->__pyx_tuple__98
#define __pyx_codeobj__6 __pyx_mstate_global->__pyx_codeobj__6
#define __pyx_tuple__106 __pyx_mstate_global->__pyx_tuple__106
#define __pyx_tuple__108 __pyx_mstate_global->__pyx_tuple__108
#define __pyx_tuple__111 __pyx_mstate_global->__pyx_tuple__111
#define __pyx_tuple__116 __pyx_mstate_global->__pyx_tuple__116
#define __pyx_tuple__119 __pyx_mstate_global->__pyx_tuple__119
#define __pyx_tuple__140 __pyx_mstate_global->__pyx_tuple__140
#define __pyx_tuple__142 __pyx_mstate_global->__pyx_tuple__142
#define __pyx_tuple__143 __pyx_mstate_global->__pyx_tuple__143
#define __pyx_tuple__144 __pyx_mstate_global->__pyx_tuple__144
#define __pyx_tuple__146 __pyx_mstate_global->__pyx_tuple__146
#define __pyx_tuple__170 __pyx_mstate_global->__pyx_tuple__170
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__24 __pyx_mstate_global->__pyx_codeobj__24
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__26 __pyx_mstate_global->__pyx_codeobj__26
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__28 __pyx_mstate_global->__pyx_codeobj__28
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__55 __pyx_mstate_global->__pyx_codeobj__55
#define __pyx_codeobj__57 __pyx_mstate_global->__pyx_codeobj__57
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__60 __pyx_mstate_global->__pyx_codeobj__60
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__65 __pyx_mstate_global->__pyx_codeobj__65
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__68 __pyx_mstate_global->__pyx_codeobj__68
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__80 __pyx_mstate_global->__pyx_codeobj__80
#define __pyx_codeobj__81 __pyx_mstate_global->__pyx_codeobj__81
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__94 __pyx_mstate_global->__pyx_codeobj__94
#define __pyx_codeobj__96 __pyx_mstate_global->__pyx_codeobj__96
#define __pyx_codeobj__97 __pyx_mstate_global->__pyx_codeobj__97
#define __pyx_codeobj__99 __pyx_mstate_global->__pyx_codeobj__99
#define __pyx_codeobj__100 __pyx_mstate_global->__pyx_codeobj__100
#define __pyx_codeobj__101 __pyx_mstate_global->__pyx_codeobj__101
#define __pyx_codeobj__102 __pyx_mstate_global->__pyx_codeobj__102
#define __pyx_codeobj__103 __pyx_mstate_global->__pyx_codeobj__103
#define __pyx_codeobj__104 __pyx_mstate_global->__pyx_codeobj__104
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__107 __pyx_mstate_global->__pyx_codeobj__107
#define __pyx_codeobj__109 __pyx_mstate_global->__pyx_codeobj__109
#define __pyx_codeobj__110 __pyx_mstate_global->__pyx_codeobj__110
#define __pyx_codeobj__112 __pyx_mstate_global->__pyx_codeobj__112
#define __pyx_codeobj__113 __pyx_mstate_global->__pyx_codeobj__113
#define __pyx_codeobj__114 __pyx_mstate_global->__pyx_codeobj__114
#define __pyx_codeobj__115 __pyx_mstate_global->__pyx_codeobj__115
#define __pyx_codeobj__117 __pyx_mstate_global->__pyx_codeobj__117
#define __pyx_codeobj__118 __pyx_mstate_global->__pyx_codeobj__118
#define __pyx_codeobj__120 __pyx_mstate_global->__pyx_codeobj__120
#define __pyx_codeobj__121 __pyx_mstate_global->__pyx_codeobj__121
#define __pyx_codeobj__122 __pyx_mstate_global->__pyx_codeobj__122
#define __pyx_codeobj__123 __pyx_mstate_global->__pyx_codeobj__123
#define __pyx_codeobj__124 __pyx_mstate_global->__pyx_codeobj__124
#define __pyx_codeobj__125 __pyx_mstate_global->__pyx_codeobj__125
#define __pyx_codeobj__126 __pyx_mstate_global->__pyx_codeobj__126
#define __pyx_codeobj__127 __pyx_mstate_global->__pyx_codeobj__127
#define __pyx_codeobj__128 __pyx_mstate_global->__pyx_codeobj__128
#define __pyx_codeobj__129 __pyx_mstate_global->__pyx_codeobj__129
#define __pyx_codeobj__130 __pyx_mstate_global->__pyx_codeobj__130
#define __pyx_codeobj__131 __pyx_mstate_global->__pyx_codeobj__131
#define __pyx_codeobj__132 __pyx_mstate_global->__pyx_codeobj__132
#define __pyx_codeobj__133 __pyx_mstate_global->__pyx_codeobj__133
#define __pyx_codeobj__134 __pyx_mstate_global->__pyx_codeobj__134
#define __pyx_codeobj__135 __pyx_mstate_global->__pyx_codeobj__135
#define __pyx_codeobj__136 __pyx_mstate_global->__pyx_codeobj__136
#define __pyx_codeobj__137 __pyx_mstate_global->__pyx_codeobj__137
#define __pyx_codeobj__138 __pyx_mstate_global->__pyx_codeobj__138
#define __pyx_codeobj__139 __pyx_mstate_global->__pyx_codeobj__139
#define __pyx_codeobj__141 __pyx_mstate_global->__pyx_codeobj__141
#define __pyx_codeobj__145 __pyx_mstate_global->__pyx_codeobj__145
#define __pyx_codeobj__147 __pyx_mstate_global->__pyx_codeobj__147
#define __pyx_codeobj__148 __pyx_mstate_global->__pyx_codeobj__148
#define __pyx_codeobj__149 __pyx_mstate_global->__pyx_codeobj__149
#define __pyx_codeobj__150 __pyx_mstate_global->__pyx_codeobj__150
#define __pyx_codeobj__151 __pyx_mstate_global->__pyx_codeobj__151
#define __pyx_codeobj__152 __pyx_mstate_global->__pyx_codeobj__152
#define __pyx_codeobj__153 __pyx_mstate_global->__pyx_codeobj__153
#define __pyx_codeobj__154 __pyx_mstate_global->__pyx_codeobj__154
#define __pyx_codeobj__155 __pyx_mstate_global->__pyx_codeobj__155
#define __pyx_codeobj__156 __pyx_mstate_global->__pyx_codeobj__156
#define __pyx_codeobj__157 __pyx_mstate_global->__pyx_codeobj__157
#define __pyx_codeobj__158 __pyx_mstate_global->__pyx_codeobj__158
#define __pyx_codeobj__159 __pyx_mstate_global->__pyx_codeobj__159
#define __pyx_codeobj__160 __pyx_mstate_global->__pyx_codeobj__160
#define __pyx_codeobj__161 __pyx_mstate_global->__pyx_codeobj__161
#define __pyx_codeobj__162 __pyx_mstate_global->__pyx_codeobj__162
#define __pyx_codeobj__163 __pyx_mstate_global->__pyx_codeobj__163
#define __pyx_codeobj__164 __pyx_mstate_global->__pyx_codeobj__164
#define __pyx_codeobj__165 __pyx_mstate_global->__pyx_codeobj__165
#define __pyx_codeobj__166 __pyx_mstate_global->__pyx_codeobj__166
#define __pyx_codeobj__167 __pyx_mstate_global->__pyx_codeobj__167
#define __pyx_codeobj__168 __pyx_mstate_global->__pyx_codeobj__168
#define __pyx_codeobj__169 __pyx_mstate_global->__pyx_codeobj__169
#define __pyx_codeobj__171 __pyx_mstate_global->__pyx_codeobj__171
#define __pyx_codeobj__172 __pyx_mstate_global->__pyx_codeobj__172
#define __pyx_codeobj__173 __pyx_mstate_global->__pyx_codeobj__173
#define __pyx_codeobj__174 __pyx_mstate_global->__pyx_codeobj__174
#define __pyx_codeobj__175 __pyx_mstate_global->__pyx_codeobj__175
#define __pyx_codeobj__176 __pyx_mstate_global->__pyx_codeobj__176
#define __pyx_codeobj__177 __pyx_mstate_global->__pyx_codeobj__177
#define __pyx_codeobj__178 __pyx_mstate_global->__pyx_codeobj__178
#define __pyx_codeobj__179 __pyx_mstate_global->__pyx_codeobj__179
#define __pyx_codeobj__180 __pyx_mstate_global->__pyx_codeobj__180
#define __pyx_codeobj__181 __pyx_mstate_global->__pyx_codeobj__181
#endif
/* #### Code section: module_code ### */

/* "python_visible.pxi":6
 * # ------------------------------------------------------------------------
 * 
 * def attribute_protected():             # <<<<<<<<<<<<<<
 *     '''
 *     attribute_protected() -> str: name of special attribute in Wrapped objects
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_1attribute_protected(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_attribute_protected, "\n    attribute_protected() -> str: name of special attribute in Wrapped objects\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_1attribute_protected = {"attribute_protected", (PyCFunction)__pyx_pw_9pyprotect_9protected_1attribute_protected, METH_NOARGS, __pyx_doc_9pyprotect_9protected_attribute_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_1attribute_protected(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("attribute_protected (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_attribute_protected(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_attribute_protected(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("attribute_protected", 0);

  /* "python_visible.pxi":10
 *     attribute_protected() -> str: name of special attribute in Wrapped objects
 *     '''
 *     return PROT_ATTR_NAME             # <<<<<<<<<<<<<<
 * 
 * def id_protected(o: object) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_9pyprotect_9protected_PROT_ATTR_NAME);
  __pyx_r = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
  goto __pyx_L0;

  /* "python_visible.pxi":6
 * # ------------------------------------------------------------------------
 * 
 * def attribute_protected():             # <<<<<<<<<<<<<<
 *     '''
 *     attribute_protected() -> str: name of special attribute in Wrapped objects
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":12
 *     return PROT_ATTR_NAME
 * 
 * def id_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     id_protected(o: object) -> int:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_3id_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_2id_protected, "\n    id_protected(o: object) -> int:\n    id of wrapped object if iswrapped(o); id of 'o' otherwise\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_3id_protected = {"id_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_3id_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_2id_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_3id_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("id_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "id_protected") < 0)) __PYX_ERR(0, 12, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("id_protected", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 12, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.id_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_2id_protected(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_2id_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("id_protected", 0);

  /* "python_visible.pxi":17
 *     id of wrapped object if iswrapped(o); id of 'o' otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).id
 *     return id(o)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  if (__pyx_t_1) {

    /* "python_visible.pxi":18
 *     '''
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).id             # <<<<<<<<<<<<<<
 *     return id(o)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":17
 *     id of wrapped object if iswrapped(o); id of 'o' otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).id
 *     return id(o)
 */
  }

  /* "python_visible.pxi":19
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).id
 *     return id(o)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_o); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":12
 *     return PROT_ATTR_NAME
 * 
 * def id_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     id_protected(o: object) -> int:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.id_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":22
 * 
 * 
 * def hash_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     hash_protected(o: object) -> int:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5hash_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_4hash_protected, "\n    hash_protected(o: object) -> int:\n    hash of wrapped object if iswrapped(o); hash of 'o' otherwise\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5hash_protected = {"hash_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5hash_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_4hash_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_5hash_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hash_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 22, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "hash_protected") < 0)) __PYX_ERR(0, 22, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("hash_protected", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 22, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.hash_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_4hash_protected(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_4hash_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_hash_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hash_protected", 0);

  /* "python_visible.pxi":27
 *     hash of wrapped object if iswrapped(o); hash of 'o' otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).hash()
 *     return hash(o)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  if (__pyx_t_1) {

    /* "python_visible.pxi":28
 *     '''
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).hash()             # <<<<<<<<<<<<<<
 *     return hash(o)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_hash); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":27
 *     hash of wrapped object if iswrapped(o); hash of 'o' otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).hash()
 *     return hash(o)
 */
  }

  /* "python_visible.pxi":29
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).hash()
 *     return hash(o)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyObject_Hash(__pyx_v_o); if (unlikely(__pyx_t_6 == ((Py_hash_t)-1))) __PYX_ERR(0, 29, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromHash_t(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":22
 * 
 * 
 * def hash_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     hash_protected(o: object) -> int:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.hash_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":32
 * 
 * 
 * def isinstance_protected(o: object, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isinstance_protected(o: object, c: type) -> bool:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7isinstance_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_6isinstance_protected, "\n    isinstance_protected(o: object, c: type) -> bool:\n    Returns: isinstance(wrapped_object, c) if iswrapped(o)\n        isinstance(o, c) otherwise\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7isinstance_protected = {"isinstance_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7isinstance_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_6isinstance_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_7isinstance_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  PyTypeObject *__pyx_v_c = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isinstance_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_c,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_c,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 32, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_c)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 32, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("isinstance_protected", 1, 2, 2, 1); __PYX_ERR(0, 32, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "isinstance_protected") < 0)) __PYX_ERR(0, 32, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_o = values[0];
    __pyx_v_c = ((PyTypeObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("isinstance_protected", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 32, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.isinstance_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyType_Type), 0, "c", 1))) __PYX_ERR(0, 32, __pyx_L1_error)
  __pyx_r = __pyx_pf_9pyprotect_9protected_6isinstance_protected(__pyx_self, __pyx_v_o, __pyx_v_c);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_6isinstance_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyTypeObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isinstance_protected", 0);

  /* "python_visible.pxi":38
 *         isinstance(o, c) otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).isinstance(c)
 *     return isinstance(o, c)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  if (__pyx_t_1) {

    /* "python_visible.pxi":39
 *     '''
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).isinstance(c)             # <<<<<<<<<<<<<<
 *     return isinstance(o, c)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_isinstance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_c)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":38
 *         isinstance(o, c) otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).isinstance(c)
 *     return isinstance(o, c)
 */
  }

  /* "python_visible.pxi":40
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).isinstance(c)
 *     return isinstance(o, c)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_v_c); 
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":32
 * 
 * 
 * def isinstance_protected(o: object, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isinstance_protected(o: object, c: type) -> bool:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.isinstance_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":43
 * 
 * 
 * def issubclass_protected(o: type, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     issubclass_protected(o: object, c: type) -> bool:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_9issubclass_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_8issubclass_protected, "\n    issubclass_protected(o: object, c: type) -> bool:\n    Returns: issubclass(wrapped_object, c) if iswrapped(o);\n        issubclass(o, c) otherwise\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_9issubclass_protected = {"issubclass_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_9issubclass_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_8issubclass_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_9issubclass_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyTypeObject *__pyx_v_o = 0;
  PyTypeObject *__pyx_v_c = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("issubclass_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_c,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_c,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 43, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_c)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 43, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("issubclass_protected", 1, 2, 2, 1); __PYX_ERR(0, 43, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "issubclass_protected") < 0)) __PYX_ERR(0, 43, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_o = ((PyTypeObject*)values[0]);
    __pyx_v_c = ((PyTypeObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("issubclass_protected", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 43, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.issubclass_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_o), (&PyType_Type), 0, "o", 1))) __PYX_ERR(0, 43, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyType_Type), 0, "c", 1))) __PYX_ERR(0, 43, __pyx_L1_error)
  __pyx_r = __pyx_pf_9pyprotect_9protected_8issubclass_protected(__pyx_self, __pyx_v_o, __pyx_v_c);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_8issubclass_protected(CYTHON_UNUSED PyObject *__pyx_self, PyTypeObject *__pyx_v_o, PyTypeObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("issubclass_protected", 0);

  /* "python_visible.pxi":49
 *         issubclass(o, c) otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).issubclass(c)
 *     return issubclass(o, c)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_o), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  if (__pyx_t_1) {

    /* "python_visible.pxi":50
 *     '''
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).issubclass(c)             # <<<<<<<<<<<<<<
 *     return issubclass(o, c)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetAttr(((PyObject *)__pyx_v_o), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_issubclass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_c)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":49
 *         issubclass(o, c) otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).issubclass(c)
 *     return issubclass(o, c)
 */
  }

  /* "python_visible.pxi":51
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).issubclass(c)
 *     return issubclass(o, c)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_IsSubclass(((PyObject *)__pyx_v_o), ((PyObject *)__pyx_v_c)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":43
 * 
 * 
 * def issubclass_protected(o: type, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     issubclass_protected(o: object, c: type) -> bool:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.issubclass_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":54
 * 
 * 
 * def instance_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     instance_of_protected(x: object, w: object) -> bool
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11instance_of_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_10instance_of_protected, "\n    instance_of_protected(x: object, w: object) -> bool\n    If iswrapped(w) and w wraps 'o':\n        Returns isinstance(x, type(o))\n    Else: returns isinstance(x, w)\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11instance_of_protected = {"instance_of_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11instance_of_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_10instance_of_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_11instance_of_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_w = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("instance_of_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_w,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_w,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_x)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_w)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("instance_of_protected", 1, 2, 2, 1); __PYX_ERR(0, 54, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "instance_of_protected") < 0)) __PYX_ERR(0, 54, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_x = values[0];
    __pyx_v_w = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("instance_of_protected", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 54, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.instance_of_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_10instance_of_protected(__pyx_self, __pyx_v_x, __pyx_v_w);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_10instance_of_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_w) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("instance_of_protected", 0);

  /* "python_visible.pxi":61
 *     Else: returns isinstance(x, w)
 *     '''
 *     if iswrapped(w):             # <<<<<<<<<<<<<<
 *         return getattr(w, PROT_ATTR_NAME).instanceof(x)
 *     return isinstance(x, w)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_w};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":62
 *     '''
 *     if iswrapped(w):
 *         return getattr(w, PROT_ATTR_NAME).instanceof(x)             # <<<<<<<<<<<<<<
 *     return isinstance(x, w)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_w, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_instanceof); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_x};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":61
 *     Else: returns isinstance(x, w)
 *     '''
 *     if iswrapped(w):             # <<<<<<<<<<<<<<
 *         return getattr(w, PROT_ATTR_NAME).instanceof(x)
 *     return isinstance(x, w)
 */
  }

  /* "python_visible.pxi":63
 *     if iswrapped(w):
 *         return getattr(w, PROT_ATTR_NAME).instanceof(x)
 *     return isinstance(x, w)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_x, __pyx_v_w); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 63, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":54
 * 
 * 
 * def instance_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     instance_of_protected(x: object, w: object) -> bool
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.instance_of_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":66
 * 
 * 
 * def subclass_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     subclass_of_protected(x: object, w: object) -> bool
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_13subclass_of_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_12subclass_of_protected, "\n    subclass_of_protected(x: object, w: object) -> bool\n    If iswrapped(w) and w wraps 'o':\n        Returns issubclass(x, type(o))\n    Else: returns issubclass(x, w)\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_13subclass_of_protected = {"subclass_of_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_13subclass_of_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_12subclass_of_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_13subclass_of_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_w = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subclass_of_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_w,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_w,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_x)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_w)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("subclass_of_protected", 1, 2, 2, 1); __PYX_ERR(0, 66, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "subclass_of_protected") < 0)) __PYX_ERR(0, 66, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_x = values[0];
    __pyx_v_w = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subclass_of_protected", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 66, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.subclass_of_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_12subclass_of_protected(__pyx_self, __pyx_v_x, __pyx_v_w);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_12subclass_of_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_w) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subclass_of_protected", 0);

  /* "python_visible.pxi":73
 *     Else: returns issubclass(x, w)
 *     '''
 *     if iswrapped(w):             # <<<<<<<<<<<<<<
 *         return getattr(w, PROT_ATTR_NAME).subclassof(x)
 *     return issubclass(x, w)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_w};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":74
 *     '''
 *     if iswrapped(w):
 *         return getattr(w, PROT_ATTR_NAME).subclassof(x)             # <<<<<<<<<<<<<<
 *     return issubclass(x, w)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_w, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_subclassof); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_x};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":73
 *     Else: returns issubclass(x, w)
 *     '''
 *     if iswrapped(w):             # <<<<<<<<<<<<<<
 *         return getattr(w, PROT_ATTR_NAME).subclassof(x)
 *     return issubclass(x, w)
 */
  }

  /* "python_visible.pxi":75
 *     if iswrapped(w):
 *         return getattr(w, PROT_ATTR_NAME).subclassof(x)
 *     return issubclass(x, w)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyObject_IsSubclass(__pyx_v_x, __pyx_v_w); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 75, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":66
 * 
 * 
 * def subclass_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     subclass_of_protected(x: object, w: object) -> bool
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.subclass_of_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":78
 * 
 * 
 * def same_class_protected(c: type, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     same_class_protected(c: type, w: object) -> bool
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15same_class_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_14same_class_protected, "\n    same_class_protected(c: type, w: object) -> bool\n    If iswrapped(w) and w wraps 'o':\n        Returns (c is type(o))\n    Else: returns (c is type(w))\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_15same_class_protected = {"same_class_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15same_class_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_14same_class_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_15same_class_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyTypeObject *__pyx_v_c = 0;
  PyObject *__pyx_v_w = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("same_class_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c,&__pyx_n_s_w,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c,&__pyx_n_s_w,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_c)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_w)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("same_class_protected", 1, 2, 2, 1); __PYX_ERR(0, 78, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "same_class_protected") < 0)) __PYX_ERR(0, 78, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_c = ((PyTypeObject*)values[0]);
    __pyx_v_w = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("same_class_protected", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 78, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.same_class_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyType_Type), 0, "c", 1))) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_r = __pyx_pf_9pyprotect_9protected_14same_class_protected(__pyx_self, __pyx_v_c, __pyx_v_w);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_14same_class_protected(CYTHON_UNUSED PyObject *__pyx_self, PyTypeObject *__pyx_v_c, PyObject *__pyx_v_w) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("same_class_protected", 0);

  /* "python_visible.pxi":85
 *     Else: returns (c is type(w))
 *     '''
 *     if iswrapped(w):             # <<<<<<<<<<<<<<
 *         return id(c) == getattr(w, PROT_ATTR_NAME).id_class
 *     return c is type(w)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_w};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":86
 *     '''
 *     if iswrapped(w):
 *         return id(c) == getattr(w, PROT_ATTR_NAME).id_class             # <<<<<<<<<<<<<<
 *     return c is type(w)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_w, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_id_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":85
 *     Else: returns (c is type(w))
 *     '''
 *     if iswrapped(w):             # <<<<<<<<<<<<<<
 *         return id(c) == getattr(w, PROT_ATTR_NAME).id_class
 *     return c is type(w)
 */
  }

  /* "python_visible.pxi":87
 *     if iswrapped(w):
 *         return id(c) == getattr(w, PROT_ATTR_NAME).id_class
 *     return c is type(w)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = (__pyx_v_c == ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_w))));
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":78
 * 
 * 
 * def same_class_protected(c: type, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     same_class_protected(c: type, w: object) -> bool
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.same_class_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":90
 * 
 * 
 * def help_protected(o: object) -> None:             # <<<<<<<<<<<<<<
 *     '''
 *     help_protected(o: object) -> None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_17help_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_16help_protected, "\n    help_protected(o: object) -> None\n    Calls help(wrapped_object) if iswrapped(o); help(o) otherwise\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_17help_protected = {"help_protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_17help_protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_16help_protected};
static PyObject *__pyx_pw_9pyprotect_9protected_17help_protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("help_protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 90, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "help_protected") < 0)) __PYX_ERR(0, 90, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("help_protected", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 90, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.help_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_16help_protected(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16help_protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("help_protected", 0);

  /* "python_visible.pxi":95
 *     Calls help(wrapped_object) if iswrapped(o); help(o) otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).help()
 *     return help(o)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  if (__pyx_t_1) {

    /* "python_visible.pxi":96
 *     '''
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).help()             # <<<<<<<<<<<<<<
 *     return help(o)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_help); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":95
 *     Calls help(wrapped_object) if iswrapped(o); help(o) otherwise
 *     '''
 *     if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).help()
 *     return help(o)
 */
  }

  /* "python_visible.pxi":97
 *     if isinstance(o, Wrapped):
 *         return getattr(o, PROT_ATTR_NAME).help()
 *     return help(o)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_help, __pyx_v_o); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":90
 * 
 * 
 * def help_protected(o: object) -> None:             # <<<<<<<<<<<<<<
 *     '''
 *     help_protected(o: object) -> None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.help_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":100
 * 
 * 
 * def contains(p: object, o: object):             # <<<<<<<<<<<<<<
 *     '''
 *     contains(p: object, o: object): bool: whether 'p' wraps 'o'
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_19contains(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_18contains, "\n    contains(p: object, o: object): bool: whether 'p' wraps 'o'\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_19contains = {"contains", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_19contains, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_18contains};
static PyObject *__pyx_pw_9pyprotect_9protected_19contains(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_p = 0;
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("contains (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_p,&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_p,&__pyx_n_s_o,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_p)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("contains", 1, 2, 2, 1); __PYX_ERR(0, 100, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "contains") < 0)) __PYX_ERR(0, 100, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_p = values[0];
    __pyx_v_o = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("contains", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 100, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_18contains(__pyx_self, __pyx_v_p, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_18contains(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("contains", 0);

  /* "python_visible.pxi":104
 *     contains(p: object, o: object): bool: whether 'p' wraps 'o'
 *     '''
 *     if isinstance(p, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(p, PROT_ATTR_NAME).id == id(o)
 *     return False
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_p, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  if (__pyx_t_1) {

    /* "python_visible.pxi":105
 *     '''
 *     if isinstance(p, Wrapped):
 *         return getattr(p, PROT_ATTR_NAME).id == id(o)             # <<<<<<<<<<<<<<
 *     return False
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_p, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_o); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":104
 *     contains(p: object, o: object): bool: whether 'p' wraps 'o'
 *     '''
 *     if isinstance(p, Wrapped):             # <<<<<<<<<<<<<<
 *         return getattr(p, PROT_ATTR_NAME).id == id(o)
 *     return False
 */
  }

  /* "python_visible.pxi":106
 *     if isinstance(p, Wrapped):
 *         return getattr(p, PROT_ATTR_NAME).id == id(o)
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "python_visible.pxi":100
 * 
 * 
 * def contains(p: object, o: object):             # <<<<<<<<<<<<<<
 *     '''
 *     contains(p: object, o: object): bool: whether 'p' wraps 'o'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.contains", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":114
 * 
 * 
 * def isimmutable(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isimmutable(o: object) -> bool: 'o' is KNOWN to be immutable
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_21isimmutable(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_20isimmutable, "\n    isimmutable(o: object) -> bool: 'o' is KNOWN to be immutable\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_21isimmutable = {"isimmutable", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_21isimmutable, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_20isimmutable};
static PyObject *__pyx_pw_9pyprotect_9protected_21isimmutable(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isimmutable (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "isimmutable") < 0)) __PYX_ERR(0, 114, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("isimmutable", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 114, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.isimmutable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_20isimmutable(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_20isimmutable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_hash_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isimmutable", 0);

  /* "python_visible.pxi":119
 *     '''
 *     # Everything in builtin module is immutable
 *     if id(o) in builtins_ids:             # <<<<<<<<<<<<<<
 *         return True
 *     # None and NotImplemented are immutable
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_9pyprotect_9protected_builtins_ids, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "python_visible.pxi":120
 *     # Everything in builtin module is immutable
 *     if id(o) in builtins_ids:
 *         return True             # <<<<<<<<<<<<<<
 *     # None and NotImplemented are immutable
 *     if o is None or o is NotImplemented:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "python_visible.pxi":119
 *     '''
 *     # Everything in builtin module is immutable
 *     if id(o) in builtins_ids:             # <<<<<<<<<<<<<<
 *         return True
 *     # None and NotImplemented are immutable
 */
  }

  /* "python_visible.pxi":122
 *         return True
 *     # None and NotImplemented are immutable
 *     if o is None or o is NotImplemented:             # <<<<<<<<<<<<<<
 *         return True
 *     if isfrozen(o):
 */
  __pyx_t_3 = (__pyx_v_o == Py_None);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_o == __pyx_builtin_NotImplemented);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "python_visible.pxi":123
 *     # None and NotImplemented are immutable
 *     if o is None or o is NotImplemented:
 *         return True             # <<<<<<<<<<<<<<
 *     if isfrozen(o):
 *         return True
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "python_visible.pxi":122
 *         return True
 *     # None and NotImplemented are immutable
 *     if o is None or o is NotImplemented:             # <<<<<<<<<<<<<<
 *         return True
 *     if isfrozen(o):
 */
  }

  /* "python_visible.pxi":124
 *     if o is None or o is NotImplemented:
 *         return True
 *     if isfrozen(o):             # <<<<<<<<<<<<<<
 *         return True
 *     if type(o) in immutable_types_set:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_isfrozen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "python_visible.pxi":125
 *         return True
 *     if isfrozen(o):
 *         return True             # <<<<<<<<<<<<<<
 *     if type(o) in immutable_types_set:
 *         # Although 'tuple' is immutable in python, for our purposes, 'tuple' does
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "python_visible.pxi":124
 *     if o is None or o is NotImplemented:
 *         return True
 *     if isfrozen(o):             # <<<<<<<<<<<<<<
 *         return True
 *     if type(o) in immutable_types_set:
 */
  }

  /* "python_visible.pxi":126
 *     if isfrozen(o):
 *         return True
 *     if type(o) in immutable_types_set:             # <<<<<<<<<<<<<<
 *         # Although 'tuple' is immutable in python, for our purposes, 'tuple' does
 *         # NOT prevent modification to MEMBERS of the tuple that may be mutable
 */
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(((PyObject *)Py_TYPE(__pyx_v_o)), __pyx_v_9pyprotect_9protected_immutable_types_set, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 126, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "python_visible.pxi":130
 *         # NOT prevent modification to MEMBERS of the tuple that may be mutable
 *         # Hence, check if 'o' has a stable hash - python hash() does this for us
 *         try:             # <<<<<<<<<<<<<<
 *             hash(o)
 *             return True
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "python_visible.pxi":131
 *         # Hence, check if 'o' has a stable hash - python hash() does this for us
 *         try:
 *             hash(o)             # <<<<<<<<<<<<<<
 *             return True
 *         except TypeError:
 */
        __pyx_t_10 = PyObject_Hash(__pyx_v_o); if (unlikely(__pyx_t_10 == ((Py_hash_t)-1))) __PYX_ERR(0, 131, __pyx_L9_error)

        /* "python_visible.pxi":132
 *         try:
 *             hash(o)
 *             return True             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             pass
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_True);
        __pyx_r = Py_True;
        goto __pyx_L13_try_return;

        /* "python_visible.pxi":130
 *         # NOT prevent modification to MEMBERS of the tuple that may be mutable
 *         # Hence, check if 'o' has a stable hash - python hash() does this for us
 *         try:             # <<<<<<<<<<<<<<
 *             hash(o)
 *             return True
 */
      }
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "python_visible.pxi":133
 *             hash(o)
 *             return True
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             pass
 *     return False
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_6) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L10_exception_handled;
      }
      goto __pyx_L11_except_error;
      __pyx_L11_except_error:;

      /* "python_visible.pxi":130
 *         # NOT prevent modification to MEMBERS of the tuple that may be mutable
 *         # Hence, check if 'o' has a stable hash - python hash() does this for us
 *         try:             # <<<<<<<<<<<<<<
 *             hash(o)
 *             return True
 */
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L13_try_return:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L0;
      __pyx_L10_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    }

    /* "python_visible.pxi":126
 *     if isfrozen(o):
 *         return True
 *     if type(o) in immutable_types_set:             # <<<<<<<<<<<<<<
 *         # Although 'tuple' is immutable in python, for our purposes, 'tuple' does
 *         # NOT prevent modification to MEMBERS of the tuple that may be mutable
 */
  }

  /* "python_visible.pxi":135
 *         except TypeError:
 *             pass
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "python_visible.pxi":114
 * 
 * 
 * def isimmutable(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isimmutable(o: object) -> bool: 'o' is KNOWN to be immutable
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.isimmutable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":138
 * 
 * 
 * def iswrapped(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     iswrapped(o: object) -> bool:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_23iswrapped(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_22iswrapped, "\n    iswrapped(o: object) -> bool:\n    'o' was created using wrap / freeze / private / protect\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_23iswrapped = {"iswrapped", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_23iswrapped, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_22iswrapped};
static PyObject *__pyx_pw_9pyprotect_9protected_23iswrapped(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iswrapped (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 138, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "iswrapped") < 0)) __PYX_ERR(0, 138, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("iswrapped", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 138, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.iswrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_22iswrapped(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_22iswrapped(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iswrapped", 0);

  /* "python_visible.pxi":143
 *     'o' was created using wrap / freeze / private / protect
 *     '''
 *     return isinstance(o, Wrapped)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":138
 * 
 * 
 * def iswrapped(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     iswrapped(o: object) -> bool:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.iswrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":146
 * 
 * 
 * def isfrozen(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isfrozen(o: object) -> bool: 'o' was created using freeze()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_25isfrozen(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_24isfrozen, "\n    isfrozen(o: object) -> bool: 'o' was created using freeze()\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_25isfrozen = {"isfrozen", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_25isfrozen, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_24isfrozen};
static PyObject *__pyx_pw_9pyprotect_9protected_25isfrozen(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isfrozen (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "isfrozen") < 0)) __PYX_ERR(0, 146, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("isfrozen", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 146, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.isfrozen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_24isfrozen(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_24isfrozen(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isfrozen", 0);

  /* "python_visible.pxi":150
 *     isfrozen(o: object) -> bool: 'o' was created using freeze()
 *     '''
 *     return isinstance(o, (             # <<<<<<<<<<<<<<
 *         Frozen, FrozenPrivate, FrozenPrivacyDict, FrozenProtected,
 *     ))
 */
  __Pyx_XDECREF(__pyx_r);

  /* "python_visible.pxi":151
 *     '''
 *     return isinstance(o, (
 *         Frozen, FrozenPrivate, FrozenPrivacyDict, FrozenProtected,             # <<<<<<<<<<<<<<
 *     ))
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Frozen); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "python_visible.pxi":150
 *     isfrozen(o: object) -> bool: 'o' was created using freeze()
 *     '''
 *     return isinstance(o, (             # <<<<<<<<<<<<<<
 *         Frozen, FrozenPrivate, FrozenPrivacyDict, FrozenProtected,
 *     ))
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_FrozenPrivate); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "python_visible.pxi":151
 *     '''
 *     return isinstance(o, (
 *         Frozen, FrozenPrivate, FrozenPrivacyDict, FrozenProtected,             # <<<<<<<<<<<<<<
 *     ))
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "python_visible.pxi":150
 *     isfrozen(o: object) -> bool: 'o' was created using freeze()
 *     '''
 *     return isinstance(o, (             # <<<<<<<<<<<<<<
 *         Frozen, FrozenPrivate, FrozenPrivacyDict, FrozenProtected,
 *     ))
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_FrozenProtected); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":146
 * 
 * 
 * def isfrozen(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isfrozen(o: object) -> bool: 'o' was created using freeze()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.isfrozen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":155
 * 
 * 
 * def isprivate(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprivate(o: object) -> bool: 'o' was created using private()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_27isprivate(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_26isprivate, "\n    isprivate(o: object) -> bool: 'o' was created using private()\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_27isprivate = {"isprivate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_27isprivate, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_26isprivate};
static PyObject *__pyx_pw_9pyprotect_9protected_27isprivate(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isprivate (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "isprivate") < 0)) __PYX_ERR(0, 155, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("isprivate", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 155, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.isprivate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_26isprivate(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_26isprivate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isprivate", 0);

  /* "python_visible.pxi":159
 *     isprivate(o: object) -> bool: 'o' was created using private()
 *     '''
 *     return isinstance(o, (             # <<<<<<<<<<<<<<
 *         Private,
 *         FrozenPrivate,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "python_visible.pxi":160
 *     '''
 *     return isinstance(o, (
 *         Private,             # <<<<<<<<<<<<<<
 *         FrozenPrivate,
 *     ))
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Private); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "python_visible.pxi":161
 *     return isinstance(o, (
 *         Private,
 *         FrozenPrivate,             # <<<<<<<<<<<<<<
 *     ))
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_FrozenPrivate); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;

  /* "python_visible.pxi":159
 *     isprivate(o: object) -> bool: 'o' was created using private()
 *     '''
 *     return isinstance(o, (             # <<<<<<<<<<<<<<
 *         Private,
 *         FrozenPrivate,
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":155
 * 
 * 
 * def isprivate(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprivate(o: object) -> bool: 'o' was created using private()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.isprivate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":165
 * 
 * 
 * def isprotected(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprotected(o: object) -> bool: 'o' was created using protect()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_29isprotected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_28isprotected, "\n    isprotected(o: object) -> bool: 'o' was created using protect()\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_29isprotected = {"isprotected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_29isprotected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_28isprotected};
static PyObject *__pyx_pw_9pyprotect_9protected_29isprotected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isprotected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 165, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "isprotected") < 0)) __PYX_ERR(0, 165, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("isprotected", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.isprotected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_28isprotected(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_28isprotected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isprotected", 0);

  /* "python_visible.pxi":169
 *     isprotected(o: object) -> bool: 'o' was created using protect()
 *     '''
 *     return isinstance(o, (             # <<<<<<<<<<<<<<
 *         Protected,
 *         FrozenProtected,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "python_visible.pxi":170
 *     '''
 *     return isinstance(o, (
 *         Protected,             # <<<<<<<<<<<<<<
 *         FrozenProtected,
 *     ))
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Protected); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "python_visible.pxi":171
 *     return isinstance(o, (
 *         Protected,
 *         FrozenProtected,             # <<<<<<<<<<<<<<
 *     ))
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_FrozenProtected); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;

  /* "python_visible.pxi":169
 *     isprotected(o: object) -> bool: 'o' was created using protect()
 *     '''
 *     return isinstance(o, (             # <<<<<<<<<<<<<<
 *         Protected,
 *         FrozenProtected,
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":165
 * 
 * 
 * def isprotected(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprotected(o: object) -> bool: 'o' was created using protect()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.isprotected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":175
 * 
 * 
 * def isreadonly(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isreadonly(o: object, a: str) -> bool:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_31isreadonly(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_30isreadonly, "\n    isreadonly(o: object, a: str) -> bool:\n    Returns-->bool: True IFF 'o' is wrapped AND 'o' makes arribute 'a'\n        read-only if present in wrapped object\n    This represents RULE of wrapped object - does not guarantee\n    that WRAPPED OBJECT has attribute 'a' or that setting attribute\n    'a' in object 'o' will not raise any exception\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_31isreadonly = {"isreadonly", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_31isreadonly, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_30isreadonly};
static PyObject *__pyx_pw_9pyprotect_9protected_31isreadonly(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_a = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isreadonly (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_a,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_a,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_a)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("isreadonly", 1, 2, 2, 1); __PYX_ERR(0, 175, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "isreadonly") < 0)) __PYX_ERR(0, 175, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_o = values[0];
    __pyx_v_a = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("isreadonly", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 175, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.isreadonly", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_a), (&PyString_Type), 0, "a", 1))) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_r = __pyx_pf_9pyprotect_9protected_30isreadonly(__pyx_self, __pyx_v_o, __pyx_v_a);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_30isreadonly(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isreadonly", 0);

  /* "python_visible.pxi":184
 *     'a' in object 'o' will not raise any exception
 *     '''
 *     try:             # <<<<<<<<<<<<<<
 *         if isimmutable(o):
 *             return True
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "python_visible.pxi":185
 *     '''
 *     try:
 *         if isimmutable(o):             # <<<<<<<<<<<<<<
 *             return True
 *     except:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_isimmutable); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 185, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_o};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 185, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 185, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_8) {

        /* "python_visible.pxi":186
 *     try:
 *         if isimmutable(o):
 *             return True             # <<<<<<<<<<<<<<
 *     except:
 *         pass
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_True);
        __pyx_r = Py_True;
        goto __pyx_L7_try_return;

        /* "python_visible.pxi":185
 *     '''
 *     try:
 *         if isimmutable(o):             # <<<<<<<<<<<<<<
 *             return True
 *     except:
 */
      }

      /* "python_visible.pxi":184
 *     'a' in object 'o' will not raise any exception
 *     '''
 *     try:             # <<<<<<<<<<<<<<
 *         if isimmutable(o):
 *             return True
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "python_visible.pxi":187
 *         if isimmutable(o):
 *             return True
 *     except:             # <<<<<<<<<<<<<<
 *         pass
 *     if isprivate(o) or isprotected(o):
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    __pyx_L7_try_return:;

    /* "python_visible.pxi":184
 *     'a' in object 'o' will not raise any exception
 *     '''
 *     try:             # <<<<<<<<<<<<<<
 *         if isimmutable(o):
 *             return True
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "python_visible.pxi":189
 *     except:
 *         pass
 *     if isprivate(o) or isprotected(o):             # <<<<<<<<<<<<<<
 *         return not getattr(o, PROT_ATTR_NAME).testop(a, 'w')
 *     else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_isprivate); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_o};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_9) {
  } else {
    __pyx_t_8 = __pyx_t_9;
    goto __pyx_L11_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_isprotected); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_o};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = __pyx_t_9;
  __pyx_L11_bool_binop_done:;
  if (__pyx_t_8) {

    /* "python_visible.pxi":190
 *         pass
 *     if isprivate(o) or isprotected(o):
 *         return not getattr(o, PROT_ATTR_NAME).testop(a, 'w')             # <<<<<<<<<<<<<<
 *     else:
 *         return False
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_testop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_a, __pyx_n_s_w};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyBool_FromLong((!__pyx_t_8)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":189
 *     except:
 *         pass
 *     if isprivate(o) or isprotected(o):             # <<<<<<<<<<<<<<
 *         return not getattr(o, PROT_ATTR_NAME).testop(a, 'w')
 *     else:
 */
  }

  /* "python_visible.pxi":192
 *         return not getattr(o, PROT_ATTR_NAME).testop(a, 'w')
 *     else:
 *         return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "python_visible.pxi":175
 * 
 * 
 * def isreadonly(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isreadonly(o: object, a: str) -> bool:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.isreadonly", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":195
 * 
 * 
 * def isvisible(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isvisible(o: object, a: str) -> bool:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_33isvisible(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_32isvisible, "\n    isvisible(o: object, a: str) -> bool:\n    Returns-->bool: False IFF 'o' is wrapped AND 'o' makes arribute 'a'\n        invisible if present in wrapped object\n    This represents RULE of wrapped object - does not guarantee\n    that WRAPPED OBJECT has attribute 'a' or that accessing attribute\n    'a' in object 'o' will not raise any exception\n\n    If 'o' is not a wrapped object, unconditionally returns False\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_33isvisible = {"isvisible", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_33isvisible, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_32isvisible};
static PyObject *__pyx_pw_9pyprotect_9protected_33isvisible(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_a = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isvisible (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_a,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_a,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_a)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("isvisible", 1, 2, 2, 1); __PYX_ERR(0, 195, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "isvisible") < 0)) __PYX_ERR(0, 195, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_o = values[0];
    __pyx_v_a = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("isvisible", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 195, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.isvisible", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_a), (&PyString_Type), 0, "a", 1))) __PYX_ERR(0, 195, __pyx_L1_error)
  __pyx_r = __pyx_pf_9pyprotect_9protected_32isvisible(__pyx_self, __pyx_v_o, __pyx_v_a);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_32isvisible(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isvisible", 0);

  /* "python_visible.pxi":206
 *     If 'o' is not a wrapped object, unconditionally returns False
 *     '''
 *     if not iswrapped(o):             # <<<<<<<<<<<<<<
 *         return False
 *     return getattr(o, PROT_ATTR_NAME).testop(a, 'r')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "python_visible.pxi":207
 *     '''
 *     if not iswrapped(o):
 *         return False             # <<<<<<<<<<<<<<
 *     return getattr(o, PROT_ATTR_NAME).testop(a, 'r')
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "python_visible.pxi":206
 *     If 'o' is not a wrapped object, unconditionally returns False
 *     '''
 *     if not iswrapped(o):             # <<<<<<<<<<<<<<
 *         return False
 *     return getattr(o, PROT_ATTR_NAME).testop(a, 'r')
 */
  }

  /* "python_visible.pxi":208
 *     if not iswrapped(o):
 *         return False
 *     return getattr(o, PROT_ATTR_NAME).testop(a, 'r')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_testop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_a, __pyx_n_s_r};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":195
 * 
 * 
 * def isvisible(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isvisible(o: object, a: str) -> bool:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.isvisible", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":216
 * 
 * 
 * def wrap(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     wrap(o: object) -> object:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_35wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_34wrap, "\n    wrap(o: object) -> object:\n    Returns: instance of Wrapped\n\n    Wrapped:\n        - Should behave just like the wrapped object, except\n          following attributes cannot be modified:\n            'getattr, __getattribute__',\n            '__delattr__', '__setattr__', '__slots__',\n        - Explicitly does NOT support pickling, and will raise\n          ProtectionError\n        - Does NOT protect CLASS of wrapped object from modification\n        - Does NOT protect __dict__ or __slots__\n\n    Useful for testing if wrapping is failing for a particular type of object\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_35wrap = {"wrap", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_35wrap, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_34wrap};
static PyObject *__pyx_pw_9pyprotect_9protected_35wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 216, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "wrap") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wrap", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 216, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_34wrap(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_34wrap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap", 0);

  /* "python_visible.pxi":233
 *     Useful for testing if wrapping is failing for a particular type of object
 *     '''
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         # Do not wrap twice
 *         return o
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":235
 *     if iswrapped(o):
 *         # Do not wrap twice
 *         return o             # <<<<<<<<<<<<<<
 *     return Wrapped(o, frozen=False)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_o);
    __pyx_r = __pyx_v_o;
    goto __pyx_L0;

    /* "python_visible.pxi":233
 *     Useful for testing if wrapping is failing for a particular type of object
 *     '''
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         # Do not wrap twice
 *         return o
 */
  }

  /* "python_visible.pxi":236
 *         # Do not wrap twice
 *         return o
 *     return Wrapped(o, frozen=False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_o);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_frozen, Py_False) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":216
 * 
 * 
 * def wrap(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     wrap(o: object) -> object:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":239
 * 
 * 
 * def freeze(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     freeze(o: object) -> object:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_37freeze(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_36freeze, "\n    freeze(o: object) -> object:\n    Returns: Instance of Frozen | FrozenPrivacyDict | FrozenPrivate |\n        FrozenProtected, depending on what 'o' is\n\n    Object returned prevents modification of ANY attribute\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_37freeze = {"freeze", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_37freeze, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_36freeze};
static PyObject *__pyx_pw_9pyprotect_9protected_37freeze(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("freeze (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "freeze") < 0)) __PYX_ERR(0, 239, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("freeze", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 239, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.freeze", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_36freeze(__pyx_self, __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_36freeze(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("freeze", 0);

  /* "python_visible.pxi":247
 *     Object returned prevents modification of ANY attribute
 *     '''
 *     if isfrozen(o):             # <<<<<<<<<<<<<<
 *         # Never freeze twice
 *         return o
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isfrozen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":249
 *     if isfrozen(o):
 *         # Never freeze twice
 *         return o             # <<<<<<<<<<<<<<
 *     elif isimmutable(o):
 *         # Object is KNOWN to be immutable - return as-is
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_o);
    __pyx_r = __pyx_v_o;
    goto __pyx_L0;

    /* "python_visible.pxi":247
 *     Object returned prevents modification of ANY attribute
 *     '''
 *     if isfrozen(o):             # <<<<<<<<<<<<<<
 *         # Never freeze twice
 *         return o
 */
  }

  /* "python_visible.pxi":250
 *         # Never freeze twice
 *         return o
 *     elif isimmutable(o):             # <<<<<<<<<<<<<<
 *         # Object is KNOWN to be immutable - return as-is
 *         return o
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isimmutable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":252
 *     elif isimmutable(o):
 *         # Object is KNOWN to be immutable - return as-is
 *         return o             # <<<<<<<<<<<<<<
 *     # Must freeze
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_o);
    __pyx_r = __pyx_v_o;
    goto __pyx_L0;

    /* "python_visible.pxi":250
 *         # Never freeze twice
 *         return o
 *     elif isimmutable(o):             # <<<<<<<<<<<<<<
 *         # Object is KNOWN to be immutable - return as-is
 *         return o
 */
  }

  /* "python_visible.pxi":256
 * 
 *     # If Wrapped, avoid double wrapping
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).freeze()
 *     return Frozen(o)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":257
 *     # If Wrapped, avoid double wrapping
 *     if iswrapped(o):
 *         return getattr(o, PROT_ATTR_NAME).freeze()             # <<<<<<<<<<<<<<
 *     return Frozen(o)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_freeze); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_3, };
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":256
 * 
 *     # If Wrapped, avoid double wrapping
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).freeze()
 *     return Frozen(o)
 */
  }

  /* "python_visible.pxi":258
 *     if iswrapped(o):
 *         return getattr(o, PROT_ATTR_NAME).freeze()
 *     return Frozen(o)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9pyprotect_9protected_Frozen), __pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":239
 * 
 * 
 * def freeze(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     freeze(o: object) -> object:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.freeze", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":261
 * 
 * 
 * def private(o: object, frozen: bool = False) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     private(o: object, frozen: bool = False) -> object:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_39private(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_38private, "\n    private(o: object, frozen: bool = False) -> object:\n        Returns: Instance of FrozenPrivate if frozen; Private otherwise\n\n    Private:\n        - Cannot access traditionally 'private' mangled python attributes\n        - Cannot access any attribute not exported by dir(o)\n        - Cannot access any unmangled double '_' attributes\n        - Cannot modify traditionally private attributes (form '_var')\n        - Cannot modify __class__ of wrapped object\n        - Cannot modify __dict__ of wrapped object\n        - Cannot modify __slots__ of wrapped object\n        - Cannot add or delete attributes\n\n    FrozenPrivate:\n        Features of Private PLUS prevents modification of ANY attribute\n\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_39private = {"private", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_39private, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_38private};
static PyObject *__pyx_pw_9pyprotect_9protected_39private(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_frozen = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("private (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,0};
    #endif
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 261, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_frozen);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 261, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "private") < 0)) __PYX_ERR(0, 261, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_o = values[0];
    __pyx_v_frozen = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("private", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 261, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.private", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_38private(__pyx_self, __pyx_v_o, __pyx_v_frozen);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_38private(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("private", 0);
  __Pyx_INCREF(__pyx_v_frozen);

  /* "python_visible.pxi":281
 *     '''
 *     # Avoid double-wrapping
 *     if frozen or isfrozen(o):             # <<<<<<<<<<<<<<
 *         frozen = True
 *     if iswrapped(o):
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_frozen); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 281, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_isfrozen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_o};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "python_visible.pxi":282
 *     # Avoid double-wrapping
 *     if frozen or isfrozen(o):
 *         frozen = True             # <<<<<<<<<<<<<<
 *     if iswrapped(o):
 *         if isprotected(o):
 */
    __Pyx_INCREF(Py_True);
    __Pyx_DECREF_SET(__pyx_v_frozen, Py_True);

    /* "python_visible.pxi":281
 *     '''
 *     # Avoid double-wrapping
 *     if frozen or isfrozen(o):             # <<<<<<<<<<<<<<
 *         frozen = True
 *     if iswrapped(o):
 */
  }

  /* "python_visible.pxi":283
 *     if frozen or isfrozen(o):
 *         frozen = True
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         if isprotected(o):
 *             return protect(o, frozen=True)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_o};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "python_visible.pxi":284
 *         frozen = True
 *     if iswrapped(o):
 *         if isprotected(o):             # <<<<<<<<<<<<<<
 *             return protect(o, frozen=True)
 *         return getattr(o, PROT_ATTR_NAME).private()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_isprotected); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_o};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "python_visible.pxi":285
 *     if iswrapped(o):
 *         if isprotected(o):
 *             return protect(o, frozen=True)             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).private()
 *     else:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_protect); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_o);
      __Pyx_GIVEREF(__pyx_v_o);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_o);
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_frozen, Py_True) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "python_visible.pxi":284
 *         frozen = True
 *     if iswrapped(o):
 *         if isprotected(o):             # <<<<<<<<<<<<<<
 *             return protect(o, frozen=True)
 *         return getattr(o, PROT_ATTR_NAME).private()
 */
    }

    /* "python_visible.pxi":286
 *         if isprotected(o):
 *             return protect(o, frozen=True)
 *         return getattr(o, PROT_ATTR_NAME).private()             # <<<<<<<<<<<<<<
 *     else:
 *         if frozen:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_private); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":283
 *     if frozen or isfrozen(o):
 *         frozen = True
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         if isprotected(o):
 *             return protect(o, frozen=True)
 */
  }

  /* "python_visible.pxi":288
 *         return getattr(o, PROT_ATTR_NAME).private()
 *     else:
 *         if frozen:             # <<<<<<<<<<<<<<
 *             return FrozenPrivate(o)
 *         else:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_frozen); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 288, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "python_visible.pxi":289
 *     else:
 *         if frozen:
 *             return FrozenPrivate(o)             # <<<<<<<<<<<<<<
 *         else:
 *             return Private(o)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivate), __pyx_v_o); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;

      /* "python_visible.pxi":288
 *         return getattr(o, PROT_ATTR_NAME).private()
 *     else:
 *         if frozen:             # <<<<<<<<<<<<<<
 *             return FrozenPrivate(o)
 *         else:
 */
    }

    /* "python_visible.pxi":291
 *             return FrozenPrivate(o)
 *         else:
 *             return Private(o)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9pyprotect_9protected_Private), __pyx_v_o); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;
    }
  }

  /* "python_visible.pxi":261
 * 
 * 
 * def private(o: object, frozen: bool = False) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     private(o: object, frozen: bool = False) -> object:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.private", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frozen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":294
 * 
 * 
 * def protect(             # <<<<<<<<<<<<<<
 *     o: object,
 *     frozen: bool = False, dynamic: bool = True,
 */

static PyObject *__pyx_pf_9pyprotect_9protected_76__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);

  /* "python_visible.pxi":298
 *     frozen: bool = False, dynamic: bool = True,
 *     hide_private: bool = False,
 *     ro_data: bool = False, ro_method: bool = True,             # <<<<<<<<<<<<<<
 *     ro=[], rw=[], hide=[],
 * ):
 */
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_False));
  __Pyx_GIVEREF(((PyObject *)Py_False));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_False));
  __Pyx_INCREF(((PyObject *)Py_True));
  __Pyx_GIVEREF(((PyObject *)Py_True));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_True));
  __Pyx_INCREF(((PyObject *)Py_False));
  __Pyx_GIVEREF(((PyObject *)Py_False));
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)Py_False));
  __Pyx_INCREF(((PyObject *)Py_False));
  __Pyx_GIVEREF(((PyObject *)Py_False));
  PyTuple_SET_ITEM(__pyx_t_1, 3, ((PyObject *)Py_False));
  __Pyx_INCREF(((PyObject *)Py_True));
  __Pyx_GIVEREF(((PyObject *)Py_True));
  PyTuple_SET_ITEM(__pyx_t_1, 4, ((PyObject *)Py_True));
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_ro);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_ro);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_ro);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_rw);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_rw);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_rw);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_hide);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_hide);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_hide);

  /* "python_visible.pxi":294
 * 
 * 
 * def protect(             # <<<<<<<<<<<<<<
 *     o: object,
 *     frozen: bool = False, dynamic: bool = True,
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_41protect(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_40protect, "\n    protect(\n        o: object,\n        frozen: bool = False, dynamic: bool = True,\n        hide_private: bool = False,\n        ro_data: bool = False, ro_method: bool = True,\n        ro=[], rw=[], hide=[],\n    ):\n\n    o: object to be wrapped\n    frozen: bool: No attribute can be modified\n        PLUS: if 'o' is NOT a module, results returned by methods,\n        including __call__ will be frozen\n    dynamic: bool: Attribute additions, deletions, type changes in wrapped\n        object are automatically considered by hide_private, ro_data,\n        ro_method, ro, rw, hide\n        If dynamic is False, it is a pledge that attributes of wrapped\n        object will not change, and visibility and mutability rules of\n        WRAPPING object use a cache to make them faster.\n        Rules imposed by Private() are always dynamic\n    hide_private: bool: Private vars (_var) will be hidden\n    ro_data: bool: Data attributes cannot be deleted or assigned to\n    ro_method: bool: Method attributes cannot be deleted or assigned to\n    ro: list of str: attributes that will be read-only\n    rw: list of str: attributes that will be read-write\n        Overrides 'ro_*'\n    hide: list of str: attributes that will be hidden\n\n    Returns-->Instance of FrozenProtected if frozen; Protected otherwise\n\n    Protected:\n        Features of Private PLUS additional restrictions on:\n            - Which attributes are VISIBLE\n            - Which attributes are WRITEABLE\n\n    FrozenProtected:\n        Features of Protected PLUS prevents modification of ANY attribute\n\n    Default settings:\n    Features of Private:\n        - Cannot access traditionally 'private' mangled python attributes\n        - Cannot access any attribute not exported by dir(o)\n        - Cannot access any unmangled double '_' attributes\n        - Cannot modify traditionally private attributes (form '_var')\n        - Cannot modify __class__ of wrapped object\n        - Cannot modify __dict__ of ""wrapped object\n        - Cannot modify __slots__ of wrapped object\n        - Cannot add or delete attributes\n    PLUS:\n        - Methods are readonly - cannot be deleted or assigned to\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_41protect = {"protect", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_41protect, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9pyprotect_9protected_40protect};
static PyObject *__pyx_pw_9pyprotect_9protected_41protect(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_frozen = 0;
  PyObject *__pyx_v_dynamic = 0;
  PyObject *__pyx_v_hide_private = 0;
  PyObject *__pyx_v_ro_data = 0;
  PyObject *__pyx_v_ro_method = 0;
  PyObject *__pyx_v_ro = 0;
  PyObject *__pyx_v_rw = 0;
  PyObject *__pyx_v_hide = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("protect (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,&__pyx_n_s_dynamic,&__pyx_n_s_hide_private,&__pyx_n_s_ro_data,&__pyx_n_s_ro_method,&__pyx_n_s_ro,&__pyx_n_s_rw,&__pyx_n_s_hide,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,&__pyx_n_s_dynamic,&__pyx_n_s_hide_private,&__pyx_n_s_ro_data,&__pyx_n_s_ro_method,&__pyx_n_s_ro,&__pyx_n_s_rw,&__pyx_n_s_hide,0};
    #endif
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[1] = ((PyObject *)((PyObject *)Py_False));
    values[2] = ((PyObject *)((PyObject *)Py_True));
    values[3] = ((PyObject *)((PyObject *)Py_False));
    values[4] = ((PyObject *)((PyObject *)Py_False));
    values[5] = ((PyObject *)((PyObject *)Py_True));
    values[6] = __pyx_dynamic_args->__pyx_arg_ro;
    values[7] = __pyx_dynamic_args->__pyx_arg_rw;
    values[8] = __pyx_dynamic_args->__pyx_arg_hide;
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  9: values[8] = __Pyx_Arg_FASTCALL(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_FASTCALL(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_frozen);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dynamic);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_hide_private);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ro_data);
          if (value) { values[4] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ro_method);
          if (value) { values[5] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ro);
          if (value) { values[6] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rw);
          if (value) { values[7] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_hide);
          if (value) { values[8] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "protect") < 0)) __PYX_ERR(0, 294, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  9: values[8] = __Pyx_Arg_FASTCALL(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_FASTCALL(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_o = values[0];
    __pyx_v_frozen = values[1];
    __pyx_v_dynamic = values[2];
    __pyx_v_hide_private = values[3];
    __pyx_v_ro_data = values[4];
    __pyx_v_ro_method = values[5];
    __pyx_v_ro = values[6];
    __pyx_v_rw = values[7];
    __pyx_v_hide = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("protect", 0, 1, 9, __pyx_nargs); __PYX_ERR(0, 294, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.protect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_40protect(__pyx_self, __pyx_v_o, __pyx_v_frozen, __pyx_v_dynamic, __pyx_v_hide_private, __pyx_v_ro_data, __pyx_v_ro_method, __pyx_v_ro, __pyx_v_rw, __pyx_v_hide);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_40protect(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen, PyObject *__pyx_v_dynamic, PyObject *__pyx_v_hide_private, PyObject *__pyx_v_ro_data, PyObject *__pyx_v_ro_method, PyObject *__pyx_v_ro, PyObject *__pyx_v_rw, PyObject *__pyx_v_hide) {
  PyObject *__pyx_v_kwargs = NULL;
  PyObject *__pyx_v_kw1 = NULL;
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_kw2 = NULL;
  PyObject *__pyx_v_rules = NULL;
  PyObject *__pyx_v_want_frozen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protect", 0);

  /* "python_visible.pxi":353
 *     '''
 *     kwargs = {
 *         'frozen': frozen,             # <<<<<<<<<<<<<<
 *         'hide_private': hide_private,
 *         'ro_data': ro_data,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_frozen, __pyx_v_frozen) < 0) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "python_visible.pxi":354
 *     kwargs = {
 *         'frozen': frozen,
 *         'hide_private': hide_private,             # <<<<<<<<<<<<<<
 *         'ro_data': ro_data,
 *         'ro_method': ro_method,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_hide_private, __pyx_v_hide_private) < 0) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "python_visible.pxi":355
 *         'frozen': frozen,
 *         'hide_private': hide_private,
 *         'ro_data': ro_data,             # <<<<<<<<<<<<<<
 *         'ro_method': ro_method,
 *         'ro': ro,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_ro_data, __pyx_v_ro_data) < 0) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "python_visible.pxi":356
 *         'hide_private': hide_private,
 *         'ro_data': ro_data,
 *         'ro_method': ro_method,             # <<<<<<<<<<<<<<
 *         'ro': ro,
 *         'rw': rw,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_ro_method, __pyx_v_ro_method) < 0) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "python_visible.pxi":357
 *         'ro_data': ro_data,
 *         'ro_method': ro_method,
 *         'ro': ro,             # <<<<<<<<<<<<<<
 *         'rw': rw,
 *         'hide': hide,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_ro, __pyx_v_ro) < 0) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "python_visible.pxi":358
 *         'ro_method': ro_method,
 *         'ro': ro,
 *         'rw': rw,             # <<<<<<<<<<<<<<
 *         'hide': hide,
 *         'dynamic': dynamic,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_rw, __pyx_v_rw) < 0) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "python_visible.pxi":359
 *         'ro': ro,
 *         'rw': rw,
 *         'hide': hide,             # <<<<<<<<<<<<<<
 *         'dynamic': dynamic,
 *     }
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_hide, __pyx_v_hide) < 0) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "python_visible.pxi":360
 *         'rw': rw,
 *         'hide': hide,
 *         'dynamic': dynamic,             # <<<<<<<<<<<<<<
 *     }
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dynamic, __pyx_v_dynamic) < 0) __PYX_ERR(0, 353, __pyx_L1_error)
  __pyx_v_kwargs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "python_visible.pxi":364
 * 
 *     # Avoid double-wrapping
 *     if isprotected(o):             # <<<<<<<<<<<<<<
 *         kw1 = getattr(o, PROT_ATTR_NAME).rules.get('kwargs', {})
 *         d = {}
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isprotected); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":365
 *     # Avoid double-wrapping
 *     if isprotected(o):
 *         kw1 = getattr(o, PROT_ATTR_NAME).rules.get('kwargs', {})             # <<<<<<<<<<<<<<
 *         d = {}
 *         for (k, v) in kw1.items():
 */
    __pyx_t_2 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_rules); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_n_s_kwargs, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_kw1 = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "python_visible.pxi":366
 *     if isprotected(o):
 *         kw1 = getattr(o, PROT_ATTR_NAME).rules.get('kwargs', {})
 *         d = {}             # <<<<<<<<<<<<<<
 *         for (k, v) in kw1.items():
 *             d[k] = v
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_d = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "python_visible.pxi":367
 *         kw1 = getattr(o, PROT_ATTR_NAME).rules.get('kwargs', {})
 *         d = {}
 *         for (k, v) in kw1.items():             # <<<<<<<<<<<<<<
 *             d[k] = v
 *         kw1 = d
 */
    __pyx_t_7 = 0;
    if (unlikely(__pyx_v_kw1 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 367, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_kw1, 0, __pyx_n_s_items, (&__pyx_t_8), (&__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_8, &__pyx_t_7, &__pyx_t_3, &__pyx_t_2, NULL, __pyx_t_4);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "python_visible.pxi":368
 *         d = {}
 *         for (k, v) in kw1.items():
 *             d[k] = v             # <<<<<<<<<<<<<<
 *         kw1 = d
 *         kw2 = dict(kwargs)
 */
      if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_v_k, __pyx_v_v) < 0))) __PYX_ERR(0, 368, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "python_visible.pxi":369
 *         for (k, v) in kw1.items():
 *             d[k] = v
 *         kw1 = d             # <<<<<<<<<<<<<<
 *         kw2 = dict(kwargs)
 *         kwargs = protected_merge_kwargs(kw1, kw2)
 */
    __Pyx_INCREF(__pyx_v_d);
    __Pyx_DECREF_SET(__pyx_v_kw1, __pyx_v_d);

    /* "python_visible.pxi":370
 *             d[k] = v
 *         kw1 = d
 *         kw2 = dict(kwargs)             # <<<<<<<<<<<<<<
 *         kwargs = protected_merge_kwargs(kw1, kw2)
 *         assert(isinstance(kwargs, dict))
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_kw2 = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "python_visible.pxi":371
 *         kw1 = d
 *         kw2 = dict(kwargs)
 *         kwargs = protected_merge_kwargs(kw1, kw2)             # <<<<<<<<<<<<<<
 *         assert(isinstance(kwargs, dict))
 *     rules = dict(protected_rules_from_kwargs(kwargs))
 */
    if (!(likely(PyDict_CheckExact(__pyx_v_kw1)) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_kw1))) __PYX_ERR(0, 371, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected_protected_merge_kwargs(((PyObject*)__pyx_v_kw1), __pyx_v_kw2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_kwargs, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "python_visible.pxi":372
 *         kw2 = dict(kwargs)
 *         kwargs = protected_merge_kwargs(kw1, kw2)
 *         assert(isinstance(kwargs, dict))             # <<<<<<<<<<<<<<
 *     rules = dict(protected_rules_from_kwargs(kwargs))
 *     assert(isinstance(rules, dict))
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_5 = PyDict_Check(__pyx_v_kwargs); 
      if (unlikely(!__pyx_t_5)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 372, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 372, __pyx_L1_error)
    #endif

    /* "python_visible.pxi":364
 * 
 *     # Avoid double-wrapping
 *     if isprotected(o):             # <<<<<<<<<<<<<<
 *         kw1 = getattr(o, PROT_ATTR_NAME).rules.get('kwargs', {})
 *         d = {}
 */
  }

  /* "python_visible.pxi":373
 *         kwargs = protected_merge_kwargs(kw1, kw2)
 *         assert(isinstance(kwargs, dict))
 *     rules = dict(protected_rules_from_kwargs(kwargs))             # <<<<<<<<<<<<<<
 *     assert(isinstance(rules, dict))
 *     want_frozen = bool(rules.get('frozen', False)) or isfrozen(o)
 */
  __pyx_t_1 = __pyx_f_9pyprotect_9protected_protected_rules_from_kwargs(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rules = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "python_visible.pxi":374
 *         assert(isinstance(kwargs, dict))
 *     rules = dict(protected_rules_from_kwargs(kwargs))
 *     assert(isinstance(rules, dict))             # <<<<<<<<<<<<<<
 *     want_frozen = bool(rules.get('frozen', False)) or isfrozen(o)
 *     if want_frozen and not isfrozen(o):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_5 = PyDict_Check(__pyx_v_rules); 
    if (unlikely(!__pyx_t_5)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 374, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 374, __pyx_L1_error)
  #endif

  /* "python_visible.pxi":375
 *     rules = dict(protected_rules_from_kwargs(kwargs))
 *     assert(isinstance(rules, dict))
 *     want_frozen = bool(rules.get('frozen', False)) or isfrozen(o)             # <<<<<<<<<<<<<<
 *     if want_frozen and not isfrozen(o):
 *         # Frozen objects remain frozen
 */
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_rules, __pyx_n_s_frozen, Py_False); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 375, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L6_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isfrozen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_L6_bool_binop_done:;
  __pyx_v_want_frozen = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "python_visible.pxi":376
 *     assert(isinstance(rules, dict))
 *     want_frozen = bool(rules.get('frozen', False)) or isfrozen(o)
 *     if want_frozen and not isfrozen(o):             # <<<<<<<<<<<<<<
 *         # Frozen objects remain frozen
 *         rules['frozen'] = True
 */
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_want_frozen); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 376, __pyx_L1_error)
  if (__pyx_t_10) {
  } else {
    __pyx_t_5 = __pyx_t_10;
    goto __pyx_L9_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_isfrozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = (!__pyx_t_10);
  __pyx_t_5 = __pyx_t_11;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_5) {

    /* "python_visible.pxi":378
 *     if want_frozen and not isfrozen(o):
 *         # Frozen objects remain frozen
 *         rules['frozen'] = True             # <<<<<<<<<<<<<<
 *     if iswrapped(o):
 *         return getattr(o, PROT_ATTR_NAME).protect(rules)
 */
    if (unlikely((PyDict_SetItem(__pyx_v_rules, __pyx_n_s_frozen, Py_True) < 0))) __PYX_ERR(0, 378, __pyx_L1_error)

    /* "python_visible.pxi":376
 *     assert(isinstance(rules, dict))
 *     want_frozen = bool(rules.get('frozen', False)) or isfrozen(o)
 *     if want_frozen and not isfrozen(o):             # <<<<<<<<<<<<<<
 *         # Frozen objects remain frozen
 *         rules['frozen'] = True
 */
  }

  /* "python_visible.pxi":379
 *         # Frozen objects remain frozen
 *         rules['frozen'] = True
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).protect(rules)
 *     else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_5) {

    /* "python_visible.pxi":380
 *         rules['frozen'] = True
 *     if iswrapped(o):
 *         return getattr(o, PROT_ATTR_NAME).protect(rules)             # <<<<<<<<<<<<<<
 *     else:
 *         if want_frozen:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_v_o, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_protect); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_rules};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "python_visible.pxi":379
 *         # Frozen objects remain frozen
 *         rules['frozen'] = True
 *     if iswrapped(o):             # <<<<<<<<<<<<<<
 *         return getattr(o, PROT_ATTR_NAME).protect(rules)
 *     else:
 */
  }

  /* "python_visible.pxi":382
 *         return getattr(o, PROT_ATTR_NAME).protect(rules)
 *     else:
 *         if want_frozen:             # <<<<<<<<<<<<<<
 *             return FrozenProtected(o, rules)
 *         else:
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_want_frozen); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 382, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "python_visible.pxi":383
 *     else:
 *         if want_frozen:
 *             return FrozenProtected(o, rules)             # <<<<<<<<<<<<<<
 *         else:
 *             return Protected(o, rules)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 383, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_o);
      __Pyx_GIVEREF(__pyx_v_o);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_o);
      __Pyx_INCREF(__pyx_v_rules);
      __Pyx_GIVEREF(__pyx_v_rules);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_rules);
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenProtected), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "python_visible.pxi":382
 *         return getattr(o, PROT_ATTR_NAME).protect(rules)
 *     else:
 *         if want_frozen:             # <<<<<<<<<<<<<<
 *             return FrozenProtected(o, rules)
 *         else:
 */
    }

    /* "python_visible.pxi":385
 *             return FrozenProtected(o, rules)
 *         else:
 *             return Protected(o, rules)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_o);
      __Pyx_GIVEREF(__pyx_v_o);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_o);
      __Pyx_INCREF(__pyx_v_rules);
      __Pyx_GIVEREF(__pyx_v_rules);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_rules);
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }

  /* "python_visible.pxi":294
 * 
 * 
 * def protect(             # <<<<<<<<<<<<<<
 *     o: object,
 *     frozen: bool = False, dynamic: bool = True,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.protect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XDECREF(__pyx_v_kw1);
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_kw2);
  __Pyx_XDECREF(__pyx_v_rules);
  __Pyx_XDECREF(__pyx_v_want_frozen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":392
 * # ------------------------------------------------------------------------
 * 
 * def never_writeable():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable() -> set(str): Attributes that are never writeable
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_43never_writeable(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_42never_writeable, "\n    never_writeable() -> set(str): Attributes that are never writeable\n    in object 'o' if iswrapped(o)\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_43never_writeable = {"never_writeable", (PyCFunction)__pyx_pw_9pyprotect_9protected_43never_writeable, METH_NOARGS, __pyx_doc_9pyprotect_9protected_42never_writeable};
static PyObject *__pyx_pw_9pyprotect_9protected_43never_writeable(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("never_writeable (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_42never_writeable(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_42never_writeable(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("never_writeable", 0);

  /* "python_visible.pxi":397
 *     in object 'o' if iswrapped(o)
 *     '''
 *     return overridden_always             # <<<<<<<<<<<<<<
 * 
 * def never_writeable_private():
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_9pyprotect_9protected_overridden_always);
  __pyx_r = __pyx_v_9pyprotect_9protected_overridden_always;
  goto __pyx_L0;

  /* "python_visible.pxi":392
 * # ------------------------------------------------------------------------
 * 
 * def never_writeable():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable() -> set(str): Attributes that are never writeable
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":399
 *     return overridden_always
 * 
 * def never_writeable_private():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable_private() -> set(str): Attributes that are never
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_45never_writeable_private(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_44never_writeable_private, "\n    never_writeable_private() -> set(str): Attributes that are never\n    writeable in object 'o' if isprivate(o)\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_45never_writeable_private = {"never_writeable_private", (PyCFunction)__pyx_pw_9pyprotect_9protected_45never_writeable_private, METH_NOARGS, __pyx_doc_9pyprotect_9protected_44never_writeable_private};
static PyObject *__pyx_pw_9pyprotect_9protected_45never_writeable_private(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("never_writeable_private (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_44never_writeable_private(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_44never_writeable_private(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("never_writeable_private", 0);

  /* "python_visible.pxi":404
 *     writeable in object 'o' if isprivate(o)
 *     '''
 *     return frozenset(set().union(             # <<<<<<<<<<<<<<
 *         overridden_always,
 *         always_frozen
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_union); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":406
 *     return frozenset(set().union(
 *         overridden_always,
 *         always_frozen             # <<<<<<<<<<<<<<
 *     ))
 * 
 */
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_9pyprotect_9protected_overridden_always, __pyx_v_9pyprotect_9protected_always_frozen};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "python_visible.pxi":404
 *     writeable in object 'o' if isprivate(o)
 *     '''
 *     return frozenset(set().union(             # <<<<<<<<<<<<<<
 *         overridden_always,
 *         always_frozen
 */
  __pyx_t_3 = __Pyx_PyFrozenSet_New(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":399
 *     return overridden_always
 * 
 * def never_writeable_private():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable_private() -> set(str): Attributes that are never
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.never_writeable_private", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":409
 *     ))
 * 
 * def hidden_pickle_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     hidden_pickle_attributes() -> set(str): Attributes that are never
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_47hidden_pickle_attributes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_46hidden_pickle_attributes, "\n    hidden_pickle_attributes() -> set(str): Attributes that are never\n    visible in object 'o' if iswrapped(o) - to disallow pickling\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_47hidden_pickle_attributes = {"hidden_pickle_attributes", (PyCFunction)__pyx_pw_9pyprotect_9protected_47hidden_pickle_attributes, METH_NOARGS, __pyx_doc_9pyprotect_9protected_46hidden_pickle_attributes};
static PyObject *__pyx_pw_9pyprotect_9protected_47hidden_pickle_attributes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hidden_pickle_attributes (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_46hidden_pickle_attributes(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_46hidden_pickle_attributes(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hidden_pickle_attributes", 0);

  /* "python_visible.pxi":414
 *     visible in object 'o' if iswrapped(o) - to disallow pickling
 *     '''
 *     return pickle_attributes             # <<<<<<<<<<<<<<
 * 
 * def always_delegated_attributes():
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_9pyprotect_9protected_pickle_attributes);
  __pyx_r = __pyx_v_9pyprotect_9protected_pickle_attributes;
  goto __pyx_L0;

  /* "python_visible.pxi":409
 *     ))
 * 
 * def hidden_pickle_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     hidden_pickle_attributes() -> set(str): Attributes that are never
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":416
 *     return pickle_attributes
 * 
 * def always_delegated_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     always_delegated_attributes() -> set(str): Attributes that are
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_49always_delegated_attributes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_48always_delegated_attributes, "\n    always_delegated_attributes() -> set(str): Attributes that are\n    always delegated to wrapped object\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_49always_delegated_attributes = {"always_delegated_attributes", (PyCFunction)__pyx_pw_9pyprotect_9protected_49always_delegated_attributes, METH_NOARGS, __pyx_doc_9pyprotect_9protected_48always_delegated_attributes};
static PyObject *__pyx_pw_9pyprotect_9protected_49always_delegated_attributes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("always_delegated_attributes (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_48always_delegated_attributes(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_48always_delegated_attributes(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("always_delegated_attributes", 0);

  /* "python_visible.pxi":421
 *     always delegated to wrapped object
 *     '''
 *     return always_delegated             # <<<<<<<<<<<<<<
 * 
 * def immutable_builtin_attributes():
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_9pyprotect_9protected_always_delegated);
  __pyx_r = __pyx_v_9pyprotect_9protected_always_delegated;
  goto __pyx_L0;

  /* "python_visible.pxi":416
 *     return pickle_attributes
 * 
 * def always_delegated_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     always_delegated_attributes() -> set(str): Attributes that are
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":423
 *     return always_delegated
 * 
 * def immutable_builtin_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     immutable_builtin_attributes() -> frozenset(str)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_51immutable_builtin_attributes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_50immutable_builtin_attributes, "\n    immutable_builtin_attributes() -> frozenset(str)\n    Returns: attributes in builtins that are immutable\n    ");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_51immutable_builtin_attributes = {"immutable_builtin_attributes", (PyCFunction)__pyx_pw_9pyprotect_9protected_51immutable_builtin_attributes, METH_NOARGS, __pyx_doc_9pyprotect_9protected_50immutable_builtin_attributes};
static PyObject *__pyx_pw_9pyprotect_9protected_51immutable_builtin_attributes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("immutable_builtin_attributes (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_50immutable_builtin_attributes(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_50immutable_builtin_attributes(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("immutable_builtin_attributes", 0);

  /* "python_visible.pxi":428
 *     Returns: attributes in builtins that are immutable
 *     '''
 *     return builtin_module_immutable_attributes             # <<<<<<<<<<<<<<
 * 
 * __all__ = [
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_9pyprotect_9protected_builtin_module_immutable_attributes);
  __pyx_r = __pyx_v_9pyprotect_9protected_builtin_module_immutable_attributes;
  goto __pyx_L0;

  /* "python_visible.pxi":423
 *     return always_delegated
 * 
 * def immutable_builtin_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     immutable_builtin_attributes() -> frozenset(str)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "python_visible.pxi":444
 * 
 * 
 * def __dir__():             # <<<<<<<<<<<<<<
 *     return __all__
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_53__dir__(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_53__dir__ = {"__dir__", (PyCFunction)__pyx_pw_9pyprotect_9protected_53__dir__, METH_NOARGS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_53__dir__(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_52__dir__(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_52__dir__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "python_visible.pxi":445
 * 
 * def __dir__():
 *     return __all__             # <<<<<<<<<<<<<<
 * 
 * class ProtectionError(Exception):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "python_visible.pxi":444
 * 
 * 
 * def __dir__():             # <<<<<<<<<<<<<<
 *     return __all__
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "global_c_functions.pxi":2
 * 
 * cdef get_protected_attr_name():             # <<<<<<<<<<<<<<
 *     '''Returns: str'''
 *     # The ONLY place where the name of the special attribute is set
 */

static PyObject *__pyx_f_9pyprotect_9protected_get_protected_attr_name(void) {
  PyObject *__pyx_v_LOCAL_PROT_ATTR_NAME = 0;
  PyObject *__pyx_v_ENV_VAR = 0;
  PyObject *__pyx_v_PROT_ATTR_SHORTEST_SUFFIX = 0;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_protected_attr_name", 0);

  /* "global_c_functions.pxi":5
 *     '''Returns: str'''
 *     # The ONLY place where the name of the special attribute is set
 *     cdef str LOCAL_PROT_ATTR_NAME = '_Protected_____'             # <<<<<<<<<<<<<<
 *     cdef str ENV_VAR = LOCAL_PROT_ATTR_NAME
 *     cdef str PROT_ATTR_SHORTEST_SUFFIX = '_____'
 */
  __Pyx_INCREF(__pyx_n_s_Protected);
  __pyx_v_LOCAL_PROT_ATTR_NAME = __pyx_n_s_Protected;

  /* "global_c_functions.pxi":6
 *     # The ONLY place where the name of the special attribute is set
 *     cdef str LOCAL_PROT_ATTR_NAME = '_Protected_____'
 *     cdef str ENV_VAR = LOCAL_PROT_ATTR_NAME             # <<<<<<<<<<<<<<
 *     cdef str PROT_ATTR_SHORTEST_SUFFIX = '_____'
 *     # Can override by setting env var '_Protected_____'
 */
  __Pyx_INCREF(__pyx_v_LOCAL_PROT_ATTR_NAME);
  __pyx_v_ENV_VAR = __pyx_v_LOCAL_PROT_ATTR_NAME;

  /* "global_c_functions.pxi":7
 *     cdef str LOCAL_PROT_ATTR_NAME = '_Protected_____'
 *     cdef str ENV_VAR = LOCAL_PROT_ATTR_NAME
 *     cdef str PROT_ATTR_SHORTEST_SUFFIX = '_____'             # <<<<<<<<<<<<<<
 *     # Can override by setting env var '_Protected_____'
 *     # Value of env var '_Protected_____' will be fixed to have EXACTLY
 */
  __Pyx_INCREF(__pyx_n_s_);
  __pyx_v_PROT_ATTR_SHORTEST_SUFFIX = __pyx_n_s_;

  /* "global_c_functions.pxi":11
 *     # Value of env var '_Protected_____' will be fixed to have EXACTLY
 *     # ONE leading underscore and AT LEAST 5 trailing underscores
 *     x = os.environ.get(ENV_VAR, None)             # <<<<<<<<<<<<<<
 *     if x is not None:
 *         if x.startswith('_'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_environ); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_ENV_VAR, Py_None};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_x = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "global_c_functions.pxi":12
 *     # ONE leading underscore and AT LEAST 5 trailing underscores
 *     x = os.environ.get(ENV_VAR, None)
 *     if x is not None:             # <<<<<<<<<<<<<<
 *         if x.startswith('_'):
 *             x = x.lstrip('_') + '_'
 */
  __pyx_t_5 = (__pyx_v_x != Py_None);
  if (__pyx_t_5) {

    /* "global_c_functions.pxi":13
 *     x = os.environ.get(ENV_VAR, None)
 *     if x is not None:
 *         if x.startswith('_'):             # <<<<<<<<<<<<<<
 *             x = x.lstrip('_') + '_'
 *         if not x.endswith(PROT_ATTR_SHORTEST_SUFFIX):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_startswith); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_s__2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 13, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "global_c_functions.pxi":14
 *     if x is not None:
 *         if x.startswith('_'):
 *             x = x.lstrip('_') + '_'             # <<<<<<<<<<<<<<
 *         if not x.endswith(PROT_ATTR_SHORTEST_SUFFIX):
 *             x = x.rstrip('_') + PROT_ATTR_SHORTEST_SUFFIX
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_s__2};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 14, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_n_s__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_x, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "global_c_functions.pxi":13
 *     x = os.environ.get(ENV_VAR, None)
 *     if x is not None:
 *         if x.startswith('_'):             # <<<<<<<<<<<<<<
 *             x = x.lstrip('_') + '_'
 *         if not x.endswith(PROT_ATTR_SHORTEST_SUFFIX):
 */
    }

    /* "global_c_functions.pxi":15
 *         if x.startswith('_'):
 *             x = x.lstrip('_') + '_'
 *         if not x.endswith(PROT_ATTR_SHORTEST_SUFFIX):             # <<<<<<<<<<<<<<
 *             x = x.rstrip('_') + PROT_ATTR_SHORTEST_SUFFIX
 *         return x
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_endswith); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_PROT_ATTR_SHORTEST_SUFFIX};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 15, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_5);
    if (__pyx_t_6) {

      /* "global_c_functions.pxi":16
 *             x = x.lstrip('_') + '_'
 *         if not x.endswith(PROT_ATTR_SHORTEST_SUFFIX):
 *             x = x.rstrip('_') + PROT_ATTR_SHORTEST_SUFFIX             # <<<<<<<<<<<<<<
 *         return x
 *     return LOCAL_PROT_ATTR_NAME
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_rstrip); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 16, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_4 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_s__2};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 16, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_PROT_ATTR_SHORTEST_SUFFIX); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 16, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_x, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "global_c_functions.pxi":15
 *         if x.startswith('_'):
 *             x = x.lstrip('_') + '_'
 *         if not x.endswith(PROT_ATTR_SHORTEST_SUFFIX):             # <<<<<<<<<<<<<<
 *             x = x.rstrip('_') + PROT_ATTR_SHORTEST_SUFFIX
 *         return x
 */
    }

    /* "global_c_functions.pxi":17
 *         if not x.endswith(PROT_ATTR_SHORTEST_SUFFIX):
 *             x = x.rstrip('_') + PROT_ATTR_SHORTEST_SUFFIX
 *         return x             # <<<<<<<<<<<<<<
 *     return LOCAL_PROT_ATTR_NAME
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_x);
    __pyx_r = __pyx_v_x;
    goto __pyx_L0;

    /* "global_c_functions.pxi":12
 *     # ONE leading underscore and AT LEAST 5 trailing underscores
 *     x = os.environ.get(ENV_VAR, None)
 *     if x is not None:             # <<<<<<<<<<<<<<
 *         if x.startswith('_'):
 *             x = x.lstrip('_') + '_'
 */
  }

  /* "global_c_functions.pxi":18
 *             x = x.rstrip('_') + PROT_ATTR_SHORTEST_SUFFIX
 *         return x
 *     return LOCAL_PROT_ATTR_NAME             # <<<<<<<<<<<<<<
 * 
 * cdef get_builtin_obj(s):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_LOCAL_PROT_ATTR_NAME);
  __pyx_r = __pyx_v_LOCAL_PROT_ATTR_NAME;
  goto __pyx_L0;

  /* "global_c_functions.pxi":2
 * 
 * cdef get_protected_attr_name():             # <<<<<<<<<<<<<<
 *     '''Returns: str'''
 *     # The ONLY place where the name of the special attribute is set
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.get_protected_attr_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_LOCAL_PROT_ATTR_NAME);
  __Pyx_XDECREF(__pyx_v_ENV_VAR);
  __Pyx_XDECREF(__pyx_v_PROT_ATTR_SHORTEST_SUFFIX);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "global_c_functions.pxi":20
 *     return LOCAL_PROT_ATTR_NAME
 * 
 * cdef get_builtin_obj(s):             # <<<<<<<<<<<<<<
 *     '''
 *     s-->str: attribute name in builtin_module
 */

static PyObject *__pyx_f_9pyprotect_9protected_get_builtin_obj(PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_builtin_obj", 0);

  /* "global_c_functions.pxi":25
 *     Returns-->object
 *     '''
 *     return getattr(builtin_module, s)             # <<<<<<<<<<<<<<
 * 
 * cdef get_immutables():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_9pyprotect_9protected_builtin_module;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "global_c_functions.pxi":20
 *     return LOCAL_PROT_ATTR_NAME
 * 
 * cdef get_builtin_obj(s):             # <<<<<<<<<<<<<<
 *     '''
 *     s-->str: attribute name in builtin_module
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.get_builtin_obj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "global_c_functions.pxi":27
 *     return getattr(builtin_module, s)
 * 
 * cdef get_immutables():             # <<<<<<<<<<<<<<
 *     '''
 *     Returns: tuple of frozenset:
 */

static PyObject *__pyx_f_9pyprotect_9protected_get_immutables(void) {
  PyObject *__pyx_v_builtin_names = 0;
  PyObject *__pyx_v_basic_data_names = 0;
  PyObject *__pyx_v_basic_mutable_data_names = 0;
  PyObject *__pyx_v_basic_immutable_data_names = 0;
  CYTHON_UNUSED PyObject *__pyx_v_basic_data_types = 0;
  CYTHON_UNUSED PyObject *__pyx_v_basic_mutable_data_types = 0;
  PyObject *__pyx_v_basic_immutable_data_types = 0;
  PyObject *__pyx_v_mapping_types = 0;
  PyObject *__pyx_v_list_types = 0;
  PyObject *__pyx_v_tuple_types = 0;
  PyObject *__pyx_v_set_types = 0;
  PyObject *__pyx_v_sequence_types = 0;
  PyObject *__pyx_v_mutable_mapping_types = 0;
  PyObject *__pyx_v_mutable_sequence_types = 0;
  PyObject *__pyx_v_mutable_set_types = 0;
  PyObject *__pyx_v_immutable_mapping_types = 0;
  PyObject *__pyx_v_immutable_sequence_ytpes = 0;
  PyObject *__pyx_v_immutable_set_types = 0;
  PyObject *__pyx_v_immutable_types = 0;
  PyObject *__pyx_v_ret_immutable_types_set = 0;
  PyObject *__pyx_v_ret_builtins_ids = 0;
  PyObject *__pyx_v_ret_builtin_module_immutable_attributes = 0;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_test_attr_name = NULL;
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_7genexpr__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr7__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr8__pyx_v_x = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_a = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_immutables", 0);

  /* "global_c_functions.pxi":59
 * 
 * 
 *     builtin_names = dir(builtin_module)             # <<<<<<<<<<<<<<
 * 
 *     basic_data_names = [
 */
  __pyx_t_1 = __pyx_v_9pyprotect_9protected_builtin_module;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Dir(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(5, 59, __pyx_L1_error)
  __pyx_v_builtin_names = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":61
 *     builtin_names = dir(builtin_module)
 * 
 *     basic_data_names = [             # <<<<<<<<<<<<<<
 *         'bool', 'bytearray', 'bytes', 'complex',
 *         'dict', 'float', 'frozenset', 'int', 'list', 'object', 'set', 'str',
 */
  __pyx_t_2 = PyList_New(17); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_bool);
  __Pyx_GIVEREF(__pyx_n_s_bool);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_bool);
  __Pyx_INCREF(__pyx_n_s_bytearray);
  __Pyx_GIVEREF(__pyx_n_s_bytearray);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_bytearray);
  __Pyx_INCREF(__pyx_n_s_bytes);
  __Pyx_GIVEREF(__pyx_n_s_bytes);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_bytes);
  __Pyx_INCREF(__pyx_n_s_complex);
  __Pyx_GIVEREF(__pyx_n_s_complex);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_complex);
  __Pyx_INCREF(__pyx_n_s_dict);
  __Pyx_GIVEREF(__pyx_n_s_dict);
  PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_dict);
  __Pyx_INCREF(__pyx_n_s_float);
  __Pyx_GIVEREF(__pyx_n_s_float);
  PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_float);
  __Pyx_INCREF(__pyx_n_s_frozenset);
  __Pyx_GIVEREF(__pyx_n_s_frozenset);
  PyList_SET_ITEM(__pyx_t_2, 6, __pyx_n_s_frozenset);
  __Pyx_INCREF(__pyx_n_s_int);
  __Pyx_GIVEREF(__pyx_n_s_int);
  PyList_SET_ITEM(__pyx_t_2, 7, __pyx_n_s_int);
  __Pyx_INCREF(__pyx_n_s_list);
  __Pyx_GIVEREF(__pyx_n_s_list);
  PyList_SET_ITEM(__pyx_t_2, 8, __pyx_n_s_list);
  __Pyx_INCREF(__pyx_n_s_object);
  __Pyx_GIVEREF(__pyx_n_s_object);
  PyList_SET_ITEM(__pyx_t_2, 9, __pyx_n_s_object);
  __Pyx_INCREF(__pyx_n_s_set);
  __Pyx_GIVEREF(__pyx_n_s_set);
  PyList_SET_ITEM(__pyx_t_2, 10, __pyx_n_s_set);
  __Pyx_INCREF(__pyx_n_s_str);
  __Pyx_GIVEREF(__pyx_n_s_str);
  PyList_SET_ITEM(__pyx_t_2, 11, __pyx_n_s_str);
  __Pyx_INCREF(__pyx_n_s_tuple);
  __Pyx_GIVEREF(__pyx_n_s_tuple);
  PyList_SET_ITEM(__pyx_t_2, 12, __pyx_n_s_tuple);
  __Pyx_INCREF(__pyx_n_s_basestring);
  __Pyx_GIVEREF(__pyx_n_s_basestring);
  PyList_SET_ITEM(__pyx_t_2, 13, __pyx_n_s_basestring);
  __Pyx_INCREF(__pyx_n_s_unichr);
  __Pyx_GIVEREF(__pyx_n_s_unichr);
  PyList_SET_ITEM(__pyx_t_2, 14, __pyx_n_s_unichr);
  __Pyx_INCREF(__pyx_n_s_unicode);
  __Pyx_GIVEREF(__pyx_n_s_unicode);
  PyList_SET_ITEM(__pyx_t_2, 15, __pyx_n_s_unicode);
  __Pyx_INCREF(__pyx_n_s_long);
  __Pyx_GIVEREF(__pyx_n_s_long);
  PyList_SET_ITEM(__pyx_t_2, 16, __pyx_n_s_long);
  __pyx_v_basic_data_names = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":66
 *         'tuple', 'basestring', 'unichr', 'unicode', 'long'
 *     ]
 *     basic_mutable_data_names = ['bytearray', 'dict', 'list']             # <<<<<<<<<<<<<<
 *     basic_immutable_data_names = [
 *         'bool', 'bytes', 'complex', 'float',
 */
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_bytearray);
  __Pyx_GIVEREF(__pyx_n_s_bytearray);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_bytearray);
  __Pyx_INCREF(__pyx_n_s_dict);
  __Pyx_GIVEREF(__pyx_n_s_dict);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_dict);
  __Pyx_INCREF(__pyx_n_s_list);
  __Pyx_GIVEREF(__pyx_n_s_list);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_list);
  __pyx_v_basic_mutable_data_names = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":67
 *     ]
 *     basic_mutable_data_names = ['bytearray', 'dict', 'list']
 *     basic_immutable_data_names = [             # <<<<<<<<<<<<<<
 *         'bool', 'bytes', 'complex', 'float',
 *         'frozenset', 'int', 'str', 'tuple', 'basestring', 'unichr',
 */
  __pyx_t_2 = PyList_New(12); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_bool);
  __Pyx_GIVEREF(__pyx_n_s_bool);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_bool);
  __Pyx_INCREF(__pyx_n_s_bytes);
  __Pyx_GIVEREF(__pyx_n_s_bytes);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_bytes);
  __Pyx_INCREF(__pyx_n_s_complex);
  __Pyx_GIVEREF(__pyx_n_s_complex);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_complex);
  __Pyx_INCREF(__pyx_n_s_float);
  __Pyx_GIVEREF(__pyx_n_s_float);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_float);
  __Pyx_INCREF(__pyx_n_s_frozenset);
  __Pyx_GIVEREF(__pyx_n_s_frozenset);
  PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_frozenset);
  __Pyx_INCREF(__pyx_n_s_int);
  __Pyx_GIVEREF(__pyx_n_s_int);
  PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_int);
  __Pyx_INCREF(__pyx_n_s_str);
  __Pyx_GIVEREF(__pyx_n_s_str);
  PyList_SET_ITEM(__pyx_t_2, 6, __pyx_n_s_str);
  __Pyx_INCREF(__pyx_n_s_tuple);
  __Pyx_GIVEREF(__pyx_n_s_tuple);
  PyList_SET_ITEM(__pyx_t_2, 7, __pyx_n_s_tuple);
  __Pyx_INCREF(__pyx_n_s_basestring);
  __Pyx_GIVEREF(__pyx_n_s_basestring);
  PyList_SET_ITEM(__pyx_t_2, 8, __pyx_n_s_basestring);
  __Pyx_INCREF(__pyx_n_s_unichr);
  __Pyx_GIVEREF(__pyx_n_s_unichr);
  PyList_SET_ITEM(__pyx_t_2, 9, __pyx_n_s_unichr);
  __Pyx_INCREF(__pyx_n_s_unicode);
  __Pyx_GIVEREF(__pyx_n_s_unicode);
  PyList_SET_ITEM(__pyx_t_2, 10, __pyx_n_s_unicode);
  __Pyx_INCREF(__pyx_n_s_long);
  __Pyx_GIVEREF(__pyx_n_s_long);
  PyList_SET_ITEM(__pyx_t_2, 11, __pyx_n_s_long);
  __pyx_v_basic_immutable_data_names = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":73
 *     ]
 * 
 *     basic_data_names = [x for x in basic_data_names if x in builtin_names]             # <<<<<<<<<<<<<<
 * 
 *     basic_mutable_data_names = [
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 73, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_v_basic_data_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 73, __pyx_L5_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 73, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_7genexpr__pyx_v_x, __pyx_v_builtin_names, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 73, __pyx_L5_error)
      if (__pyx_t_5) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_7genexpr__pyx_v_x))) __PYX_ERR(5, 73, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_x); __pyx_7genexpr__pyx_v_x = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_x); __pyx_7genexpr__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_basic_data_names, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":75
 *     basic_data_names = [x for x in basic_data_names if x in builtin_names]
 * 
 *     basic_mutable_data_names = [             # <<<<<<<<<<<<<<
 *         x for x in basic_mutable_data_names if
 *         x in builtin_names
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 75, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":76
 * 
 *     basic_mutable_data_names = [
 *         x for x in basic_mutable_data_names if             # <<<<<<<<<<<<<<
 *         x in builtin_names
 *     ]
 */
    __pyx_t_1 = __pyx_v_basic_mutable_data_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 76, __pyx_L12_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 76, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":77
 *     basic_mutable_data_names = [
 *         x for x in basic_mutable_data_names if
 *         x in builtin_names             # <<<<<<<<<<<<<<
 *     ]
 *     basic_immutable_data_names = [
 */
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr1__pyx_v_x, __pyx_v_builtin_names, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 77, __pyx_L12_error)

      /* "global_c_functions.pxi":76
 * 
 *     basic_mutable_data_names = [
 *         x for x in basic_mutable_data_names if             # <<<<<<<<<<<<<<
 *         x in builtin_names
 *     ]
 */
      if (__pyx_t_5) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_8genexpr1__pyx_v_x))) __PYX_ERR(5, 75, __pyx_L12_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x); __pyx_8genexpr1__pyx_v_x = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x); __pyx_8genexpr1__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_basic_mutable_data_names, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":79
 *         x in builtin_names
 *     ]
 *     basic_immutable_data_names = [             # <<<<<<<<<<<<<<
 *         x for x in basic_immutable_data_names if
 *         x in builtin_names
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 79, __pyx_L19_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":80
 *     ]
 *     basic_immutable_data_names = [
 *         x for x in basic_immutable_data_names if             # <<<<<<<<<<<<<<
 *         x in builtin_names
 *     ]
 */
    __pyx_t_1 = __pyx_v_basic_immutable_data_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 80, __pyx_L19_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 80, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":81
 *     basic_immutable_data_names = [
 *         x for x in basic_immutable_data_names if
 *         x in builtin_names             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr2__pyx_v_x, __pyx_v_builtin_names, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 81, __pyx_L19_error)

      /* "global_c_functions.pxi":80
 *     ]
 *     basic_immutable_data_names = [
 *         x for x in basic_immutable_data_names if             # <<<<<<<<<<<<<<
 *         x in builtin_names
 *     ]
 */
      if (__pyx_t_5) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_8genexpr2__pyx_v_x))) __PYX_ERR(5, 79, __pyx_L19_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_x); __pyx_8genexpr2__pyx_v_x = 0;
    goto __pyx_L23_exit_scope;
    __pyx_L19_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_x); __pyx_8genexpr2__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L23_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_basic_immutable_data_names, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":84
 *     ]
 * 
 *     basic_data_types = [get_builtin_obj(x) for x in basic_data_names]             # <<<<<<<<<<<<<<
 *     basic_mutable_data_types = [
 *         get_builtin_obj(x) for x in basic_mutable_data_names
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 84, __pyx_L26_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_v_basic_data_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 84, __pyx_L26_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 84, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __pyx_f_9pyprotect_9protected_get_builtin_obj(__pyx_8genexpr3__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 84, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(5, 84, __pyx_L26_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_x); __pyx_8genexpr3__pyx_v_x = 0;
    goto __pyx_L29_exit_scope;
    __pyx_L26_error:;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_x); __pyx_8genexpr3__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L29_exit_scope:;
  } /* exit inner scope */
  __pyx_v_basic_data_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":85
 * 
 *     basic_data_types = [get_builtin_obj(x) for x in basic_data_names]
 *     basic_mutable_data_types = [             # <<<<<<<<<<<<<<
 *         get_builtin_obj(x) for x in basic_mutable_data_names
 *     ]
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 85, __pyx_L32_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":86
 *     basic_data_types = [get_builtin_obj(x) for x in basic_data_names]
 *     basic_mutable_data_types = [
 *         get_builtin_obj(x) for x in basic_mutable_data_names             # <<<<<<<<<<<<<<
 *     ]
 *     basic_immutable_data_types = [
 */
    __pyx_t_1 = __pyx_v_basic_mutable_data_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 86, __pyx_L32_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 86, __pyx_L32_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __pyx_f_9pyprotect_9protected_get_builtin_obj(__pyx_8genexpr4__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 86, __pyx_L32_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(5, 85, __pyx_L32_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_x); __pyx_8genexpr4__pyx_v_x = 0;
    goto __pyx_L35_exit_scope;
    __pyx_L32_error:;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_x); __pyx_8genexpr4__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L35_exit_scope:;
  } /* exit inner scope */
  __pyx_v_basic_mutable_data_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":88
 *         get_builtin_obj(x) for x in basic_mutable_data_names
 *     ]
 *     basic_immutable_data_types = [             # <<<<<<<<<<<<<<
 *         get_builtin_obj(x) for x in basic_immutable_data_names
 *     ]
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 88, __pyx_L38_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":89
 *     ]
 *     basic_immutable_data_types = [
 *         get_builtin_obj(x) for x in basic_immutable_data_names             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
    __pyx_t_1 = __pyx_v_basic_immutable_data_names; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 89, __pyx_L38_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 89, __pyx_L38_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __pyx_f_9pyprotect_9protected_get_builtin_obj(__pyx_8genexpr5__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 89, __pyx_L38_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(5, 88, __pyx_L38_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_x); __pyx_8genexpr5__pyx_v_x = 0;
    goto __pyx_L41_exit_scope;
    __pyx_L38_error:;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_x); __pyx_8genexpr5__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L41_exit_scope:;
  } /* exit inner scope */
  __pyx_v_basic_immutable_data_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":92
 *     ]
 * 
 *     mapping_types = [dict, CollectionsABC.MutableMapping, CollectionsABC.Mapping]             # <<<<<<<<<<<<<<
 *     list_types = [list, CollectionsABC.MutableSequence]
 *     tuple_types = [tuple, CollectionsABC.Sequence]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&PyDict_Type));
  __Pyx_GIVEREF((PyObject *)(&PyDict_Type));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PyDict_Type)));
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_v_mapping_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":93
 * 
 *     mapping_types = [dict, CollectionsABC.MutableMapping, CollectionsABC.Mapping]
 *     list_types = [list, CollectionsABC.MutableSequence]             # <<<<<<<<<<<<<<
 *     tuple_types = [tuple, CollectionsABC.Sequence]
 *     set_types = [set, CollectionsABC.Set, CollectionsABC.MutableSet]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&PyList_Type));
  __Pyx_GIVEREF((PyObject *)(&PyList_Type));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PyList_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_v_list_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":94
 *     mapping_types = [dict, CollectionsABC.MutableMapping, CollectionsABC.Mapping]
 *     list_types = [list, CollectionsABC.MutableSequence]
 *     tuple_types = [tuple, CollectionsABC.Sequence]             # <<<<<<<<<<<<<<
 *     set_types = [set, CollectionsABC.Set, CollectionsABC.MutableSet]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Sequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&PyTuple_Type));
  __Pyx_GIVEREF((PyObject *)(&PyTuple_Type));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PyTuple_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_v_tuple_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":95
 *     list_types = [list, CollectionsABC.MutableSequence]
 *     tuple_types = [tuple, CollectionsABC.Sequence]
 *     set_types = [set, CollectionsABC.Set, CollectionsABC.MutableSet]             # <<<<<<<<<<<<<<
 * 
 *     sequence_types = tuple_types + list_types
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Set); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&PySet_Type));
  __Pyx_GIVEREF((PyObject *)(&PySet_Type));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PySet_Type)));
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_v_set_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":97
 *     set_types = [set, CollectionsABC.Set, CollectionsABC.MutableSet]
 * 
 *     sequence_types = tuple_types + list_types             # <<<<<<<<<<<<<<
 * 
 *     mutable_sequence_types = list_types
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_tuple_types, __pyx_v_list_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sequence_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":99
 *     sequence_types = tuple_types + list_types
 * 
 *     mutable_sequence_types = list_types             # <<<<<<<<<<<<<<
 *     mutable_mapping_types = [dict, CollectionsABC.MutableMapping]
 *     mutable_set_types = [set, CollectionsABC.MutableSet]
 */
  __Pyx_INCREF(__pyx_v_list_types);
  __pyx_v_mutable_sequence_types = __pyx_v_list_types;

  /* "global_c_functions.pxi":100
 * 
 *     mutable_sequence_types = list_types
 *     mutable_mapping_types = [dict, CollectionsABC.MutableMapping]             # <<<<<<<<<<<<<<
 *     mutable_set_types = [set, CollectionsABC.MutableSet]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&PyDict_Type));
  __Pyx_GIVEREF((PyObject *)(&PyDict_Type));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PyDict_Type)));
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_mutable_mapping_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":101
 *     mutable_sequence_types = list_types
 *     mutable_mapping_types = [dict, CollectionsABC.MutableMapping]
 *     mutable_set_types = [set, CollectionsABC.MutableSet]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&PySet_Type));
  __Pyx_GIVEREF((PyObject *)(&PySet_Type));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PySet_Type)));
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_mutable_set_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":104
 * 
 * 
 *     immutable_mapping_types = [             # <<<<<<<<<<<<<<
 *         x for x in mapping_types if
 *         x not in mutable_mapping_types
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 104, __pyx_L44_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":105
 * 
 *     immutable_mapping_types = [
 *         x for x in mapping_types if             # <<<<<<<<<<<<<<
 *         x not in mutable_mapping_types
 *     ]
 */
    __pyx_t_1 = __pyx_v_mapping_types; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 105, __pyx_L44_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 105, __pyx_L44_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":106
 *     immutable_mapping_types = [
 *         x for x in mapping_types if
 *         x not in mutable_mapping_types             # <<<<<<<<<<<<<<
 *     ]
 *     immutable_sequence_ytpes = [
 */
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr6__pyx_v_x, __pyx_v_mutable_mapping_types, Py_NE)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 106, __pyx_L44_error)

      /* "global_c_functions.pxi":105
 * 
 *     immutable_mapping_types = [
 *         x for x in mapping_types if             # <<<<<<<<<<<<<<
 *         x not in mutable_mapping_types
 *     ]
 */
      if (__pyx_t_5) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_8genexpr6__pyx_v_x))) __PYX_ERR(5, 104, __pyx_L44_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_x); __pyx_8genexpr6__pyx_v_x = 0;
    goto __pyx_L48_exit_scope;
    __pyx_L44_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_x); __pyx_8genexpr6__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L48_exit_scope:;
  } /* exit inner scope */
  __pyx_v_immutable_mapping_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":108
 *         x not in mutable_mapping_types
 *     ]
 *     immutable_sequence_ytpes = [             # <<<<<<<<<<<<<<
 *         x for x in sequence_types if
 *         x not in mutable_sequence_types
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 108, __pyx_L51_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":109
 *     ]
 *     immutable_sequence_ytpes = [
 *         x for x in sequence_types if             # <<<<<<<<<<<<<<
 *         x not in mutable_sequence_types
 *     ]
 */
    __pyx_t_1 = __pyx_v_sequence_types; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 109, __pyx_L51_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 109, __pyx_L51_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":110
 *     immutable_sequence_ytpes = [
 *         x for x in sequence_types if
 *         x not in mutable_sequence_types             # <<<<<<<<<<<<<<
 *     ]
 *     immutable_set_types = [x for x in set_types if x not in mutable_set_types]
 */
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr7__pyx_v_x, __pyx_v_mutable_sequence_types, Py_NE)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 110, __pyx_L51_error)

      /* "global_c_functions.pxi":109
 *     ]
 *     immutable_sequence_ytpes = [
 *         x for x in sequence_types if             # <<<<<<<<<<<<<<
 *         x not in mutable_sequence_types
 *     ]
 */
      if (__pyx_t_5) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_8genexpr7__pyx_v_x))) __PYX_ERR(5, 108, __pyx_L51_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_x); __pyx_8genexpr7__pyx_v_x = 0;
    goto __pyx_L55_exit_scope;
    __pyx_L51_error:;
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_x); __pyx_8genexpr7__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L55_exit_scope:;
  } /* exit inner scope */
  __pyx_v_immutable_sequence_ytpes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":112
 *         x not in mutable_sequence_types
 *     ]
 *     immutable_set_types = [x for x in set_types if x not in mutable_set_types]             # <<<<<<<<<<<<<<
 * 
 *     immutable_types = list(set(
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 112, __pyx_L58_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_v_set_types; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 112, __pyx_L58_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 112, __pyx_L58_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr8__pyx_v_x, __pyx_v_mutable_set_types, Py_NE)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 112, __pyx_L58_error)
      if (__pyx_t_5) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_8genexpr8__pyx_v_x))) __PYX_ERR(5, 112, __pyx_L58_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_x); __pyx_8genexpr8__pyx_v_x = 0;
    goto __pyx_L62_exit_scope;
    __pyx_L58_error:;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_x); __pyx_8genexpr8__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L62_exit_scope:;
  } /* exit inner scope */
  __pyx_v_immutable_set_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":115
 * 
 *     immutable_types = list(set(
 *         basic_immutable_data_types +             # <<<<<<<<<<<<<<
 *         immutable_mapping_types +
 *         immutable_sequence_ytpes +
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_basic_immutable_data_types, __pyx_v_immutable_mapping_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "global_c_functions.pxi":116
 *     immutable_types = list(set(
 *         basic_immutable_data_types +
 *         immutable_mapping_types +             # <<<<<<<<<<<<<<
 *         immutable_sequence_ytpes +
 *         immutable_set_types
 */
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_immutable_sequence_ytpes); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":117
 *         basic_immutable_data_types +
 *         immutable_mapping_types +
 *         immutable_sequence_ytpes +             # <<<<<<<<<<<<<<
 *         immutable_set_types
 *     ))
 */
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_v_immutable_set_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "global_c_functions.pxi":114
 *     immutable_set_types = [x for x in set_types if x not in mutable_set_types]
 * 
 *     immutable_types = list(set(             # <<<<<<<<<<<<<<
 *         basic_immutable_data_types +
 *         immutable_mapping_types +
 */
  __pyx_t_1 = PySet_New(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_immutable_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":120
 *         immutable_set_types
 *     ))
 *     immutable_types = [x for x in immutable_types if isinstance(x, type)]             # <<<<<<<<<<<<<<
 * 
 *     # Since builtin_module is by default writeable in Python and attributes
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 120, __pyx_L65_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_v_immutable_types; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 120, __pyx_L65_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 120, __pyx_L65_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_5 = PyType_Check(__pyx_8genexpr9__pyx_v_x); 
      if (__pyx_t_5) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_8genexpr9__pyx_v_x))) __PYX_ERR(5, 120, __pyx_L65_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_x); __pyx_8genexpr9__pyx_v_x = 0;
    goto __pyx_L69_exit_scope;
    __pyx_L65_error:;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_x); __pyx_8genexpr9__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L69_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_immutable_types, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":125
 *     # in builtin_module can be overwritten, we only track attributes
 *     # that do not allow __class__ attribute to be overwritten (crude test)
 *     s = set()             # <<<<<<<<<<<<<<
 *     test_attr_name = '__class__'
 *     for a in builtin_names:
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_s = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":126
 *     # that do not allow __class__ attribute to be overwritten (crude test)
 *     s = set()
 *     test_attr_name = '__class__'             # <<<<<<<<<<<<<<
 *     for a in builtin_names:
 *         try:
 */
  __Pyx_INCREF(__pyx_n_s_class);
  __pyx_v_test_attr_name = __pyx_n_s_class;

  /* "global_c_functions.pxi":127
 *     s = set()
 *     test_attr_name = '__class__'
 *     for a in builtin_names:             # <<<<<<<<<<<<<<
 *         try:
 *             x = getattr(builtin_module, a)
 */
  if (unlikely(__pyx_v_builtin_names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(5, 127, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_builtin_names; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
  for (;;) {
    if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 127, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_a, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "global_c_functions.pxi":128
 *     test_attr_name = '__class__'
 *     for a in builtin_names:
 *         try:             # <<<<<<<<<<<<<<
 *             x = getattr(builtin_module, a)
 *             try:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "global_c_functions.pxi":129
 *     for a in builtin_names:
 *         try:
 *             x = getattr(builtin_module, a)             # <<<<<<<<<<<<<<
 *             try:
 *                 setattr(x, test_attr_name, getattr(x, test_attr_name))
 */
        __pyx_t_1 = __pyx_v_9pyprotect_9protected_builtin_module;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_a); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 129, __pyx_L72_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "global_c_functions.pxi":130
 *         try:
 *             x = getattr(builtin_module, a)
 *             try:             # <<<<<<<<<<<<<<
 *                 setattr(x, test_attr_name, getattr(x, test_attr_name))
 *                 continue
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_11);
          /*try:*/ {

            /* "global_c_functions.pxi":131
 *             x = getattr(builtin_module, a)
 *             try:
 *                 setattr(x, test_attr_name, getattr(x, test_attr_name))             # <<<<<<<<<<<<<<
 *                 continue
 *             except:
 */
            __pyx_t_4 = __Pyx_GetAttr(__pyx_v_x, __pyx_v_test_attr_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 131, __pyx_L80_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_12 = PyObject_SetAttr(__pyx_v_x, __pyx_v_test_attr_name, __pyx_t_4); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(5, 131, __pyx_L80_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "global_c_functions.pxi":132
 *             try:
 *                 setattr(x, test_attr_name, getattr(x, test_attr_name))
 *                 continue             # <<<<<<<<<<<<<<
 *             except:
 *                 s.add(a)
 */
            goto __pyx_L86_try_continue;

            /* "global_c_functions.pxi":130
 *         try:
 *             x = getattr(builtin_module, a)
 *             try:             # <<<<<<<<<<<<<<
 *                 setattr(x, test_attr_name, getattr(x, test_attr_name))
 *                 continue
 */
          }
          __pyx_L80_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "global_c_functions.pxi":133
 *                 setattr(x, test_attr_name, getattr(x, test_attr_name))
 *                 continue
 *             except:             # <<<<<<<<<<<<<<
 *                 s.add(a)
 *         except:
 */
          /*except:*/ {
            __Pyx_AddTraceback("pyprotect.protected.get_immutables", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_13) < 0) __PYX_ERR(5, 133, __pyx_L82_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_13);

            /* "global_c_functions.pxi":134
 *                 continue
 *             except:
 *                 s.add(a)             # <<<<<<<<<<<<<<
 *         except:
 *             continue
 */
            __pyx_t_12 = PySet_Add(__pyx_v_s, __pyx_v_a); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(5, 134, __pyx_L82_except_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            goto __pyx_L81_exception_handled;
          }
          __pyx_L82_except_error:;

          /* "global_c_functions.pxi":130
 *         try:
 *             x = getattr(builtin_module, a)
 *             try:             # <<<<<<<<<<<<<<
 *                 setattr(x, test_attr_name, getattr(x, test_attr_name))
 *                 continue
 */
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
          goto __pyx_L72_error;
          __pyx_L86_try_continue:;
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
          goto __pyx_L78_try_continue;
          __pyx_L81_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        }

        /* "global_c_functions.pxi":128
 *     test_attr_name = '__class__'
 *     for a in builtin_names:
 *         try:             # <<<<<<<<<<<<<<
 *             x = getattr(builtin_module, a)
 *             try:
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L79_try_end;
      __pyx_L72_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "global_c_functions.pxi":135
 *             except:
 *                 s.add(a)
 *         except:             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
      /*except:*/ {
        __Pyx_AddTraceback("pyprotect.protected.get_immutables", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_13, &__pyx_t_1, &__pyx_t_4) < 0) __PYX_ERR(5, 135, __pyx_L74_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);

        /* "global_c_functions.pxi":136
 *                 s.add(a)
 *         except:
 *             continue             # <<<<<<<<<<<<<<
 * 
 *     ret_immutable_types_set = frozenset(immutable_types)
 */
        goto __pyx_L91_except_continue;
        __pyx_L91_except_continue:;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L78_try_continue;
      }
      __pyx_L74_except_error:;

      /* "global_c_functions.pxi":128
 *     test_attr_name = '__class__'
 *     for a in builtin_names:
 *         try:             # <<<<<<<<<<<<<<
 *             x = getattr(builtin_module, a)
 *             try:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L78_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L70_continue;
      __pyx_L79_try_end:;
    }

    /* "global_c_functions.pxi":127
 *     s = set()
 *     test_attr_name = '__class__'
 *     for a in builtin_names:             # <<<<<<<<<<<<<<
 *         try:
 *             x = getattr(builtin_module, a)
 */
    __pyx_L70_continue:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":138
 *             continue
 * 
 *     ret_immutable_types_set = frozenset(immutable_types)             # <<<<<<<<<<<<<<
 *     ret_builtin_module_immutable_attributes = frozenset(s)
 *     ret_builtins_ids = frozenset([
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_v_immutable_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ret_immutable_types_set = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":139
 * 
 *     ret_immutable_types_set = frozenset(immutable_types)
 *     ret_builtin_module_immutable_attributes = frozenset(s)             # <<<<<<<<<<<<<<
 *     ret_builtins_ids = frozenset([
 *         id(getattr(builtin_module, a)) for a in builtin_names
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ret_builtin_module_immutable_attributes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":140
 *     ret_immutable_types_set = frozenset(immutable_types)
 *     ret_builtin_module_immutable_attributes = frozenset(s)
 *     ret_builtins_ids = frozenset([             # <<<<<<<<<<<<<<
 *         id(getattr(builtin_module, a)) for a in builtin_names
 *         if a in ret_builtin_module_immutable_attributes
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 140, __pyx_L94_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":141
 *     ret_builtin_module_immutable_attributes = frozenset(s)
 *     ret_builtins_ids = frozenset([
 *         id(getattr(builtin_module, a)) for a in builtin_names             # <<<<<<<<<<<<<<
 *         if a in ret_builtin_module_immutable_attributes
 *     ])
 */
    if (unlikely(__pyx_v_builtin_names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(5, 141, __pyx_L94_error)
    }
    __pyx_t_4 = __pyx_v_builtin_names; __Pyx_INCREF(__pyx_t_4); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(5, 141, __pyx_L94_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 141, __pyx_L94_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_a, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "global_c_functions.pxi":142
 *     ret_builtins_ids = frozenset([
 *         id(getattr(builtin_module, a)) for a in builtin_names
 *         if a in ret_builtin_module_immutable_attributes             # <<<<<<<<<<<<<<
 *     ])
 * 
 */
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_9genexpr10__pyx_v_a, __pyx_v_ret_builtin_module_immutable_attributes, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(5, 142, __pyx_L94_error)
      if (__pyx_t_5) {

        /* "global_c_functions.pxi":141
 *     ret_builtin_module_immutable_attributes = frozenset(s)
 *     ret_builtins_ids = frozenset([
 *         id(getattr(builtin_module, a)) for a in builtin_names             # <<<<<<<<<<<<<<
 *         if a in ret_builtin_module_immutable_attributes
 *     ])
 */
        __pyx_t_1 = __pyx_v_9pyprotect_9protected_builtin_module;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_13 = __Pyx_GetAttr(__pyx_t_1, __pyx_9genexpr10__pyx_v_a); if (unlikely(!__pyx_t_13)) __PYX_ERR(5, 141, __pyx_L94_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 141, __pyx_L94_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(5, 140, __pyx_L94_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "global_c_functions.pxi":142
 *     ret_builtins_ids = frozenset([
 *         id(getattr(builtin_module, a)) for a in builtin_names
 *         if a in ret_builtin_module_immutable_attributes             # <<<<<<<<<<<<<<
 *     ])
 * 
 */
      }

      /* "global_c_functions.pxi":141
 *     ret_builtin_module_immutable_attributes = frozenset(s)
 *     ret_builtins_ids = frozenset([
 *         id(getattr(builtin_module, a)) for a in builtin_names             # <<<<<<<<<<<<<<
 *         if a in ret_builtin_module_immutable_attributes
 *     ])
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_a); __pyx_9genexpr10__pyx_v_a = 0;
    goto __pyx_L98_exit_scope;
    __pyx_L94_error:;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_a); __pyx_9genexpr10__pyx_v_a = 0;
    goto __pyx_L1_error;
    __pyx_L98_exit_scope:;
  } /* exit inner scope */

  /* "global_c_functions.pxi":140
 *     ret_immutable_types_set = frozenset(immutable_types)
 *     ret_builtin_module_immutable_attributes = frozenset(s)
 *     ret_builtins_ids = frozenset([             # <<<<<<<<<<<<<<
 *         id(getattr(builtin_module, a)) for a in builtin_names
 *         if a in ret_builtin_module_immutable_attributes
 */
  __pyx_t_4 = __Pyx_PyFrozenSet_New(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ret_builtins_ids = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "global_c_functions.pxi":145
 *     ])
 * 
 *     return (             # <<<<<<<<<<<<<<
 *         ret_immutable_types_set,
 *         ret_builtin_module_immutable_attributes,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "global_c_functions.pxi":146
 * 
 *     return (
 *         ret_immutable_types_set,             # <<<<<<<<<<<<<<
 *         ret_builtin_module_immutable_attributes,
 *         ret_builtins_ids,
 */
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_ret_immutable_types_set);
  __Pyx_GIVEREF(__pyx_v_ret_immutable_types_set);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_ret_immutable_types_set);
  __Pyx_INCREF(__pyx_v_ret_builtin_module_immutable_attributes);
  __Pyx_GIVEREF(__pyx_v_ret_builtin_module_immutable_attributes);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_ret_builtin_module_immutable_attributes);
  __Pyx_INCREF(__pyx_v_ret_builtins_ids);
  __Pyx_GIVEREF(__pyx_v_ret_builtins_ids);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_ret_builtins_ids);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "global_c_functions.pxi":27
 *     return getattr(builtin_module, s)
 * 
 * cdef get_immutables():             # <<<<<<<<<<<<<<
 *     '''
 *     Returns: tuple of frozenset:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pyprotect.protected.get_immutables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_builtin_names);
  __Pyx_XDECREF(__pyx_v_basic_data_names);
  __Pyx_XDECREF(__pyx_v_basic_mutable_data_names);
  __Pyx_XDECREF(__pyx_v_basic_immutable_data_names);
  __Pyx_XDECREF(__pyx_v_basic_data_types);
  __Pyx_XDECREF(__pyx_v_basic_mutable_data_types);
  __Pyx_XDECREF(__pyx_v_basic_immutable_data_types);
  __Pyx_XDECREF(__pyx_v_mapping_types);
  __Pyx_XDECREF(__pyx_v_list_types);
  __Pyx_XDECREF(__pyx_v_tuple_types);
  __Pyx_XDECREF(__pyx_v_set_types);
  __Pyx_XDECREF(__pyx_v_sequence_types);
  __Pyx_XDECREF(__pyx_v_mutable_mapping_types);
  __Pyx_XDECREF(__pyx_v_mutable_sequence_types);
  __Pyx_XDECREF(__pyx_v_mutable_set_types);
  __Pyx_XDECREF(__pyx_v_immutable_mapping_types);
  __Pyx_XDECREF(__pyx_v_immutable_sequence_ytpes);
  __Pyx_XDECREF(__pyx_v_immutable_set_types);
  __Pyx_XDECREF(__pyx_v_immutable_types);
  __Pyx_XDECREF(__pyx_v_ret_immutable_types_set);
  __Pyx_XDECREF(__pyx_v_ret_builtins_ids);
  __Pyx_XDECREF(__pyx_v_ret_builtin_module_immutable_attributes);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_test_attr_name);
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_x);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_a);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "global_c_functions.pxi":159
 *     Called once by protect() before Protected class initialization
 *     '''
 *     def _build_regex(alist):             # <<<<<<<<<<<<<<
 *         _ret = ''
 *         _rl = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_27protected_rules_from_kwargs_1_build_regex(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_27protected_rules_from_kwargs_1_build_regex = {"_build_regex", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_27protected_rules_from_kwargs_1_build_regex, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_27protected_rules_from_kwargs_1_build_regex(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_alist = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_build_regex (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_alist,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_alist,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alist)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 159, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_build_regex") < 0)) __PYX_ERR(5, 159, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_alist = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_build_regex", 1, 1, 1, __pyx_nargs); __PYX_ERR(5, 159, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.protected_rules_from_kwargs._build_regex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_27protected_rules_from_kwargs__build_regex(__pyx_self, __pyx_v_alist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_27protected_rules_from_kwargs__build_regex(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_alist) {
  PyObject *__pyx_v__ret = NULL;
  PyObject *__pyx_v__rl = NULL;
  PyObject *__pyx_v__x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_build_regex", 0);

  /* "global_c_functions.pxi":160
 *     '''
 *     def _build_regex(alist):
 *         _ret = ''             # <<<<<<<<<<<<<<
 *         _rl = []
 *         if not alist:
 */
  __Pyx_INCREF(__pyx_kp_s__3);
  __pyx_v__ret = __pyx_kp_s__3;

  /* "global_c_functions.pxi":161
 *     def _build_regex(alist):
 *         _ret = ''
 *         _rl = []             # <<<<<<<<<<<<<<
 *         if not alist:
 *             return re.compile(_ret)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__rl = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "global_c_functions.pxi":162
 *         _ret = ''
 *         _rl = []
 *         if not alist:             # <<<<<<<<<<<<<<
 *             return re.compile(_ret)
 *         for _x in alist:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_alist); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 162, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "global_c_functions.pxi":163
 *         _rl = []
 *         if not alist:
 *             return re.compile(_ret)             # <<<<<<<<<<<<<<
 *         for _x in alist:
 *             if not isinstance(_x, str):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_re); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_compile); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v__ret};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "global_c_functions.pxi":162
 *         _ret = ''
 *         _rl = []
 *         if not alist:             # <<<<<<<<<<<<<<
 *             return re.compile(_ret)
 *         for _x in alist:
 */
  }

  /* "global_c_functions.pxi":164
 *         if not alist:
 *             return re.compile(_ret)
 *         for _x in alist:             # <<<<<<<<<<<<<<
 *             if not isinstance(_x, str):
 *                 continue
 */
  if (likely(PyList_CheckExact(__pyx_v_alist)) || PyTuple_CheckExact(__pyx_v_alist)) {
    __pyx_t_1 = __pyx_v_alist; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_alist); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 164, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(5, 164, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(5, 164, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_8(__pyx_t_1);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(5, 164, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v__x, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "global_c_functions.pxi":165
 *             return re.compile(_ret)
 *         for _x in alist:
 *             if not isinstance(_x, str):             # <<<<<<<<<<<<<<
 *                 continue
 *             if attr_identifier.match(_x):
 */
    __pyx_t_3 = PyString_Check(__pyx_v__x); 
    __pyx_t_2 = (!__pyx_t_3);
    if (__pyx_t_2) {

      /* "global_c_functions.pxi":166
 *         for _x in alist:
 *             if not isinstance(_x, str):
 *                 continue             # <<<<<<<<<<<<<<
 *             if attr_identifier.match(_x):
 *                 _rl += ['^%s$' % (_x,)]
 */
      goto __pyx_L4_continue;

      /* "global_c_functions.pxi":165
 *             return re.compile(_ret)
 *         for _x in alist:
 *             if not isinstance(_x, str):             # <<<<<<<<<<<<<<
 *                 continue
 *             if attr_identifier.match(_x):
 */
    }

    /* "global_c_functions.pxi":167
 *             if not isinstance(_x, str):
 *                 continue
 *             if attr_identifier.match(_x):             # <<<<<<<<<<<<<<
 *                 _rl += ['^%s$' % (_x,)]
 *         for _x in _rl:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_attr_identifier, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v__x};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 167, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "global_c_functions.pxi":168
 *                 continue
 *             if attr_identifier.match(_x):
 *                 _rl += ['^%s$' % (_x,)]             # <<<<<<<<<<<<<<
 *         for _x in _rl:
 *             if _ret:
 */
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v__x);
      __Pyx_GIVEREF(__pyx_v__x);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__x);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_s, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v__rl, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v__rl, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":167
 *             if not isinstance(_x, str):
 *                 continue
 *             if attr_identifier.match(_x):             # <<<<<<<<<<<<<<
 *                 _rl += ['^%s$' % (_x,)]
 *         for _x in _rl:
 */
    }

    /* "global_c_functions.pxi":164
 *         if not alist:
 *             return re.compile(_ret)
 *         for _x in alist:             # <<<<<<<<<<<<<<
 *             if not isinstance(_x, str):
 *                 continue
 */
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "global_c_functions.pxi":169
 *             if attr_identifier.match(_x):
 *                 _rl += ['^%s$' % (_x,)]
 *         for _x in _rl:             # <<<<<<<<<<<<<<
 *             if _ret:
 *                 _ret = _ret + '|' + _x
 */
  __pyx_t_1 = __pyx_v__rl; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(5, 169, __pyx_L1_error)
    #else
    __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v__x, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "global_c_functions.pxi":170
 *                 _rl += ['^%s$' % (_x,)]
 *         for _x in _rl:
 *             if _ret:             # <<<<<<<<<<<<<<
 *                 _ret = _ret + '|' + _x
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v__ret); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 170, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "global_c_functions.pxi":171
 *         for _x in _rl:
 *             if _ret:
 *                 _ret = _ret + '|' + _x             # <<<<<<<<<<<<<<
 *             else:
 *                 _ret = _x
 */
      __pyx_t_4 = PyNumber_Add(__pyx_v__ret, __pyx_kp_s__4); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_v__x); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v__ret, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "global_c_functions.pxi":170
 *                 _rl += ['^%s$' % (_x,)]
 *         for _x in _rl:
 *             if _ret:             # <<<<<<<<<<<<<<
 *                 _ret = _ret + '|' + _x
 *             else:
 */
      goto __pyx_L10;
    }

    /* "global_c_functions.pxi":173
 *                 _ret = _ret + '|' + _x
 *             else:
 *                 _ret = _x             # <<<<<<<<<<<<<<
 *         return re.compile(_ret)
 * 
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v__x);
      __Pyx_DECREF_SET(__pyx_v__ret, __pyx_v__x);
    }
    __pyx_L10:;

    /* "global_c_functions.pxi":169
 *             if attr_identifier.match(_x):
 *                 _rl += ['^%s$' % (_x,)]
 *         for _x in _rl:             # <<<<<<<<<<<<<<
 *             if _ret:
 *                 _ret = _ret + '|' + _x
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "global_c_functions.pxi":174
 *             else:
 *                 _ret = _x
 *         return re.compile(_ret)             # <<<<<<<<<<<<<<
 * 
 *     ro_method = bool(kwargs.get('ro_method', False))
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_re); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_compile); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v__ret};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "global_c_functions.pxi":159
 *     Called once by protect() before Protected class initialization
 *     '''
 *     def _build_regex(alist):             # <<<<<<<<<<<<<<
 *         _ret = ''
 *         _rl = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyprotect.protected.protected_rules_from_kwargs._build_regex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__ret);
  __Pyx_XDECREF(__pyx_v__rl);
  __Pyx_XDECREF(__pyx_v__x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "global_c_functions.pxi":153
 * 
 * 
 * cdef protected_rules_from_kwargs(kwargs):             # <<<<<<<<<<<<<<
 *     '''
 *     kwargs-->dict
 */

static PyObject *__pyx_f_9pyprotect_9protected_protected_rules_from_kwargs(PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v__build_regex = 0;
  PyObject *__pyx_v_ro_method = NULL;
  PyObject *__pyx_v_ro_data = NULL;
  PyObject *__pyx_v_hide_private = NULL;
  PyObject *__pyx_v_ro = NULL;
  PyObject *__pyx_v_rw = NULL;
  PyObject *__pyx_v_hide = NULL;
  PyObject *__pyx_v_hide_regex = NULL;
  PyObject *__pyx_v_ro_regex = NULL;
  PyObject *__pyx_v_rw_regex = NULL;
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_kw = NULL;
  PyObject *__pyx_9genexpr11__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_rules_from_kwargs", 0);

  /* "global_c_functions.pxi":159
 *     Called once by protect() before Protected class initialization
 *     '''
 *     def _build_regex(alist):             # <<<<<<<<<<<<<<
 *         _ret = ''
 *         _rl = []
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_27protected_rules_from_kwargs_1_build_regex, 0, __pyx_n_s_protected_rules_from_kwargs_loca, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__6)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__build_regex = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "global_c_functions.pxi":176
 *         return re.compile(_ret)
 * 
 *     ro_method = bool(kwargs.get('ro_method', False))             # <<<<<<<<<<<<<<
 *     ro_data = bool(kwargs.get('ro_data', False))
 *     hide_private = kwargs.get('hide_private', False)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ro_method = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":177
 * 
 *     ro_method = bool(kwargs.get('ro_method', False))
 *     ro_data = bool(kwargs.get('ro_data', False))             # <<<<<<<<<<<<<<
 *     hide_private = kwargs.get('hide_private', False)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ro_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "global_c_functions.pxi":178
 *     ro_method = bool(kwargs.get('ro_method', False))
 *     ro_data = bool(kwargs.get('ro_data', False))
 *     hide_private = kwargs.get('hide_private', False)             # <<<<<<<<<<<<<<
 * 
 *     ro = [
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_hide_private = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":180
 *     hide_private = kwargs.get('hide_private', False)
 * 
 *     ro = [             # <<<<<<<<<<<<<<
 *         x for x in list(kwargs.get('ro', []))
 *         if isinstance(x, str) and attr_identifier.match(x)
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 180, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":181
 * 
 *     ro = [
 *         x for x in list(kwargs.get('ro', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 181, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 181, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_n_s_ro, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 181, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 181, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(5, 181, __pyx_L5_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 181, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":182
 *     ro = [
 *         x for x in list(kwargs.get('ro', []))
 *         if isinstance(x, str) and attr_identifier.match(x)             # <<<<<<<<<<<<<<
 *     ]
 *     rw = [
 */
      __pyx_t_9 = PyString_Check(__pyx_9genexpr11__pyx_v_x); 
      if (__pyx_t_9) {
      } else {
        __pyx_t_3 = __pyx_t_9;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_attr_identifier, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 182, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_9genexpr11__pyx_v_x};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 182, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(5, 182, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_3 = __pyx_t_9;
      __pyx_L9_bool_binop_done:;
      if (__pyx_t_3) {

        /* "global_c_functions.pxi":181
 * 
 *     ro = [
 *         x for x in list(kwargs.get('ro', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_9genexpr11__pyx_v_x))) __PYX_ERR(5, 180, __pyx_L5_error)

        /* "global_c_functions.pxi":182
 *     ro = [
 *         x for x in list(kwargs.get('ro', []))
 *         if isinstance(x, str) and attr_identifier.match(x)             # <<<<<<<<<<<<<<
 *     ]
 *     rw = [
 */
      }

      /* "global_c_functions.pxi":181
 * 
 *     ro = [
 *         x for x in list(kwargs.get('ro', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_x); __pyx_9genexpr11__pyx_v_x = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_x); __pyx_9genexpr11__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */
  __pyx_v_ro = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":184
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 *     rw = [             # <<<<<<<<<<<<<<
 *         x for x in list(kwargs.get('rw', []))
 *         if isinstance(x, str) and attr_identifier.match(x)
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 184, __pyx_L14_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":185
 *     ]
 *     rw = [
 *         x for x in list(kwargs.get('rw', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 185, __pyx_L14_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 185, __pyx_L14_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_n_s_rw, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 185, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 185, __pyx_L14_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(5, 185, __pyx_L14_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 185, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":186
 *     rw = [
 *         x for x in list(kwargs.get('rw', []))
 *         if isinstance(x, str) and attr_identifier.match(x)             # <<<<<<<<<<<<<<
 *     ]
 *     hide = [
 */
      __pyx_t_9 = PyString_Check(__pyx_9genexpr12__pyx_v_x); 
      if (__pyx_t_9) {
      } else {
        __pyx_t_3 = __pyx_t_9;
        goto __pyx_L18_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_attr_identifier, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 186, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_9genexpr12__pyx_v_x};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 186, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(5, 186, __pyx_L14_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_3 = __pyx_t_9;
      __pyx_L18_bool_binop_done:;
      if (__pyx_t_3) {

        /* "global_c_functions.pxi":185
 *     ]
 *     rw = [
 *         x for x in list(kwargs.get('rw', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_9genexpr12__pyx_v_x))) __PYX_ERR(5, 184, __pyx_L14_error)

        /* "global_c_functions.pxi":186
 *     rw = [
 *         x for x in list(kwargs.get('rw', []))
 *         if isinstance(x, str) and attr_identifier.match(x)             # <<<<<<<<<<<<<<
 *     ]
 *     hide = [
 */
      }

      /* "global_c_functions.pxi":185
 *     ]
 *     rw = [
 *         x for x in list(kwargs.get('rw', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_x); __pyx_9genexpr12__pyx_v_x = 0;
    goto __pyx_L20_exit_scope;
    __pyx_L14_error:;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_x); __pyx_9genexpr12__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L20_exit_scope:;
  } /* exit inner scope */
  __pyx_v_rw = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":188
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 *     hide = [             # <<<<<<<<<<<<<<
 *         x for x in list(kwargs.get('hide', []))
 *         if isinstance(x, str) and attr_identifier.match(x)
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 188, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "global_c_functions.pxi":189
 *     ]
 *     hide = [
 *         x for x in list(kwargs.get('hide', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 189, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 189, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_n_s_hide, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 189, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 189, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(5, 189, __pyx_L23_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 189, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_x, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "global_c_functions.pxi":190
 *     hide = [
 *         x for x in list(kwargs.get('hide', []))
 *         if isinstance(x, str) and attr_identifier.match(x)             # <<<<<<<<<<<<<<
 *     ]
 *     ro = frozenset(ro)
 */
      __pyx_t_9 = PyString_Check(__pyx_9genexpr13__pyx_v_x); 
      if (__pyx_t_9) {
      } else {
        __pyx_t_3 = __pyx_t_9;
        goto __pyx_L27_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_attr_identifier, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 190, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_9genexpr13__pyx_v_x};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 190, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(5, 190, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_3 = __pyx_t_9;
      __pyx_L27_bool_binop_done:;
      if (__pyx_t_3) {

        /* "global_c_functions.pxi":189
 *     ]
 *     hide = [
 *         x for x in list(kwargs.get('hide', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_9genexpr13__pyx_v_x))) __PYX_ERR(5, 188, __pyx_L23_error)

        /* "global_c_functions.pxi":190
 *     hide = [
 *         x for x in list(kwargs.get('hide', []))
 *         if isinstance(x, str) and attr_identifier.match(x)             # <<<<<<<<<<<<<<
 *     ]
 *     ro = frozenset(ro)
 */
      }

      /* "global_c_functions.pxi":189
 *     ]
 *     hide = [
 *         x for x in list(kwargs.get('hide', []))             # <<<<<<<<<<<<<<
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_x); __pyx_9genexpr13__pyx_v_x = 0;
    goto __pyx_L29_exit_scope;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_x); __pyx_9genexpr13__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L29_exit_scope:;
  } /* exit inner scope */
  __pyx_v_hide = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":192
 *         if isinstance(x, str) and attr_identifier.match(x)
 *     ]
 *     ro = frozenset(ro)             # <<<<<<<<<<<<<<
 *     rw = frozenset(rw)
 *     hide = frozenset(hide)
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_v_ro); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_ro, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":193
 *     ]
 *     ro = frozenset(ro)
 *     rw = frozenset(rw)             # <<<<<<<<<<<<<<
 *     hide = frozenset(hide)
 * 
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_v_rw); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_rw, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":194
 *     ro = frozenset(ro)
 *     rw = frozenset(rw)
 *     hide = frozenset(hide)             # <<<<<<<<<<<<<<
 * 
 *     # Build regexes
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_v_hide); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_hide, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":197
 * 
 *     # Build regexes
 *     hide_regex = _build_regex(hide)             # <<<<<<<<<<<<<<
 *     ro_regex = _build_regex(ro)
 *     rw_regex = _build_regex(rw)
 */
  __pyx_t_2 = __pyx_pf_9pyprotect_9protected_27protected_rules_from_kwargs__build_regex(__pyx_v__build_regex, __pyx_v_hide); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_hide_regex = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":198
 *     # Build regexes
 *     hide_regex = _build_regex(hide)
 *     ro_regex = _build_regex(ro)             # <<<<<<<<<<<<<<
 *     rw_regex = _build_regex(rw)
 * 
 */
  __pyx_t_2 = __pyx_pf_9pyprotect_9protected_27protected_rules_from_kwargs__build_regex(__pyx_v__build_regex, __pyx_v_ro); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ro_regex = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":199
 *     hide_regex = _build_regex(hide)
 *     ro_regex = _build_regex(ro)
 *     rw_regex = _build_regex(rw)             # <<<<<<<<<<<<<<
 * 
 *     d = {
 */
  __pyx_t_2 = __pyx_pf_9pyprotect_9protected_27protected_rules_from_kwargs__build_regex(__pyx_v__build_regex, __pyx_v_rw); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_rw_regex = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":202
 * 
 *     d = {
 *         'hide_private': hide_private,             # <<<<<<<<<<<<<<
 *         'hide_regex': hide_regex,
 *         'ro_regex': ro_regex,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_hide_private, __pyx_v_hide_private) < 0) __PYX_ERR(5, 202, __pyx_L1_error)

  /* "global_c_functions.pxi":203
 *     d = {
 *         'hide_private': hide_private,
 *         'hide_regex': hide_regex,             # <<<<<<<<<<<<<<
 *         'ro_regex': ro_regex,
 *         'rw_regex': rw_regex,
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_hide_regex, __pyx_v_hide_regex) < 0) __PYX_ERR(5, 202, __pyx_L1_error)

  /* "global_c_functions.pxi":204
 *         'hide_private': hide_private,
 *         'hide_regex': hide_regex,
 *         'ro_regex': ro_regex,             # <<<<<<<<<<<<<<
 *         'rw_regex': rw_regex,
 *         'ro_method': bool(ro_method),
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ro_regex, __pyx_v_ro_regex) < 0) __PYX_ERR(5, 202, __pyx_L1_error)

  /* "global_c_functions.pxi":205
 *         'hide_regex': hide_regex,
 *         'ro_regex': ro_regex,
 *         'rw_regex': rw_regex,             # <<<<<<<<<<<<<<
 *         'ro_method': bool(ro_method),
 *         'ro_data': bool(ro_data),
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_rw_regex, __pyx_v_rw_regex) < 0) __PYX_ERR(5, 202, __pyx_L1_error)

  /* "global_c_functions.pxi":206
 *         'ro_regex': ro_regex,
 *         'rw_regex': rw_regex,
 *         'ro_method': bool(ro_method),             # <<<<<<<<<<<<<<
 *         'ro_data': bool(ro_data),
 *     }
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_ro_method); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 206, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ro_method, __pyx_t_1) < 0) __PYX_ERR(5, 202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "global_c_functions.pxi":207
 *         'rw_regex': rw_regex,
 *         'ro_method': bool(ro_method),
 *         'ro_data': bool(ro_data),             # <<<<<<<<<<<<<<
 *     }
 *     d['dynamic'] = kwargs.get('dynamic', False)
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_ro_data); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 207, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ro_data, __pyx_t_1) < 0) __PYX_ERR(5, 202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_d = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":209
 *         'ro_data': bool(ro_data),
 *     }
 *     d['dynamic'] = kwargs.get('dynamic', False)             # <<<<<<<<<<<<<<
 *     d['frozen'] = bool(kwargs.get('frozen', False))
 *     d['kwargs'] = kwargs
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_n_s_dynamic, __pyx_t_1) < 0))) __PYX_ERR(5, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "global_c_functions.pxi":210
 *     }
 *     d['dynamic'] = kwargs.get('dynamic', False)
 *     d['frozen'] = bool(kwargs.get('frozen', False))             # <<<<<<<<<<<<<<
 *     d['kwargs'] = kwargs
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_kwargs, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_n_s_frozen, __pyx_t_2) < 0))) __PYX_ERR(5, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":211
 *     d['dynamic'] = kwargs.get('dynamic', False)
 *     d['frozen'] = bool(kwargs.get('frozen', False))
 *     d['kwargs'] = kwargs             # <<<<<<<<<<<<<<
 * 
 *     d['attr_type_check'] = False
 */
  if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_n_s_kwargs, __pyx_v_kwargs) < 0))) __PYX_ERR(5, 211, __pyx_L1_error)

  /* "global_c_functions.pxi":213
 *     d['kwargs'] = kwargs
 * 
 *     d['attr_type_check'] = False             # <<<<<<<<<<<<<<
 *     for kw in ('ro_method', 'ro_data'):
 *         if bool(d.get(kw, False)):
 */
  if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_n_s_attr_type_check, Py_False) < 0))) __PYX_ERR(5, 213, __pyx_L1_error)

  /* "global_c_functions.pxi":214
 * 
 *     d['attr_type_check'] = False
 *     for kw in ('ro_method', 'ro_data'):             # <<<<<<<<<<<<<<
 *         if bool(d.get(kw, False)):
 *             d['attr_type_check'] = True
 */
  __pyx_t_2 = __pyx_tuple__12; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
  for (;;) {
    if (__pyx_t_8 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(5, 214, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_kw, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "global_c_functions.pxi":215
 *     d['attr_type_check'] = False
 *     for kw in ('ro_method', 'ro_data'):
 *         if bool(d.get(kw, False)):             # <<<<<<<<<<<<<<
 *             d['attr_type_check'] = True
 * 
 */
    __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_d, __pyx_v_kw, Py_False); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 215, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((!(!__pyx_t_3))) {

      /* "global_c_functions.pxi":216
 *     for kw in ('ro_method', 'ro_data'):
 *         if bool(d.get(kw, False)):
 *             d['attr_type_check'] = True             # <<<<<<<<<<<<<<
 * 
 *     return d
 */
      if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_n_s_attr_type_check, Py_True) < 0))) __PYX_ERR(5, 216, __pyx_L1_error)

      /* "global_c_functions.pxi":215
 *     d['attr_type_check'] = False
 *     for kw in ('ro_method', 'ro_data'):
 *         if bool(d.get(kw, False)):             # <<<<<<<<<<<<<<
 *             d['attr_type_check'] = True
 * 
 */
    }

    /* "global_c_functions.pxi":214
 * 
 *     d['attr_type_check'] = False
 *     for kw in ('ro_method', 'ro_data'):             # <<<<<<<<<<<<<<
 *         if bool(d.get(kw, False)):
 *             d['attr_type_check'] = True
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":218
 *             d['attr_type_check'] = True
 * 
 *     return d             # <<<<<<<<<<<<<<
 * 
 * cdef protected_merge_kwargs(kw1: dict, kw2: dict):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_d);
  __pyx_r = __pyx_v_d;
  goto __pyx_L0;

  /* "global_c_functions.pxi":153
 * 
 * 
 * cdef protected_rules_from_kwargs(kwargs):             # <<<<<<<<<<<<<<
 *     '''
 *     kwargs-->dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.protected_rules_from_kwargs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__build_regex);
  __Pyx_XDECREF(__pyx_v_ro_method);
  __Pyx_XDECREF(__pyx_v_ro_data);
  __Pyx_XDECREF(__pyx_v_hide_private);
  __Pyx_XDECREF(__pyx_v_ro);
  __Pyx_XDECREF(__pyx_v_rw);
  __Pyx_XDECREF(__pyx_v_hide);
  __Pyx_XDECREF(__pyx_v_hide_regex);
  __Pyx_XDECREF(__pyx_v_ro_regex);
  __Pyx_XDECREF(__pyx_v_rw_regex);
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_kw);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "global_c_functions.pxi":220
 *     return d
 * 
 * cdef protected_merge_kwargs(kw1: dict, kw2: dict):             # <<<<<<<<<<<<<<
 *     '''
 *     Merges kw1 and kw2 to return dict with most restrictive options
 */

static PyObject *__pyx_f_9pyprotect_9protected_protected_merge_kwargs(PyObject *__pyx_v_kw1, PyObject *__pyx_v_kw2) {
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_v_s1 = NULL;
  PyObject *__pyx_v_s2 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_merge_kwargs", 0);
  __Pyx_INCREF(__pyx_v_kw1);
  __Pyx_INCREF(__pyx_v_kw2);

  /* "global_c_functions.pxi":227
 *     Called once by protect() before Protected class initialization
 *     '''
 *     (kw1, kw2) = (dict(kw1), dict(kw2))             # <<<<<<<<<<<<<<
 *     d = {}
 *     # Permissive bool options - must be 'and-ed'
 */
  __pyx_t_1 = PyDict_Copy(__pyx_v_kw1); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kw2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_kw1, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_kw2, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":228
 *     '''
 *     (kw1, kw2) = (dict(kw1), dict(kw2))
 *     d = {}             # <<<<<<<<<<<<<<
 *     # Permissive bool options - must be 'and-ed'
 *     # dynamic defaults to True while add defaults to False
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_d = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_c_functions.pxi":231
 *     # Permissive bool options - must be 'and-ed'
 *     # dynamic defaults to True while add defaults to False
 *     a = 'dynamic'             # <<<<<<<<<<<<<<
 *     d[a] = (kw1.get(a, True) and kw2.get(a, True))
 * 
 */
  __Pyx_INCREF(__pyx_n_s_dynamic);
  __pyx_v_a = __pyx_n_s_dynamic;

  /* "global_c_functions.pxi":232
 *     # dynamic defaults to True while add defaults to False
 *     a = 'dynamic'
 *     d[a] = (kw1.get(a, True) and kw2.get(a, True))             # <<<<<<<<<<<<<<
 * 
 *     # Restrictive bool options must be 'or-ed'
 */
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw1, __pyx_v_a, Py_True); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 232, __pyx_L1_error)
  if (__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw2, __pyx_v_a, Py_True); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_L3_bool_binop_done:;
  if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_v_a, __pyx_t_2) < 0))) __PYX_ERR(5, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":235
 * 
 *     # Restrictive bool options must be 'or-ed'
 *     for a in (             # <<<<<<<<<<<<<<
 *         'frozen', 'hide_private', 'ro_data', 'ro_method',
 *     ):
 */
  __pyx_t_2 = __pyx_tuple__13; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= 4) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(5, 235, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF_SET(__pyx_v_a, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "global_c_functions.pxi":238
 *         'frozen', 'hide_private', 'ro_data', 'ro_method',
 *     ):
 *         d[a] = (kw1.get(a, False) or kw2.get(a, False))             # <<<<<<<<<<<<<<
 * 
 *     # Restrictive lists (non-bool) are unioned
 */
    __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw1, __pyx_v_a, Py_False); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 238, __pyx_L1_error)
    if (!__pyx_t_3) {
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_1 = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw2, __pyx_v_a, Py_False); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_1 = __pyx_t_5;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_L7_bool_binop_done:;
    if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_v_a, __pyx_t_1) < 0))) __PYX_ERR(5, 238, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "global_c_functions.pxi":235
 * 
 *     # Restrictive bool options must be 'or-ed'
 *     for a in (             # <<<<<<<<<<<<<<
 *         'frozen', 'hide_private', 'ro_data', 'ro_method',
 *     ):
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":241
 * 
 *     # Restrictive lists (non-bool) are unioned
 *     for a in (             # <<<<<<<<<<<<<<
 *         'ro', 'hide',
 *     ):
 */
  __pyx_t_2 = __pyx_tuple__14; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= 2) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(5, 241, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF_SET(__pyx_v_a, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "global_c_functions.pxi":244
 *         'ro', 'hide',
 *     ):
 *         s1 = set(list(kw1.get(a, [])))             # <<<<<<<<<<<<<<
 *         s2 = set(list(kw2.get(a, [])))
 *         d[a] = list(
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw1, __pyx_v_a, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_s1, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "global_c_functions.pxi":245
 *     ):
 *         s1 = set(list(kw1.get(a, [])))
 *         s2 = set(list(kw2.get(a, [])))             # <<<<<<<<<<<<<<
 *         d[a] = list(
 *             s1.union(s2)
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw2, __pyx_v_a, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PySet_New(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_s2, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "global_c_functions.pxi":247
 *         s2 = set(list(kw2.get(a, [])))
 *         d[a] = list(
 *             s1.union(s2)             # <<<<<<<<<<<<<<
 *         )
 *     # Permissive lists (non-bool) are intersected
 */
    __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_union, __pyx_v_s1, __pyx_v_s2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "global_c_functions.pxi":246
 *         s1 = set(list(kw1.get(a, [])))
 *         s2 = set(list(kw2.get(a, [])))
 *         d[a] = list(             # <<<<<<<<<<<<<<
 *             s1.union(s2)
 *         )
 */
    __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_v_a, __pyx_t_5) < 0))) __PYX_ERR(5, 246, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "global_c_functions.pxi":241
 * 
 *     # Restrictive lists (non-bool) are unioned
 *     for a in (             # <<<<<<<<<<<<<<
 *         'ro', 'hide',
 *     ):
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":250
 *         )
 *     # Permissive lists (non-bool) are intersected
 *     for a in (             # <<<<<<<<<<<<<<
 *         'rw',
 *     ):
 */
  __pyx_t_2 = __pyx_tuple__15; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= 1) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(5, 250, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF_SET(__pyx_v_a, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "global_c_functions.pxi":253
 *         'rw',
 *     ):
 *         s1 = set(list(kw1.get(a, [])))             # <<<<<<<<<<<<<<
 *         s2 = set(list(kw2.get(a, [])))
 *         d[a] = list(
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw1, __pyx_v_a, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PySet_New(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_s1, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "global_c_functions.pxi":254
 *     ):
 *         s1 = set(list(kw1.get(a, [])))
 *         s2 = set(list(kw2.get(a, [])))             # <<<<<<<<<<<<<<
 *         d[a] = list(
 *             s1.intersection(s2)
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_kw2, __pyx_v_a, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_s2, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "global_c_functions.pxi":256
 *         s2 = set(list(kw2.get(a, [])))
 *         d[a] = list(
 *             s1.intersection(s2)             # <<<<<<<<<<<<<<
 *         )
 *     return d
 */
    __pyx_t_5 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_intersection, __pyx_v_s1, __pyx_v_s2); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "global_c_functions.pxi":255
 *         s1 = set(list(kw1.get(a, [])))
 *         s2 = set(list(kw2.get(a, [])))
 *         d[a] = list(             # <<<<<<<<<<<<<<
 *             s1.intersection(s2)
 *         )
 */
    __pyx_t_1 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_v_a, __pyx_t_1) < 0))) __PYX_ERR(5, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "global_c_functions.pxi":250
 *         )
 *     # Permissive lists (non-bool) are intersected
 *     for a in (             # <<<<<<<<<<<<<<
 *         'rw',
 *     ):
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "global_c_functions.pxi":258
 *             s1.intersection(s2)
 *         )
 *     return d             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_d);
  __pyx_r = __pyx_v_d;
  goto __pyx_L0;

  /* "global_c_functions.pxi":220
 *     return d
 * 
 * cdef protected_merge_kwargs(kw1: dict, kw2: dict):             # <<<<<<<<<<<<<<
 *     '''
 *     Merges kw1 and kw2 to return dict with most restrictive options
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.protected_merge_kwargs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XDECREF(__pyx_v_s1);
  __Pyx_XDECREF(__pyx_v_s2);
  __Pyx_XDECREF(__pyx_v_kw1);
  __Pyx_XDECREF(__pyx_v_kw2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "global_c_functions.pxi":261
 * 
 * 
 * cdef privatedict(o, cn, frozen=False, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *     '''
 *     o-->Mapping (to be wrapped)
 */

static PyObject *__pyx_f_9pyprotect_9protected_privatedict(PyObject *__pyx_v_o, PyObject *__pyx_v_cn, struct __pyx_opt_args_9pyprotect_9protected_privatedict *__pyx_optional_args) {
  PyObject *__pyx_v_frozen = ((PyObject *)Py_False);
  PyObject *__pyx_v_oldstyle_class = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("privatedict", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_frozen = __pyx_optional_args->frozen;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_oldstyle_class = __pyx_optional_args->oldstyle_class;
      }
    }
  }

  /* "global_c_functions.pxi":266
 *     Returns-->FrozenPrivacyDict if frozen; Privacybject otherwise
 *     '''
 *     if frozen:             # <<<<<<<<<<<<<<
 *         if isinstance (o, FrozenPrivacyDict):
 *             return o
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_frozen); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(5, 266, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "global_c_functions.pxi":267
 *     '''
 *     if frozen:
 *         if isinstance (o, FrozenPrivacyDict):             # <<<<<<<<<<<<<<
 *             return o
 *         return FrozenPrivacyDict(o, cn, oldstyle_class)
 */
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict); 
    if (__pyx_t_1) {

      /* "global_c_functions.pxi":268
 *     if frozen:
 *         if isinstance (o, FrozenPrivacyDict):
 *             return o             # <<<<<<<<<<<<<<
 *         return FrozenPrivacyDict(o, cn, oldstyle_class)
 *     else:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_o);
      __pyx_r = __pyx_v_o;
      goto __pyx_L0;

      /* "global_c_functions.pxi":267
 *     '''
 *     if frozen:
 *         if isinstance (o, FrozenPrivacyDict):             # <<<<<<<<<<<<<<
 *             return o
 *         return FrozenPrivacyDict(o, cn, oldstyle_class)
 */
    }

    /* "global_c_functions.pxi":269
 *         if isinstance (o, FrozenPrivacyDict):
 *             return o
 *         return FrozenPrivacyDict(o, cn, oldstyle_class)             # <<<<<<<<<<<<<<
 *     else:
 *         if isinstance (o, FrozenPrivacyDict):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_o);
    __Pyx_GIVEREF(__pyx_v_o);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_o);
    __Pyx_INCREF(__pyx_v_cn);
    __Pyx_GIVEREF(__pyx_v_cn);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_cn);
    __Pyx_INCREF(__pyx_v_oldstyle_class);
    __Pyx_GIVEREF(__pyx_v_oldstyle_class);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_oldstyle_class);
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "global_c_functions.pxi":266
 *     Returns-->FrozenPrivacyDict if frozen; Privacybject otherwise
 *     '''
 *     if frozen:             # <<<<<<<<<<<<<<
 *         if isinstance (o, FrozenPrivacyDict):
 *             return o
 */
  }

  /* "global_c_functions.pxi":271
 *         return FrozenPrivacyDict(o, cn, oldstyle_class)
 *     else:
 *         if isinstance (o, FrozenPrivacyDict):             # <<<<<<<<<<<<<<
 *             # Underlying already frozen
 *             return o
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict); 
    if (__pyx_t_1) {

      /* "global_c_functions.pxi":273
 *         if isinstance (o, FrozenPrivacyDict):
 *             # Underlying already frozen
 *             return o             # <<<<<<<<<<<<<<
 *         elif isinstance(o, PrivacyDict):
 *             return o
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_o);
      __pyx_r = __pyx_v_o;
      goto __pyx_L0;

      /* "global_c_functions.pxi":271
 *         return FrozenPrivacyDict(o, cn, oldstyle_class)
 *     else:
 *         if isinstance (o, FrozenPrivacyDict):             # <<<<<<<<<<<<<<
 *             # Underlying already frozen
 *             return o
 */
    }

    /* "global_c_functions.pxi":274
 *             # Underlying already frozen
 *             return o
 *         elif isinstance(o, PrivacyDict):             # <<<<<<<<<<<<<<
 *             return o
 *         return PrivacyDict(o, cn, oldstyle_class)
 */
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_PrivacyDict); 
    if (__pyx_t_1) {

      /* "global_c_functions.pxi":275
 *             return o
 *         elif isinstance(o, PrivacyDict):
 *             return o             # <<<<<<<<<<<<<<
 *         return PrivacyDict(o, cn, oldstyle_class)
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_o);
      __pyx_r = __pyx_v_o;
      goto __pyx_L0;

      /* "global_c_functions.pxi":274
 *             # Underlying already frozen
 *             return o
 *         elif isinstance(o, PrivacyDict):             # <<<<<<<<<<<<<<
 *             return o
 *         return PrivacyDict(o, cn, oldstyle_class)
 */
    }

    /* "global_c_functions.pxi":276
 *         elif isinstance(o, PrivacyDict):
 *             return o
 *         return PrivacyDict(o, cn, oldstyle_class)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_o);
    __Pyx_GIVEREF(__pyx_v_o);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
    __Pyx_INCREF(__pyx_v_cn);
    __Pyx_GIVEREF(__pyx_v_cn);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_cn);
    __Pyx_INCREF(__pyx_v_oldstyle_class);
    __Pyx_GIVEREF(__pyx_v_oldstyle_class);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_oldstyle_class);
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "global_c_functions.pxi":261
 * 
 * 
 * cdef privatedict(o, cn, frozen=False, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *     '''
 *     o-->Mapping (to be wrapped)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.privatedict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":39
 *     cdef object attributes_map
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         id_val,
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_id_val = 0;
  PyObject *__pyx_v_id_class = 0;
  PyObject *__pyx_v_hash_val = 0;
  PyObject *__pyx_v_isinstance_val = 0;
  PyObject *__pyx_v_issubclass_val = 0;
  PyObject *__pyx_v_instanceof = 0;
  PyObject *__pyx_v_subclassof = 0;
  PyObject *__pyx_v_help_val = 0;
  PyObject *__pyx_v_help_str = 0;
  PyObject *__pyx_v_testop = 0;
  PyObject *__pyx_v_rules = 0;
  PyObject *__pyx_v_freeze = 0;
  PyObject *__pyx_v_private = 0;
  PyObject *__pyx_v_protect = 0;
  PyObject *__pyx_v_multiwrapped = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_id_val,&__pyx_n_s_id_class,&__pyx_n_s_hash_val,&__pyx_n_s_isinstance_val,&__pyx_n_s_issubclass_val,&__pyx_n_s_instanceof,&__pyx_n_s_subclassof,&__pyx_n_s_help_val,&__pyx_n_s_help_str,&__pyx_n_s_testop,&__pyx_n_s_rules,&__pyx_n_s_freeze,&__pyx_n_s_private,&__pyx_n_s_protect,&__pyx_n_s_multiwrapped,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_id_val,&__pyx_n_s_id_class,&__pyx_n_s_hash_val,&__pyx_n_s_isinstance_val,&__pyx_n_s_issubclass_val,&__pyx_n_s_instanceof,&__pyx_n_s_subclassof,&__pyx_n_s_help_val,&__pyx_n_s_help_str,&__pyx_n_s_testop,&__pyx_n_s_rules,&__pyx_n_s_freeze,&__pyx_n_s_private,&__pyx_n_s_protect,&__pyx_n_s_multiwrapped,0};
    #endif
    PyObject* values[15] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case 15: values[14] = __Pyx_Arg_VARARGS(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = __Pyx_Arg_VARARGS(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = __Pyx_Arg_VARARGS(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = __Pyx_Arg_VARARGS(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = __Pyx_Arg_VARARGS(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = __Pyx_Arg_VARARGS(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = __Pyx_Arg_VARARGS(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_id_val)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_id_class)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 1); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_hash_val)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 2); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_isinstance_val)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 3); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_issubclass_val)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 4); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_instanceof)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 5); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_subclassof)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 6); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_help_val)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 7); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_help_str)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 8); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_testop)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 9); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rules)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 10); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_freeze)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 11); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_private)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 12); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protect)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 13); __PYX_ERR(1, 39, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_multiwrapped)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, 14); __PYX_ERR(1, 39, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 39, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 15)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
      values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
      values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
      values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
      values[8] = __Pyx_Arg_VARARGS(__pyx_args, 8);
      values[9] = __Pyx_Arg_VARARGS(__pyx_args, 9);
      values[10] = __Pyx_Arg_VARARGS(__pyx_args, 10);
      values[11] = __Pyx_Arg_VARARGS(__pyx_args, 11);
      values[12] = __Pyx_Arg_VARARGS(__pyx_args, 12);
      values[13] = __Pyx_Arg_VARARGS(__pyx_args, 13);
      values[14] = __Pyx_Arg_VARARGS(__pyx_args, 14);
    }
    __pyx_v_id_val = values[0];
    __pyx_v_id_class = values[1];
    __pyx_v_hash_val = values[2];
    __pyx_v_isinstance_val = values[3];
    __pyx_v_issubclass_val = values[4];
    __pyx_v_instanceof = values[5];
    __pyx_v_subclassof = values[6];
    __pyx_v_help_val = values[7];
    __pyx_v_help_str = values[8];
    __pyx_v_testop = values[9];
    __pyx_v_rules = values[10];
    __pyx_v_freeze = values[11];
    __pyx_v_private = values[12];
    __pyx_v_protect = values[13];
    __pyx_v_multiwrapped = values[14];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 15, 15, __pyx_nargs); __PYX_ERR(1, 39, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData___init__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), __pyx_v_id_val, __pyx_v_id_class, __pyx_v_hash_val, __pyx_v_isinstance_val, __pyx_v_issubclass_val, __pyx_v_instanceof, __pyx_v_subclassof, __pyx_v_help_val, __pyx_v_help_str, __pyx_v_testop, __pyx_v_rules, __pyx_v_freeze, __pyx_v_private, __pyx_v_protect, __pyx_v_multiwrapped);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData___init__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_id_val, PyObject *__pyx_v_id_class, PyObject *__pyx_v_hash_val, PyObject *__pyx_v_isinstance_val, PyObject *__pyx_v_issubclass_val, PyObject *__pyx_v_instanceof, PyObject *__pyx_v_subclassof, PyObject *__pyx_v_help_val, PyObject *__pyx_v_help_str, PyObject *__pyx_v_testop, PyObject *__pyx_v_rules, PyObject *__pyx_v_freeze, PyObject *__pyx_v_private, PyObject *__pyx_v_protect, PyObject *__pyx_v_multiwrapped) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ProtectionData.pxi":57
 *         multiwrapped,
 *     ):
 *         self.id = id_val             # <<<<<<<<<<<<<<
 *         self.id_class = id_class
 *         self.hash = hash_val
 */
  __Pyx_INCREF(__pyx_v_id_val);
  __Pyx_GIVEREF(__pyx_v_id_val);
  __Pyx_GOTREF(__pyx_v_self->id);
  __Pyx_DECREF(__pyx_v_self->id);
  __pyx_v_self->id = __pyx_v_id_val;

  /* "ProtectionData.pxi":58
 *     ):
 *         self.id = id_val
 *         self.id_class = id_class             # <<<<<<<<<<<<<<
 *         self.hash = hash_val
 *         self.isinstance = isinstance_val
 */
  __Pyx_INCREF(__pyx_v_id_class);
  __Pyx_GIVEREF(__pyx_v_id_class);
  __Pyx_GOTREF(__pyx_v_self->id_class);
  __Pyx_DECREF(__pyx_v_self->id_class);
  __pyx_v_self->id_class = __pyx_v_id_class;

  /* "ProtectionData.pxi":59
 *         self.id = id_val
 *         self.id_class = id_class
 *         self.hash = hash_val             # <<<<<<<<<<<<<<
 *         self.isinstance = isinstance_val
 *         self.issubclass = issubclass_val
 */
  __Pyx_INCREF(__pyx_v_hash_val);
  __Pyx_GIVEREF(__pyx_v_hash_val);
  __Pyx_GOTREF(__pyx_v_self->hash);
  __Pyx_DECREF(__pyx_v_self->hash);
  __pyx_v_self->hash = __pyx_v_hash_val;

  /* "ProtectionData.pxi":60
 *         self.id_class = id_class
 *         self.hash = hash_val
 *         self.isinstance = isinstance_val             # <<<<<<<<<<<<<<
 *         self.issubclass = issubclass_val
 *         self.instanceof = instanceof
 */
  __Pyx_INCREF(__pyx_v_isinstance_val);
  __Pyx_GIVEREF(__pyx_v_isinstance_val);
  __Pyx_GOTREF(__pyx_v_self->isinstance);
  __Pyx_DECREF(__pyx_v_self->isinstance);
  __pyx_v_self->isinstance = __pyx_v_isinstance_val;

  /* "ProtectionData.pxi":61
 *         self.hash = hash_val
 *         self.isinstance = isinstance_val
 *         self.issubclass = issubclass_val             # <<<<<<<<<<<<<<
 *         self.instanceof = instanceof
 *         self.subclassof = subclassof
 */
  __Pyx_INCREF(__pyx_v_issubclass_val);
  __Pyx_GIVEREF(__pyx_v_issubclass_val);
  __Pyx_GOTREF(__pyx_v_self->issubclass);
  __Pyx_DECREF(__pyx_v_self->issubclass);
  __pyx_v_self->issubclass = __pyx_v_issubclass_val;

  /* "ProtectionData.pxi":62
 *         self.isinstance = isinstance_val
 *         self.issubclass = issubclass_val
 *         self.instanceof = instanceof             # <<<<<<<<<<<<<<
 *         self.subclassof = subclassof
 *         self.help = help_val
 */
  __Pyx_INCREF(__pyx_v_instanceof);
  __Pyx_GIVEREF(__pyx_v_instanceof);
  __Pyx_GOTREF(__pyx_v_self->instanceof);
  __Pyx_DECREF(__pyx_v_self->instanceof);
  __pyx_v_self->instanceof = __pyx_v_instanceof;

  /* "ProtectionData.pxi":63
 *         self.issubclass = issubclass_val
 *         self.instanceof = instanceof
 *         self.subclassof = subclassof             # <<<<<<<<<<<<<<
 *         self.help = help_val
 *         self.help_str = help_str
 */
  __Pyx_INCREF(__pyx_v_subclassof);
  __Pyx_GIVEREF(__pyx_v_subclassof);
  __Pyx_GOTREF(__pyx_v_self->subclassof);
  __Pyx_DECREF(__pyx_v_self->subclassof);
  __pyx_v_self->subclassof = __pyx_v_subclassof;

  /* "ProtectionData.pxi":64
 *         self.instanceof = instanceof
 *         self.subclassof = subclassof
 *         self.help = help_val             # <<<<<<<<<<<<<<
 *         self.help_str = help_str
 *         self.testop = testop
 */
  __Pyx_INCREF(__pyx_v_help_val);
  __Pyx_GIVEREF(__pyx_v_help_val);
  __Pyx_GOTREF(__pyx_v_self->help);
  __Pyx_DECREF(__pyx_v_self->help);
  __pyx_v_self->help = __pyx_v_help_val;

  /* "ProtectionData.pxi":65
 *         self.subclassof = subclassof
 *         self.help = help_val
 *         self.help_str = help_str             # <<<<<<<<<<<<<<
 *         self.testop = testop
 *         self.rules = rules
 */
  __Pyx_INCREF(__pyx_v_help_str);
  __Pyx_GIVEREF(__pyx_v_help_str);
  __Pyx_GOTREF(__pyx_v_self->help_str);
  __Pyx_DECREF(__pyx_v_self->help_str);
  __pyx_v_self->help_str = __pyx_v_help_str;

  /* "ProtectionData.pxi":66
 *         self.help = help_val
 *         self.help_str = help_str
 *         self.testop = testop             # <<<<<<<<<<<<<<
 *         self.rules = rules
 *         self.freeze = freeze
 */
  __Pyx_INCREF(__pyx_v_testop);
  __Pyx_GIVEREF(__pyx_v_testop);
  __Pyx_GOTREF(__pyx_v_self->testop);
  __Pyx_DECREF(__pyx_v_self->testop);
  __pyx_v_self->testop = __pyx_v_testop;

  /* "ProtectionData.pxi":67
 *         self.help_str = help_str
 *         self.testop = testop
 *         self.rules = rules             # <<<<<<<<<<<<<<
 *         self.freeze = freeze
 *         self.multiwrapped = multiwrapped
 */
  __Pyx_INCREF(__pyx_v_rules);
  __Pyx_GIVEREF(__pyx_v_rules);
  __Pyx_GOTREF(__pyx_v_self->rules);
  __Pyx_DECREF(__pyx_v_self->rules);
  __pyx_v_self->rules = __pyx_v_rules;

  /* "ProtectionData.pxi":68
 *         self.testop = testop
 *         self.rules = rules
 *         self.freeze = freeze             # <<<<<<<<<<<<<<
 *         self.multiwrapped = multiwrapped
 *         self.private = private
 */
  __Pyx_INCREF(__pyx_v_freeze);
  __Pyx_GIVEREF(__pyx_v_freeze);
  __Pyx_GOTREF(__pyx_v_self->freeze);
  __Pyx_DECREF(__pyx_v_self->freeze);
  __pyx_v_self->freeze = __pyx_v_freeze;

  /* "ProtectionData.pxi":69
 *         self.rules = rules
 *         self.freeze = freeze
 *         self.multiwrapped = multiwrapped             # <<<<<<<<<<<<<<
 *         self.private = private
 *         self.protect = protect
 */
  __Pyx_INCREF(__pyx_v_multiwrapped);
  __Pyx_GIVEREF(__pyx_v_multiwrapped);
  __Pyx_GOTREF(__pyx_v_self->multiwrapped);
  __Pyx_DECREF(__pyx_v_self->multiwrapped);
  __pyx_v_self->multiwrapped = __pyx_v_multiwrapped;

  /* "ProtectionData.pxi":70
 *         self.freeze = freeze
 *         self.multiwrapped = multiwrapped
 *         self.private = private             # <<<<<<<<<<<<<<
 *         self.protect = protect
 * 
 */
  __Pyx_INCREF(__pyx_v_private);
  __Pyx_GIVEREF(__pyx_v_private);
  __Pyx_GOTREF(__pyx_v_self->private);
  __Pyx_DECREF(__pyx_v_self->private);
  __pyx_v_self->private = __pyx_v_private;

  /* "ProtectionData.pxi":71
 *         self.multiwrapped = multiwrapped
 *         self.private = private
 *         self.protect = protect             # <<<<<<<<<<<<<<
 * 
 *         self.attributes_map = {
 */
  __Pyx_INCREF(__pyx_v_protect);
  __Pyx_GIVEREF(__pyx_v_protect);
  __Pyx_GOTREF(__pyx_v_self->protect);
  __Pyx_DECREF(__pyx_v_self->protect);
  __pyx_v_self->protect = __pyx_v_protect;

  /* "ProtectionData.pxi":74
 * 
 *         self.attributes_map = {
 *             'id': self.id,             # <<<<<<<<<<<<<<
 *             'id_class': self.id_class,
 *             'hash': self.hash,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_id, __pyx_v_self->id) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":75
 *         self.attributes_map = {
 *             'id': self.id,
 *             'id_class': self.id_class,             # <<<<<<<<<<<<<<
 *             'hash': self.hash,
 *             'isinstance': self.isinstance,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_id_class, __pyx_v_self->id_class) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":76
 *             'id': self.id,
 *             'id_class': self.id_class,
 *             'hash': self.hash,             # <<<<<<<<<<<<<<
 *             'isinstance': self.isinstance,
 *             'issubclass': self.issubclass,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_hash, __pyx_v_self->hash) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":77
 *             'id_class': self.id_class,
 *             'hash': self.hash,
 *             'isinstance': self.isinstance,             # <<<<<<<<<<<<<<
 *             'issubclass': self.issubclass,
 *             'instanceof': self.instanceof,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_isinstance, __pyx_v_self->isinstance) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":78
 *             'hash': self.hash,
 *             'isinstance': self.isinstance,
 *             'issubclass': self.issubclass,             # <<<<<<<<<<<<<<
 *             'instanceof': self.instanceof,
 *             'subclassof': self.subclassof,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_issubclass, __pyx_v_self->issubclass) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":79
 *             'isinstance': self.isinstance,
 *             'issubclass': self.issubclass,
 *             'instanceof': self.instanceof,             # <<<<<<<<<<<<<<
 *             'subclassof': self.subclassof,
 *             'help': self.help,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_instanceof, __pyx_v_self->instanceof) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":80
 *             'issubclass': self.issubclass,
 *             'instanceof': self.instanceof,
 *             'subclassof': self.subclassof,             # <<<<<<<<<<<<<<
 *             'help': self.help,
 *             'help_str': self.help_str,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_subclassof, __pyx_v_self->subclassof) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":81
 *             'instanceof': self.instanceof,
 *             'subclassof': self.subclassof,
 *             'help': self.help,             # <<<<<<<<<<<<<<
 *             'help_str': self.help_str,
 *             'testop': self.testop,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_help, __pyx_v_self->help) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":82
 *             'subclassof': self.subclassof,
 *             'help': self.help,
 *             'help_str': self.help_str,             # <<<<<<<<<<<<<<
 *             'testop': self.testop,
 *             'rules': self.rules,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_help_str, __pyx_v_self->help_str) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":83
 *             'help': self.help,
 *             'help_str': self.help_str,
 *             'testop': self.testop,             # <<<<<<<<<<<<<<
 *             'rules': self.rules,
 *             'freeze': self.freeze,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_testop, __pyx_v_self->testop) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":84
 *             'help_str': self.help_str,
 *             'testop': self.testop,
 *             'rules': self.rules,             # <<<<<<<<<<<<<<
 *             'freeze': self.freeze,
 *             'private': self.private,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_rules, __pyx_v_self->rules) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":85
 *             'testop': self.testop,
 *             'rules': self.rules,
 *             'freeze': self.freeze,             # <<<<<<<<<<<<<<
 *             'private': self.private,
 *             'protect': self.protect,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_freeze, __pyx_v_self->freeze) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":86
 *             'rules': self.rules,
 *             'freeze': self.freeze,
 *             'private': self.private,             # <<<<<<<<<<<<<<
 *             'protect': self.protect,
 *             'multiwrapped': self.multiwrapped,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_private, __pyx_v_self->private) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":87
 *             'freeze': self.freeze,
 *             'private': self.private,
 *             'protect': self.protect,             # <<<<<<<<<<<<<<
 *             'multiwrapped': self.multiwrapped,
 *             '__class__': __ProtectionData,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_protect, __pyx_v_self->protect) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":88
 *             'private': self.private,
 *             'protect': self.protect,
 *             'multiwrapped': self.multiwrapped,             # <<<<<<<<<<<<<<
 *             '__class__': __ProtectionData,
 *         }
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_multiwrapped, __pyx_v_self->multiwrapped) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":89
 *             'protect': self.protect,
 *             'multiwrapped': self.multiwrapped,
 *             '__class__': __ProtectionData,             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_class, ((PyObject *)__pyx_ptype_9pyprotect_9protected___ProtectionData)) < 0) __PYX_ERR(1, 74, __pyx_L1_error)

  /* "ProtectionData.pxi":73
 *         self.protect = protect
 * 
 *         self.attributes_map = {             # <<<<<<<<<<<<<<
 *             'id': self.id,
 *             'id_class': self.id_class,
 */
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->attributes_map);
  __Pyx_DECREF(__pyx_v_self->attributes_map);
  __pyx_v_self->attributes_map = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "ProtectionData.pxi":39
 *     cdef object attributes_map
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         id_val,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":92
 *         }
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         if a in self.attributes_map:
 *             return self.attributes_map.get(a)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattribute__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_2__getattribute__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_2__getattribute__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_v_missing_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattribute__", 0);

  /* "ProtectionData.pxi":93
 * 
 *     def __getattribute__(self, a):
 *         if a in self.attributes_map:             # <<<<<<<<<<<<<<
 *             return self.attributes_map.get(a)
 *         missing_msg = "Object '%s' has no attribute '%s'" % (
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_self->attributes_map, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 93, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "ProtectionData.pxi":94
 *     def __getattribute__(self, a):
 *         if a in self.attributes_map:
 *             return self.attributes_map.get(a)             # <<<<<<<<<<<<<<
 *         missing_msg = "Object '%s' has no attribute '%s'" % (
 *             '__ProtectionData',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->attributes_map, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_a};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "ProtectionData.pxi":93
 * 
 *     def __getattribute__(self, a):
 *         if a in self.attributes_map:             # <<<<<<<<<<<<<<
 *             return self.attributes_map.get(a)
 *         missing_msg = "Object '%s' has no attribute '%s'" % (
 */
  }

  /* "ProtectionData.pxi":97
 *         missing_msg = "Object '%s' has no attribute '%s'" % (
 *             '__ProtectionData',
 *             str(a)             # <<<<<<<<<<<<<<
 *         )
 *         raise AttributeError(missing_msg)
 */
  __pyx_t_2 = __Pyx_PyObject_Str(__pyx_v_a); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "ProtectionData.pxi":96
 *             return self.attributes_map.get(a)
 *         missing_msg = "Object '%s' has no attribute '%s'" % (
 *             '__ProtectionData',             # <<<<<<<<<<<<<<
 *             str(a)
 *         )
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_ProtectionData);
  __Pyx_GIVEREF(__pyx_n_s_ProtectionData);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_ProtectionData);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "ProtectionData.pxi":95
 *         if a in self.attributes_map:
 *             return self.attributes_map.get(a)
 *         missing_msg = "Object '%s' has no attribute '%s'" % (             # <<<<<<<<<<<<<<
 *             '__ProtectionData',
 *             str(a)
 */
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Object_s_has_no_attribute_s, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_missing_msg = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "ProtectionData.pxi":99
 *             str(a)
 *         )
 *         raise AttributeError(missing_msg)             # <<<<<<<<<<<<<<
 * 
 *     def __setattr__(self, a, val):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_v_missing_msg); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(1, 99, __pyx_L1_error)

  /* "ProtectionData.pxi":92
 *         }
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         if a in self.attributes_map:
 *             return self.attributes_map.get(a)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__getattribute__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_missing_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":101
 *         raise AttributeError(missing_msg)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Object is read-only')
 * 
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_4__setattr__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_a), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4__setattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_a, CYTHON_UNUSED PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setattr__", 0);

  /* "ProtectionData.pxi":102
 * 
 *     def __setattr__(self, a, val):
 *         raise ProtectionError('Object is read-only')             # <<<<<<<<<<<<<<
 * 
 *     def __delattr__(self, a):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_s_Object_is_read_only};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 102, __pyx_L1_error)

  /* "ProtectionData.pxi":101
 *         raise AttributeError(missing_msg)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Object is read-only')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__setattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":104
 *         raise ProtectionError('Object is read-only')
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Object is read-only')
 * 
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_6__delattr__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6__delattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delattr__", 0);

  /* "ProtectionData.pxi":105
 * 
 *     def __delattr__(self, a):
 *         raise ProtectionError('Object is read-only')             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_s_Object_is_read_only};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 105, __pyx_L1_error)

  /* "ProtectionData.pxi":104
 *         raise ProtectionError('Object is read-only')
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Object is read-only')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__delattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":107
 *         raise ProtectionError('Object is read-only')
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return list(self.attributes_map.keys())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_16__ProtectionData_9__dir__ = {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_16__ProtectionData_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__dir__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__dir__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_8__dir__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_8__dir__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "ProtectionData.pxi":108
 * 
 *     def __dir__(self):
 *         return list(self.attributes_map.keys())             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->attributes_map, __pyx_n_s_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "ProtectionData.pxi":107
 *         raise ProtectionError('Object is read-only')
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return list(self.attributes_map.keys())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":22
 *         multiwrapped: method (no args) -> bool
 *     '''
 *     cdef public object id             # <<<<<<<<<<<<<<
 *     cdef public object id_class
 *     cdef public object hash
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_2id___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_2id___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->id);
  __pyx_r = __pyx_v_self->id;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_2id_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_2id_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->id);
  __Pyx_DECREF(__pyx_v_self->id);
  __pyx_v_self->id = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_2id_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_2id_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->id);
  __Pyx_DECREF(__pyx_v_self->id);
  __pyx_v_self->id = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":23
 *     '''
 *     cdef public object id
 *     cdef public object id_class             # <<<<<<<<<<<<<<
 *     cdef public object hash
 *     cdef public object isinstance
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->id_class);
  __pyx_r = __pyx_v_self->id_class;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->id_class);
  __Pyx_DECREF(__pyx_v_self->id_class);
  __pyx_v_self->id_class = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8id_class_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->id_class);
  __Pyx_DECREF(__pyx_v_self->id_class);
  __pyx_v_self->id_class = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":24
 *     cdef public object id
 *     cdef public object id_class
 *     cdef public object hash             # <<<<<<<<<<<<<<
 *     cdef public object isinstance
 *     cdef public object issubclass
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->hash);
  __pyx_r = __pyx_v_self->hash;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->hash);
  __Pyx_DECREF(__pyx_v_self->hash);
  __pyx_v_self->hash = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4hash_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->hash);
  __Pyx_DECREF(__pyx_v_self->hash);
  __pyx_v_self->hash = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":25
 *     cdef public object id_class
 *     cdef public object hash
 *     cdef public object isinstance             # <<<<<<<<<<<<<<
 *     cdef public object issubclass
 *     cdef public object instanceof
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->isinstance);
  __pyx_r = __pyx_v_self->isinstance;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->isinstance);
  __Pyx_DECREF(__pyx_v_self->isinstance);
  __pyx_v_self->isinstance = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10isinstance_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->isinstance);
  __Pyx_DECREF(__pyx_v_self->isinstance);
  __pyx_v_self->isinstance = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":26
 *     cdef public object hash
 *     cdef public object isinstance
 *     cdef public object issubclass             # <<<<<<<<<<<<<<
 *     cdef public object instanceof
 *     cdef public object subclassof
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->issubclass);
  __pyx_r = __pyx_v_self->issubclass;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->issubclass);
  __Pyx_DECREF(__pyx_v_self->issubclass);
  __pyx_v_self->issubclass = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10issubclass_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->issubclass);
  __Pyx_DECREF(__pyx_v_self->issubclass);
  __pyx_v_self->issubclass = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":27
 *     cdef public object isinstance
 *     cdef public object issubclass
 *     cdef public object instanceof             # <<<<<<<<<<<<<<
 *     cdef public object subclassof
 *     cdef public object help
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->instanceof);
  __pyx_r = __pyx_v_self->instanceof;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->instanceof);
  __Pyx_DECREF(__pyx_v_self->instanceof);
  __pyx_v_self->instanceof = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10instanceof_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->instanceof);
  __Pyx_DECREF(__pyx_v_self->instanceof);
  __pyx_v_self->instanceof = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":28
 *     cdef public object issubclass
 *     cdef public object instanceof
 *     cdef public object subclassof             # <<<<<<<<<<<<<<
 *     cdef public object help
 *     cdef public object help_str
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->subclassof);
  __pyx_r = __pyx_v_self->subclassof;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->subclassof);
  __Pyx_DECREF(__pyx_v_self->subclassof);
  __pyx_v_self->subclassof = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_10subclassof_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->subclassof);
  __Pyx_DECREF(__pyx_v_self->subclassof);
  __pyx_v_self->subclassof = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":29
 *     cdef public object instanceof
 *     cdef public object subclassof
 *     cdef public object help             # <<<<<<<<<<<<<<
 *     cdef public object help_str
 *     cdef public object testop
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_4help___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_4help___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->help);
  __pyx_r = __pyx_v_self->help;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_4help_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4help_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->help);
  __Pyx_DECREF(__pyx_v_self->help);
  __pyx_v_self->help = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_4help_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_4help_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->help);
  __Pyx_DECREF(__pyx_v_self->help);
  __pyx_v_self->help = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":30
 *     cdef public object subclassof
 *     cdef public object help
 *     cdef public object help_str             # <<<<<<<<<<<<<<
 *     cdef public object testop
 *     cdef public object rules
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->help_str);
  __pyx_r = __pyx_v_self->help_str;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->help_str);
  __Pyx_DECREF(__pyx_v_self->help_str);
  __pyx_v_self->help_str = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_8help_str_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->help_str);
  __Pyx_DECREF(__pyx_v_self->help_str);
  __pyx_v_self->help_str = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":31
 *     cdef public object help
 *     cdef public object help_str
 *     cdef public object testop             # <<<<<<<<<<<<<<
 *     cdef public object rules
 *     cdef public object freeze
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->testop);
  __pyx_r = __pyx_v_self->testop;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->testop);
  __Pyx_DECREF(__pyx_v_self->testop);
  __pyx_v_self->testop = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6testop_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->testop);
  __Pyx_DECREF(__pyx_v_self->testop);
  __pyx_v_self->testop = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":32
 *     cdef public object help_str
 *     cdef public object testop
 *     cdef public object rules             # <<<<<<<<<<<<<<
 *     cdef public object freeze
 *     cdef public object private
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->rules);
  __pyx_r = __pyx_v_self->rules;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->rules);
  __Pyx_DECREF(__pyx_v_self->rules);
  __pyx_v_self->rules = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_5rules_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->rules);
  __Pyx_DECREF(__pyx_v_self->rules);
  __pyx_v_self->rules = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":33
 *     cdef public object testop
 *     cdef public object rules
 *     cdef public object freeze             # <<<<<<<<<<<<<<
 *     cdef public object private
 *     cdef public object protect
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->freeze);
  __pyx_r = __pyx_v_self->freeze;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->freeze);
  __Pyx_DECREF(__pyx_v_self->freeze);
  __pyx_v_self->freeze = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_6freeze_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->freeze);
  __Pyx_DECREF(__pyx_v_self->freeze);
  __pyx_v_self->freeze = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":34
 *     cdef public object rules
 *     cdef public object freeze
 *     cdef public object private             # <<<<<<<<<<<<<<
 *     cdef public object protect
 *     cdef public object multiwrapped
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_7private___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_7private___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->private);
  __pyx_r = __pyx_v_self->private;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_7private_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7private_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->private);
  __Pyx_DECREF(__pyx_v_self->private);
  __pyx_v_self->private = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_7private_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7private_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->private);
  __Pyx_DECREF(__pyx_v_self->private);
  __pyx_v_self->private = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":35
 *     cdef public object freeze
 *     cdef public object private
 *     cdef public object protect             # <<<<<<<<<<<<<<
 *     cdef public object multiwrapped
 *     cdef object attributes_map
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->protect);
  __pyx_r = __pyx_v_self->protect;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->protect);
  __Pyx_DECREF(__pyx_v_self->protect);
  __pyx_v_self->protect = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_7protect_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->protect);
  __Pyx_DECREF(__pyx_v_self->protect);
  __pyx_v_self->protect = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "ProtectionData.pxi":36
 *     cdef public object private
 *     cdef public object protect
 *     cdef public object multiwrapped             # <<<<<<<<<<<<<<
 *     cdef object attributes_map
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped___get__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped___get__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->multiwrapped);
  __pyx_r = __pyx_v_self->multiwrapped;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped_2__set__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped_2__set__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->multiwrapped);
  __Pyx_DECREF(__pyx_v_self->multiwrapped);
  __pyx_v_self->multiwrapped = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped_4__del__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_16__ProtectionData_12multiwrapped_4__del__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->multiwrapped);
  __Pyx_DECREF(__pyx_v_self->multiwrapped);
  __pyx_v_self->multiwrapped = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_16__ProtectionData_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_16__ProtectionData_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_10__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_10__reduce_cython__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.attributes_map, self.freeze, self.hash, self.help, self.help_str, self.id, self.id_class, self.instanceof, self.isinstance, self.issubclass, self.multiwrapped, self.private, self.protect, self.rules, self.subclassof, self.testop)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(16); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->attributes_map);
  __Pyx_GIVEREF(__pyx_v_self->attributes_map);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->attributes_map);
  __Pyx_INCREF(__pyx_v_self->freeze);
  __Pyx_GIVEREF(__pyx_v_self->freeze);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->freeze);
  __Pyx_INCREF(__pyx_v_self->hash);
  __Pyx_GIVEREF(__pyx_v_self->hash);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->hash);
  __Pyx_INCREF(__pyx_v_self->help);
  __Pyx_GIVEREF(__pyx_v_self->help);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->help);
  __Pyx_INCREF(__pyx_v_self->help_str);
  __Pyx_GIVEREF(__pyx_v_self->help_str);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->help_str);
  __Pyx_INCREF(__pyx_v_self->id);
  __Pyx_GIVEREF(__pyx_v_self->id);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_self->id);
  __Pyx_INCREF(__pyx_v_self->id_class);
  __Pyx_GIVEREF(__pyx_v_self->id_class);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_self->id_class);
  __Pyx_INCREF(__pyx_v_self->instanceof);
  __Pyx_GIVEREF(__pyx_v_self->instanceof);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_v_self->instanceof);
  __Pyx_INCREF(__pyx_v_self->isinstance);
  __Pyx_GIVEREF(__pyx_v_self->isinstance);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_v_self->isinstance);
  __Pyx_INCREF(__pyx_v_self->issubclass);
  __Pyx_GIVEREF(__pyx_v_self->issubclass);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_v_self->issubclass);
  __Pyx_INCREF(__pyx_v_self->multiwrapped);
  __Pyx_GIVEREF(__pyx_v_self->multiwrapped);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_v_self->multiwrapped);
  __Pyx_INCREF(__pyx_v_self->private);
  __Pyx_GIVEREF(__pyx_v_self->private);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_v_self->private);
  __Pyx_INCREF(__pyx_v_self->protect);
  __Pyx_GIVEREF(__pyx_v_self->protect);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_v_self->protect);
  __Pyx_INCREF(__pyx_v_self->rules);
  __Pyx_GIVEREF(__pyx_v_self->rules);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_v_self->rules);
  __Pyx_INCREF(__pyx_v_self->subclassof);
  __Pyx_GIVEREF(__pyx_v_self->subclassof);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_v_self->subclassof);
  __Pyx_INCREF(__pyx_v_self->testop);
  __Pyx_GIVEREF(__pyx_v_self->testop);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_v_self->testop);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.attributes_map, self.freeze, self.hash, self.help, self.help_str, self.id, self.id_class, self.instanceof, self.isinstance, self.issubclass, self.multiwrapped, self.private, self.protect, self.rules, self.subclassof, self.testop)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.attributes_map, self.freeze, self.hash, self.help, self.help_str, self.id, self.id_class, self.instanceof, self.isinstance, self.issubclass, self.multiwrapped, self.private, self.protect, self.rules, self.subclassof, self.testop)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.attributes_map is not None or self.freeze is not None or self.hash is not None or self.help is not None or self.help_str is not None or self.id is not None or self.id_class is not None or self.instanceof is not None or self.isinstance is not None or self.issubclass is not None or self.multiwrapped is not None or self.private is not None or self.protect is not None or self.rules is not None or self.subclassof is not None or self.testop is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.attributes_map, self.freeze, self.hash, self.help, self.help_str, self.id, self.id_class, self.instanceof, self.isinstance, self.issubclass, self.multiwrapped, self.private, self.protect, self.rules, self.subclassof, self.testop)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.attributes_map is not None or self.freeze is not None or self.hash is not None or self.help is not None or self.help_str is not None or self.id is not None or self.id_class is not None or self.instanceof is not None or self.isinstance is not None or self.issubclass is not None or self.multiwrapped is not None or self.private is not None or self.protect is not None or self.rules is not None or self.subclassof is not None or self.testop is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->attributes_map != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->freeze != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->hash != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->help != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->help_str != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->id != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->id_class != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->instanceof != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->isinstance != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->issubclass != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->multiwrapped != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->private != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->protect != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->rules != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->subclassof != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->testop != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.attributes_map is not None or self.freeze is not None or self.hash is not None or self.help is not None or self.help_str is not None or self.id is not None or self.id_class is not None or self.instanceof is not None or self.isinstance is not None or self.issubclass is not None or self.multiwrapped is not None or self.private is not None or self.protect is not None or self.rules is not None or self.subclassof is not None or self.testop is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.attributes_map is not None or self.freeze is not None or self.hash is not None or self.help is not None or self.help_str is not None or self.id is not None or self.id_class is not None or self.instanceof is not None or self.isinstance is not None or self.issubclass is not None or self.multiwrapped is not None or self.private is not None or self.protect is not None or self.rules is not None or self.subclassof is not None or self.testop is not None
 *     if use_setstate:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle___ProtectionData); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_262487005);
    __Pyx_GIVEREF(__pyx_int_262487005);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_262487005);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.attributes_map is not None or self.freeze is not None or self.hash is not None or self.help is not None or self.help_str is not None or self.id is not None or self.id_class is not None or self.instanceof is not None or self.isinstance is not None or self.issubclass is not None or self.multiwrapped is not None or self.private is not None or self.protect is not None or self.rules is not None or self.subclassof is not None or self.testop is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, None), state
 *     else:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___ProtectionData__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle___ProtectionData); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_262487005);
    __Pyx_GIVEREF(__pyx_int_262487005);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_262487005);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___ProtectionData__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_16__ProtectionData_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_16__ProtectionData_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_16__ProtectionData_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_16__ProtectionData_12__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_16__ProtectionData_12__setstate_cython__(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___ProtectionData__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle___ProtectionData__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___ProtectionData__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.__ProtectionData.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":11
 *     cdef bint frozen
 * 
 *     def __init__(self, o, frozen=False):             # <<<<<<<<<<<<<<
 *         '''
 *         o: object to be wrapped
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_5Proxy_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_5Proxy___init__, "\n        o: object to be wrapped\n        frozen: bool: If True, no attribute can be modified\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_5Proxy___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_5Proxy_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_frozen = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,0};
    #endif
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 11, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_frozen);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 11, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(2, 11, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_o = values[0];
    __pyx_v_frozen = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, __pyx_nargs); __PYX_ERR(2, 11, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy___init__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_o, __pyx_v_frozen);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_5Proxy___init__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Proxy.pxi":16
 *         frozen: bool: If True, no attribute can be modified
 *         '''
 *         self.pvt_o = o             # <<<<<<<<<<<<<<
 *         self.frozen = bool(frozen)
 * 
 */
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_v_o;

  /* "Proxy.pxi":17
 *         '''
 *         self.pvt_o = o
 *         self.frozen = bool(frozen)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_frozen); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_v_self->frozen = (!(!__pyx_t_1));

  /* "Proxy.pxi":11
 *     cdef bint frozen
 * 
 *     def __init__(self, o, frozen=False):             # <<<<<<<<<<<<<<
 *         '''
 *         o: object to be wrapped
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":25
 *     # __repr__, __str__ and __bytes__:
 *     # We do not want the default cython implementations of Wrapped
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.pvt_o)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_2__repr__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_2__repr__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Proxy.pxi":26
 *     # We do not want the default cython implementations of Wrapped
 *     def __repr__(self):
 *         return repr(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":25
 *     # __repr__, __str__ and __bytes__:
 *     # We do not want the default cython implementations of Wrapped
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":28
 *         return repr(self.pvt_o)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(self.pvt_o)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_5__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_5__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_4__str__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_4__str__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Proxy.pxi":29
 * 
 *     def __str__(self):
 *         return str(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     def __bytes__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_self->pvt_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":28
 *         return repr(self.pvt_o)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":31
 *         return str(self.pvt_o)
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(self.pvt_o)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_7__bytes__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_7__bytes__ = {"__bytes__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_7__bytes__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_7__bytes__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bytes__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__bytes__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__bytes__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_6__bytes__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_6__bytes__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__bytes__", 0);

  /* "Proxy.pxi":32
 * 
 *     def __bytes__(self):
 *         return bytes(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_v_self->pvt_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":31
 *         return str(self.pvt_o)
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__bytes__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":38
 *     # --------------------------------------------------------------------
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if iswrapped(self.pvt_o):
 *             # We claim to be avoiding double-wrapping, so this code
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_9__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_9__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__call__", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_8__call__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_8__call__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "Proxy.pxi":39
 * 
 *     def __call__(self, *args, **kwargs):
 *         if iswrapped(self.pvt_o):             # <<<<<<<<<<<<<<
 *             # We claim to be avoiding double-wrapping, so this code
 *             # should never be reached
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_self->pvt_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(2, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "Proxy.pxi":42
 *             # We claim to be avoiding double-wrapping, so this code
 *             # should never be reached
 *             raise RuntimeError('Double-wrapped!')             # <<<<<<<<<<<<<<
 * 
 *         x = self.pvt_o(*args, **kwargs)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 42, __pyx_L1_error)

    /* "Proxy.pxi":39
 * 
 *     def __call__(self, *args, **kwargs):
 *         if iswrapped(self.pvt_o):             # <<<<<<<<<<<<<<
 *             # We claim to be avoiding double-wrapping, so this code
 *             # should never be reached
 */
  }

  /* "Proxy.pxi":44
 *             raise RuntimeError('Double-wrapped!')
 * 
 *         x = self.pvt_o(*args, **kwargs)             # <<<<<<<<<<<<<<
 *         if self.frozen and not isimmutable(x):
 *             x = Frozen(x)
 */
  __pyx_t_1 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_self->pvt_o, __pyx_v_args, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_x = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Proxy.pxi":45
 * 
 *         x = self.pvt_o(*args, **kwargs)
 *         if self.frozen and not isimmutable(x):             # <<<<<<<<<<<<<<
 *             x = Frozen(x)
 *         return x
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_5 = __pyx_v_self->frozen;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_isimmutable); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_x};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(2, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (!__pyx_t_6);
  __pyx_t_5 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Proxy.pxi":46
 *         x = self.pvt_o(*args, **kwargs)
 *         if self.frozen and not isimmutable(x):
 *             x = Frozen(x)             # <<<<<<<<<<<<<<
 *         return x
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9pyprotect_9protected_Frozen), __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_x, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Proxy.pxi":45
 * 
 *         x = self.pvt_o(*args, **kwargs)
 *         if self.frozen and not isimmutable(x):             # <<<<<<<<<<<<<<
 *             x = Frozen(x)
 *         return x
 */
  }

  /* "Proxy.pxi":47
 *         if self.frozen and not isimmutable(x):
 *             x = Frozen(x)
 *         return x             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_x);
  __pyx_r = __pyx_v_x;
  goto __pyx_L0;

  /* "Proxy.pxi":38
 *     # --------------------------------------------------------------------
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if iswrapped(self.pvt_o):
 *             # We claim to be avoiding double-wrapping, so this code
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9pyprotect_9protected_5Proxy_12generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Proxy.pxi":49
 *         return x
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         for x in iter(self.pvt_o):
 *             if self.frozen:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_11__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_10__iter__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_10__iter__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct____iter__(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(2, 49, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9pyprotect_9protected_5Proxy_12generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_Proxy___iter, __pyx_n_s_pyprotect_protected); if (unlikely(!gen)) __PYX_ERR(2, 49, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9pyprotect_9protected_5Proxy_12generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 49, __pyx_L1_error)

  /* "Proxy.pxi":50
 * 
 *     def __iter__(self):
 *         for x in iter(self.pvt_o):             # <<<<<<<<<<<<<<
 *             if self.frozen:
 *                 x = freeze(x)
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 50, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(2, 50, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 50, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(2, 50, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 50, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 50, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_x);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_x, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Proxy.pxi":51
 *     def __iter__(self):
 *         for x in iter(self.pvt_o):
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 x = freeze(x)
 *             yield x
 */
    if (__pyx_cur_scope->__pyx_v_self->frozen) {

      /* "Proxy.pxi":52
 *         for x in iter(self.pvt_o):
 *             if self.frozen:
 *                 x = freeze(x)             # <<<<<<<<<<<<<<
 *             yield x
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_freeze); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_x};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 52, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_x);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_x, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "Proxy.pxi":51
 *     def __iter__(self):
 *         for x in iter(self.pvt_o):
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 x = freeze(x)
 *             yield x
 */
    }

    /* "Proxy.pxi":53
 *             if self.frozen:
 *                 x = freeze(x)
 *             yield x             # <<<<<<<<<<<<<<
 * 
 *     # Representations - return immutable
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_x);
    __pyx_r = __pyx_cur_scope->__pyx_v_x;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 53, __pyx_L1_error)

    /* "Proxy.pxi":50
 * 
 *     def __iter__(self):
 *         for x in iter(self.pvt_o):             # <<<<<<<<<<<<<<
 *             if self.frozen:
 *                 x = freeze(x)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "Proxy.pxi":49
 *         return x
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         for x in iter(self.pvt_o):
 *             if self.frozen:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":56
 * 
 *     # Representations - return immutable
 *     def __format__(self, val):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__format__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_14__format__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_14__format__ = {"__format__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_14__format__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_14__format__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_val = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__format__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_val,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_val,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_val)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 56, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__format__") < 0)) __PYX_ERR(2, 56, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_val = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__format__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 56, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__format__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_13__format__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_val);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_13__format__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__format__", 0);

  /* "Proxy.pxi":57
 *     # Representations - return immutable
 *     def __format__(self, val):
 *         return self.pvt_o.__format__(val)             # <<<<<<<<<<<<<<
 * 
 *     # Truth value testing - non-mutating, returning immutable bool
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":56
 * 
 *     # Representations - return immutable
 *     def __format__(self, val):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__format__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__format__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":60
 * 
 *     # Truth value testing - non-mutating, returning immutable bool
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return bool(self.pvt_o)
 * 
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_5Proxy_16__bool__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9pyprotect_9protected_5Proxy_16__bool__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bool__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_15__bool__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_5Proxy_15__bool__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__bool__", 0);

  /* "Proxy.pxi":61
 *     # Truth value testing - non-mutating, returning immutable bool
 *     def __bool__(self):
 *         return bool(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     # Emulating container types - non-mutating
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->pvt_o); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 61, __pyx_L1_error)
  __pyx_r = (!(!__pyx_t_1));
  goto __pyx_L0;

  /* "Proxy.pxi":60
 * 
 *     # Truth value testing - non-mutating, returning immutable bool
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return bool(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__bool__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":64
 * 
 *     # Emulating container types - non-mutating
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         x = self.pvt_o.__getitem__(key)
 *         if self.frozen:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_18__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_18__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_17__getitem__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_17__getitem__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "Proxy.pxi":65
 *     # Emulating container types - non-mutating
 *     def __getitem__(self, key):
 *         x = self.pvt_o.__getitem__(key)             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             x = freeze(x)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_getitem); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_x = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":66
 *     def __getitem__(self, key):
 *         x = self.pvt_o.__getitem__(key)
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             x = freeze(x)
 *         return x
 */
  if (__pyx_v_self->frozen) {

    /* "Proxy.pxi":67
 *         x = self.pvt_o.__getitem__(key)
 *         if self.frozen:
 *             x = freeze(x)             # <<<<<<<<<<<<<<
 *         return x
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_freeze); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_x};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Proxy.pxi":66
 *     def __getitem__(self, key):
 *         x = self.pvt_o.__getitem__(key)
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             x = freeze(x)
 *         return x
 */
  }

  /* "Proxy.pxi":68
 *         if self.frozen:
 *             x = freeze(x)
 *         return x             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_x);
  __pyx_r = __pyx_v_x;
  goto __pyx_L0;

  /* "Proxy.pxi":64
 * 
 *     # Emulating container types - non-mutating
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         x = self.pvt_o.__getitem__(key)
 *         if self.frozen:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":70
 *         return x
 * 
 *     def __contains__(self, val):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__contains__(val)
 * 
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_5Proxy_20__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_5Proxy_20__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_19__contains__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_5Proxy_19__contains__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "Proxy.pxi":71
 * 
 *     def __contains__(self, val):
 *         return self.pvt_o.__contains__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  goto __pyx_L0;

  /* "Proxy.pxi":70
 *         return x
 * 
 *     def __contains__(self, val):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__contains__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":73
 *         return self.pvt_o.__contains__(val)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__len__()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_9pyprotect_9protected_5Proxy_22__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_9pyprotect_9protected_5Proxy_22__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_21__len__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_9pyprotect_9protected_5Proxy_21__len__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "Proxy.pxi":74
 * 
 *     def __len__(self):
 *         return self.pvt_o.__len__()             # <<<<<<<<<<<<<<
 * 
 *     def __length_hint__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "Proxy.pxi":73
 *         return self.pvt_o.__contains__(val)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__len__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":76
 *         return self.pvt_o.__len__()
 * 
 *     def __length_hint__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__length_hint__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_24__length_hint__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_24__length_hint__ = {"__length_hint__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_24__length_hint__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_24__length_hint__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__length_hint__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__length_hint__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__length_hint__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_23__length_hint__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_23__length_hint__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__length_hint__", 0);

  /* "Proxy.pxi":77
 * 
 *     def __length_hint__(self):
 *         return self.pvt_o.__length_hint__()             # <<<<<<<<<<<<<<
 * 
 *     # Unary numeric operations - return immutable
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_length_hint); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":76
 *         return self.pvt_o.__len__()
 * 
 *     def __length_hint__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__length_hint__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__length_hint__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":80
 * 
 *     # Unary numeric operations - return immutable
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__neg__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_26__neg__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_26__neg__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neg__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_25__neg__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_25__neg__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__neg__", 0);

  /* "Proxy.pxi":81
 *     # Unary numeric operations - return immutable
 *     def __neg__(self):
 *         return self.pvt_o.__neg__()             # <<<<<<<<<<<<<<
 * 
 *     def __pos__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_neg); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":80
 * 
 *     # Unary numeric operations - return immutable
 *     def __neg__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__neg__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__neg__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":83
 *         return self.pvt_o.__neg__()
 * 
 *     def __pos__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__pos__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_28__pos__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_28__pos__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pos__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_27__pos__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_27__pos__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pos__", 0);

  /* "Proxy.pxi":84
 * 
 *     def __pos__(self):
 *         return self.pvt_o.__pos__()             # <<<<<<<<<<<<<<
 * 
 *     def __abs__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":83
 *         return self.pvt_o.__neg__()
 * 
 *     def __pos__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__pos__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__pos__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":86
 *         return self.pvt_o.__pos__()
 * 
 *     def __abs__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__abs__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_30__abs__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_30__abs__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__abs__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_29__abs__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_29__abs__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__abs__", 0);

  /* "Proxy.pxi":87
 * 
 *     def __abs__(self):
 *         return self.pvt_o.__abs__()             # <<<<<<<<<<<<<<
 * 
 *     def __invert__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_abs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":86
 *         return self.pvt_o.__pos__()
 * 
 *     def __abs__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__abs__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__abs__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":89
 *         return self.pvt_o.__abs__()
 * 
 *     def __invert__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__invert__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_32__invert__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_32__invert__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__invert__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_31__invert__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_31__invert__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__invert__", 0);

  /* "Proxy.pxi":90
 * 
 *     def __invert__(self):
 *         return self.pvt_o.__invert__()             # <<<<<<<<<<<<<<
 * 
 *     def __complex__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_invert); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":89
 *         return self.pvt_o.__abs__()
 * 
 *     def __invert__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__invert__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__invert__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":92
 *         return self.pvt_o.__invert__()
 * 
 *     def __complex__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__complex__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_34__complex__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_34__complex__ = {"__complex__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_34__complex__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_34__complex__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__complex__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__complex__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__complex__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_33__complex__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_33__complex__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__complex__", 0);

  /* "Proxy.pxi":93
 * 
 *     def __complex__(self):
 *         return self.pvt_o.__complex__()             # <<<<<<<<<<<<<<
 * 
 *     def __int__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_complex_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":92
 *         return self.pvt_o.__invert__()
 * 
 *     def __complex__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__complex__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__complex__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":95
 *         return self.pvt_o.__complex__()
 * 
 *     def __int__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__int__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_36__int__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_36__int__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__int__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_35__int__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_35__int__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__int__", 0);

  /* "Proxy.pxi":96
 * 
 *     def __int__(self):
 *         return self.pvt_o.__int__()             # <<<<<<<<<<<<<<
 * 
 *     def __float__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":95
 *         return self.pvt_o.__complex__()
 * 
 *     def __int__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__int__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__int__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":98
 *         return self.pvt_o.__int__()
 * 
 *     def __float__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__float__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_38__float__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_38__float__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__float__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_37__float__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_37__float__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__float__", 0);

  /* "Proxy.pxi":99
 * 
 *     def __float__(self):
 *         return self.pvt_o.__float__()             # <<<<<<<<<<<<<<
 * 
 *     def __index__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_float_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":98
 *         return self.pvt_o.__int__()
 * 
 *     def __float__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__float__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__float__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":101
 *         return self.pvt_o.__float__()
 * 
 *     def __index__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__index__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_40__index__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_40__index__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__index__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_39__index__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_39__index__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__index__", 0);

  /* "Proxy.pxi":102
 * 
 *     def __index__(self):
 *         return self.pvt_o.__index__()             # <<<<<<<<<<<<<<
 * 
 *     def __round__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":101
 *         return self.pvt_o.__float__()
 * 
 *     def __index__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__index__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__index__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":104
 *         return self.pvt_o.__index__()
 * 
 *     def __round__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__round__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_42__round__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_42__round__ = {"__round__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_42__round__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_42__round__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__round__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__round__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__round__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_41__round__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_41__round__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__round__", 0);

  /* "Proxy.pxi":105
 * 
 *     def __round__(self):
 *         return self.pvt_o.__round__()             # <<<<<<<<<<<<<<
 * 
 *     def __trunc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_round); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":104
 *         return self.pvt_o.__index__()
 * 
 *     def __round__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__round__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__round__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":107
 *         return self.pvt_o.__round__()
 * 
 *     def __trunc__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__trunc__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_44__trunc__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_44__trunc__ = {"__trunc__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_44__trunc__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_44__trunc__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__trunc__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__trunc__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__trunc__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_43__trunc__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_43__trunc__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__trunc__", 0);

  /* "Proxy.pxi":108
 * 
 *     def __trunc__(self):
 *         return self.pvt_o.__trunc__()             # <<<<<<<<<<<<<<
 * 
 *     def __floor__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_trunc); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":107
 *         return self.pvt_o.__round__()
 * 
 *     def __trunc__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__trunc__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__trunc__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":110
 *         return self.pvt_o.__trunc__()
 * 
 *     def __floor__(self):             # <<<<<<<<<<<<<<
 *         return math.floor(self.pvt_o)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_46__floor__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_46__floor__ = {"__floor__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_46__floor__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_46__floor__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__floor__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__floor__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__floor__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_45__floor__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_45__floor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__floor__", 0);

  /* "Proxy.pxi":111
 * 
 *     def __floor__(self):
 *         return math.floor(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     def __ceil__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_math); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_floor); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->pvt_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":110
 *         return self.pvt_o.__trunc__()
 * 
 *     def __floor__(self):             # <<<<<<<<<<<<<<
 *         return math.floor(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__floor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":113
 *         return math.floor(self.pvt_o)
 * 
 *     def __ceil__(self):             # <<<<<<<<<<<<<<
 *         return math.ceil(self.pvt_o)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_48__ceil__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_48__ceil__ = {"__ceil__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_48__ceil__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_48__ceil__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ceil__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__ceil__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__ceil__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_47__ceil__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_47__ceil__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ceil__", 0);

  /* "Proxy.pxi":114
 * 
 *     def __ceil__(self):
 *         return math.ceil(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     # The numeric operations below - including the reversed versions
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_math); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ceil); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->pvt_o};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":113
 *         return math.floor(self.pvt_o)
 * 
 *     def __ceil__(self):             # <<<<<<<<<<<<<<
 *         return math.ceil(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__ceil__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":121
 * 
 *     # Numeric types - return immutable
 *     def __add__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__radd__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_50__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_50__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_49__add__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_49__add__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 0);

  /* "Proxy.pxi":122
 *     # Numeric types - return immutable
 *     def __add__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__radd__', None)):
 *                 return val.__radd__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":123
 *     def __add__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__radd__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__radd__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_radd, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 123, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":124
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__radd__', None)):
 *                 return val.__radd__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_radd); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 124, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":123
 *     def __add__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__radd__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__radd__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":126
 *                 return val.__radd__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__add__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":122
 *     # Numeric types - return immutable
 *     def __add__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__radd__', None)):
 *                 return val.__radd__(self)
 */
  }

  /* "Proxy.pxi":127
 *             else:
 *                 return NotImplemented
 *         return self.__add__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __mul__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_add); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":121
 * 
 *     # Numeric types - return immutable
 *     def __add__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__radd__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":129
 *         return self.__add__(val)
 * 
 *     def __mul__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmul__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_52__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_52__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_51__mul__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_51__mul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__mul__", 0);

  /* "Proxy.pxi":130
 * 
 *     def __mul__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rmul__', None)):
 *                 return val.__rmul__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":131
 *     def __mul__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmul__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rmul__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rmul, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 131, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":132
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmul__', None)):
 *                 return val.__rmul__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rmul); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":131
 *     def __mul__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmul__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rmul__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":134
 *                 return val.__rmul__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__mul__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":130
 * 
 *     def __mul__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rmul__', None)):
 *                 return val.__rmul__(self)
 */
  }

  /* "Proxy.pxi":135
 *             else:
 *                 return NotImplemented
 *         return self.__mul__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mul); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":129
 *         return self.__add__(val)
 * 
 *     def __mul__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmul__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":137
 *         return self.__mul__(val)
 * 
 *     def __sub__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rsub__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_54__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_54__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_53__sub__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_53__sub__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sub__", 0);

  /* "Proxy.pxi":138
 * 
 *     def __sub__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rsub__', None)):
 *                 return val.__rsub__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":139
 *     def __sub__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rsub__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rsub__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rsub, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":140
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rsub__', None)):
 *                 return val.__rsub__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rsub); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":139
 *     def __sub__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rsub__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rsub__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":142
 *                 return val.__rsub__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__sub__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":138
 * 
 *     def __sub__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rsub__', None)):
 *                 return val.__rsub__(self)
 */
  }

  /* "Proxy.pxi":143
 *             else:
 *                 return NotImplemented
 *         return self.__sub__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __matmul__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sub); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":137
 *         return self.__mul__(val)
 * 
 *     def __sub__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rsub__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":145
 *         return self.__sub__(val)
 * 
 *     def __matmul__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmatmul__', None)):
 */

/* Python wrapper */
#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_56__matmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_56__matmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__matmul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_55__matmul__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_VERSION_HEX >= 0x03050000)*/

#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_55__matmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__matmul__", 0);

  /* "Proxy.pxi":146
 * 
 *     def __matmul__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rmatmul__', None)):
 *                 return val.__rmatmul__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":147
 *     def __matmul__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmatmul__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rmatmul__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rmatmul, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 147, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":148
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmatmul__', None)):
 *                 return val.__rmatmul__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rmatmul); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":147
 *     def __matmul__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmatmul__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rmatmul__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":150
 *                 return val.__rmatmul__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__matmul__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":146
 * 
 *     def __matmul__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rmatmul__', None)):
 *                 return val.__rmatmul__(self)
 */
  }

  /* "Proxy.pxi":151
 *             else:
 *                 return NotImplemented
 *         return self.__matmul__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __truediv__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matmul); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":145
 *         return self.__sub__(val)
 * 
 *     def __matmul__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmatmul__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__matmul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_VERSION_HEX >= 0x03050000)*/

/* "Proxy.pxi":153
 *         return self.__matmul__(val)
 * 
 *     def __truediv__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_58__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_58__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__truediv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_57__truediv__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_57__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__truediv__", 0);

  /* "Proxy.pxi":154
 * 
 *     def __truediv__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rtruediv__', None)):
 *                 return val.__rtruediv__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":155
 *     def __truediv__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rtruediv__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rtruediv, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 155, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":156
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):
 *                 return val.__rtruediv__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rtruediv); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":155
 *     def __truediv__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rtruediv__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":158
 *                 return val.__rtruediv__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__truediv__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":154
 * 
 *     def __truediv__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rtruediv__', None)):
 *                 return val.__rtruediv__(self)
 */
  }

  /* "Proxy.pxi":159
 *             else:
 *                 return NotImplemented
 *         return self.__truediv__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __floordiv__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_truediv); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":153
 *         return self.__matmul__(val)
 * 
 *     def __truediv__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__truediv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":161
 *         return self.__truediv__(val)
 * 
 *     def __floordiv__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_60__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_60__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__floordiv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_59__floordiv__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_59__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__floordiv__", 0);

  /* "Proxy.pxi":162
 * 
 *     def __floordiv__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rtruediv__', None)):
 *                 return val.__rtruediv__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":163
 *     def __floordiv__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rtruediv__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rtruediv, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 163, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":164
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):
 *                 return val.__rtruediv__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rtruediv); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":163
 *     def __floordiv__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rtruediv__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":166
 *                 return val.__rtruediv__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__floordiv__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":162
 * 
 *     def __floordiv__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rtruediv__', None)):
 *                 return val.__rtruediv__(self)
 */
  }

  /* "Proxy.pxi":167
 *             else:
 *                 return NotImplemented
 *         return self.__floordiv__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __mod__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_floordiv); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":161
 *         return self.__truediv__(val)
 * 
 *     def __floordiv__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rtruediv__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__floordiv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":169
 *         return self.__floordiv__(val)
 * 
 *     def __mod__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmod__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_62__mod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_62__mod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mod__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_61__mod__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_61__mod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__mod__", 0);

  /* "Proxy.pxi":170
 * 
 *     def __mod__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rmod__', None)):
 *                 return val.__rmod__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":171
 *     def __mod__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmod__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rmod__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rmod, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 171, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":172
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmod__', None)):
 *                 return val.__rmod__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rmod); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":171
 *     def __mod__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmod__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rmod__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":174
 *                 return val.__rmod__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__mod__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":170
 * 
 *     def __mod__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rmod__', None)):
 *                 return val.__rmod__(self)
 */
  }

  /* "Proxy.pxi":175
 *             else:
 *                 return NotImplemented
 *         return self.__mod__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __divmod__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mod); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":169
 *         return self.__floordiv__(val)
 * 
 *     def __mod__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rmod__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__mod__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":177
 *         return self.__mod__(val)
 * 
 *     def __divmod__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rdivmod__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_64__divmod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_64__divmod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__divmod__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_63__divmod__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_63__divmod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__divmod__", 0);

  /* "Proxy.pxi":178
 * 
 *     def __divmod__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rdivmod__', None)):
 *                 return val.__rdivmod__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":179
 *     def __divmod__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rdivmod__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rdivmod__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rdivmod, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 179, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":180
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rdivmod__', None)):
 *                 return val.__rdivmod__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rdivmod); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":179
 *     def __divmod__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rdivmod__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rdivmod__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":182
 *                 return val.__rdivmod__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__divmod__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":178
 * 
 *     def __divmod__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rdivmod__', None)):
 *                 return val.__rdivmod__(self)
 */
  }

  /* "Proxy.pxi":183
 *             else:
 *                 return NotImplemented
 *         return self.__divmod__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __pow__(self, val, mod):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_divmod); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":177
 *         return self.__mod__(val)
 * 
 *     def __divmod__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rdivmod__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__divmod__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":185
 *         return self.__divmod__(val)
 * 
 *     def __pow__(self, val, mod):             # <<<<<<<<<<<<<<
 *         return self.__pow__(val, mod)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_66__pow__(PyObject *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_66__pow__(PyObject *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pow__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_65__pow__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val), ((PyObject *)__pyx_v_mod));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_65__pow__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pow__", 0);

  /* "Proxy.pxi":186
 * 
 *     def __pow__(self, val, mod):
 *         return self.__pow__(val, mod)             # <<<<<<<<<<<<<<
 * 
 *     def __lshift__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pow); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_val, __pyx_v_mod};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":185
 *         return self.__divmod__(val)
 * 
 *     def __pow__(self, val, mod):             # <<<<<<<<<<<<<<
 *         return self.__pow__(val, mod)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__pow__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":188
 *         return self.__pow__(val, mod)
 * 
 *     def __lshift__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rlshift__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_68__lshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_68__lshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__lshift__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_67__lshift__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_67__lshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__lshift__", 0);

  /* "Proxy.pxi":189
 * 
 *     def __lshift__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rlshift__', None)):
 *                 return val.__rlshift__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":190
 *     def __lshift__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rlshift__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rlshift__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rlshift, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 190, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":191
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rlshift__', None)):
 *                 return val.__rlshift__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rlshift); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":190
 *     def __lshift__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rlshift__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rlshift__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":193
 *                 return val.__rlshift__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__lshift__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":189
 * 
 *     def __lshift__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rlshift__', None)):
 *                 return val.__rlshift__(self)
 */
  }

  /* "Proxy.pxi":194
 *             else:
 *                 return NotImplemented
 *         return self.__lshift__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rshift__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lshift); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":188
 *         return self.__pow__(val, mod)
 * 
 *     def __lshift__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rlshift__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__lshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":196
 *         return self.__lshift__(val)
 * 
 *     def __rshift__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rrshift__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_70__rshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_70__rshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rshift__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_69__rshift__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_69__rshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rshift__", 0);

  /* "Proxy.pxi":197
 * 
 *     def __rshift__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rrshift__', None)):
 *                 return val.__rrshift__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":198
 *     def __rshift__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rrshift__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rrshift__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rrshift, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":199
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rrshift__', None)):
 *                 return val.__rrshift__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rrshift); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":198
 *     def __rshift__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rrshift__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rrshift__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":201
 *                 return val.__rrshift__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__rshift__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":197
 * 
 *     def __rshift__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rrshift__', None)):
 *                 return val.__rrshift__(self)
 */
  }

  /* "Proxy.pxi":202
 *             else:
 *                 return NotImplemented
 *         return self.__rshift__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __and__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rshift); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":196
 *         return self.__lshift__(val)
 * 
 *     def __rshift__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rrshift__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":204
 *         return self.__rshift__(val)
 * 
 *     def __and__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rand__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_72__and__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_72__and__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__and__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_71__and__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_71__and__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__and__", 0);

  /* "Proxy.pxi":205
 * 
 *     def __and__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rand__', None)):
 *                 return val.__rand__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":206
 *     def __and__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rand__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rand__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rand, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 206, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":207
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rand__', None)):
 *                 return val.__rand__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rand); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 207, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":206
 *     def __and__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rand__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rand__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":209
 *                 return val.__rand__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__and__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":205
 * 
 *     def __and__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rand__', None)):
 *                 return val.__rand__(self)
 */
  }

  /* "Proxy.pxi":210
 *             else:
 *                 return NotImplemented
 *         return self.__and__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __or__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_and); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":204
 *         return self.__rshift__(val)
 * 
 *     def __and__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rand__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__and__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":212
 *         return self.__and__(val)
 * 
 *     def __or__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__ror__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_74__or__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_74__or__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__or__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_73__or__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_73__or__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__or__", 0);

  /* "Proxy.pxi":213
 * 
 *     def __or__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__ror__', None)):
 *                 return val.__ror__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":214
 *     def __or__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__ror__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__ror__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_ror, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":215
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__ror__', None)):
 *                 return val.__ror__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_ror); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 215, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":214
 *     def __or__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__ror__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__ror__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":217
 *                 return val.__ror__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__or__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":213
 * 
 *     def __or__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__ror__', None)):
 *                 return val.__ror__(self)
 */
  }

  /* "Proxy.pxi":218
 *             else:
 *                 return NotImplemented
 *         return self.__or__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __xor__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_or); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":212
 *         return self.__and__(val)
 * 
 *     def __or__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__ror__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__or__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":220
 *         return self.__or__(val)
 * 
 *     def __xor__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rxor__', None)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_76__xor__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_76__xor__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__xor__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_75__xor__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_75__xor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__xor__", 0);

  /* "Proxy.pxi":221
 * 
 *     def __xor__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rxor__', None)):
 *                 return val.__rxor__(self)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "Proxy.pxi":222
 *     def __xor__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rxor__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rxor__(self)
 *             else:
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_s_rxor, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 222, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Proxy.pxi":223
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rxor__', None)):
 *                 return val.__rxor__(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_rxor); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 223, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Proxy.pxi":222
 *     def __xor__(self, val):
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rxor__', None)):             # <<<<<<<<<<<<<<
 *                 return val.__rxor__(self)
 *             else:
 */
    }

    /* "Proxy.pxi":225
 *                 return val.__rxor__(self)
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 *         return self.__xor__(val)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_builtin_NotImplemented);
      __pyx_r = __pyx_builtin_NotImplemented;
      goto __pyx_L0;
    }

    /* "Proxy.pxi":221
 * 
 *     def __xor__(self, val):
 *         if not isinstance(self, Wrapped):             # <<<<<<<<<<<<<<
 *             if callable(getattr(val, '__rxor__', None)):
 *                 return val.__rxor__(self)
 */
  }

  /* "Proxy.pxi":226
 *             else:
 *                 return NotImplemented
 *         return self.__xor__(val)             # <<<<<<<<<<<<<<
 * 
 *     # Numeric types - reversed - - return immutable
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_xor); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_val};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":220
 *         return self.__or__(val)
 * 
 *     def __xor__(self, val):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, Wrapped):
 *             if callable(getattr(val, '__rxor__', None)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__xor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":229
 * 
 *     # Numeric types - reversed - - return immutable
 *     def __radd__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__radd__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_78__radd__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_78__radd__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__radd__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_77__radd__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_77__radd__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__radd__", 0);

  /* "Proxy.pxi":230
 *     # Numeric types - reversed - - return immutable
 *     def __radd__(self, val):
 *         return self.__radd__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rmul__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_radd); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":229
 * 
 *     # Numeric types - reversed - - return immutable
 *     def __radd__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__radd__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__radd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":232
 *         return self.__radd__(val)
 * 
 *     def __rmul__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rmul__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_80__rmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_80__rmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rmul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_79__rmul__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_79__rmul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rmul__", 0);

  /* "Proxy.pxi":233
 * 
 *     def __rmul__(self, val):
 *         return self.__rmul__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rsub__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rmul); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":232
 *         return self.__radd__(val)
 * 
 *     def __rmul__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rmul__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rmul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":235
 *         return self.__rmul__(val)
 * 
 *     def __rsub__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rsub__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_82__rsub__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_82__rsub__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rsub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_81__rsub__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_81__rsub__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rsub__", 0);

  /* "Proxy.pxi":236
 * 
 *     def __rsub__(self, val):
 *         return self.__rsub__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rmatmul__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rsub); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":235
 *         return self.__rmul__(val)
 * 
 *     def __rsub__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rsub__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rsub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":238
 *         return self.__rsub__(val)
 * 
 *     def __rmatmul__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rmatmul__(val)
 * 
 */

/* Python wrapper */
#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_84__rmatmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_84__rmatmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rmatmul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_83__rmatmul__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_VERSION_HEX >= 0x03050000)*/

#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_83__rmatmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rmatmul__", 0);

  /* "Proxy.pxi":239
 * 
 *     def __rmatmul__(self, val):
 *         return self.__rmatmul__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rtruediv__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rmatmul); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":238
 *         return self.__rsub__(val)
 * 
 *     def __rmatmul__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rmatmul__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rmatmul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_VERSION_HEX >= 0x03050000)*/

/* "Proxy.pxi":241
 *         return self.__rmatmul__(val)
 * 
 *     def __rtruediv__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rtruediv__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_86__rtruediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_86__rtruediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rtruediv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_85__rtruediv__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_85__rtruediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rtruediv__", 0);

  /* "Proxy.pxi":242
 * 
 *     def __rtruediv__(self, val):
 *         return self.__rtruediv__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rfloordiv__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rtruediv); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":241
 *         return self.__rmatmul__(val)
 * 
 *     def __rtruediv__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rtruediv__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rtruediv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":244
 *         return self.__rtruediv__(val)
 * 
 *     def __rfloordiv__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rfloordiv__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_88__rfloordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_88__rfloordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rfloordiv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_87__rfloordiv__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_87__rfloordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rfloordiv__", 0);

  /* "Proxy.pxi":245
 * 
 *     def __rfloordiv__(self, val):
 *         return self.__rfloordiv__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rmod__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_rfloordiv); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":244
 *         return self.__rtruediv__(val)
 * 
 *     def __rfloordiv__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rfloordiv__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rfloordiv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":247
 *         return self.__rfloordiv__(val)
 * 
 *     def __rmod__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rmod__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_90__rmod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_90__rmod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rmod__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_89__rmod__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_89__rmod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rmod__", 0);

  /* "Proxy.pxi":248
 * 
 *     def __rmod__(self, val):
 *         return self.__rmod__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rdivmod__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rmod); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":247
 *         return self.__rfloordiv__(val)
 * 
 *     def __rmod__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rmod__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rmod__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":250
 *         return self.__rmod__(val)
 * 
 *     def __rdivmod__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rdivmod__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_92__rdivmod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_92__rdivmod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rdivmod__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_91__rdivmod__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_91__rdivmod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rdivmod__", 0);

  /* "Proxy.pxi":251
 * 
 *     def __rdivmod__(self, val):
 *         return self.__rdivmod__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rpow__(self, val, mod):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rdivmod); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":250
 *         return self.__rmod__(val)
 * 
 *     def __rdivmod__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rdivmod__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rdivmod__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":253
 *         return self.__rdivmod__(val)
 * 
 *     def __rpow__(self, val, mod):             # <<<<<<<<<<<<<<
 *         return self.__rpow__(val, mod)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_94__rpow__(PyObject *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_94__rpow__(PyObject *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rpow__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_93__rpow__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val), ((PyObject *)__pyx_v_mod));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_93__rpow__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val, PyObject *__pyx_v_mod) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rpow__", 0);

  /* "Proxy.pxi":254
 * 
 *     def __rpow__(self, val, mod):
 *         return self.__rpow__(val, mod)             # <<<<<<<<<<<<<<
 * 
 *     def __rlshift__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rpow); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_val, __pyx_v_mod};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":253
 *         return self.__rdivmod__(val)
 * 
 *     def __rpow__(self, val, mod):             # <<<<<<<<<<<<<<
 *         return self.__rpow__(val, mod)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rpow__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":256
 *         return self.__rpow__(val, mod)
 * 
 *     def __rlshift__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rlshift__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_96__rlshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_96__rlshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rlshift__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_95__rlshift__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_95__rlshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rlshift__", 0);

  /* "Proxy.pxi":257
 * 
 *     def __rlshift__(self, val):
 *         return self.__rlshift__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rrshift__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rlshift); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":256
 *         return self.__rpow__(val, mod)
 * 
 *     def __rlshift__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rlshift__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rlshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":259
 *         return self.__rlshift__(val)
 * 
 *     def __rrshift__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rrshift__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_98__rrshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_98__rrshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rrshift__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_97__rrshift__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_97__rrshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rrshift__", 0);

  /* "Proxy.pxi":260
 * 
 *     def __rrshift__(self, val):
 *         return self.__rrshift__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rand__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rrshift); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":259
 *         return self.__rlshift__(val)
 * 
 *     def __rrshift__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rrshift__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rrshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":262
 *         return self.__rrshift__(val)
 * 
 *     def __rand__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rand__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_100__rand__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_100__rand__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rand__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_99__rand__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_99__rand__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rand__", 0);

  /* "Proxy.pxi":263
 * 
 *     def __rand__(self, val):
 *         return self.__rand__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __ror__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rand); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":262
 *         return self.__rrshift__(val)
 * 
 *     def __rand__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rand__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rand__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":265
 *         return self.__rand__(val)
 * 
 *     def __ror__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__ror__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_102__ror__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_102__ror__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ror__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_101__ror__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_101__ror__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ror__", 0);

  /* "Proxy.pxi":266
 * 
 *     def __ror__(self, val):
 *         return self.__ror__(val)             # <<<<<<<<<<<<<<
 * 
 *     def __rxor__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ror); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":265
 *         return self.__rand__(val)
 * 
 *     def __ror__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__ror__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__ror__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":268
 *         return self.__ror__(val)
 * 
 *     def __rxor__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rxor__(val)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_104__rxor__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_104__rxor__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rxor__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_103__rxor__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_103__rxor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rxor__", 0);

  /* "Proxy.pxi":269
 * 
 *     def __rxor__(self, val):
 *         return self.__rxor__(val)             # <<<<<<<<<<<<<<
 * 
 *     # Numeric types - augmented assignments - mutating
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rxor); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":268
 *         return self.__ror__(val)
 * 
 *     def __rxor__(self, val):             # <<<<<<<<<<<<<<
 *         return self.__rxor__(val)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__rxor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":272
 * 
 *     # Numeric types - augmented assignments - mutating
 *     def __iadd__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_106__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_106__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iadd__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_105__iadd__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_105__iadd__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iadd__", 0);

  /* "Proxy.pxi":273
 *     # Numeric types - augmented assignments - mutating
 *     def __iadd__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o += val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":274
 *     def __iadd__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o += val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 274, __pyx_L1_error)

    /* "Proxy.pxi":273
 *     # Numeric types - augmented assignments - mutating
 *     def __iadd__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o += val
 */
  }

  /* "Proxy.pxi":275
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o += val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":276
 *             raise frozen_error
 *         self.pvt_o += val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __imul__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":272
 * 
 *     # Numeric types - augmented assignments - mutating
 *     def __iadd__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":278
 *         return self
 * 
 *     def __imul__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_108__imul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_108__imul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__imul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_107__imul__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_107__imul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__imul__", 0);

  /* "Proxy.pxi":279
 * 
 *     def __imul__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o *= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":280
 *     def __imul__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o *= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 280, __pyx_L1_error)

    /* "Proxy.pxi":279
 * 
 *     def __imul__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o *= val
 */
  }

  /* "Proxy.pxi":281
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o *= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceMultiply(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":282
 *             raise frozen_error
 *         self.pvt_o *= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __isub__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":278
 *         return self
 * 
 *     def __imul__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__imul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":284
 *         return self
 * 
 *     def __isub__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_110__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_110__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__isub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_109__isub__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_109__isub__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__isub__", 0);

  /* "Proxy.pxi":285
 * 
 *     def __isub__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o -= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":286
 *     def __isub__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o -= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 286, __pyx_L1_error)

    /* "Proxy.pxi":285
 * 
 *     def __isub__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o -= val
 */
  }

  /* "Proxy.pxi":287
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o -= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceSubtract(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":288
 *             raise frozen_error
 *         self.pvt_o -= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __imod__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":284
 *         return self
 * 
 *     def __isub__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__isub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":290
 *         return self
 * 
 *     def __imod__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_112__imod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_112__imod__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__imod__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_111__imod__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_111__imod__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__imod__", 0);

  /* "Proxy.pxi":291
 * 
 *     def __imod__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o  %= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":292
 *     def __imod__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o  %= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 292, __pyx_L1_error)

    /* "Proxy.pxi":291
 * 
 *     def __imod__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o  %= val
 */
  }

  /* "Proxy.pxi":293
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o  %= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceRemainder(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":294
 *             raise frozen_error
 *         self.pvt_o  %= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __ilshift__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":290
 *         return self
 * 
 *     def __imod__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__imod__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":296
 *         return self
 * 
 *     def __ilshift__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_114__ilshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_114__ilshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ilshift__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_113__ilshift__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_113__ilshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ilshift__", 0);

  /* "Proxy.pxi":297
 * 
 *     def __ilshift__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o <<= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":298
 *     def __ilshift__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o <<= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 298, __pyx_L1_error)

    /* "Proxy.pxi":297
 * 
 *     def __ilshift__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o <<= val
 */
  }

  /* "Proxy.pxi":299
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o <<= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceLshift(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":300
 *             raise frozen_error
 *         self.pvt_o <<= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __irshift__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":296
 *         return self
 * 
 *     def __ilshift__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__ilshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":302
 *         return self
 * 
 *     def __irshift__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_116__irshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_116__irshift__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__irshift__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_115__irshift__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_115__irshift__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__irshift__", 0);

  /* "Proxy.pxi":303
 * 
 *     def __irshift__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o >>= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":304
 *     def __irshift__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o >>= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 304, __pyx_L1_error)

    /* "Proxy.pxi":303
 * 
 *     def __irshift__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o >>= val
 */
  }

  /* "Proxy.pxi":305
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o >>= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceRshift(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":306
 *             raise frozen_error
 *         self.pvt_o >>= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __iand__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":302
 *         return self
 * 
 *     def __irshift__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__irshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":308
 *         return self
 * 
 *     def __iand__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_118__iand__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_118__iand__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iand__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_117__iand__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_117__iand__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iand__", 0);

  /* "Proxy.pxi":309
 * 
 *     def __iand__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o &= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":310
 *     def __iand__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o &= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 310, __pyx_L1_error)

    /* "Proxy.pxi":309
 * 
 *     def __iand__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o &= val
 */
  }

  /* "Proxy.pxi":311
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o &= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceAnd(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":312
 *             raise frozen_error
 *         self.pvt_o &= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __ior__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":308
 *         return self
 * 
 *     def __iand__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__iand__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":314
 *         return self
 * 
 *     def __ior__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_120__ior__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_120__ior__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ior__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_119__ior__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_119__ior__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ior__", 0);

  /* "Proxy.pxi":315
 * 
 *     def __ior__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o |= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":316
 *     def __ior__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o |= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 316, __pyx_L1_error)

    /* "Proxy.pxi":315
 * 
 *     def __ior__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o |= val
 */
  }

  /* "Proxy.pxi":317
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o |= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceOr(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":318
 *             raise frozen_error
 *         self.pvt_o |= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __ixor__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":314
 *         return self
 * 
 *     def __ior__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__ior__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":320
 *         return self
 * 
 *     def __ixor__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_122__ixor__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_122__ixor__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ixor__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_121__ixor__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_121__ixor__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ixor__", 0);

  /* "Proxy.pxi":321
 * 
 *     def __ixor__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o ^= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":322
 *     def __ixor__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o ^= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 322, __pyx_L1_error)

    /* "Proxy.pxi":321
 * 
 *     def __ixor__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o ^= val
 */
  }

  /* "Proxy.pxi":323
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o ^= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceXor(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":324
 *             raise frozen_error
 *         self.pvt_o ^= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __ipow__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":320
 *         return self
 * 
 *     def __ixor__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__ixor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":326
 *         return self
 * 
 *     def __ipow__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_124__ipow__(PyObject *__pyx_v_self, PyObject *__pyx_v_val, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_124__ipow__(PyObject *__pyx_v_self, PyObject *__pyx_v_val, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ipow__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_123__ipow__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_123__ipow__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ipow__", 0);

  /* "Proxy.pxi":327
 * 
 *     def __ipow__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o **= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":328
 *     def __ipow__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o **= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 328, __pyx_L1_error)

    /* "Proxy.pxi":327
 * 
 *     def __ipow__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o **= val
 */
  }

  /* "Proxy.pxi":329
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o **= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlacePower(__pyx_v_self->pvt_o, __pyx_v_val, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":330
 *             raise frozen_error
 *         self.pvt_o **= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __itruediv__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":326
 *         return self
 * 
 *     def __ipow__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__ipow__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":332
 *         return self
 * 
 *     def __itruediv__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_126__itruediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_126__itruediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__itruediv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_125__itruediv__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_125__itruediv__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__itruediv__", 0);

  /* "Proxy.pxi":333
 * 
 *     def __itruediv__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o /= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":334
 *     def __itruediv__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o /= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 334, __pyx_L1_error)

    /* "Proxy.pxi":333
 * 
 *     def __itruediv__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o /= val
 */
  }

  /* "Proxy.pxi":335
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o /= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = __Pyx_PyNumber_InPlaceDivide(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":336
 *             raise frozen_error
 *         self.pvt_o /= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __ifloordiv__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":332
 *         return self
 * 
 *     def __itruediv__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__itruediv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":338
 *         return self
 * 
 *     def __ifloordiv__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_128__ifloordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_128__ifloordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ifloordiv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_127__ifloordiv__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_127__ifloordiv__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ifloordiv__", 0);

  /* "Proxy.pxi":339
 * 
 *     def __ifloordiv__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o //= val
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":340
 *     def __ifloordiv__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o //= val
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 340, __pyx_L1_error)

    /* "Proxy.pxi":339
 * 
 *     def __ifloordiv__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o //= val
 */
  }

  /* "Proxy.pxi":341
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o //= val             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceFloorDivide(__pyx_v_self->pvt_o, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pvt_o);
  __Pyx_DECREF(__pyx_v_self->pvt_o);
  __pyx_v_self->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Proxy.pxi":342
 *             raise frozen_error
 *         self.pvt_o //= val
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __imatmul__(self, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":338
 *         return self
 * 
 *     def __ifloordiv__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__ifloordiv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":344
 *         return self
 * 
 *     def __imatmul__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_130__imatmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_130__imatmul__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__imatmul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_129__imatmul__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_VERSION_HEX >= 0x03050000)*/

#if PY_VERSION_HEX >= 0x03050000
static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_129__imatmul__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__imatmul__", 0);

  /* "Proxy.pxi":345
 * 
 *     def __imatmul__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o.__imatmul__(val)
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":346
 *     def __imatmul__(self, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o.__imatmul__(val)
 *         return self
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 346, __pyx_L1_error)

    /* "Proxy.pxi":345
 * 
 *     def __imatmul__(self, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o.__imatmul__(val)
 */
  }

  /* "Proxy.pxi":347
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o.__imatmul__(val)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_imatmul); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Proxy.pxi":348
 *             raise frozen_error
 *         self.pvt_o.__imatmul__(val)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     # Mutating methods of containers
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "Proxy.pxi":344
 *         return self
 * 
 *     def __imatmul__(self, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__imatmul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_VERSION_HEX >= 0x03050000)*/

/* "Proxy.pxi":351
 * 
 *     # Mutating methods of containers
 *     def __setitem__(self, key, val):             # <<<<<<<<<<<<<<
 *         if isinstance(
 *             self.pvt_o,
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_5Proxy_132__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_5Proxy_132__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_131__setitem__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_key), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_5Proxy_131__setitem__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "Proxy.pxi":355
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *             )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Proxy.pxi":356
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Proxy.pxi":353
 *     def __setitem__(self, key, val):
 *         if isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_1 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_1);

  /* "Proxy.pxi":355
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *             )
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_1, __pyx_t_2); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":353
 *     def __setitem__(self, key, val):
 *         if isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_1 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_1);

  /* "Proxy.pxi":356
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_1, __pyx_t_3); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":352
 *     # Mutating methods of containers
 *     def __setitem__(self, key, val):
 *         if isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (__pyx_t_4) {

    /* "Proxy.pxi":359
 *             )
 *         ):
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 raise frozen_error
 *         self.pvt_o.__setitem__(key, val)
 */
    if (unlikely(__pyx_v_self->frozen)) {

      /* "Proxy.pxi":360
 *         ):
 *             if self.frozen:
 *                 raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o.__setitem__(key, val)
 * 
 */
      __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
      __PYX_ERR(2, 360, __pyx_L1_error)

      /* "Proxy.pxi":359
 *             )
 *         ):
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 raise frozen_error
 *         self.pvt_o.__setitem__(key, val)
 */
    }

    /* "Proxy.pxi":352
 *     # Mutating methods of containers
 *     def __setitem__(self, key, val):
 *         if isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  }

  /* "Proxy.pxi":361
 *             if self.frozen:
 *                 raise frozen_error
 *         self.pvt_o.__setitem__(key, val)             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, key):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_setitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_key, __pyx_v_val};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":351
 * 
 *     # Mutating methods of containers
 *     def __setitem__(self, key, val):             # <<<<<<<<<<<<<<
 *         if isinstance(
 *             self.pvt_o,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":363
 *         self.pvt_o.__setitem__(key, val)
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         if isinstance(
 *             self.pvt_o,
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_5Proxy_134__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_9pyprotect_9protected_5Proxy_134__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_133__delitem__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_5Proxy_133__delitem__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "Proxy.pxi":367
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *             )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Proxy.pxi":368
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Proxy.pxi":365
 *     def __delitem__(self, key):
 *         if isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_1 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_1);

  /* "Proxy.pxi":367
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *             )
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_1, __pyx_t_2); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":365
 *     def __delitem__(self, key):
 *         if isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_1 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_1);

  /* "Proxy.pxi":368
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_1, __pyx_t_3); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":364
 * 
 *     def __delitem__(self, key):
 *         if isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (__pyx_t_4) {

    /* "Proxy.pxi":371
 *             )
 *         ):
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 raise frozen_error
 *         self.pvt_o.__delitem__(key)
 */
    if (unlikely(__pyx_v_self->frozen)) {

      /* "Proxy.pxi":372
 *         ):
 *             if self.frozen:
 *                 raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o.__delitem__(key)
 * 
 */
      __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
      __PYX_ERR(2, 372, __pyx_L1_error)

      /* "Proxy.pxi":371
 *             )
 *         ):
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 raise frozen_error
 *         self.pvt_o.__delitem__(key)
 */
    }

    /* "Proxy.pxi":364
 * 
 *     def __delitem__(self, key):
 *         if isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  }

  /* "Proxy.pxi":373
 *             if self.frozen:
 *                 raise frozen_error
 *         self.pvt_o.__delitem__(key)             # <<<<<<<<<<<<<<
 * 
 *     # Just pass to wrapped
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_delitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_key};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":363
 *         self.pvt_o.__setitem__(key, val)
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         if isinstance(
 *             self.pvt_o,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":376
 * 
 *     # Just pass to wrapped
 *     def __instancecheck__(self, inst):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_136__instancecheck__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_136__instancecheck__ = {"__instancecheck__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_136__instancecheck__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_136__instancecheck__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_inst = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__instancecheck__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inst,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inst,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_inst)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 376, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__instancecheck__") < 0)) __PYX_ERR(2, 376, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_inst = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__instancecheck__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 376, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__instancecheck__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_135__instancecheck__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_inst);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_135__instancecheck__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_inst) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__instancecheck__", 0);

  /* "Proxy.pxi":377
 *     # Just pass to wrapped
 *     def __instancecheck__(self, inst):
 *         return self.pvt_o.__instancecheck__(inst)             # <<<<<<<<<<<<<<
 * 
 *     def __subclasscheck__(self, subclass):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_instancecheck); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_inst};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":376
 * 
 *     # Just pass to wrapped
 *     def __instancecheck__(self, inst):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__instancecheck__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":379
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 *     def __subclasscheck__(self, subclass):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__subclasscheck__(subclass)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_138__subclasscheck__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_138__subclasscheck__ = {"__subclasscheck__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_138__subclasscheck__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_138__subclasscheck__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_subclass = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__subclasscheck__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_subclass,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_subclass,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_subclass)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 379, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__subclasscheck__") < 0)) __PYX_ERR(2, 379, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_subclass = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__subclasscheck__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 379, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__subclasscheck__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_137__subclasscheck__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_subclass);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_137__subclasscheck__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_subclass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__subclasscheck__", 0);

  /* "Proxy.pxi":380
 * 
 *     def __subclasscheck__(self, subclass):
 *         return self.pvt_o.__subclasscheck__(subclass)             # <<<<<<<<<<<<<<
 * 
 *     # Context managers - intrinsic behavior - pass to wrapped
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_subclasscheck); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_subclass};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":379
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 *     def __subclasscheck__(self, subclass):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__subclasscheck__(subclass)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__subclasscheck__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":383
 * 
 *     # Context managers - intrinsic behavior - pass to wrapped
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__enter__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_140__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_140__enter__ = {"__enter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_140__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_140__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__enter__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__enter__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_139__enter__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_139__enter__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__enter__", 0);

  /* "Proxy.pxi":384
 *     # Context managers - intrinsic behavior - pass to wrapped
 *     def __enter__(self):
 *         return self.pvt_o.__enter__()             # <<<<<<<<<<<<<<
 * 
 *     def __exit__(self, exc_type, exc_value, tb):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_enter); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":383
 * 
 *     # Context managers - intrinsic behavior - pass to wrapped
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__enter__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__enter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":386
 *         return self.pvt_o.__enter__()
 * 
 *     def __exit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__exit(exc_type, exc_value, tb)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_142__exit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_142__exit__ = {"__exit__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_142__exit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_142__exit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_exc_type = 0;
  PyObject *__pyx_v_exc_value = 0;
  PyObject *__pyx_v_tb = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc_type,&__pyx_n_s_exc_value,&__pyx_n_s_tb,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc_type,&__pyx_n_s_exc_value,&__pyx_n_s_tb,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 386, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc_value)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 386, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, 1); __PYX_ERR(2, 386, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tb)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 386, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, 2); __PYX_ERR(2, 386, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__exit__") < 0)) __PYX_ERR(2, 386, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_exc_type = values[0];
    __pyx_v_exc_value = values[1];
    __pyx_v_tb = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 386, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_141__exit__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_exc_type, __pyx_v_exc_value, __pyx_v_tb);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_141__exit__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_exc_type, PyObject *__pyx_v_exc_value, PyObject *__pyx_v_tb) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__exit__", 0);

  /* "Proxy.pxi":387
 * 
 *     def __exit__(self, exc_type, exc_value, tb):
 *         return self.pvt_o.__exit(exc_type, exc_value, tb)             # <<<<<<<<<<<<<<
 * 
 *     # Async context managers - intrinsic behavior - pass to wrapped
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_Proxy__exit); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_exc_type, __pyx_v_exc_value, __pyx_v_tb};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":386
 *         return self.pvt_o.__enter__()
 * 
 *     def __exit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__exit(exc_type, exc_value, tb)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":390
 * 
 *     # Async context managers - intrinsic behavior - pass to wrapped
 *     def __aenter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aenter__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_144__aenter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_144__aenter__ = {"__aenter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_144__aenter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_144__aenter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aenter__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__aenter__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__aenter__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_143__aenter__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_143__aenter__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__aenter__", 0);

  /* "Proxy.pxi":391
 *     # Async context managers - intrinsic behavior - pass to wrapped
 *     def __aenter__(self):
 *         return self.pvt_o.__aenter__()             # <<<<<<<<<<<<<<
 * 
 *     def __aexit__(self, exc_type, exc_value, tb):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_aenter); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":390
 * 
 *     # Async context managers - intrinsic behavior - pass to wrapped
 *     def __aenter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aenter__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__aenter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":393
 *         return self.pvt_o.__aenter__()
 * 
 *     def __aexit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aexit(exc_type, exc_value, tb)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_146__aexit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_146__aexit__ = {"__aexit__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_146__aexit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_146__aexit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_exc_type = 0;
  PyObject *__pyx_v_exc_value = 0;
  PyObject *__pyx_v_tb = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aexit__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc_type,&__pyx_n_s_exc_value,&__pyx_n_s_tb,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc_type,&__pyx_n_s_exc_value,&__pyx_n_s_tb,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 393, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc_value)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 393, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__aexit__", 1, 3, 3, 1); __PYX_ERR(2, 393, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tb)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 393, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__aexit__", 1, 3, 3, 2); __PYX_ERR(2, 393, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__aexit__") < 0)) __PYX_ERR(2, 393, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_exc_type = values[0];
    __pyx_v_exc_value = values[1];
    __pyx_v_tb = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__aexit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 393, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__aexit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_145__aexit__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_exc_type, __pyx_v_exc_value, __pyx_v_tb);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_145__aexit__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_exc_type, PyObject *__pyx_v_exc_value, PyObject *__pyx_v_tb) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__aexit__", 0);

  /* "Proxy.pxi":394
 * 
 *     def __aexit__(self, exc_type, exc_value, tb):
 *         return self.pvt_o.__aexit(exc_type, exc_value, tb)             # <<<<<<<<<<<<<<
 * 
 *     # Customizing positional arguments in class pattern matching
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_Proxy__aexit); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_exc_type, __pyx_v_exc_value, __pyx_v_tb};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":393
 *         return self.pvt_o.__aenter__()
 * 
 *     def __aexit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aexit(exc_type, exc_value, tb)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__aexit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":398
 *     # Customizing positional arguments in class pattern matching
 *     # Returns tuple of strings (immutable) - pass to wrapped
 *     def __match_args__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__match_args__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_148__match_args__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_148__match_args__ = {"__match_args__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_148__match_args__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_148__match_args__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__match_args__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__match_args__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__match_args__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_147__match_args__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_147__match_args__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__match_args__", 0);

  /* "Proxy.pxi":399
 *     # Returns tuple of strings (immutable) - pass to wrapped
 *     def __match_args__(self):
 *         return self.pvt_o.__match_args__()             # <<<<<<<<<<<<<<
 * 
 *     # Coroutine objects - intrinsic behavior - pass to wrapped
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_match_args); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":398
 *     # Customizing positional arguments in class pattern matching
 *     # Returns tuple of strings (immutable) - pass to wrapped
 *     def __match_args__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__match_args__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__match_args__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":402
 * 
 *     # Coroutine objects - intrinsic behavior - pass to wrapped
 *     def send(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.send(value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_150send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_150send = {"send", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_150send, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_150send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("send (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 402, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "send") < 0)) __PYX_ERR(2, 402, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("send", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 402, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_149send(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_149send(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("send", 0);

  /* "Proxy.pxi":403
 *     # Coroutine objects - intrinsic behavior - pass to wrapped
 *     def send(self, value):
 *         return self.pvt_o.send(value)             # <<<<<<<<<<<<<<
 * 
 *     def throw(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_send); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":402
 * 
 *     # Coroutine objects - intrinsic behavior - pass to wrapped
 *     def send(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.send(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":405
 *         return self.pvt_o.send(value)
 * 
 *     def throw(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.throw(value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_152throw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_152throw = {"throw", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_152throw, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_152throw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("throw (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 405, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "throw") < 0)) __PYX_ERR(2, 405, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("throw", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 405, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.throw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_151throw(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_151throw(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("throw", 0);

  /* "Proxy.pxi":406
 * 
 *     def throw(self, value):
 *         return self.pvt_o.throw(value)             # <<<<<<<<<<<<<<
 * 
 *     # Implementing descriptors - mutate object
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_throw); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":405
 *         return self.pvt_o.send(value)
 * 
 *     def throw(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.throw(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.throw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":409
 * 
 *     # Implementing descriptors - mutate object
 *     def __set__(self, inst, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_5Proxy_154__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_5Proxy_154__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_153__set__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_inst), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_5Proxy_153__set__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "Proxy.pxi":410
 *     # Implementing descriptors - mutate object
 *     def __set__(self, inst, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o.__set__(inst, val)
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":411
 *     def __set__(self, inst, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o.__set__(inst, val)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 411, __pyx_L1_error)

    /* "Proxy.pxi":410
 *     # Implementing descriptors - mutate object
 *     def __set__(self, inst, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o.__set__(inst, val)
 */
  }

  /* "Proxy.pxi":412
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o.__set__(inst, val)             # <<<<<<<<<<<<<<
 * 
 *     def __delete__(self, inst):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_set_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_inst, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Proxy.pxi":409
 * 
 *     # Implementing descriptors - mutate object
 *     def __set__(self, inst, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":414
 *         self.pvt_o.__set__(inst, val)
 * 
 *     def __delete__(self, inst):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_5Proxy_156__delete__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst); /*proto*/
static int __pyx_pw_9pyprotect_9protected_5Proxy_156__delete__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delete__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_155__delete__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), ((PyObject *)__pyx_v_inst));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_5Proxy_155__delete__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_inst) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delete__", 0);

  /* "Proxy.pxi":415
 * 
 *     def __delete__(self, inst):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o.__delete__(inst)
 */
  if (unlikely(__pyx_v_self->frozen)) {

    /* "Proxy.pxi":416
 *     def __delete__(self, inst):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         self.pvt_o.__delete__(inst)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 416, __pyx_L1_error)

    /* "Proxy.pxi":415
 * 
 *     def __delete__(self, inst):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         self.pvt_o.__delete__(inst)
 */
  }

  /* "Proxy.pxi":417
 *         if self.frozen:
 *             raise frozen_error
 *         self.pvt_o.__delete__(inst)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_delete); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_inst};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Proxy.pxi":414
 *         self.pvt_o.__set__(inst, val)
 * 
 *     def __delete__(self, inst):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__delete__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":428
 *     # --------------------------------------------------------------------
 * 
 *     def clear(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_158clear(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_158clear = {"clear", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_158clear, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_158clear(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_157clear(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_157clear(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);

  /* "Proxy.pxi":429
 * 
 *     def clear(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":432
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSet,
 *                 types.FrameType,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":433
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSet,             # <<<<<<<<<<<<<<
 *                 types.FrameType,
 *             )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":434
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSet,
 *                 types.FrameType,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FrameType); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":430
 *     def clear(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":432
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSet,
 *                 types.FrameType,
 */
  __pyx_t_7 = PyObject_IsInstance(__pyx_t_2, __pyx_t_3); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L6_bool_binop_done;
  }

  /* "Proxy.pxi":430
 *     def clear(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":433
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSet,             # <<<<<<<<<<<<<<
 *                 types.FrameType,
 *             )
 */
  __pyx_t_7 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L6_bool_binop_done;
  }

  /* "Proxy.pxi":430
 *     def clear(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":434
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSet,
 *                 types.FrameType,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __pyx_t_7 = PyObject_IsInstance(__pyx_t_2, __pyx_t_5); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;

  /* "Proxy.pxi":429
 * 
 *     def clear(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":437
 *             )
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.clear(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 437, __pyx_L1_error)

    /* "Proxy.pxi":429
 * 
 *     def clear(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  }

  /* "Proxy.pxi":438
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.clear(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def setdefault(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_clear); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_v_args, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":428
 *     # --------------------------------------------------------------------
 * 
 *     def clear(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":440
 *         return self.pvt_o.clear(*args, **kwargs)
 * 
 *     def setdefault(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,CollectionsABC.MutableMapping
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_160setdefault(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_160setdefault = {"setdefault", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_160setdefault, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_160setdefault(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setdefault (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "setdefault", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_159setdefault(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_159setdefault(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setdefault", 0);

  /* "Proxy.pxi":441
 * 
 *     def setdefault(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,CollectionsABC.MutableMapping
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":442
 *     def setdefault(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,CollectionsABC.MutableMapping             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":441
 * 
 *     def setdefault(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,CollectionsABC.MutableMapping
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":444
 *             self.pvt_o,CollectionsABC.MutableMapping
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.setdefault(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 444, __pyx_L1_error)

    /* "Proxy.pxi":441
 * 
 *     def setdefault(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,CollectionsABC.MutableMapping
 *         ):
 */
  }

  /* "Proxy.pxi":445
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.setdefault(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def pop(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_setdefault); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":440
 *         return self.pvt_o.clear(*args, **kwargs)
 * 
 *     def setdefault(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,CollectionsABC.MutableMapping
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.setdefault", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":447
 *         return self.pvt_o.setdefault(*args, **kwargs)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_162pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_162pop = {"pop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_162pop, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_162pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pop", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_161pop(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_161pop(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 0);

  /* "Proxy.pxi":448
 * 
 *     def pop(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":451
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *                 CollectionsABC.MutableSet,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":452
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSet,
 *             )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":453
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,
 *                 CollectionsABC.MutableSet,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSet); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":449
 *     def pop(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":451
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableMapping,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *                 CollectionsABC.MutableSet,
 */
  __pyx_t_7 = PyObject_IsInstance(__pyx_t_2, __pyx_t_3); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L6_bool_binop_done;
  }

  /* "Proxy.pxi":449
 *     def pop(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":452
 *             (
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSet,
 *             )
 */
  __pyx_t_7 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L6_bool_binop_done;
  }

  /* "Proxy.pxi":449
 *     def pop(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableMapping,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":453
 *                 CollectionsABC.MutableMapping,
 *                 CollectionsABC.MutableSequence,
 *                 CollectionsABC.MutableSet,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __pyx_t_7 = PyObject_IsInstance(__pyx_t_2, __pyx_t_5); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;

  /* "Proxy.pxi":448
 * 
 *     def pop(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":456
 *             )
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.pop(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 456, __pyx_L1_error)

    /* "Proxy.pxi":448
 * 
 *     def pop(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  }

  /* "Proxy.pxi":457
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.pop(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def popitem(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_pop); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_v_args, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":447
 *         return self.pvt_o.setdefault(*args, **kwargs)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":459
 *         return self.pvt_o.pop(*args, **kwargs)
 * 
 *     def popitem(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_164popitem(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_164popitem = {"popitem", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_164popitem, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_164popitem(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("popitem (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "popitem", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_163popitem(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_163popitem(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("popitem", 0);

  /* "Proxy.pxi":460
 * 
 *     def popitem(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":461
 *     def popitem(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":460
 * 
 *     def popitem(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":463
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.popitem(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 463, __pyx_L1_error)

    /* "Proxy.pxi":460
 * 
 *     def popitem(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 */
  }

  /* "Proxy.pxi":464
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.popitem(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def update(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_popitem); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":459
 *         return self.pvt_o.pop(*args, **kwargs)
 * 
 *     def popitem(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.popitem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":466
 *         return self.pvt_o.popitem(*args, **kwargs)
 * 
 *     def update(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_166update(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_166update = {"update", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_166update, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_166update(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "update", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_165update(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_165update(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);

  /* "Proxy.pxi":467
 * 
 *     def update(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":468
 *     def update(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":467
 * 
 *     def update(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":470
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.update(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 470, __pyx_L1_error)

    /* "Proxy.pxi":467
 * 
 *     def update(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableMapping
 *         ):
 */
  }

  /* "Proxy.pxi":471
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.update(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":466
 *         return self.pvt_o.popitem(*args, **kwargs)
 * 
 *     def update(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":474
 * 
 * 
 *     def append(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_168append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_168append = {"append", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_168append, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_168append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "append", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_167append(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_167append(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);

  /* "Proxy.pxi":475
 * 
 *     def append(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":476
 *     def append(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":475
 * 
 *     def append(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":478
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.append(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 478, __pyx_L1_error)

    /* "Proxy.pxi":475
 * 
 *     def append(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  }

  /* "Proxy.pxi":479
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.append(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def extend(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_append); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":474
 * 
 * 
 *     def append(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":481
 *         return self.pvt_o.append(*args, **kwargs)
 * 
 *     def extend(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_170extend(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_170extend = {"extend", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_170extend, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_170extend(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "extend", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_169extend(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_169extend(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "Proxy.pxi":482
 * 
 *     def extend(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":483
 *     def extend(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":482
 * 
 *     def extend(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":485
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.extend(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 485, __pyx_L1_error)

    /* "Proxy.pxi":482
 * 
 *     def extend(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  }

  /* "Proxy.pxi":486
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.extend(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_extend); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":481
 *         return self.pvt_o.append(*args, **kwargs)
 * 
 *     def extend(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":488
 *         return self.pvt_o.extend(*args, **kwargs)
 * 
 *     def insert(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_172insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_172insert = {"insert", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_172insert, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_172insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "insert", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_171insert(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_171insert(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insert", 0);

  /* "Proxy.pxi":489
 * 
 *     def insert(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":490
 *     def insert(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":489
 * 
 *     def insert(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":492
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.insert(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 492, __pyx_L1_error)

    /* "Proxy.pxi":489
 * 
 *     def insert(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  }

  /* "Proxy.pxi":493
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.insert(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def sort(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_insert); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":488
 *         return self.pvt_o.extend(*args, **kwargs)
 * 
 *     def insert(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":495
 *         return self.pvt_o.insert(*args, **kwargs)
 * 
 *     def sort(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_174sort(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_174sort = {"sort", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_174sort, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_174sort(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sort (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sort", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_173sort(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_173sort(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sort", 0);

  /* "Proxy.pxi":496
 * 
 *     def sort(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":497
 *     def sort(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":496
 * 
 *     def sort(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 496, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":499
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.sort(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 499, __pyx_L1_error)

    /* "Proxy.pxi":496
 * 
 *     def sort(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSequence
 *         ):
 */
  }

  /* "Proxy.pxi":500
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.sort(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def add(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_sort); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":495
 *         return self.pvt_o.insert(*args, **kwargs)
 * 
 *     def sort(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.sort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":502
 *         return self.pvt_o.sort(*args, **kwargs)
 * 
 *     def add(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_176add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_176add = {"add", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_176add, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_176add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "add", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_175add(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_175add(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 0);

  /* "Proxy.pxi":503
 * 
 *     def add(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":504
 *     def add(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":503
 * 
 *     def add(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":506
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.add(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 506, __pyx_L1_error)

    /* "Proxy.pxi":503
 * 
 *     def add(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 */
  }

  /* "Proxy.pxi":507
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.add(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def discard(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_add_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":502
 *         return self.pvt_o.sort(*args, **kwargs)
 * 
 *     def add(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":509
 *         return self.pvt_o.add(*args, **kwargs)
 * 
 *     def discard(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_178discard(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_178discard = {"discard", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_178discard, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_178discard(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("discard (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "discard", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_177discard(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_177discard(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("discard", 0);

  /* "Proxy.pxi":510
 * 
 *     def discard(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":511
 *     def discard(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet             # <<<<<<<<<<<<<<
 *         ):
 *             raise frozen_error
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Proxy.pxi":510
 * 
 *     def discard(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 510, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":513
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.discard(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 513, __pyx_L1_error)

    /* "Proxy.pxi":510
 * 
 *     def discard(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o, CollectionsABC.MutableSet
 *         ):
 */
  }

  /* "Proxy.pxi":514
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.discard(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def remove(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_discard); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":509
 *         return self.pvt_o.add(*args, **kwargs)
 * 
 *     def discard(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.discard", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":516
 *         return self.pvt_o.discard(*args, **kwargs)
 * 
 *     def remove(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_180remove(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_180remove = {"remove", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_180remove, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_180remove(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "remove", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_179remove(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_179remove(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove", 0);

  /* "Proxy.pxi":517
 * 
 *     def remove(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":520
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableSet,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *             )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSet); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":521
 *             (
 *                 CollectionsABC.MutableSet,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Proxy.pxi":518
 *     def remove(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableSet,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":520
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableSet,             # <<<<<<<<<<<<<<
 *                 CollectionsABC.MutableSequence,
 *             )
 */
  __pyx_t_6 = PyObject_IsInstance(__pyx_t_2, __pyx_t_3); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }

  /* "Proxy.pxi":518
 *     def remove(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableSet,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":521
 *             (
 *                 CollectionsABC.MutableSet,
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __pyx_t_6 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;

  /* "Proxy.pxi":517
 * 
 *     def remove(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":524
 *             )
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.remove(*args, **kwargs)
 * 
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 524, __pyx_L1_error)

    /* "Proxy.pxi":517
 * 
 *     def remove(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  }

  /* "Proxy.pxi":525
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.remove(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def reverse(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_remove); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_v_args, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":516
 *         return self.pvt_o.discard(*args, **kwargs)
 * 
 *     def remove(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Proxy.pxi":527
 *         return self.pvt_o.remove(*args, **kwargs)
 * 
 *     def reverse(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_182reverse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_182reverse = {"reverse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_182reverse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_182reverse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reverse (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "reverse", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_181reverse(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_181reverse(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reverse", 0);

  /* "Proxy.pxi":528
 * 
 *     def reverse(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  if (__pyx_v_self->frozen) {
  } else {
    __pyx_t_1 = __pyx_v_self->frozen;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Proxy.pxi":529
 *     def reverse(self, *args, **kwargs):
 *         if self.frozen and isinstance(
 *             self.pvt_o,             # <<<<<<<<<<<<<<
 *             (
 *                 CollectionsABC.MutableSequence,
 */
  __pyx_t_2 = __pyx_v_self->pvt_o;
  __Pyx_INCREF(__pyx_t_2);

  /* "Proxy.pxi":531
 *             self.pvt_o,
 *             (
 *                 CollectionsABC.MutableSequence,             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MutableSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "Proxy.pxi":528
 * 
 *     def reverse(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_2, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Proxy.pxi":534
 *             )
 *         ):
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         return self.pvt_o.reverse(*args, **kwargs)
 *     # --------------------------------------------------------------------
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(2, 534, __pyx_L1_error)

    /* "Proxy.pxi":528
 * 
 *     def reverse(self, *args, **kwargs):
 *         if self.frozen and isinstance(             # <<<<<<<<<<<<<<
 *             self.pvt_o,
 *             (
 */
  }

  /* "Proxy.pxi":535
 *         ):
 *             raise frozen_error
 *         return self.pvt_o.reverse(*args, **kwargs)             # <<<<<<<<<<<<<<
 *     # --------------------------------------------------------------------
 *     # End of mutating methods of containers
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pvt_o, __pyx_n_s_reverse); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Proxy.pxi":527
 *         return self.pvt_o.remove(*args, **kwargs)
 * 
 *     def reverse(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.reverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_184__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_184__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_184__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_184__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_183__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_183__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.frozen, self.pvt_o)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->pvt_o);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->pvt_o);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.frozen, self.pvt_o)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.frozen, self.pvt_o)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.pvt_o is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.frozen, self.pvt_o)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.pvt_o is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->pvt_o != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.pvt_o is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.pvt_o is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Proxy); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_152356376);
    __Pyx_GIVEREF(__pyx_int_152356376);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_152356376);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.pvt_o is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, None), state
 *     else:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Proxy__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Proxy); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_152356376);
    __Pyx_GIVEREF(__pyx_int_152356376);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_152356376);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Proxy__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_186__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_5Proxy_186__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_186__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_5Proxy_186__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_5Proxy_185__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_5Proxy_185__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Proxy__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Proxy__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Proxy__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Proxy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":23
 *     cdef object hidden_private_attr
 * 
 *     def __init__(self, o, frozen=False, rules=None, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *         '''
 *         o: object to be wrapped
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_7Wrapped_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_7Wrapped___init__, "\n        o: object to be wrapped\n        frozen: bool: If True, no attribute can be modified\n        rules: dict (if called from Protected.__init__) or None\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_7Wrapped___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_7Wrapped_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_frozen = 0;
  PyObject *__pyx_v_rules = 0;
  PyObject *__pyx_v_oldstyle_class = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,&__pyx_n_s_rules,&__pyx_n_s_oldstyle_class,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,&__pyx_n_s_rules,&__pyx_n_s_oldstyle_class,0};
    #endif
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject *)Py_False);
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 23, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_frozen);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 23, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rules);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 23, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_oldstyle_class);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 23, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(3, 23, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_o = values[0];
    __pyx_v_frozen = values[1];
    __pyx_v_rules = values[2];
    __pyx_v_oldstyle_class = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 4, __pyx_nargs); __PYX_ERR(3, 23, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped___init__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_o, __pyx_v_frozen, __pyx_v_rules, __pyx_v_oldstyle_class);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_7Wrapped___init__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen, PyObject *__pyx_v_rules, PyObject *__pyx_v_oldstyle_class) {
  PyTypeObject *__pyx_v_protect_class = NULL;
  PyTypeObject *__pyx_v_private_class = NULL;
  PyObject *__pyx_v_id_class = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_rules);

  /* "Wrapped_Frozen.pxi":29
 *         rules: dict (if called from Protected.__init__) or None
 *         '''
 *         if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *             # We claim to be avoiding double-wrapping, so this exception
 *             # should never be raised
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  if (unlikely(__pyx_t_1)) {

    /* "Wrapped_Frozen.pxi":32
 *             # We claim to be avoiding double-wrapping, so this exception
 *             # should never be raised
 *             raise RuntimeError('Double-wrapped!')             # <<<<<<<<<<<<<<
 * 
 *         self.pvt_o = o
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(3, 32, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":29
 *         rules: dict (if called from Protected.__init__) or None
 *         '''
 *         if isinstance(o, Wrapped):             # <<<<<<<<<<<<<<
 *             # We claim to be avoiding double-wrapping, so this exception
 *             # should never be raised
 */
  }

  /* "Wrapped_Frozen.pxi":34
 *             raise RuntimeError('Double-wrapped!')
 * 
 *         self.pvt_o = o             # <<<<<<<<<<<<<<
 *         self.frozen = bool(frozen)
 *         if oldstyle_class is None:
 */
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.pvt_o);
  __pyx_v_self->__pyx_base.pvt_o = __pyx_v_o;

  /* "Wrapped_Frozen.pxi":35
 * 
 *         self.pvt_o = o
 *         self.frozen = bool(frozen)             # <<<<<<<<<<<<<<
 *         if oldstyle_class is None:
 *             self.oldstyle_class = False
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_frozen); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 35, __pyx_L1_error)
  __pyx_v_self->__pyx_base.frozen = (!(!__pyx_t_1));

  /* "Wrapped_Frozen.pxi":36
 *         self.pvt_o = o
 *         self.frozen = bool(frozen)
 *         if oldstyle_class is None:             # <<<<<<<<<<<<<<
 *             self.oldstyle_class = False
 *         else:
 */
  __pyx_t_1 = (__pyx_v_oldstyle_class == Py_None);
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":37
 *         self.frozen = bool(frozen)
 *         if oldstyle_class is None:
 *             self.oldstyle_class = False             # <<<<<<<<<<<<<<
 *         else:
 *             self.oldstyle_class = oldstyle_class
 */
    __pyx_v_self->oldstyle_class = 0;

    /* "Wrapped_Frozen.pxi":36
 *         self.pvt_o = o
 *         self.frozen = bool(frozen)
 *         if oldstyle_class is None:             # <<<<<<<<<<<<<<
 *             self.oldstyle_class = False
 *         else:
 */
    goto __pyx_L4;
  }

  /* "Wrapped_Frozen.pxi":39
 *             self.oldstyle_class = False
 *         else:
 *             self.oldstyle_class = oldstyle_class             # <<<<<<<<<<<<<<
 *         if PY2:
 *             # In PY2 old-style classes don't have __class__ attribute !
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_oldstyle_class); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 39, __pyx_L1_error)
    __pyx_v_self->oldstyle_class = __pyx_t_1;
  }
  __pyx_L4:;

  /* "Wrapped_Frozen.pxi":40
 *         else:
 *             self.oldstyle_class = oldstyle_class
 *         if PY2:             # <<<<<<<<<<<<<<
 *             # In PY2 old-style classes don't have __class__ attribute !
 *             # When such CLASSES (not INSTANCES of such classes) are wrapped
 */
  if (__pyx_v_9pyprotect_9protected_PY2) {

    /* "Wrapped_Frozen.pxi":48
 *             # CCC is a REGEX, otherwise # CCC is self.cn
 *             # Instances of such classes CAN be wrapped normally.
 *             if hasattr(o, '__class__'):             # <<<<<<<<<<<<<<
 *                 if type(o) is type:
 *                     if self.cn is None:
 */
    __pyx_t_1 = __Pyx_HasAttr(__pyx_v_o, __pyx_n_s_class); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 48, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "Wrapped_Frozen.pxi":49
 *             # Instances of such classes CAN be wrapped normally.
 *             if hasattr(o, '__class__'):
 *                 if type(o) is type:             # <<<<<<<<<<<<<<
 *                     if self.cn is None:
 *                         self.cn = o.__name__
 */
      __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_o)) == ((PyObject *)(&PyType_Type)));
      if (__pyx_t_1) {

        /* "Wrapped_Frozen.pxi":50
 *             if hasattr(o, '__class__'):
 *                 if type(o) is type:
 *                     if self.cn is None:             # <<<<<<<<<<<<<<
 *                         self.cn = o.__name__
 *                 else:
 */
        __pyx_t_1 = (__pyx_v_self->cn == ((PyObject*)Py_None));
        if (__pyx_t_1) {

          /* "Wrapped_Frozen.pxi":51
 *                 if type(o) is type:
 *                     if self.cn is None:
 *                         self.cn = o.__name__             # <<<<<<<<<<<<<<
 *                 else:
 *                     if self.cn is None:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_o, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 51, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(3, 51, __pyx_L1_error)
          __Pyx_GIVEREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_v_self->cn);
          __Pyx_DECREF(__pyx_v_self->cn);
          __pyx_v_self->cn = ((PyObject*)__pyx_t_2);
          __pyx_t_2 = 0;

          /* "Wrapped_Frozen.pxi":50
 *             if hasattr(o, '__class__'):
 *                 if type(o) is type:
 *                     if self.cn is None:             # <<<<<<<<<<<<<<
 *                         self.cn = o.__name__
 *                 else:
 */
        }

        /* "Wrapped_Frozen.pxi":49
 *             # Instances of such classes CAN be wrapped normally.
 *             if hasattr(o, '__class__'):
 *                 if type(o) is type:             # <<<<<<<<<<<<<<
 *                     if self.cn is None:
 *                         self.cn = o.__name__
 */
        goto __pyx_L7;
      }

      /* "Wrapped_Frozen.pxi":53
 *                         self.cn = o.__name__
 *                 else:
 *                     if self.cn is None:             # <<<<<<<<<<<<<<
 *                         self.cn = str(o.__class__.__name__)
 *             else:
 */
      /*else*/ {
        __pyx_t_1 = (__pyx_v_self->cn == ((PyObject*)Py_None));
        if (__pyx_t_1) {

          /* "Wrapped_Frozen.pxi":54
 *                 else:
 *                     if self.cn is None:
 *                         self.cn = str(o.__class__.__name__)             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.cn is None:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_o, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 54, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 54, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Str(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 54, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (!(likely(PyString_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(3, 54, __pyx_L1_error)
          __Pyx_GIVEREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_v_self->cn);
          __Pyx_DECREF(__pyx_v_self->cn);
          __pyx_v_self->cn = ((PyObject*)__pyx_t_2);
          __pyx_t_2 = 0;

          /* "Wrapped_Frozen.pxi":53
 *                         self.cn = o.__name__
 *                 else:
 *                     if self.cn is None:             # <<<<<<<<<<<<<<
 *                         self.cn = str(o.__class__.__name__)
 *             else:
 */
        }
      }
      __pyx_L7:;

      /* "Wrapped_Frozen.pxi":48
 *             # CCC is a REGEX, otherwise # CCC is self.cn
 *             # Instances of such classes CAN be wrapped normally.
 *             if hasattr(o, '__class__'):             # <<<<<<<<<<<<<<
 *                 if type(o) is type:
 *                     if self.cn is None:
 */
      goto __pyx_L6;
    }

    /* "Wrapped_Frozen.pxi":56
 *                         self.cn = str(o.__class__.__name__)
 *             else:
 *                 if self.cn is None:             # <<<<<<<<<<<<<<
 *                     self.cn = 'Unknown_OldStyle_Class'
 *                 if oldstyle_class is None:
 */
    /*else*/ {
      __pyx_t_1 = (__pyx_v_self->cn == ((PyObject*)Py_None));
      if (__pyx_t_1) {

        /* "Wrapped_Frozen.pxi":57
 *             else:
 *                 if self.cn is None:
 *                     self.cn = 'Unknown_OldStyle_Class'             # <<<<<<<<<<<<<<
 *                 if oldstyle_class is None:
 *                     self.oldstyle_class = True
 */
        __Pyx_INCREF(__pyx_n_s_Unknown_OldStyle_Class);
        __Pyx_GIVEREF(__pyx_n_s_Unknown_OldStyle_Class);
        __Pyx_GOTREF(__pyx_v_self->cn);
        __Pyx_DECREF(__pyx_v_self->cn);
        __pyx_v_self->cn = __pyx_n_s_Unknown_OldStyle_Class;

        /* "Wrapped_Frozen.pxi":56
 *                         self.cn = str(o.__class__.__name__)
 *             else:
 *                 if self.cn is None:             # <<<<<<<<<<<<<<
 *                     self.cn = 'Unknown_OldStyle_Class'
 *                 if oldstyle_class is None:
 */
      }

      /* "Wrapped_Frozen.pxi":58
 *                 if self.cn is None:
 *                     self.cn = 'Unknown_OldStyle_Class'
 *                 if oldstyle_class is None:             # <<<<<<<<<<<<<<
 *                     self.oldstyle_class = True
 *         else:
 */
      __pyx_t_1 = (__pyx_v_oldstyle_class == Py_None);
      if (__pyx_t_1) {

        /* "Wrapped_Frozen.pxi":59
 *                     self.cn = 'Unknown_OldStyle_Class'
 *                 if oldstyle_class is None:
 *                     self.oldstyle_class = True             # <<<<<<<<<<<<<<
 *         else:
 *             if type(o) is type:
 */
        __pyx_v_self->oldstyle_class = 1;

        /* "Wrapped_Frozen.pxi":58
 *                 if self.cn is None:
 *                     self.cn = 'Unknown_OldStyle_Class'
 *                 if oldstyle_class is None:             # <<<<<<<<<<<<<<
 *                     self.oldstyle_class = True
 *         else:
 */
      }
    }
    __pyx_L6:;

    /* "Wrapped_Frozen.pxi":40
 *         else:
 *             self.oldstyle_class = oldstyle_class
 *         if PY2:             # <<<<<<<<<<<<<<
 *             # In PY2 old-style classes don't have __class__ attribute !
 *             # When such CLASSES (not INSTANCES of such classes) are wrapped
 */
    goto __pyx_L5;
  }

  /* "Wrapped_Frozen.pxi":61
 *                     self.oldstyle_class = True
 *         else:
 *             if type(o) is type:             # <<<<<<<<<<<<<<
 *                 if self.cn is None:
 *                     self.cn = o.__name__
 */
  /*else*/ {
    __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_o)) == ((PyObject *)(&PyType_Type)));
    if (__pyx_t_1) {

      /* "Wrapped_Frozen.pxi":62
 *         else:
 *             if type(o) is type:
 *                 if self.cn is None:             # <<<<<<<<<<<<<<
 *                     self.cn = o.__name__
 *             else:
 */
      __pyx_t_1 = (__pyx_v_self->cn == ((PyObject*)Py_None));
      if (__pyx_t_1) {

        /* "Wrapped_Frozen.pxi":63
 *             if type(o) is type:
 *                 if self.cn is None:
 *                     self.cn = o.__name__             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.cn is None:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_o, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 63, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(3, 63, __pyx_L1_error)
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_self->cn);
        __Pyx_DECREF(__pyx_v_self->cn);
        __pyx_v_self->cn = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "Wrapped_Frozen.pxi":62
 *         else:
 *             if type(o) is type:
 *                 if self.cn is None:             # <<<<<<<<<<<<<<
 *                     self.cn = o.__name__
 *             else:
 */
      }

      /* "Wrapped_Frozen.pxi":61
 *                     self.oldstyle_class = True
 *         else:
 *             if type(o) is type:             # <<<<<<<<<<<<<<
 *                 if self.cn is None:
 *                     self.cn = o.__name__
 */
      goto __pyx_L12;
    }

    /* "Wrapped_Frozen.pxi":65
 *                     self.cn = o.__name__
 *             else:
 *                 if self.cn is None:             # <<<<<<<<<<<<<<
 *                     self.cn = str(o.__class__.__name__)
 * 
 */
    /*else*/ {
      __pyx_t_1 = (__pyx_v_self->cn == ((PyObject*)Py_None));
      if (__pyx_t_1) {

        /* "Wrapped_Frozen.pxi":66
 *             else:
 *                 if self.cn is None:
 *                     self.cn = str(o.__class__.__name__)             # <<<<<<<<<<<<<<
 * 
 *         # self.hidden_private_attr is set in Wrapped.__init__ but
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_o, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 66, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 66, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_Str(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 66, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyString_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(3, 66, __pyx_L1_error)
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_self->cn);
        __Pyx_DECREF(__pyx_v_self->cn);
        __pyx_v_self->cn = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "Wrapped_Frozen.pxi":65
 *                     self.cn = o.__name__
 *             else:
 *                 if self.cn is None:             # <<<<<<<<<<<<<<
 *                     self.cn = str(o.__class__.__name__)
 * 
 */
      }
    }
    __pyx_L12:;
  }
  __pyx_L5:;

  /* "Wrapped_Frozen.pxi":70
 *         # self.hidden_private_attr is set in Wrapped.__init__ but
 *         # only used in Private and descendants
 *         if self.oldstyle_class:             # <<<<<<<<<<<<<<
 *             self.hidden_private_attr = re.compile(
 *                 mangled_private_attr_regex_fmt % (
 */
  if (__pyx_v_self->oldstyle_class) {

    /* "Wrapped_Frozen.pxi":71
 *         # only used in Private and descendants
 *         if self.oldstyle_class:
 *             self.hidden_private_attr = re.compile(             # <<<<<<<<<<<<<<
 *                 mangled_private_attr_regex_fmt % (
 *                     mangled_private_attr_classname_regex,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compile); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Wrapped_Frozen.pxi":73
 *             self.hidden_private_attr = re.compile(
 *                 mangled_private_attr_regex_fmt % (
 *                     mangled_private_attr_classname_regex,             # <<<<<<<<<<<<<<
 *                 )
 *             )
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_9pyprotect_9protected_mangled_private_attr_classname_regex);
    __Pyx_GIVEREF(__pyx_v_9pyprotect_9protected_mangled_private_attr_classname_regex);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_9pyprotect_9protected_mangled_private_attr_classname_regex);

    /* "Wrapped_Frozen.pxi":72
 *         if self.oldstyle_class:
 *             self.hidden_private_attr = re.compile(
 *                 mangled_private_attr_regex_fmt % (             # <<<<<<<<<<<<<<
 *                     mangled_private_attr_classname_regex,
 *                 )
 */
    __pyx_t_5 = PyNumber_Remainder(__pyx_v_9pyprotect_9protected_mangled_private_attr_regex_fmt, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "Wrapped_Frozen.pxi":71
 *         # only used in Private and descendants
 *         if self.oldstyle_class:
 *             self.hidden_private_attr = re.compile(             # <<<<<<<<<<<<<<
 *                 mangled_private_attr_regex_fmt % (
 *                     mangled_private_attr_classname_regex,
 */
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->hidden_private_attr);
    __Pyx_DECREF(__pyx_v_self->hidden_private_attr);
    __pyx_v_self->hidden_private_attr = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Wrapped_Frozen.pxi":70
 *         # self.hidden_private_attr is set in Wrapped.__init__ but
 *         # only used in Private and descendants
 *         if self.oldstyle_class:             # <<<<<<<<<<<<<<
 *             self.hidden_private_attr = re.compile(
 *                 mangled_private_attr_regex_fmt % (
 */
    goto __pyx_L15;
  }

  /* "Wrapped_Frozen.pxi":77
 *             )
 *         else:
 *             self.hidden_private_attr = re.compile(             # <<<<<<<<<<<<<<
 *                 mangled_private_attr_regex_fmt % (
 *                     self.cn,
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_re); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_compile); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Wrapped_Frozen.pxi":79
 *             self.hidden_private_attr = re.compile(
 *                 mangled_private_attr_regex_fmt % (
 *                     self.cn,             # <<<<<<<<<<<<<<
 *                 )
 *             )
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_self->cn);
    __Pyx_GIVEREF(__pyx_v_self->cn);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->cn);

    /* "Wrapped_Frozen.pxi":78
 *         else:
 *             self.hidden_private_attr = re.compile(
 *                 mangled_private_attr_regex_fmt % (             # <<<<<<<<<<<<<<
 *                     self.cn,
 *                 )
 */
    __pyx_t_3 = PyNumber_Remainder(__pyx_v_9pyprotect_9protected_mangled_private_attr_regex_fmt, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 77, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }

    /* "Wrapped_Frozen.pxi":77
 *             )
 *         else:
 *             self.hidden_private_attr = re.compile(             # <<<<<<<<<<<<<<
 *                 mangled_private_attr_regex_fmt % (
 *                     self.cn,
 */
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->hidden_private_attr);
    __Pyx_DECREF(__pyx_v_self->hidden_private_attr);
    __pyx_v_self->hidden_private_attr = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L15:;

  /* "Wrapped_Frozen.pxi":84
 * 
 *         # Special code to avoid double-wrapping of Protected
 *         if rules is None:             # <<<<<<<<<<<<<<
 *             rules = dict(self.get_rules())
 * 
 */
  __pyx_t_1 = (__pyx_v_rules == Py_None);
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":85
 *         # Special code to avoid double-wrapping of Protected
 *         if rules is None:
 *             rules = dict(self.get_rules())             # <<<<<<<<<<<<<<
 * 
 *         if self.frozen:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->get_rules(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_rules, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Wrapped_Frozen.pxi":84
 * 
 *         # Special code to avoid double-wrapping of Protected
 *         if rules is None:             # <<<<<<<<<<<<<<
 *             rules = dict(self.get_rules())
 * 
 */
  }

  /* "Wrapped_Frozen.pxi":87
 *             rules = dict(self.get_rules())
 * 
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             protect_class = FrozenProtected
 *             private_class = FrozenPrivate
 */
  if (__pyx_v_self->__pyx_base.frozen) {

    /* "Wrapped_Frozen.pxi":88
 * 
 *         if self.frozen:
 *             protect_class = FrozenProtected             # <<<<<<<<<<<<<<
 *             private_class = FrozenPrivate
 *         else:
 */
    __Pyx_INCREF((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenProtected);
    __pyx_v_protect_class = __pyx_ptype_9pyprotect_9protected_FrozenProtected;

    /* "Wrapped_Frozen.pxi":89
 *         if self.frozen:
 *             protect_class = FrozenProtected
 *             private_class = FrozenPrivate             # <<<<<<<<<<<<<<
 *         else:
 *             protect_class = Protected
 */
    __Pyx_INCREF((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivate);
    __pyx_v_private_class = __pyx_ptype_9pyprotect_9protected_FrozenPrivate;

    /* "Wrapped_Frozen.pxi":87
 *             rules = dict(self.get_rules())
 * 
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             protect_class = FrozenProtected
 *             private_class = FrozenPrivate
 */
    goto __pyx_L17;
  }

  /* "Wrapped_Frozen.pxi":91
 *             private_class = FrozenPrivate
 *         else:
 *             protect_class = Protected             # <<<<<<<<<<<<<<
 *             private_class = Private
 *         if isinstance(self.pvt_o, type):
 */
  /*else*/ {
    __Pyx_INCREF((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected);
    __pyx_v_protect_class = __pyx_ptype_9pyprotect_9protected_Protected;

    /* "Wrapped_Frozen.pxi":92
 *         else:
 *             protect_class = Protected
 *             private_class = Private             # <<<<<<<<<<<<<<
 *         if isinstance(self.pvt_o, type):
 *             id_class = id(self.pvt_o)
 */
    __Pyx_INCREF((PyObject *)__pyx_ptype_9pyprotect_9protected_Private);
    __pyx_v_private_class = __pyx_ptype_9pyprotect_9protected_Private;
  }
  __pyx_L17:;

  /* "Wrapped_Frozen.pxi":93
 *             protect_class = Protected
 *             private_class = Private
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             id_class = id(self.pvt_o)
 *         else:
 */
  __pyx_t_5 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_1 = PyType_Check(__pyx_t_5); 
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":94
 *             private_class = Private
 *         if isinstance(self.pvt_o, type):
 *             id_class = id(self.pvt_o)             # <<<<<<<<<<<<<<
 *         else:
 *             id_class = id(type(self.pvt_o))
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_self->__pyx_base.pvt_o); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_id_class = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Wrapped_Frozen.pxi":93
 *             protect_class = Protected
 *             private_class = Private
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             id_class = id(self.pvt_o)
 *         else:
 */
    goto __pyx_L18;
  }

  /* "Wrapped_Frozen.pxi":96
 *             id_class = id(self.pvt_o)
 *         else:
 *             id_class = id(type(self.pvt_o))             # <<<<<<<<<<<<<<
 * 
 *         self.protected_attribute = __ProtectionData(
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)Py_TYPE(__pyx_v_self->__pyx_base.pvt_o))); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_id_class = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __pyx_L18:;

  /* "Wrapped_Frozen.pxi":99
 * 
 *         self.protected_attribute = __ProtectionData(
 *             id_val=id(self.pvt_o),             # <<<<<<<<<<<<<<
 *             id_class=id_class,
 *             hash_val=__HiddenPartial(self.hash_protected, self),
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(15); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_self->__pyx_base.pvt_o); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_id_val, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":100
 *         self.protected_attribute = __ProtectionData(
 *             id_val=id(self.pvt_o),
 *             id_class=id_class,             # <<<<<<<<<<<<<<
 *             hash_val=__HiddenPartial(self.hash_protected, self),
 *             isinstance_val=__HiddenPartial(self.isinstance_protected, self),
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_id_class, __pyx_v_id_class) < 0) __PYX_ERR(3, 99, __pyx_L1_error)

  /* "Wrapped_Frozen.pxi":101
 *             id_val=id(self.pvt_o),
 *             id_class=id_class,
 *             hash_val=__HiddenPartial(self.hash_protected, self),             # <<<<<<<<<<<<<<
 *             isinstance_val=__HiddenPartial(self.isinstance_protected, self),
 *             issubclass_val=__HiddenPartial(self.issubclass_protected, self),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->hash_protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_hash_val, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":102
 *             id_class=id_class,
 *             hash_val=__HiddenPartial(self.hash_protected, self),
 *             isinstance_val=__HiddenPartial(self.isinstance_protected, self),             # <<<<<<<<<<<<<<
 *             issubclass_val=__HiddenPartial(self.issubclass_protected, self),
 *             instanceof=__HiddenPartial(self.instanceof_protected, self),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->isinstance_protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_isinstance_val, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":103
 *             hash_val=__HiddenPartial(self.hash_protected, self),
 *             isinstance_val=__HiddenPartial(self.isinstance_protected, self),
 *             issubclass_val=__HiddenPartial(self.issubclass_protected, self),             # <<<<<<<<<<<<<<
 *             instanceof=__HiddenPartial(self.instanceof_protected, self),
 *             subclassof=__HiddenPartial(self.subclassof_protected, self),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->issubclass_protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_issubclass_val, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":104
 *             isinstance_val=__HiddenPartial(self.isinstance_protected, self),
 *             issubclass_val=__HiddenPartial(self.issubclass_protected, self),
 *             instanceof=__HiddenPartial(self.instanceof_protected, self),             # <<<<<<<<<<<<<<
 *             subclassof=__HiddenPartial(self.subclassof_protected, self),
 *             help_val=__HiddenPartial(self.help_protected, self),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->instanceof_protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_instanceof, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":105
 *             issubclass_val=__HiddenPartial(self.issubclass_protected, self),
 *             instanceof=__HiddenPartial(self.instanceof_protected, self),
 *             subclassof=__HiddenPartial(self.subclassof_protected, self),             # <<<<<<<<<<<<<<
 *             help_val=__HiddenPartial(self.help_protected, self),
 *             help_str=__HiddenPartial(self.help_str_protected, self),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->subclassof_protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_subclassof, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":106
 *             instanceof=__HiddenPartial(self.instanceof_protected, self),
 *             subclassof=__HiddenPartial(self.subclassof_protected, self),
 *             help_val=__HiddenPartial(self.help_protected, self),             # <<<<<<<<<<<<<<
 *             help_str=__HiddenPartial(self.help_str_protected, self),
 *             testop=__HiddenPartial(self.testop, self),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->help_protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_help_val, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":107
 *             subclassof=__HiddenPartial(self.subclassof_protected, self),
 *             help_val=__HiddenPartial(self.help_protected, self),
 *             help_str=__HiddenPartial(self.help_str_protected, self),             # <<<<<<<<<<<<<<
 *             testop=__HiddenPartial(self.testop, self),
 *             rules=rules,
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->help_str_protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_help_str, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":108
 *             help_val=__HiddenPartial(self.help_protected, self),
 *             help_str=__HiddenPartial(self.help_str_protected, self),
 *             testop=__HiddenPartial(self.testop, self),             # <<<<<<<<<<<<<<
 *             rules=rules,
 *             freeze=__HiddenPartial(self.freeze, self),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->testop); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_testop, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":109
 *             help_str=__HiddenPartial(self.help_str_protected, self),
 *             testop=__HiddenPartial(self.testop, self),
 *             rules=rules,             # <<<<<<<<<<<<<<
 *             freeze=__HiddenPartial(self.freeze, self),
 *             private=__HiddenPartial(private_class, self.pvt_o),
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_rules, __pyx_v_rules) < 0) __PYX_ERR(3, 99, __pyx_L1_error)

  /* "Wrapped_Frozen.pxi":110
 *             testop=__HiddenPartial(self.testop, self),
 *             rules=rules,
 *             freeze=__HiddenPartial(self.freeze, self),             # <<<<<<<<<<<<<<
 *             private=__HiddenPartial(private_class, self.pvt_o),
 *             protect=__HiddenPartial(protect_class, self.pvt_o),
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->freeze); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_freeze, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":111
 *             rules=rules,
 *             freeze=__HiddenPartial(self.freeze, self),
 *             private=__HiddenPartial(private_class, self.pvt_o),             # <<<<<<<<<<<<<<
 *             protect=__HiddenPartial(protect_class, self.pvt_o),
 *             multiwrapped=__HiddenPartial(self.multiwrapped, self),
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_private_class);
  __Pyx_GIVEREF((PyObject *)__pyx_v_private_class);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_private_class));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->__pyx_base.pvt_o);
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_private, __pyx_t_3) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Wrapped_Frozen.pxi":112
 *             freeze=__HiddenPartial(self.freeze, self),
 *             private=__HiddenPartial(private_class, self.pvt_o),
 *             protect=__HiddenPartial(protect_class, self.pvt_o),             # <<<<<<<<<<<<<<
 *             multiwrapped=__HiddenPartial(self.multiwrapped, self),
 *         )
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_protect_class);
  __Pyx_GIVEREF((PyObject *)__pyx_v_protect_class);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_protect_class));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->__pyx_base.pvt_o);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_protect, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":113
 *             private=__HiddenPartial(private_class, self.pvt_o),
 *             protect=__HiddenPartial(protect_class, self.pvt_o),
 *             multiwrapped=__HiddenPartial(self.multiwrapped, self),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_2 = __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->multiwrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_multiwrapped, __pyx_t_2) < 0) __PYX_ERR(3, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":98
 *             id_class = id(type(self.pvt_o))
 * 
 *         self.protected_attribute = __ProtectionData(             # <<<<<<<<<<<<<<
 *             id_val=id(self.pvt_o),
 *             id_class=id_class,
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___ProtectionData), __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v_self->protected_attribute);
  __pyx_v_self->protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":23
 *     cdef object hidden_private_attr
 * 
 *     def __init__(self, o, frozen=False, rules=None, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *         '''
 *         o: object to be wrapped
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_protect_class);
  __Pyx_XDECREF((PyObject *)__pyx_v_private_class);
  __Pyx_XDECREF(__pyx_v_id_class);
  __Pyx_XDECREF(__pyx_v_rules);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":120
 *     # --------------------------------------------------------------------
 * 
 *     cdef attr_hidden(self, attr):             # <<<<<<<<<<<<<<
 *         '''
 *         Central place where we decide if an attribute or key in a
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_attr_hidden(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("attr_hidden", 0);

  /* "Wrapped_Frozen.pxi":125
 *         PrivacyDict is hidden
 *         '''
 *         if unmangled_private_attr.match(attr):             # <<<<<<<<<<<<<<
 *             return True
 *         if self.hidden_private_attr.match(attr):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_unmangled_private_attr, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_attr};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(3, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Wrapped_Frozen.pxi":126
 *         '''
 *         if unmangled_private_attr.match(attr):
 *             return True             # <<<<<<<<<<<<<<
 *         if self.hidden_private_attr.match(attr):
 *             return True
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":125
 *         PrivacyDict is hidden
 *         '''
 *         if unmangled_private_attr.match(attr):             # <<<<<<<<<<<<<<
 *             return True
 *         if self.hidden_private_attr.match(attr):
 */
  }

  /* "Wrapped_Frozen.pxi":127
 *         if unmangled_private_attr.match(attr):
 *             return True
 *         if self.hidden_private_attr.match(attr):             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->hidden_private_attr, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_attr};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(3, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Wrapped_Frozen.pxi":128
 *             return True
 *         if self.hidden_private_attr.match(attr):
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cdef fif(self, o):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":127
 *         if unmangled_private_attr.match(attr):
 *             return True
 *         if self.hidden_private_attr.match(attr):             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
  }

  /* "Wrapped_Frozen.pxi":120
 *     # --------------------------------------------------------------------
 * 
 *     cdef attr_hidden(self, attr):             # <<<<<<<<<<<<<<
 *         '''
 *         Central place where we decide if an attribute or key in a
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.attr_hidden", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":130
 *             return True
 * 
 *     cdef fif(self, o):             # <<<<<<<<<<<<<<
 *         '''
 *         fif = Freeze If Frozen
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_fif(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fif", 0);

  /* "Wrapped_Frozen.pxi":136
 *         Returns-->o or Frozen(o)
 *         '''
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             return freeze(o)
 *         return o
 */
  if (__pyx_v_self->__pyx_base.frozen) {

    /* "Wrapped_Frozen.pxi":137
 *         '''
 *         if self.frozen:
 *             return freeze(o)             # <<<<<<<<<<<<<<
 *         return o
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_freeze); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_o};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":136
 *         Returns-->o or Frozen(o)
 *         '''
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             return freeze(o)
 *         return o
 */
  }

  /* "Wrapped_Frozen.pxi":138
 *         if self.frozen:
 *             return freeze(o)
 *         return o             # <<<<<<<<<<<<<<
 * 
 *     cdef freeze(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_o);
  __pyx_r = __pyx_v_o;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":130
 *             return True
 * 
 *     cdef fif(self, o):             # <<<<<<<<<<<<<<
 *         '''
 *         fif = Freeze If Frozen
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.fif", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":140
 *         return o
 * 
 *     cdef freeze(self):             # <<<<<<<<<<<<<<
 *         '''Smartly avoid double wrapping when freezing a Wrapped object'''
 *         if self.frozen:
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_freeze(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("freeze", 0);

  /* "Wrapped_Frozen.pxi":142
 *     cdef freeze(self):
 *         '''Smartly avoid double wrapping when freezing a Wrapped object'''
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             return self
 *         if isinstance(self, Protected):
 */
  if (__pyx_v_self->__pyx_base.frozen) {

    /* "Wrapped_Frozen.pxi":143
 *         '''Smartly avoid double wrapping when freezing a Wrapped object'''
 *         if self.frozen:
 *             return self             # <<<<<<<<<<<<<<
 *         if isinstance(self, Protected):
 *             d = {}
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":142
 *     cdef freeze(self):
 *         '''Smartly avoid double wrapping when freezing a Wrapped object'''
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             return self
 *         if isinstance(self, Protected):
 */
  }

  /* "Wrapped_Frozen.pxi":144
 *         if self.frozen:
 *             return self
 *         if isinstance(self, Protected):             # <<<<<<<<<<<<<<
 *             d = {}
 *             d.update(self.rules)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Protected); 
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":145
 *             return self
 *         if isinstance(self, Protected):
 *             d = {}             # <<<<<<<<<<<<<<
 *             d.update(self.rules)
 *             d['frozen'] = True
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_d = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Wrapped_Frozen.pxi":146
 *         if isinstance(self, Protected):
 *             d = {}
 *             d.update(self.rules)             # <<<<<<<<<<<<<<
 *             d['frozen'] = True
 *             return FrozenProtected(self.pvt_o, d)
 */
    __pyx_t_2 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_d, __pyx_v_self->rules); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Wrapped_Frozen.pxi":147
 *             d = {}
 *             d.update(self.rules)
 *             d['frozen'] = True             # <<<<<<<<<<<<<<
 *             return FrozenProtected(self.pvt_o, d)
 *         elif isinstance(self, Private):
 */
    if (unlikely((PyDict_SetItem(__pyx_v_d, __pyx_n_s_frozen, Py_True) < 0))) __PYX_ERR(3, 147, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":148
 *             d.update(self.rules)
 *             d['frozen'] = True
 *             return FrozenProtected(self.pvt_o, d)             # <<<<<<<<<<<<<<
 *         elif isinstance(self, Private):
 *             return FrozenPrivate(self.pvt_o)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.pvt_o);
    __Pyx_GIVEREF(__pyx_v_self->__pyx_base.pvt_o);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->__pyx_base.pvt_o);
    __Pyx_INCREF(__pyx_v_d);
    __Pyx_GIVEREF(__pyx_v_d);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_d);
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenProtected), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":144
 *         if self.frozen:
 *             return self
 *         if isinstance(self, Protected):             # <<<<<<<<<<<<<<
 *             d = {}
 *             d.update(self.rules)
 */
  }

  /* "Wrapped_Frozen.pxi":149
 *             d['frozen'] = True
 *             return FrozenProtected(self.pvt_o, d)
 *         elif isinstance(self, Private):             # <<<<<<<<<<<<<<
 *             return FrozenPrivate(self.pvt_o)
 *         if isinstance(self, PrivacyDict):
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Private); 
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":150
 *             return FrozenProtected(self.pvt_o, d)
 *         elif isinstance(self, Private):
 *             return FrozenPrivate(self.pvt_o)             # <<<<<<<<<<<<<<
 *         if isinstance(self, PrivacyDict):
 *             return FrozenPrivacyDict(self.pvt_o, cn=self.cn)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivate), __pyx_v_self->__pyx_base.pvt_o); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":149
 *             d['frozen'] = True
 *             return FrozenProtected(self.pvt_o, d)
 *         elif isinstance(self, Private):             # <<<<<<<<<<<<<<
 *             return FrozenPrivate(self.pvt_o)
 *         if isinstance(self, PrivacyDict):
 */
  }

  /* "Wrapped_Frozen.pxi":151
 *         elif isinstance(self, Private):
 *             return FrozenPrivate(self.pvt_o)
 *         if isinstance(self, PrivacyDict):             # <<<<<<<<<<<<<<
 *             return FrozenPrivacyDict(self.pvt_o, cn=self.cn)
 *         else:
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_PrivacyDict); 
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":152
 *             return FrozenPrivate(self.pvt_o)
 *         if isinstance(self, PrivacyDict):
 *             return FrozenPrivacyDict(self.pvt_o, cn=self.cn)             # <<<<<<<<<<<<<<
 *         else:
 *             return Frozen(self.pvt_o)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.pvt_o);
    __Pyx_GIVEREF(__pyx_v_self->__pyx_base.pvt_o);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.pvt_o);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_cn, __pyx_v_self->cn) < 0) __PYX_ERR(3, 152, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict), __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":151
 *         elif isinstance(self, Private):
 *             return FrozenPrivate(self.pvt_o)
 *         if isinstance(self, PrivacyDict):             # <<<<<<<<<<<<<<
 *             return FrozenPrivacyDict(self.pvt_o, cn=self.cn)
 *         else:
 */
  }

  /* "Wrapped_Frozen.pxi":154
 *             return FrozenPrivacyDict(self.pvt_o, cn=self.cn)
 *         else:
 *             return Frozen(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     cdef multiwrapped(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9pyprotect_9protected_Frozen), __pyx_v_self->__pyx_base.pvt_o); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "Wrapped_Frozen.pxi":140
 *         return o
 * 
 *     cdef freeze(self):             # <<<<<<<<<<<<<<
 *         '''Smartly avoid double wrapping when freezing a Wrapped object'''
 *         if self.frozen:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.freeze", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":156
 *             return Frozen(self.pvt_o)
 * 
 *     cdef multiwrapped(self):             # <<<<<<<<<<<<<<
 *         '''For testing'''
 *         return isinstance(self.pvt_o, Wrapped)
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_multiwrapped(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("multiwrapped", 0);

  /* "Wrapped_Frozen.pxi":158
 *     cdef multiwrapped(self):
 *         '''For testing'''
 *         return isinstance(self.pvt_o, Wrapped)             # <<<<<<<<<<<<<<
 * 
 *     cdef id_protected(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_t_1, __pyx_ptype_9pyprotect_9protected_Wrapped); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":156
 *             return Frozen(self.pvt_o)
 * 
 *     cdef multiwrapped(self):             # <<<<<<<<<<<<<<
 *         '''For testing'''
 *         return isinstance(self.pvt_o, Wrapped)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.multiwrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":160
 *         return isinstance(self.pvt_o, Wrapped)
 * 
 *     cdef id_protected(self):             # <<<<<<<<<<<<<<
 *         return id(self.pvt_o)
 * 
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_id_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("id_protected", 0);

  /* "Wrapped_Frozen.pxi":161
 * 
 *     cdef id_protected(self):
 *         return id(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     cdef hash_protected(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_self->__pyx_base.pvt_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":160
 *         return isinstance(self.pvt_o, Wrapped)
 * 
 *     cdef id_protected(self):             # <<<<<<<<<<<<<<
 *         return id(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.id_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":163
 *         return id(self.pvt_o)
 * 
 *     cdef hash_protected(self):             # <<<<<<<<<<<<<<
 *         return hash(self.pvt_o)
 * 
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_hash_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hash_protected", 0);

  /* "Wrapped_Frozen.pxi":164
 * 
 *     cdef hash_protected(self):
 *         return hash(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     cdef isinstance_protected(self, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(3, 164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":163
 *         return id(self.pvt_o)
 * 
 *     cdef hash_protected(self):             # <<<<<<<<<<<<<<
 *         return hash(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.hash_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":166
 *         return hash(self.pvt_o)
 * 
 *     cdef isinstance_protected(self, c):             # <<<<<<<<<<<<<<
 *         return isinstance(self.pvt_o, c)
 * 
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_isinstance_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isinstance_protected", 0);

  /* "Wrapped_Frozen.pxi":167
 * 
 *     cdef isinstance_protected(self, c):
 *         return isinstance(self.pvt_o, c)             # <<<<<<<<<<<<<<
 * 
 *     cdef issubclass_protected(self, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_t_1, __pyx_v_c); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":166
 *         return hash(self.pvt_o)
 * 
 *     cdef isinstance_protected(self, c):             # <<<<<<<<<<<<<<
 *         return isinstance(self.pvt_o, c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.isinstance_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":169
 *         return isinstance(self.pvt_o, c)
 * 
 *     cdef issubclass_protected(self, c):             # <<<<<<<<<<<<<<
 *         if isinstance(self.pvt_o, type):
 *             return issubclass(self.pvt_o, c)
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_issubclass_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("issubclass_protected", 0);

  /* "Wrapped_Frozen.pxi":170
 * 
 *     cdef issubclass_protected(self, c):
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             return issubclass(self.pvt_o, c)
 *         else:
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyType_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Wrapped_Frozen.pxi":171
 *     cdef issubclass_protected(self, c):
 *         if isinstance(self.pvt_o, type):
 *             return issubclass(self.pvt_o, c)             # <<<<<<<<<<<<<<
 *         else:
 *             return issubclass(type(self.pvt_o), c)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_IsSubclass(__pyx_t_1, __pyx_v_c); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 171, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":170
 * 
 *     cdef issubclass_protected(self, c):
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             return issubclass(self.pvt_o, c)
 *         else:
 */
  }

  /* "Wrapped_Frozen.pxi":173
 *             return issubclass(self.pvt_o, c)
 *         else:
 *             return issubclass(type(self.pvt_o), c)             # <<<<<<<<<<<<<<
 * 
 *     cdef instanceof_protected(self, c):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyObject_IsSubclass(((PyObject *)Py_TYPE(__pyx_v_self->__pyx_base.pvt_o)), __pyx_v_c); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 173, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "Wrapped_Frozen.pxi":169
 *         return isinstance(self.pvt_o, c)
 * 
 *     cdef issubclass_protected(self, c):             # <<<<<<<<<<<<<<
 *         if isinstance(self.pvt_o, type):
 *             return issubclass(self.pvt_o, c)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.issubclass_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":175
 *             return issubclass(type(self.pvt_o), c)
 * 
 *     cdef instanceof_protected(self, c):             # <<<<<<<<<<<<<<
 *         if isinstance(self.pvt_o, type):
 *             return isinstance(c, self.pvt_o)
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_instanceof_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("instanceof_protected", 0);

  /* "Wrapped_Frozen.pxi":176
 * 
 *     cdef instanceof_protected(self, c):
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             return isinstance(c, self.pvt_o)
 *         else:
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyType_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Wrapped_Frozen.pxi":177
 *     cdef instanceof_protected(self, c):
 *         if isinstance(self.pvt_o, type):
 *             return isinstance(c, self.pvt_o)             # <<<<<<<<<<<<<<
 *         else:
 *             return isinstance(c, type(self.pvt_o))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_c, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 177, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":176
 * 
 *     cdef instanceof_protected(self, c):
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             return isinstance(c, self.pvt_o)
 *         else:
 */
  }

  /* "Wrapped_Frozen.pxi":179
 *             return isinstance(c, self.pvt_o)
 *         else:
 *             return isinstance(c, type(self.pvt_o))             # <<<<<<<<<<<<<<
 * 
 *     cdef subclassof_protected(self, c):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_c, ((PyObject *)Py_TYPE(__pyx_v_self->__pyx_base.pvt_o))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 179, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "Wrapped_Frozen.pxi":175
 *             return issubclass(type(self.pvt_o), c)
 * 
 *     cdef instanceof_protected(self, c):             # <<<<<<<<<<<<<<
 *         if isinstance(self.pvt_o, type):
 *             return isinstance(c, self.pvt_o)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.instanceof_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":181
 *             return isinstance(c, type(self.pvt_o))
 * 
 *     cdef subclassof_protected(self, c):             # <<<<<<<<<<<<<<
 *         if isinstance(self.pvt_o, type):
 *             return issubclass(c, self.pvt_o)
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_subclassof_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subclassof_protected", 0);

  /* "Wrapped_Frozen.pxi":182
 * 
 *     cdef subclassof_protected(self, c):
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             return issubclass(c, self.pvt_o)
 *         else:
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyType_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Wrapped_Frozen.pxi":183
 *     cdef subclassof_protected(self, c):
 *         if isinstance(self.pvt_o, type):
 *             return issubclass(c, self.pvt_o)             # <<<<<<<<<<<<<<
 *         else:
 *             return issubclass(c, type(self.pvt_o))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_IsSubclass(__pyx_v_c, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":182
 * 
 *     cdef subclassof_protected(self, c):
 *         if isinstance(self.pvt_o, type):             # <<<<<<<<<<<<<<
 *             return issubclass(c, self.pvt_o)
 *         else:
 */
  }

  /* "Wrapped_Frozen.pxi":185
 *             return issubclass(c, self.pvt_o)
 *         else:
 *             return issubclass(c, type(self.pvt_o))             # <<<<<<<<<<<<<<
 * 
 *     cdef help_protected(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyObject_IsSubclass(__pyx_v_c, ((PyObject *)Py_TYPE(__pyx_v_self->__pyx_base.pvt_o))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 185, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "Wrapped_Frozen.pxi":181
 *             return isinstance(c, type(self.pvt_o))
 * 
 *     cdef subclassof_protected(self, c):             # <<<<<<<<<<<<<<
 *         if isinstance(self.pvt_o, type):
 *             return issubclass(c, self.pvt_o)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.subclassof_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":187
 *             return issubclass(c, type(self.pvt_o))
 * 
 *     cdef help_protected(self):             # <<<<<<<<<<<<<<
 *         return help(self.pvt_o)
 * 
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_help_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("help_protected", 0);

  /* "Wrapped_Frozen.pxi":188
 * 
 *     cdef help_protected(self):
 *         return help(self.pvt_o)             # <<<<<<<<<<<<<<
 * 
 *     cdef help_str_protected(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_help, __pyx_v_self->__pyx_base.pvt_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":187
 *             return issubclass(c, type(self.pvt_o))
 * 
 *     cdef help_protected(self):             # <<<<<<<<<<<<<<
 *         return help(self.pvt_o)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.help_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":190
 *         return help(self.pvt_o)
 * 
 *     cdef help_str_protected(self):             # <<<<<<<<<<<<<<
 *         return '\n'.join(
 *             pydoc.render_doc(self.pvt_o).splitlines()[2:]
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_help_str_protected(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("help_str_protected", 0);

  /* "Wrapped_Frozen.pxi":191
 * 
 *     cdef help_str_protected(self):
 *         return '\n'.join(             # <<<<<<<<<<<<<<
 *             pydoc.render_doc(self.pvt_o).splitlines()[2:]
 *         ).rstrip('\n') + '\n'
 */
  __Pyx_XDECREF(__pyx_r);

  /* "Wrapped_Frozen.pxi":192
 *     cdef help_str_protected(self):
 *         return '\n'.join(
 *             pydoc.render_doc(self.pvt_o).splitlines()[2:]             # <<<<<<<<<<<<<<
 *         ).rstrip('\n') + '\n'
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pydoc); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_render_doc); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_self->__pyx_base.pvt_o};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_splitlines); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_2, 2, 0, NULL, NULL, &__pyx_slice__18, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":191
 * 
 *     cdef help_str_protected(self):
 *         return '\n'.join(             # <<<<<<<<<<<<<<
 *             pydoc.render_doc(self.pvt_o).splitlines()[2:]
 *         ).rstrip('\n') + '\n'
 */
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__17, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Wrapped_Frozen.pxi":193
 *         return '\n'.join(
 *             pydoc.render_doc(self.pvt_o).splitlines()[2:]
 *         ).rstrip('\n') + '\n'             # <<<<<<<<<<<<<<
 * 
 *     cdef visible(self, a):
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_rstrip); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_s__17};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__17); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":190
 *         return help(self.pvt_o)
 * 
 *     cdef help_str_protected(self):             # <<<<<<<<<<<<<<
 *         return '\n'.join(
 *             pydoc.render_doc(self.pvt_o).splitlines()[2:]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.help_str_protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":195
 *         ).rstrip('\n') + '\n'
 * 
 *     cdef visible(self, a):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_visible(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("visible", 0);

  /* "Wrapped_Frozen.pxi":196
 * 
 *     cdef visible(self, a):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef writeable(self, a):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":195
 *         ).rstrip('\n') + '\n'
 * 
 *     cdef visible(self, a):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":198
 *         return True
 * 
 *     cdef writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         if a in special_attributes or a in overridden_always:
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_writeable(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeable", 0);

  /* "Wrapped_Frozen.pxi":200
 *     cdef writeable(self, a):
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         if a in special_attributes or a in overridden_always:             # <<<<<<<<<<<<<<
 *             return False
 *         return not self.frozen
 */
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_special_attributes, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 200, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_overridden_always, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 200, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":201
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         if a in special_attributes or a in overridden_always:
 *             return False             # <<<<<<<<<<<<<<
 *         return not self.frozen
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":200
 *     cdef writeable(self, a):
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         if a in special_attributes or a in overridden_always:             # <<<<<<<<<<<<<<
 *             return False
 *         return not self.frozen
 */
  }

  /* "Wrapped_Frozen.pxi":202
 *         if a in special_attributes or a in overridden_always:
 *             return False
 *         return not self.frozen             # <<<<<<<<<<<<<<
 * 
 *     cdef testop(self, a, op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBool_FromLong((!__pyx_v_self->__pyx_base.frozen)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":198
 *         return True
 * 
 *     cdef writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         if a in special_attributes or a in overridden_always:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.writeable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":204
 *         return not self.frozen
 * 
 *     cdef testop(self, a, op):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_testop(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("testop", 0);

  /* "Wrapped_Frozen.pxi":210
 *         Returns-->bool
 *         '''
 *         if op == 'r':             # <<<<<<<<<<<<<<
 *             return hasattr(self, a)
 *         elif op == 'w':
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_r, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 210, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":211
 *         '''
 *         if op == 'r':
 *             return hasattr(self, a)             # <<<<<<<<<<<<<<
 *         elif op == 'w':
 *             if not self.writeable(a):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_HasAttr(((PyObject *)__pyx_v_self), __pyx_v_a); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 211, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":210
 *         Returns-->bool
 *         '''
 *         if op == 'r':             # <<<<<<<<<<<<<<
 *             return hasattr(self, a)
 *         elif op == 'w':
 */
  }

  /* "Wrapped_Frozen.pxi":212
 *         if op == 'r':
 *             return hasattr(self, a)
 *         elif op == 'w':             # <<<<<<<<<<<<<<
 *             if not self.writeable(a):
 *                 return False
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_w, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 212, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":213
 *             return hasattr(self, a)
 *         elif op == 'w':
 *             if not self.writeable(a):             # <<<<<<<<<<<<<<
 *                 return False
 *             return not self.frozen
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->writeable(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (!__pyx_t_1);
    if (__pyx_t_3) {

      /* "Wrapped_Frozen.pxi":214
 *         elif op == 'w':
 *             if not self.writeable(a):
 *                 return False             # <<<<<<<<<<<<<<
 *             return not self.frozen
 *         return False
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "Wrapped_Frozen.pxi":213
 *             return hasattr(self, a)
 *         elif op == 'w':
 *             if not self.writeable(a):             # <<<<<<<<<<<<<<
 *                 return False
 *             return not self.frozen
 */
    }

    /* "Wrapped_Frozen.pxi":215
 *             if not self.writeable(a):
 *                 return False
 *             return not self.frozen             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBool_FromLong((!__pyx_v_self->__pyx_base.frozen)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":212
 *         if op == 'r':
 *             return hasattr(self, a)
 *         elif op == 'w':             # <<<<<<<<<<<<<<
 *             if not self.writeable(a):
 *                 return False
 */
  }

  /* "Wrapped_Frozen.pxi":216
 *                 return False
 *             return not self.frozen
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef get_rules(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":204
 *         return not self.frozen
 * 
 *     cdef testop(self, a, op):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.testop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":218
 *         return False
 * 
 *     cdef get_rules(self):             # <<<<<<<<<<<<<<
 *         return dict()
 * 
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_get_rules(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_rules", 0);

  /* "Wrapped_Frozen.pxi":219
 * 
 *     cdef get_rules(self):
 *         return dict()             # <<<<<<<<<<<<<<
 * 
 *     cdef comparator(self, other, op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":218
 *         return False
 * 
 *     cdef get_rules(self):             # <<<<<<<<<<<<<<
 *         return dict()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.get_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":249
 * 
 *         '''
 *         def pass_to_wrapped():             # <<<<<<<<<<<<<<
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_10comparator_1pass_to_wrapped(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_7Wrapped_10comparator_pass_to_wrapped, "Trap RecursionError if object is too deeply nested");
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7Wrapped_10comparator_1pass_to_wrapped = {"pass_to_wrapped", (PyCFunction)__pyx_pw_9pyprotect_9protected_7Wrapped_10comparator_1pass_to_wrapped, METH_NOARGS, __pyx_doc_9pyprotect_9protected_7Wrapped_10comparator_pass_to_wrapped};
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_10comparator_1pass_to_wrapped(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pass_to_wrapped (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_10comparator_pass_to_wrapped(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_10comparator_pass_to_wrapped(PyObject *__pyx_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *__pyx_cur_scope;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pass_to_wrapped", 0);
  __pyx_outer_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Wrapped_Frozen.pxi":251
 *         def pass_to_wrapped():
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o < other
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_op)) { __Pyx_RaiseClosureNameError("op"); __PYX_ERR(3, 251, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_LT); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(3, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Wrapped_Frozen.pxi":252
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o < other
 *                 except RecursionError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":253
 *             if op == Py_LT:
 *                 try:
 *                     return self.pvt_o < other             # <<<<<<<<<<<<<<
 *                 except RecursionError:
 *                     return NotImplemented
 */
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(3, 253, __pyx_L4_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(3, 253, __pyx_L4_error) }
        __pyx_t_2 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_other, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 253, __pyx_L4_error)
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        goto __pyx_L8_try_return;

        /* "Wrapped_Frozen.pxi":252
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o < other
 *                 except RecursionError:
 */
      }
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Wrapped_Frozen.pxi":254
 *                 try:
 *                     return self.pvt_o < other
 *                 except RecursionError:             # <<<<<<<<<<<<<<
 *                     return NotImplemented
 *             elif op == Py_EQ:
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_1, &__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 254, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_1, __pyx_t_7);
      __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_7 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator.pass_to_wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(3, 254, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);

        /* "Wrapped_Frozen.pxi":255
 *                     return self.pvt_o < other
 *                 except RecursionError:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 *             elif op == Py_EQ:
 *                 try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_builtin_NotImplemented);
        __pyx_r = __pyx_builtin_NotImplemented;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "Wrapped_Frozen.pxi":252
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o < other
 *                 except RecursionError:
 */
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L8_try_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
    }

    /* "Wrapped_Frozen.pxi":251
 *         def pass_to_wrapped():
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o < other
 */
  }

  /* "Wrapped_Frozen.pxi":256
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_EQ:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o == other
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_op)) { __Pyx_RaiseClosureNameError("op"); __PYX_ERR(3, 256, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(3, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Wrapped_Frozen.pxi":257
 *                     return NotImplemented
 *             elif op == Py_EQ:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o == other
 *                 except RecursionError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":258
 *             elif op == Py_EQ:
 *                 try:
 *                     return self.pvt_o == other             # <<<<<<<<<<<<<<
 *                 except RecursionError:
 *                     return NotImplemented
 */
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(3, 258, __pyx_L12_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(3, 258, __pyx_L12_error) }
        __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 258, __pyx_L12_error)
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L16_try_return;

        /* "Wrapped_Frozen.pxi":257
 *                     return NotImplemented
 *             elif op == Py_EQ:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o == other
 *                 except RecursionError:
 */
      }
      __pyx_L12_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Wrapped_Frozen.pxi":259
 *                 try:
 *                     return self.pvt_o == other
 *                 except RecursionError:             # <<<<<<<<<<<<<<
 *                     return NotImplemented
 *             elif op == Py_GT:
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_2, &__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 259, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_2, __pyx_t_7);
      __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_7 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator.pass_to_wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(3, 259, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);

        /* "Wrapped_Frozen.pxi":260
 *                     return self.pvt_o == other
 *                 except RecursionError:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 *             elif op == Py_GT:
 *                 try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_builtin_NotImplemented);
        __pyx_r = __pyx_builtin_NotImplemented;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L15_except_return;
      }
      goto __pyx_L14_except_error;
      __pyx_L14_except_error:;

      /* "Wrapped_Frozen.pxi":257
 *                     return NotImplemented
 *             elif op == Py_EQ:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o == other
 *                 except RecursionError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L16_try_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L0;
      __pyx_L15_except_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L0;
    }

    /* "Wrapped_Frozen.pxi":256
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_EQ:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o == other
 */
  }

  /* "Wrapped_Frozen.pxi":261
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_GT:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o > other
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_op)) { __Pyx_RaiseClosureNameError("op"); __PYX_ERR(3, 261, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_GT); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(3, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Wrapped_Frozen.pxi":262
 *                     return NotImplemented
 *             elif op == Py_GT:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o > other
 *                 except RecursionError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":263
 *             elif op == Py_GT:
 *                 try:
 *                     return self.pvt_o > other             # <<<<<<<<<<<<<<
 *                 except RecursionError:
 *                     return NotImplemented
 */
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(3, 263, __pyx_L20_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(3, 263, __pyx_L20_error) }
        __pyx_t_2 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_other, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 263, __pyx_L20_error)
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        goto __pyx_L24_try_return;

        /* "Wrapped_Frozen.pxi":262
 *                     return NotImplemented
 *             elif op == Py_GT:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o > other
 *                 except RecursionError:
 */
      }
      __pyx_L20_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Wrapped_Frozen.pxi":264
 *                 try:
 *                     return self.pvt_o > other
 *                 except RecursionError:             # <<<<<<<<<<<<<<
 *                     return NotImplemented
 *             elif op == Py_LE:
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_1, &__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 264, __pyx_L22_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_1, __pyx_t_7);
      __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_7 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator.pass_to_wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(3, 264, __pyx_L22_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);

        /* "Wrapped_Frozen.pxi":265
 *                     return self.pvt_o > other
 *                 except RecursionError:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 *             elif op == Py_LE:
 *                 try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_builtin_NotImplemented);
        __pyx_r = __pyx_builtin_NotImplemented;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L23_except_return;
      }
      goto __pyx_L22_except_error;
      __pyx_L22_except_error:;

      /* "Wrapped_Frozen.pxi":262
 *                     return NotImplemented
 *             elif op == Py_GT:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o > other
 *                 except RecursionError:
 */
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L24_try_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L23_except_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
    }

    /* "Wrapped_Frozen.pxi":261
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_GT:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o > other
 */
  }

  /* "Wrapped_Frozen.pxi":266
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_LE:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o <= other
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_op)) { __Pyx_RaiseClosureNameError("op"); __PYX_ERR(3, 266, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_LE); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(3, 266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Wrapped_Frozen.pxi":267
 *                     return NotImplemented
 *             elif op == Py_LE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o <= other
 *                 except RecursionError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":268
 *             elif op == Py_LE:
 *                 try:
 *                     return self.pvt_o <= other             # <<<<<<<<<<<<<<
 *                 except RecursionError:
 *                     return NotImplemented
 */
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(3, 268, __pyx_L28_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(3, 268, __pyx_L28_error) }
        __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_other, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 268, __pyx_L28_error)
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L32_try_return;

        /* "Wrapped_Frozen.pxi":267
 *                     return NotImplemented
 *             elif op == Py_LE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o <= other
 *                 except RecursionError:
 */
      }
      __pyx_L28_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Wrapped_Frozen.pxi":269
 *                 try:
 *                     return self.pvt_o <= other
 *                 except RecursionError:             # <<<<<<<<<<<<<<
 *                     return NotImplemented
 *             elif op == Py_NE:
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_2, &__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 269, __pyx_L30_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_2, __pyx_t_7);
      __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_7 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator.pass_to_wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(3, 269, __pyx_L30_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);

        /* "Wrapped_Frozen.pxi":270
 *                     return self.pvt_o <= other
 *                 except RecursionError:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 *             elif op == Py_NE:
 *                 try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_builtin_NotImplemented);
        __pyx_r = __pyx_builtin_NotImplemented;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L31_except_return;
      }
      goto __pyx_L30_except_error;
      __pyx_L30_except_error:;

      /* "Wrapped_Frozen.pxi":267
 *                     return NotImplemented
 *             elif op == Py_LE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o <= other
 *                 except RecursionError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L32_try_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L0;
      __pyx_L31_except_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L0;
    }

    /* "Wrapped_Frozen.pxi":266
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_LE:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o <= other
 */
  }

  /* "Wrapped_Frozen.pxi":271
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_NE:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o != other
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_op)) { __Pyx_RaiseClosureNameError("op"); __PYX_ERR(3, 271, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(3, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Wrapped_Frozen.pxi":272
 *                     return NotImplemented
 *             elif op == Py_NE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o != other
 *                 except RecursionError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":273
 *             elif op == Py_NE:
 *                 try:
 *                     return self.pvt_o != other             # <<<<<<<<<<<<<<
 *                 except RecursionError:
 *                     return NotImplemented
 */
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(3, 273, __pyx_L36_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(3, 273, __pyx_L36_error) }
        __pyx_t_2 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_other, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 273, __pyx_L36_error)
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        goto __pyx_L40_try_return;

        /* "Wrapped_Frozen.pxi":272
 *                     return NotImplemented
 *             elif op == Py_NE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o != other
 *                 except RecursionError:
 */
      }
      __pyx_L36_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Wrapped_Frozen.pxi":274
 *                 try:
 *                     return self.pvt_o != other
 *                 except RecursionError:             # <<<<<<<<<<<<<<
 *                     return NotImplemented
 *             elif op == Py_GE:
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_1, &__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 274, __pyx_L38_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_1, __pyx_t_7);
      __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_7 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator.pass_to_wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(3, 274, __pyx_L38_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);

        /* "Wrapped_Frozen.pxi":275
 *                     return self.pvt_o != other
 *                 except RecursionError:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 *             elif op == Py_GE:
 *                 try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_builtin_NotImplemented);
        __pyx_r = __pyx_builtin_NotImplemented;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L39_except_return;
      }
      goto __pyx_L38_except_error;
      __pyx_L38_except_error:;

      /* "Wrapped_Frozen.pxi":272
 *                     return NotImplemented
 *             elif op == Py_NE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o != other
 *                 except RecursionError:
 */
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L40_try_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L39_except_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
    }

    /* "Wrapped_Frozen.pxi":271
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_NE:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o != other
 */
  }

  /* "Wrapped_Frozen.pxi":276
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_GE:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o >= other
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_op)) { __Pyx_RaiseClosureNameError("op"); __PYX_ERR(3, 276, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_GE); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(3, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Wrapped_Frozen.pxi":277
 *                     return NotImplemented
 *             elif op == Py_GE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o >= other
 *                 except RecursionError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":278
 *             elif op == Py_GE:
 *                 try:
 *                     return self.pvt_o >= other             # <<<<<<<<<<<<<<
 *                 except RecursionError:
 *                     return NotImplemented
 */
        __Pyx_XDECREF(__pyx_r);
        if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(3, 278, __pyx_L44_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(3, 278, __pyx_L44_error) }
        __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_self->__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_other, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 278, __pyx_L44_error)
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L48_try_return;

        /* "Wrapped_Frozen.pxi":277
 *                     return NotImplemented
 *             elif op == Py_GE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o >= other
 *                 except RecursionError:
 */
      }
      __pyx_L44_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Wrapped_Frozen.pxi":279
 *                 try:
 *                     return self.pvt_o >= other
 *                 except RecursionError:             # <<<<<<<<<<<<<<
 *                     return NotImplemented
 *             else:
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_2, &__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_RecursionError); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 279, __pyx_L46_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_2, __pyx_t_7);
      __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_7 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator.pass_to_wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(3, 279, __pyx_L46_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);

        /* "Wrapped_Frozen.pxi":280
 *                     return self.pvt_o >= other
 *                 except RecursionError:
 *                     return NotImplemented             # <<<<<<<<<<<<<<
 *             else:
 *                 return NotImplemented
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_builtin_NotImplemented);
        __pyx_r = __pyx_builtin_NotImplemented;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L47_except_return;
      }
      goto __pyx_L46_except_error;
      __pyx_L46_except_error:;

      /* "Wrapped_Frozen.pxi":277
 *                     return NotImplemented
 *             elif op == Py_GE:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.pvt_o >= other
 *                 except RecursionError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L48_try_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L0;
      __pyx_L47_except_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L0;
    }

    /* "Wrapped_Frozen.pxi":276
 *                 except RecursionError:
 *                     return NotImplemented
 *             elif op == Py_GE:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.pvt_o >= other
 */
  }

  /* "Wrapped_Frozen.pxi":282
 *                     return NotImplemented
 *             else:
 *                 return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *         if not iswrapped(other):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;
  }

  /* "Wrapped_Frozen.pxi":249
 * 
 *         '''
 *         def pass_to_wrapped():             # <<<<<<<<<<<<<<
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator.pass_to_wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":221
 *         return dict()
 * 
 *     cdef comparator(self, other, op):             # <<<<<<<<<<<<<<
 *         '''
 *         Operations:
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_comparator(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *__pyx_cur_scope;
  PyObject *__pyx_v_pass_to_wrapped = 0;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_d1 = NULL;
  PyObject *__pyx_v_d2 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("comparator", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_1_comparator(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(3, 221, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_other = __pyx_v_other;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_other);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_other);
  __pyx_cur_scope->__pyx_v_op = __pyx_v_op;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_op);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_op);

  /* "Wrapped_Frozen.pxi":249
 * 
 *         '''
 *         def pass_to_wrapped():             # <<<<<<<<<<<<<<
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7Wrapped_10comparator_1pass_to_wrapped, 0, __pyx_n_s_Wrapped_comparator_locals_pass_t, ((PyObject*)__pyx_cur_scope), __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pass_to_wrapped = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":284
 *                 return NotImplemented
 * 
 *         if not iswrapped(other):             # <<<<<<<<<<<<<<
 *             return pass_to_wrapped()
 *         # If we got here, other is Wrapped
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iswrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_cur_scope->__pyx_v_other};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(3, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "Wrapped_Frozen.pxi":285
 * 
 *         if not iswrapped(other):
 *             return pass_to_wrapped()             # <<<<<<<<<<<<<<
 *         # If we got here, other is Wrapped
 *         # Only equality / inequality are supported. Neither object
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_pf_9pyprotect_9protected_7Wrapped_10comparator_pass_to_wrapped(__pyx_v_pass_to_wrapped); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":284
 *                 return NotImplemented
 * 
 *         if not iswrapped(other):             # <<<<<<<<<<<<<<
 *             return pass_to_wrapped()
 *         # If we got here, other is Wrapped
 */
  }

  /* "Wrapped_Frozen.pxi":289
 *         # Only equality / inequality are supported. Neither object
 *         # can access object wrapped by the other for other comparisons.
 *         if op not in (Py_NE, Py_EQ):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         res = (
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_op);
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_op;
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(3, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_6 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(3, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_t_5;
  __pyx_L5_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __pyx_t_6;
  if (__pyx_t_5) {

    /* "Wrapped_Frozen.pxi":290
 *         # can access object wrapped by the other for other comparisons.
 *         if op not in (Py_NE, Py_EQ):
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         res = (
 *             type(self) == type(other) and
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":289
 *         # Only equality / inequality are supported. Neither object
 *         # can access object wrapped by the other for other comparisons.
 *         if op not in (Py_NE, Py_EQ):             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         res = (
 */
  }

  /* "Wrapped_Frozen.pxi":292
 *             return NotImplemented
 *         res = (
 *             type(self) == type(other) and             # <<<<<<<<<<<<<<
 *             id_protected(self) == id_protected(other)
 *         )
 */
  __pyx_t_2 = PyObject_RichCompare(((PyObject *)Py_TYPE(((PyObject *)__pyx_cur_scope->__pyx_v_self))), ((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_other)), Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 292, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(3, 292, __pyx_L1_error)
  if (__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L7_bool_binop_done;
  }

  /* "Wrapped_Frozen.pxi":293
 *         res = (
 *             type(self) == type(other) and
 *             id_protected(self) == id_protected(other)             # <<<<<<<<<<<<<<
 *         )
 *         if not isprotected(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_id_protected); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_id_protected); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_cur_scope->__pyx_v_other};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_1 = __pyx_t_7;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_L7_bool_binop_done:;
  __pyx_v_res = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":295
 *             id_protected(self) == id_protected(other)
 *         )
 *         if not isprotected(self):             # <<<<<<<<<<<<<<
 *             if op == Py_EQ:
 *                 return res
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_isprotected); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(3, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "Wrapped_Frozen.pxi":296
 *         )
 *         if not isprotected(self):
 *             if op == Py_EQ:             # <<<<<<<<<<<<<<
 *                 return res
 *             elif op == Py_NE:
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 296, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 296, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_6) {

      /* "Wrapped_Frozen.pxi":297
 *         if not isprotected(self):
 *             if op == Py_EQ:
 *                 return res             # <<<<<<<<<<<<<<
 *             elif op == Py_NE:
 *                 return not res
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_res);
      __pyx_r = __pyx_v_res;
      goto __pyx_L0;

      /* "Wrapped_Frozen.pxi":296
 *         )
 *         if not isprotected(self):
 *             if op == Py_EQ:             # <<<<<<<<<<<<<<
 *                 return res
 *             elif op == Py_NE:
 */
    }

    /* "Wrapped_Frozen.pxi":298
 *             if op == Py_EQ:
 *                 return res
 *             elif op == Py_NE:             # <<<<<<<<<<<<<<
 *                 return not res
 *         else:
 */
    __pyx_t_7 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_6) {

      /* "Wrapped_Frozen.pxi":299
 *                 return res
 *             elif op == Py_NE:
 *                 return not res             # <<<<<<<<<<<<<<
 *         else:
 *             # Protected
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 299, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_6)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "Wrapped_Frozen.pxi":298
 *             if op == Py_EQ:
 *                 return res
 *             elif op == Py_NE:             # <<<<<<<<<<<<<<
 *                 return not res
 *         else:
 */
    }

    /* "Wrapped_Frozen.pxi":295
 *             id_protected(self) == id_protected(other)
 *         )
 *         if not isprotected(self):             # <<<<<<<<<<<<<<
 *             if op == Py_EQ:
 *                 return res
 */
    goto __pyx_L9;
  }

  /* "Wrapped_Frozen.pxi":302
 *         else:
 *             # Protected
 *             d1 = dict(self.rules)             # <<<<<<<<<<<<<<
 *             d2 = dict(getattr(other, PROT_ATTR_NAME).rules)
 *             # Don't consider kwargs, just the rules
 */
  /*else*/ {
    if (unlikely(__pyx_cur_scope->__pyx_v_self->rules == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
      __PYX_ERR(3, 302, __pyx_L1_error)
    }
    __pyx_t_1 = PyDict_Copy(__pyx_cur_scope->__pyx_v_self->rules); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_d1 = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Wrapped_Frozen.pxi":303
 *             # Protected
 *             d1 = dict(self.rules)
 *             d2 = dict(getattr(other, PROT_ATTR_NAME).rules)             # <<<<<<<<<<<<<<
 *             # Don't consider kwargs, just the rules
 *             try:
 */
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_other;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_GetAttr(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_rules); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_d2 = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "Wrapped_Frozen.pxi":305
 *             d2 = dict(getattr(other, PROT_ATTR_NAME).rules)
 *             # Don't consider kwargs, just the rules
 *             try:             # <<<<<<<<<<<<<<
 *                 del d1['kwargs']
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":306
 *             # Don't consider kwargs, just the rules
 *             try:
 *                 del d1['kwargs']             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 pass
 */
        if (unlikely((PyDict_DelItem(__pyx_v_d1, __pyx_n_s_kwargs) < 0))) __PYX_ERR(3, 306, __pyx_L11_error)

        /* "Wrapped_Frozen.pxi":305
 *             d2 = dict(getattr(other, PROT_ATTR_NAME).rules)
 *             # Don't consider kwargs, just the rules
 *             try:             # <<<<<<<<<<<<<<
 *                 del d1['kwargs']
 *             except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L16_try_end;
      __pyx_L11_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Wrapped_Frozen.pxi":307
 *             try:
 *                 del d1['kwargs']
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 pass
 *             try:
 */
      __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_4) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L12_exception_handled;
      }
      goto __pyx_L13_except_error;
      __pyx_L13_except_error:;

      /* "Wrapped_Frozen.pxi":305
 *             d2 = dict(getattr(other, PROT_ATTR_NAME).rules)
 *             # Don't consider kwargs, just the rules
 *             try:             # <<<<<<<<<<<<<<
 *                 del d1['kwargs']
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L12_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_L16_try_end:;
    }

    /* "Wrapped_Frozen.pxi":309
 *             except KeyError:
 *                 pass
 *             try:             # <<<<<<<<<<<<<<
 *                 del d2['kwargs']
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "Wrapped_Frozen.pxi":310
 *                 pass
 *             try:
 *                 del d2['kwargs']             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 pass
 */
        if (unlikely((PyDict_DelItem(__pyx_v_d2, __pyx_n_s_kwargs) < 0))) __PYX_ERR(3, 310, __pyx_L17_error)

        /* "Wrapped_Frozen.pxi":309
 *             except KeyError:
 *                 pass
 *             try:             # <<<<<<<<<<<<<<
 *                 del d2['kwargs']
 *             except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L22_try_end;
      __pyx_L17_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Wrapped_Frozen.pxi":311
 *             try:
 *                 del d2['kwargs']
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 pass
 *             res = res and d1 == d2
 */
      __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_4) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L18_exception_handled;
      }
      goto __pyx_L19_except_error;
      __pyx_L19_except_error:;

      /* "Wrapped_Frozen.pxi":309
 *             except KeyError:
 *                 pass
 *             try:             # <<<<<<<<<<<<<<
 *                 del d2['kwargs']
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L18_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      __pyx_L22_try_end:;
    }

    /* "Wrapped_Frozen.pxi":313
 *             except KeyError:
 *                 pass
 *             res = res and d1 == d2             # <<<<<<<<<<<<<<
 *             if op == Py_EQ:
 *                 return res
 */
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 313, __pyx_L1_error)
    if (__pyx_t_6) {
    } else {
      __Pyx_INCREF(__pyx_v_res);
      __pyx_t_3 = __pyx_v_res;
      goto __pyx_L23_bool_binop_done;
    }
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_d1, __pyx_v_d2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 313, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_3 = __pyx_t_7;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_L23_bool_binop_done:;
    __Pyx_DECREF_SET(__pyx_v_res, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Wrapped_Frozen.pxi":314
 *                 pass
 *             res = res and d1 == d2
 *             if op == Py_EQ:             # <<<<<<<<<<<<<<
 *                 return res
 *             elif op == Py_NE:
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 314, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 314, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_6) {

      /* "Wrapped_Frozen.pxi":315
 *             res = res and d1 == d2
 *             if op == Py_EQ:
 *                 return res             # <<<<<<<<<<<<<<
 *             elif op == Py_NE:
 *                 return not res
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_res);
      __pyx_r = __pyx_v_res;
      goto __pyx_L0;

      /* "Wrapped_Frozen.pxi":314
 *                 pass
 *             res = res and d1 == d2
 *             if op == Py_EQ:             # <<<<<<<<<<<<<<
 *                 return res
 *             elif op == Py_NE:
 */
    }

    /* "Wrapped_Frozen.pxi":316
 *             if op == Py_EQ:
 *                 return res
 *             elif op == Py_NE:             # <<<<<<<<<<<<<<
 *                 return not res
 * 
 */
    __pyx_t_7 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_op, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_6) {

      /* "Wrapped_Frozen.pxi":317
 *                 return res
 *             elif op == Py_NE:
 *                 return not res             # <<<<<<<<<<<<<<
 * 
 *     cdef wrapped_getattr(self, a):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 317, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyBool_FromLong((!__pyx_t_6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "Wrapped_Frozen.pxi":316
 *             if op == Py_EQ:
 *                 return res
 *             elif op == Py_NE:             # <<<<<<<<<<<<<<
 *                 return not res
 * 
 */
    }
  }
  __pyx_L9:;

  /* "Wrapped_Frozen.pxi":221
 *         return dict()
 * 
 *     cdef comparator(self, other, op):             # <<<<<<<<<<<<<<
 *         '''
 *         Operations:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.comparator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pass_to_wrapped);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_d1);
  __Pyx_XDECREF(__pyx_v_d2);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":319
 *                 return not res
 * 
 *     cdef wrapped_getattr(self, a):             # <<<<<<<<<<<<<<
 *         # protected_attribute
 *         if a == PROT_ATTR_NAME:
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_getattr(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_v_delegated = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrapped_getattr", 0);

  /* "Wrapped_Frozen.pxi":321
 *     cdef wrapped_getattr(self, a):
 *         # protected_attribute
 *         if a == PROT_ATTR_NAME:             # <<<<<<<<<<<<<<
 *             return self.protected_attribute
 *         if a in overridden_always:
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_a, __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 321, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":322
 *         # protected_attribute
 *         if a == PROT_ATTR_NAME:
 *             return self.protected_attribute             # <<<<<<<<<<<<<<
 *         if a in overridden_always:
 *             return __HiddenPartial(getattr(Wrapped, a), self)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self->protected_attribute);
    __pyx_r = ((PyObject *)__pyx_v_self->protected_attribute);
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":321
 *     cdef wrapped_getattr(self, a):
 *         # protected_attribute
 *         if a == PROT_ATTR_NAME:             # <<<<<<<<<<<<<<
 *             return self.protected_attribute
 *         if a in overridden_always:
 */
  }

  /* "Wrapped_Frozen.pxi":323
 *         if a == PROT_ATTR_NAME:
 *             return self.protected_attribute
 *         if a in overridden_always:             # <<<<<<<<<<<<<<
 *             return __HiddenPartial(getattr(Wrapped, a), self)
 * 
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_overridden_always, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 323, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":324
 *             return self.protected_attribute
 *         if a in overridden_always:
 *             return __HiddenPartial(getattr(Wrapped, a), self)             # <<<<<<<<<<<<<<
 * 
 *         # PREVENT pickling - doesn't work even if methods are implemented,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_v_a); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":323
 *         if a == PROT_ATTR_NAME:
 *             return self.protected_attribute
 *         if a in overridden_always:             # <<<<<<<<<<<<<<
 *             return __HiddenPartial(getattr(Wrapped, a), self)
 * 
 */
  }

  /* "Wrapped_Frozen.pxi":327
 * 
 *         # PREVENT pickling - doesn't work even if methods are implemented,
 *         if a in pickle_attributes:             # <<<<<<<<<<<<<<
 *             raise AttributeError('Wrapped object cannot be pickled')
 * 
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_pickle_attributes, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 327, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "Wrapped_Frozen.pxi":328
 *         # PREVENT pickling - doesn't work even if methods are implemented,
 *         if a in pickle_attributes:
 *             raise AttributeError('Wrapped object cannot be pickled')             # <<<<<<<<<<<<<<
 * 
 *         delegated = getattr(self.pvt_o, a, None)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_AttributeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(3, 328, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":327
 * 
 *         # PREVENT pickling - doesn't work even if methods are implemented,
 *         if a in pickle_attributes:             # <<<<<<<<<<<<<<
 *             raise AttributeError('Wrapped object cannot be pickled')
 * 
 */
  }

  /* "Wrapped_Frozen.pxi":330
 *             raise AttributeError('Wrapped object cannot be pickled')
 * 
 *         delegated = getattr(self.pvt_o, a, None)             # <<<<<<<<<<<<<<
 *         if a in always_delegated:
 *             return delegated
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_v_a, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_delegated = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Wrapped_Frozen.pxi":331
 * 
 *         delegated = getattr(self.pvt_o, a, None)
 *         if a in always_delegated:             # <<<<<<<<<<<<<<
 *             return delegated
 * 
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_always_delegated, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 331, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":332
 *         delegated = getattr(self.pvt_o, a, None)
 *         if a in always_delegated:
 *             return delegated             # <<<<<<<<<<<<<<
 * 
 *         # Container mutating methods - implemented and selectively blocked
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_delegated);
    __pyx_r = __pyx_v_delegated;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":331
 * 
 *         delegated = getattr(self.pvt_o, a, None)
 *         if a in always_delegated:             # <<<<<<<<<<<<<<
 *             return delegated
 * 
 */
  }

  /* "Wrapped_Frozen.pxi":335
 * 
 *         # Container mutating methods - implemented and selectively blocked
 *         if a in m_block and hasattr(Wrapped, a):             # <<<<<<<<<<<<<<
 *             return __HiddenPartial(getattr(Wrapped, a), self)
 *         # Any non-method or missing attribute or special callable method
 */
  if (unlikely(__pyx_v_9pyprotect_9protected_m_block == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(3, 335, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PySet_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_m_block, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(3, 335, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_v_a); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 335, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_4;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":336
 *         # Container mutating methods - implemented and selectively blocked
 *         if a in m_block and hasattr(Wrapped, a):
 *             return __HiddenPartial(getattr(Wrapped, a), self)             # <<<<<<<<<<<<<<
 *         # Any non-method or missing attribute or special callable method
 *         # that is not delegated or blocked
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_v_a); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Wrapped_Frozen.pxi":335
 * 
 *         # Container mutating methods - implemented and selectively blocked
 *         if a in m_block and hasattr(Wrapped, a):             # <<<<<<<<<<<<<<
 *             return __HiddenPartial(getattr(Wrapped, a), self)
 *         # Any non-method or missing attribute or special callable method
 */
  }

  /* "Wrapped_Frozen.pxi":339
 *         # Any non-method or missing attribute or special callable method
 *         # that is not delegated or blocked
 *         if delegated is None:             # <<<<<<<<<<<<<<
 *             if delegated in dir(self.pvt_o):
 *                 return delegated
 */
  __pyx_t_1 = (__pyx_v_delegated == Py_None);
  if (__pyx_t_1) {

    /* "Wrapped_Frozen.pxi":340
 *         # that is not delegated or blocked
 *         if delegated is None:
 *             if delegated in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *                 return delegated
 *             raise AttributeError(
 */
    __pyx_t_3 = __pyx_v_self->__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = PyObject_Dir(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_delegated, __pyx_t_2, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 340, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "Wrapped_Frozen.pxi":341
 *         if delegated is None:
 *             if delegated in dir(self.pvt_o):
 *                 return delegated             # <<<<<<<<<<<<<<
 *             raise AttributeError(
 *                 "Object Wrapped('%s') has no attribute '%s'" % (self.cn, a)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_delegated);
      __pyx_r = __pyx_v_delegated;
      goto __pyx_L0;

      /* "Wrapped_Frozen.pxi":340
 *         # that is not delegated or blocked
 *         if delegated is None:
 *             if delegated in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *                 return delegated
 *             raise AttributeError(
 */
    }

    /* "Wrapped_Frozen.pxi":343
 *                 return delegated
 *             raise AttributeError(
 *                 "Object Wrapped('%s') has no attribute '%s'" % (self.cn, a)             # <<<<<<<<<<<<<<
 *             )
 *         # If frozen, freeze all the way down
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_self->cn);
    __Pyx_GIVEREF(__pyx_v_self->cn);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->cn);
    __Pyx_INCREF(__pyx_v_a);
    __Pyx_GIVEREF(__pyx_v_a);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_a);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Object_Wrapped_s_has_no_attribut, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Wrapped_Frozen.pxi":342
 *             if delegated in dir(self.pvt_o):
 *                 return delegated
 *             raise AttributeError(             # <<<<<<<<<<<<<<
 *                 "Object Wrapped('%s') has no attribute '%s'" % (self.cn, a)
 *             )
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(3, 342, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":339
 *         # Any non-method or missing attribute or special callable method
 *         # that is not delegated or blocked
 *         if delegated is None:             # <<<<<<<<<<<<<<
 *             if delegated in dir(self.pvt_o):
 *                 return delegated
 */
  }

  /* "Wrapped_Frozen.pxi":346
 *             )
 *         # If frozen, freeze all the way down
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             # Specifically in case of ModuleType, make resultant
 *             # object behave closest to a python module:
 */
  if (__pyx_v_self->__pyx_base.frozen) {

    /* "Wrapped_Frozen.pxi":352
 *             #   FROM the module by methods, classes are not
 *             # However module's __dict__ is still frozen
 *             if a == '__dict__' or not isinstance(self.pvt_o, types.ModuleType):             # <<<<<<<<<<<<<<
 *                 delegated = freeze(delegated)
 *         return delegated
 */
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_a, __pyx_n_s_dict_2, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(3, 352, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = __pyx_v_self->__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ModuleType); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = PyObject_IsInstance(__pyx_t_2, __pyx_t_5); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 352, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = (!__pyx_t_4);
    __pyx_t_1 = __pyx_t_6;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_1) {

      /* "Wrapped_Frozen.pxi":353
 *             # However module's __dict__ is still frozen
 *             if a == '__dict__' or not isinstance(self.pvt_o, types.ModuleType):
 *                 delegated = freeze(delegated)             # <<<<<<<<<<<<<<
 *         return delegated
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_freeze); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_delegated};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 353, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_delegated, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Wrapped_Frozen.pxi":352
 *             #   FROM the module by methods, classes are not
 *             # However module's __dict__ is still frozen
 *             if a == '__dict__' or not isinstance(self.pvt_o, types.ModuleType):             # <<<<<<<<<<<<<<
 *                 delegated = freeze(delegated)
 *         return delegated
 */
    }

    /* "Wrapped_Frozen.pxi":346
 *             )
 *         # If frozen, freeze all the way down
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             # Specifically in case of ModuleType, make resultant
 *             # object behave closest to a python module:
 */
  }

  /* "Wrapped_Frozen.pxi":354
 *             if a == '__dict__' or not isinstance(self.pvt_o, types.ModuleType):
 *                 delegated = freeze(delegated)
 *         return delegated             # <<<<<<<<<<<<<<
 * 
 *     cdef wrapped_check_setattr(self, a, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_delegated);
  __pyx_r = __pyx_v_delegated;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":319
 *                 return not res
 * 
 *     cdef wrapped_getattr(self, a):             # <<<<<<<<<<<<<<
 *         # protected_attribute
 *         if a == PROT_ATTR_NAME:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.wrapped_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_delegated);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":356
 *         return delegated
 * 
 *     cdef wrapped_check_setattr(self, a, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_check_setattr(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, CYTHON_UNUSED PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrapped_check_setattr", 0);

  /* "Wrapped_Frozen.pxi":357
 * 
 *     cdef wrapped_check_setattr(self, a, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:
 */
  if (unlikely(__pyx_v_self->__pyx_base.frozen)) {

    /* "Wrapped_Frozen.pxi":358
 *     cdef wrapped_check_setattr(self, a, val):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         if a in overridden_always or a in special_attributes:
 *             raise ProtectionError('Cannot modify attribute: %s' % (a,))
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(3, 358, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":357
 * 
 *     cdef wrapped_check_setattr(self, a, val):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:
 */
  }

  /* "Wrapped_Frozen.pxi":359
 *         if self.frozen:
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:             # <<<<<<<<<<<<<<
 *             raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 */
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_overridden_always, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 359, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_special_attributes, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 359, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Wrapped_Frozen.pxi":360
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:
 *             raise ProtectionError('Cannot modify attribute: %s' % (a,))             # <<<<<<<<<<<<<<
 * 
 *     cdef wrapped_check_delattr(self, a):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_a);
    __Pyx_GIVEREF(__pyx_v_a);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_a);
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_modify_attribute_s, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 360, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 360, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":359
 *         if self.frozen:
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:             # <<<<<<<<<<<<<<
 *             raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 */
  }

  /* "Wrapped_Frozen.pxi":356
 *         return delegated
 * 
 *     cdef wrapped_check_setattr(self, a, val):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.wrapped_check_setattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":362
 *             raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 *     cdef wrapped_check_delattr(self, a):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_check_delattr(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrapped_check_delattr", 0);

  /* "Wrapped_Frozen.pxi":363
 * 
 *     cdef wrapped_check_delattr(self, a):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:
 */
  if (unlikely(__pyx_v_self->__pyx_base.frozen)) {

    /* "Wrapped_Frozen.pxi":364
 *     cdef wrapped_check_delattr(self, a):
 *         if self.frozen:
 *             raise frozen_error             # <<<<<<<<<<<<<<
 *         if a in overridden_always or a in special_attributes:
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 */
    __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
    __PYX_ERR(3, 364, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":363
 * 
 *     cdef wrapped_check_delattr(self, a):
 *         if self.frozen:             # <<<<<<<<<<<<<<
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:
 */
  }

  /* "Wrapped_Frozen.pxi":365
 *         if self.frozen:
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:             # <<<<<<<<<<<<<<
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 *         if not hasattr(self.pvt_o, a) and a in self.__dir__():
 */
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_overridden_always, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 365, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_special_attributes, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(3, 365, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Wrapped_Frozen.pxi":366
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))             # <<<<<<<<<<<<<<
 *         if not hasattr(self.pvt_o, a) and a in self.__dir__():
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_a);
    __Pyx_GIVEREF(__pyx_v_a);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_a);
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_delete_attribute_s, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 366, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":365
 *         if self.frozen:
 *             raise frozen_error
 *         if a in overridden_always or a in special_attributes:             # <<<<<<<<<<<<<<
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 *         if not hasattr(self.pvt_o, a) and a in self.__dir__():
 */
  }

  /* "Wrapped_Frozen.pxi":367
 *         if a in overridden_always or a in special_attributes:
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 *         if not hasattr(self.pvt_o, a) and a in self.__dir__():             # <<<<<<<<<<<<<<
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_3, __pyx_v_a); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = (!__pyx_t_2);
  if (__pyx_t_8) {
  } else {
    __pyx_t_1 = __pyx_t_8;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dir); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_6, };
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_t_3, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(3, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_8;
  __pyx_L8_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "Wrapped_Frozen.pxi":368
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 *         if not hasattr(self.pvt_o, a) and a in self.__dir__():
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))             # <<<<<<<<<<<<<<
 * 
 *     cdef wrapped_dir(self):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_a);
    __Pyx_GIVEREF(__pyx_v_a);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_a);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_delete_attribute_s, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 368, __pyx_L1_error)

    /* "Wrapped_Frozen.pxi":367
 *         if a in overridden_always or a in special_attributes:
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 *         if not hasattr(self.pvt_o, a) and a in self.__dir__():             # <<<<<<<<<<<<<<
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 */
  }

  /* "Wrapped_Frozen.pxi":362
 *             raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 *     cdef wrapped_check_delattr(self, a):             # <<<<<<<<<<<<<<
 *         if self.frozen:
 *             raise frozen_error
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.wrapped_check_delattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":370
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 *     cdef wrapped_dir(self):             # <<<<<<<<<<<<<<
 *         res_set = special_attributes
 *         delegated = set(dir(self.pvt_o))
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_dir(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_v_res_set = NULL;
  PyObject *__pyx_v_delegated = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrapped_dir", 0);

  /* "Wrapped_Frozen.pxi":371
 * 
 *     cdef wrapped_dir(self):
 *         res_set = special_attributes             # <<<<<<<<<<<<<<
 *         delegated = set(dir(self.pvt_o))
 *         res_set = res_set.union(delegated)
 */
  __Pyx_INCREF(__pyx_v_9pyprotect_9protected_special_attributes);
  __pyx_v_res_set = __pyx_v_9pyprotect_9protected_special_attributes;

  /* "Wrapped_Frozen.pxi":372
 *     cdef wrapped_dir(self):
 *         res_set = special_attributes
 *         delegated = set(dir(self.pvt_o))             # <<<<<<<<<<<<<<
 *         res_set = res_set.union(delegated)
 *         res_set = res_set.difference(pickle_attributes)
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Dir(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySet_New(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_delegated = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":373
 *         res_set = special_attributes
 *         delegated = set(dir(self.pvt_o))
 *         res_set = res_set.union(delegated)             # <<<<<<<<<<<<<<
 *         res_set = res_set.difference(pickle_attributes)
 *         return list(res_set)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_res_set, __pyx_n_s_union); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_delegated};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_res_set, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":374
 *         delegated = set(dir(self.pvt_o))
 *         res_set = res_set.union(delegated)
 *         res_set = res_set.difference(pickle_attributes)             # <<<<<<<<<<<<<<
 *         return list(res_set)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_res_set, __pyx_n_s_difference); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_9pyprotect_9protected_pickle_attributes};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_res_set, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":375
 *         res_set = res_set.union(delegated)
 *         res_set = res_set.difference(pickle_attributes)
 *         return list(res_set)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(__pyx_v_res_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":370
 *             raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 *     cdef wrapped_dir(self):             # <<<<<<<<<<<<<<
 *         res_set = special_attributes
 *         delegated = set(dir(self.pvt_o))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.wrapped_dir", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res_set);
  __Pyx_XDECREF(__pyx_v_delegated);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":381
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.wrapped_getattr(a)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattribute__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_2__getattribute__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattribute__", 0);

  /* "Wrapped_Frozen.pxi":382
 * 
 *     def __getattribute__(self, a):
 *         return self.wrapped_getattr(a)             # <<<<<<<<<<<<<<
 * 
 *     def __setattr__(self, a, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->wrapped_getattr(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":381
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.wrapped_getattr(a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__getattribute__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":384
 *         return self.wrapped_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.wrapped_check_setattr(a, val)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_7Wrapped_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_7Wrapped_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_4__setattr__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), ((PyObject *)__pyx_v_a), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_7Wrapped_4__setattr__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setattr__", 0);

  /* "Wrapped_Frozen.pxi":386
 *     def __setattr__(self, a, val):
 *         # Only checks and raises exceptions
 *         self.wrapped_check_setattr(a, val)             # <<<<<<<<<<<<<<
 *         setattr(self.pvt_o, a, val)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->wrapped_check_setattr(__pyx_v_self, __pyx_v_a, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":387
 *         # Only checks and raises exceptions
 *         self.wrapped_check_setattr(a, val)
 *         setattr(self.pvt_o, a, val)             # <<<<<<<<<<<<<<
 * 
 *     def __delattr__(self, a):
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_SetAttr(__pyx_t_1, __pyx_v_a, __pyx_v_val); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":384
 *         return self.wrapped_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.wrapped_check_setattr(a, val)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__setattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":389
 *         setattr(self.pvt_o, a, val)
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.wrapped_check_delattr(a)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_7Wrapped_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static int __pyx_pw_9pyprotect_9protected_7Wrapped_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_6__delattr__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_7Wrapped_6__delattr__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delattr__", 0);

  /* "Wrapped_Frozen.pxi":391
 *     def __delattr__(self, a):
 *         # Only checks and raises exceptions
 *         self.wrapped_check_delattr(a)             # <<<<<<<<<<<<<<
 *         delattr(self.pvt_o, a)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->wrapped_check_delattr(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":392
 *         # Only checks and raises exceptions
 *         self.wrapped_check_delattr(a)
 *         delattr(self.pvt_o, a)             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_DelAttr(__pyx_t_1, __pyx_v_a); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Wrapped_Frozen.pxi":389
 *         setattr(self.pvt_o, a, val)
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.wrapped_check_delattr(a)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__delattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":394
 *         delattr(self.pvt_o, a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.wrapped_dir()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7Wrapped_9__dir__ = {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Wrapped_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__dir__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__dir__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_8__dir__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_8__dir__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "Wrapped_Frozen.pxi":395
 * 
 *     def __dir__(self):
 *         return self.wrapped_dir()             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->wrapped_dir(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":394
 *         delattr(self.pvt_o, a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.wrapped_dir()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":397
 *         return self.wrapped_dir()
 * 
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_11__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_10__richcmp__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_10__richcmp__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "Wrapped_Frozen.pxi":399
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 *     # Needs to be class-specific
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->comparator(__pyx_v_self, __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":397
 *         return self.wrapped_dir()
 * 
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":403
 *     # Needs to be class-specific
 *     # Depends on pvt_o being hashable
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((
 *             id(type(self)),
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_7Wrapped_13__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_7Wrapped_13__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_12__hash__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_7Wrapped_12__hash__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_hash_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Wrapped_Frozen.pxi":405
 *     def __hash__(self):
 *         return hash((
 *             id(type(self)),             # <<<<<<<<<<<<<<
 *             str(self.get_rules()),
 *             id(self.pvt_o),
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Wrapped_Frozen.pxi":406
 *         return hash((
 *             id(type(self)),
 *             str(self.get_rules()),             # <<<<<<<<<<<<<<
 *             id(self.pvt_o),
 *             hash(self.pvt_o)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped *)__pyx_v_self->__pyx_vtab)->get_rules(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Str(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Wrapped_Frozen.pxi":407
 *             id(type(self)),
 *             str(self.get_rules()),
 *             id(self.pvt_o),             # <<<<<<<<<<<<<<
 *             hash(self.pvt_o)
 *         ))
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_self->__pyx_base.pvt_o); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Wrapped_Frozen.pxi":408
 *             str(self.get_rules()),
 *             id(self.pvt_o),
 *             hash(self.pvt_o)             # <<<<<<<<<<<<<<
 *         ))
 * 
 */
  __pyx_t_4 = __pyx_v_self->__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(3, 408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromHash_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "Wrapped_Frozen.pxi":405
 *     def __hash__(self):
 *         return hash((
 *             id(type(self)),             # <<<<<<<<<<<<<<
 *             str(self.get_rules()),
 *             id(self.pvt_o),
 */
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;

  /* "Wrapped_Frozen.pxi":404
 *     # Depends on pvt_o being hashable
 *     def __hash__(self):
 *         return hash((             # <<<<<<<<<<<<<<
 *             id(type(self)),
 *             str(self.get_rules()),
 */
  __pyx_t_5 = PyObject_Hash(__pyx_t_6); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(3, 404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":403
 *     # Needs to be class-specific
 *     # Depends on pvt_o being hashable
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((
 *             id(type(self)),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7Wrapped_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Wrapped_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_14__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_14__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->cn);
  __Pyx_GIVEREF(__pyx_v_self->cn);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 4, ((PyObject *)__pyx_v_self->protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->rules);
  __Pyx_GIVEREF(__pyx_v_self->rules);
  PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Wrapped); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, None), state
 *     else:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Wrapped__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Wrapped); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Wrapped__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7Wrapped_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Wrapped_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_7Wrapped_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Wrapped_16__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Wrapped_16__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Wrapped__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Wrapped__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Wrapped__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Wrapped.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":418
 *     Subclass of Wrapped that is automatically frozen
 *     '''
 *     def __init__(self, o):             # <<<<<<<<<<<<<<
 *         '''o-->object to be wrapped'''
 *         Wrapped.__init__(self, o, frozen=True)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_6Frozen_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_6Frozen___init__, "o-->object to be wrapped");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_6Frozen___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_6Frozen_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 418, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(3, 418, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(3, 418, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Frozen.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_6Frozen___init__(((struct __pyx_obj_9pyprotect_9protected_Frozen *)__pyx_v_self), __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_6Frozen___init__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self, PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Wrapped_Frozen.pxi":420
 *     def __init__(self, o):
 *         '''o-->object to be wrapped'''
 *         Wrapped.__init__(self, o, frozen=True)             # <<<<<<<<<<<<<<
 * 
 *     # Python / cython does not automatically use parent __hash__
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_o);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_frozen, Py_True) < 0) __PYX_ERR(3, 420, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Wrapped_Frozen.pxi":418
 *     Subclass of Wrapped that is automatically frozen
 *     '''
 *     def __init__(self, o):             # <<<<<<<<<<<<<<
 *         '''o-->object to be wrapped'''
 *         Wrapped.__init__(self, o, frozen=True)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Frozen.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":423
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_6Frozen_3__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_6Frozen_3__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_6Frozen_2__hash__(((struct __pyx_obj_9pyprotect_9protected_Frozen *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_6Frozen_2__hash__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_hash_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Wrapped_Frozen.pxi":424
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):
 *         return Wrapped.__hash__(self)             # <<<<<<<<<<<<<<
 * 
 *     # __richcmp__ needs to be class-specific
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_hash_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":423
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Frozen.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Wrapped_Frozen.pxi":427
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_6Frozen_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_6Frozen_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_6Frozen_4__richcmp__(((struct __pyx_obj_9pyprotect_9protected_Frozen *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_6Frozen_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "Wrapped_Frozen.pxi":429
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Frozen *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.comparator(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Wrapped_Frozen.pxi":427
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.Frozen.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_6Frozen_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_6Frozen_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_6Frozen_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_6Frozen_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_6Frozen_6__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_Frozen *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_6Frozen_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 4, ((PyObject *)__pyx_v_self->__pyx_base.protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.rules);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.rules);
  PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->__pyx_base.rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Frozen); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, None), state
 *     else:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Frozen__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Frozen.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Frozen__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_6Frozen_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_6Frozen_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_6Frozen_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_6Frozen_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Frozen.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_6Frozen_8__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_Frozen *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_6Frozen_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Frozen__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Frozen__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Frozen__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Frozen.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":12
 *     '''
 * 
 *     def __init__(self, o, cn, frozen=False, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->dict (any Mapping or MutableMapping type)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_11PrivacyDict_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_11PrivacyDict___init__, "\n        o-->dict (any Mapping or MutableMapping type)\n        cn-->str: class name\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_11PrivacyDict___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_11PrivacyDict_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_cn = 0;
  PyObject *__pyx_v_frozen = 0;
  PyObject *__pyx_v_oldstyle_class = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_cn,&__pyx_n_s_frozen,&__pyx_n_s_oldstyle_class,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_cn,&__pyx_n_s_frozen,&__pyx_n_s_oldstyle_class,0};
    #endif
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)Py_False);
    values[3] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 12, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 12, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(4, 12, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_frozen);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 12, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_oldstyle_class);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 12, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(4, 12, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_o = values[0];
    __pyx_v_cn = values[1];
    __pyx_v_frozen = values[2];
    __pyx_v_oldstyle_class = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, __pyx_nargs); __PYX_ERR(4, 12, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict___init__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self), __pyx_v_o, __pyx_v_cn, __pyx_v_frozen, __pyx_v_oldstyle_class);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_11PrivacyDict___init__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_cn, PyObject *__pyx_v_frozen, PyObject *__pyx_v_oldstyle_class) {
  PyObject *__pyx_v_C = NULL;
  PyObject *__pyx_v_t = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":19
 *         # In PY2 __dict__ is a dictproxy - not in types module and
 *         # not instance of dict or CollectionsABC.Mapping
 *         if PY2:             # <<<<<<<<<<<<<<
 *             class C(object):
 *                 pass
 */
  if (__pyx_v_9pyprotect_9protected_PY2) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":20
 *         # not instance of dict or CollectionsABC.Mapping
 *         if PY2:
 *             class C(object):             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
    __pyx_t_1 = __Pyx_PEP560_update_bases(__pyx_tuple__22); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_C, __pyx_n_s_init___locals_C, (PyObject *) NULL, __pyx_n_s_pyprotect_protected, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_1 != __pyx_tuple__22) {
      if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_tuple__22) < 0))) __PYX_ERR(4, 20, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_C, __pyx_t_1, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_C = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":23
 *                 pass
 * 
 *             t = type(C.__dict__)             # <<<<<<<<<<<<<<
 *         else:
 *             t = CollectionsABC.Mapping
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_C, __pyx_n_s_dict_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_t_1)));
    __pyx_v_t = ((PyObject *)Py_TYPE(__pyx_t_1));
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":19
 *         # In PY2 __dict__ is a dictproxy - not in types module and
 *         # not instance of dict or CollectionsABC.Mapping
 *         if PY2:             # <<<<<<<<<<<<<<
 *             class C(object):
 *                 pass
 */
    goto __pyx_L3;
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":25
 *             t = type(C.__dict__)
 *         else:
 *             t = CollectionsABC.Mapping             # <<<<<<<<<<<<<<
 *         if not isinstance(o, (t, dict)):
 *             raise TypeError('o: Invalid type: %s' % (o.__class__.__name__,))
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CollectionsABC); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_t = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":26
 *         else:
 *             t = CollectionsABC.Mapping
 *         if not isinstance(o, (t, dict)):             # <<<<<<<<<<<<<<
 *             raise TypeError('o: Invalid type: %s' % (o.__class__.__name__,))
 * 
 */
  __Pyx_INCREF(__pyx_v_t);
  __pyx_t_2 = __pyx_v_t;
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_o, __pyx_t_2); 
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_6 = PyDict_Check(__pyx_v_o); 
  __pyx_t_5 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (unlikely(__pyx_t_6)) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":27
 *             t = CollectionsABC.Mapping
 *         if not isinstance(o, (t, dict)):
 *             raise TypeError('o: Invalid type: %s' % (o.__class__.__name__,))             # <<<<<<<<<<<<<<
 * 
 *         self.cn = str(cn)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_o, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_o_Invalid_type_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 27, __pyx_L1_error)

    /* "PrivacyDict_FrozenPrivacyDict.pxi":26
 *         else:
 *             t = CollectionsABC.Mapping
 *         if not isinstance(o, (t, dict)):             # <<<<<<<<<<<<<<
 *             raise TypeError('o: Invalid type: %s' % (o.__class__.__name__,))
 * 
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":29
 *             raise TypeError('o: Invalid type: %s' % (o.__class__.__name__,))
 * 
 *         self.cn = str(cn)             # <<<<<<<<<<<<<<
 *         Wrapped.__init__(self, o=o, frozen=frozen, oldstyle_class=oldstyle_class)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Str(__pyx_v_cn); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyString_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(4, 29, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.cn);
  __pyx_v_self->__pyx_base.cn = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":30
 * 
 *         self.cn = str(cn)
 *         Wrapped.__init__(self, o=o, frozen=frozen, oldstyle_class=oldstyle_class)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_v_o) < 0) __PYX_ERR(4, 30, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_frozen, __pyx_v_frozen) < 0) __PYX_ERR(4, 30, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_oldstyle_class, __pyx_v_oldstyle_class) < 0) __PYX_ERR(4, 30, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":12
 *     '''
 * 
 *     def __init__(self, o, cn, frozen=False, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->dict (any Mapping or MutableMapping type)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_C);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":40
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         if self.attr_hidden(a):
 *             raise KeyError(a)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattribute__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_2__getattribute__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_v_py2_map = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattribute__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":41
 * 
 *     def __getattribute__(self, a):
 *         if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *             raise KeyError(a)
 *         # Cannot modify CLASS of wrapped object
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":42
 *     def __getattribute__(self, a):
 *         if self.attr_hidden(a):
 *             raise KeyError(a)             # <<<<<<<<<<<<<<
 *         # Cannot modify CLASS of wrapped object
 *         if a == '__class__':
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(4, 42, __pyx_L1_error)

    /* "PrivacyDict_FrozenPrivacyDict.pxi":41
 * 
 *     def __getattribute__(self, a):
 *         if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *             raise KeyError(a)
 *         # Cannot modify CLASS of wrapped object
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":44
 *             raise KeyError(a)
 *         # Cannot modify CLASS of wrapped object
 *         if a == '__class__':             # <<<<<<<<<<<<<<
 *             return freeze(self.pvt_o.__class__)
 *         if a in set([
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_a, __pyx_n_s_class, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 44, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":45
 *         # Cannot modify CLASS of wrapped object
 *         if a == '__class__':
 *             return freeze(self.pvt_o.__class__)             # <<<<<<<<<<<<<<
 *         if a in set([
 *             '__getitem__', '__setitem__', '__delitem__',
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_freeze); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 45, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":44
 *             raise KeyError(a)
 *         # Cannot modify CLASS of wrapped object
 *         if a == '__class__':             # <<<<<<<<<<<<<<
 *             return freeze(self.pvt_o.__class__)
 *         if a in set([
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":46
 *         if a == '__class__':
 *             return freeze(self.pvt_o.__class__)
 *         if a in set([             # <<<<<<<<<<<<<<
 *             '__getitem__', '__setitem__', '__delitem__',
 *             '__str__', '__repr__',
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PySet_Add(__pyx_t_1, __pyx_n_s_getitem) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_setitem) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_delitem) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_str_2) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_repr) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_copy) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_iterkeys) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_iteritems) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_itervalues) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_viewkeys) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_viewitems) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_viewvalues) < 0) __PYX_ERR(4, 46, __pyx_L1_error)
  __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_v_a, __pyx_t_1, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":53
 *             'viewkeys', 'viewitems', 'viewvalues',
 *         ]):
 *             return __HiddenPartial(getattr(PrivacyDict, a), self)             # <<<<<<<<<<<<<<
 *         if a in set([
 *             'keys', 'items', 'values', 'copy',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":46
 *         if a == '__class__':
 *             return freeze(self.pvt_o.__class__)
 *         if a in set([             # <<<<<<<<<<<<<<
 *             '__getitem__', '__setitem__', '__delitem__',
 *             '__str__', '__repr__',
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":54
 *         ]):
 *             return __HiddenPartial(getattr(PrivacyDict, a), self)
 *         if a in set([             # <<<<<<<<<<<<<<
 *             'keys', 'items', 'values', 'copy',
 *         ]):
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PySet_Add(__pyx_t_1, __pyx_n_s_keys) < 0) __PYX_ERR(4, 54, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_items) < 0) __PYX_ERR(4, 54, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_values) < 0) __PYX_ERR(4, 54, __pyx_L1_error)
  if (PySet_Add(__pyx_t_1, __pyx_n_s_copy) < 0) __PYX_ERR(4, 54, __pyx_L1_error)
  __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_v_a, __pyx_t_1, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":58
 *         ]):
 *             py2_map = {
 *                 'keys': __HiddenPartial(getattr(PrivacyDict, 'keys_py2'), self),             # <<<<<<<<<<<<<<
 *                 'items': __HiddenPartial(getattr(PrivacyDict, 'items_py2'), self),
 *                 'values': __HiddenPartial(getattr(PrivacyDict, 'values_py2'), self),
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_n_s_keys_py2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_keys, __pyx_t_3) < 0) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":59
 *             py2_map = {
 *                 'keys': __HiddenPartial(getattr(PrivacyDict, 'keys_py2'), self),
 *                 'items': __HiddenPartial(getattr(PrivacyDict, 'items_py2'), self),             # <<<<<<<<<<<<<<
 *                 'values': __HiddenPartial(getattr(PrivacyDict, 'values_py2'), self),
 *             }
 */
    __pyx_t_3 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_n_s_items_py2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_items, __pyx_t_3) < 0) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":60
 *                 'keys': __HiddenPartial(getattr(PrivacyDict, 'keys_py2'), self),
 *                 'items': __HiddenPartial(getattr(PrivacyDict, 'items_py2'), self),
 *                 'values': __HiddenPartial(getattr(PrivacyDict, 'values_py2'), self),             # <<<<<<<<<<<<<<
 *             }
 *             if PY2 and a in py2_map:
 */
    __pyx_t_3 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_n_s_values_py2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_values, __pyx_t_3) < 0) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_py2_map = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":62
 *                 'values': __HiddenPartial(getattr(PrivacyDict, 'values_py2'), self),
 *             }
 *             if PY2 and a in py2_map:             # <<<<<<<<<<<<<<
 *                 return py2_map[a]
 *             return __HiddenPartial(getattr(PrivacyDict, a), self)
 */
    if (__pyx_v_9pyprotect_9protected_PY2) {
    } else {
      __pyx_t_2 = __pyx_v_9pyprotect_9protected_PY2;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_7 = (__Pyx_PyDict_ContainsTF(__pyx_v_a, __pyx_v_py2_map, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(4, 62, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_7;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_2) {

      /* "PrivacyDict_FrozenPrivacyDict.pxi":63
 *             }
 *             if PY2 and a in py2_map:
 *                 return py2_map[a]             # <<<<<<<<<<<<<<
 *             return __HiddenPartial(getattr(PrivacyDict, a), self)
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_py2_map, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 63, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "PrivacyDict_FrozenPrivacyDict.pxi":62
 *                 'values': __HiddenPartial(getattr(PrivacyDict, 'values_py2'), self),
 *             }
 *             if PY2 and a in py2_map:             # <<<<<<<<<<<<<<
 *                 return py2_map[a]
 *             return __HiddenPartial(getattr(PrivacyDict, a), self)
 */
    }

    /* "PrivacyDict_FrozenPrivacyDict.pxi":64
 *             if PY2 and a in py2_map:
 *                 return py2_map[a]
 *             return __HiddenPartial(getattr(PrivacyDict, a), self)             # <<<<<<<<<<<<<<
 * 
 *         # If frozen, freeze all the way down
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":54
 *         ]):
 *             return __HiddenPartial(getattr(PrivacyDict, a), self)
 *         if a in set([             # <<<<<<<<<<<<<<
 *             'keys', 'items', 'values', 'copy',
 *         ]):
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":67
 * 
 *         # If frozen, freeze all the way down
 *         return self.wrapped_getattr(a)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.wrapped_getattr(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":40
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         if self.attr_hidden(a):
 *             raise KeyError(a)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__getattribute__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py2_map);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":69
 *         return self.wrapped_getattr(a)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         if self.attr_hidden(key):
 *             raise KeyError(key)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_4__getitem__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_4__getitem__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":70
 * 
 *     def __getitem__(self, key):
 *         if self.attr_hidden(key):             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return self.fif(self.pvt_o.__getitem__(key))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":71
 *     def __getitem__(self, key):
 *         if self.attr_hidden(key):
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 *         return self.fif(self.pvt_o.__getitem__(key))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(4, 71, __pyx_L1_error)

    /* "PrivacyDict_FrozenPrivacyDict.pxi":70
 * 
 *     def __getitem__(self, key):
 *         if self.attr_hidden(key):             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return self.fif(self.pvt_o.__getitem__(key))
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":72
 *         if self.attr_hidden(key):
 *             raise KeyError(key)
 *         return self.fif(self.pvt_o.__getitem__(key))             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_n_s_getitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":69
 *         return self.wrapped_getattr(a)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         if self.attr_hidden(key):
 *             raise KeyError(key)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":74
 *         return self.fif(self.pvt_o.__getitem__(key))
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         if self.attr_hidden(key):
 *             raise KeyError(key)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_11PrivacyDict_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_9pyprotect_9protected_11PrivacyDict_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_6__delitem__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_11PrivacyDict_6__delitem__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_nodel_msg = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":75
 * 
 *     def __delitem__(self, key):
 *         if self.attr_hidden(key):             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         nodel_msg = 'Cannot delete private attribute: %s.%s' % (self.cn, str(key))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":76
 *     def __delitem__(self, key):
 *         if self.attr_hidden(key):
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 *         nodel_msg = 'Cannot delete private attribute: %s.%s' % (self.cn, str(key))
 *         if ro_private_attr.match(key):
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(4, 76, __pyx_L1_error)

    /* "PrivacyDict_FrozenPrivacyDict.pxi":75
 * 
 *     def __delitem__(self, key):
 *         if self.attr_hidden(key):             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         nodel_msg = 'Cannot delete private attribute: %s.%s' % (self.cn, str(key))
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":77
 *         if self.attr_hidden(key):
 *             raise KeyError(key)
 *         nodel_msg = 'Cannot delete private attribute: %s.%s' % (self.cn, str(key))             # <<<<<<<<<<<<<<
 *         if ro_private_attr.match(key):
 *             raise ProtectionError(nodel_msg)
 */
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_delete_private_attribute, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_nodel_msg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":78
 *             raise KeyError(key)
 *         nodel_msg = 'Cannot delete private attribute: %s.%s' % (self.cn, str(key))
 *         if ro_private_attr.match(key):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nodel_msg)
 *         Wrapped.__delitem__(self, key)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_ro_private_attr, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":79
 *         nodel_msg = 'Cannot delete private attribute: %s.%s' % (self.cn, str(key))
 *         if ro_private_attr.match(key):
 *             raise ProtectionError(nodel_msg)             # <<<<<<<<<<<<<<
 *         Wrapped.__delitem__(self, key)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_nodel_msg};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(4, 79, __pyx_L1_error)

    /* "PrivacyDict_FrozenPrivacyDict.pxi":78
 *             raise KeyError(key)
 *         nodel_msg = 'Cannot delete private attribute: %s.%s' % (self.cn, str(key))
 *         if ro_private_attr.match(key):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nodel_msg)
 *         Wrapped.__delitem__(self, key)
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":80
 *         if ro_private_attr.match(key):
 *             raise ProtectionError(nodel_msg)
 *         Wrapped.__delitem__(self, key)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, key, val):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_delitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_v_self), __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":74
 *         return self.fif(self.pvt_o.__getitem__(key))
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         if self.attr_hidden(key):
 *             raise KeyError(key)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nodel_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":82
 *         Wrapped.__delitem__(self, key)
 * 
 *     def __setitem__(self, key, val):             # <<<<<<<<<<<<<<
 *         nopvt_msg = 'Cannot set private attribute: %s.%s' % (self.cn, str(key))
 * 
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_11PrivacyDict_9__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_11PrivacyDict_9__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_8__setitem__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self), ((PyObject *)__pyx_v_key), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_11PrivacyDict_8__setitem__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_val) {
  PyObject *__pyx_v_nopvt_msg = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":83
 * 
 *     def __setitem__(self, key, val):
 *         nopvt_msg = 'Cannot set private attribute: %s.%s' % (self.cn, str(key))             # <<<<<<<<<<<<<<
 * 
 *         if self.attr_hidden(key):
 */
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_set_private_attribute_s_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nopvt_msg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":85
 *         nopvt_msg = 'Cannot set private attribute: %s.%s' % (self.cn, str(key))
 * 
 *         if self.attr_hidden(key):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nopvt_msg)
 *         if ro_private_attr.match(key):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":86
 * 
 *         if self.attr_hidden(key):
 *             raise ProtectionError(nopvt_msg)             # <<<<<<<<<<<<<<
 *         if ro_private_attr.match(key):
 *             raise ProtectionError(nopvt_msg)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_nopvt_msg};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(4, 86, __pyx_L1_error)

    /* "PrivacyDict_FrozenPrivacyDict.pxi":85
 *         nopvt_msg = 'Cannot set private attribute: %s.%s' % (self.cn, str(key))
 * 
 *         if self.attr_hidden(key):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nopvt_msg)
 *         if ro_private_attr.match(key):
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":87
 *         if self.attr_hidden(key):
 *             raise ProtectionError(nopvt_msg)
 *         if ro_private_attr.match(key):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nopvt_msg)
 *         Wrapped.__setitem__(self, key, val)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_ro_private_attr, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "PrivacyDict_FrozenPrivacyDict.pxi":88
 *             raise ProtectionError(nopvt_msg)
 *         if ro_private_attr.match(key):
 *             raise ProtectionError(nopvt_msg)             # <<<<<<<<<<<<<<
 *         Wrapped.__setitem__(self, key, val)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_nopvt_msg};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 88, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(4, 88, __pyx_L1_error)

    /* "PrivacyDict_FrozenPrivacyDict.pxi":87
 *         if self.attr_hidden(key):
 *             raise ProtectionError(nopvt_msg)
 *         if ro_private_attr.match(key):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nopvt_msg)
 *         Wrapped.__setitem__(self, key, val)
 */
  }

  /* "PrivacyDict_FrozenPrivacyDict.pxi":89
 *         if ro_private_attr.match(key):
 *             raise ProtectionError(nopvt_msg)
 *         Wrapped.__setitem__(self, key, val)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_setitem); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, ((PyObject *)__pyx_v_self), __pyx_v_key, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":82
 *         Wrapped.__delitem__(self, key)
 * 
 *     def __setitem__(self, key, val):             # <<<<<<<<<<<<<<
 *         nopvt_msg = 'Cannot set private attribute: %s.%s' % (self.cn, str(key))
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nopvt_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":91
 *         Wrapped.__setitem__(self, key, val)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         ret = dict([x for x in self.items()])
 *         return repr(ret)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_10__repr__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_10__repr__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":92
 * 
 *     def __repr__(self):
 *         ret = dict([x for x in self.items()])             # <<<<<<<<<<<<<<
 *         return repr(ret)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 92, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(4, 92, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(((PyObject *)__pyx_v_self), 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 92, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, NULL, NULL, &__pyx_t_6, __pyx_t_5);
      if (unlikely(__pyx_t_7 == 0)) break;
      if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(4, 92, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_x, __pyx_t_6);
      __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr14__pyx_v_x))) __PYX_ERR(4, 92, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_x); __pyx_9genexpr14__pyx_v_x = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_x); __pyx_9genexpr14__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ret = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":93
 *     def __repr__(self):
 *         ret = dict([x for x in self.items()])
 *         return repr(ret)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_Repr(__pyx_v_ret); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":91
 *         Wrapped.__setitem__(self, key, val)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         ret = dict([x for x in self.items()])
 *         return repr(ret)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":95
 *         return repr(ret)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         ret = dict([x for x in self.items()])
 *         return str(ret)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_13__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_13__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_12__str__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_12__str__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_9genexpr15__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":96
 * 
 *     def __str__(self):
 *         ret = dict([x for x in self.items()])             # <<<<<<<<<<<<<<
 *         return str(ret)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 96, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(4, 96, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(((PyObject *)__pyx_v_self), 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 96, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, NULL, NULL, &__pyx_t_6, __pyx_t_5);
      if (unlikely(__pyx_t_7 == 0)) break;
      if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(4, 96, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_x, __pyx_t_6);
      __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr15__pyx_v_x))) __PYX_ERR(4, 96, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_x); __pyx_9genexpr15__pyx_v_x = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_x); __pyx_9genexpr15__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ret = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":97
 *     def __str__(self):
 *         ret = dict([x for x in self.items()])
 *         return str(ret)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_Str(__pyx_v_ret); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":95
 *         return repr(ret)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         ret = dict([x for x in self.items()])
 *         return str(ret)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":99
 *         return str(ret)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         # TODO: Something strange here: Just wrapping self.pvt_o and returning
 *         # as a new FrozenPrivacyDict doesn't seem to adopt all the behavior
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_15copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_15copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_15copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_15copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_14copy(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_14copy(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  struct __pyx_opt_args_9pyprotect_9protected_privatedict __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":103
 *         # as a new FrozenPrivacyDict doesn't seem to adopt all the behavior
 *         # of PrivacyDict!
 *         d = {}             # <<<<<<<<<<<<<<
 *         d.update(dict(self.items()))
 *         # return privatedict(d, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_d = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":104
 *         # of PrivacyDict!
 *         d = {}
 *         d.update(dict(self.items()))             # <<<<<<<<<<<<<<
 *         # return privatedict(d, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 *         return privatedict(
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_d, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":106
 *         d.update(dict(self.items()))
 *         # return privatedict(d, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 *         return privatedict(             # <<<<<<<<<<<<<<
 *             self.pvt_o, self.cn,
 *             frozen=True, oldstyle_class=self.oldstyle_class
 */
  __Pyx_XDECREF(__pyx_r);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":107
 *         # return privatedict(d, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 *         return privatedict(
 *             self.pvt_o, self.cn,             # <<<<<<<<<<<<<<
 *             frozen=True, oldstyle_class=self.oldstyle_class
 *         )
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->__pyx_base.cn;
  __Pyx_INCREF(__pyx_t_2);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":108
 *         return privatedict(
 *             self.pvt_o, self.cn,
 *             frozen=True, oldstyle_class=self.oldstyle_class             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":106
 *         d.update(dict(self.items()))
 *         # return privatedict(d, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 *         return privatedict(             # <<<<<<<<<<<<<<
 *             self.pvt_o, self.cn,
 *             frozen=True, oldstyle_class=self.oldstyle_class
 */
  __pyx_t_6.__pyx_n = 2;
  __pyx_t_6.frozen = Py_True;
  __pyx_t_6.oldstyle_class = __pyx_t_3;
  __pyx_t_5 = __pyx_f_9pyprotect_9protected_privatedict(__pyx_t_1, __pyx_t_2, &__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":99
 *         return str(ret)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         # TODO: Something strange here: Just wrapping self.pvt_o and returning
 *         # as a new FrozenPrivacyDict doesn't seem to adopt all the behavior
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_18generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "PrivacyDict_FrozenPrivacyDict.pxi":115
 *     # --------------------------------------------------------------------
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_17keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_17keys = {"keys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_17keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_17keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("keys", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "keys", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_16keys(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_16keys(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_2_keys(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(4, 115, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9pyprotect_9protected_11PrivacyDict_18generator1, __pyx_codeobj__23, (PyObject *) __pyx_cur_scope, __pyx_n_s_keys, __pyx_n_s_PrivacyDict_keys, __pyx_n_s_pyprotect_protected); if (unlikely(!gen)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_18generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *__pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 115, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":116
 * 
 *     def keys(self):
 *         for k in self.pvt_o.keys():             # <<<<<<<<<<<<<<
 *             if self.attr_hidden(k):
 *                 continue
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(4, 116, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o, 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":117
 *     def keys(self):
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):             # <<<<<<<<<<<<<<
 *                 continue
 *             yield k
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "PrivacyDict_FrozenPrivacyDict.pxi":118
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):
 *                 continue             # <<<<<<<<<<<<<<
 *             yield k
 * 
 */
      goto __pyx_L4_continue;

      /* "PrivacyDict_FrozenPrivacyDict.pxi":117
 *     def keys(self):
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):             # <<<<<<<<<<<<<<
 *                 continue
 *             yield k
 */
    }

    /* "PrivacyDict_FrozenPrivacyDict.pxi":119
 *             if self.attr_hidden(k):
 *                 continue
 *             yield k             # <<<<<<<<<<<<<<
 * 
 *     def items(self):
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k);
    __pyx_r = __pyx_cur_scope->__pyx_v_k;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 119, __pyx_L1_error)
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":115
 *     # --------------------------------------------------------------------
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_21generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "PrivacyDict_FrozenPrivacyDict.pxi":121
 *             yield k
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_20items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_20items = {"items", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_20items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_20items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("items", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "items", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_19items(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_19items(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_3_items(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(4, 121, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9pyprotect_9protected_11PrivacyDict_21generator2, __pyx_codeobj__24, (PyObject *) __pyx_cur_scope, __pyx_n_s_items, __pyx_n_s_PrivacyDict_items, __pyx_n_s_pyprotect_protected); if (unlikely(!gen)) __PYX_ERR(4, 121, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_21generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *__pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 121, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":122
 * 
 *     def items(self):
 *         for k in self.keys():             # <<<<<<<<<<<<<<
 *             v = self.pvt_o[k]
 *             yield self.fif((k, v))
 */
  __pyx_t_2 = 0;
  if (unlikely(((PyObject *)__pyx_cur_scope->__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(4, 122, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_cur_scope->__pyx_v_self), 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":123
 *     def items(self):
 *         for k in self.keys():
 *             v = self.pvt_o[k]             # <<<<<<<<<<<<<<
 *             yield self.fif((k, v))
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":124
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 *             yield self.fif((k, v))             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_cur_scope->__pyx_v_k);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_cur_scope->__pyx_v_v);
    __pyx_t_7 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_cur_scope->__pyx_v_self), __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 124, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":121
 *             yield k
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_24generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "PrivacyDict_FrozenPrivacyDict.pxi":126
 *             yield self.fif((k, v))
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_23values(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_23values = {"values", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_23values, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_23values(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("values", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "values", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_22values(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_22values(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_4_values(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(4, 126, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9pyprotect_9protected_11PrivacyDict_24generator3, __pyx_codeobj__25, (PyObject *) __pyx_cur_scope, __pyx_n_s_values, __pyx_n_s_PrivacyDict_values, __pyx_n_s_pyprotect_protected); if (unlikely(!gen)) __PYX_ERR(4, 126, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_24generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *__pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 126, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":127
 * 
 *     def values(self):
 *         for k in self.keys():             # <<<<<<<<<<<<<<
 *             v = self.pvt_o[k]
 *             yield self.fif(v)
 */
  __pyx_t_2 = 0;
  if (unlikely(((PyObject *)__pyx_cur_scope->__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(4, 127, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_cur_scope->__pyx_v_self), 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":128
 *     def values(self):
 *         for k in self.keys():
 *             v = self.pvt_o[k]             # <<<<<<<<<<<<<<
 *             yield self.fif(v)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":129
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 *             yield self.fif(v)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_v); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 129, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":126
 *             yield self.fif((k, v))
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":135
 *     # --------------------------------------------------------------------
 * 
 *     def keys_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields keys
 *         ret = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_26keys_py2(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_26keys_py2 = {"keys_py2", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_26keys_py2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_26keys_py2(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys_py2 (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("keys_py2", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "keys_py2", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_25keys_py2(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_25keys_py2(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys_py2", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":137
 *     def keys_py2(self):
 *         # In PY2 returns list, in PY3, yields keys
 *         ret = []             # <<<<<<<<<<<<<<
 *         for k in self.pvt_o.iterkeys():
 *             if self.attr_hidden(k):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":138
 *         # In PY2 returns list, in PY3, yields keys
 *         ret = []
 *         for k in self.pvt_o.iterkeys():             # <<<<<<<<<<<<<<
 *             if self.attr_hidden(k):
 *                 continue
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base.pvt_o == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "iterkeys");
    __PYX_ERR(4, 138, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_self->__pyx_base.__pyx_base.pvt_o, 0, __pyx_n_s_iterkeys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":139
 *         ret = []
 *         for k in self.pvt_o.iterkeys():
 *             if self.attr_hidden(k):             # <<<<<<<<<<<<<<
 *                 continue
 *             ret.append(k)
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(4, 139, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "PrivacyDict_FrozenPrivacyDict.pxi":140
 *         for k in self.pvt_o.iterkeys():
 *             if self.attr_hidden(k):
 *                 continue             # <<<<<<<<<<<<<<
 *             ret.append(k)
 *         return ret
 */
      goto __pyx_L3_continue;

      /* "PrivacyDict_FrozenPrivacyDict.pxi":139
 *         ret = []
 *         for k in self.pvt_o.iterkeys():
 *             if self.attr_hidden(k):             # <<<<<<<<<<<<<<
 *                 continue
 *             ret.append(k)
 */
    }

    /* "PrivacyDict_FrozenPrivacyDict.pxi":141
 *             if self.attr_hidden(k):
 *                 continue
 *             ret.append(k)             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_v_k); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(4, 141, __pyx_L1_error)
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":142
 *                 continue
 *             ret.append(k)
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def items_py2(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":135
 *     # --------------------------------------------------------------------
 * 
 *     def keys_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields keys
 *         ret = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.keys_py2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":144
 *         return ret
 * 
 *     def items_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields items
 *         ret = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_28items_py2(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_28items_py2 = {"items_py2", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_28items_py2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_28items_py2(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items_py2 (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("items_py2", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "items_py2", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_27items_py2(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_27items_py2(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items_py2", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":146
 *     def items_py2(self):
 *         # In PY2 returns list, in PY3, yields items
 *         ret = []             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":147
 *         # In PY2 returns list, in PY3, yields items
 *         ret = []
 *         for k in self.keys():             # <<<<<<<<<<<<<<
 *             v = self.pvt_o[k]
 *             ret.append((k, v))
 */
  __pyx_t_2 = 0;
  if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(4, 147, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_v_self), 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":148
 *         ret = []
 *         for k in self.keys():
 *             v = self.pvt_o[k]             # <<<<<<<<<<<<<<
 *             ret.append((k, v))
 *         return self.fif(ret)
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":149
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 *             ret.append((k, v))             # <<<<<<<<<<<<<<
 *         return self.fif(ret)
 * 
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_k);
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_v);
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_5); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(4, 149, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":150
 *             v = self.pvt_o[k]
 *             ret.append((k, v))
 *         return self.fif(ret)             # <<<<<<<<<<<<<<
 * 
 *     def values_py2(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_ret); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":144
 *         return ret
 * 
 *     def items_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields items
 *         ret = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.items_py2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":152
 *         return self.fif(ret)
 * 
 *     def values_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields values
 *         ret = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_30values_py2(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_30values_py2 = {"values_py2", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_30values_py2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_30values_py2(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values_py2 (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("values_py2", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "values_py2", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_29values_py2(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_29values_py2(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values_py2", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":154
 *     def values_py2(self):
 *         # In PY2 returns list, in PY3, yields values
 *         ret = []             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":155
 *         # In PY2 returns list, in PY3, yields values
 *         ret = []
 *         for k in self.keys():             # <<<<<<<<<<<<<<
 *             v = self.pvt_o[k]
 *             ret.append(v)
 */
  __pyx_t_2 = 0;
  if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(4, 155, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_v_self), 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":156
 *         ret = []
 *         for k in self.keys():
 *             v = self.pvt_o[k]             # <<<<<<<<<<<<<<
 *             ret.append(v)
 *         return self.fif(ret)
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":157
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 *             ret.append(v)             # <<<<<<<<<<<<<<
 *         return self.fif(ret)
 * 
 */
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_v_v); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(4, 157, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":158
 *             v = self.pvt_o[k]
 *             ret.append(v)
 *         return self.fif(ret)             # <<<<<<<<<<<<<<
 * 
 *     def iterkeys(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_ret); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":152
 *         return self.fif(ret)
 * 
 *     def values_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields values
 *         ret = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.values_py2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_33generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "PrivacyDict_FrozenPrivacyDict.pxi":160
 *         return self.fif(ret)
 * 
 *     def iterkeys(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.pvt_o.keys():
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_32iterkeys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_32iterkeys = {"iterkeys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_32iterkeys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_32iterkeys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iterkeys (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("iterkeys", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "iterkeys", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_31iterkeys(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_31iterkeys(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iterkeys", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_5_iterkeys(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(4, 160, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9pyprotect_9protected_11PrivacyDict_33generator4, __pyx_codeobj__26, (PyObject *) __pyx_cur_scope, __pyx_n_s_iterkeys, __pyx_n_s_PrivacyDict_iterkeys, __pyx_n_s_pyprotect_protected); if (unlikely(!gen)) __PYX_ERR(4, 160, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.iterkeys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_33generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *__pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iterkeys", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 160, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":162
 *     def iterkeys(self):
 *         # PY2 only
 *         for k in self.pvt_o.keys():             # <<<<<<<<<<<<<<
 *             if self.attr_hidden(k):
 *                 continue
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(4, 162, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o, 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":163
 *         # PY2 only
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):             # <<<<<<<<<<<<<<
 *                 continue
 *             yield k
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(4, 163, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "PrivacyDict_FrozenPrivacyDict.pxi":164
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):
 *                 continue             # <<<<<<<<<<<<<<
 *             yield k
 * 
 */
      goto __pyx_L4_continue;

      /* "PrivacyDict_FrozenPrivacyDict.pxi":163
 *         # PY2 only
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):             # <<<<<<<<<<<<<<
 *                 continue
 *             yield k
 */
    }

    /* "PrivacyDict_FrozenPrivacyDict.pxi":165
 *             if self.attr_hidden(k):
 *                 continue
 *             yield k             # <<<<<<<<<<<<<<
 * 
 *     def iteritems(self):
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k);
    __pyx_r = __pyx_cur_scope->__pyx_v_k;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 165, __pyx_L1_error)
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":160
 *         return self.fif(ret)
 * 
 *     def iterkeys(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.pvt_o.keys():
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("iterkeys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_36generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "PrivacyDict_FrozenPrivacyDict.pxi":167
 *             yield k
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_35iteritems(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_35iteritems = {"iteritems", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_35iteritems, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_35iteritems(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iteritems (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("iteritems", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "iteritems", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_34iteritems(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_34iteritems(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iteritems", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_6_iteritems(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(4, 167, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9pyprotect_9protected_11PrivacyDict_36generator5, __pyx_codeobj__27, (PyObject *) __pyx_cur_scope, __pyx_n_s_iteritems, __pyx_n_s_PrivacyDict_iteritems, __pyx_n_s_pyprotect_protected); if (unlikely(!gen)) __PYX_ERR(4, 167, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.iteritems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_36generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *__pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iteritems", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 167, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":169
 *     def iteritems(self):
 *         # PY2 only
 *         for k in self.iterkeys():             # <<<<<<<<<<<<<<
 *             v = self.pvt_o[k]
 *             yield self.fif((k, v))
 */
  __pyx_t_2 = 0;
  if (unlikely(((PyObject *)__pyx_cur_scope->__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "iterkeys");
    __PYX_ERR(4, 169, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_cur_scope->__pyx_v_self), 0, __pyx_n_s_iterkeys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":170
 *         # PY2 only
 *         for k in self.iterkeys():
 *             v = self.pvt_o[k]             # <<<<<<<<<<<<<<
 *             yield self.fif((k, v))
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":171
 *         for k in self.iterkeys():
 *             v = self.pvt_o[k]
 *             yield self.fif((k, v))             # <<<<<<<<<<<<<<
 * 
 *     def itervalues(self):
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_cur_scope->__pyx_v_k);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_cur_scope->__pyx_v_v);
    __pyx_t_7 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_cur_scope->__pyx_v_self), __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 171, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":167
 *             yield k
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("iteritems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_39generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "PrivacyDict_FrozenPrivacyDict.pxi":173
 *             yield self.fif((k, v))
 * 
 *     def itervalues(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_38itervalues(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_38itervalues = {"itervalues", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_38itervalues, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_38itervalues(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("itervalues (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("itervalues", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "itervalues", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_37itervalues(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_37itervalues(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("itervalues", 0);
  __pyx_cur_scope = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_7_itervalues(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(4, 173, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9pyprotect_9protected_11PrivacyDict_39generator6, __pyx_codeobj__28, (PyObject *) __pyx_cur_scope, __pyx_n_s_itervalues, __pyx_n_s_PrivacyDict_itervalues, __pyx_n_s_pyprotect_protected); if (unlikely(!gen)) __PYX_ERR(4, 173, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.itervalues", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9pyprotect_9protected_11PrivacyDict_39generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *__pyx_cur_scope = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("itervalues", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 173, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":175
 *     def itervalues(self):
 *         # PY2 only
 *         for k in self.iterkeys():             # <<<<<<<<<<<<<<
 *             yield self.fif(self.pvt_o[k])
 * 
 */
  __pyx_t_2 = 0;
  if (unlikely(((PyObject *)__pyx_cur_scope->__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "iterkeys");
    __PYX_ERR(4, 175, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_cur_scope->__pyx_v_self), 0, __pyx_n_s_iterkeys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(4, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "PrivacyDict_FrozenPrivacyDict.pxi":176
 *         # PY2 only
 *         for k in self.iterkeys():
 *             yield self.fif(self.pvt_o[k])             # <<<<<<<<<<<<<<
 * 
 *     def viewitems(self):
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_cur_scope->__pyx_v_self), __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(4, 176, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":173
 *             yield self.fif((k, v))
 * 
 *     def itervalues(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("itervalues", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":178
 *             yield self.fif(self.pvt_o[k])
 * 
 *     def viewitems(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.items()))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_41viewitems(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_41viewitems = {"viewitems", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_41viewitems, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_41viewitems(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("viewitems (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("viewitems", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "viewitems", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_40viewitems(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_40viewitems(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("viewitems", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":179
 * 
 *     def viewitems(self):
 *         return self.fif(set(self.items()))             # <<<<<<<<<<<<<<
 * 
 *     def viewvalues(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":178
 *             yield self.fif(self.pvt_o[k])
 * 
 *     def viewitems(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.items()))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.viewitems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":181
 *         return self.fif(set(self.items()))
 * 
 *     def viewvalues(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.values()))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_43viewvalues(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_43viewvalues = {"viewvalues", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_43viewvalues, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_43viewvalues(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("viewvalues (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("viewvalues", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "viewvalues", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_42viewvalues(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_42viewvalues(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("viewvalues", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":182
 * 
 *     def viewvalues(self):
 *         return self.fif(set(self.values()))             # <<<<<<<<<<<<<<
 * 
 *     def viewkeys(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":181
 *         return self.fif(set(self.items()))
 * 
 *     def viewvalues(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.values()))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.viewvalues", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":184
 *         return self.fif(set(self.values()))
 * 
 *     def viewkeys(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.keys()))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_45viewkeys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_45viewkeys = {"viewkeys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_45viewkeys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_45viewkeys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("viewkeys (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("viewkeys", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "viewkeys", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_44viewkeys(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_44viewkeys(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("viewkeys", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":185
 * 
 *     def viewkeys(self):
 *         return self.fif(set(self.keys()))             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.fif(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":184
 *         return self.fif(set(self.values()))
 * 
 *     def viewkeys(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.keys()))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.viewkeys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":193
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_11PrivacyDict_47__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_11PrivacyDict_47__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_46__hash__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_11PrivacyDict_46__hash__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_hash_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":194
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):
 *         return Wrapped.__hash__(self)             # <<<<<<<<<<<<<<
 * 
 *     # __richcmp__ needs to be class-specific
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_hash_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":193
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":197
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_49__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_49__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_48__richcmp__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_48__richcmp__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":199
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.comparator(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":197
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_51__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_51__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_51__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_51__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_50__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_50__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 4, ((PyObject *)__pyx_v_self->__pyx_base.protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.rules);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.rules);
  PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->__pyx_base.rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_PrivacyDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, None), state
 *     else:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PrivacyDict__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_PrivacyDict); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrivacyDict__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_53__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_11PrivacyDict_53__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_53__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_11PrivacyDict_53__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_11PrivacyDict_52__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_11PrivacyDict_52__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PrivacyDict__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_PrivacyDict__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrivacyDict__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.PrivacyDict.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":208
 *     Subclass of PrivacyDict that is automatically frozen
 *     '''
 *     def __init__(self, o, cn, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *         '''o-->object to be wrapped'''
 *         PrivacyDict.__init__(
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_17FrozenPrivacyDict___init__, "o-->object to be wrapped");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_17FrozenPrivacyDict___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_cn = 0;
  PyObject *__pyx_v_oldstyle_class = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_cn,&__pyx_n_s_oldstyle_class,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_cn,&__pyx_n_s_oldstyle_class,0};
    #endif
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 208, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cn)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 208, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(4, 208, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_oldstyle_class);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 208, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(4, 208, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_o = values[0];
    __pyx_v_cn = values[1];
    __pyx_v_oldstyle_class = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, __pyx_nargs); __PYX_ERR(4, 208, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivacyDict.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict___init__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *)__pyx_v_self), __pyx_v_o, __pyx_v_cn, __pyx_v_oldstyle_class);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict___init__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_cn, PyObject *__pyx_v_oldstyle_class) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":210
 *     def __init__(self, o, cn, oldstyle_class=None):
 *         '''o-->object to be wrapped'''
 *         PrivacyDict.__init__(             # <<<<<<<<<<<<<<
 *             self, o, cn, frozen=True, oldstyle_class=oldstyle_class
 *         )
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":211
 *         '''o-->object to be wrapped'''
 *         PrivacyDict.__init__(
 *             self, o, cn, frozen=True, oldstyle_class=oldstyle_class             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_o);
  __Pyx_INCREF(__pyx_v_cn);
  __Pyx_GIVEREF(__pyx_v_cn);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_cn);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_frozen, Py_True) < 0) __PYX_ERR(4, 211, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_oldstyle_class, __pyx_v_oldstyle_class) < 0) __PYX_ERR(4, 211, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":210
 *     def __init__(self, o, cn, oldstyle_class=None):
 *         '''o-->object to be wrapped'''
 *         PrivacyDict.__init__(             # <<<<<<<<<<<<<<
 *             self, o, cn, frozen=True, oldstyle_class=oldstyle_class
 *         )
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":208
 *     Subclass of PrivacyDict that is automatically frozen
 *     '''
 *     def __init__(self, o, cn, oldstyle_class=None):             # <<<<<<<<<<<<<<
 *         '''o-->object to be wrapped'''
 *         PrivacyDict.__init__(
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivacyDict.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":215
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_3__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_3__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_2__hash__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_2__hash__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_hash_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":216
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):
 *         return Wrapped.__hash__(self)             # <<<<<<<<<<<<<<
 * 
 *     # __richcmp__ needs to be class-specific
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_hash_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":215
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivacyDict.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "PrivacyDict_FrozenPrivacyDict.pxi":219
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_4__richcmp__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":221
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivacyDict *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.comparator(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "PrivacyDict_FrozenPrivacyDict.pxi":219
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivacyDict.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_17FrozenPrivacyDict_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_6__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 4, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->__pyx_base.__pyx_base.rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_FrozenPrivacyDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, None), state
 *     else:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FrozenPrivacyDict__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_FrozenPrivacyDict); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivacyDict.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivacyDict__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_17FrozenPrivacyDict_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivacyDict.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_8__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_17FrozenPrivacyDict_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FrozenPrivacyDict__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivacyDict__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivacyDict__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivacyDict.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":15
 *     '''
 * 
 *     def __init__(self, o, frozen=False, rules=None):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->object to be wrapped
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_7Private_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_7Private___init__, "\n        o-->object to be wrapped\n        frozen--bool: If True, no direct attribute can be modified\n        rules: dict (if called from Protected.__init__) or None\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_7Private___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_7Private_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_frozen = 0;
  PyObject *__pyx_v_rules = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,&__pyx_n_s_rules,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_frozen,&__pyx_n_s_rules,0};
    #endif
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)Py_False);
    values[2] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 15, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_frozen);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 15, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rules);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 15, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(7, 15, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_o = values[0];
    __pyx_v_frozen = values[1];
    __pyx_v_rules = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, __pyx_nargs); __PYX_ERR(7, 15, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Private.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private___init__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), __pyx_v_o, __pyx_v_frozen, __pyx_v_rules);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_7Private___init__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_frozen, PyObject *__pyx_v_rules) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Private_FrozenPrivate.pxi":21
 *         rules: dict (if called from Protected.__init__) or None
 *         '''
 *         Wrapped.__init__(self, o, frozen=frozen, rules=rules)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_o);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_frozen, __pyx_v_frozen) < 0) __PYX_ERR(7, 21, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_rules, __pyx_v_rules) < 0) __PYX_ERR(7, 21, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Private_FrozenPrivate.pxi":15
 *     '''
 * 
 *     def __init__(self, o, frozen=False, rules=None):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->object to be wrapped
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Private.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":27
 *     # --------------------------------------------------------------------
 * 
 *     cdef private_visible(self, a):             # <<<<<<<<<<<<<<
 *         '''Share with Private-derived'''
 *         if a in special_attributes:
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_visible(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("private_visible", 0);

  /* "Private_FrozenPrivate.pxi":29
 *     cdef private_visible(self, a):
 *         '''Share with Private-derived'''
 *         if a in special_attributes:             # <<<<<<<<<<<<<<
 *             return True
 *         if self.attr_hidden(a):
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_special_attributes, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(7, 29, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Private_FrozenPrivate.pxi":30
 *         '''Share with Private-derived'''
 *         if a in special_attributes:
 *             return True             # <<<<<<<<<<<<<<
 *         if self.attr_hidden(a):
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":29
 *     cdef private_visible(self, a):
 *         '''Share with Private-derived'''
 *         if a in special_attributes:             # <<<<<<<<<<<<<<
 *             return True
 *         if self.attr_hidden(a):
 */
  }

  /* "Private_FrozenPrivate.pxi":31
 *         if a in special_attributes:
 *             return True
 *         if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if a not in dir(self.pvt_o):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(7, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "Private_FrozenPrivate.pxi":32
 *             return True
 *         if self.attr_hidden(a):
 *             return False             # <<<<<<<<<<<<<<
 *         if a not in dir(self.pvt_o):
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":31
 *         if a in special_attributes:
 *             return True
 *         if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if a not in dir(self.pvt_o):
 */
  }

  /* "Private_FrozenPrivate.pxi":33
 *         if self.attr_hidden(a):
 *             return False
 *         if a not in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *             return False
 *         # Special case for PY2 that does not seem to obey __dir__ for modules
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Dir(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_t_3, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(7, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "Private_FrozenPrivate.pxi":34
 *             return False
 *         if a not in dir(self.pvt_o):
 *             return False             # <<<<<<<<<<<<<<
 *         # Special case for PY2 that does not seem to obey __dir__ for modules
 *         # Also applies to PY3 < 3.7
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":33
 *         if self.attr_hidden(a):
 *             return False
 *         if a not in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *             return False
 *         # Special case for PY2 that does not seem to obey __dir__ for modules
 */
  }

  /* "Private_FrozenPrivate.pxi":38
 *         # Also applies to PY3 < 3.7
 *         if (
 *             isinstance(self.pvt_o, types.ModuleType) and             # <<<<<<<<<<<<<<
 *             (
 *                 PY2 or (sys.version_info.major, sys.version_info.minor) < (3, 7)
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ModuleType); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_3, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(7, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L7_bool_binop_done;
  }

  /* "Private_FrozenPrivate.pxi":40
 *             isinstance(self.pvt_o, types.ModuleType) and
 *             (
 *                 PY2 or (sys.version_info.major, sys.version_info.minor) < (3, 7)             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  if (!__pyx_v_9pyprotect_9protected_PY2) {
  } else {
    __pyx_t_1 = __pyx_v_9pyprotect_9protected_PY2;
    goto __pyx_L7_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_version_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_major); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_version_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_minor); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_tuple__29, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L7_bool_binop_done:;

  /* "Private_FrozenPrivate.pxi":37
 *         # Special case for PY2 that does not seem to obey __dir__ for modules
 *         # Also applies to PY3 < 3.7
 *         if (             # <<<<<<<<<<<<<<
 *             isinstance(self.pvt_o, types.ModuleType) and
 *             (
 */
  if (__pyx_t_1) {

    /* "Private_FrozenPrivate.pxi":44
 *         ):
 *             if (
 *                 hasattr(self.pvt_o, '__dir__') and             # <<<<<<<<<<<<<<
 *                 callable(self.pvt_o.__dir__)
 *             ):
 */
    __pyx_t_3 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_HasAttr(__pyx_t_3, __pyx_n_s_dir); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(7, 44, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L11_bool_binop_done;
    }

    /* "Private_FrozenPrivate.pxi":45
 *             if (
 *                 hasattr(self.pvt_o, '__dir__') and
 *                 callable(self.pvt_o.__dir__)             # <<<<<<<<<<<<<<
 *             ):
 *                 if a not in self.pvt_o.__dir__():
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_n_s_dir); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyCallable_Check(__pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(7, 45, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_5;
    __pyx_L11_bool_binop_done:;

    /* "Private_FrozenPrivate.pxi":43
 *             )
 *         ):
 *             if (             # <<<<<<<<<<<<<<
 *                 hasattr(self.pvt_o, '__dir__') and
 *                 callable(self.pvt_o.__dir__)
 */
    if (__pyx_t_1) {

      /* "Private_FrozenPrivate.pxi":47
 *                 callable(self.pvt_o.__dir__)
 *             ):
 *                 if a not in self.pvt_o.__dir__():             # <<<<<<<<<<<<<<
 *                     return False
 *         return True
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.pvt_o, __pyx_n_s_dir); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[1] = {__pyx_t_4, };
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_t_3, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(7, 47, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_1) {

        /* "Private_FrozenPrivate.pxi":48
 *             ):
 *                 if a not in self.pvt_o.__dir__():
 *                     return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L0;

        /* "Private_FrozenPrivate.pxi":47
 *                 callable(self.pvt_o.__dir__)
 *             ):
 *                 if a not in self.pvt_o.__dir__():             # <<<<<<<<<<<<<<
 *                     return False
 *         return True
 */
      }

      /* "Private_FrozenPrivate.pxi":43
 *             )
 *         ):
 *             if (             # <<<<<<<<<<<<<<
 *                 hasattr(self.pvt_o, '__dir__') and
 *                 callable(self.pvt_o.__dir__)
 */
    }

    /* "Private_FrozenPrivate.pxi":37
 *         # Special case for PY2 that does not seem to obey __dir__ for modules
 *         # Also applies to PY3 < 3.7
 *         if (             # <<<<<<<<<<<<<<
 *             isinstance(self.pvt_o, types.ModuleType) and
 *             (
 */
  }

  /* "Private_FrozenPrivate.pxi":49
 *                 if a not in self.pvt_o.__dir__():
 *                     return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef private_writeable(self, a):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":27
 *     # --------------------------------------------------------------------
 * 
 *     cdef private_visible(self, a):             # <<<<<<<<<<<<<<
 *         '''Share with Private-derived'''
 *         if a in special_attributes:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Private.private_visible", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":51
 *         return True
 * 
 *     cdef private_writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Shared with Private-derived
 *         # writeable implies visible. not visible implies not writeable
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_writeable(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("private_writeable", 0);

  /* "Private_FrozenPrivate.pxi":54
 *         # Shared with Private-derived
 *         # writeable implies visible. not visible implies not writeable
 *         if not self.visible(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if ro_private_attr.match(a):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.visible(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(7, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "Private_FrozenPrivate.pxi":55
 *         # writeable implies visible. not visible implies not writeable
 *         if not self.visible(a):
 *             return False             # <<<<<<<<<<<<<<
 *         if ro_private_attr.match(a):
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":54
 *         # Shared with Private-derived
 *         # writeable implies visible. not visible implies not writeable
 *         if not self.visible(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if ro_private_attr.match(a):
 */
  }

  /* "Private_FrozenPrivate.pxi":56
 *         if not self.visible(a):
 *             return False
 *         if ro_private_attr.match(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if a in special_attributes:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_ro_private_attr, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_a};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(7, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Private_FrozenPrivate.pxi":57
 *             return False
 *         if ro_private_attr.match(a):
 *             return False             # <<<<<<<<<<<<<<
 *         if a in special_attributes:
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":56
 *         if not self.visible(a):
 *             return False
 *         if ro_private_attr.match(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if a in special_attributes:
 */
  }

  /* "Private_FrozenPrivate.pxi":58
 *         if ro_private_attr.match(a):
 *             return False
 *         if a in special_attributes:             # <<<<<<<<<<<<<<
 *             return False
 *         if a in always_frozen:
 */
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_special_attributes, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(7, 58, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Private_FrozenPrivate.pxi":59
 *             return False
 *         if a in special_attributes:
 *             return False             # <<<<<<<<<<<<<<
 *         if a in always_frozen:
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":58
 *         if ro_private_attr.match(a):
 *             return False
 *         if a in special_attributes:             # <<<<<<<<<<<<<<
 *             return False
 *         if a in always_frozen:
 */
  }

  /* "Private_FrozenPrivate.pxi":60
 *         if a in special_attributes:
 *             return False
 *         if a in always_frozen:             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_always_frozen, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(7, 60, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Private_FrozenPrivate.pxi":61
 *             return False
 *         if a in always_frozen:
 *             return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":60
 *         if a in special_attributes:
 *             return False
 *         if a in always_frozen:             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  }

  /* "Private_FrozenPrivate.pxi":62
 *         if a in always_frozen:
 *             return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef visible(self, a):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":51
 *         return True
 * 
 *     cdef private_writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Shared with Private-derived
 *         # writeable implies visible. not visible implies not writeable
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Private.private_writeable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":64
 *         return True
 * 
 *     cdef visible(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         return self.private_visible(a)
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_visible(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visible", 0);

  /* "Private_FrozenPrivate.pxi":66
 *     cdef visible(self, a):
 *         # Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         return self.private_visible(a)             # <<<<<<<<<<<<<<
 * 
 *     cdef writeable(self, a):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->private_visible(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":64
 *         return True
 * 
 *     cdef visible(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         return self.private_visible(a)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Private.visible", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":68
 *         return self.private_visible(a)
 * 
 *     cdef writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         return self.private_writeable(a)
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_writeable(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeable", 0);

  /* "Private_FrozenPrivate.pxi":70
 *     cdef writeable(self, a):
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         return self.private_writeable(a)             # <<<<<<<<<<<<<<
 * 
 *     cdef private_getattr(self, a):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->private_writeable(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":68
 *         return self.private_visible(a)
 * 
 *     cdef writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         return self.private_writeable(a)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Private.writeable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":72
 *         return self.private_writeable(a)
 * 
 *     cdef private_getattr(self, a):             # <<<<<<<<<<<<<<
 *         # Cannot access any attribute not exported by dir(pvt_o)
 *         # cannot access any unmangled double '_' attributes
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_getattr(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  struct __pyx_opt_args_9pyprotect_9protected_privatedict __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("private_getattr", 0);

  /* "Private_FrozenPrivate.pxi":75
 *         # Cannot access any attribute not exported by dir(pvt_o)
 *         # cannot access any unmangled double '_' attributes
 *         if not self.visible(a):             # <<<<<<<<<<<<<<
 *             raise AttributeError(
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.visible(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(7, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (unlikely(__pyx_t_3)) {

    /* "Private_FrozenPrivate.pxi":77
 *         if not self.visible(a):
 *             raise AttributeError(
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)             # <<<<<<<<<<<<<<
 *             )
 *         if a in overridden_always:
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
    __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.cn);
    __Pyx_INCREF(__pyx_v_a);
    __Pyx_GIVEREF(__pyx_v_a);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_a);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Object_Private_s_has_no_attribut, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Private_FrozenPrivate.pxi":76
 *         # cannot access any unmangled double '_' attributes
 *         if not self.visible(a):
 *             raise AttributeError(             # <<<<<<<<<<<<<<
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 *             )
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(7, 76, __pyx_L1_error)

    /* "Private_FrozenPrivate.pxi":75
 *         # Cannot access any attribute not exported by dir(pvt_o)
 *         # cannot access any unmangled double '_' attributes
 *         if not self.visible(a):             # <<<<<<<<<<<<<<
 *             raise AttributeError(
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 */
  }

  /* "Private_FrozenPrivate.pxi":79
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 *             )
 *         if a in overridden_always:             # <<<<<<<<<<<<<<
 *             return functools.partial(getattr(Private, a), self)
 * 
 */
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_overridden_always, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(7, 79, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Private_FrozenPrivate.pxi":80
 *             )
 *         if a in overridden_always:
 *             return functools.partial(getattr(Private, a), self)             # <<<<<<<<<<<<<<
 * 
 *         if a in always_frozen:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_functools); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_partial); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Private), __pyx_v_a); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_4, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Private_FrozenPrivate.pxi":79
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 *             )
 *         if a in overridden_always:             # <<<<<<<<<<<<<<
 *             return functools.partial(getattr(Private, a), self)
 * 
 */
  }

  /* "Private_FrozenPrivate.pxi":82
 *             return functools.partial(getattr(Private, a), self)
 * 
 *         if a in always_frozen:             # <<<<<<<<<<<<<<
 *             x = getattr(self.pvt_o, a)
 *             if a == '__dict__':
 */
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_always_frozen, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(7, 82, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Private_FrozenPrivate.pxi":83
 * 
 *         if a in always_frozen:
 *             x = getattr(self.pvt_o, a)             # <<<<<<<<<<<<<<
 *             if a == '__dict__':
 *                 return privatedict(x, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 */
    __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_a); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_x = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Private_FrozenPrivate.pxi":84
 *         if a in always_frozen:
 *             x = getattr(self.pvt_o, a)
 *             if a == '__dict__':             # <<<<<<<<<<<<<<
 *                 return privatedict(x, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 *             else:
 */
    __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_a, __pyx_n_s_dict_2, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(7, 84, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "Private_FrozenPrivate.pxi":85
 *             x = getattr(self.pvt_o, a)
 *             if a == '__dict__':
 *                 return privatedict(x, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)             # <<<<<<<<<<<<<<
 *             else:
 *                 return freeze(x)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __pyx_v_self->__pyx_base.cn;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8.__pyx_n = 2;
      __pyx_t_8.frozen = Py_True;
      __pyx_t_8.oldstyle_class = __pyx_t_1;
      __pyx_t_4 = __pyx_f_9pyprotect_9protected_privatedict(__pyx_v_x, __pyx_t_5, &__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "Private_FrozenPrivate.pxi":84
 *         if a in always_frozen:
 *             x = getattr(self.pvt_o, a)
 *             if a == '__dict__':             # <<<<<<<<<<<<<<
 *                 return privatedict(x, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 *             else:
 */
    }

    /* "Private_FrozenPrivate.pxi":87
 *                 return privatedict(x, self.cn, frozen=True, oldstyle_class=self.oldstyle_class)
 *             else:
 *                 return freeze(x)             # <<<<<<<<<<<<<<
 *         return self.wrapped_getattr(a)
 * 
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_freeze); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_x};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 87, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;
    }

    /* "Private_FrozenPrivate.pxi":82
 *             return functools.partial(getattr(Private, a), self)
 * 
 *         if a in always_frozen:             # <<<<<<<<<<<<<<
 *             x = getattr(self.pvt_o, a)
 *             if a == '__dict__':
 */
  }

  /* "Private_FrozenPrivate.pxi":88
 *             else:
 *                 return freeze(x)
 *         return self.wrapped_getattr(a)             # <<<<<<<<<<<<<<
 * 
 *     cdef private_check_setattr(self, a, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.wrapped_getattr(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":72
 *         return self.private_writeable(a)
 * 
 *     cdef private_getattr(self, a):             # <<<<<<<<<<<<<<
 *         # Cannot access any attribute not exported by dir(pvt_o)
 *         # cannot access any unmangled double '_' attributes
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.Private.private_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":90
 *         return self.wrapped_getattr(a)
 * 
 *     cdef private_check_setattr(self, a, val):             # <<<<<<<<<<<<<<
 *         nopvt_msg = 'Cannot set attribute: %s.%s' % (self.cn, str(a))
 *         noadd_msg = 'Cannot add attribute: %s.%s' % (self.cn, str(a))
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_check_setattr(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  PyObject *__pyx_v_nopvt_msg = NULL;
  PyObject *__pyx_v_noadd_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("private_check_setattr", 0);

  /* "Private_FrozenPrivate.pxi":91
 * 
 *     cdef private_check_setattr(self, a, val):
 *         nopvt_msg = 'Cannot set attribute: %s.%s' % (self.cn, str(a))             # <<<<<<<<<<<<<<
 *         noadd_msg = 'Cannot add attribute: %s.%s' % (self.cn, str(a))
 *         if not self.writeable(a):
 */
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_set_attribute_s_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nopvt_msg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Private_FrozenPrivate.pxi":92
 *     cdef private_check_setattr(self, a, val):
 *         nopvt_msg = 'Cannot set attribute: %s.%s' % (self.cn, str(a))
 *         noadd_msg = 'Cannot add attribute: %s.%s' % (self.cn, str(a))             # <<<<<<<<<<<<<<
 *         if not self.writeable(a):
 *             raise ProtectionError(nopvt_msg)
 */
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_add_attribute_s_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_noadd_msg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Private_FrozenPrivate.pxi":93
 *         nopvt_msg = 'Cannot set attribute: %s.%s' % (self.cn, str(a))
 *         noadd_msg = 'Cannot add attribute: %s.%s' % (self.cn, str(a))
 *         if not self.writeable(a):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nopvt_msg)
 *         if a not in dir(self.pvt_o):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.writeable(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(7, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (!__pyx_t_3);
  if (unlikely(__pyx_t_4)) {

    /* "Private_FrozenPrivate.pxi":94
 *         noadd_msg = 'Cannot add attribute: %s.%s' % (self.cn, str(a))
 *         if not self.writeable(a):
 *             raise ProtectionError(nopvt_msg)             # <<<<<<<<<<<<<<
 *         if a not in dir(self.pvt_o):
 *             raise ProtectionError(noadd_msg)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_nopvt_msg};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(7, 94, __pyx_L1_error)

    /* "Private_FrozenPrivate.pxi":93
 *         nopvt_msg = 'Cannot set attribute: %s.%s' % (self.cn, str(a))
 *         noadd_msg = 'Cannot add attribute: %s.%s' % (self.cn, str(a))
 *         if not self.writeable(a):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(nopvt_msg)
 *         if a not in dir(self.pvt_o):
 */
  }

  /* "Private_FrozenPrivate.pxi":95
 *         if not self.writeable(a):
 *             raise ProtectionError(nopvt_msg)
 *         if a not in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(noadd_msg)
 *         self.wrapped_check_setattr(a, val)
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Dir(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_t_2, Py_NE)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(7, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "Private_FrozenPrivate.pxi":96
 *             raise ProtectionError(nopvt_msg)
 *         if a not in dir(self.pvt_o):
 *             raise ProtectionError(noadd_msg)             # <<<<<<<<<<<<<<
 *         self.wrapped_check_setattr(a, val)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_noadd_msg};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(7, 96, __pyx_L1_error)

    /* "Private_FrozenPrivate.pxi":95
 *         if not self.writeable(a):
 *             raise ProtectionError(nopvt_msg)
 *         if a not in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *             raise ProtectionError(noadd_msg)
 *         self.wrapped_check_setattr(a, val)
 */
  }

  /* "Private_FrozenPrivate.pxi":97
 *         if a not in dir(self.pvt_o):
 *             raise ProtectionError(noadd_msg)
 *         self.wrapped_check_setattr(a, val)             # <<<<<<<<<<<<<<
 * 
 *     cdef private_check_delattr(self, a):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.wrapped_check_setattr(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a, __pyx_v_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Private_FrozenPrivate.pxi":90
 *         return self.wrapped_getattr(a)
 * 
 *     cdef private_check_setattr(self, a, val):             # <<<<<<<<<<<<<<
 *         nopvt_msg = 'Cannot set attribute: %s.%s' % (self.cn, str(a))
 *         noadd_msg = 'Cannot add attribute: %s.%s' % (self.cn, str(a))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Private.private_check_setattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nopvt_msg);
  __Pyx_XDECREF(__pyx_v_noadd_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":99
 *         self.wrapped_check_setattr(a, val)
 * 
 *     cdef private_check_delattr(self, a):             # <<<<<<<<<<<<<<
 *         nodel_msg = 'Cannot delete attribute: %s.%s' % (self.cn, str(a))
 *         if not hasattr(self.pvt_o, a):
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_check_delattr(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_v_nodel_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("private_check_delattr", 0);

  /* "Private_FrozenPrivate.pxi":100
 * 
 *     cdef private_check_delattr(self, a):
 *         nodel_msg = 'Cannot delete attribute: %s.%s' % (self.cn, str(a))             # <<<<<<<<<<<<<<
 *         if not hasattr(self.pvt_o, a):
 *             raise AttributeError(
 */
  __pyx_t_1 = __Pyx_PyObject_Str(__pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_delete_attribute_s_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nodel_msg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Private_FrozenPrivate.pxi":101
 *     cdef private_check_delattr(self, a):
 *         nodel_msg = 'Cannot delete attribute: %s.%s' % (self.cn, str(a))
 *         if not hasattr(self.pvt_o, a):             # <<<<<<<<<<<<<<
 *             raise AttributeError(
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_HasAttr(__pyx_t_1, __pyx_v_a); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(7, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (!__pyx_t_3);
  if (unlikely(__pyx_t_4)) {

    /* "Private_FrozenPrivate.pxi":103
 *         if not hasattr(self.pvt_o, a):
 *             raise AttributeError(
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)             # <<<<<<<<<<<<<<
 *             )
 *         raise ProtectionError(nodel_msg)
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
    __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.cn);
    __Pyx_INCREF(__pyx_v_a);
    __Pyx_GIVEREF(__pyx_v_a);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_a);
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Object_Private_s_has_no_attribut, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Private_FrozenPrivate.pxi":102
 *         nodel_msg = 'Cannot delete attribute: %s.%s' % (self.cn, str(a))
 *         if not hasattr(self.pvt_o, a):
 *             raise AttributeError(             # <<<<<<<<<<<<<<
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 *             )
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(7, 102, __pyx_L1_error)

    /* "Private_FrozenPrivate.pxi":101
 *     cdef private_check_delattr(self, a):
 *         nodel_msg = 'Cannot delete attribute: %s.%s' % (self.cn, str(a))
 *         if not hasattr(self.pvt_o, a):             # <<<<<<<<<<<<<<
 *             raise AttributeError(
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 */
  }

  /* "Private_FrozenPrivate.pxi":105
 *                 "Object Private('%s') has no attribute '%s'" % (self.cn, a)
 *             )
 *         raise ProtectionError(nodel_msg)             # <<<<<<<<<<<<<<
 * 
 *     cdef private_dir(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_nodel_msg};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(7, 105, __pyx_L1_error)

  /* "Private_FrozenPrivate.pxi":99
 *         self.wrapped_check_setattr(a, val)
 * 
 *     cdef private_check_delattr(self, a):             # <<<<<<<<<<<<<<
 *         nodel_msg = 'Cannot delete attribute: %s.%s' % (self.cn, str(a))
 *         if not hasattr(self.pvt_o, a):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Private.private_check_delattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XDECREF(__pyx_v_nodel_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":107
 *         raise ProtectionError(nodel_msg)
 * 
 *     cdef private_dir(self):             # <<<<<<<<<<<<<<
 *         return [
 *             x for x in self.wrapped_dir()
 */

static PyObject *__pyx_f_9pyprotect_9protected_7Private_private_dir(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self) {
  PyObject *__pyx_9genexpr16__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("private_dir", 0);

  /* "Private_FrozenPrivate.pxi":108
 * 
 *     cdef private_dir(self):
 *         return [             # <<<<<<<<<<<<<<
 *             x for x in self.wrapped_dir()
 *             if self.private_visible(x)
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 108, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "Private_FrozenPrivate.pxi":109
 *     cdef private_dir(self):
 *         return [
 *             x for x in self.wrapped_dir()             # <<<<<<<<<<<<<<
 *             if self.private_visible(x)
 *         ]
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.wrapped_dir(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 109, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 109, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 109, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(7, 109, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 109, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(7, 109, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 109, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(7, 109, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_x, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Private_FrozenPrivate.pxi":110
 *         return [
 *             x for x in self.wrapped_dir()
 *             if self.private_visible(x)             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->private_visible(__pyx_v_self, __pyx_9genexpr16__pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 110, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(7, 110, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {

        /* "Private_FrozenPrivate.pxi":109
 *     cdef private_dir(self):
 *         return [
 *             x for x in self.wrapped_dir()             # <<<<<<<<<<<<<<
 *             if self.private_visible(x)
 *         ]
 */
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr16__pyx_v_x))) __PYX_ERR(7, 108, __pyx_L5_error)

        /* "Private_FrozenPrivate.pxi":110
 *         return [
 *             x for x in self.wrapped_dir()
 *             if self.private_visible(x)             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      }

      /* "Private_FrozenPrivate.pxi":109
 *     cdef private_dir(self):
 *         return [
 *             x for x in self.wrapped_dir()             # <<<<<<<<<<<<<<
 *             if self.private_visible(x)
 *         ]
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_x); __pyx_9genexpr16__pyx_v_x = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_x); __pyx_9genexpr16__pyx_v_x = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":107
 *         raise ProtectionError(nodel_msg)
 * 
 *     cdef private_dir(self):             # <<<<<<<<<<<<<<
 *         return [
 *             x for x in self.wrapped_dir()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Private.private_dir", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":117
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.private_getattr(a)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattribute__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_2__getattribute__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Private_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattribute__", 0);

  /* "Private_FrozenPrivate.pxi":118
 * 
 *     def __getattribute__(self, a):
 *         return self.private_getattr(a)             # <<<<<<<<<<<<<<
 * 
 *     def __setattr__(self, a, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->private_getattr(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":117
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.private_getattr(a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Private.__getattribute__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":120
 *         return self.private_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.private_check_setattr(a, val)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_7Private_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_7Private_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_4__setattr__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), ((PyObject *)__pyx_v_a), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_7Private_4__setattr__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setattr__", 0);

  /* "Private_FrozenPrivate.pxi":122
 *     def __setattr__(self, a, val):
 *         # Only checks and raises exceptions
 *         self.private_check_setattr(a, val)             # <<<<<<<<<<<<<<
 *         setattr(self.pvt_o, a, val)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->private_check_setattr(__pyx_v_self, __pyx_v_a, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Private_FrozenPrivate.pxi":123
 *         # Only checks and raises exceptions
 *         self.private_check_setattr(a, val)
 *         setattr(self.pvt_o, a, val)             # <<<<<<<<<<<<<<
 * 
 *     def __delattr__(self, a):
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_SetAttr(__pyx_t_1, __pyx_v_a, __pyx_v_val); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(7, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Private_FrozenPrivate.pxi":120
 *         return self.private_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.private_check_setattr(a, val)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Private.__setattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":125
 *         setattr(self.pvt_o, a, val)
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.private_check_delattr(a)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_7Private_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static int __pyx_pw_9pyprotect_9protected_7Private_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_6__delattr__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_7Private_6__delattr__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delattr__", 0);

  /* "Private_FrozenPrivate.pxi":127
 *     def __delattr__(self, a):
 *         # Only checks and raises exceptions
 *         self.private_check_delattr(a)             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->private_check_delattr(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Private_FrozenPrivate.pxi":125
 *         setattr(self.pvt_o, a, val)
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.private_check_delattr(a)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Private.__delattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":129
 *         self.private_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.private_dir()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7Private_9__dir__ = {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Private_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__dir__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__dir__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_8__dir__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Private_8__dir__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "Private_FrozenPrivate.pxi":130
 * 
 *     def __dir__(self):
 *         return self.private_dir()             # <<<<<<<<<<<<<<
 * 
 *     # Python / cython does not automatically use parent __hash__
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->private_dir(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":129
 *         self.private_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.private_dir()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Private.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":133
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_7Private_11__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_7Private_11__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_10__hash__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_7Private_10__hash__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_hash_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Private_FrozenPrivate.pxi":134
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):
 *         return Wrapped.__hash__(self)             # <<<<<<<<<<<<<<
 * 
 *     # __richcmp__ needs to be class-specific
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_hash_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(7, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":133
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Private.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":137
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_13__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_13__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_12__richcmp__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Private_12__richcmp__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "Private_FrozenPrivate.pxi":139
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Private *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.comparator(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":137
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.Private.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7Private_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Private_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_14__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Private_14__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 4, ((PyObject *)__pyx_v_self->__pyx_base.protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base.__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.rules);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.rules);
  PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->__pyx_base.rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Private); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, None), state
 *     else:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Private__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Private); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Private.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Private__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_7Private_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Private_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_7Private_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Private.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_7Private_16__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_7Private_16__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Private__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Private__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Private__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Private.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":148
 *     Subclass of Private that is automatically frozen
 *     '''
 *     def __init__(self, o):             # <<<<<<<<<<<<<<
 *         '''o-->object to be wrapped'''
 *         Private.__init__(self, o, frozen=True)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_13FrozenPrivate_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_13FrozenPrivate___init__, "o-->object to be wrapped");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_13FrozenPrivate___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_13FrozenPrivate_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 148, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(7, 148, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_o = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(7, 148, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivate.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_13FrozenPrivate___init__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *)__pyx_v_self), __pyx_v_o);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_13FrozenPrivate___init__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self, PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Private_FrozenPrivate.pxi":150
 *     def __init__(self, o):
 *         '''o-->object to be wrapped'''
 *         Private.__init__(self, o, frozen=True)             # <<<<<<<<<<<<<<
 * 
 *     # Python / cython does not automatically use parent __hash__
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Private), __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_o);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_frozen, Py_True) < 0) __PYX_ERR(7, 150, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Private_FrozenPrivate.pxi":148
 *     Subclass of Private that is automatically frozen
 *     '''
 *     def __init__(self, o):             # <<<<<<<<<<<<<<
 *         '''o-->object to be wrapped'''
 *         Private.__init__(self, o, frozen=True)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivate.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":153
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_13FrozenPrivate_3__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_13FrozenPrivate_3__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_13FrozenPrivate_2__hash__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_13FrozenPrivate_2__hash__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_hash_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Private_FrozenPrivate.pxi":154
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):
 *         return Wrapped.__hash__(self)             # <<<<<<<<<<<<<<
 * 
 *     # __richcmp__ needs to be class-specific
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_hash_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(7, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":153
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivate.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Private_FrozenPrivate.pxi":157
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_13FrozenPrivate_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_13FrozenPrivate_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_13FrozenPrivate_4__richcmp__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_13FrozenPrivate_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "Private_FrozenPrivate.pxi":159
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivate *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.comparator(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Private_FrozenPrivate.pxi":157
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivate.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_13FrozenPrivate_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_13FrozenPrivate_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_13FrozenPrivate_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_13FrozenPrivate_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_13FrozenPrivate_6__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_13FrozenPrivate_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 4, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->__pyx_base.__pyx_base.rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cn, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_FrozenPrivate); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.cn is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, None), state
 *     else:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FrozenPrivate__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_FrozenPrivate); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_209109265);
    __Pyx_GIVEREF(__pyx_int_209109265);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_209109265);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivate.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivate__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_13FrozenPrivate_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_13FrozenPrivate_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_13FrozenPrivate_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_13FrozenPrivate_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivate.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_13FrozenPrivate_8__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_13FrozenPrivate_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FrozenPrivate__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivate__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivate__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.FrozenPrivate.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":13
 *     cdef list dir_out
 * 
 *     def __init__(self, o, rules):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->object to be wrapped
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_9Protected_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_9Protected___init__, "\n        o-->object to be wrapped\n        rules-->dict: returned by protected_rules_from_kwargs\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_9Protected___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_9Protected_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_rules = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_rules,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_rules,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 13, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rules)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 13, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(8, 13, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(8, 13, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_o = values[0];
    __pyx_v_rules = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(8, 13, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Protected.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected___init__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self), __pyx_v_o, __pyx_v_rules);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_9Protected___init__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_rules) {
  PyObject *__pyx_v_frozen = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Protected_FrozenProtected.pxi":18
 *         rules-->dict: returned by protected_rules_from_kwargs
 *         '''
 *         self.rules = rules             # <<<<<<<<<<<<<<
 *         frozen = bool(rules.get('frozen', False))
 *         Private.__init__(self, o, frozen=frozen)
 */
  if (!(likely(PyDict_CheckExact(__pyx_v_rules))||((__pyx_v_rules) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_rules))) __PYX_ERR(8, 18, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_rules;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  __pyx_v_self->__pyx_base.__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":19
 *         '''
 *         self.rules = rules
 *         frozen = bool(rules.get('frozen', False))             # <<<<<<<<<<<<<<
 *         Private.__init__(self, o, frozen=frozen)
 *         self.frozen = frozen
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rules, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_frozen = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Protected_FrozenProtected.pxi":20
 *         self.rules = rules
 *         frozen = bool(rules.get('frozen', False))
 *         Private.__init__(self, o, frozen=frozen)             # <<<<<<<<<<<<<<
 *         self.frozen = frozen
 *         self.process_rules(rules)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Private), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_o);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_frozen, __pyx_v_frozen) < 0) __PYX_ERR(8, 20, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Protected_FrozenProtected.pxi":21
 *         frozen = bool(rules.get('frozen', False))
 *         Private.__init__(self, o, frozen=frozen)
 *         self.frozen = frozen             # <<<<<<<<<<<<<<
 *         self.process_rules(rules)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_frozen); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(8, 21, __pyx_L1_error)
  __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.frozen = __pyx_t_3;

  /* "Protected_FrozenProtected.pxi":22
 *         Private.__init__(self, o, frozen=frozen)
 *         self.frozen = frozen
 *         self.process_rules(rules)             # <<<<<<<<<<<<<<
 * 
 *     # --------------------------------------------------------------------
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->process_rules(__pyx_v_self, __pyx_v_rules); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Protected_FrozenProtected.pxi":13
 *     cdef list dir_out
 * 
 *     def __init__(self, o, rules):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->object to be wrapped
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frozen);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":28
 *     # --------------------------------------------------------------------
 * 
 *     cdef get_rules(self):             # <<<<<<<<<<<<<<
 *         if self.rules is None:
 *             return dict()
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_get_rules(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_rules", 0);

  /* "Protected_FrozenProtected.pxi":29
 * 
 *     cdef get_rules(self):
 *         if self.rules is None:             # <<<<<<<<<<<<<<
 *             return dict()
 *         return self.rules
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.__pyx_base.rules == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "Protected_FrozenProtected.pxi":30
 *     cdef get_rules(self):
 *         if self.rules is None:
 *             return dict()             # <<<<<<<<<<<<<<
 *         return self.rules
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":29
 * 
 *     cdef get_rules(self):
 *         if self.rules is None:             # <<<<<<<<<<<<<<
 *             return dict()
 *         return self.rules
 */
  }

  /* "Protected_FrozenProtected.pxi":31
 *         if self.rules is None:
 *             return dict()
 *         return self.rules             # <<<<<<<<<<<<<<
 * 
 *     cdef process_rules(self, rules):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  __pyx_r = __pyx_v_self->__pyx_base.__pyx_base.rules;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":28
 *     # --------------------------------------------------------------------
 * 
 *     cdef get_rules(self):             # <<<<<<<<<<<<<<
 *         if self.rules is None:
 *             return dict()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.Protected.get_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":33
 *         return self.rules
 * 
 *     cdef process_rules(self, rules):             # <<<<<<<<<<<<<<
 *         '''
 *         rules-->dict
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_process_rules(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_rules) {
  PyObject *__pyx_9genexpr17__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr17__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("process_rules", 0);

  /* "Protected_FrozenProtected.pxi":38
 *         Called once at object wrapping time
 *         '''
 *         self.dir_out = []             # <<<<<<<<<<<<<<
 *         # frozen does NOT override dynamic
 *         if bool(rules.get('dynamic', False)):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->dir_out);
  __Pyx_DECREF(__pyx_v_self->dir_out);
  __pyx_v_self->dir_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":40
 *         self.dir_out = []
 *         # frozen does NOT override dynamic
 *         if bool(rules.get('dynamic', False)):             # <<<<<<<<<<<<<<
 *             self.acl_cache = None
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rules, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((!(!__pyx_t_3))) {

    /* "Protected_FrozenProtected.pxi":41
 *         # frozen does NOT override dynamic
 *         if bool(rules.get('dynamic', False)):
 *             self.acl_cache = None             # <<<<<<<<<<<<<<
 *         else:
 *             self.acl_cache = {}
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->acl_cache);
    __Pyx_DECREF(__pyx_v_self->acl_cache);
    __pyx_v_self->acl_cache = ((PyObject*)Py_None);

    /* "Protected_FrozenProtected.pxi":40
 *         self.dir_out = []
 *         # frozen does NOT override dynamic
 *         if bool(rules.get('dynamic', False)):             # <<<<<<<<<<<<<<
 *             self.acl_cache = None
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Protected_FrozenProtected.pxi":43
 *             self.acl_cache = None
 *         else:
 *             self.acl_cache = {}             # <<<<<<<<<<<<<<
 *             self.build_cache()
 *             # Make dir() pre-computed
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->acl_cache);
    __Pyx_DECREF(__pyx_v_self->acl_cache);
    __pyx_v_self->acl_cache = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Protected_FrozenProtected.pxi":44
 *         else:
 *             self.acl_cache = {}
 *             self.build_cache()             # <<<<<<<<<<<<<<
 *             # Make dir() pre-computed
 *             self.dir_out = [
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->build_cache(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Protected_FrozenProtected.pxi":46
 *             self.build_cache()
 *             # Make dir() pre-computed
 *             self.dir_out = [             # <<<<<<<<<<<<<<
 *                     k for (k, v) in self.acl_cache.items()
 *                     if v.get('r', False)
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 46, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "Protected_FrozenProtected.pxi":47
 *             # Make dir() pre-computed
 *             self.dir_out = [
 *                     k for (k, v) in self.acl_cache.items()             # <<<<<<<<<<<<<<
 *                     if v.get('r', False)
 *             ]
 */
      __pyx_t_4 = 0;
      if (unlikely(__pyx_v_self->acl_cache == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(8, 47, __pyx_L6_error)
      }
      __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_self->acl_cache, 1, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(8, 47, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_7;
      __pyx_t_7 = 0;
      while (1) {
        __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_5, &__pyx_t_4, &__pyx_t_7, &__pyx_t_8, NULL, __pyx_t_6);
        if (unlikely(__pyx_t_9 == 0)) break;
        if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(8, 47, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_k, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_v, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "Protected_FrozenProtected.pxi":48
 *             self.dir_out = [
 *                     k for (k, v) in self.acl_cache.items()
 *                     if v.get('r', False)             # <<<<<<<<<<<<<<
 *             ]
 * 
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr17__pyx_v_v, __pyx_n_s_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(8, 48, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(8, 48, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 48, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_3) {

          /* "Protected_FrozenProtected.pxi":47
 *             # Make dir() pre-computed
 *             self.dir_out = [
 *                     k for (k, v) in self.acl_cache.items()             # <<<<<<<<<<<<<<
 *                     if v.get('r', False)
 *             ]
 */
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_9genexpr17__pyx_v_k))) __PYX_ERR(8, 46, __pyx_L6_error)

          /* "Protected_FrozenProtected.pxi":48
 *             self.dir_out = [
 *                     k for (k, v) in self.acl_cache.items()
 *                     if v.get('r', False)             # <<<<<<<<<<<<<<
 *             ]
 * 
 */
        }
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_k); __pyx_9genexpr17__pyx_v_k = 0;
      __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_v); __pyx_9genexpr17__pyx_v_v = 0;
      goto __pyx_L10_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_k); __pyx_9genexpr17__pyx_v_k = 0;
      __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_v); __pyx_9genexpr17__pyx_v_v = 0;
      goto __pyx_L1_error;
      __pyx_L10_exit_scope:;
    } /* exit inner scope */

    /* "Protected_FrozenProtected.pxi":46
 *             self.build_cache()
 *             # Make dir() pre-computed
 *             self.dir_out = [             # <<<<<<<<<<<<<<
 *                     k for (k, v) in self.acl_cache.items()
 *                     if v.get('r', False)
 */
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->dir_out);
    __Pyx_DECREF(__pyx_v_self->dir_out);
    __pyx_v_self->dir_out = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "Protected_FrozenProtected.pxi":33
 *         return self.rules
 * 
 *     cdef process_rules(self, rules):             # <<<<<<<<<<<<<<
 *         '''
 *         rules-->dict
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyprotect.protected.Protected.process_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":51
 *             ]
 * 
 *     cdef build_cache(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Called once at object wrapping time
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_build_cache(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self) {
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_hidden_d = NULL;
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_visible __pyx_t_7;
  struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_writeable __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build_cache", 0);

  /* "Protected_FrozenProtected.pxi":55
 *         Called once at object wrapping time
 *         '''
 *         if self.acl_cache is None:             # <<<<<<<<<<<<<<
 *             return
 *         d = self.acl_cache
 */
  __pyx_t_1 = (__pyx_v_self->acl_cache == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "Protected_FrozenProtected.pxi":56
 *         '''
 *         if self.acl_cache is None:
 *             return             # <<<<<<<<<<<<<<
 *         d = self.acl_cache
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":55
 *         Called once at object wrapping time
 *         '''
 *         if self.acl_cache is None:             # <<<<<<<<<<<<<<
 *             return
 *         d = self.acl_cache
 */
  }

  /* "Protected_FrozenProtected.pxi":57
 *         if self.acl_cache is None:
 *             return
 *         d = self.acl_cache             # <<<<<<<<<<<<<<
 * 
 *         hidden_d = {'r': False, 'w': False}
 */
  __pyx_t_2 = __pyx_v_self->acl_cache;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_d = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Protected_FrozenProtected.pxi":59
 *         d = self.acl_cache
 * 
 *         hidden_d = {'r': False, 'w': False}             # <<<<<<<<<<<<<<
 * 
 *         for a in dir(self.pvt_o):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_r, Py_False) < 0) __PYX_ERR(8, 59, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_w, Py_False) < 0) __PYX_ERR(8, 59, __pyx_L1_error)
  __pyx_v_hidden_d = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Protected_FrozenProtected.pxi":61
 *         hidden_d = {'r': False, 'w': False}
 * 
 *         for a in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *             if self.attr_hidden(a):
 *                 self.acl_cache[a] = hidden_d
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Dir(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 61, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(8, 61, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 61, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(8, 61, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 61, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(8, 61, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_a, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Protected_FrozenProtected.pxi":62
 * 
 *         for a in dir(self.pvt_o):
 *             if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *                 self.acl_cache[a] = hidden_d
 *                 continue
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(8, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "Protected_FrozenProtected.pxi":63
 *         for a in dir(self.pvt_o):
 *             if self.attr_hidden(a):
 *                 self.acl_cache[a] = hidden_d             # <<<<<<<<<<<<<<
 *                 continue
 *             d = {
 */
      if (unlikely(__pyx_v_self->acl_cache == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(8, 63, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_self->acl_cache, __pyx_v_a, __pyx_v_hidden_d) < 0))) __PYX_ERR(8, 63, __pyx_L1_error)

      /* "Protected_FrozenProtected.pxi":64
 *             if self.attr_hidden(a):
 *                 self.acl_cache[a] = hidden_d
 *                 continue             # <<<<<<<<<<<<<<
 *             d = {
 *                 'r': self.protected_visible(a, use_cache=False),
 */
      goto __pyx_L4_continue;

      /* "Protected_FrozenProtected.pxi":62
 * 
 *         for a in dir(self.pvt_o):
 *             if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *                 self.acl_cache[a] = hidden_d
 *                 continue
 */
    }

    /* "Protected_FrozenProtected.pxi":66
 *                 continue
 *             d = {
 *                 'r': self.protected_visible(a, use_cache=False),             # <<<<<<<<<<<<<<
 *                 'w': self.protected_writeable(a, use_cache=False),
 *             }
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7.__pyx_n = 1;
    __pyx_t_7.use_cache = Py_False;
    __pyx_t_6 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_visible(__pyx_v_self, __pyx_v_a, &__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_r, __pyx_t_6) < 0) __PYX_ERR(8, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Protected_FrozenProtected.pxi":67
 *             d = {
 *                 'r': self.protected_visible(a, use_cache=False),
 *                 'w': self.protected_writeable(a, use_cache=False),             # <<<<<<<<<<<<<<
 *             }
 *             self.acl_cache[a] = d
 */
    __pyx_t_8.__pyx_n = 1;
    __pyx_t_8.use_cache = Py_False;
    __pyx_t_6 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_writeable(__pyx_v_self, __pyx_v_a, &__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_w, __pyx_t_6) < 0) __PYX_ERR(8, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_d, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "Protected_FrozenProtected.pxi":69
 *                 'w': self.protected_writeable(a, use_cache=False),
 *             }
 *             self.acl_cache[a] = d             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
    if (unlikely(__pyx_v_self->acl_cache == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(8, 69, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__pyx_v_self->acl_cache, __pyx_v_a, __pyx_v_d) < 0))) __PYX_ERR(8, 69, __pyx_L1_error)

    /* "Protected_FrozenProtected.pxi":70
 *             }
 *             self.acl_cache[a] = d
 *             continue             # <<<<<<<<<<<<<<
 * 
 *     cdef check_1_op(self, a, op, use_cache=True):
 */
    goto __pyx_L4_continue;

    /* "Protected_FrozenProtected.pxi":61
 *         hidden_d = {'r': False, 'w': False}
 * 
 *         for a in dir(self.pvt_o):             # <<<<<<<<<<<<<<
 *             if self.attr_hidden(a):
 *                 self.acl_cache[a] = hidden_d
 */
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Protected_FrozenProtected.pxi":51
 *             ]
 * 
 *     cdef build_cache(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Called once at object wrapping time
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.Protected.build_cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_hidden_d);
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":72
 *             continue
 * 
 *     cdef check_1_op(self, a, op, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_check_1_op(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op, struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op *__pyx_optional_args) {
  PyObject *__pyx_v_use_cache = ((PyObject *)Py_True);
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_attr_type_check = NULL;
  PyObject *__pyx_v_ro_method = NULL;
  PyObject *__pyx_v_ro_data = NULL;
  PyObject *__pyx_v_hide_private = NULL;
  PyObject *__pyx_v_def_d = NULL;
  int __pyx_v_bMethod;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_1_op", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_use_cache = __pyx_optional_args->use_cache;
    }
  }

  /* "Protected_FrozenProtected.pxi":82
 *         Only called from protected_visible() and protected_writeable()
 *         '''
 *         if op not in ('r', 'w'):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __Pyx_INCREF(__pyx_v_op);
  __pyx_t_1 = __pyx_v_op;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_r, Py_NE)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 82, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_w, Py_NE)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 82, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_2;
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":83
 *         '''
 *         if op not in ('r', 'w'):
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         d = self.rules
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":82
 *         Only called from protected_visible() and protected_writeable()
 *         '''
 *         if op not in ('r', 'w'):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "Protected_FrozenProtected.pxi":85
 *             return False
 * 
 *         d = self.rules             # <<<<<<<<<<<<<<
 *         attr_type_check = bool(d.get('attr_type_check', False))
 *         ro_method = bool(d.get('ro_method', False))
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.rules;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_d = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":86
 * 
 *         d = self.rules
 *         attr_type_check = bool(d.get('attr_type_check', False))             # <<<<<<<<<<<<<<
 *         ro_method = bool(d.get('ro_method', False))
 *         ro_data = bool(d.get('ro_data', False))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_d, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_attr_type_check = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Protected_FrozenProtected.pxi":87
 *         d = self.rules
 *         attr_type_check = bool(d.get('attr_type_check', False))
 *         ro_method = bool(d.get('ro_method', False))             # <<<<<<<<<<<<<<
 *         ro_data = bool(d.get('ro_data', False))
 *         hide_private = bool(d.get('hide_private', False))
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_d, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ro_method = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":88
 *         attr_type_check = bool(d.get('attr_type_check', False))
 *         ro_method = bool(d.get('ro_method', False))
 *         ro_data = bool(d.get('ro_data', False))             # <<<<<<<<<<<<<<
 *         hide_private = bool(d.get('hide_private', False))
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_d, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_ro_data = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Protected_FrozenProtected.pxi":89
 *         ro_method = bool(d.get('ro_method', False))
 *         ro_data = bool(d.get('ro_data', False))
 *         hide_private = bool(d.get('hide_private', False))             # <<<<<<<<<<<<<<
 * 
 *         if use_cache and self.acl_cache is not None:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_d, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_hide_private = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":91
 *         hide_private = bool(d.get('hide_private', False))
 * 
 *         if use_cache and self.acl_cache is not None:             # <<<<<<<<<<<<<<
 *             def_d = {'r': True, 'w': True}
 *             d = self.acl_cache.get(a, def_d)
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_use_cache); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 91, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->acl_cache != ((PyObject*)Py_None));
  __pyx_t_3 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":92
 * 
 *         if use_cache and self.acl_cache is not None:
 *             def_d = {'r': True, 'w': True}             # <<<<<<<<<<<<<<
 *             d = self.acl_cache.get(a, def_d)
 *             return d[op]
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_r, Py_True) < 0) __PYX_ERR(8, 92, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_w, Py_True) < 0) __PYX_ERR(8, 92, __pyx_L1_error)
    __pyx_v_def_d = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Protected_FrozenProtected.pxi":93
 *         if use_cache and self.acl_cache is not None:
 *             def_d = {'r': True, 'w': True}
 *             d = self.acl_cache.get(a, def_d)             # <<<<<<<<<<<<<<
 *             return d[op]
 * 
 */
    if (unlikely(__pyx_v_self->acl_cache == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
      __PYX_ERR(8, 93, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_self->acl_cache, __pyx_v_a, __pyx_v_def_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_d, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Protected_FrozenProtected.pxi":94
 *             def_d = {'r': True, 'w': True}
 *             d = self.acl_cache.get(a, def_d)
 *             return d[op]             # <<<<<<<<<<<<<<
 * 
 *         # If we got here, we need DYNAMIC lookup
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_d, __pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":91
 *         hide_private = bool(d.get('hide_private', False))
 * 
 *         if use_cache and self.acl_cache is not None:             # <<<<<<<<<<<<<<
 *             def_d = {'r': True, 'w': True}
 *             d = self.acl_cache.get(a, def_d)
 */
  }

  /* "Protected_FrozenProtected.pxi":102
 *         # Enforce hiding of private mangled attributes
 *         # This is already done in Private !
 *         if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.attr_hidden(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":103
 *         # This is already done in Private !
 *         if self.attr_hidden(a):
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         if op == 'r':
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":102
 *         # Enforce hiding of private mangled attributes
 *         # This is already done in Private !
 *         if self.attr_hidden(a):             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "Protected_FrozenProtected.pxi":105
 *             return False
 * 
 *         if op == 'r':             # <<<<<<<<<<<<<<
 *             # special_attributes always visible
 *             if a in special_attributes:
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_r, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 105, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":107
 *         if op == 'r':
 *             # special_attributes always visible
 *             if a in special_attributes:             # <<<<<<<<<<<<<<
 *                 return True
 *             # always_frozen are .... always frozen
 */
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_special_attributes, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 107, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":108
 *             # special_attributes always visible
 *             if a in special_attributes:
 *                 return True             # <<<<<<<<<<<<<<
 *             # always_frozen are .... always frozen
 *             if a in always_frozen:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":107
 *         if op == 'r':
 *             # special_attributes always visible
 *             if a in special_attributes:             # <<<<<<<<<<<<<<
 *                 return True
 *             # always_frozen are .... always frozen
 */
    }

    /* "Protected_FrozenProtected.pxi":110
 *                 return True
 *             # always_frozen are .... always frozen
 *             if a in always_frozen:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_always_frozen, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 110, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":111
 *             # always_frozen are .... always frozen
 *             if a in always_frozen:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *             if hide_private and ro_private_attr.match(a):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":110
 *                 return True
 *             # always_frozen are .... always frozen
 *             if a in always_frozen:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }

    /* "Protected_FrozenProtected.pxi":113
 *                 return True
 * 
 *             if hide_private and ro_private_attr.match(a):             # <<<<<<<<<<<<<<
 *                 return False
 *             if (
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_hide_private); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 113, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_ro_private_attr, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_a};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 113, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":114
 * 
 *             if hide_private and ro_private_attr.match(a):
 *                 return False             # <<<<<<<<<<<<<<
 *             if (
 *                 'hide_regex' in d and
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":113
 *                 return True
 * 
 *             if hide_private and ro_private_attr.match(a):             # <<<<<<<<<<<<<<
 *                 return False
 *             if (
 */
    }

    /* "Protected_FrozenProtected.pxi":116
 *                 return False
 *             if (
 *                 'hide_regex' in d and             # <<<<<<<<<<<<<<
 *                 d['hide_regex'].pattern and
 *                 d['hide_regex'].match(a)
 */
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_hide_regex, __pyx_v_d, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 116, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L17_bool_binop_done;
    }

    /* "Protected_FrozenProtected.pxi":117
 *             if (
 *                 'hide_regex' in d and
 *                 d['hide_regex'].pattern and             # <<<<<<<<<<<<<<
 *                 d['hide_regex'].match(a)
 *             ):
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_hide_regex); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_pattern); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L17_bool_binop_done;
    }

    /* "Protected_FrozenProtected.pxi":118
 *                 'hide_regex' in d and
 *                 d['hide_regex'].pattern and
 *                 d['hide_regex'].match(a)             # <<<<<<<<<<<<<<
 *             ):
 *                 return False
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_hide_regex); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_a};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 118, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_L17_bool_binop_done:;

    /* "Protected_FrozenProtected.pxi":115
 *             if hide_private and ro_private_attr.match(a):
 *                 return False
 *             if (             # <<<<<<<<<<<<<<
 *                 'hide_regex' in d and
 *                 d['hide_regex'].pattern and
 */
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":120
 *                 d['hide_regex'].match(a)
 *             ):
 *                 return False             # <<<<<<<<<<<<<<
 *         elif op == 'w':
 *             if self.frozen:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":115
 *             if hide_private and ro_private_attr.match(a):
 *                 return False
 *             if (             # <<<<<<<<<<<<<<
 *                 'hide_regex' in d and
 *                 d['hide_regex'].pattern and
 */
    }

    /* "Protected_FrozenProtected.pxi":105
 *             return False
 * 
 *         if op == 'r':             # <<<<<<<<<<<<<<
 *             # special_attributes always visible
 *             if a in special_attributes:
 */
    goto __pyx_L10;
  }

  /* "Protected_FrozenProtected.pxi":121
 *             ):
 *                 return False
 *         elif op == 'w':             # <<<<<<<<<<<<<<
 *             if self.frozen:
 *                 return False
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_w, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 121, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":122
 *                 return False
 *         elif op == 'w':
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 return False
 *             # special_attributes never writeable
 */
    if (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.frozen) {

      /* "Protected_FrozenProtected.pxi":123
 *         elif op == 'w':
 *             if self.frozen:
 *                 return False             # <<<<<<<<<<<<<<
 *             # special_attributes never writeable
 *             if a in special_attributes:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":122
 *                 return False
 *         elif op == 'w':
 *             if self.frozen:             # <<<<<<<<<<<<<<
 *                 return False
 *             # special_attributes never writeable
 */
    }

    /* "Protected_FrozenProtected.pxi":125
 *                 return False
 *             # special_attributes never writeable
 *             if a in special_attributes:             # <<<<<<<<<<<<<<
 *                 return False
 *             if ro_private_attr.match(a):
 */
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_special_attributes, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 125, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":126
 *             # special_attributes never writeable
 *             if a in special_attributes:
 *                 return False             # <<<<<<<<<<<<<<
 *             if ro_private_attr.match(a):
 *                 return False
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":125
 *                 return False
 *             # special_attributes never writeable
 *             if a in special_attributes:             # <<<<<<<<<<<<<<
 *                 return False
 *             if ro_private_attr.match(a):
 */
    }

    /* "Protected_FrozenProtected.pxi":127
 *             if a in special_attributes:
 *                 return False
 *             if ro_private_attr.match(a):             # <<<<<<<<<<<<<<
 *                 return False
 *             # rw overrides ro_*
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_9pyprotect_9protected_ro_private_attr, __pyx_n_s_match); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_a};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 127, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":128
 *                 return False
 *             if ro_private_attr.match(a):
 *                 return False             # <<<<<<<<<<<<<<
 *             # rw overrides ro_*
 *             if (
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":127
 *             if a in special_attributes:
 *                 return False
 *             if ro_private_attr.match(a):             # <<<<<<<<<<<<<<
 *                 return False
 *             # rw overrides ro_*
 */
    }

    /* "Protected_FrozenProtected.pxi":131
 *             # rw overrides ro_*
 *             if (
 *                 'rw_regex' in d and             # <<<<<<<<<<<<<<
 *                 d['rw_regex'].pattern and
 *                 d['rw_regex'].match(a)
 */
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_rw_regex, __pyx_v_d, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 131, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L24_bool_binop_done;
    }

    /* "Protected_FrozenProtected.pxi":132
 *             if (
 *                 'rw_regex' in d and
 *                 d['rw_regex'].pattern and             # <<<<<<<<<<<<<<
 *                 d['rw_regex'].match(a)
 *             ):
 */
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_rw_regex); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_pattern); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L24_bool_binop_done;
    }

    /* "Protected_FrozenProtected.pxi":133
 *                 'rw_regex' in d and
 *                 d['rw_regex'].pattern and
 *                 d['rw_regex'].match(a)             # <<<<<<<<<<<<<<
 *             ):
 *                 return True
 */
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_rw_regex); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_match); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_a};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 133, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_L24_bool_binop_done:;

    /* "Protected_FrozenProtected.pxi":130
 *                 return False
 *             # rw overrides ro_*
 *             if (             # <<<<<<<<<<<<<<
 *                 'rw_regex' in d and
 *                 d['rw_regex'].pattern and
 */
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":135
 *                 d['rw_regex'].match(a)
 *             ):
 *                 return True             # <<<<<<<<<<<<<<
 *             if (
 *                 'ro_regex' in d and
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":130
 *                 return False
 *             # rw overrides ro_*
 *             if (             # <<<<<<<<<<<<<<
 *                 'rw_regex' in d and
 *                 d['rw_regex'].pattern and
 */
    }

    /* "Protected_FrozenProtected.pxi":137
 *                 return True
 *             if (
 *                 'ro_regex' in d and             # <<<<<<<<<<<<<<
 *                 d['ro_regex'].pattern and
 *                 d['ro_regex'].match(a)
 */
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_ro_regex, __pyx_v_d, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 137, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L28_bool_binop_done;
    }

    /* "Protected_FrozenProtected.pxi":138
 *             if (
 *                 'ro_regex' in d and
 *                 d['ro_regex'].pattern and             # <<<<<<<<<<<<<<
 *                 d['ro_regex'].match(a)
 *             ):
 */
    __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_ro_regex); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_pattern); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 138, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L28_bool_binop_done;
    }

    /* "Protected_FrozenProtected.pxi":139
 *                 'ro_regex' in d and
 *                 d['ro_regex'].pattern and
 *                 d['ro_regex'].match(a)             # <<<<<<<<<<<<<<
 *             ):
 *                 return False
 */
    __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_d, __pyx_n_s_ro_regex); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_a};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 139, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_L28_bool_binop_done:;

    /* "Protected_FrozenProtected.pxi":136
 *             ):
 *                 return True
 *             if (             # <<<<<<<<<<<<<<
 *                 'ro_regex' in d and
 *                 d['ro_regex'].pattern and
 */
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":141
 *                 d['ro_regex'].match(a)
 *             ):
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         if attr_type_check is True:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "Protected_FrozenProtected.pxi":136
 *             ):
 *                 return True
 *             if (             # <<<<<<<<<<<<<<
 *                 'ro_regex' in d and
 *                 d['ro_regex'].pattern and
 */
    }

    /* "Protected_FrozenProtected.pxi":121
 *             ):
 *                 return False
 *         elif op == 'w':             # <<<<<<<<<<<<<<
 *             if self.frozen:
 *                 return False
 */
  }
  __pyx_L10:;

  /* "Protected_FrozenProtected.pxi":143
 *                 return False
 * 
 *         if attr_type_check is True:             # <<<<<<<<<<<<<<
 *             if op == 'w' and (ro_method or ro_data):
 *                 bMethod = callable(getattr(self.pvt_o, a))
 */
  __pyx_t_3 = (__pyx_v_attr_type_check == Py_True);
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":144
 * 
 *         if attr_type_check is True:
 *             if op == 'w' and (ro_method or ro_data):             # <<<<<<<<<<<<<<
 *                 bMethod = callable(getattr(self.pvt_o, a))
 *                 if ro_method:
 */
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_w, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 144, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L33_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_ro_method); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 144, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L33_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_ro_data); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 144, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_2;
    __pyx_L33_bool_binop_done:;
    if (__pyx_t_3) {

      /* "Protected_FrozenProtected.pxi":145
 *         if attr_type_check is True:
 *             if op == 'w' and (ro_method or ro_data):
 *                 bMethod = callable(getattr(self.pvt_o, a))             # <<<<<<<<<<<<<<
 *                 if ro_method:
 *                     return not bMethod
 */
      __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_a); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_3 = __Pyx_PyCallable_Check(__pyx_t_4); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(8, 145, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_bMethod = __pyx_t_3;

      /* "Protected_FrozenProtected.pxi":146
 *             if op == 'w' and (ro_method or ro_data):
 *                 bMethod = callable(getattr(self.pvt_o, a))
 *                 if ro_method:             # <<<<<<<<<<<<<<
 *                     return not bMethod
 *                 elif ro_data:
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_ro_method); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 146, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "Protected_FrozenProtected.pxi":147
 *                 bMethod = callable(getattr(self.pvt_o, a))
 *                 if ro_method:
 *                     return not bMethod             # <<<<<<<<<<<<<<
 *                 elif ro_data:
 *                     return bMethod
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyBool_FromLong((!__pyx_v_bMethod)); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L0;

        /* "Protected_FrozenProtected.pxi":146
 *             if op == 'w' and (ro_method or ro_data):
 *                 bMethod = callable(getattr(self.pvt_o, a))
 *                 if ro_method:             # <<<<<<<<<<<<<<
 *                     return not bMethod
 *                 elif ro_data:
 */
      }

      /* "Protected_FrozenProtected.pxi":148
 *                 if ro_method:
 *                     return not bMethod
 *                 elif ro_data:             # <<<<<<<<<<<<<<
 *                     return bMethod
 *         return True
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_ro_data); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 148, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "Protected_FrozenProtected.pxi":149
 *                     return not bMethod
 *                 elif ro_data:
 *                     return bMethod             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_bMethod); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L0;

        /* "Protected_FrozenProtected.pxi":148
 *                 if ro_method:
 *                     return not bMethod
 *                 elif ro_data:             # <<<<<<<<<<<<<<
 *                     return bMethod
 *         return True
 */
      }

      /* "Protected_FrozenProtected.pxi":144
 * 
 *         if attr_type_check is True:
 *             if op == 'w' and (ro_method or ro_data):             # <<<<<<<<<<<<<<
 *                 bMethod = callable(getattr(self.pvt_o, a))
 *                 if ro_method:
 */
    }

    /* "Protected_FrozenProtected.pxi":143
 *                 return False
 * 
 *         if attr_type_check is True:             # <<<<<<<<<<<<<<
 *             if op == 'w' and (ro_method or ro_data):
 *                 bMethod = callable(getattr(self.pvt_o, a))
 */
  }

  /* "Protected_FrozenProtected.pxi":150
 *                 elif ro_data:
 *                     return bMethod
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef protected_visible(self, a, use_cache=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":72
 *             continue
 * 
 *     cdef check_1_op(self, a, op, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.Protected.check_1_op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_attr_type_check);
  __Pyx_XDECREF(__pyx_v_ro_method);
  __Pyx_XDECREF(__pyx_v_ro_data);
  __Pyx_XDECREF(__pyx_v_hide_private);
  __Pyx_XDECREF(__pyx_v_def_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":152
 *         return True
 * 
 *     cdef protected_visible(self, a, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_visible(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_visible *__pyx_optional_args) {
  PyObject *__pyx_v_use_cache = ((PyObject *)Py_True);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_visible", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_use_cache = __pyx_optional_args->use_cache;
    }
  }
  __Pyx_INCREF(__pyx_v_use_cache);

  /* "Protected_FrozenProtected.pxi":161
 *         Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         '''
 *         if not self.private_visible(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if self.acl_cache is None:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.private_visible(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":162
 *         '''
 *         if not self.private_visible(a):
 *             return False             # <<<<<<<<<<<<<<
 *         if self.acl_cache is None:
 *             use_cache = False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":161
 *         Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         '''
 *         if not self.private_visible(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if self.acl_cache is None:
 */
  }

  /* "Protected_FrozenProtected.pxi":163
 *         if not self.private_visible(a):
 *             return False
 *         if self.acl_cache is None:             # <<<<<<<<<<<<<<
 *             use_cache = False
 *         return self.check_1_op(a=a, op='r', use_cache=use_cache)
 */
  __pyx_t_3 = (__pyx_v_self->acl_cache == ((PyObject*)Py_None));
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":164
 *             return False
 *         if self.acl_cache is None:
 *             use_cache = False             # <<<<<<<<<<<<<<
 *         return self.check_1_op(a=a, op='r', use_cache=use_cache)
 * 
 */
    __Pyx_INCREF(Py_False);
    __Pyx_DECREF_SET(__pyx_v_use_cache, Py_False);

    /* "Protected_FrozenProtected.pxi":163
 *         if not self.private_visible(a):
 *             return False
 *         if self.acl_cache is None:             # <<<<<<<<<<<<<<
 *             use_cache = False
 *         return self.check_1_op(a=a, op='r', use_cache=use_cache)
 */
  }

  /* "Protected_FrozenProtected.pxi":165
 *         if self.acl_cache is None:
 *             use_cache = False
 *         return self.check_1_op(a=a, op='r', use_cache=use_cache)             # <<<<<<<<<<<<<<
 * 
 *     cdef protected_writeable(self, a, use_cache=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.use_cache = __pyx_v_use_cache;
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->check_1_op(__pyx_v_self, __pyx_v_a, __pyx_n_s_r, &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":152
 *         return True
 * 
 *     cdef protected_visible(self, a, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.protected_visible", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_use_cache);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":167
 *         return self.check_1_op(a=a, op='r', use_cache=use_cache)
 * 
 *     cdef protected_writeable(self, a, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_writeable(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_writeable *__pyx_optional_args) {
  PyObject *__pyx_v_use_cache = ((PyObject *)Py_True);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_writeable", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_use_cache = __pyx_optional_args->use_cache;
    }
  }
  __Pyx_INCREF(__pyx_v_use_cache);

  /* "Protected_FrozenProtected.pxi":176
 *         Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         '''
 *         if not self.private_writeable(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if self.acl_cache is None:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.private_writeable(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":177
 *         '''
 *         if not self.private_writeable(a):
 *             return False             # <<<<<<<<<<<<<<
 *         if self.acl_cache is None:
 *             use_cache = False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":176
 *         Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         '''
 *         if not self.private_writeable(a):             # <<<<<<<<<<<<<<
 *             return False
 *         if self.acl_cache is None:
 */
  }

  /* "Protected_FrozenProtected.pxi":178
 *         if not self.private_writeable(a):
 *             return False
 *         if self.acl_cache is None:             # <<<<<<<<<<<<<<
 *             use_cache = False
 *         return self.check_1_op(a=a, op='w', use_cache=use_cache)
 */
  __pyx_t_3 = (__pyx_v_self->acl_cache == ((PyObject*)Py_None));
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":179
 *             return False
 *         if self.acl_cache is None:
 *             use_cache = False             # <<<<<<<<<<<<<<
 *         return self.check_1_op(a=a, op='w', use_cache=use_cache)
 * 
 */
    __Pyx_INCREF(Py_False);
    __Pyx_DECREF_SET(__pyx_v_use_cache, Py_False);

    /* "Protected_FrozenProtected.pxi":178
 *         if not self.private_writeable(a):
 *             return False
 *         if self.acl_cache is None:             # <<<<<<<<<<<<<<
 *             use_cache = False
 *         return self.check_1_op(a=a, op='w', use_cache=use_cache)
 */
  }

  /* "Protected_FrozenProtected.pxi":180
 *         if self.acl_cache is None:
 *             use_cache = False
 *         return self.check_1_op(a=a, op='w', use_cache=use_cache)             # <<<<<<<<<<<<<<
 * 
 *     cdef aclcheck(self, a, op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.use_cache = __pyx_v_use_cache;
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->check_1_op(__pyx_v_self, __pyx_v_a, __pyx_n_s_w, &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":167
 *         return self.check_1_op(a=a, op='r', use_cache=use_cache)
 * 
 *     cdef protected_writeable(self, a, use_cache=True):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.protected_writeable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_use_cache);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":182
 *         return self.check_1_op(a=a, op='w', use_cache=use_cache)
 * 
 *     cdef aclcheck(self, a, op):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_aclcheck(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op) {
  PyObject *__pyx_v_ro_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("aclcheck", 0);

  /* "Protected_FrozenProtected.pxi":190
 *         Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         '''
 *         ro_msg = 'Read only attribute: %s' % (a,)             # <<<<<<<<<<<<<<
 * 
 *         if self.frozen and op in ('w', 'd'):
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_a);
  __Pyx_GIVEREF(__pyx_v_a);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_a);
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Read_only_attribute_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ro_msg = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Protected_FrozenProtected.pxi":192
 *         ro_msg = 'Read only attribute: %s' % (a,)
 * 
 *         if self.frozen and op in ('w', 'd'):             # <<<<<<<<<<<<<<
 *             # Module hack
 *             if not isinstance(self.pvt_o, types.ModuleType):
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.frozen) {
  } else {
    __pyx_t_3 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.frozen;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_v_op);
  __pyx_t_2 = __pyx_v_op;
  __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_w, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(8, 192, __pyx_L1_error)
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_d, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(8, 192, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_5;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_t_4;
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":194
 *         if self.frozen and op in ('w', 'd'):
 *             # Module hack
 *             if not isinstance(self.pvt_o, types.ModuleType):             # <<<<<<<<<<<<<<
 *                 raise frozen_error
 *         if op == 'r':
 */
    __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ModuleType); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = PyObject_IsInstance(__pyx_t_2, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(8, 194, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (!__pyx_t_3);
    if (unlikely(__pyx_t_5)) {

      /* "Protected_FrozenProtected.pxi":195
 *             # Module hack
 *             if not isinstance(self.pvt_o, types.ModuleType):
 *                 raise frozen_error             # <<<<<<<<<<<<<<
 *         if op == 'r':
 *             if not self.visible(a):
 */
      __Pyx_Raise(__pyx_v_9pyprotect_9protected_frozen_error, 0, 0, 0);
      __PYX_ERR(8, 195, __pyx_L1_error)

      /* "Protected_FrozenProtected.pxi":194
 *         if self.frozen and op in ('w', 'd'):
 *             # Module hack
 *             if not isinstance(self.pvt_o, types.ModuleType):             # <<<<<<<<<<<<<<
 *                 raise frozen_error
 *         if op == 'r':
 */
    }

    /* "Protected_FrozenProtected.pxi":192
 *         ro_msg = 'Read only attribute: %s' % (a,)
 * 
 *         if self.frozen and op in ('w', 'd'):             # <<<<<<<<<<<<<<
 *             # Module hack
 *             if not isinstance(self.pvt_o, types.ModuleType):
 */
  }

  /* "Protected_FrozenProtected.pxi":196
 *             if not isinstance(self.pvt_o, types.ModuleType):
 *                 raise frozen_error
 *         if op == 'r':             # <<<<<<<<<<<<<<
 *             if not self.visible(a):
 *                 raise AttributeError(
 */
  __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_r, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(8, 196, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Protected_FrozenProtected.pxi":197
 *                 raise frozen_error
 *         if op == 'r':
 *             if not self.visible(a):             # <<<<<<<<<<<<<<
 *                 raise AttributeError(
 *                     "Object Protected('%s') has no attribute '%s'" % (self.cn, a)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.visible(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(8, 197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = (!__pyx_t_5);
    if (unlikely(__pyx_t_3)) {

      /* "Protected_FrozenProtected.pxi":199
 *             if not self.visible(a):
 *                 raise AttributeError(
 *                     "Object Protected('%s') has no attribute '%s'" % (self.cn, a)             # <<<<<<<<<<<<<<
 *                 )
 *             return     # OK
 */
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
      __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_self->__pyx_base.__pyx_base.cn);
      __Pyx_INCREF(__pyx_v_a);
      __Pyx_GIVEREF(__pyx_v_a);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_a);
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Object_Protected_s_has_no_attrib, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Protected_FrozenProtected.pxi":198
 *         if op == 'r':
 *             if not self.visible(a):
 *                 raise AttributeError(             # <<<<<<<<<<<<<<
 *                     "Object Protected('%s') has no attribute '%s'" % (self.cn, a)
 *                 )
 */
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(8, 198, __pyx_L1_error)

      /* "Protected_FrozenProtected.pxi":197
 *                 raise frozen_error
 *         if op == 'r':
 *             if not self.visible(a):             # <<<<<<<<<<<<<<
 *                 raise AttributeError(
 *                     "Object Protected('%s') has no attribute '%s'" % (self.cn, a)
 */
    }

    /* "Protected_FrozenProtected.pxi":201
 *                     "Object Protected('%s') has no attribute '%s'" % (self.cn, a)
 *                 )
 *             return     # OK             # <<<<<<<<<<<<<<
 *         elif op == 'w':
 *             if not self.writeable(a):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":196
 *             if not isinstance(self.pvt_o, types.ModuleType):
 *                 raise frozen_error
 *         if op == 'r':             # <<<<<<<<<<<<<<
 *             if not self.visible(a):
 *                 raise AttributeError(
 */
  }

  /* "Protected_FrozenProtected.pxi":202
 *                 )
 *             return     # OK
 *         elif op == 'w':             # <<<<<<<<<<<<<<
 *             if not self.writeable(a):
 *                 raise ProtectionError(ro_msg)
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_w, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 202, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Protected_FrozenProtected.pxi":203
 *             return     # OK
 *         elif op == 'w':
 *             if not self.writeable(a):             # <<<<<<<<<<<<<<
 *                 raise ProtectionError(ro_msg)
 *             return   # OK
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.writeable(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 203, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (!__pyx_t_3);
    if (unlikely(__pyx_t_5)) {

      /* "Protected_FrozenProtected.pxi":204
 *         elif op == 'w':
 *             if not self.writeable(a):
 *                 raise ProtectionError(ro_msg)             # <<<<<<<<<<<<<<
 *             return   # OK
 *         elif op == 'd':
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_7 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_ro_msg};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(8, 204, __pyx_L1_error)

      /* "Protected_FrozenProtected.pxi":203
 *             return     # OK
 *         elif op == 'w':
 *             if not self.writeable(a):             # <<<<<<<<<<<<<<
 *                 raise ProtectionError(ro_msg)
 *             return   # OK
 */
    }

    /* "Protected_FrozenProtected.pxi":205
 *             if not self.writeable(a):
 *                 raise ProtectionError(ro_msg)
 *             return   # OK             # <<<<<<<<<<<<<<
 *         elif op == 'd':
 *             return   # OK
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":202
 *                 )
 *             return     # OK
 *         elif op == 'w':             # <<<<<<<<<<<<<<
 *             if not self.writeable(a):
 *                 raise ProtectionError(ro_msg)
 */
  }

  /* "Protected_FrozenProtected.pxi":206
 *                 raise ProtectionError(ro_msg)
 *             return   # OK
 *         elif op == 'd':             # <<<<<<<<<<<<<<
 *             return   # OK
 * 
 */
  __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_v_op, __pyx_n_s_d, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(8, 206, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Protected_FrozenProtected.pxi":207
 *             return   # OK
 *         elif op == 'd':
 *             return   # OK             # <<<<<<<<<<<<<<
 * 
 *     cdef protected_getattr(self, a):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":206
 *                 raise ProtectionError(ro_msg)
 *             return   # OK
 *         elif op == 'd':             # <<<<<<<<<<<<<<
 *             return   # OK
 * 
 */
  }

  /* "Protected_FrozenProtected.pxi":182
 *         return self.check_1_op(a=a, op='w', use_cache=use_cache)
 * 
 *     cdef aclcheck(self, a, op):             # <<<<<<<<<<<<<<
 *         '''
 *         a-->str: attribute name
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.Protected.aclcheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ro_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":209
 *             return   # OK
 * 
 *     cdef protected_getattr(self, a):             # <<<<<<<<<<<<<<
 *         self.aclcheck(a=a, op='r')
 *         x = self.private_getattr(a)
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_getattr(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_getattr", 0);

  /* "Protected_FrozenProtected.pxi":210
 * 
 *     cdef protected_getattr(self, a):
 *         self.aclcheck(a=a, op='r')             # <<<<<<<<<<<<<<
 *         x = self.private_getattr(a)
 *         # Can always read PROT_ATTR_NAME, even with hide_private == True
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->aclcheck(__pyx_v_self, __pyx_v_a, __pyx_n_s_r); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":211
 *     cdef protected_getattr(self, a):
 *         self.aclcheck(a=a, op='r')
 *         x = self.private_getattr(a)             # <<<<<<<<<<<<<<
 *         # Can always read PROT_ATTR_NAME, even with hide_private == True
 *         if a == PROT_ATTR_NAME:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.private_getattr(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_x = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":213
 *         x = self.private_getattr(a)
 *         # Can always read PROT_ATTR_NAME, even with hide_private == True
 *         if a == PROT_ATTR_NAME:             # <<<<<<<<<<<<<<
 *             return x
 *         if a in m_block and hasattr(Wrapped, a):
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_a, __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 213, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "Protected_FrozenProtected.pxi":214
 *         # Can always read PROT_ATTR_NAME, even with hide_private == True
 *         if a == PROT_ATTR_NAME:
 *             return x             # <<<<<<<<<<<<<<
 *         if a in m_block and hasattr(Wrapped, a):
 *             return x
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_x);
    __pyx_r = __pyx_v_x;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":213
 *         x = self.private_getattr(a)
 *         # Can always read PROT_ATTR_NAME, even with hide_private == True
 *         if a == PROT_ATTR_NAME:             # <<<<<<<<<<<<<<
 *             return x
 *         if a in m_block and hasattr(Wrapped, a):
 */
  }

  /* "Protected_FrozenProtected.pxi":215
 *         if a == PROT_ATTR_NAME:
 *             return x
 *         if a in m_block and hasattr(Wrapped, a):             # <<<<<<<<<<<<<<
 *             return x
 *         try:
 */
  if (unlikely(__pyx_v_9pyprotect_9protected_m_block == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(8, 215, __pyx_L1_error)
  }
  __pyx_t_3 = (__Pyx_PySet_ContainsTF(__pyx_v_a, __pyx_v_9pyprotect_9protected_m_block, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(8, 215, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_v_a); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(8, 215, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "Protected_FrozenProtected.pxi":216
 *             return x
 *         if a in m_block and hasattr(Wrapped, a):
 *             return x             # <<<<<<<<<<<<<<
 *         try:
 *             self.aclcheck(a=a, op='w')
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_x);
    __pyx_r = __pyx_v_x;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":215
 *         if a == PROT_ATTR_NAME:
 *             return x
 *         if a in m_block and hasattr(Wrapped, a):             # <<<<<<<<<<<<<<
 *             return x
 *         try:
 */
  }

  /* "Protected_FrozenProtected.pxi":217
 *         if a in m_block and hasattr(Wrapped, a):
 *             return x
 *         try:             # <<<<<<<<<<<<<<
 *             self.aclcheck(a=a, op='w')
 *             return x
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "Protected_FrozenProtected.pxi":218
 *             return x
 *         try:
 *             self.aclcheck(a=a, op='w')             # <<<<<<<<<<<<<<
 *             return x
 *         except:   # not writeable for any reason
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->aclcheck(__pyx_v_self, __pyx_v_a, __pyx_n_s_w); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 218, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Protected_FrozenProtected.pxi":219
 *         try:
 *             self.aclcheck(a=a, op='w')
 *             return x             # <<<<<<<<<<<<<<
 *         except:   # not writeable for any reason
 *             return freeze(x)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_x);
      __pyx_r = __pyx_v_x;
      goto __pyx_L11_try_return;

      /* "Protected_FrozenProtected.pxi":217
 *         if a in m_block and hasattr(Wrapped, a):
 *             return x
 *         try:             # <<<<<<<<<<<<<<
 *             self.aclcheck(a=a, op='w')
 *             return x
 */
    }
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Protected_FrozenProtected.pxi":220
 *             self.aclcheck(a=a, op='w')
 *             return x
 *         except:   # not writeable for any reason             # <<<<<<<<<<<<<<
 *             return freeze(x)
 * 
 */
    /*except:*/ {
      __Pyx_AddTraceback("pyprotect.protected.Protected.protected_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(8, 220, __pyx_L9_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "Protected_FrozenProtected.pxi":221
 *             return x
 *         except:   # not writeable for any reason
 *             return freeze(x)             # <<<<<<<<<<<<<<
 * 
 *     cdef protected_check_setattr(self, a, val):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_freeze); if (unlikely(!__pyx_t_10)) __PYX_ERR(8, 221, __pyx_L9_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_12 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_x};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(8, 221, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __pyx_r = __pyx_t_9;
      __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L10_except_return;
    }
    __pyx_L9_except_error:;

    /* "Protected_FrozenProtected.pxi":217
 *         if a in m_block and hasattr(Wrapped, a):
 *             return x
 *         try:             # <<<<<<<<<<<<<<
 *             self.aclcheck(a=a, op='w')
 *             return x
 */
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L11_try_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
    __pyx_L10_except_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
  }

  /* "Protected_FrozenProtected.pxi":209
 *             return   # OK
 * 
 *     cdef protected_getattr(self, a):             # <<<<<<<<<<<<<<
 *         self.aclcheck(a=a, op='r')
 *         x = self.private_getattr(a)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pyprotect.protected.Protected.protected_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":223
 *             return freeze(x)
 * 
 *     cdef protected_check_setattr(self, a, val):             # <<<<<<<<<<<<<<
 *         self.aclcheck(a=a, op='w')
 *         self.private_check_setattr(a, val)
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_check_setattr(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_check_setattr", 0);

  /* "Protected_FrozenProtected.pxi":224
 * 
 *     cdef protected_check_setattr(self, a, val):
 *         self.aclcheck(a=a, op='w')             # <<<<<<<<<<<<<<
 *         self.private_check_setattr(a, val)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->aclcheck(__pyx_v_self, __pyx_v_a, __pyx_n_s_w); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":225
 *     cdef protected_check_setattr(self, a, val):
 *         self.aclcheck(a=a, op='w')
 *         self.private_check_setattr(a, val)             # <<<<<<<<<<<<<<
 * 
 *     cdef protected_check_delattr(self, a):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.private_check_setattr(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), __pyx_v_a, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":223
 *             return freeze(x)
 * 
 *     cdef protected_check_setattr(self, a, val):             # <<<<<<<<<<<<<<
 *         self.aclcheck(a=a, op='w')
 *         self.private_check_setattr(a, val)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.protected_check_setattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":227
 *         self.private_check_setattr(a, val)
 * 
 *     cdef protected_check_delattr(self, a):             # <<<<<<<<<<<<<<
 *         self.aclcheck(a=a, op='d')
 *         self.private_check_delattr(a)
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_check_delattr(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_check_delattr", 0);

  /* "Protected_FrozenProtected.pxi":228
 * 
 *     cdef protected_check_delattr(self, a):
 *         self.aclcheck(a=a, op='d')             # <<<<<<<<<<<<<<
 *         self.private_check_delattr(a)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->aclcheck(__pyx_v_self, __pyx_v_a, __pyx_n_s_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":229
 *     cdef protected_check_delattr(self, a):
 *         self.aclcheck(a=a, op='d')
 *         self.private_check_delattr(a)             # <<<<<<<<<<<<<<
 * 
 *     cdef protected_dir(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.private_check_delattr(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self), __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":227
 *         self.private_check_setattr(a, val)
 * 
 *     cdef protected_check_delattr(self, a):             # <<<<<<<<<<<<<<
 *         self.aclcheck(a=a, op='d')
 *         self.private_check_delattr(a)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.protected_check_delattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":231
 *         self.private_check_delattr(a)
 * 
 *     cdef protected_dir(self):             # <<<<<<<<<<<<<<
 *         if bool(self.rules.get('dynamic', True)):
 *             return [
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_protected_dir(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self) {
  PyObject *__pyx_9genexpr18__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("protected_dir", 0);

  /* "Protected_FrozenProtected.pxi":232
 * 
 *     cdef protected_dir(self):
 *         if bool(self.rules.get('dynamic', True)):             # <<<<<<<<<<<<<<
 *             return [
 *                 x for x in self.private_dir()
 */
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base.rules == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
    __PYX_ERR(8, 232, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_self->__pyx_base.__pyx_base.rules, __pyx_n_s_dynamic, Py_True); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if ((!(!__pyx_t_2))) {

    /* "Protected_FrozenProtected.pxi":233
 *     cdef protected_dir(self):
 *         if bool(self.rules.get('dynamic', True)):
 *             return [             # <<<<<<<<<<<<<<
 *                 x for x in self.private_dir()
 *                 if self.visible(x)
 */
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 233, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "Protected_FrozenProtected.pxi":234
 *         if bool(self.rules.get('dynamic', True)):
 *             return [
 *                 x for x in self.private_dir()             # <<<<<<<<<<<<<<
 *                 if self.visible(x)
 *             ]
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.private_dir(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 234, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 234, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 234, __pyx_L6_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(8, 234, __pyx_L6_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 234, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(8, 234, __pyx_L6_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 234, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_6(__pyx_t_4);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(8, 234, __pyx_L6_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_x, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Protected_FrozenProtected.pxi":235
 *             return [
 *                 x for x in self.private_dir()
 *                 if self.visible(x)             # <<<<<<<<<<<<<<
 *             ]
 *         else:
 */
        __pyx_t_3 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.visible(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_9genexpr18__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 235, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 235, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_2) {

          /* "Protected_FrozenProtected.pxi":234
 *         if bool(self.rules.get('dynamic', True)):
 *             return [
 *                 x for x in self.private_dir()             # <<<<<<<<<<<<<<
 *                 if self.visible(x)
 *             ]
 */
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr18__pyx_v_x))) __PYX_ERR(8, 233, __pyx_L6_error)

          /* "Protected_FrozenProtected.pxi":235
 *             return [
 *                 x for x in self.private_dir()
 *                 if self.visible(x)             # <<<<<<<<<<<<<<
 *             ]
 *         else:
 */
        }

        /* "Protected_FrozenProtected.pxi":234
 *         if bool(self.rules.get('dynamic', True)):
 *             return [
 *                 x for x in self.private_dir()             # <<<<<<<<<<<<<<
 *                 if self.visible(x)
 *             ]
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_x); __pyx_9genexpr18__pyx_v_x = 0;
      goto __pyx_L10_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_x); __pyx_9genexpr18__pyx_v_x = 0;
      goto __pyx_L1_error;
      __pyx_L10_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Protected_FrozenProtected.pxi":232
 * 
 *     cdef protected_dir(self):
 *         if bool(self.rules.get('dynamic', True)):             # <<<<<<<<<<<<<<
 *             return [
 *                 x for x in self.private_dir()
 */
  }

  /* "Protected_FrozenProtected.pxi":238
 *             ]
 *         else:
 *             return self.dir_out             # <<<<<<<<<<<<<<
 * 
 *     cdef visible(self, a):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->dir_out);
    __pyx_r = __pyx_v_self->dir_out;
    goto __pyx_L0;
  }

  /* "Protected_FrozenProtected.pxi":231
 *         self.private_check_delattr(a)
 * 
 *     cdef protected_dir(self):             # <<<<<<<<<<<<<<
 *         if bool(self.rules.get('dynamic', True)):
 *             return [
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.Protected.protected_dir", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":240
 *             return self.dir_out
 * 
 *     cdef visible(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         return self.protected_visible(a)
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_visible(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visible", 0);

  /* "Protected_FrozenProtected.pxi":242
 *     cdef visible(self, a):
 *         # Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         return self.protected_visible(a)             # <<<<<<<<<<<<<<
 * 
 *     cdef writeable(self, a):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_visible(__pyx_v_self, __pyx_v_a, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":240
 *             return self.dir_out
 * 
 *     cdef visible(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __getattribute__, __setattr__, __delattr__
 *         return self.protected_visible(a)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.visible", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":244
 *         return self.protected_visible(a)
 * 
 *     cdef writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         return self.protected_writeable(a)
 */

static PyObject *__pyx_f_9pyprotect_9protected_9Protected_writeable(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writeable", 0);

  /* "Protected_FrozenProtected.pxi":246
 *     cdef writeable(self, a):
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         return self.protected_writeable(a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_writeable(__pyx_v_self, __pyx_v_a, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":244
 *         return self.protected_visible(a)
 * 
 *     cdef writeable(self, a):             # <<<<<<<<<<<<<<
 *         # Needs to be FAST - called in __setattr__, __delattr__
 *         return self.protected_writeable(a)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.writeable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":253
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.protected_getattr(a)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattribute__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_2__getattribute__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_2__getattribute__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattribute__", 0);

  /* "Protected_FrozenProtected.pxi":254
 * 
 *     def __getattribute__(self, a):
 *         return self.protected_getattr(a)             # <<<<<<<<<<<<<<
 * 
 *     def __setattr__(self, a, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_getattr(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":253
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.protected_getattr(a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__getattribute__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":256
 *         return self.protected_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.protected_check_setattr(a, val)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_9Protected_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_9Protected_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_4__setattr__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self), ((PyObject *)__pyx_v_a), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_9Protected_4__setattr__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setattr__", 0);

  /* "Protected_FrozenProtected.pxi":258
 *     def __setattr__(self, a, val):
 *         # Only checks and raises exceptions
 *         self.protected_check_setattr(a, val)             # <<<<<<<<<<<<<<
 *         setattr(self.pvt_o, a, val)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_check_setattr(__pyx_v_self, __pyx_v_a, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":259
 *         # Only checks and raises exceptions
 *         self.protected_check_setattr(a, val)
 *         setattr(self.pvt_o, a, val)             # <<<<<<<<<<<<<<
 * 
 *     def __delattr__(self, a):
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_SetAttr(__pyx_t_1, __pyx_v_a, __pyx_v_val); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(8, 259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":256
 *         return self.protected_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.protected_check_setattr(a, val)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__setattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":261
 *         setattr(self.pvt_o, a, val)
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.protected_check_delattr(a)
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_9Protected_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static int __pyx_pw_9pyprotect_9protected_9Protected_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_6__delattr__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_9Protected_6__delattr__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delattr__", 0);

  /* "Protected_FrozenProtected.pxi":263
 *     def __delattr__(self, a):
 *         # Only checks and raises exceptions
 *         self.protected_check_delattr(a)             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_check_delattr(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":261
 *         setattr(self.pvt_o, a, val)
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         # Only checks and raises exceptions
 *         self.protected_check_delattr(a)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__delattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":265
 *         self.protected_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.protected_dir()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_9Protected_9__dir__ = {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_9Protected_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__dir__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__dir__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_8__dir__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_8__dir__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "Protected_FrozenProtected.pxi":266
 * 
 *     def __dir__(self):
 *         return self.protected_dir()             # <<<<<<<<<<<<<<
 * 
 *     # Python / cython does not automatically use parent __hash__
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->protected_dir(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":265
 *         self.protected_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.protected_dir()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":269
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_9Protected_11__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_9Protected_11__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_10__hash__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_9Protected_10__hash__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_hash_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Protected_FrozenProtected.pxi":270
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):
 *         return Wrapped.__hash__(self)             # <<<<<<<<<<<<<<
 * 
 *     # __richcmp__ needs to be class-specific
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_hash_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(8, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":269
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":273
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_13__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_13__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_12__richcmp__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_12__richcmp__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "Protected_FrozenProtected.pxi":275
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_Protected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.comparator(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":273
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_9Protected_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_9Protected_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_14__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_14__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(9); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->acl_cache);
  __Pyx_GIVEREF(__pyx_v_self->acl_cache);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->acl_cache);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->__pyx_base.__pyx_base.cn);
  __Pyx_INCREF(__pyx_v_self->dir_out);
  __Pyx_GIVEREF(__pyx_v_self->dir_out);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->dir_out);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 6, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.rules);
  PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_v_self->__pyx_base.__pyx_base.rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->acl_cache != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->dir_out != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Protected); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_218598457);
    __Pyx_GIVEREF(__pyx_int_218598457);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_218598457);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, None), state
 *     else:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Protected__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Protected); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_218598457);
    __Pyx_GIVEREF(__pyx_int_218598457);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_218598457);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Protected__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_9Protected_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_9Protected_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_9Protected_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.Protected.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_9Protected_16__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_9Protected_16__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Protected__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Protected__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Protected__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.Protected.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":283
 *     Subclass of Protected that is automatically frozen
 *     '''
 *     def __init__(self, o, rules):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->object to be wrapped
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_15FrozenProtected_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_15FrozenProtected___init__, "\n        o-->object to be wrapped\n        rules-->dict: returned by protected_rules_from_kwargs\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_15FrozenProtected___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_15FrozenProtected_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_o = 0;
  PyObject *__pyx_v_rules = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_rules,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_o,&__pyx_n_s_rules,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_o)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 283, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rules)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 283, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(8, 283, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(8, 283, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_o = values[0];
    __pyx_v_rules = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(8, 283, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.FrozenProtected.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15FrozenProtected___init__(((struct __pyx_obj_9pyprotect_9protected_FrozenProtected *)__pyx_v_self), __pyx_v_o, __pyx_v_rules);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_15FrozenProtected___init__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self, PyObject *__pyx_v_o, PyObject *__pyx_v_rules) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Protected_FrozenProtected.pxi":288
 *         rules-->dict: returned by protected_rules_from_kwargs
 *         '''
 *         rules['frozen'] = True             # <<<<<<<<<<<<<<
 *         Protected.__init__(self, o, rules)
 * 
 */
  if (unlikely((PyObject_SetItem(__pyx_v_rules, __pyx_n_s_frozen, Py_True) < 0))) __PYX_ERR(8, 288, __pyx_L1_error)

  /* "Protected_FrozenProtected.pxi":289
 *         '''
 *         rules['frozen'] = True
 *         Protected.__init__(self, o, rules)             # <<<<<<<<<<<<<<
 * 
 *     # Python / cython does not automatically use parent __hash__
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_o, __pyx_v_rules};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Protected_FrozenProtected.pxi":283
 *     Subclass of Protected that is automatically frozen
 *     '''
 *     def __init__(self, o, rules):             # <<<<<<<<<<<<<<
 *         '''
 *         o-->object to be wrapped
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.FrozenProtected.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":292
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_15FrozenProtected_3__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_15FrozenProtected_3__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15FrozenProtected_2__hash__(((struct __pyx_obj_9pyprotect_9protected_FrozenProtected *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_15FrozenProtected_2__hash__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_hash_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Protected_FrozenProtected.pxi":293
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):
 *         return Wrapped.__hash__(self)             # <<<<<<<<<<<<<<
 * 
 *     # __richcmp__ needs to be class-specific
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_hash_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_5 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(8, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":292
 * 
 *     # Python / cython does not automatically use parent __hash__
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return Wrapped.__hash__(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.FrozenProtected.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Protected_FrozenProtected.pxi":296
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15FrozenProtected_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_15FrozenProtected_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15FrozenProtected_4__richcmp__(((struct __pyx_obj_9pyprotect_9protected_FrozenProtected *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15FrozenProtected_4__richcmp__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "Protected_FrozenProtected.pxi":298
 *     def __richcmp__(self, other, int op):
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_9pyprotect_9protected_FrozenProtected *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.comparator(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v_self), __pyx_v_other, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Protected_FrozenProtected.pxi":296
 * 
 *     # __richcmp__ needs to be class-specific
 *     def __richcmp__(self, other, int op):             # <<<<<<<<<<<<<<
 *         '''Use common method for all Wrapped objects'''
 *         return self.comparator(other, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.FrozenProtected.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15FrozenProtected_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_15FrozenProtected_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15FrozenProtected_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_15FrozenProtected_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15FrozenProtected_6__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected_FrozenProtected *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15FrozenProtected_6__reduce_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.frozen); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.oldstyle_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(9); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.acl_cache);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.acl_cache);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.acl_cache);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.cn);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.cn);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.cn);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.dir_out);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.dir_out);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->__pyx_base.dir_out);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.hidden_private_attr);
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.protected_attribute);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.protected_attribute);
  PyTuple_SET_ITEM(__pyx_t_3, 6, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.protected_attribute));
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.pvt_o);
  PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.rules);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.rules);
  PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.rules);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.acl_cache, self.cn, self.dir_out, self.frozen, self.hidden_private_attr, self.oldstyle_class, self.protected_attribute, self.pvt_o, self.rules)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.acl_cache != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.cn != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.dir_out != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.hidden_private_attr != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.protected_attribute) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.pvt_o != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.rules != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_FrozenProtected); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_218598457);
    __Pyx_GIVEREF(__pyx_int_218598457);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_218598457);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.acl_cache is not None or self.cn is not None or self.dir_out is not None or self.hidden_private_attr is not None or self.protected_attribute is not None or self.pvt_o is not None or self.rules is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, None), state
 *     else:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FrozenProtected__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_FrozenProtected); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_218598457);
    __Pyx_GIVEREF(__pyx_int_218598457);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_218598457);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyprotect.protected.FrozenProtected.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenProtected__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15FrozenProtected_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_15FrozenProtected_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15FrozenProtected_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_15FrozenProtected_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.FrozenProtected.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15FrozenProtected_8__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected_FrozenProtected *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15FrozenProtected_8__setstate_cython__(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FrozenProtected__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenProtected__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenProtected__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.FrozenProtected.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":10
 *     cdef object kwargs
 * 
 *     def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         '''
 *         args, kwargs: passed to functools.partial
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_15__HiddenPartial_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_9pyprotect_9protected_15__HiddenPartial___init__, "\n        args, kwargs: passed to functools.partial\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_9pyprotect_9protected_15__HiddenPartial___init__;
#endif
static int __pyx_pw_9pyprotect_9protected_15__HiddenPartial_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 1))) return -1;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return -1;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return -1;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial___init__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_15__HiddenPartial___init__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "HiddenPartial.pxi":14
 *         args, kwargs: passed to functools.partial
 *         '''
 *         self.args = args             # <<<<<<<<<<<<<<
 *         self.kwargs = kwargs
 * 
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = __pyx_v_args;

  /* "HiddenPartial.pxi":15
 *         '''
 *         self.args = args
 *         self.kwargs = kwargs             # <<<<<<<<<<<<<<
 * 
 *     cdef wrapped_getattr(self, a):
 */
  __Pyx_INCREF(__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_v_kwargs);
  __Pyx_GOTREF(__pyx_v_self->kwargs);
  __Pyx_DECREF(__pyx_v_self->kwargs);
  __pyx_v_self->kwargs = __pyx_v_kwargs;

  /* "HiddenPartial.pxi":10
 *     cdef object kwargs
 * 
 *     def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         '''
 *         args, kwargs: passed to functools.partial
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":17
 *         self.kwargs = kwargs
 * 
 *     cdef wrapped_getattr(self, a):             # <<<<<<<<<<<<<<
 *         if a in ['__call__']:
 *             return functools.partial(getattr(Wrapped, a), self)
 */

static PyObject *__pyx_f_9pyprotect_9protected_15__HiddenPartial_wrapped_getattr(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrapped_getattr", 0);

  /* "HiddenPartial.pxi":18
 * 
 *     cdef wrapped_getattr(self, a):
 *         if a in ['__call__']:             # <<<<<<<<<<<<<<
 *             return functools.partial(getattr(Wrapped, a), self)
 *         raise AttributeError(
 */
  __Pyx_INCREF(__pyx_v_a);
  __pyx_t_1 = __pyx_v_a;
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_call, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(9, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_2;
  if (__pyx_t_3) {

    /* "HiddenPartial.pxi":19
 *     cdef wrapped_getattr(self, a):
 *         if a in ['__call__']:
 *             return functools.partial(getattr(Wrapped, a), self)             # <<<<<<<<<<<<<<
 *         raise AttributeError(
 *             "Object __HiddenPartial has no attribute '%s'" % (a,)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_functools); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_partial); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_v_a); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_4, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 19, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "HiddenPartial.pxi":18
 * 
 *     cdef wrapped_getattr(self, a):
 *         if a in ['__call__']:             # <<<<<<<<<<<<<<
 *             return functools.partial(getattr(Wrapped, a), self)
 *         raise AttributeError(
 */
  }

  /* "HiddenPartial.pxi":21
 *             return functools.partial(getattr(Wrapped, a), self)
 *         raise AttributeError(
 *             "Object __HiddenPartial has no attribute '%s'" % (a,)             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_a);
  __Pyx_GIVEREF(__pyx_v_a);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_a);
  __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Object___HiddenPartial_has_no_at, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "HiddenPartial.pxi":20
 *         if a in ['__call__']:
 *             return functools.partial(getattr(Wrapped, a), self)
 *         raise AttributeError(             # <<<<<<<<<<<<<<
 *             "Object __HiddenPartial has no attribute '%s'" % (a,)
 *         )
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(9, 20, __pyx_L1_error)

  /* "HiddenPartial.pxi":17
 *         self.kwargs = kwargs
 * 
 *     cdef wrapped_getattr(self, a):             # <<<<<<<<<<<<<<
 *         if a in ['__call__']:
 *             return functools.partial(getattr(Wrapped, a), self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.wrapped_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":28
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.wrapped_getattr(a)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_3__getattribute__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattribute__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_2__getattribute__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_2__getattribute__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattribute__", 0);

  /* "HiddenPartial.pxi":29
 * 
 *     def __getattribute__(self, a):
 *         return self.wrapped_getattr(a)             # <<<<<<<<<<<<<<
 * 
 *     def __setattr__(self, a, val):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_9pyprotect_9protected___HiddenPartial *)__pyx_v_self->__pyx_vtab)->wrapped_getattr(__pyx_v_self, __pyx_v_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "HiddenPartial.pxi":28
 *     # --------------------------------------------------------------------
 * 
 *     def __getattribute__(self, a):             # <<<<<<<<<<<<<<
 *         return self.wrapped_getattr(a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__getattribute__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":31
 *         return self.wrapped_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_15__HiddenPartial_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_9pyprotect_9protected_15__HiddenPartial_5__setattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_4__setattr__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self), ((PyObject *)__pyx_v_a), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_15__HiddenPartial_4__setattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a, CYTHON_UNUSED PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setattr__", 0);

  /* "HiddenPartial.pxi":32
 * 
 *     def __setattr__(self, a, val):
 *         raise ProtectionError('Cannot modify attribute: %s' % (a,))             # <<<<<<<<<<<<<<
 * 
 *     def __delattr__(self, a):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_a);
  __Pyx_GIVEREF(__pyx_v_a);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_a);
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_modify_attribute_s, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(9, 32, __pyx_L1_error)

  /* "HiddenPartial.pxi":31
 *         return self.wrapped_getattr(a)
 * 
 *     def __setattr__(self, a, val):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__setattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":34
 *         raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 */

/* Python wrapper */
static int __pyx_pw_9pyprotect_9protected_15__HiddenPartial_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static int __pyx_pw_9pyprotect_9protected_15__HiddenPartial_7__delattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delattr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_6__delattr__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9pyprotect_9protected_15__HiddenPartial_6__delattr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delattr__", 0);

  /* "HiddenPartial.pxi":35
 * 
 *     def __delattr__(self, a):
 *         raise ProtectionError('Cannot delete attribute: %s' % (a,))             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_a);
  __Pyx_GIVEREF(__pyx_v_a);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_a);
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_delete_attribute_s, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(9, 35, __pyx_L1_error)

  /* "HiddenPartial.pxi":34
 *         raise ProtectionError('Cannot modify attribute: %s' % (a,))
 * 
 *     def __delattr__(self, a):             # <<<<<<<<<<<<<<
 *         raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__delattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":37
 *         raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['__call__']
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_15__HiddenPartial_9__dir__ = {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_9__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__dir__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__dir__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_8__dir__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_8__dir__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "HiddenPartial.pxi":38
 * 
 *     def __dir__(self):
 *         return ['__call__']             # <<<<<<<<<<<<<<
 * 
 *     # Needs to be class-specific
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_call);
  __Pyx_GIVEREF(__pyx_n_s_call);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_call);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "HiddenPartial.pxi":37
 *         raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['__call__']
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":41
 * 
 *     # Needs to be class-specific
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((
 *             hash(self.args),
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_9pyprotect_9protected_15__HiddenPartial_11__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_9pyprotect_9protected_15__HiddenPartial_11__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_10__hash__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_9pyprotect_9protected_15__HiddenPartial_10__hash__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "HiddenPartial.pxi":43
 *     def __hash__(self):
 *         return hash((
 *             hash(self.args),             # <<<<<<<<<<<<<<
 *             str(self.kwargs),
 *         ))
 */
  __pyx_t_1 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(9, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "HiddenPartial.pxi":44
 *         return hash((
 *             hash(self.args),
 *             str(self.kwargs),             # <<<<<<<<<<<<<<
 *         ))
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Str(__pyx_v_self->kwargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "HiddenPartial.pxi":43
 *     def __hash__(self):
 *         return hash((
 *             hash(self.args),             # <<<<<<<<<<<<<<
 *             str(self.kwargs),
 *         ))
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;

  /* "HiddenPartial.pxi":42
 *     # Needs to be class-specific
 *     def __hash__(self):
 *         return hash((             # <<<<<<<<<<<<<<
 *             hash(self.args),
 *             str(self.kwargs),
 */
  __pyx_t_2 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(9, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "HiddenPartial.pxi":41
 * 
 *     # Needs to be class-specific
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((
 *             hash(self.args),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":49
 *     # __repr__, __str__ and __bytes__:
 *     # We do not want the default cython implementations
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(type(functools.partial))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_13__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_13__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_12__repr__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_12__repr__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "HiddenPartial.pxi":50
 *     # We do not want the default cython implementations
 *     def __repr__(self):
 *         return repr(type(functools.partial))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_functools); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Repr(((PyObject *)Py_TYPE(__pyx_t_2))); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "HiddenPartial.pxi":49
 *     # __repr__, __str__ and __bytes__:
 *     # We do not want the default cython implementations
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(type(functools.partial))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":52
 *         return repr(type(functools.partial))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(type(functools.partial))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_15__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_15__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_14__str__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_14__str__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "HiddenPartial.pxi":53
 * 
 *     def __str__(self):
 *         return str(type(functools.partial))             # <<<<<<<<<<<<<<
 * 
 *     def __bytes__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_functools); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Str(((PyObject *)Py_TYPE(__pyx_t_2))); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "HiddenPartial.pxi":52
 *         return repr(type(functools.partial))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(type(functools.partial))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":55
 *         return str(type(functools.partial))
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(type(functools.partial))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_17__bytes__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_15__HiddenPartial_17__bytes__ = {"__bytes__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_17__bytes__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_17__bytes__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bytes__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__bytes__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__bytes__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_16__bytes__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_16__bytes__(CYTHON_UNUSED struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__bytes__", 0);

  /* "HiddenPartial.pxi":56
 * 
 *     def __bytes__(self):
 *         return bytes(type(functools.partial))             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_functools); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), ((PyObject *)Py_TYPE(__pyx_t_2))); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "HiddenPartial.pxi":55
 *         return str(type(functools.partial))
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(type(functools.partial))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__bytes__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "HiddenPartial.pxi":58
 *         return bytes(type(functools.partial))
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         f = functools.partial(*self.args, **self.kwargs)
 *         return f(*args, **kwargs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_19__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_19__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__call__", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_18__call__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_18__call__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "HiddenPartial.pxi":59
 * 
 *     def __call__(self, *args, **kwargs):
 *         f = functools.partial(*self.args, **self.kwargs)             # <<<<<<<<<<<<<<
 *         return f(*args, **kwargs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_functools); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_self->args); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_self->kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
    __PYX_ERR(9, 59, __pyx_L1_error)
  }
  if (likely(PyDict_CheckExact(__pyx_v_self->kwargs))) {
    __pyx_t_3 = PyDict_Copy(__pyx_v_self->kwargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg((PyObject*)&PyDict_Type, __pyx_v_self->kwargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_f = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "HiddenPartial.pxi":60
 *     def __call__(self, *args, **kwargs):
 *         f = functools.partial(*self.args, **self.kwargs)
 *         return f(*args, **kwargs)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_f, __pyx_v_args, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "HiddenPartial.pxi":58
 *         return bytes(type(functools.partial))
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         f = functools.partial(*self.args, **self.kwargs)
 *         return f(*args, **kwargs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_15__HiddenPartial_21__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_20__reduce_cython__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_20__reduce_cython__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.args, self.kwargs)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->args);
  __Pyx_GIVEREF(__pyx_v_self->args);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->args);
  __Pyx_INCREF(__pyx_v_self->kwargs);
  __Pyx_GIVEREF(__pyx_v_self->kwargs);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->kwargs);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.args, self.kwargs)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict_2, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.args, self.kwargs)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.args is not None or self.kwargs is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.args, self.kwargs)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.args is not None or self.kwargs is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->args != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->kwargs != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.kwargs is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.args is not None or self.kwargs is not None
 *     if use_setstate:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle___HiddenPartial); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_155231502);
    __Pyx_GIVEREF(__pyx_int_155231502);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_155231502);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.args is not None or self.kwargs is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, None), state
 *     else:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___HiddenPartial__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle___HiddenPartial); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_155231502);
    __Pyx_GIVEREF(__pyx_int_155231502);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_155231502);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___HiddenPartial__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_15__HiddenPartial_23__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_15__HiddenPartial_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_15__HiddenPartial_22__setstate_cython__(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_15__HiddenPartial_22__setstate_cython__(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___HiddenPartial__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle___HiddenPartial__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___HiddenPartial__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyprotect.protected.__HiddenPartial.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle___ProtectionData(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_55__pyx_unpickle___ProtectionData(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_55__pyx_unpickle___ProtectionData = {"__pyx_unpickle___ProtectionData", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_55__pyx_unpickle___ProtectionData, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_55__pyx_unpickle___ProtectionData(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle___ProtectionData (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___ProtectionData", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___ProtectionData", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle___ProtectionData") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___ProtectionData", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle___ProtectionData", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_54__pyx_unpickle___ProtectionData(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_54__pyx_unpickle___ProtectionData(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___ProtectionData", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfa53bdd, 0xec20346, 0x12e646a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__32, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfa53bdd, 0xec20346, 0x12e646a):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum
 *     __pyx_result = __ProtectionData.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xfa53bdd, 0xec20346, 0x12e646a):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = __ProtectionData.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfa53bdd, 0xec20346, 0x12e646a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum
 *     __pyx_result = __ProtectionData.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected___ProtectionData), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum
 *     __pyx_result = __ProtectionData.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = __ProtectionData.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle___ProtectionData__set_state(((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum
 *     __pyx_result = __ProtectionData.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___ProtectionData(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle___ProtectionData", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle___ProtectionData__set_state(struct __pyx_obj_9pyprotect_9protected___ProtectionData *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___ProtectionData__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->attributes_map);
  __Pyx_DECREF(__pyx_v___pyx_result->attributes_map);
  __pyx_v___pyx_result->attributes_map = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->freeze);
  __Pyx_DECREF(__pyx_v___pyx_result->freeze);
  __pyx_v___pyx_result->freeze = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->hash);
  __Pyx_DECREF(__pyx_v___pyx_result->hash);
  __pyx_v___pyx_result->hash = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->help);
  __Pyx_DECREF(__pyx_v___pyx_result->help);
  __pyx_v___pyx_result->help = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->help_str);
  __Pyx_DECREF(__pyx_v___pyx_result->help_str);
  __pyx_v___pyx_result->help_str = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->id);
  __Pyx_DECREF(__pyx_v___pyx_result->id);
  __pyx_v___pyx_result->id = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->id_class);
  __Pyx_DECREF(__pyx_v___pyx_result->id_class);
  __pyx_v___pyx_result->id_class = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->instanceof);
  __Pyx_DECREF(__pyx_v___pyx_result->instanceof);
  __pyx_v___pyx_result->instanceof = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->isinstance);
  __Pyx_DECREF(__pyx_v___pyx_result->isinstance);
  __pyx_v___pyx_result->isinstance = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->issubclass);
  __Pyx_DECREF(__pyx_v___pyx_result->issubclass);
  __pyx_v___pyx_result->issubclass = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->multiwrapped);
  __Pyx_DECREF(__pyx_v___pyx_result->multiwrapped);
  __pyx_v___pyx_result->multiwrapped = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->private);
  __Pyx_DECREF(__pyx_v___pyx_result->private);
  __pyx_v___pyx_result->private = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->protect);
  __Pyx_DECREF(__pyx_v___pyx_result->protect);
  __pyx_v___pyx_result->protect = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->rules);
  __Pyx_DECREF(__pyx_v___pyx_result->rules);
  __pyx_v___pyx_result->rules = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->subclassof);
  __Pyx_DECREF(__pyx_v___pyx_result->subclassof);
  __pyx_v___pyx_result->subclassof = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->testop);
  __Pyx_DECREF(__pyx_v___pyx_result->testop);
  __pyx_v___pyx_result->testop = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 16);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle___ProtectionData__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Proxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_57__pyx_unpickle_Proxy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_57__pyx_unpickle_Proxy = {"__pyx_unpickle_Proxy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_57__pyx_unpickle_Proxy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_57__pyx_unpickle_Proxy(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Proxy (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Proxy", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Proxy", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Proxy") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Proxy", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Proxy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_56__pyx_unpickle_Proxy(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_56__pyx_unpickle_Proxy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Proxy", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x914c618, 0x9a3f7ee, 0x2fd7cdd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__34, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x914c618, 0x9a3f7ee, 0x2fd7cdd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))" % __pyx_checksum
 *     __pyx_result = Proxy.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x914c618, 0x9a3f7ee, 0x2fd7cdd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Proxy.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x914c618, 0x9a3f7ee, 0x2fd7cdd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))" % __pyx_checksum
 *     __pyx_result = Proxy.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Proxy__set_state(<Proxy> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))" % __pyx_checksum
 *     __pyx_result = Proxy.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Proxy__set_state(<Proxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Proxy.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Proxy__set_state(<Proxy> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Proxy__set_state(Proxy __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Proxy__set_state(((struct __pyx_obj_9pyprotect_9protected_Proxy *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x914c618, 0x9a3f7ee, 0x2fd7cdd) = (frozen, pvt_o))" % __pyx_checksum
 *     __pyx_result = Proxy.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Proxy__set_state(<Proxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Proxy__set_state(<Proxy> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Proxy__set_state(Proxy __pyx_result, tuple __pyx_state):
 *     __pyx_result.frozen = __pyx_state[0]; __pyx_result.pvt_o = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Proxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Proxy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Proxy__set_state(<Proxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Proxy__set_state(Proxy __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.frozen = __pyx_state[0]; __pyx_result.pvt_o = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Proxy__set_state(struct __pyx_obj_9pyprotect_9protected_Proxy *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Proxy__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Proxy__set_state(Proxy __pyx_result, tuple __pyx_state):
 *     __pyx_result.frozen = __pyx_state[0]; __pyx_result.pvt_o = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->pvt_o);
  __pyx_v___pyx_result->pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Proxy__set_state(Proxy __pyx_result, tuple __pyx_state):
 *     __pyx_result.frozen = __pyx_state[0]; __pyx_result.pvt_o = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.frozen = __pyx_state[0]; __pyx_result.pvt_o = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Proxy__set_state(Proxy __pyx_result, tuple __pyx_state):
 *     __pyx_result.frozen = __pyx_state[0]; __pyx_result.pvt_o = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Proxy__set_state(<Proxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Proxy__set_state(Proxy __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.frozen = __pyx_state[0]; __pyx_result.pvt_o = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Proxy__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Wrapped(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_59__pyx_unpickle_Wrapped(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_59__pyx_unpickle_Wrapped = {"__pyx_unpickle_Wrapped", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_59__pyx_unpickle_Wrapped, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_59__pyx_unpickle_Wrapped(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Wrapped (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Wrapped", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Wrapped", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Wrapped") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Wrapped", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_58__pyx_unpickle_Wrapped(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_58__pyx_unpickle_Wrapped(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Wrapped", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Wrapped.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Wrapped.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Wrapped.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Wrapped.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Wrapped.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Wrapped__set_state(((struct __pyx_obj_9pyprotect_9protected_Wrapped *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Wrapped.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Wrapped(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Wrapped", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Wrapped__set_state(struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Wrapped__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->cn);
  __Pyx_DECREF(__pyx_v___pyx_result->cn);
  __pyx_v___pyx_result->cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->hidden_private_attr);
  __pyx_v___pyx_result->hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->protected_attribute);
  __pyx_v___pyx_result->protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->rules);
  __Pyx_DECREF(__pyx_v___pyx_result->rules);
  __pyx_v___pyx_result->rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Wrapped__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Frozen(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_61__pyx_unpickle_Frozen(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_61__pyx_unpickle_Frozen = {"__pyx_unpickle_Frozen", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_61__pyx_unpickle_Frozen, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_61__pyx_unpickle_Frozen(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Frozen (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Frozen", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Frozen", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Frozen") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Frozen", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Frozen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_60__pyx_unpickle_Frozen(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_60__pyx_unpickle_Frozen(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Frozen", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Frozen.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Frozen.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Frozen.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Frozen__set_state(<Frozen> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Frozen), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Frozen.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Frozen__set_state(<Frozen> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Frozen.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Frozen__set_state(<Frozen> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Frozen__set_state(Frozen __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Frozen__set_state(((struct __pyx_obj_9pyprotect_9protected_Frozen *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Frozen.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Frozen__set_state(<Frozen> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Frozen__set_state(<Frozen> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Frozen__set_state(Frozen __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Frozen(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Frozen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Frozen__set_state(<Frozen> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Frozen__set_state(Frozen __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Frozen__set_state(struct __pyx_obj_9pyprotect_9protected_Frozen *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Frozen__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Frozen__set_state(Frozen __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.cn);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.cn);
  __pyx_v___pyx_result->__pyx_base.cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.hidden_private_attr);
  __pyx_v___pyx_result->__pyx_base.hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.protected_attribute);
  __pyx_v___pyx_result->__pyx_base.protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.rules);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.rules);
  __pyx_v___pyx_result->__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Frozen__set_state(Frozen __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Frozen__set_state(Frozen __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Frozen__set_state(<Frozen> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Frozen__set_state(Frozen __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Frozen__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PrivacyDict(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_63__pyx_unpickle_PrivacyDict(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_63__pyx_unpickle_PrivacyDict = {"__pyx_unpickle_PrivacyDict", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_63__pyx_unpickle_PrivacyDict, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_63__pyx_unpickle_PrivacyDict(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PrivacyDict (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PrivacyDict", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PrivacyDict", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PrivacyDict") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PrivacyDict", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_PrivacyDict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_62__pyx_unpickle_PrivacyDict(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_62__pyx_unpickle_PrivacyDict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PrivacyDict", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = PrivacyDict.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = PrivacyDict.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = PrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_PrivacyDict__set_state(((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = PrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PrivacyDict(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_PrivacyDict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_PrivacyDict__set_state(struct __pyx_obj_9pyprotect_9protected_PrivacyDict *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PrivacyDict__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.cn);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.cn);
  __pyx_v___pyx_result->__pyx_base.cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.hidden_private_attr);
  __pyx_v___pyx_result->__pyx_base.hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.protected_attribute);
  __pyx_v___pyx_result->__pyx_base.protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.rules);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.rules);
  __pyx_v___pyx_result->__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_PrivacyDict__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FrozenPrivacyDict(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_65__pyx_unpickle_FrozenPrivacyDict(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_65__pyx_unpickle_FrozenPrivacyDict = {"__pyx_unpickle_FrozenPrivacyDict", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_65__pyx_unpickle_FrozenPrivacyDict, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_65__pyx_unpickle_FrozenPrivacyDict(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenPrivacyDict (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenPrivacyDict", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenPrivacyDict", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FrozenPrivacyDict") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenPrivacyDict", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenPrivacyDict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_64__pyx_unpickle_FrozenPrivacyDict(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_64__pyx_unpickle_FrozenPrivacyDict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenPrivacyDict", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivacyDict.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FrozenPrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivacyDict.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenPrivacyDict__set_state(<FrozenPrivacyDict> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FrozenPrivacyDict__set_state(<FrozenPrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FrozenPrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenPrivacyDict__set_state(<FrozenPrivacyDict> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivacyDict__set_state(FrozenPrivacyDict __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivacyDict__set_state(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivacyDict.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FrozenPrivacyDict__set_state(<FrozenPrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenPrivacyDict__set_state(<FrozenPrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FrozenPrivacyDict__set_state(FrozenPrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FrozenPrivacyDict(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenPrivacyDict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FrozenPrivacyDict__set_state(<FrozenPrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivacyDict__set_state(FrozenPrivacyDict __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivacyDict__set_state(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenPrivacyDict__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivacyDict__set_state(FrozenPrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.cn);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.cn);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.rules);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.rules);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FrozenPrivacyDict__set_state(FrozenPrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FrozenPrivacyDict__set_state(FrozenPrivacyDict __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FrozenPrivacyDict__set_state(<FrozenPrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivacyDict__set_state(FrozenPrivacyDict __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenPrivacyDict__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Private(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_67__pyx_unpickle_Private(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_67__pyx_unpickle_Private = {"__pyx_unpickle_Private", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_67__pyx_unpickle_Private, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_67__pyx_unpickle_Private(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Private (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Private", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Private", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Private") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Private", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Private", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_66__pyx_unpickle_Private(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_66__pyx_unpickle_Private(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Private", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Private.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Private.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Private.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Private), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Private.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Private.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Private__set_state(((struct __pyx_obj_9pyprotect_9protected_Private *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Private.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Private(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Private", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Private__set_state(struct __pyx_obj_9pyprotect_9protected_Private *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Private__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.cn);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.cn);
  __pyx_v___pyx_result->__pyx_base.cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.hidden_private_attr);
  __pyx_v___pyx_result->__pyx_base.hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.protected_attribute);
  __pyx_v___pyx_result->__pyx_base.protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.rules);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.rules);
  __pyx_v___pyx_result->__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Private__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FrozenPrivate(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_69__pyx_unpickle_FrozenPrivate(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_69__pyx_unpickle_FrozenPrivate = {"__pyx_unpickle_FrozenPrivate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_69__pyx_unpickle_FrozenPrivate, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_69__pyx_unpickle_FrozenPrivate(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenPrivate (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenPrivate", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenPrivate", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FrozenPrivate") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenPrivate", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenPrivate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_68__pyx_unpickle_FrozenPrivate(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_68__pyx_unpickle_FrozenPrivate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenPrivate", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivate.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FrozenPrivate.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc76c111, 0x6dc25c3, 0x05bd181):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivate.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenPrivate__set_state(<FrozenPrivate> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivate), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivate.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FrozenPrivate__set_state(<FrozenPrivate> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FrozenPrivate.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenPrivate__set_state(<FrozenPrivate> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivate__set_state(FrozenPrivate __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivate__set_state(((struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc76c111, 0x6dc25c3, 0x05bd181) = (cn, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenPrivate.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FrozenPrivate__set_state(<FrozenPrivate> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenPrivate__set_state(<FrozenPrivate> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FrozenPrivate__set_state(FrozenPrivate __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FrozenPrivate(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenPrivate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FrozenPrivate__set_state(<FrozenPrivate> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivate__set_state(FrozenPrivate __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenPrivate__set_state(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenPrivate__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivate__set_state(FrozenPrivate __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.cn);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.cn);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.rules);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.rules);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FrozenPrivate__set_state(FrozenPrivate __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FrozenPrivate__set_state(FrozenPrivate __pyx_result, tuple __pyx_state):
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FrozenPrivate__set_state(<FrozenPrivate> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenPrivate__set_state(FrozenPrivate __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenPrivate__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Protected(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_71__pyx_unpickle_Protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_71__pyx_unpickle_Protected = {"__pyx_unpickle_Protected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_71__pyx_unpickle_Protected, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_71__pyx_unpickle_Protected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Protected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Protected", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Protected", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Protected") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Protected", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_70__pyx_unpickle_Protected(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_70__pyx_unpickle_Protected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Protected", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__36, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Protected.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Protected.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Protected.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Protected.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Protected.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_Protected__set_state(((struct __pyx_obj_9pyprotect_9protected_Protected *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = Protected.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Protected(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_Protected__set_state(struct __pyx_obj_9pyprotect_9protected_Protected *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Protected__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->acl_cache);
  __Pyx_DECREF(__pyx_v___pyx_result->acl_cache);
  __pyx_v___pyx_result->acl_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.cn);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.cn);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->dir_out);
  __Pyx_DECREF(__pyx_v___pyx_result->dir_out);
  __pyx_v___pyx_result->dir_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.rules);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.rules);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 9);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_Protected__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FrozenProtected(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_73__pyx_unpickle_FrozenProtected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_73__pyx_unpickle_FrozenProtected = {"__pyx_unpickle_FrozenProtected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_73__pyx_unpickle_FrozenProtected, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_73__pyx_unpickle_FrozenProtected(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenProtected (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenProtected", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenProtected", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FrozenProtected") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FrozenProtected", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenProtected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_72__pyx_unpickle_FrozenProtected(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_72__pyx_unpickle_FrozenProtected(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenProtected", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__36, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenProtected.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FrozenProtected.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd078c39, 0xac2112f, 0x4a83031):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenProtected.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenProtected__set_state(<FrozenProtected> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenProtected), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenProtected.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FrozenProtected__set_state(<FrozenProtected> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FrozenProtected.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenProtected__set_state(<FrozenProtected> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenProtected__set_state(FrozenProtected __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenProtected__set_state(((struct __pyx_obj_9pyprotect_9protected_FrozenProtected *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd078c39, 0xac2112f, 0x4a83031) = (acl_cache, cn, dir_out, frozen, hidden_private_attr, oldstyle_class, protected_attribute, pvt_o, rules))" % __pyx_checksum
 *     __pyx_result = FrozenProtected.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FrozenProtected__set_state(<FrozenProtected> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FrozenProtected__set_state(<FrozenProtected> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FrozenProtected__set_state(FrozenProtected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FrozenProtected(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenProtected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FrozenProtected__set_state(<FrozenProtected> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenProtected__set_state(FrozenProtected __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle_FrozenProtected__set_state(struct __pyx_obj_9pyprotect_9protected_FrozenProtected *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FrozenProtected__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenProtected__set_state(FrozenProtected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.acl_cache);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.acl_cache);
  __pyx_v___pyx_result->__pyx_base.acl_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.cn);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.cn);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.cn = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.dir_out);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.dir_out);
  __pyx_v___pyx_result->__pyx_base.dir_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.__pyx_base.frozen = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.hidden_private_attr);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.hidden_private_attr);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.hidden_private_attr = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.oldstyle_class = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9pyprotect_9protected___ProtectionData))))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.protected_attribute);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.protected_attribute);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.__pyx_base.pvt_o);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.__pyx_base.pvt_o = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.rules);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.rules);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.rules = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FrozenProtected__set_state(FrozenProtected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 9);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[9])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FrozenProtected__set_state(FrozenProtected __pyx_result, tuple __pyx_state):
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[9])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FrozenProtected__set_state(<FrozenProtected> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FrozenProtected__set_state(FrozenProtected __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle_FrozenProtected__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle___HiddenPartial(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_9pyprotect_9protected_75__pyx_unpickle___HiddenPartial(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9pyprotect_9protected_75__pyx_unpickle___HiddenPartial = {"__pyx_unpickle___HiddenPartial", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_75__pyx_unpickle___HiddenPartial, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9pyprotect_9protected_75__pyx_unpickle___HiddenPartial(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle___HiddenPartial (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___HiddenPartial", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___HiddenPartial", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle___HiddenPartial") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___HiddenPartial", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle___HiddenPartial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9pyprotect_9protected_74__pyx_unpickle___HiddenPartial(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9pyprotect_9protected_74__pyx_unpickle___HiddenPartial(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___HiddenPartial", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x940a50e, 0xc8cf91d, 0xf0cf4c1):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__37, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x940a50e, 0xc8cf91d, 0xf0cf4c1):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))" % __pyx_checksum
 *     __pyx_result = __HiddenPartial.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x940a50e, 0xc8cf91d, 0xf0cf4c1):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = __HiddenPartial.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x940a50e, 0xc8cf91d, 0xf0cf4c1):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))" % __pyx_checksum
 *     __pyx_result = __HiddenPartial.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___HiddenPartial__set_state(<__HiddenPartial> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))" % __pyx_checksum
 *     __pyx_result = __HiddenPartial.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___HiddenPartial__set_state(<__HiddenPartial> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = __HiddenPartial.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___HiddenPartial__set_state(<__HiddenPartial> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___HiddenPartial__set_state(__HiddenPartial __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_9pyprotect_9protected___pyx_unpickle___HiddenPartial__set_state(((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x940a50e, 0xc8cf91d, 0xf0cf4c1) = (args, kwargs))" % __pyx_checksum
 *     __pyx_result = __HiddenPartial.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___HiddenPartial__set_state(<__HiddenPartial> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___HiddenPartial__set_state(<__HiddenPartial> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___HiddenPartial__set_state(__HiddenPartial __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.kwargs = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___HiddenPartial(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle___HiddenPartial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle___HiddenPartial__set_state(<__HiddenPartial> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___HiddenPartial__set_state(__HiddenPartial __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.kwargs = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_9pyprotect_9protected___pyx_unpickle___HiddenPartial__set_state(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___HiddenPartial__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___HiddenPartial__set_state(__HiddenPartial __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.kwargs = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->args);
  __Pyx_DECREF(__pyx_v___pyx_result->args);
  __pyx_v___pyx_result->args = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->kwargs);
  __Pyx_DECREF(__pyx_v___pyx_result->kwargs);
  __pyx_v___pyx_result->kwargs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle___HiddenPartial__set_state(__HiddenPartial __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.kwargs = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.kwargs = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___HiddenPartial__set_state(__HiddenPartial __pyx_result, tuple __pyx_state):
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.kwargs = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___HiddenPartial__set_state(<__HiddenPartial> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___HiddenPartial__set_state(__HiddenPartial __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.args = __pyx_state[0]; __pyx_result.kwargs = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyprotect.protected.__pyx_unpickle___HiddenPartial__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self")
 * cdef object __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(object (*f)(Wrapped) ):
 *     def wrap(Wrapped self):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped')"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_wrap, "wrap(self: 'Wrapped')");
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_1wrap = {"wrap", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_1wrap, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "wrap") < 0)) __PYX_ERR(6, 67, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = ((struct __pyx_obj_9pyprotect_9protected_Wrapped *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wrap", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped, 1, "self", 0))) __PYX_ERR(6, 67, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_wrap(__pyx_self, __pyx_v_self);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_wrap(PyObject *__pyx_self, struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":69
 *     def wrap(Wrapped self):
 *         """wrap(self: 'Wrapped')"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_f(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self")
 * cdef object __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(object (*f)(Wrapped) ):
 *     def wrap(Wrapped self):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped')"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self")
 * cdef object __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(object (*f)(Wrapped) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self):
 *         """wrap(self: 'Wrapped')"""
 */

static PyObject *__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(PyObject *(*__pyx_v_f)(struct __pyx_obj_9pyprotect_9protected_Wrapped *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(6, 66, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self")
 * cdef object __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(object (*f)(Wrapped) ):
 *     def wrap(Wrapped self):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped')"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_11cfunc_dot_to_py_53__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_1wrap, 0, __pyx_n_s_Pyx_CFunc_object__lParenWrappe, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":70
 *         """wrap(self: 'Wrapped')"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self")
 * cdef object __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(object (*f)(Wrapped) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self):
 *         """wrap(self: 'Wrapped')"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(object (*f)(Wrapped, object) ):
 *     def wrap(Wrapped self, object c):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped', c)"""
 *         return f(self, c)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_wrap, "wrap(self: 'Wrapped', c)");
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_1wrap = {"wrap", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_1wrap, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self = 0;
  PyObject *__pyx_v_c = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_c,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_c,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_c)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("wrap", 1, 2, 2, 1); __PYX_ERR(6, 67, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "wrap") < 0)) __PYX_ERR(6, 67, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = ((struct __pyx_obj_9pyprotect_9protected_Wrapped *)values[0]);
    __pyx_v_c = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wrap", 1, 2, 2, __pyx_nargs); __PYX_ERR(6, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped, 1, "self", 0))) __PYX_ERR(6, 67, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_wrap(__pyx_self, __pyx_v_self, __pyx_v_c);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_wrap(PyObject *__pyx_self, struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_c) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":69
 *     def wrap(Wrapped self, object c):
 *         """wrap(self: 'Wrapped', c)"""
 *         return f(self, c)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_f(__pyx_v_self, __pyx_v_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(object (*f)(Wrapped, object) ):
 *     def wrap(Wrapped self, object c):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped', c)"""
 *         return f(self, c)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(object (*f)(Wrapped, object) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self, object c):
 *         """wrap(self: 'Wrapped', c)"""
 */

static PyObject *__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(PyObject *(*__pyx_v_f)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(6, 66, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(object (*f)(Wrapped, object) ):
 *     def wrap(Wrapped self, object c):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped', c)"""
 *         return f(self, c)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_11cfunc_dot_to_py_70__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_1wrap, 0, __pyx_n_s_Pyx_CFunc_object__lParenWrappe_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":70
 *         """wrap(self: 'Wrapped', c)"""
 *         return f(self, c)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(object (*f)(Wrapped, object) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self, object c):
 *         """wrap(self: 'Wrapped', c)"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(object (*f)(Wrapped, object, object) ):
 *     def wrap(Wrapped self, object a, object op):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped', a, op)"""
 *         return f(self, a, op)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_wrap, "wrap(self: 'Wrapped', a, op)");
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_1wrap = {"wrap", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_1wrap, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self = 0;
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_op = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  {
    #if CYTHON_USE_MODULE_STATE
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_a,&__pyx_n_s_op,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_a,&__pyx_n_s_op,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_a)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("wrap", 1, 3, 3, 1); __PYX_ERR(6, 67, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_op)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("wrap", 1, 3, 3, 2); __PYX_ERR(6, 67, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "wrap") < 0)) __PYX_ERR(6, 67, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_self = ((struct __pyx_obj_9pyprotect_9protected_Wrapped *)values[0]);
    __pyx_v_a = values[1];
    __pyx_v_op = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wrap", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_9pyprotect_9protected_Wrapped, 1, "self", 0))) __PYX_ERR(6, 67, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_wrap(__pyx_self, __pyx_v_self, __pyx_v_a, __pyx_v_op);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_wrap(PyObject *__pyx_self, struct __pyx_obj_9pyprotect_9protected_Wrapped *__pyx_v_self, PyObject *__pyx_v_a, PyObject *__pyx_v_op) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":69
 *     def wrap(Wrapped self, object a, object op):
 *         """wrap(self: 'Wrapped', a, op)"""
 *         return f(self, a, op)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_f(__pyx_v_self, __pyx_v_a, __pyx_v_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(object (*f)(Wrapped, object, object) ):
 *     def wrap(Wrapped self, object a, object op):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped', a, op)"""
 *         return f(self, a, op)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(object (*f)(Wrapped, object, object) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self, object a, object op):
 *         """wrap(self: 'Wrapped', a, op)"""
 */

static PyObject *__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(PyObject *(*__pyx_v_f)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(6, 66, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(object (*f)(Wrapped, object, object) ):
 *     def wrap(Wrapped self, object a, object op):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped', a, op)"""
 *         return f(self, a, op)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_11cfunc_dot_to_py_88__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_1wrap, 0, __pyx_n_s_Pyx_CFunc_object__lParenWrappe_3, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":70
 *         """wrap(self: 'Wrapped', a, op)"""
 *         return f(self, a, op)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(object (*f)(Wrapped, object, object) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self, object a, object op):
 *         """wrap(self: 'Wrapped', a, op)"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_9pyprotect_9protected___ProtectionData(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected___ProtectionData *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)o);
  p->id = Py_None; Py_INCREF(Py_None);
  p->id_class = Py_None; Py_INCREF(Py_None);
  p->hash = Py_None; Py_INCREF(Py_None);
  p->isinstance = Py_None; Py_INCREF(Py_None);
  p->issubclass = Py_None; Py_INCREF(Py_None);
  p->instanceof = Py_None; Py_INCREF(Py_None);
  p->subclassof = Py_None; Py_INCREF(Py_None);
  p->help = Py_None; Py_INCREF(Py_None);
  p->help_str = Py_None; Py_INCREF(Py_None);
  p->testop = Py_None; Py_INCREF(Py_None);
  p->rules = Py_None; Py_INCREF(Py_None);
  p->freeze = Py_None; Py_INCREF(Py_None);
  p->private = Py_None; Py_INCREF(Py_None);
  p->protect = Py_None; Py_INCREF(Py_None);
  p->multiwrapped = Py_None; Py_INCREF(Py_None);
  p->attributes_map = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___ProtectionData(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___ProtectionData *p = (struct __pyx_obj_9pyprotect_9protected___ProtectionData *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->id);
  Py_CLEAR(p->id_class);
  Py_CLEAR(p->hash);
  Py_CLEAR(p->isinstance);
  Py_CLEAR(p->issubclass);
  Py_CLEAR(p->instanceof);
  Py_CLEAR(p->subclassof);
  Py_CLEAR(p->help);
  Py_CLEAR(p->help_str);
  Py_CLEAR(p->testop);
  Py_CLEAR(p->rules);
  Py_CLEAR(p->freeze);
  Py_CLEAR(p->private);
  Py_CLEAR(p->protect);
  Py_CLEAR(p->multiwrapped);
  Py_CLEAR(p->attributes_map);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_9pyprotect_9protected___ProtectionData(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___ProtectionData *p = (struct __pyx_obj_9pyprotect_9protected___ProtectionData *)o;
  if (p->id) {
    e = (*v)(p->id, a); if (e) return e;
  }
  if (p->id_class) {
    e = (*v)(p->id_class, a); if (e) return e;
  }
  if (p->hash) {
    e = (*v)(p->hash, a); if (e) return e;
  }
  if (p->isinstance) {
    e = (*v)(p->isinstance, a); if (e) return e;
  }
  if (p->issubclass) {
    e = (*v)(p->issubclass, a); if (e) return e;
  }
  if (p->instanceof) {
    e = (*v)(p->instanceof, a); if (e) return e;
  }
  if (p->subclassof) {
    e = (*v)(p->subclassof, a); if (e) return e;
  }
  if (p->help) {
    e = (*v)(p->help, a); if (e) return e;
  }
  if (p->help_str) {
    e = (*v)(p->help_str, a); if (e) return e;
  }
  if (p->testop) {
    e = (*v)(p->testop, a); if (e) return e;
  }
  if (p->rules) {
    e = (*v)(p->rules, a); if (e) return e;
  }
  if (p->freeze) {
    e = (*v)(p->freeze, a); if (e) return e;
  }
  if (p->private) {
    e = (*v)(p->private, a); if (e) return e;
  }
  if (p->protect) {
    e = (*v)(p->protect, a); if (e) return e;
  }
  if (p->multiwrapped) {
    e = (*v)(p->multiwrapped, a); if (e) return e;
  }
  if (p->attributes_map) {
    e = (*v)(p->attributes_map, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9pyprotect_9protected___ProtectionData(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9pyprotect_9protected___ProtectionData *p = (struct __pyx_obj_9pyprotect_9protected___ProtectionData *)o;
  tmp = ((PyObject*)p->id);
  p->id = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->id_class);
  p->id_class = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->hash);
  p->hash = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->isinstance);
  p->isinstance = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->issubclass);
  p->issubclass = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->instanceof);
  p->instanceof = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->subclassof);
  p->subclassof = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->help);
  p->help = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->help_str);
  p->help_str = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->testop);
  p->testop = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->rules);
  p->rules = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->freeze);
  p->freeze = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->private);
  p->private = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->protect);
  p->protect = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->multiwrapped);
  p->multiwrapped = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->attributes_map);
  p->attributes_map = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_getattro_9pyprotect_9protected___ProtectionData(PyObject *o, PyObject *n) {
  PyObject *v = __pyx_pw_9pyprotect_9protected_16__ProtectionData_3__getattribute__(o, n);
  return v;
}

static int __pyx_tp_setattro_9pyprotect_9protected___ProtectionData(PyObject *o, PyObject *n, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_5__setattr__(o, n, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_7__delattr__(o, n);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_2id_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_id_class(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_id_class(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_8id_class_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_hash(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_4hash_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_isinstance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_isinstance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10isinstance_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_issubclass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_issubclass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10issubclass_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_instanceof(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_instanceof(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10instanceof_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_subclassof(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_subclassof(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_10subclassof_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_help(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_help(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_4help_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_help_str(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_help_str(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_8help_str_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_testop(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_testop(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_6testop_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_rules(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_rules(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_5rules_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_freeze(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_freeze(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_6freeze_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_private(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_private(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_7private_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_protect(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_protect(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_7protect_5__del__(o);
  }
}

static PyObject *__pyx_getprop_9pyprotect_9protected_16__ProtectionData_multiwrapped(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_1__get__(o);
}

static int __pyx_setprop_9pyprotect_9protected_16__ProtectionData_multiwrapped(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_3__set__(o, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_16__ProtectionData_12multiwrapped_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_9pyprotect_9protected___ProtectionData[] = {
  {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_16__ProtectionData_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_16__ProtectionData_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_16__ProtectionData_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9pyprotect_9protected___ProtectionData[] = {
  {(char *)"id", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_id, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_id, (char *)0, 0},
  {(char *)"id_class", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_id_class, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_id_class, (char *)0, 0},
  {(char *)"hash", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_hash, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_hash, (char *)0, 0},
  {(char *)"isinstance", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_isinstance, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_isinstance, (char *)0, 0},
  {(char *)"issubclass", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_issubclass, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_issubclass, (char *)0, 0},
  {(char *)"instanceof", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_instanceof, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_instanceof, (char *)0, 0},
  {(char *)"subclassof", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_subclassof, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_subclassof, (char *)0, 0},
  {(char *)"help", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_help, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_help, (char *)0, 0},
  {(char *)"help_str", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_help_str, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_help_str, (char *)0, 0},
  {(char *)"testop", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_testop, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_testop, (char *)0, 0},
  {(char *)"rules", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_rules, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_rules, (char *)0, 0},
  {(char *)"freeze", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_freeze, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_freeze, (char *)0, 0},
  {(char *)"private", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_private, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_private, (char *)0, 0},
  {(char *)"protect", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_protect, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_protect, (char *)0, 0},
  {(char *)"multiwrapped", __pyx_getprop_9pyprotect_9protected_16__ProtectionData_multiwrapped, __pyx_setprop_9pyprotect_9protected_16__ProtectionData_multiwrapped, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___ProtectionData_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___ProtectionData},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_9pyprotect_9protected___ProtectionData},
  {Py_tp_setattro, (void *)__pyx_tp_setattro_9pyprotect_9protected___ProtectionData},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Attributes:\n        id: int\n        id_class: int\n        hash: method (no args) -> int\n        isinstance: method: isinstance(o)\n        issubclass: method: issubclass(o)\n        instanceof: method: instanceof(o)\n        subclassof: method: subclassof(o)\n        help: method (no args)\n        help_str: method (no args) -> str\n        testop: method: testop(a: str, op: str) -> bool\n        rules: dict\n        freeze: method (no args) -> Wrapped\n        private: method (no args) -> Private\n        multiwrapped: method (no args) -> bool\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___ProtectionData},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected___ProtectionData},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected___ProtectionData},
  {Py_tp_getset, (void *)__pyx_getsets_9pyprotect_9protected___ProtectionData},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_16__ProtectionData_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___ProtectionData},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___ProtectionData_spec = {
  "pyprotect.protected.__ProtectionData",
  sizeof(struct __pyx_obj_9pyprotect_9protected___ProtectionData),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___ProtectionData_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___ProtectionData = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__ProtectionData", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___ProtectionData), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___ProtectionData, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_9pyprotect_9protected___ProtectionData, /*tp_getattro*/
  __pyx_tp_setattro_9pyprotect_9protected___ProtectionData, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Attributes:\n        id: int\n        id_class: int\n        hash: method (no args) -> int\n        isinstance: method: isinstance(o)\n        issubclass: method: issubclass(o)\n        instanceof: method: instanceof(o)\n        subclassof: method: subclassof(o)\n        help: method (no args)\n        help_str: method (no args) -> str\n        testop: method: testop(a: str, op: str) -> bool\n        rules: dict\n        freeze: method (no args) -> Wrapped\n        private: method (no args) -> Private\n        multiwrapped: method (no args) -> bool\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___ProtectionData, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected___ProtectionData, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected___ProtectionData, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9pyprotect_9protected___ProtectionData, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_16__ProtectionData_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___ProtectionData, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9pyprotect_9protected_Proxy(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_Proxy *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9pyprotect_9protected_Proxy *)o);
  p->pvt_o = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected_Proxy(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected_Proxy *p = (struct __pyx_obj_9pyprotect_9protected_Proxy *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9pyprotect_9protected_Proxy) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->pvt_o);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_9pyprotect_9protected_Proxy(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected_Proxy *p = (struct __pyx_obj_9pyprotect_9protected_Proxy *)o;
  if (p->pvt_o) {
    e = (*v)(p->pvt_o, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9pyprotect_9protected_Proxy(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9pyprotect_9protected_Proxy *p = (struct __pyx_obj_9pyprotect_9protected_Proxy *)o;
  tmp = ((PyObject*)p->pvt_o);
  p->pvt_o = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_9pyprotect_9protected_Proxy(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_9pyprotect_9protected_Proxy(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_5Proxy_132__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_5Proxy_134__delitem__(o, i);
  }
}

static int __pyx_tp_descr_set_9pyprotect_9protected_Proxy(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_5Proxy_154__set__(o, i, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_5Proxy_156__delete__(o, i);
  }
}

static CYTHON_INLINE PyObject *__pyx_nb_add_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_add : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_add);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_add_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_add == &__pyx_nb_add_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_add == &__pyx_nb_add_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_78__radd__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_50__add__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_add == &__pyx_nb_add_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_78__radd__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_subtract_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_subtract : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_subtract);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_subtract_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_subtract == &__pyx_nb_subtract_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_subtract == &__pyx_nb_subtract_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_82__rsub__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_54__sub__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_subtract == &__pyx_nb_subtract_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_82__rsub__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_multiply_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_multiply : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_multiply);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_multiply_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_multiply == &__pyx_nb_multiply_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_multiply == &__pyx_nb_multiply_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_80__rmul__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_52__mul__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_multiply == &__pyx_nb_multiply_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_80__rmul__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_remainder_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_remainder : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_remainder);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_remainder_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_remainder == &__pyx_nb_remainder_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_remainder == &__pyx_nb_remainder_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_90__rmod__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_62__mod__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_remainder == &__pyx_nb_remainder_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_90__rmod__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_divmod_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_divmod : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_divmod);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_divmod_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_divmod == &__pyx_nb_divmod_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_divmod == &__pyx_nb_divmod_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_92__rdivmod__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_64__divmod__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_divmod == &__pyx_nb_divmod_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_92__rdivmod__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_power_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right , PyObject* extra_arg) {
    ternaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_power : NULL;
#else
    slot = (ternaryfunc) PyType_GetSlot(type, Py_nb_power);
#endif
    return slot ? slot(left, right , extra_arg) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_power_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right , PyObject* extra_arg) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_power == &__pyx_nb_power_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_power == &__pyx_nb_power_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_94__rpow__(right, left, extra_arg);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_66__pow__(left, right, extra_arg);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_power == &__pyx_nb_power_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_94__rpow__(right, left, extra_arg);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_lshift_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_lshift : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_lshift);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_lshift_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_lshift == &__pyx_nb_lshift_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_lshift == &__pyx_nb_lshift_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_96__rlshift__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_68__lshift__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_lshift == &__pyx_nb_lshift_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_96__rlshift__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_rshift_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_rshift : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_rshift);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_rshift_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_rshift == &__pyx_nb_rshift_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_rshift == &__pyx_nb_rshift_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_98__rrshift__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_70__rshift__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_rshift == &__pyx_nb_rshift_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_98__rrshift__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_and_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_and : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_and);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_and_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_and == &__pyx_nb_and_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_and == &__pyx_nb_and_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_100__rand__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_72__and__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_and == &__pyx_nb_and_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_100__rand__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_xor_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_xor : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_xor);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_xor_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_xor == &__pyx_nb_xor_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_xor == &__pyx_nb_xor_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_104__rxor__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_76__xor__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_xor == &__pyx_nb_xor_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_104__rxor__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_or_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_or : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_or);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_or_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_or == &__pyx_nb_or_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_or == &__pyx_nb_or_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_102__ror__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_74__or__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_or == &__pyx_nb_or_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_102__ror__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_floor_divide_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_floor_divide : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_floor_divide);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_floor_divide_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_floor_divide == &__pyx_nb_floor_divide_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_floor_divide == &__pyx_nb_floor_divide_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_88__rfloordiv__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_60__floordiv__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_floor_divide == &__pyx_nb_floor_divide_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_88__rfloordiv__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



static CYTHON_INLINE PyObject *__pyx_nb_true_divide_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_true_divide : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_true_divide);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_true_divide_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_true_divide == &__pyx_nb_true_divide_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_true_divide == &__pyx_nb_true_divide_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_86__rtruediv__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_58__truediv__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_true_divide == &__pyx_nb_true_divide_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_86__rtruediv__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}



#if PY_VERSION_HEX >= 0x03050000
static CYTHON_INLINE PyObject *__pyx_nb_matrix_multiply_9pyprotect_9protected_Proxy_maybe_call_slot(PyTypeObject* type, PyObject *left, PyObject *right ) {
    binaryfunc slot;
#if CYTHON_USE_TYPE_SLOTS || PY_MAJOR_VERSION < 3 || CYTHON_COMPILING_IN_PYPY
    slot = type->tp_as_number ? type->tp_as_number->nb_matrix_multiply : NULL;
#else
    slot = (binaryfunc) PyType_GetSlot(type, Py_nb_matrix_multiply);
#endif
    return slot ? slot(left, right ) : __Pyx_NewRef(Py_NotImplemented);
}
static PyObject *__pyx_nb_matrix_multiply_9pyprotect_9protected_Proxy(PyObject *left, PyObject *right ) {
    int maybe_self_is_left, maybe_self_is_right = 0;
    maybe_self_is_left = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
            || (Py_TYPE(left)->tp_as_number && Py_TYPE(left)->tp_as_number->nb_matrix_multiply == &__pyx_nb_matrix_multiply_9pyprotect_9protected_Proxy)
#endif
            || __Pyx_TypeCheck(left, __pyx_ptype_9pyprotect_9protected_Proxy);
    if (!(1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_matrix_multiply == &__pyx_nb_matrix_multiply_9pyprotect_9protected_Proxy)
#endif
                || __Pyx_TypeCheck(right, __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_left) {
        PyObject *res;
        if (maybe_self_is_right && 1 && !(1)) {
            res = __pyx_pw_9pyprotect_9protected_5Proxy_84__rmatmul__(right, left);
            if (res != Py_NotImplemented) return res;
            Py_DECREF(res);
            maybe_self_is_right = 0;
        }
        res = __pyx_pw_9pyprotect_9protected_5Proxy_56__matmul__(left, right);
        if (res != Py_NotImplemented) return res;
        Py_DECREF(res);
    }
    if ((1)) {
        maybe_self_is_right = Py_TYPE(left) == Py_TYPE(right)
#if CYTHON_USE_TYPE_SLOTS
                || (Py_TYPE(right)->tp_as_number && Py_TYPE(right)->tp_as_number->nb_matrix_multiply == &__pyx_nb_matrix_multiply_9pyprotect_9protected_Proxy)
#endif
                || PyType_IsSubtype(Py_TYPE(right), __pyx_ptype_9pyprotect_9protected_Proxy);
    }
    if (maybe_self_is_right) {
        return __pyx_pw_9pyprotect_9protected_5Proxy_84__rmatmul__(right, left);
    }
    return __Pyx_NewRef(Py_NotImplemented);
}


#endif

static PyObject *__pyx_specialmethod___pyx_pw_9pyprotect_9protected_5Proxy_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__(self);
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_Proxy[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__bytes__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_7__bytes__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__format__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_14__format__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__length_hint__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_24__length_hint__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__complex__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_34__complex__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__round__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_42__round__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__trunc__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_44__trunc__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__floor__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_46__floor__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__ceil__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_48__ceil__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__rshift__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_70__rshift__, METH_O|METH_COEXIST, 0},
  {"__radd__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_78__radd__, METH_O|METH_COEXIST, 0},
  {"__rmul__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_80__rmul__, METH_O|METH_COEXIST, 0},
  {"__rsub__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_82__rsub__, METH_O|METH_COEXIST, 0},
  #if PY_VERSION_HEX >= 0x03050000
  {"__rmatmul__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_84__rmatmul__, METH_O|METH_COEXIST, 0},
  #endif
  {"__rtruediv__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_86__rtruediv__, METH_O|METH_COEXIST, 0},
  {"__rfloordiv__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_88__rfloordiv__, METH_O|METH_COEXIST, 0},
  {"__rmod__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_90__rmod__, METH_O|METH_COEXIST, 0},
  {"__rdivmod__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_92__rdivmod__, METH_O|METH_COEXIST, 0},
  {"__rlshift__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_96__rlshift__, METH_O|METH_COEXIST, 0},
  {"__rrshift__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_98__rrshift__, METH_O|METH_COEXIST, 0},
  {"__rand__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_100__rand__, METH_O|METH_COEXIST, 0},
  {"__ror__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_102__ror__, METH_O|METH_COEXIST, 0},
  {"__rxor__", (PyCFunction)__pyx_pw_9pyprotect_9protected_5Proxy_104__rxor__, METH_O|METH_COEXIST, 0},
  {"__instancecheck__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_136__instancecheck__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__subclasscheck__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_138__subclasscheck__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__enter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_140__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__exit__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_142__exit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__aenter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_144__aenter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__aexit__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_146__aexit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__match_args__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_148__match_args__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"send", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_150send, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"throw", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_152throw, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"clear", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_158clear, METH_VARARGS|METH_KEYWORDS, 0},
  {"setdefault", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_160setdefault, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_162pop, METH_VARARGS|METH_KEYWORDS, 0},
  {"popitem", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_164popitem, METH_VARARGS|METH_KEYWORDS, 0},
  {"update", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_166update, METH_VARARGS|METH_KEYWORDS, 0},
  {"append", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_168append, METH_VARARGS|METH_KEYWORDS, 0},
  {"extend", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_170extend, METH_VARARGS|METH_KEYWORDS, 0},
  {"insert", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_172insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"sort", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_174sort, METH_VARARGS|METH_KEYWORDS, 0},
  {"add", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_176add, METH_VARARGS|METH_KEYWORDS, 0},
  {"discard", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_178discard, METH_VARARGS|METH_KEYWORDS, 0},
  {"remove", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_180remove, METH_VARARGS|METH_KEYWORDS, 0},
  {"reverse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_182reverse, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_184__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_5Proxy_186__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_Proxy_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected_Proxy},
  {Py_tp_repr, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_3__repr__},
  {Py_nb_add, (void *)__pyx_nb_add_9pyprotect_9protected_Proxy},
  {Py_nb_subtract, (void *)__pyx_nb_subtract_9pyprotect_9protected_Proxy},
  {Py_nb_multiply, (void *)__pyx_nb_multiply_9pyprotect_9protected_Proxy},
  {Py_nb_remainder, (void *)__pyx_nb_remainder_9pyprotect_9protected_Proxy},
  {Py_nb_divmod, (void *)__pyx_nb_divmod_9pyprotect_9protected_Proxy},
  {Py_nb_power, (void *)__pyx_nb_power_9pyprotect_9protected_Proxy},
  {Py_nb_negative, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_26__neg__},
  {Py_nb_positive, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_28__pos__},
  {Py_nb_absolute, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_30__abs__},
  {Py_nb_bool, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_16__bool__},
  {Py_nb_invert, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_32__invert__},
  {Py_nb_lshift, (void *)__pyx_nb_lshift_9pyprotect_9protected_Proxy},
  {Py_nb_rshift, (void *)__pyx_nb_rshift_9pyprotect_9protected_Proxy},
  {Py_nb_and, (void *)__pyx_nb_and_9pyprotect_9protected_Proxy},
  {Py_nb_xor, (void *)__pyx_nb_xor_9pyprotect_9protected_Proxy},
  {Py_nb_or, (void *)__pyx_nb_or_9pyprotect_9protected_Proxy},
  {Py_nb_int, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_36__int__},
  #if PY_MAJOR_VERSION < 3
  {Py_nb_long, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_36__int__},
  #endif
  {Py_nb_float, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_38__float__},
  {Py_nb_inplace_add, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_106__iadd__},
  {Py_nb_inplace_subtract, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_110__isub__},
  {Py_nb_inplace_multiply, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_108__imul__},
  {Py_nb_inplace_remainder, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_112__imod__},
  {Py_nb_inplace_power, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_124__ipow__},
  {Py_nb_inplace_lshift, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_114__ilshift__},
  {Py_nb_inplace_rshift, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_116__irshift__},
  {Py_nb_inplace_and, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_118__iand__},
  {Py_nb_inplace_xor, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_122__ixor__},
  {Py_nb_inplace_or, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_120__ior__},
  {Py_nb_floor_divide, (void *)__pyx_nb_floor_divide_9pyprotect_9protected_Proxy},
  {Py_nb_true_divide, (void *)__pyx_nb_true_divide_9pyprotect_9protected_Proxy},
  {Py_nb_inplace_floor_divide, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_128__ifloordiv__},
  {Py_nb_inplace_true_divide, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_126__itruediv__},
  {Py_nb_index, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_40__index__},
  #if PY_VERSION_HEX >= 0x03050000
  {Py_nb_matrix_multiply, (void *)__pyx_nb_matrix_multiply_9pyprotect_9protected_Proxy},
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  {Py_nb_inplace_matrix_multiply, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_130__imatmul__},
  #endif
  {Py_sq_length, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_22__len__},
  {Py_sq_item, (void *)__pyx_sq_item_9pyprotect_9protected_Proxy},
  {Py_sq_contains, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_20__contains__},
  {Py_mp_length, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_22__len__},
  {Py_mp_subscript, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_18__getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_9pyprotect_9protected_Proxy},
  {Py_tp_call, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_9__call__},
  {Py_tp_str, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_5__str__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    This is an object wrapper / proxy that implements the Python\n    special dunder methods\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Proxy},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Proxy},
  {Py_tp_iter, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_11__iter__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_Proxy},
  {Py_tp_descr_set, (void *)__pyx_tp_descr_set_9pyprotect_9protected_Proxy},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_5Proxy_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_Proxy},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_Proxy_spec = {
  "pyprotect.protected.Proxy",
  sizeof(struct __pyx_obj_9pyprotect_9protected_Proxy),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_Proxy_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_Proxy = {
  __pyx_nb_add_9pyprotect_9protected_Proxy, /*nb_add*/
  __pyx_nb_subtract_9pyprotect_9protected_Proxy, /*nb_subtract*/
  __pyx_nb_multiply_9pyprotect_9protected_Proxy, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  __pyx_nb_remainder_9pyprotect_9protected_Proxy, /*nb_remainder*/
  __pyx_nb_divmod_9pyprotect_9protected_Proxy, /*nb_divmod*/
  __pyx_nb_power_9pyprotect_9protected_Proxy, /*nb_power*/
  __pyx_pw_9pyprotect_9protected_5Proxy_26__neg__, /*nb_negative*/
  __pyx_pw_9pyprotect_9protected_5Proxy_28__pos__, /*nb_positive*/
  __pyx_pw_9pyprotect_9protected_5Proxy_30__abs__, /*nb_absolute*/
  __pyx_pw_9pyprotect_9protected_5Proxy_16__bool__, /*nb_bool*/
  __pyx_pw_9pyprotect_9protected_5Proxy_32__invert__, /*nb_invert*/
  __pyx_nb_lshift_9pyprotect_9protected_Proxy, /*nb_lshift*/
  __pyx_nb_rshift_9pyprotect_9protected_Proxy, /*nb_rshift*/
  __pyx_nb_and_9pyprotect_9protected_Proxy, /*nb_and*/
  __pyx_nb_xor_9pyprotect_9protected_Proxy, /*nb_xor*/
  __pyx_nb_or_9pyprotect_9protected_Proxy, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  __pyx_pw_9pyprotect_9protected_5Proxy_36__int__, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  __pyx_pw_9pyprotect_9protected_5Proxy_36__int__, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  __pyx_pw_9pyprotect_9protected_5Proxy_38__float__, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  __pyx_pw_9pyprotect_9protected_5Proxy_106__iadd__, /*nb_inplace_add*/
  __pyx_pw_9pyprotect_9protected_5Proxy_110__isub__, /*nb_inplace_subtract*/
  __pyx_pw_9pyprotect_9protected_5Proxy_108__imul__, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  __pyx_pw_9pyprotect_9protected_5Proxy_112__imod__, /*nb_inplace_remainder*/
  __pyx_pw_9pyprotect_9protected_5Proxy_124__ipow__, /*nb_inplace_power*/
  __pyx_pw_9pyprotect_9protected_5Proxy_114__ilshift__, /*nb_inplace_lshift*/
  __pyx_pw_9pyprotect_9protected_5Proxy_116__irshift__, /*nb_inplace_rshift*/
  __pyx_pw_9pyprotect_9protected_5Proxy_118__iand__, /*nb_inplace_and*/
  __pyx_pw_9pyprotect_9protected_5Proxy_122__ixor__, /*nb_inplace_xor*/
  __pyx_pw_9pyprotect_9protected_5Proxy_120__ior__, /*nb_inplace_or*/
  __pyx_nb_floor_divide_9pyprotect_9protected_Proxy, /*nb_floor_divide*/
  __pyx_nb_true_divide_9pyprotect_9protected_Proxy, /*nb_true_divide*/
  __pyx_pw_9pyprotect_9protected_5Proxy_128__ifloordiv__, /*nb_inplace_floor_divide*/
  __pyx_pw_9pyprotect_9protected_5Proxy_126__itruediv__, /*nb_inplace_true_divide*/
  __pyx_pw_9pyprotect_9protected_5Proxy_40__index__, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  __pyx_nb_matrix_multiply_9pyprotect_9protected_Proxy, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  __pyx_pw_9pyprotect_9protected_5Proxy_130__imatmul__, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_Proxy = {
  __pyx_pw_9pyprotect_9protected_5Proxy_22__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_9pyprotect_9protected_Proxy, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_9pyprotect_9protected_5Proxy_20__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Proxy = {
  __pyx_pw_9pyprotect_9protected_5Proxy_22__len__, /*mp_length*/
  __pyx_pw_9pyprotect_9protected_5Proxy_18__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_9pyprotect_9protected_Proxy, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_9pyprotect_9protected_Proxy = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""Proxy", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_Proxy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Proxy, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, /*tp_repr*/
  &__pyx_tp_as_number_Proxy, /*tp_as_number*/
  &__pyx_tp_as_sequence_Proxy, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Proxy, /*tp_as_mapping*/
  0, /*tp_hash*/
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  __pyx_pw_9pyprotect_9protected_5Proxy_5__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    This is an object wrapper / proxy that implements the Python\n    special dunder methods\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Proxy, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Proxy, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_Proxy, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  __pyx_tp_descr_set_9pyprotect_9protected_Proxy, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_5Proxy_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_Proxy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped __pyx_vtable_9pyprotect_9protected_Wrapped;

static PyObject *__pyx_tp_new_9pyprotect_9protected_Wrapped(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_Wrapped *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_Proxy(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_Wrapped *)o);
  p->__pyx_vtab = __pyx_vtabptr_9pyprotect_9protected_Wrapped;
  p->protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)Py_None); Py_INCREF(Py_None);
  p->cn = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->rules = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->hidden_private_attr = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected_Wrapped(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected_Wrapped *p = (struct __pyx_obj_9pyprotect_9protected_Wrapped *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9pyprotect_9protected_Wrapped) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->protected_attribute);
  Py_CLEAR(p->cn);
  Py_CLEAR(p->rules);
  Py_CLEAR(p->hidden_private_attr);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_9pyprotect_9protected_Proxy(o);
}

static int __pyx_tp_traverse_9pyprotect_9protected_Wrapped(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected_Wrapped *p = (struct __pyx_obj_9pyprotect_9protected_Wrapped *)o;
  e = __pyx_tp_traverse_9pyprotect_9protected_Proxy(o, v, a); if (e) return e;
  if (p->protected_attribute) {
    e = (*v)(((PyObject *)p->protected_attribute), a); if (e) return e;
  }
  if (p->rules) {
    e = (*v)(p->rules, a); if (e) return e;
  }
  if (p->hidden_private_attr) {
    e = (*v)(p->hidden_private_attr, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9pyprotect_9protected_Wrapped(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9pyprotect_9protected_Wrapped *p = (struct __pyx_obj_9pyprotect_9protected_Wrapped *)o;
  __pyx_tp_clear_9pyprotect_9protected_Proxy(o);
  tmp = ((PyObject*)p->protected_attribute);
  p->protected_attribute = ((struct __pyx_obj_9pyprotect_9protected___ProtectionData *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->rules);
  p->rules = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->hidden_private_attr);
  p->hidden_private_attr = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_getattro_9pyprotect_9protected_Wrapped(PyObject *o, PyObject *n) {
  PyObject *v = __pyx_pw_9pyprotect_9protected_7Wrapped_3__getattribute__(o, n);
  return v;
}

static int __pyx_tp_setattro_9pyprotect_9protected_Wrapped(PyObject *o, PyObject *n, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_7Wrapped_5__setattr__(o, n, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_7Wrapped_7__delattr__(o, n);
  }
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_Wrapped[] = {
  {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Wrapped_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Wrapped_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Wrapped_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_Wrapped_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected_Wrapped},
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_7Wrapped_13__hash__},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_9pyprotect_9protected_Wrapped},
  {Py_tp_setattro, (void *)__pyx_tp_setattro_9pyprotect_9protected_Wrapped},
  {Py_tp_doc, (void *)PyDoc_STR("\n    This is an object wrapper / proxy that adds the 'frozen' parameter\n    If frozen is False, should behave just like the wrapped object, except\n    following attributes cannot be modified:\n        '__getattribute__', '__delattr__', '__setattr__', '__slots__',\n    If frozen is True, prevents modification of ANY attribute\n    WITHOUT frozen == True:\n        - Does NOT protect CLASS of wrapped object from modification\n        - Does NOT protect __dict__ or __slots__\n    Implements all known special methods for classes under collections etc\n    The one difference is that a Wrapped instance explicitly does NOT\n    support pickling, and will raise a ProtectionError\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Wrapped},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Wrapped},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_7Wrapped_11__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_Wrapped},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_7Wrapped_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_Wrapped},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_Wrapped_spec = {
  "pyprotect.protected.Wrapped",
  sizeof(struct __pyx_obj_9pyprotect_9protected_Wrapped),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_Wrapped_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected_Wrapped = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""Wrapped", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_Wrapped), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Wrapped, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_7Wrapped_13__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_5__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  __pyx_tp_getattro_9pyprotect_9protected_Wrapped, /*tp_getattro*/
  __pyx_tp_setattro_9pyprotect_9protected_Wrapped, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    This is an object wrapper / proxy that adds the 'frozen' parameter\n    If frozen is False, should behave just like the wrapped object, except\n    following attributes cannot be modified:\n        '__getattribute__', '__delattr__', '__setattr__', '__slots__',\n    If frozen is True, prevents modification of ANY attribute\n    WITHOUT frozen == True:\n        - Does NOT protect CLASS of wrapped object from modification\n        - Does NOT protect __dict__ or __slots__\n    Implements all known special methods for classes under collections etc\n    The one difference is that a Wrapped instance explicitly does NOT\n    support pickling, and will raise a ProtectionError\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Wrapped, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Wrapped, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_7Wrapped_11__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_Wrapped, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_7Wrapped_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_Wrapped, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_Frozen __pyx_vtable_9pyprotect_9protected_Frozen;

static PyObject *__pyx_tp_new_9pyprotect_9protected_Frozen(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_Frozen *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_Wrapped(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_Frozen *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped*)__pyx_vtabptr_9pyprotect_9protected_Frozen;
  return o;
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_Frozen[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_6Frozen_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_6Frozen_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_Frozen_slots[] = {
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_6Frozen_3__hash__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Subclass of Wrapped that is automatically frozen\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Wrapped},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Wrapped},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_6Frozen_5__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_Frozen},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_6Frozen_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_Frozen},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_Frozen_spec = {
  "pyprotect.protected.Frozen",
  sizeof(struct __pyx_obj_9pyprotect_9protected_Frozen),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_Frozen_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected_Frozen = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""Frozen", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_Frozen), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Wrapped, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_6Frozen_3__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_5__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Subclass of Wrapped that is automatically frozen\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Wrapped, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Wrapped, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_6Frozen_5__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_Frozen, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_6Frozen_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_Frozen, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_PrivacyDict __pyx_vtable_9pyprotect_9protected_PrivacyDict;

static PyObject *__pyx_tp_new_9pyprotect_9protected_PrivacyDict(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_PrivacyDict *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_Wrapped(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_PrivacyDict *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped*)__pyx_vtabptr_9pyprotect_9protected_PrivacyDict;
  return o;
}
static PyObject *__pyx_sq_item_9pyprotect_9protected_PrivacyDict(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_9pyprotect_9protected_PrivacyDict(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_11PrivacyDict_9__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_11PrivacyDict_7__delitem__(o, i);
  }
}

static PyObject *__pyx_tp_getattro_9pyprotect_9protected_PrivacyDict(PyObject *o, PyObject *n) {
  PyObject *v = __pyx_pw_9pyprotect_9protected_11PrivacyDict_3__getattribute__(o, n);
  return v;
}

static PyObject *__pyx_specialmethod___pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__(self);
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_PrivacyDict[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_15copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"keys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_17keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"items", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_20items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"values", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_23values, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"keys_py2", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_26keys_py2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"items_py2", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_28items_py2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"values_py2", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_30values_py2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"iterkeys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_32iterkeys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"iteritems", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_35iteritems, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"itervalues", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_38itervalues, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"viewitems", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_41viewitems, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"viewvalues", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_43viewvalues, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"viewkeys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_45viewkeys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_51__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_11PrivacyDict_53__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_PrivacyDict_slots[] = {
  {Py_tp_repr, (void *)__pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__},
  {Py_sq_item, (void *)__pyx_sq_item_9pyprotect_9protected_PrivacyDict},
  {Py_mp_subscript, (void *)__pyx_pw_9pyprotect_9protected_11PrivacyDict_5__getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_9pyprotect_9protected_PrivacyDict},
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_11PrivacyDict_47__hash__},
  {Py_tp_str, (void *)__pyx_pw_9pyprotect_9protected_11PrivacyDict_13__str__},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_9pyprotect_9protected_PrivacyDict},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Like types.MappingProxyType - with following additional functionality:\n        - Cannot access traditionally 'private' mangled python attributes\n        - Cannot modify traditionally private attributes (form '_var')\n        - Cannot modify CLASS of wrapped object\n        - TODO: Cannot access any attribute not exported by dir(pvt_o)\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Wrapped},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Wrapped},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_11PrivacyDict_49__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_PrivacyDict},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_11PrivacyDict_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_PrivacyDict},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_PrivacyDict_spec = {
  "pyprotect.protected.PrivacyDict",
  sizeof(struct __pyx_obj_9pyprotect_9protected_PrivacyDict),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_PrivacyDict_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_PrivacyDict = {
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_22__len__, /*sq_length*/
  #else
  0, /*sq_length*/
  #endif
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_9pyprotect_9protected_PrivacyDict, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_20__contains__, /*sq_contains*/
  #else
  0, /*sq_contains*/
  #endif
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_PrivacyDict = {
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_22__len__, /*mp_length*/
  #else
  0, /*mp_length*/
  #endif
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_5__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_9pyprotect_9protected_PrivacyDict, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_9pyprotect_9protected_PrivacyDict = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""PrivacyDict", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_PrivacyDict), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Wrapped, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_PrivacyDict, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_PrivacyDict, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_47__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_13__str__, /*tp_str*/
  __pyx_tp_getattro_9pyprotect_9protected_PrivacyDict, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Like types.MappingProxyType - with following additional functionality:\n        - Cannot access traditionally 'private' mangled python attributes\n        - Cannot modify traditionally private attributes (form '_var')\n        - Cannot modify CLASS of wrapped object\n        - TODO: Cannot access any attribute not exported by dir(pvt_o)\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Wrapped, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Wrapped, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_49__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_PrivacyDict, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_PrivacyDict, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivacyDict __pyx_vtable_9pyprotect_9protected_FrozenPrivacyDict;

static PyObject *__pyx_tp_new_9pyprotect_9protected_FrozenPrivacyDict(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_PrivacyDict(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped*)__pyx_vtabptr_9pyprotect_9protected_FrozenPrivacyDict;
  return o;
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_FrozenPrivacyDict[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_FrozenPrivacyDict_slots[] = {
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_3__hash__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Subclass of PrivacyDict that is automatically frozen\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Wrapped},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Wrapped},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_5__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_FrozenPrivacyDict},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_FrozenPrivacyDict},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_FrozenPrivacyDict_spec = {
  "pyprotect.protected.FrozenPrivacyDict",
  sizeof(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_FrozenPrivacyDict_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected_FrozenPrivacyDict = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""FrozenPrivacyDict", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_FrozenPrivacyDict), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Wrapped, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_11__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_3__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_11PrivacyDict_13__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Subclass of PrivacyDict that is automatically frozen\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Wrapped, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Wrapped, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_5__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_FrozenPrivacyDict, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_17FrozenPrivacyDict_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_FrozenPrivacyDict, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_Private __pyx_vtable_9pyprotect_9protected_Private;

static PyObject *__pyx_tp_new_9pyprotect_9protected_Private(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_Private *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_Wrapped(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_Private *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped*)__pyx_vtabptr_9pyprotect_9protected_Private;
  return o;
}

static PyObject *__pyx_tp_getattro_9pyprotect_9protected_Private(PyObject *o, PyObject *n) {
  PyObject *v = __pyx_pw_9pyprotect_9protected_7Private_3__getattribute__(o, n);
  return v;
}

static int __pyx_tp_setattro_9pyprotect_9protected_Private(PyObject *o, PyObject *n, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_7Private_5__setattr__(o, n, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_7Private_7__delattr__(o, n);
  }
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_Private[] = {
  {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Private_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Private_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_7Private_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_Private_slots[] = {
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_7Private_11__hash__},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_9pyprotect_9protected_Private},
  {Py_tp_setattro, (void *)__pyx_tp_setattro_9pyprotect_9protected_Private},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Subclass of Wrapped with following additional functionality:\n        - Cannot access traditionally 'private' mangled python attributes\n        - Cannot modify traditionally private attributes (form '_var')\n        - Cannot modify CLASS of wrapped object\n        - Cannot modify __dict__ or __slots__ of wrapped object\n        - DONE: Cannot access any attribute not exported by dir(pvt_o)\n        - DONE: cannot access any unmangled double '_' attributes\n        - DONE: Cannot add or delete attributes\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Wrapped},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Wrapped},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_7Private_13__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_Private},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_7Private_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_Private},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_Private_spec = {
  "pyprotect.protected.Private",
  sizeof(struct __pyx_obj_9pyprotect_9protected_Private),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_Private_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected_Private = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""Private", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_Private), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Wrapped, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_7Private_11__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_5__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  __pyx_tp_getattro_9pyprotect_9protected_Private, /*tp_getattro*/
  __pyx_tp_setattro_9pyprotect_9protected_Private, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Subclass of Wrapped with following additional functionality:\n        - Cannot access traditionally 'private' mangled python attributes\n        - Cannot modify traditionally private attributes (form '_var')\n        - Cannot modify CLASS of wrapped object\n        - Cannot modify __dict__ or __slots__ of wrapped object\n        - DONE: Cannot access any attribute not exported by dir(pvt_o)\n        - DONE: cannot access any unmangled double '_' attributes\n        - DONE: Cannot add or delete attributes\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Wrapped, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Wrapped, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_7Private_13__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_Private, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_7Private_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_Private, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_FrozenPrivate __pyx_vtable_9pyprotect_9protected_FrozenPrivate;

static PyObject *__pyx_tp_new_9pyprotect_9protected_FrozenPrivate(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_Private(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_FrozenPrivate *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped*)__pyx_vtabptr_9pyprotect_9protected_FrozenPrivate;
  return o;
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_FrozenPrivate[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_13FrozenPrivate_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_13FrozenPrivate_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_FrozenPrivate_slots[] = {
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_13FrozenPrivate_3__hash__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Subclass of Private that is automatically frozen\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Wrapped},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Wrapped},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_13FrozenPrivate_5__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_FrozenPrivate},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_13FrozenPrivate_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_FrozenPrivate},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_FrozenPrivate_spec = {
  "pyprotect.protected.FrozenPrivate",
  sizeof(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_FrozenPrivate_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected_FrozenPrivate = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""FrozenPrivate", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_FrozenPrivate), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Wrapped, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_13FrozenPrivate_3__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_5__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Subclass of Private that is automatically frozen\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Wrapped, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Wrapped, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_13FrozenPrivate_5__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_FrozenPrivate, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_13FrozenPrivate_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_FrozenPrivate, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_Protected __pyx_vtable_9pyprotect_9protected_Protected;

static PyObject *__pyx_tp_new_9pyprotect_9protected_Protected(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_Protected *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_Private(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_Protected *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped*)__pyx_vtabptr_9pyprotect_9protected_Protected;
  p->acl_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->dir_out = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected_Protected(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected_Protected *p = (struct __pyx_obj_9pyprotect_9protected_Protected *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9pyprotect_9protected_Protected) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->acl_cache);
  Py_CLEAR(p->dir_out);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_9pyprotect_9protected_Wrapped(o);
}

static int __pyx_tp_traverse_9pyprotect_9protected_Protected(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected_Protected *p = (struct __pyx_obj_9pyprotect_9protected_Protected *)o;
  e = __pyx_tp_traverse_9pyprotect_9protected_Wrapped(o, v, a); if (e) return e;
  if (p->acl_cache) {
    e = (*v)(p->acl_cache, a); if (e) return e;
  }
  if (p->dir_out) {
    e = (*v)(p->dir_out, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9pyprotect_9protected_Protected(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9pyprotect_9protected_Protected *p = (struct __pyx_obj_9pyprotect_9protected_Protected *)o;
  __pyx_tp_clear_9pyprotect_9protected_Wrapped(o);
  tmp = ((PyObject*)p->acl_cache);
  p->acl_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->dir_out);
  p->dir_out = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_getattro_9pyprotect_9protected_Protected(PyObject *o, PyObject *n) {
  PyObject *v = __pyx_pw_9pyprotect_9protected_9Protected_3__getattribute__(o, n);
  return v;
}

static int __pyx_tp_setattro_9pyprotect_9protected_Protected(PyObject *o, PyObject *n, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_9Protected_5__setattr__(o, n, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_9Protected_7__delattr__(o, n);
  }
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_Protected[] = {
  {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_9Protected_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_9Protected_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_9Protected_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_Protected_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected_Protected},
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_9Protected_11__hash__},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_9pyprotect_9protected_Protected},
  {Py_tp_setattro, (void *)__pyx_tp_setattro_9pyprotect_9protected_Protected},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Subclass of Private that further restriction of:\n        - Which attributes are VISIBLE\n        - Which attributes are WRITEABLE\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Protected},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Protected},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_9Protected_13__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_Protected},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_9Protected_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_Protected},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_Protected_spec = {
  "pyprotect.protected.Protected",
  sizeof(struct __pyx_obj_9pyprotect_9protected_Protected),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_Protected_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected_Protected = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""Protected", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_Protected), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Protected, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_9Protected_11__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_5__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  __pyx_tp_getattro_9pyprotect_9protected_Protected, /*tp_getattro*/
  __pyx_tp_setattro_9pyprotect_9protected_Protected, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Subclass of Private that further restriction of:\n        - Which attributes are VISIBLE\n        - Which attributes are WRITEABLE\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Protected, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Protected, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_9Protected_13__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_Protected, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_9Protected_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_Protected, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected_FrozenProtected __pyx_vtable_9pyprotect_9protected_FrozenProtected;

static PyObject *__pyx_tp_new_9pyprotect_9protected_FrozenProtected(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected_FrozenProtected *p;
  PyObject *o = __pyx_tp_new_9pyprotect_9protected_Protected(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_9pyprotect_9protected_FrozenProtected *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_9pyprotect_9protected_Wrapped*)__pyx_vtabptr_9pyprotect_9protected_FrozenProtected;
  return o;
}

static PyMethodDef __pyx_methods_9pyprotect_9protected_FrozenProtected[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15FrozenProtected_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15FrozenProtected_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected_FrozenProtected_slots[] = {
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_15FrozenProtected_3__hash__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Subclass of Protected that is automatically frozen\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected_Protected},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected_Protected},
  {Py_tp_richcompare, (void *)__pyx_pw_9pyprotect_9protected_15FrozenProtected_5__richcmp__},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected_FrozenProtected},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_15FrozenProtected_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected_FrozenProtected},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected_FrozenProtected_spec = {
  "pyprotect.protected.FrozenProtected",
  sizeof(struct __pyx_obj_9pyprotect_9protected_FrozenProtected),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected_FrozenProtected_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected_FrozenProtected = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""FrozenProtected", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected_FrozenProtected), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected_Protected, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_15FrozenProtected_3__hash__, /*tp_hash*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_9__call__, /*tp_call*/
  #else
  0, /*tp_call*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_5__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Subclass of Protected that is automatically frozen\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected_Protected, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected_Protected, /*tp_clear*/
  __pyx_pw_9pyprotect_9protected_15FrozenProtected_5__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_9pyprotect_9protected_5Proxy_11__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected_FrozenProtected, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_15FrozenProtected_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected_FrozenProtected, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9pyprotect_9protected___HiddenPartial __pyx_vtable_9pyprotect_9protected___HiddenPartial;

static PyObject *__pyx_tp_new_9pyprotect_9protected___HiddenPartial(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9pyprotect_9protected___HiddenPartial *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)o);
  p->__pyx_vtab = __pyx_vtabptr_9pyprotect_9protected___HiddenPartial;
  p->args = Py_None; Py_INCREF(Py_None);
  p->kwargs = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___HiddenPartial(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___HiddenPartial *p = (struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !_PyGC_FINALIZED(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9pyprotect_9protected___HiddenPartial) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->args);
  Py_CLEAR(p->kwargs);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_9pyprotect_9protected___HiddenPartial(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___HiddenPartial *p = (struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)o;
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  if (p->kwargs) {
    e = (*v)(p->kwargs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9pyprotect_9protected___HiddenPartial(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9pyprotect_9protected___HiddenPartial *p = (struct __pyx_obj_9pyprotect_9protected___HiddenPartial *)o;
  tmp = ((PyObject*)p->args);
  p->args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->kwargs);
  p->kwargs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_getattro_9pyprotect_9protected___HiddenPartial(PyObject *o, PyObject *n) {
  PyObject *v = __pyx_pw_9pyprotect_9protected_15__HiddenPartial_3__getattribute__(o, n);
  return v;
}

static int __pyx_tp_setattro_9pyprotect_9protected___HiddenPartial(PyObject *o, PyObject *n, PyObject *v) {
  if (v) {
    return __pyx_pw_9pyprotect_9protected_15__HiddenPartial_5__setattr__(o, n, v);
  }
  else {
    return __pyx_pw_9pyprotect_9protected_15__HiddenPartial_7__delattr__(o, n);
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_9pyprotect_9protected_15__HiddenPartial_13__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_9pyprotect_9protected_15__HiddenPartial_13__repr__(self);
}

static PyMethodDef __pyx_methods_9pyprotect_9protected___HiddenPartial[] = {
  {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_9__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_9pyprotect_9protected_15__HiddenPartial_13__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__bytes__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_17__bytes__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___HiddenPartial_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___HiddenPartial},
  {Py_tp_repr, (void *)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_13__repr__},
  {Py_tp_hash, (void *)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_11__hash__},
  {Py_tp_call, (void *)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_19__call__},
  {Py_tp_str, (void *)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_15__str__},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_9pyprotect_9protected___HiddenPartial},
  {Py_tp_setattro, (void *)__pyx_tp_setattro_9pyprotect_9protected___HiddenPartial},
  {Py_tp_doc, (void *)PyDoc_STR("\n    This is an object wrapper / proxy for functools.partial\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___HiddenPartial},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected___HiddenPartial},
  {Py_tp_methods, (void *)__pyx_methods_9pyprotect_9protected___HiddenPartial},
  {Py_tp_init, (void *)__pyx_pw_9pyprotect_9protected_15__HiddenPartial_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___HiddenPartial},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___HiddenPartial_spec = {
  "pyprotect.protected.__HiddenPartial",
  sizeof(struct __pyx_obj_9pyprotect_9protected___HiddenPartial),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___HiddenPartial_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___HiddenPartial = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__HiddenPartial", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___HiddenPartial), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___HiddenPartial, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_9pyprotect_9protected_15__HiddenPartial_13__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_9pyprotect_9protected_15__HiddenPartial_11__hash__, /*tp_hash*/
  __pyx_pw_9pyprotect_9protected_15__HiddenPartial_19__call__, /*tp_call*/
  __pyx_pw_9pyprotect_9protected_15__HiddenPartial_15__str__, /*tp_str*/
  __pyx_tp_getattro_9pyprotect_9protected___HiddenPartial, /*tp_getattro*/
  __pyx_tp_setattro_9pyprotect_9protected___HiddenPartial, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    This is an object wrapper / proxy for functools.partial\n    "), /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___HiddenPartial, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected___HiddenPartial, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9pyprotect_9protected___HiddenPartial, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_9pyprotect_9protected_15__HiddenPartial_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___HiddenPartial, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct____iter__[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct____iter__[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_x);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct____iter__[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct____iter__++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_x) {
    e = (*v)(p->__pyx_v_x, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct____iter___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct____iter__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct____iter__},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct____iter__},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct____iter___spec = {
  "pyprotect.protected.__pyx_scope_struct____iter__",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct____iter___slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct____iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_1_comparator[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct_1_comparator = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_1_comparator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_1_comparator > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_1_comparator[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_1_comparator];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_1_comparator(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_op);
  Py_CLEAR(p->__pyx_v_other);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_1_comparator < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct_1_comparator[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_1_comparator++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_1_comparator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *)o;
  if (p->__pyx_v_op) {
    e = (*v)(p->__pyx_v_op, a); if (e) return e;
  }
  if (p->__pyx_v_other) {
    e = (*v)(p->__pyx_v_other, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9pyprotect_9protected___pyx_scope_struct_1_comparator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator *)o;
  tmp = ((PyObject*)p->__pyx_v_op);
  p->__pyx_v_op = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_other);
  p->__pyx_v_other = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_9pyprotect_9protected_Wrapped *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_1_comparator},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_1_comparator},
  {Py_tp_clear, (void *)__pyx_tp_clear_9pyprotect_9protected___pyx_scope_struct_1_comparator},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_1_comparator},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator_spec = {
  "pyprotect.protected.__pyx_scope_struct_1_comparator",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct_1_comparator", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_1_comparator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_1_comparator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_1_comparator, /*tp_traverse*/
  __pyx_tp_clear_9pyprotect_9protected___pyx_scope_struct_1_comparator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_1_comparator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_2_keys[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct_2_keys = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_2_keys(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_2_keys > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_2_keys[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_2_keys];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_2_keys(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_2_keys < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct_2_keys[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_2_keys++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_2_keys(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys *)o;
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_2_keys},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_2_keys},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_2_keys},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys_spec = {
  "pyprotect.protected.__pyx_scope_struct_2_keys",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct_2_keys", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_2_keys), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_2_keys, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_2_keys, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_2_keys, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_3_items[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct_3_items = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_3_items(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_3_items > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_3_items[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_3_items];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_3_items(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_3_items < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct_3_items[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_3_items++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_3_items(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items *)o;
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_3_items},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_3_items},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_3_items},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items_spec = {
  "pyprotect.protected.__pyx_scope_struct_3_items",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct_3_items", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_3_items), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_3_items, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_3_items, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_3_items, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_4_values[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct_4_values = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_4_values(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_4_values > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_4_values[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_4_values];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_4_values(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_4_values < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct_4_values[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_4_values++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_4_values(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values *)o;
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_4_values},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_4_values},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_4_values},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values_spec = {
  "pyprotect.protected.__pyx_scope_struct_4_values",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct_4_values", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_4_values), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_4_values, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_4_values, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_4_values, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_5_iterkeys[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct_5_iterkeys = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_5_iterkeys(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_5_iterkeys > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_5_iterkeys[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_5_iterkeys];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_5_iterkeys(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_5_iterkeys < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct_5_iterkeys[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_5_iterkeys++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_5_iterkeys(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys *)o;
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_5_iterkeys},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_5_iterkeys},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_5_iterkeys},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys_spec = {
  "pyprotect.protected.__pyx_scope_struct_5_iterkeys",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct_5_iterkeys", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_5_iterkeys), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_5_iterkeys, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_5_iterkeys, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_5_iterkeys, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_6_iteritems[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct_6_iteritems = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_6_iteritems(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_6_iteritems > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_6_iteritems[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_6_iteritems];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_6_iteritems(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_6_iteritems < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct_6_iteritems[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_6_iteritems++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_6_iteritems(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems *)o;
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_6_iteritems},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_6_iteritems},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_6_iteritems},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems_spec = {
  "pyprotect.protected.__pyx_scope_struct_6_iteritems",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct_6_iteritems", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_6_iteritems), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_6_iteritems, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_6_iteritems, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_6_iteritems, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_7_itervalues[8];
static int __pyx_freecount_9pyprotect_9protected___pyx_scope_struct_7_itervalues = 0;

static PyObject *__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_7_itervalues(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_7_itervalues > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues)))) {
    o = (PyObject*)__pyx_freelist_9pyprotect_9protected___pyx_scope_struct_7_itervalues[--__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_7_itervalues];
    memset(o, 0, sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_7_itervalues(PyObject *o) {
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_7_itervalues < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues)))) {
    __pyx_freelist_9pyprotect_9protected___pyx_scope_struct_7_itervalues[__pyx_freecount_9pyprotect_9protected___pyx_scope_struct_7_itervalues++] = ((struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_7_itervalues(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *p = (struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues *)o;
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_7_itervalues},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_7_itervalues},
  {Py_tp_new, (void *)__pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_7_itervalues},
  {0, 0},
};
static PyType_Spec __pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues_spec = {
  "pyprotect.protected.__pyx_scope_struct_7_itervalues",
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues_slots,
};
#else

static PyTypeObject __pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct_7_itervalues", /*tp_name*/
  sizeof(struct __pyx_obj_9pyprotect_9protected___pyx_scope_struct_7_itervalues), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9pyprotect_9protected___pyx_scope_struct_7_itervalues, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9pyprotect_9protected___pyx_scope_struct_7_itervalues, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9pyprotect_9protected___pyx_scope_struct_7_itervalues, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self},
  {Py_tp_new, (void *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self},
  {0, 0},
};
static PyType_Spec __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_spec = {
  "pyprotect.protected.__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self",
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_slots,
};
#else

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c},
  {Py_tp_new, (void *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c},
  {0, 0},
};
static PyType_Spec __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_spec = {
  "pyprotect.protected.__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c",
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_slots,
};
#else

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op},
  {Py_tp_new, (void *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op},
  {0, 0},
};
static PyType_Spec __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_spec = {
  "pyprotect.protected.__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op",
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_slots,
};
#else

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyprotect.protected.""__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  #if CYTHON_USE_MODULE_STATE
  {0, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 1},
  {0, __pyx_k_0_1, sizeof(__pyx_k_0_1), 0, 0, 1, 0},
  {0, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {0, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {0, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {0, __pyx_k_Cannot_add_attribute_s_s, sizeof(__pyx_k_Cannot_add_attribute_s_s), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_delete_attribute_s, sizeof(__pyx_k_Cannot_delete_attribute_s), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_delete_attribute_s_s, sizeof(__pyx_k_Cannot_delete_attribute_s_s), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_delete_private_attribute, sizeof(__pyx_k_Cannot_delete_private_attribute), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_modify_attribute_s, sizeof(__pyx_k_Cannot_modify_attribute_s), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_set_attribute_s_s, sizeof(__pyx_k_Cannot_set_attribute_s_s), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_set_private_attribute_s_s, sizeof(__pyx_k_Cannot_set_private_attribute_s_s), 0, 0, 1, 0},
  {0, __pyx_k_CollectionsABC, sizeof(__pyx_k_CollectionsABC), 0, 0, 1, 1},
  {0, __pyx_k_Double_wrapped, sizeof(__pyx_k_Double_wrapped), 0, 0, 1, 0},
  {0, __pyx_k_FrameType, sizeof(__pyx_k_FrameType), 0, 0, 1, 1},
  {0, __pyx_k_Frozen, sizeof(__pyx_k_Frozen), 0, 0, 1, 1},
  {0, __pyx_k_FrozenPrivacyDict, sizeof(__pyx_k_FrozenPrivacyDict), 0, 0, 1, 1},
  {0, __pyx_k_FrozenPrivacyDict___reduce_cytho, sizeof(__pyx_k_FrozenPrivacyDict___reduce_cytho), 0, 0, 1, 1},
  {0, __pyx_k_FrozenPrivacyDict___setstate_cyt, sizeof(__pyx_k_FrozenPrivacyDict___setstate_cyt), 0, 0, 1, 1},
  {0, __pyx_k_FrozenPrivate, sizeof(__pyx_k_FrozenPrivate), 0, 0, 1, 1},
  {0, __pyx_k_FrozenPrivate___reduce_cython, sizeof(__pyx_k_FrozenPrivate___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_FrozenPrivate___setstate_cython, sizeof(__pyx_k_FrozenPrivate___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_FrozenProtected, sizeof(__pyx_k_FrozenProtected), 0, 0, 1, 1},
  {0, __pyx_k_FrozenProtected___reduce_cython, sizeof(__pyx_k_FrozenProtected___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_FrozenProtected___setstate_cytho, sizeof(__pyx_k_FrozenProtected___setstate_cytho), 0, 0, 1, 1},
  {0, __pyx_k_Frozen___reduce_cython, sizeof(__pyx_k_Frozen___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Frozen___setstate_cython, sizeof(__pyx_k_Frozen___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_HiddenPartial, sizeof(__pyx_k_HiddenPartial), 0, 0, 1, 1},
  {0, __pyx_k_HiddenPartial___bytes, sizeof(__pyx_k_HiddenPartial___bytes), 0, 0, 1, 1},
  {0, __pyx_k_HiddenPartial___dir, sizeof(__pyx_k_HiddenPartial___dir), 0, 0, 1, 1},
  {0, __pyx_k_HiddenPartial___reduce_cython, sizeof(__pyx_k_HiddenPartial___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_HiddenPartial___setstate_cytho, sizeof(__pyx_k_HiddenPartial___setstate_cytho), 0, 0, 1, 1},
  {0, __pyx_k_HiddenPartial_pxi, sizeof(__pyx_k_HiddenPartial_pxi), 0, 0, 1, 0},
  {0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {0, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {0, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
  {0, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
  {0, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
  {0, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {0, __pyx_k_Mapping, sizeof(__pyx_k_Mapping), 0, 0, 1, 1},
  {0, __pyx_k_ModuleType, sizeof(__pyx_k_ModuleType), 0, 0, 1, 1},
  {0, __pyx_k_MutableMapping, sizeof(__pyx_k_MutableMapping), 0, 0, 1, 1},
  {0, __pyx_k_MutableSequence, sizeof(__pyx_k_MutableSequence), 0, 0, 1, 1},
  {0, __pyx_k_MutableSet, sizeof(__pyx_k_MutableSet), 0, 0, 1, 1},
  {0, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
  {0, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {0, __pyx_k_Object_Private_s_has_no_attribut, sizeof(__pyx_k_Object_Private_s_has_no_attribut), 0, 0, 1, 0},
  {0, __pyx_k_Object_Protected_s_has_no_attrib, sizeof(__pyx_k_Object_Protected_s_has_no_attrib), 0, 0, 1, 0},
  {0, __pyx_k_Object_Wrapped_s_has_no_attribut, sizeof(__pyx_k_Object_Wrapped_s_has_no_attribut), 0, 0, 1, 0},
  {0, __pyx_k_Object___HiddenPartial_has_no_at, sizeof(__pyx_k_Object___HiddenPartial_has_no_at), 0, 0, 1, 0},
  {0, __pyx_k_Object_is_read_only, sizeof(__pyx_k_Object_is_read_only), 0, 0, 1, 0},
  {0, __pyx_k_Object_s_has_no_attribute_s, sizeof(__pyx_k_Object_s_has_no_attribute_s), 0, 0, 1, 0},
  {0, __pyx_k_PYPY, sizeof(__pyx_k_PYPY), 0, 0, 1, 1},
  {0, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict, sizeof(__pyx_k_PrivacyDict), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_FrozenPrivacyDict_px, sizeof(__pyx_k_PrivacyDict_FrozenPrivacyDict_px), 0, 0, 1, 0},
  {0, __pyx_k_PrivacyDict___reduce_cython, sizeof(__pyx_k_PrivacyDict___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict___setstate_cython, sizeof(__pyx_k_PrivacyDict___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_copy, sizeof(__pyx_k_PrivacyDict_copy), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_items, sizeof(__pyx_k_PrivacyDict_items), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_items_py2, sizeof(__pyx_k_PrivacyDict_items_py2), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_iteritems, sizeof(__pyx_k_PrivacyDict_iteritems), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_iterkeys, sizeof(__pyx_k_PrivacyDict_iterkeys), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_itervalues, sizeof(__pyx_k_PrivacyDict_itervalues), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_keys, sizeof(__pyx_k_PrivacyDict_keys), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_keys_py2, sizeof(__pyx_k_PrivacyDict_keys_py2), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_values, sizeof(__pyx_k_PrivacyDict_values), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_values_py2, sizeof(__pyx_k_PrivacyDict_values_py2), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_viewitems, sizeof(__pyx_k_PrivacyDict_viewitems), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_viewkeys, sizeof(__pyx_k_PrivacyDict_viewkeys), 0, 0, 1, 1},
  {0, __pyx_k_PrivacyDict_viewvalues, sizeof(__pyx_k_PrivacyDict_viewvalues), 0, 0, 1, 1},
  {0, __pyx_k_Private, sizeof(__pyx_k_Private), 0, 0, 1, 1},
  {0, __pyx_k_Private_FrozenPrivate_pxi, sizeof(__pyx_k_Private_FrozenPrivate_pxi), 0, 0, 1, 0},
  {0, __pyx_k_Private___dir, sizeof(__pyx_k_Private___dir), 0, 0, 1, 1},
  {0, __pyx_k_Private___reduce_cython, sizeof(__pyx_k_Private___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Private___setstate_cython, sizeof(__pyx_k_Private___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Protected, sizeof(__pyx_k_Protected), 0, 0, 1, 1},
  {0, __pyx_k_Protected_2, sizeof(__pyx_k_Protected_2), 0, 0, 1, 1},
  {0, __pyx_k_Protected_FrozenProtected_pxi, sizeof(__pyx_k_Protected_FrozenProtected_pxi), 0, 0, 1, 0},
  {0, __pyx_k_Protected___dir, sizeof(__pyx_k_Protected___dir), 0, 0, 1, 1},
  {0, __pyx_k_Protected___reduce_cython, sizeof(__pyx_k_Protected___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Protected___setstate_cython, sizeof(__pyx_k_Protected___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_ProtectionData, sizeof(__pyx_k_ProtectionData), 0, 0, 1, 1},
  {0, __pyx_k_ProtectionData___dir, sizeof(__pyx_k_ProtectionData___dir), 0, 0, 1, 1},
  {0, __pyx_k_ProtectionData___reduce_cython, sizeof(__pyx_k_ProtectionData___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_ProtectionData___setstate_cyth, sizeof(__pyx_k_ProtectionData___setstate_cyth), 0, 0, 1, 1},
  {0, __pyx_k_ProtectionData_pxi, sizeof(__pyx_k_ProtectionData_pxi), 0, 0, 1, 0},
  {0, __pyx_k_ProtectionError, sizeof(__pyx_k_ProtectionError), 0, 0, 1, 1},
  {0, __pyx_k_Proxy, sizeof(__pyx_k_Proxy), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___aenter, sizeof(__pyx_k_Proxy___aenter), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___aexit, sizeof(__pyx_k_Proxy___aexit), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___bytes, sizeof(__pyx_k_Proxy___bytes), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___ceil, sizeof(__pyx_k_Proxy___ceil), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___complex, sizeof(__pyx_k_Proxy___complex), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___enter, sizeof(__pyx_k_Proxy___enter), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___exit, sizeof(__pyx_k_Proxy___exit), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___floor, sizeof(__pyx_k_Proxy___floor), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___format, sizeof(__pyx_k_Proxy___format), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___instancecheck, sizeof(__pyx_k_Proxy___instancecheck), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___iter, sizeof(__pyx_k_Proxy___iter), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___length_hint, sizeof(__pyx_k_Proxy___length_hint), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___match_args, sizeof(__pyx_k_Proxy___match_args), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___reduce_cython, sizeof(__pyx_k_Proxy___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___round, sizeof(__pyx_k_Proxy___round), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___setstate_cython, sizeof(__pyx_k_Proxy___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___subclasscheck, sizeof(__pyx_k_Proxy___subclasscheck), 0, 0, 1, 1},
  {0, __pyx_k_Proxy___trunc, sizeof(__pyx_k_Proxy___trunc), 0, 0, 1, 1},
  {0, __pyx_k_Proxy__aexit, sizeof(__pyx_k_Proxy__aexit), 0, 0, 1, 1},
  {0, __pyx_k_Proxy__exit, sizeof(__pyx_k_Proxy__exit), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_add, sizeof(__pyx_k_Proxy_add), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_append, sizeof(__pyx_k_Proxy_append), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_clear, sizeof(__pyx_k_Proxy_clear), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_discard, sizeof(__pyx_k_Proxy_discard), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_extend, sizeof(__pyx_k_Proxy_extend), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_insert, sizeof(__pyx_k_Proxy_insert), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_pop, sizeof(__pyx_k_Proxy_pop), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_popitem, sizeof(__pyx_k_Proxy_popitem), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_pxi, sizeof(__pyx_k_Proxy_pxi), 0, 0, 1, 0},
  {0, __pyx_k_Proxy_remove, sizeof(__pyx_k_Proxy_remove), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_reverse, sizeof(__pyx_k_Proxy_reverse), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_send, sizeof(__pyx_k_Proxy_send), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_setdefault, sizeof(__pyx_k_Proxy_setdefault), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_sort, sizeof(__pyx_k_Proxy_sort), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_throw, sizeof(__pyx_k_Proxy_throw), 0, 0, 1, 1},
  {0, __pyx_k_Proxy_update, sizeof(__pyx_k_Proxy_update), 0, 0, 1, 1},
  {0, __pyx_k_PyPy, sizeof(__pyx_k_PyPy), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_CFunc_object__lParenWrappe, sizeof(__pyx_k_Pyx_CFunc_object__lParenWrappe), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_CFunc_object__lParenWrappe_2, sizeof(__pyx_k_Pyx_CFunc_object__lParenWrappe_2), 0, 0, 1, 1},
  {0, __pyx_k_Pyx_CFunc_object__lParenWrappe_3, sizeof(__pyx_k_Pyx_CFunc_object__lParenWrappe_3), 0, 0, 1, 1},
  {0, __pyx_k_Read_only_attribute_s, sizeof(__pyx_k_Read_only_attribute_s), 0, 0, 1, 0},
  {0, __pyx_k_RecursionError, sizeof(__pyx_k_RecursionError), 0, 0, 1, 1},
  {0, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {0, __pyx_k_Sequence, sizeof(__pyx_k_Sequence), 0, 0, 1, 1},
  {0, __pyx_k_Set, sizeof(__pyx_k_Set), 0, 0, 1, 1},
  {0, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {0, __pyx_k_Unknown_OldStyle_Class, sizeof(__pyx_k_Unknown_OldStyle_Class), 0, 0, 1, 1},
  {0, __pyx_k_Wrapped, sizeof(__pyx_k_Wrapped), 0, 0, 1, 1},
  {0, __pyx_k_Wrapped_Frozen_pxi, sizeof(__pyx_k_Wrapped_Frozen_pxi), 0, 0, 1, 0},
  {0, __pyx_k_Wrapped___dir, sizeof(__pyx_k_Wrapped___dir), 0, 0, 1, 1},
  {0, __pyx_k_Wrapped___reduce_cython, sizeof(__pyx_k_Wrapped___reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_Wrapped___setstate_cython, sizeof(__pyx_k_Wrapped___setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_Wrapped_comparator_locals_pass_t, sizeof(__pyx_k_Wrapped_comparator_locals_pass_t), 0, 0, 1, 1},
  {0, __pyx_k_Wrapped_object_cannot_be_pickled, sizeof(__pyx_k_Wrapped_object_cannot_be_pickled), 0, 0, 1, 0},
  {0, __pyx_k__17, sizeof(__pyx_k__17), 0, 0, 1, 0},
  {0, __pyx_k__182, sizeof(__pyx_k__182), 0, 0, 1, 1},
  {0, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 1},
  {0, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {0, __pyx_k__33, sizeof(__pyx_k__33), 0, 1, 0, 0},
  {0, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {0, __pyx_k__44, sizeof(__pyx_k__44), 0, 0, 1, 1},
  {0, __pyx_k__88, sizeof(__pyx_k__88), 0, 0, 1, 0},
  {0, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {0, __pyx_k_a_zA_Z_a_zA_Z0_9, sizeof(__pyx_k_a_zA_Z_a_zA_Z0_9), 0, 0, 1, 0},
  {0, __pyx_k_a_zA_Z_a_zA_Z0_9_2, sizeof(__pyx_k_a_zA_Z_a_zA_Z0_9_2), 0, 0, 1, 0},
  {0, __pyx_k_abc, sizeof(__pyx_k_abc), 0, 0, 1, 1},
  {0, __pyx_k_abs, sizeof(__pyx_k_abs), 0, 0, 1, 1},
  {0, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {0, __pyx_k_add_2, sizeof(__pyx_k_add_2), 0, 0, 1, 1},
  {0, __pyx_k_aenter, sizeof(__pyx_k_aenter), 0, 0, 1, 1},
  {0, __pyx_k_aexit, sizeof(__pyx_k_aexit), 0, 0, 1, 1},
  {0, __pyx_k_alist, sizeof(__pyx_k_alist), 0, 0, 1, 1},
  {0, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {0, __pyx_k_always_delegated_attributes, sizeof(__pyx_k_always_delegated_attributes), 0, 0, 1, 1},
  {0, __pyx_k_and, sizeof(__pyx_k_and), 0, 0, 1, 1},
  {0, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {0, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {0, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
  {0, __pyx_k_attr_type_check, sizeof(__pyx_k_attr_type_check), 0, 0, 1, 1},
  {0, __pyx_k_attribute_protected, sizeof(__pyx_k_attribute_protected), 0, 0, 1, 1},
  {0, __pyx_k_basestring, sizeof(__pyx_k_basestring), 0, 0, 1, 1},
  {0, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
  {0, __pyx_k_bool_2, sizeof(__pyx_k_bool_2), 0, 0, 1, 1},
  {0, __pyx_k_build_regex, sizeof(__pyx_k_build_regex), 0, 0, 1, 1},
  {0, __pyx_k_builtin, sizeof(__pyx_k_builtin), 0, 0, 1, 1},
  {0, __pyx_k_builtins, sizeof(__pyx_k_builtins), 0, 0, 1, 1},
  {0, __pyx_k_bytearray, sizeof(__pyx_k_bytearray), 0, 0, 1, 1},
  {0, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {0, __pyx_k_bytes_2, sizeof(__pyx_k_bytes_2), 0, 0, 1, 1},
  {0, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {0, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
  {0, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 0, 1, 1},
  {0, __pyx_k_ceil_2, sizeof(__pyx_k_ceil_2), 0, 0, 1, 1},
  {0, __pyx_k_cfunc_to_py, sizeof(__pyx_k_cfunc_to_py), 0, 0, 1, 1},
  {0, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {0, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
  {0, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
  {0, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {0, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {0, __pyx_k_cmp, sizeof(__pyx_k_cmp), 0, 0, 1, 1},
  {0, __pyx_k_cn, sizeof(__pyx_k_cn), 0, 0, 1, 1},
  {0, __pyx_k_code, sizeof(__pyx_k_code), 0, 0, 1, 1},
  {0, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {0, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 1},
  {0, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {0, __pyx_k_complex, sizeof(__pyx_k_complex), 0, 0, 1, 1},
  {0, __pyx_k_complex_2, sizeof(__pyx_k_complex_2), 0, 0, 1, 1},
  {0, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {0, __pyx_k_contains_2, sizeof(__pyx_k_contains_2), 0, 0, 1, 1},
  {0, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {0, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {0, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {0, __pyx_k_delattr, sizeof(__pyx_k_delattr), 0, 0, 1, 1},
  {0, __pyx_k_delete, sizeof(__pyx_k_delete), 0, 0, 1, 1},
  {0, __pyx_k_delitem, sizeof(__pyx_k_delitem), 0, 0, 1, 1},
  {0, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {0, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
  {0, __pyx_k_dict_3, sizeof(__pyx_k_dict_3), 0, 0, 1, 1},
  {0, __pyx_k_difference, sizeof(__pyx_k_difference), 0, 0, 1, 1},
  {0, __pyx_k_dir, sizeof(__pyx_k_dir), 0, 0, 1, 1},
  {0, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
  {0, __pyx_k_discard, sizeof(__pyx_k_discard), 0, 0, 1, 1},
  {0, __pyx_k_divmod, sizeof(__pyx_k_divmod), 0, 0, 1, 1},
  {0, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {0, __pyx_k_dynamic, sizeof(__pyx_k_dynamic), 0, 0, 1, 1},
  {0, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
  {0, __pyx_k_endswith, sizeof(__pyx_k_endswith), 0, 0, 1, 1},
  {0, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {0, __pyx_k_environ, sizeof(__pyx_k_environ), 0, 0, 1, 1},
  {0, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {0, __pyx_k_exc_type, sizeof(__pyx_k_exc_type), 0, 0, 1, 1},
  {0, __pyx_k_exc_value, sizeof(__pyx_k_exc_value), 0, 0, 1, 1},
  {0, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {0, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {0, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {0, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {0, __pyx_k_float_2, sizeof(__pyx_k_float_2), 0, 0, 1, 1},
  {0, __pyx_k_floor, sizeof(__pyx_k_floor), 0, 0, 1, 1},
  {0, __pyx_k_floor_2, sizeof(__pyx_k_floor_2), 0, 0, 1, 1},
  {0, __pyx_k_floordiv, sizeof(__pyx_k_floordiv), 0, 0, 1, 1},
  {0, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {0, __pyx_k_freeze, sizeof(__pyx_k_freeze), 0, 0, 1, 1},
  {0, __pyx_k_frozen, sizeof(__pyx_k_frozen), 0, 0, 1, 1},
  {0, __pyx_k_frozenset, sizeof(__pyx_k_frozenset), 0, 0, 1, 1},
  {0, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {0, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
  {0, __pyx_k_ge, sizeof(__pyx_k_ge), 0, 0, 1, 1},
  {0, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {0, __pyx_k_getattribute, sizeof(__pyx_k_getattribute), 0, 0, 1, 1},
  {0, __pyx_k_getitem, sizeof(__pyx_k_getitem), 0, 0, 1, 1},
  {0, __pyx_k_getsate, sizeof(__pyx_k_getsate), 0, 0, 1, 1},
  {0, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {0, __pyx_k_global_c_functions_pxi, sizeof(__pyx_k_global_c_functions_pxi), 0, 0, 1, 0},
  {0, __pyx_k_gt, sizeof(__pyx_k_gt), 0, 0, 1, 1},
  {0, __pyx_k_hash, sizeof(__pyx_k_hash), 0, 0, 1, 1},
  {0, __pyx_k_hash_2, sizeof(__pyx_k_hash_2), 0, 0, 1, 1},
  {0, __pyx_k_hash_protected, sizeof(__pyx_k_hash_protected), 0, 0, 1, 1},
  {0, __pyx_k_hash_val, sizeof(__pyx_k_hash_val), 0, 0, 1, 1},
  {0, __pyx_k_help, sizeof(__pyx_k_help), 0, 0, 1, 1},
  {0, __pyx_k_help_protected, sizeof(__pyx_k_help_protected), 0, 0, 1, 1},
  {0, __pyx_k_help_str, sizeof(__pyx_k_help_str), 0, 0, 1, 1},
  {0, __pyx_k_help_val, sizeof(__pyx_k_help_val), 0, 0, 1, 1},
  {0, __pyx_k_hidden_pickle_attributes, sizeof(__pyx_k_hidden_pickle_attributes), 0, 0, 1, 1},
  {0, __pyx_k_hide, sizeof(__pyx_k_hide), 0, 0, 1, 1},
  {0, __pyx_k_hide_private, sizeof(__pyx_k_hide_private), 0, 0, 1, 1},
  {0, __pyx_k_hide_regex, sizeof(__pyx_k_hide_regex), 0, 0, 1, 1},
  {0, __pyx_k_iadd, sizeof(__pyx_k_iadd), 0, 0, 1, 1},
  {0, __pyx_k_iand, sizeof(__pyx_k_iand), 0, 0, 1, 1},
  {0, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {0, __pyx_k_id_class, sizeof(__pyx_k_id_class), 0, 0, 1, 1},
  {0, __pyx_k_id_protected, sizeof(__pyx_k_id_protected), 0, 0, 1, 1},
  {0, __pyx_k_id_val, sizeof(__pyx_k_id_val), 0, 0, 1, 1},
  {0, __pyx_k_ifloordiv, sizeof(__pyx_k_ifloordiv), 0, 0, 1, 1},
  {0, __pyx_k_ilshift, sizeof(__pyx_k_ilshift), 0, 0, 1, 1},
  {0, __pyx_k_imatmul, sizeof(__pyx_k_imatmul), 0, 0, 1, 1},
  {0, __pyx_k_immutable_builtin_attributes, sizeof(__pyx_k_immutable_builtin_attributes), 0, 0, 1, 1},
  {0, __pyx_k_imod, sizeof(__pyx_k_imod), 0, 0, 1, 1},
  {0, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {0, __pyx_k_imul, sizeof(__pyx_k_imul), 0, 0, 1, 1},
  {0, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {0, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {0, __pyx_k_init___locals_C, sizeof(__pyx_k_init___locals_C), 0, 0, 1, 1},
  {0, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
  {0, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {0, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {0, __pyx_k_inst, sizeof(__pyx_k_inst), 0, 0, 1, 1},
  {0, __pyx_k_instance_of_protected, sizeof(__pyx_k_instance_of_protected), 0, 0, 1, 1},
  {0, __pyx_k_instancecheck, sizeof(__pyx_k_instancecheck), 0, 0, 1, 1},
  {0, __pyx_k_instanceof, sizeof(__pyx_k_instanceof), 0, 0, 1, 1},
  {0, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
  {0, __pyx_k_int_2, sizeof(__pyx_k_int_2), 0, 0, 1, 1},
  {0, __pyx_k_intersection, sizeof(__pyx_k_intersection), 0, 0, 1, 1},
  {0, __pyx_k_invert, sizeof(__pyx_k_invert), 0, 0, 1, 1},
  {0, __pyx_k_ior, sizeof(__pyx_k_ior), 0, 0, 1, 1},
  {0, __pyx_k_ipow, sizeof(__pyx_k_ipow), 0, 0, 1, 1},
  {0, __pyx_k_irshift, sizeof(__pyx_k_irshift), 0, 0, 1, 1},
  {0, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
  {0, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
  {0, __pyx_k_isfrozen, sizeof(__pyx_k_isfrozen), 0, 0, 1, 1},
  {0, __pyx_k_isimmutable, sizeof(__pyx_k_isimmutable), 0, 0, 1, 1},
  {0, __pyx_k_isinstance, sizeof(__pyx_k_isinstance), 0, 0, 1, 1},
  {0, __pyx_k_isinstance_protected, sizeof(__pyx_k_isinstance_protected), 0, 0, 1, 1},
  {0, __pyx_k_isinstance_val, sizeof(__pyx_k_isinstance_val), 0, 0, 1, 1},
  {0, __pyx_k_isprivate, sizeof(__pyx_k_isprivate), 0, 0, 1, 1},
  {0, __pyx_k_isprotected, sizeof(__pyx_k_isprotected), 0, 0, 1, 1},
  {0, __pyx_k_isreadonly, sizeof(__pyx_k_isreadonly), 0, 0, 1, 1},
  {0, __pyx_k_issubclass, sizeof(__pyx_k_issubclass), 0, 0, 1, 1},
  {0, __pyx_k_issubclass_protected, sizeof(__pyx_k_issubclass_protected), 0, 0, 1, 1},
  {0, __pyx_k_issubclass_val, sizeof(__pyx_k_issubclass_val), 0, 0, 1, 1},
  {0, __pyx_k_isub, sizeof(__pyx_k_isub), 0, 0, 1, 1},
  {0, __pyx_k_isvisible, sizeof(__pyx_k_isvisible), 0, 0, 1, 1},
  {0, __pyx_k_iswrapped, sizeof(__pyx_k_iswrapped), 0, 0, 1, 1},
  {0, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {0, __pyx_k_items_py2, sizeof(__pyx_k_items_py2), 0, 0, 1, 1},
  {0, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {0, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {0, __pyx_k_iterkeys, sizeof(__pyx_k_iterkeys), 0, 0, 1, 1},
  {0, __pyx_k_itervalues, sizeof(__pyx_k_itervalues), 0, 0, 1, 1},
  {0, __pyx_k_itruediv, sizeof(__pyx_k_itruediv), 0, 0, 1, 1},
  {0, __pyx_k_ixor, sizeof(__pyx_k_ixor), 0, 0, 1, 1},
  {0, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {0, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {0, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {0, __pyx_k_keys_py2, sizeof(__pyx_k_keys_py2), 0, 0, 1, 1},
  {0, __pyx_k_kw1, sizeof(__pyx_k_kw1), 0, 0, 1, 1},
  {0, __pyx_k_kw2, sizeof(__pyx_k_kw2), 0, 0, 1, 1},
  {0, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {0, __pyx_k_le, sizeof(__pyx_k_le), 0, 0, 1, 1},
  {0, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {0, __pyx_k_length_hint, sizeof(__pyx_k_length_hint), 0, 0, 1, 1},
  {0, __pyx_k_list, sizeof(__pyx_k_list), 0, 0, 1, 1},
  {0, __pyx_k_long, sizeof(__pyx_k_long), 0, 0, 1, 1},
  {0, __pyx_k_lshift, sizeof(__pyx_k_lshift), 0, 0, 1, 1},
  {0, __pyx_k_lstrip, sizeof(__pyx_k_lstrip), 0, 0, 1, 1},
  {0, __pyx_k_lt, sizeof(__pyx_k_lt), 0, 0, 1, 1},
  {0, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {0, __pyx_k_major, sizeof(__pyx_k_major), 0, 0, 1, 1},
  {0, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {0, __pyx_k_match_args, sizeof(__pyx_k_match_args), 0, 0, 1, 1},
  {0, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
  {0, __pyx_k_matmul, sizeof(__pyx_k_matmul), 0, 0, 1, 1},
  {0, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {0, __pyx_k_minor, sizeof(__pyx_k_minor), 0, 0, 1, 1},
  {0, __pyx_k_mod, sizeof(__pyx_k_mod), 0, 0, 1, 1},
  {0, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {0, __pyx_k_modules, sizeof(__pyx_k_modules), 0, 0, 1, 1},
  {0, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
  {0, __pyx_k_mul, sizeof(__pyx_k_mul), 0, 0, 1, 1},
  {0, __pyx_k_multiwrapped, sizeof(__pyx_k_multiwrapped), 0, 0, 1, 1},
  {0, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {0, __pyx_k_ne, sizeof(__pyx_k_ne), 0, 0, 1, 1},
  {0, __pyx_k_neg, sizeof(__pyx_k_neg), 0, 0, 1, 1},
  {0, __pyx_k_never_writeable, sizeof(__pyx_k_never_writeable), 0, 0, 1, 1},
  {0, __pyx_k_never_writeable_private, sizeof(__pyx_k_never_writeable_private), 0, 0, 1, 1},
  {0, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {0, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
  {0, __pyx_k_o_Invalid_type_s, sizeof(__pyx_k_o_Invalid_type_s), 0, 0, 1, 0},
  {0, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {0, __pyx_k_oldstyle_class, sizeof(__pyx_k_oldstyle_class), 0, 0, 1, 1},
  {0, __pyx_k_op, sizeof(__pyx_k_op), 0, 0, 1, 1},
  {0, __pyx_k_or, sizeof(__pyx_k_or), 0, 0, 1, 1},
  {0, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {0, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {0, __pyx_k_package, sizeof(__pyx_k_package), 0, 0, 1, 1},
  {0, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
  {0, __pyx_k_pass_to_wrapped, sizeof(__pyx_k_pass_to_wrapped), 0, 0, 1, 1},
  {0, __pyx_k_pattern, sizeof(__pyx_k_pattern), 0, 0, 1, 1},
  {0, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {0, __pyx_k_platform, sizeof(__pyx_k_platform), 0, 0, 1, 1},
  {0, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {0, __pyx_k_popitem, sizeof(__pyx_k_popitem), 0, 0, 1, 1},
  {0, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {0, __pyx_k_pow, sizeof(__pyx_k_pow), 0, 0, 1, 1},
  {0, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {0, __pyx_k_private, sizeof(__pyx_k_private), 0, 0, 1, 1},
  {0, __pyx_k_protect, sizeof(__pyx_k_protect), 0, 0, 1, 1},
  {0, __pyx_k_protected_rules_from_kwargs_loca, sizeof(__pyx_k_protected_rules_from_kwargs_loca), 0, 0, 1, 1},
  {0, __pyx_k_py2_function_attrs_rw, sizeof(__pyx_k_py2_function_attrs_rw), 0, 0, 1, 1},
  {0, __pyx_k_pydoc, sizeof(__pyx_k_pydoc), 0, 0, 1, 1},
  {0, __pyx_k_pyprotect_protected, sizeof(__pyx_k_pyprotect_protected), 0, 0, 1, 1},
  {0, __pyx_k_python_implementation, sizeof(__pyx_k_python_implementation), 0, 0, 1, 1},
  {0, __pyx_k_python_visible_pxi, sizeof(__pyx_k_python_visible_pxi), 0, 0, 1, 0},
  {0, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {0, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {0, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {0, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {0, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_Frozen, sizeof(__pyx_k_pyx_unpickle_Frozen), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_FrozenPrivacyDict, sizeof(__pyx_k_pyx_unpickle_FrozenPrivacyDict), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_FrozenPrivate, sizeof(__pyx_k_pyx_unpickle_FrozenPrivate), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_FrozenProtected, sizeof(__pyx_k_pyx_unpickle_FrozenProtected), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_PrivacyDict, sizeof(__pyx_k_pyx_unpickle_PrivacyDict), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_Private, sizeof(__pyx_k_pyx_unpickle_Private), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_Protected, sizeof(__pyx_k_pyx_unpickle_Protected), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_Proxy, sizeof(__pyx_k_pyx_unpickle_Proxy), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_Wrapped, sizeof(__pyx_k_pyx_unpickle_Wrapped), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle___HiddenPartial, sizeof(__pyx_k_pyx_unpickle___HiddenPartial), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle___ProtectionData, sizeof(__pyx_k_pyx_unpickle___ProtectionData), 0, 0, 1, 1},
  {0, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {0, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {0, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {0, __pyx_k_radd, sizeof(__pyx_k_radd), 0, 0, 1, 1},
  {0, __pyx_k_rand, sizeof(__pyx_k_rand), 0, 0, 1, 1},
  {0, __pyx_k_rdivmod, sizeof(__pyx_k_rdivmod), 0, 0, 1, 1},
  {0, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {0, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {0, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {0, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {0, __pyx_k_render_doc, sizeof(__pyx_k_render_doc), 0, 0, 1, 1},
  {0, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {0, __pyx_k_ret, sizeof(__pyx_k_ret), 0, 0, 1, 1},
  {0, __pyx_k_ret_2, sizeof(__pyx_k_ret_2), 0, 0, 1, 1},
  {0, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
  {0, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
  {0, __pyx_k_rfloordiv, sizeof(__pyx_k_rfloordiv), 0, 0, 1, 1},
  {0, __pyx_k_rl, sizeof(__pyx_k_rl), 0, 0, 1, 1},
  {0, __pyx_k_rlshift, sizeof(__pyx_k_rlshift), 0, 0, 1, 1},
  {0, __pyx_k_rmatmul, sizeof(__pyx_k_rmatmul), 0, 0, 1, 1},
  {0, __pyx_k_rmod, sizeof(__pyx_k_rmod), 0, 0, 1, 1},
  {0, __pyx_k_rmul, sizeof(__pyx_k_rmul), 0, 0, 1, 1},
  {0, __pyx_k_ro, sizeof(__pyx_k_ro), 0, 0, 1, 1},
  {0, __pyx_k_ro_data, sizeof(__pyx_k_ro_data), 0, 0, 1, 1},
  {0, __pyx_k_ro_method, sizeof(__pyx_k_ro_method), 0, 0, 1, 1},
  {0, __pyx_k_ro_regex, sizeof(__pyx_k_ro_regex), 0, 0, 1, 1},
  {0, __pyx_k_ror, sizeof(__pyx_k_ror), 0, 0, 1, 1},
  {0, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {0, __pyx_k_rpow, sizeof(__pyx_k_rpow), 0, 0, 1, 1},
  {0, __pyx_k_rrshift, sizeof(__pyx_k_rrshift), 0, 0, 1, 1},
  {0, __pyx_k_rshift, sizeof(__pyx_k_rshift), 0, 0, 1, 1},
  {0, __pyx_k_rstrip, sizeof(__pyx_k_rstrip), 0, 0, 1, 1},
  {0, __pyx_k_rsub, sizeof(__pyx_k_rsub), 0, 0, 1, 1},
  {0, __pyx_k_rtruediv, sizeof(__pyx_k_rtruediv), 0, 0, 1, 1},
  {0, __pyx_k_rules, sizeof(__pyx_k_rules), 0, 0, 1, 1},
  {0, __pyx_k_rw, sizeof(__pyx_k_rw), 0, 0, 1, 1},
  {0, __pyx_k_rw_regex, sizeof(__pyx_k_rw_regex), 0, 0, 1, 1},
  {0, __pyx_k_rxor, sizeof(__pyx_k_rxor), 0, 0, 1, 1},
  {0, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 0},
  {0, __pyx_k_s_________0_1, sizeof(__pyx_k_s_________0_1), 0, 0, 1, 0},
  {0, __pyx_k_same_class_protected, sizeof(__pyx_k_same_class_protected), 0, 0, 1, 1},
  {0, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {0, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {0, __pyx_k_set, sizeof(__pyx_k_set), 0, 0, 1, 1},
  {0, __pyx_k_set_2, sizeof(__pyx_k_set_2), 0, 0, 1, 1},
  {0, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
  {0, __pyx_k_setattr, sizeof(__pyx_k_setattr), 0, 0, 1, 1},
  {0, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
  {0, __pyx_k_setitem, sizeof(__pyx_k_setitem), 0, 0, 1, 1},
  {0, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {0, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_slots, sizeof(__pyx_k_slots), 0, 0, 1, 1},
  {0, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {0, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {0, __pyx_k_splitlines, sizeof(__pyx_k_splitlines), 0, 0, 1, 1},
  {0, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {0, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {0, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {0, __pyx_k_str_2, sizeof(__pyx_k_str_2), 0, 0, 1, 1},
  {0, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {0, __pyx_k_sub, sizeof(__pyx_k_sub), 0, 0, 1, 1},
  {0, __pyx_k_subclass, sizeof(__pyx_k_subclass), 0, 0, 1, 1},
  {0, __pyx_k_subclass_of_protected, sizeof(__pyx_k_subclass_of_protected), 0, 0, 1, 1},
  {0, __pyx_k_subclasscheck, sizeof(__pyx_k_subclasscheck), 0, 0, 1, 1},
  {0, __pyx_k_subclassof, sizeof(__pyx_k_subclassof), 0, 0, 1, 1},
  {0, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {0, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {0, __pyx_k_tb, sizeof(__pyx_k_tb), 0, 0, 1, 1},
  {0, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, __pyx_k_testop, sizeof(__pyx_k_testop), 0, 0, 1, 1},
  {0, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {0, __pyx_k_truediv, sizeof(__pyx_k_truediv), 0, 0, 1, 1},
  {0, __pyx_k_trunc, sizeof(__pyx_k_trunc), 0, 0, 1, 1},
  {0, __pyx_k_tuple, sizeof(__pyx_k_tuple), 0, 0, 1, 1},
  {0, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {0, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {0, __pyx_k_unichr, sizeof(__pyx_k_unichr), 0, 0, 1, 1},
  {0, __pyx_k_unicode, sizeof(__pyx_k_unicode), 0, 0, 1, 1},
  {0, __pyx_k_union, sizeof(__pyx_k_union), 0, 0, 1, 1},
  {0, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {0, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
  {0, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {0, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {0, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {0, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {0, __pyx_k_values_py2, sizeof(__pyx_k_values_py2), 0, 0, 1, 1},
  {0, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
  {0, __pyx_k_viewitems, sizeof(__pyx_k_viewitems), 0, 0, 1, 1},
  {0, __pyx_k_viewkeys, sizeof(__pyx_k_viewkeys), 0, 0, 1, 1},
  {0, __pyx_k_viewvalues, sizeof(__pyx_k_viewvalues), 0, 0, 1, 1},
  {0, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {0, __pyx_k_want_frozen, sizeof(__pyx_k_want_frozen), 0, 0, 1, 1},
  {0, __pyx_k_weakref, sizeof(__pyx_k_weakref), 0, 0, 1, 1},
  {0, __pyx_k_wrap, sizeof(__pyx_k_wrap), 0, 0, 1, 1},
  {0, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {0, __pyx_k_x_2, sizeof(__pyx_k_x_2), 0, 0, 1, 1},
  {0, __pyx_k_xor, sizeof(__pyx_k_xor), 0, 0, 1, 1},
  #else
  {&__pyx_n_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 1},
  {&__pyx_kp_s_0_1, __pyx_k_0_1, sizeof(__pyx_k_0_1), 0, 0, 1, 0},
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_kp_s_Cannot_add_attribute_s_s, __pyx_k_Cannot_add_attribute_s_s, sizeof(__pyx_k_Cannot_add_attribute_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_delete_attribute_s, __pyx_k_Cannot_delete_attribute_s, sizeof(__pyx_k_Cannot_delete_attribute_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_delete_attribute_s_s, __pyx_k_Cannot_delete_attribute_s_s, sizeof(__pyx_k_Cannot_delete_attribute_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_delete_private_attribute, __pyx_k_Cannot_delete_private_attribute, sizeof(__pyx_k_Cannot_delete_private_attribute), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_modify_attribute_s, __pyx_k_Cannot_modify_attribute_s, sizeof(__pyx_k_Cannot_modify_attribute_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_set_attribute_s_s, __pyx_k_Cannot_set_attribute_s_s, sizeof(__pyx_k_Cannot_set_attribute_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_set_private_attribute_s_s, __pyx_k_Cannot_set_private_attribute_s_s, sizeof(__pyx_k_Cannot_set_private_attribute_s_s), 0, 0, 1, 0},
  {&__pyx_n_s_CollectionsABC, __pyx_k_CollectionsABC, sizeof(__pyx_k_CollectionsABC), 0, 0, 1, 1},
  {&__pyx_kp_s_Double_wrapped, __pyx_k_Double_wrapped, sizeof(__pyx_k_Double_wrapped), 0, 0, 1, 0},
  {&__pyx_n_s_FrameType, __pyx_k_FrameType, sizeof(__pyx_k_FrameType), 0, 0, 1, 1},
  {&__pyx_n_s_Frozen, __pyx_k_Frozen, sizeof(__pyx_k_Frozen), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenPrivacyDict, __pyx_k_FrozenPrivacyDict, sizeof(__pyx_k_FrozenPrivacyDict), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenPrivacyDict___reduce_cytho, __pyx_k_FrozenPrivacyDict___reduce_cytho, sizeof(__pyx_k_FrozenPrivacyDict___reduce_cytho), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenPrivacyDict___setstate_cyt, __pyx_k_FrozenPrivacyDict___setstate_cyt, sizeof(__pyx_k_FrozenPrivacyDict___setstate_cyt), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenPrivate, __pyx_k_FrozenPrivate, sizeof(__pyx_k_FrozenPrivate), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenPrivate___reduce_cython, __pyx_k_FrozenPrivate___reduce_cython, sizeof(__pyx_k_FrozenPrivate___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenPrivate___setstate_cython, __pyx_k_FrozenPrivate___setstate_cython, sizeof(__pyx_k_FrozenPrivate___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenProtected, __pyx_k_FrozenProtected, sizeof(__pyx_k_FrozenProtected), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenProtected___reduce_cython, __pyx_k_FrozenProtected___reduce_cython, sizeof(__pyx_k_FrozenProtected___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_FrozenProtected___setstate_cytho, __pyx_k_FrozenProtected___setstate_cytho, sizeof(__pyx_k_FrozenProtected___setstate_cytho), 0, 0, 1, 1},
  {&__pyx_n_s_Frozen___reduce_cython, __pyx_k_Frozen___reduce_cython, sizeof(__pyx_k_Frozen___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Frozen___setstate_cython, __pyx_k_Frozen___setstate_cython, sizeof(__pyx_k_Frozen___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_HiddenPartial, __pyx_k_HiddenPartial, sizeof(__pyx_k_HiddenPartial), 0, 0, 1, 1},
  {&__pyx_n_s_HiddenPartial___bytes, __pyx_k_HiddenPartial___bytes, sizeof(__pyx_k_HiddenPartial___bytes), 0, 0, 1, 1},
  {&__pyx_n_s_HiddenPartial___dir, __pyx_k_HiddenPartial___dir, sizeof(__pyx_k_HiddenPartial___dir), 0, 0, 1, 1},
  {&__pyx_n_s_HiddenPartial___reduce_cython, __pyx_k_HiddenPartial___reduce_cython, sizeof(__pyx_k_HiddenPartial___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_HiddenPartial___setstate_cytho, __pyx_k_HiddenPartial___setstate_cytho, sizeof(__pyx_k_HiddenPartial___setstate_cytho), 0, 0, 1, 1},
  {&__pyx_kp_s_HiddenPartial_pxi, __pyx_k_HiddenPartial_pxi, sizeof(__pyx_k_HiddenPartial_pxi), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_Mapping, __pyx_k_Mapping, sizeof(__pyx_k_Mapping), 0, 0, 1, 1},
  {&__pyx_n_s_ModuleType, __pyx_k_ModuleType, sizeof(__pyx_k_ModuleType), 0, 0, 1, 1},
  {&__pyx_n_s_MutableMapping, __pyx_k_MutableMapping, sizeof(__pyx_k_MutableMapping), 0, 0, 1, 1},
  {&__pyx_n_s_MutableSequence, __pyx_k_MutableSequence, sizeof(__pyx_k_MutableSequence), 0, 0, 1, 1},
  {&__pyx_n_s_MutableSet, __pyx_k_MutableSet, sizeof(__pyx_k_MutableSet), 0, 0, 1, 1},
  {&__pyx_n_s_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_kp_s_Object_Private_s_has_no_attribut, __pyx_k_Object_Private_s_has_no_attribut, sizeof(__pyx_k_Object_Private_s_has_no_attribut), 0, 0, 1, 0},
  {&__pyx_kp_s_Object_Protected_s_has_no_attrib, __pyx_k_Object_Protected_s_has_no_attrib, sizeof(__pyx_k_Object_Protected_s_has_no_attrib), 0, 0, 1, 0},
  {&__pyx_kp_s_Object_Wrapped_s_has_no_attribut, __pyx_k_Object_Wrapped_s_has_no_attribut, sizeof(__pyx_k_Object_Wrapped_s_has_no_attribut), 0, 0, 1, 0},
  {&__pyx_kp_s_Object___HiddenPartial_has_no_at, __pyx_k_Object___HiddenPartial_has_no_at, sizeof(__pyx_k_Object___HiddenPartial_has_no_at), 0, 0, 1, 0},
  {&__pyx_kp_s_Object_is_read_only, __pyx_k_Object_is_read_only, sizeof(__pyx_k_Object_is_read_only), 0, 0, 1, 0},
  {&__pyx_kp_s_Object_s_has_no_attribute_s, __pyx_k_Object_s_has_no_attribute_s, sizeof(__pyx_k_Object_s_has_no_attribute_s), 0, 0, 1, 0},
  {&__pyx_n_s_PYPY, __pyx_k_PYPY, sizeof(__pyx_k_PYPY), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict, __pyx_k_PrivacyDict, sizeof(__pyx_k_PrivacyDict), 0, 0, 1, 1},
  {&__pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_k_PrivacyDict_FrozenPrivacyDict_px, sizeof(__pyx_k_PrivacyDict_FrozenPrivacyDict_px), 0, 0, 1, 0},
  {&__pyx_n_s_PrivacyDict___reduce_cython, __pyx_k_PrivacyDict___reduce_cython, sizeof(__pyx_k_PrivacyDict___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict___setstate_cython, __pyx_k_PrivacyDict___setstate_cython, sizeof(__pyx_k_PrivacyDict___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_copy, __pyx_k_PrivacyDict_copy, sizeof(__pyx_k_PrivacyDict_copy), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_items, __pyx_k_PrivacyDict_items, sizeof(__pyx_k_PrivacyDict_items), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_items_py2, __pyx_k_PrivacyDict_items_py2, sizeof(__pyx_k_PrivacyDict_items_py2), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_iteritems, __pyx_k_PrivacyDict_iteritems, sizeof(__pyx_k_PrivacyDict_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_iterkeys, __pyx_k_PrivacyDict_iterkeys, sizeof(__pyx_k_PrivacyDict_iterkeys), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_itervalues, __pyx_k_PrivacyDict_itervalues, sizeof(__pyx_k_PrivacyDict_itervalues), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_keys, __pyx_k_PrivacyDict_keys, sizeof(__pyx_k_PrivacyDict_keys), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_keys_py2, __pyx_k_PrivacyDict_keys_py2, sizeof(__pyx_k_PrivacyDict_keys_py2), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_values, __pyx_k_PrivacyDict_values, sizeof(__pyx_k_PrivacyDict_values), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_values_py2, __pyx_k_PrivacyDict_values_py2, sizeof(__pyx_k_PrivacyDict_values_py2), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_viewitems, __pyx_k_PrivacyDict_viewitems, sizeof(__pyx_k_PrivacyDict_viewitems), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_viewkeys, __pyx_k_PrivacyDict_viewkeys, sizeof(__pyx_k_PrivacyDict_viewkeys), 0, 0, 1, 1},
  {&__pyx_n_s_PrivacyDict_viewvalues, __pyx_k_PrivacyDict_viewvalues, sizeof(__pyx_k_PrivacyDict_viewvalues), 0, 0, 1, 1},
  {&__pyx_n_s_Private, __pyx_k_Private, sizeof(__pyx_k_Private), 0, 0, 1, 1},
  {&__pyx_kp_s_Private_FrozenPrivate_pxi, __pyx_k_Private_FrozenPrivate_pxi, sizeof(__pyx_k_Private_FrozenPrivate_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_Private___dir, __pyx_k_Private___dir, sizeof(__pyx_k_Private___dir), 0, 0, 1, 1},
  {&__pyx_n_s_Private___reduce_cython, __pyx_k_Private___reduce_cython, sizeof(__pyx_k_Private___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Private___setstate_cython, __pyx_k_Private___setstate_cython, sizeof(__pyx_k_Private___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Protected, __pyx_k_Protected, sizeof(__pyx_k_Protected), 0, 0, 1, 1},
  {&__pyx_n_s_Protected_2, __pyx_k_Protected_2, sizeof(__pyx_k_Protected_2), 0, 0, 1, 1},
  {&__pyx_kp_s_Protected_FrozenProtected_pxi, __pyx_k_Protected_FrozenProtected_pxi, sizeof(__pyx_k_Protected_FrozenProtected_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_Protected___dir, __pyx_k_Protected___dir, sizeof(__pyx_k_Protected___dir), 0, 0, 1, 1},
  {&__pyx_n_s_Protected___reduce_cython, __pyx_k_Protected___reduce_cython, sizeof(__pyx_k_Protected___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Protected___setstate_cython, __pyx_k_Protected___setstate_cython, sizeof(__pyx_k_Protected___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_ProtectionData, __pyx_k_ProtectionData, sizeof(__pyx_k_ProtectionData), 0, 0, 1, 1},
  {&__pyx_n_s_ProtectionData___dir, __pyx_k_ProtectionData___dir, sizeof(__pyx_k_ProtectionData___dir), 0, 0, 1, 1},
  {&__pyx_n_s_ProtectionData___reduce_cython, __pyx_k_ProtectionData___reduce_cython, sizeof(__pyx_k_ProtectionData___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_ProtectionData___setstate_cyth, __pyx_k_ProtectionData___setstate_cyth, sizeof(__pyx_k_ProtectionData___setstate_cyth), 0, 0, 1, 1},
  {&__pyx_kp_s_ProtectionData_pxi, __pyx_k_ProtectionData_pxi, sizeof(__pyx_k_ProtectionData_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_ProtectionError, __pyx_k_ProtectionError, sizeof(__pyx_k_ProtectionError), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy, __pyx_k_Proxy, sizeof(__pyx_k_Proxy), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___aenter, __pyx_k_Proxy___aenter, sizeof(__pyx_k_Proxy___aenter), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___aexit, __pyx_k_Proxy___aexit, sizeof(__pyx_k_Proxy___aexit), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___bytes, __pyx_k_Proxy___bytes, sizeof(__pyx_k_Proxy___bytes), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___ceil, __pyx_k_Proxy___ceil, sizeof(__pyx_k_Proxy___ceil), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___complex, __pyx_k_Proxy___complex, sizeof(__pyx_k_Proxy___complex), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___enter, __pyx_k_Proxy___enter, sizeof(__pyx_k_Proxy___enter), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___exit, __pyx_k_Proxy___exit, sizeof(__pyx_k_Proxy___exit), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___floor, __pyx_k_Proxy___floor, sizeof(__pyx_k_Proxy___floor), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___format, __pyx_k_Proxy___format, sizeof(__pyx_k_Proxy___format), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___instancecheck, __pyx_k_Proxy___instancecheck, sizeof(__pyx_k_Proxy___instancecheck), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___iter, __pyx_k_Proxy___iter, sizeof(__pyx_k_Proxy___iter), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___length_hint, __pyx_k_Proxy___length_hint, sizeof(__pyx_k_Proxy___length_hint), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___match_args, __pyx_k_Proxy___match_args, sizeof(__pyx_k_Proxy___match_args), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___reduce_cython, __pyx_k_Proxy___reduce_cython, sizeof(__pyx_k_Proxy___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___round, __pyx_k_Proxy___round, sizeof(__pyx_k_Proxy___round), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___setstate_cython, __pyx_k_Proxy___setstate_cython, sizeof(__pyx_k_Proxy___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___subclasscheck, __pyx_k_Proxy___subclasscheck, sizeof(__pyx_k_Proxy___subclasscheck), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy___trunc, __pyx_k_Proxy___trunc, sizeof(__pyx_k_Proxy___trunc), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy__aexit, __pyx_k_Proxy__aexit, sizeof(__pyx_k_Proxy__aexit), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy__exit, __pyx_k_Proxy__exit, sizeof(__pyx_k_Proxy__exit), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_add, __pyx_k_Proxy_add, sizeof(__pyx_k_Proxy_add), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_append, __pyx_k_Proxy_append, sizeof(__pyx_k_Proxy_append), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_clear, __pyx_k_Proxy_clear, sizeof(__pyx_k_Proxy_clear), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_discard, __pyx_k_Proxy_discard, sizeof(__pyx_k_Proxy_discard), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_extend, __pyx_k_Proxy_extend, sizeof(__pyx_k_Proxy_extend), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_insert, __pyx_k_Proxy_insert, sizeof(__pyx_k_Proxy_insert), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_pop, __pyx_k_Proxy_pop, sizeof(__pyx_k_Proxy_pop), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_popitem, __pyx_k_Proxy_popitem, sizeof(__pyx_k_Proxy_popitem), 0, 0, 1, 1},
  {&__pyx_kp_s_Proxy_pxi, __pyx_k_Proxy_pxi, sizeof(__pyx_k_Proxy_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_Proxy_remove, __pyx_k_Proxy_remove, sizeof(__pyx_k_Proxy_remove), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_reverse, __pyx_k_Proxy_reverse, sizeof(__pyx_k_Proxy_reverse), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_send, __pyx_k_Proxy_send, sizeof(__pyx_k_Proxy_send), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_setdefault, __pyx_k_Proxy_setdefault, sizeof(__pyx_k_Proxy_setdefault), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_sort, __pyx_k_Proxy_sort, sizeof(__pyx_k_Proxy_sort), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_throw, __pyx_k_Proxy_throw, sizeof(__pyx_k_Proxy_throw), 0, 0, 1, 1},
  {&__pyx_n_s_Proxy_update, __pyx_k_Proxy_update, sizeof(__pyx_k_Proxy_update), 0, 0, 1, 1},
  {&__pyx_n_s_PyPy, __pyx_k_PyPy, sizeof(__pyx_k_PyPy), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_object__lParenWrappe, __pyx_k_Pyx_CFunc_object__lParenWrappe, sizeof(__pyx_k_Pyx_CFunc_object__lParenWrappe), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_object__lParenWrappe_2, __pyx_k_Pyx_CFunc_object__lParenWrappe_2, sizeof(__pyx_k_Pyx_CFunc_object__lParenWrappe_2), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_object__lParenWrappe_3, __pyx_k_Pyx_CFunc_object__lParenWrappe_3, sizeof(__pyx_k_Pyx_CFunc_object__lParenWrappe_3), 0, 0, 1, 1},
  {&__pyx_kp_s_Read_only_attribute_s, __pyx_k_Read_only_attribute_s, sizeof(__pyx_k_Read_only_attribute_s), 0, 0, 1, 0},
  {&__pyx_n_s_RecursionError, __pyx_k_RecursionError, sizeof(__pyx_k_RecursionError), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_Sequence, __pyx_k_Sequence, sizeof(__pyx_k_Sequence), 0, 0, 1, 1},
  {&__pyx_n_s_Set, __pyx_k_Set, sizeof(__pyx_k_Set), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_Unknown_OldStyle_Class, __pyx_k_Unknown_OldStyle_Class, sizeof(__pyx_k_Unknown_OldStyle_Class), 0, 0, 1, 1},
  {&__pyx_n_s_Wrapped, __pyx_k_Wrapped, sizeof(__pyx_k_Wrapped), 0, 0, 1, 1},
  {&__pyx_kp_s_Wrapped_Frozen_pxi, __pyx_k_Wrapped_Frozen_pxi, sizeof(__pyx_k_Wrapped_Frozen_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_Wrapped___dir, __pyx_k_Wrapped___dir, sizeof(__pyx_k_Wrapped___dir), 0, 0, 1, 1},
  {&__pyx_n_s_Wrapped___reduce_cython, __pyx_k_Wrapped___reduce_cython, sizeof(__pyx_k_Wrapped___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Wrapped___setstate_cython, __pyx_k_Wrapped___setstate_cython, sizeof(__pyx_k_Wrapped___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Wrapped_comparator_locals_pass_t, __pyx_k_Wrapped_comparator_locals_pass_t, sizeof(__pyx_k_Wrapped_comparator_locals_pass_t), 0, 0, 1, 1},
  {&__pyx_kp_s_Wrapped_object_cannot_be_pickled, __pyx_k_Wrapped_object_cannot_be_pickled, sizeof(__pyx_k_Wrapped_object_cannot_be_pickled), 0, 0, 1, 0},
  {&__pyx_kp_s__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 0, 1, 0},
  {&__pyx_n_s__182, __pyx_k__182, sizeof(__pyx_k__182), 0, 0, 1, 1},
  {&__pyx_n_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 1},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_u__33, __pyx_k__33, sizeof(__pyx_k__33), 0, 1, 0, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_n_s__44, __pyx_k__44, sizeof(__pyx_k__44), 0, 0, 1, 1},
  {&__pyx_kp_s__88, __pyx_k__88, sizeof(__pyx_k__88), 0, 0, 1, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_kp_s_a_zA_Z_a_zA_Z0_9, __pyx_k_a_zA_Z_a_zA_Z0_9, sizeof(__pyx_k_a_zA_Z_a_zA_Z0_9), 0, 0, 1, 0},
  {&__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2, __pyx_k_a_zA_Z_a_zA_Z0_9_2, sizeof(__pyx_k_a_zA_Z_a_zA_Z0_9_2), 0, 0, 1, 0},
  {&__pyx_n_s_abc, __pyx_k_abc, sizeof(__pyx_k_abc), 0, 0, 1, 1},
  {&__pyx_n_s_abs, __pyx_k_abs, sizeof(__pyx_k_abs), 0, 0, 1, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_add_2, __pyx_k_add_2, sizeof(__pyx_k_add_2), 0, 0, 1, 1},
  {&__pyx_n_s_aenter, __pyx_k_aenter, sizeof(__pyx_k_aenter), 0, 0, 1, 1},
  {&__pyx_n_s_aexit, __pyx_k_aexit, sizeof(__pyx_k_aexit), 0, 0, 1, 1},
  {&__pyx_n_s_alist, __pyx_k_alist, sizeof(__pyx_k_alist), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_always_delegated_attributes, __pyx_k_always_delegated_attributes, sizeof(__pyx_k_always_delegated_attributes), 0, 0, 1, 1},
  {&__pyx_n_s_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
  {&__pyx_n_s_attr_type_check, __pyx_k_attr_type_check, sizeof(__pyx_k_attr_type_check), 0, 0, 1, 1},
  {&__pyx_n_s_attribute_protected, __pyx_k_attribute_protected, sizeof(__pyx_k_attribute_protected), 0, 0, 1, 1},
  {&__pyx_n_s_basestring, __pyx_k_basestring, sizeof(__pyx_k_basestring), 0, 0, 1, 1},
  {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
  {&__pyx_n_s_bool_2, __pyx_k_bool_2, sizeof(__pyx_k_bool_2), 0, 0, 1, 1},
  {&__pyx_n_s_build_regex, __pyx_k_build_regex, sizeof(__pyx_k_build_regex), 0, 0, 1, 1},
  {&__pyx_n_s_builtin, __pyx_k_builtin, sizeof(__pyx_k_builtin), 0, 0, 1, 1},
  {&__pyx_n_s_builtins, __pyx_k_builtins, sizeof(__pyx_k_builtins), 0, 0, 1, 1},
  {&__pyx_n_s_bytearray, __pyx_k_bytearray, sizeof(__pyx_k_bytearray), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_bytes_2, __pyx_k_bytes_2, sizeof(__pyx_k_bytes_2), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
  {&__pyx_n_s_ceil, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 0, 1, 1},
  {&__pyx_n_s_ceil_2, __pyx_k_ceil_2, sizeof(__pyx_k_ceil_2), 0, 0, 1, 1},
  {&__pyx_n_s_cfunc_to_py, __pyx_k_cfunc_to_py, sizeof(__pyx_k_cfunc_to_py), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
  {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cmp, __pyx_k_cmp, sizeof(__pyx_k_cmp), 0, 0, 1, 1},
  {&__pyx_n_s_cn, __pyx_k_cn, sizeof(__pyx_k_cn), 0, 0, 1, 1},
  {&__pyx_n_s_code, __pyx_k_code, sizeof(__pyx_k_code), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_collections_abc, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 1},
  {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {&__pyx_n_s_complex, __pyx_k_complex, sizeof(__pyx_k_complex), 0, 0, 1, 1},
  {&__pyx_n_s_complex_2, __pyx_k_complex_2, sizeof(__pyx_k_complex_2), 0, 0, 1, 1},
  {&__pyx_n_s_contains, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {&__pyx_n_s_contains_2, __pyx_k_contains_2, sizeof(__pyx_k_contains_2), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_delattr, __pyx_k_delattr, sizeof(__pyx_k_delattr), 0, 0, 1, 1},
  {&__pyx_n_s_delete, __pyx_k_delete, sizeof(__pyx_k_delete), 0, 0, 1, 1},
  {&__pyx_n_s_delitem, __pyx_k_delitem, sizeof(__pyx_k_delitem), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
  {&__pyx_n_s_dict_3, __pyx_k_dict_3, sizeof(__pyx_k_dict_3), 0, 0, 1, 1},
  {&__pyx_n_s_difference, __pyx_k_difference, sizeof(__pyx_k_difference), 0, 0, 1, 1},
  {&__pyx_n_s_dir, __pyx_k_dir, sizeof(__pyx_k_dir), 0, 0, 1, 1},
  {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
  {&__pyx_n_s_discard, __pyx_k_discard, sizeof(__pyx_k_discard), 0, 0, 1, 1},
  {&__pyx_n_s_divmod, __pyx_k_divmod, sizeof(__pyx_k_divmod), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dynamic, __pyx_k_dynamic, sizeof(__pyx_k_dynamic), 0, 0, 1, 1},
  {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
  {&__pyx_n_s_endswith, __pyx_k_endswith, sizeof(__pyx_k_endswith), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_environ, __pyx_k_environ, sizeof(__pyx_k_environ), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_exc_type, __pyx_k_exc_type, sizeof(__pyx_k_exc_type), 0, 0, 1, 1},
  {&__pyx_n_s_exc_value, __pyx_k_exc_value, sizeof(__pyx_k_exc_value), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_float_2, __pyx_k_float_2, sizeof(__pyx_k_float_2), 0, 0, 1, 1},
  {&__pyx_n_s_floor, __pyx_k_floor, sizeof(__pyx_k_floor), 0, 0, 1, 1},
  {&__pyx_n_s_floor_2, __pyx_k_floor_2, sizeof(__pyx_k_floor_2), 0, 0, 1, 1},
  {&__pyx_n_s_floordiv, __pyx_k_floordiv, sizeof(__pyx_k_floordiv), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_freeze, __pyx_k_freeze, sizeof(__pyx_k_freeze), 0, 0, 1, 1},
  {&__pyx_n_s_frozen, __pyx_k_frozen, sizeof(__pyx_k_frozen), 0, 0, 1, 1},
  {&__pyx_n_s_frozenset, __pyx_k_frozenset, sizeof(__pyx_k_frozenset), 0, 0, 1, 1},
  {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
  {&__pyx_n_s_ge, __pyx_k_ge, sizeof(__pyx_k_ge), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getattribute, __pyx_k_getattribute, sizeof(__pyx_k_getattribute), 0, 0, 1, 1},
  {&__pyx_n_s_getitem, __pyx_k_getitem, sizeof(__pyx_k_getitem), 0, 0, 1, 1},
  {&__pyx_n_s_getsate, __pyx_k_getsate, sizeof(__pyx_k_getsate), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_global_c_functions_pxi, __pyx_k_global_c_functions_pxi, sizeof(__pyx_k_global_c_functions_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_gt, __pyx_k_gt, sizeof(__pyx_k_gt), 0, 0, 1, 1},
  {&__pyx_n_s_hash, __pyx_k_hash, sizeof(__pyx_k_hash), 0, 0, 1, 1},
  {&__pyx_n_s_hash_2, __pyx_k_hash_2, sizeof(__pyx_k_hash_2), 0, 0, 1, 1},
  {&__pyx_n_s_hash_protected, __pyx_k_hash_protected, sizeof(__pyx_k_hash_protected), 0, 0, 1, 1},
  {&__pyx_n_s_hash_val, __pyx_k_hash_val, sizeof(__pyx_k_hash_val), 0, 0, 1, 1},
  {&__pyx_n_s_help, __pyx_k_help, sizeof(__pyx_k_help), 0, 0, 1, 1},
  {&__pyx_n_s_help_protected, __pyx_k_help_protected, sizeof(__pyx_k_help_protected), 0, 0, 1, 1},
  {&__pyx_n_s_help_str, __pyx_k_help_str, sizeof(__pyx_k_help_str), 0, 0, 1, 1},
  {&__pyx_n_s_help_val, __pyx_k_help_val, sizeof(__pyx_k_help_val), 0, 0, 1, 1},
  {&__pyx_n_s_hidden_pickle_attributes, __pyx_k_hidden_pickle_attributes, sizeof(__pyx_k_hidden_pickle_attributes), 0, 0, 1, 1},
  {&__pyx_n_s_hide, __pyx_k_hide, sizeof(__pyx_k_hide), 0, 0, 1, 1},
  {&__pyx_n_s_hide_private, __pyx_k_hide_private, sizeof(__pyx_k_hide_private), 0, 0, 1, 1},
  {&__pyx_n_s_hide_regex, __pyx_k_hide_regex, sizeof(__pyx_k_hide_regex), 0, 0, 1, 1},
  {&__pyx_n_s_iadd, __pyx_k_iadd, sizeof(__pyx_k_iadd), 0, 0, 1, 1},
  {&__pyx_n_s_iand, __pyx_k_iand, sizeof(__pyx_k_iand), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_id_class, __pyx_k_id_class, sizeof(__pyx_k_id_class), 0, 0, 1, 1},
  {&__pyx_n_s_id_protected, __pyx_k_id_protected, sizeof(__pyx_k_id_protected), 0, 0, 1, 1},
  {&__pyx_n_s_id_val, __pyx_k_id_val, sizeof(__pyx_k_id_val), 0, 0, 1, 1},
  {&__pyx_n_s_ifloordiv, __pyx_k_ifloordiv, sizeof(__pyx_k_ifloordiv), 0, 0, 1, 1},
  {&__pyx_n_s_ilshift, __pyx_k_ilshift, sizeof(__pyx_k_ilshift), 0, 0, 1, 1},
  {&__pyx_n_s_imatmul, __pyx_k_imatmul, sizeof(__pyx_k_imatmul), 0, 0, 1, 1},
  {&__pyx_n_s_immutable_builtin_attributes, __pyx_k_immutable_builtin_attributes, sizeof(__pyx_k_immutable_builtin_attributes), 0, 0, 1, 1},
  {&__pyx_n_s_imod, __pyx_k_imod, sizeof(__pyx_k_imod), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_imul, __pyx_k_imul, sizeof(__pyx_k_imul), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_init___locals_C, __pyx_k_init___locals_C, sizeof(__pyx_k_init___locals_C), 0, 0, 1, 1},
  {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
  {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_inst, __pyx_k_inst, sizeof(__pyx_k_inst), 0, 0, 1, 1},
  {&__pyx_n_s_instance_of_protected, __pyx_k_instance_of_protected, sizeof(__pyx_k_instance_of_protected), 0, 0, 1, 1},
  {&__pyx_n_s_instancecheck, __pyx_k_instancecheck, sizeof(__pyx_k_instancecheck), 0, 0, 1, 1},
  {&__pyx_n_s_instanceof, __pyx_k_instanceof, sizeof(__pyx_k_instanceof), 0, 0, 1, 1},
  {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
  {&__pyx_n_s_int_2, __pyx_k_int_2, sizeof(__pyx_k_int_2), 0, 0, 1, 1},
  {&__pyx_n_s_intersection, __pyx_k_intersection, sizeof(__pyx_k_intersection), 0, 0, 1, 1},
  {&__pyx_n_s_invert, __pyx_k_invert, sizeof(__pyx_k_invert), 0, 0, 1, 1},
  {&__pyx_n_s_ior, __pyx_k_ior, sizeof(__pyx_k_ior), 0, 0, 1, 1},
  {&__pyx_n_s_ipow, __pyx_k_ipow, sizeof(__pyx_k_ipow), 0, 0, 1, 1},
  {&__pyx_n_s_irshift, __pyx_k_irshift, sizeof(__pyx_k_irshift), 0, 0, 1, 1},
  {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
  {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
  {&__pyx_n_s_isfrozen, __pyx_k_isfrozen, sizeof(__pyx_k_isfrozen), 0, 0, 1, 1},
  {&__pyx_n_s_isimmutable, __pyx_k_isimmutable, sizeof(__pyx_k_isimmutable), 0, 0, 1, 1},
  {&__pyx_n_s_isinstance, __pyx_k_isinstance, sizeof(__pyx_k_isinstance), 0, 0, 1, 1},
  {&__pyx_n_s_isinstance_protected, __pyx_k_isinstance_protected, sizeof(__pyx_k_isinstance_protected), 0, 0, 1, 1},
  {&__pyx_n_s_isinstance_val, __pyx_k_isinstance_val, sizeof(__pyx_k_isinstance_val), 0, 0, 1, 1},
  {&__pyx_n_s_isprivate, __pyx_k_isprivate, sizeof(__pyx_k_isprivate), 0, 0, 1, 1},
  {&__pyx_n_s_isprotected, __pyx_k_isprotected, sizeof(__pyx_k_isprotected), 0, 0, 1, 1},
  {&__pyx_n_s_isreadonly, __pyx_k_isreadonly, sizeof(__pyx_k_isreadonly), 0, 0, 1, 1},
  {&__pyx_n_s_issubclass, __pyx_k_issubclass, sizeof(__pyx_k_issubclass), 0, 0, 1, 1},
  {&__pyx_n_s_issubclass_protected, __pyx_k_issubclass_protected, sizeof(__pyx_k_issubclass_protected), 0, 0, 1, 1},
  {&__pyx_n_s_issubclass_val, __pyx_k_issubclass_val, sizeof(__pyx_k_issubclass_val), 0, 0, 1, 1},
  {&__pyx_n_s_isub, __pyx_k_isub, sizeof(__pyx_k_isub), 0, 0, 1, 1},
  {&__pyx_n_s_isvisible, __pyx_k_isvisible, sizeof(__pyx_k_isvisible), 0, 0, 1, 1},
  {&__pyx_n_s_iswrapped, __pyx_k_iswrapped, sizeof(__pyx_k_iswrapped), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_items_py2, __pyx_k_items_py2, sizeof(__pyx_k_items_py2), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_iterkeys, __pyx_k_iterkeys, sizeof(__pyx_k_iterkeys), 0, 0, 1, 1},
  {&__pyx_n_s_itervalues, __pyx_k_itervalues, sizeof(__pyx_k_itervalues), 0, 0, 1, 1},
  {&__pyx_n_s_itruediv, __pyx_k_itruediv, sizeof(__pyx_k_itruediv), 0, 0, 1, 1},
  {&__pyx_n_s_ixor, __pyx_k_ixor, sizeof(__pyx_k_ixor), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_keys_py2, __pyx_k_keys_py2, sizeof(__pyx_k_keys_py2), 0, 0, 1, 1},
  {&__pyx_n_s_kw1, __pyx_k_kw1, sizeof(__pyx_k_kw1), 0, 0, 1, 1},
  {&__pyx_n_s_kw2, __pyx_k_kw2, sizeof(__pyx_k_kw2), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_le, __pyx_k_le, sizeof(__pyx_k_le), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_length_hint, __pyx_k_length_hint, sizeof(__pyx_k_length_hint), 0, 0, 1, 1},
  {&__pyx_n_s_list, __pyx_k_list, sizeof(__pyx_k_list), 0, 0, 1, 1},
  {&__pyx_n_s_long, __pyx_k_long, sizeof(__pyx_k_long), 0, 0, 1, 1},
  {&__pyx_n_s_lshift, __pyx_k_lshift, sizeof(__pyx_k_lshift), 0, 0, 1, 1},
  {&__pyx_n_s_lstrip, __pyx_k_lstrip, sizeof(__pyx_k_lstrip), 0, 0, 1, 1},
  {&__pyx_n_s_lt, __pyx_k_lt, sizeof(__pyx_k_lt), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_major, __pyx_k_major, sizeof(__pyx_k_major), 0, 0, 1, 1},
  {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {&__pyx_n_s_match_args, __pyx_k_match_args, sizeof(__pyx_k_match_args), 0, 0, 1, 1},
  {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
  {&__pyx_n_s_matmul, __pyx_k_matmul, sizeof(__pyx_k_matmul), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_minor, __pyx_k_minor, sizeof(__pyx_k_minor), 0, 0, 1, 1},
  {&__pyx_n_s_mod, __pyx_k_mod, sizeof(__pyx_k_mod), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_modules, __pyx_k_modules, sizeof(__pyx_k_modules), 0, 0, 1, 1},
  {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
  {&__pyx_n_s_mul, __pyx_k_mul, sizeof(__pyx_k_mul), 0, 0, 1, 1},
  {&__pyx_n_s_multiwrapped, __pyx_k_multiwrapped, sizeof(__pyx_k_multiwrapped), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_ne, __pyx_k_ne, sizeof(__pyx_k_ne), 0, 0, 1, 1},
  {&__pyx_n_s_neg, __pyx_k_neg, sizeof(__pyx_k_neg), 0, 0, 1, 1},
  {&__pyx_n_s_never_writeable, __pyx_k_never_writeable, sizeof(__pyx_k_never_writeable), 0, 0, 1, 1},
  {&__pyx_n_s_never_writeable_private, __pyx_k_never_writeable_private, sizeof(__pyx_k_never_writeable_private), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
  {&__pyx_kp_s_o_Invalid_type_s, __pyx_k_o_Invalid_type_s, sizeof(__pyx_k_o_Invalid_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_oldstyle_class, __pyx_k_oldstyle_class, sizeof(__pyx_k_oldstyle_class), 0, 0, 1, 1},
  {&__pyx_n_s_op, __pyx_k_op, sizeof(__pyx_k_op), 0, 0, 1, 1},
  {&__pyx_n_s_or, __pyx_k_or, sizeof(__pyx_k_or), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_package, __pyx_k_package, sizeof(__pyx_k_package), 0, 0, 1, 1},
  {&__pyx_n_s_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
  {&__pyx_n_s_pass_to_wrapped, __pyx_k_pass_to_wrapped, sizeof(__pyx_k_pass_to_wrapped), 0, 0, 1, 1},
  {&__pyx_n_s_pattern, __pyx_k_pattern, sizeof(__pyx_k_pattern), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_platform, __pyx_k_platform, sizeof(__pyx_k_platform), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_popitem, __pyx_k_popitem, sizeof(__pyx_k_popitem), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_pow, __pyx_k_pow, sizeof(__pyx_k_pow), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_private, __pyx_k_private, sizeof(__pyx_k_private), 0, 0, 1, 1},
  {&__pyx_n_s_protect, __pyx_k_protect, sizeof(__pyx_k_protect), 0, 0, 1, 1},
  {&__pyx_n_s_protected_rules_from_kwargs_loca, __pyx_k_protected_rules_from_kwargs_loca, sizeof(__pyx_k_protected_rules_from_kwargs_loca), 0, 0, 1, 1},
  {&__pyx_n_s_py2_function_attrs_rw, __pyx_k_py2_function_attrs_rw, sizeof(__pyx_k_py2_function_attrs_rw), 0, 0, 1, 1},
  {&__pyx_n_s_pydoc, __pyx_k_pydoc, sizeof(__pyx_k_pydoc), 0, 0, 1, 1},
  {&__pyx_n_s_pyprotect_protected, __pyx_k_pyprotect_protected, sizeof(__pyx_k_pyprotect_protected), 0, 0, 1, 1},
  {&__pyx_n_s_python_implementation, __pyx_k_python_implementation, sizeof(__pyx_k_python_implementation), 0, 0, 1, 1},
  {&__pyx_kp_s_python_visible_pxi, __pyx_k_python_visible_pxi, sizeof(__pyx_k_python_visible_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Frozen, __pyx_k_pyx_unpickle_Frozen, sizeof(__pyx_k_pyx_unpickle_Frozen), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_FrozenPrivacyDict, __pyx_k_pyx_unpickle_FrozenPrivacyDict, sizeof(__pyx_k_pyx_unpickle_FrozenPrivacyDict), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_FrozenPrivate, __pyx_k_pyx_unpickle_FrozenPrivate, sizeof(__pyx_k_pyx_unpickle_FrozenPrivate), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_FrozenProtected, __pyx_k_pyx_unpickle_FrozenProtected, sizeof(__pyx_k_pyx_unpickle_FrozenProtected), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_PrivacyDict, __pyx_k_pyx_unpickle_PrivacyDict, sizeof(__pyx_k_pyx_unpickle_PrivacyDict), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Private, __pyx_k_pyx_unpickle_Private, sizeof(__pyx_k_pyx_unpickle_Private), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Protected, __pyx_k_pyx_unpickle_Protected, sizeof(__pyx_k_pyx_unpickle_Protected), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Proxy, __pyx_k_pyx_unpickle_Proxy, sizeof(__pyx_k_pyx_unpickle_Proxy), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Wrapped, __pyx_k_pyx_unpickle_Wrapped, sizeof(__pyx_k_pyx_unpickle_Wrapped), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle___HiddenPartial, __pyx_k_pyx_unpickle___HiddenPartial, sizeof(__pyx_k_pyx_unpickle___HiddenPartial), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle___ProtectionData, __pyx_k_pyx_unpickle___ProtectionData, sizeof(__pyx_k_pyx_unpickle___ProtectionData), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_s_radd, __pyx_k_radd, sizeof(__pyx_k_radd), 0, 0, 1, 1},
  {&__pyx_n_s_rand, __pyx_k_rand, sizeof(__pyx_k_rand), 0, 0, 1, 1},
  {&__pyx_n_s_rdivmod, __pyx_k_rdivmod, sizeof(__pyx_k_rdivmod), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_render_doc, __pyx_k_render_doc, sizeof(__pyx_k_render_doc), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_ret, __pyx_k_ret, sizeof(__pyx_k_ret), 0, 0, 1, 1},
  {&__pyx_n_s_ret_2, __pyx_k_ret_2, sizeof(__pyx_k_ret_2), 0, 0, 1, 1},
  {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
  {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
  {&__pyx_n_s_rfloordiv, __pyx_k_rfloordiv, sizeof(__pyx_k_rfloordiv), 0, 0, 1, 1},
  {&__pyx_n_s_rl, __pyx_k_rl, sizeof(__pyx_k_rl), 0, 0, 1, 1},
  {&__pyx_n_s_rlshift, __pyx_k_rlshift, sizeof(__pyx_k_rlshift), 0, 0, 1, 1},
  {&__pyx_n_s_rmatmul, __pyx_k_rmatmul, sizeof(__pyx_k_rmatmul), 0, 0, 1, 1},
  {&__pyx_n_s_rmod, __pyx_k_rmod, sizeof(__pyx_k_rmod), 0, 0, 1, 1},
  {&__pyx_n_s_rmul, __pyx_k_rmul, sizeof(__pyx_k_rmul), 0, 0, 1, 1},
  {&__pyx_n_s_ro, __pyx_k_ro, sizeof(__pyx_k_ro), 0, 0, 1, 1},
  {&__pyx_n_s_ro_data, __pyx_k_ro_data, sizeof(__pyx_k_ro_data), 0, 0, 1, 1},
  {&__pyx_n_s_ro_method, __pyx_k_ro_method, sizeof(__pyx_k_ro_method), 0, 0, 1, 1},
  {&__pyx_n_s_ro_regex, __pyx_k_ro_regex, sizeof(__pyx_k_ro_regex), 0, 0, 1, 1},
  {&__pyx_n_s_ror, __pyx_k_ror, sizeof(__pyx_k_ror), 0, 0, 1, 1},
  {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {&__pyx_n_s_rpow, __pyx_k_rpow, sizeof(__pyx_k_rpow), 0, 0, 1, 1},
  {&__pyx_n_s_rrshift, __pyx_k_rrshift, sizeof(__pyx_k_rrshift), 0, 0, 1, 1},
  {&__pyx_n_s_rshift, __pyx_k_rshift, sizeof(__pyx_k_rshift), 0, 0, 1, 1},
  {&__pyx_n_s_rstrip, __pyx_k_rstrip, sizeof(__pyx_k_rstrip), 0, 0, 1, 1},
  {&__pyx_n_s_rsub, __pyx_k_rsub, sizeof(__pyx_k_rsub), 0, 0, 1, 1},
  {&__pyx_n_s_rtruediv, __pyx_k_rtruediv, sizeof(__pyx_k_rtruediv), 0, 0, 1, 1},
  {&__pyx_n_s_rules, __pyx_k_rules, sizeof(__pyx_k_rules), 0, 0, 1, 1},
  {&__pyx_n_s_rw, __pyx_k_rw, sizeof(__pyx_k_rw), 0, 0, 1, 1},
  {&__pyx_n_s_rw_regex, __pyx_k_rw_regex, sizeof(__pyx_k_rw_regex), 0, 0, 1, 1},
  {&__pyx_n_s_rxor, __pyx_k_rxor, sizeof(__pyx_k_rxor), 0, 0, 1, 1},
  {&__pyx_kp_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_________0_1, __pyx_k_s_________0_1, sizeof(__pyx_k_s_________0_1), 0, 0, 1, 0},
  {&__pyx_n_s_same_class_protected, __pyx_k_same_class_protected, sizeof(__pyx_k_same_class_protected), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_set, __pyx_k_set, sizeof(__pyx_k_set), 0, 0, 1, 1},
  {&__pyx_n_s_set_2, __pyx_k_set_2, sizeof(__pyx_k_set_2), 0, 0, 1, 1},
  {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
  {&__pyx_n_s_setattr, __pyx_k_setattr, sizeof(__pyx_k_setattr), 0, 0, 1, 1},
  {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
  {&__pyx_n_s_setitem, __pyx_k_setitem, sizeof(__pyx_k_setitem), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_slots, __pyx_k_slots, sizeof(__pyx_k_slots), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {&__pyx_n_s_splitlines, __pyx_k_splitlines, sizeof(__pyx_k_splitlines), 0, 0, 1, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_str_2, __pyx_k_str_2, sizeof(__pyx_k_str_2), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_sub, __pyx_k_sub, sizeof(__pyx_k_sub), 0, 0, 1, 1},
  {&__pyx_n_s_subclass, __pyx_k_subclass, sizeof(__pyx_k_subclass), 0, 0, 1, 1},
  {&__pyx_n_s_subclass_of_protected, __pyx_k_subclass_of_protected, sizeof(__pyx_k_subclass_of_protected), 0, 0, 1, 1},
  {&__pyx_n_s_subclasscheck, __pyx_k_subclasscheck, sizeof(__pyx_k_subclasscheck), 0, 0, 1, 1},
  {&__pyx_n_s_subclassof, __pyx_k_subclassof, sizeof(__pyx_k_subclassof), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_tb, __pyx_k_tb, sizeof(__pyx_k_tb), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_testop, __pyx_k_testop, sizeof(__pyx_k_testop), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_truediv, __pyx_k_truediv, sizeof(__pyx_k_truediv), 0, 0, 1, 1},
  {&__pyx_n_s_trunc, __pyx_k_trunc, sizeof(__pyx_k_trunc), 0, 0, 1, 1},
  {&__pyx_n_s_tuple, __pyx_k_tuple, sizeof(__pyx_k_tuple), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {&__pyx_n_s_unichr, __pyx_k_unichr, sizeof(__pyx_k_unichr), 0, 0, 1, 1},
  {&__pyx_n_s_unicode, __pyx_k_unicode, sizeof(__pyx_k_unicode), 0, 0, 1, 1},
  {&__pyx_n_s_union, __pyx_k_union, sizeof(__pyx_k_union), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_values_py2, __pyx_k_values_py2, sizeof(__pyx_k_values_py2), 0, 0, 1, 1},
  {&__pyx_n_s_version_info, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
  {&__pyx_n_s_viewitems, __pyx_k_viewitems, sizeof(__pyx_k_viewitems), 0, 0, 1, 1},
  {&__pyx_n_s_viewkeys, __pyx_k_viewkeys, sizeof(__pyx_k_viewkeys), 0, 0, 1, 1},
  {&__pyx_n_s_viewvalues, __pyx_k_viewvalues, sizeof(__pyx_k_viewvalues), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_want_frozen, __pyx_k_want_frozen, sizeof(__pyx_k_want_frozen), 0, 0, 1, 1},
  {&__pyx_n_s_weakref, __pyx_k_weakref, sizeof(__pyx_k_weakref), 0, 0, 1, 1},
  {&__pyx_n_s_wrap, __pyx_k_wrap, sizeof(__pyx_k_wrap), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_x_2, __pyx_k_x_2, sizeof(__pyx_k_x_2), 0, 0, 1, 1},
  {&__pyx_n_s_xor, __pyx_k_xor, sizeof(__pyx_k_xor), 0, 0, 1, 1},
  #endif
  {0, 0, 0, 0, 0, 0, 0}
};
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(0, 19, __pyx_L1_error)
  __pyx_builtin_help = __Pyx_GetBuiltinName(__pyx_n_s_help); if (!__pyx_builtin_help) __PYX_ERR(0, 97, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 372, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(1, 99, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(2, 42, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(3, 307, __pyx_L1_error)
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(4, 20, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "global_c_functions.pxi":159
 *     Called once by protect() before Protected class initialization
 *     '''
 *     def _build_regex(alist):             # <<<<<<<<<<<<<<
 *         _ret = ''
 *         _rl = []
 */
  __pyx_tuple__5 = PyTuple_Pack(4, __pyx_n_s_alist, __pyx_n_s_ret, __pyx_n_s_rl, __pyx_n_s_x_2); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(5, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __pyx_codeobj__6 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__5, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_global_c_functions_pxi, __pyx_n_s_build_regex, 159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__6)) __PYX_ERR(5, 159, __pyx_L1_error)

  /* "global_c_functions.pxi":176
 *         return re.compile(_ret)
 * 
 *     ro_method = bool(kwargs.get('ro_method', False))             # <<<<<<<<<<<<<<
 *     ro_data = bool(kwargs.get('ro_data', False))
 *     hide_private = kwargs.get('hide_private', False)
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_n_s_ro_method, Py_False); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(5, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "global_c_functions.pxi":177
 * 
 *     ro_method = bool(kwargs.get('ro_method', False))
 *     ro_data = bool(kwargs.get('ro_data', False))             # <<<<<<<<<<<<<<
 *     hide_private = kwargs.get('hide_private', False)
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_n_s_ro_data, Py_False); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(5, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "global_c_functions.pxi":178
 *     ro_method = bool(kwargs.get('ro_method', False))
 *     ro_data = bool(kwargs.get('ro_data', False))
 *     hide_private = kwargs.get('hide_private', False)             # <<<<<<<<<<<<<<
 * 
 *     ro = [
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_n_s_hide_private, Py_False); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(5, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "global_c_functions.pxi":209
 *         'ro_data': bool(ro_data),
 *     }
 *     d['dynamic'] = kwargs.get('dynamic', False)             # <<<<<<<<<<<<<<
 *     d['frozen'] = bool(kwargs.get('frozen', False))
 *     d['kwargs'] = kwargs
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_n_s_dynamic, Py_False); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(5, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "global_c_functions.pxi":210
 *     }
 *     d['dynamic'] = kwargs.get('dynamic', False)
 *     d['frozen'] = bool(kwargs.get('frozen', False))             # <<<<<<<<<<<<<<
 *     d['kwargs'] = kwargs
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_n_s_frozen, Py_False); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(5, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "global_c_functions.pxi":214
 * 
 *     d['attr_type_check'] = False
 *     for kw in ('ro_method', 'ro_data'):             # <<<<<<<<<<<<<<
 *         if bool(d.get(kw, False)):
 *             d['attr_type_check'] = True
 */
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_n_s_ro_method, __pyx_n_s_ro_data); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(5, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "global_c_functions.pxi":236
 *     # Restrictive bool options must be 'or-ed'
 *     for a in (
 *         'frozen', 'hide_private', 'ro_data', 'ro_method',             # <<<<<<<<<<<<<<
 *     ):
 *         d[a] = (kw1.get(a, False) or kw2.get(a, False))
 */
  __pyx_tuple__13 = PyTuple_Pack(4, __pyx_n_s_frozen, __pyx_n_s_hide_private, __pyx_n_s_ro_data, __pyx_n_s_ro_method); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(5, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "global_c_functions.pxi":242
 *     # Restrictive lists (non-bool) are unioned
 *     for a in (
 *         'ro', 'hide',             # <<<<<<<<<<<<<<
 *     ):
 *         s1 = set(list(kw1.get(a, [])))
 */
  __pyx_tuple__14 = PyTuple_Pack(2, __pyx_n_s_ro, __pyx_n_s_hide); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(5, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "global_c_functions.pxi":251
 *     # Permissive lists (non-bool) are intersected
 *     for a in (
 *         'rw',             # <<<<<<<<<<<<<<
 *     ):
 *         s1 = set(list(kw1.get(a, [])))
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_n_s_rw); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(5, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "Proxy.pxi":42
 *             # We claim to be avoiding double-wrapping, so this code
 *             # should never be reached
 *             raise RuntimeError('Double-wrapped!')             # <<<<<<<<<<<<<<
 * 
 *         x = self.pvt_o(*args, **kwargs)
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Double_wrapped); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "Wrapped_Frozen.pxi":192
 *     cdef help_str_protected(self):
 *         return '\n'.join(
 *             pydoc.render_doc(self.pvt_o).splitlines()[2:]             # <<<<<<<<<<<<<<
 *         ).rstrip('\n') + '\n'
 * 
 */
  __pyx_slice__18 = PySlice_New(__pyx_int_2, Py_None, Py_None); if (unlikely(!__pyx_slice__18)) __PYX_ERR(3, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__18);
  __Pyx_GIVEREF(__pyx_slice__18);

  /* "Wrapped_Frozen.pxi":249
 * 
 *         '''
 *         def pass_to_wrapped():             # <<<<<<<<<<<<<<
 *             '''Trap RecursionError if object is too deeply nested'''
 *             if op == Py_LT:
 */
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Wrapped_Frozen_pxi, __pyx_n_s_pass_to_wrapped, 249, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(3, 249, __pyx_L1_error)

  /* "Wrapped_Frozen.pxi":328
 *         # PREVENT pickling - doesn't work even if methods are implemented,
 *         if a in pickle_attributes:
 *             raise AttributeError('Wrapped object cannot be pickled')             # <<<<<<<<<<<<<<
 * 
 *         delegated = getattr(self.pvt_o, a, None)
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_Wrapped_object_cannot_be_pickled); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(3, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":20
 *         # not instance of dict or CollectionsABC.Mapping
 *         if PY2:
 *             class C(object):             # <<<<<<<<<<<<<<
 *                 pass
 * 
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "Private_FrozenPrivate.pxi":40
 *             isinstance(self.pvt_o, types.ModuleType) and
 *             (
 *                 PY2 or (sys.version_info.major, sys.version_info.minor) < (3, 7)             # <<<<<<<<<<<<<<
 *             )
 *         ):
 */
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_int_3, __pyx_int_7); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "Protected_FrozenProtected.pxi":48
 *             self.dir_out = [
 *                     k for (k, v) in self.acl_cache.items()
 *                     if v.get('r', False)             # <<<<<<<<<<<<<<
 *             ]
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(2, __pyx_n_s_r, Py_False); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(8, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "Protected_FrozenProtected.pxi":86
 * 
 *         d = self.rules
 *         attr_type_check = bool(d.get('attr_type_check', False))             # <<<<<<<<<<<<<<
 *         ro_method = bool(d.get('ro_method', False))
 *         ro_data = bool(d.get('ro_data', False))
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_attr_type_check, Py_False); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfa53bdd, 0xec20346, 0x12e646a):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xfa53bdd, 0xec20346, 0x12e646a) = (attributes_map, freeze, hash, help, help_str, id, id_class, instanceof, isinstance, issubclass, multiwrapped, private, protect, rules, subclassof, testop))" % __pyx_checksum
 */
  __pyx_tuple__32 = PyTuple_Pack(3, __pyx_int_262487005, __pyx_int_247595846, __pyx_int_19817578); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_tuple__34 = PyTuple_Pack(3, __pyx_int_152356376, __pyx_int_161740782, __pyx_int_50167005); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_tuple__35 = PyTuple_Pack(3, __pyx_int_209109265, __pyx_int_115090883, __pyx_int_6017409); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_tuple__36 = PyTuple_Pack(3, __pyx_int_218598457, __pyx_int_180490543, __pyx_int_78131249); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_tuple__37 = PyTuple_Pack(3, __pyx_int_155231502, __pyx_int_210565405, __pyx_int_252507329); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self")
 * cdef object __Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self(object (*f)(Wrapped) ):
 *     def wrap(Wrapped self):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Wrapped')"""
 *         return f(self)
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 67, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(6, 67, __pyx_L1_error)
  __pyx_tuple__40 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_c); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 67, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(6, 67, __pyx_L1_error)
  __pyx_tuple__42 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_a, __pyx_n_s_op); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 67, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(6, 67, __pyx_L1_error)

  /* "imports.pxi":8
 *     PY2 = False
 *     builtin_module = sys.modules['builtins']
 *     import collections.abc as CollectionsABC             # <<<<<<<<<<<<<<
 * else:
 *     PY2 = True
 */
  __pyx_tuple__45 = PyTuple_Pack(2, __pyx_n_s_collections, __pyx_n_s_abc); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(12, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "python_visible.pxi":6
 * # ------------------------------------------------------------------------
 * 
 * def attribute_protected():             # <<<<<<<<<<<<<<
 *     '''
 *     attribute_protected() -> str: name of special attribute in Wrapped objects
 */
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_attribute_protected, 6, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 6, __pyx_L1_error)

  /* "python_visible.pxi":12
 *     return PROT_ATTR_NAME
 * 
 * def id_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     id_protected(o: object) -> int:
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_n_s_o); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_id_protected, 12, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 12, __pyx_L1_error)

  /* "python_visible.pxi":22
 * 
 * 
 * def hash_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     hash_protected(o: object) -> int:
 */
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_hash_protected, 22, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 22, __pyx_L1_error)

  /* "python_visible.pxi":32
 * 
 * 
 * def isinstance_protected(o: object, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isinstance_protected(o: object, c: type) -> bool:
 */
  __pyx_tuple__50 = PyTuple_Pack(2, __pyx_n_s_o, __pyx_n_s_c); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_isinstance_protected, 32, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 32, __pyx_L1_error)

  /* "python_visible.pxi":43
 * 
 * 
 * def issubclass_protected(o: type, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     issubclass_protected(o: object, c: type) -> bool:
 */
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_issubclass_protected, 43, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 43, __pyx_L1_error)

  /* "python_visible.pxi":54
 * 
 * 
 * def instance_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     instance_of_protected(x: object, w: object) -> bool
 */
  __pyx_tuple__53 = PyTuple_Pack(2, __pyx_n_s_x, __pyx_n_s_w); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_instance_of_protected, 54, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 54, __pyx_L1_error)

  /* "python_visible.pxi":66
 * 
 * 
 * def subclass_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     subclass_of_protected(x: object, w: object) -> bool
 */
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_subclass_of_protected, 66, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 66, __pyx_L1_error)

  /* "python_visible.pxi":78
 * 
 * 
 * def same_class_protected(c: type, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     same_class_protected(c: type, w: object) -> bool
 */
  __pyx_tuple__56 = PyTuple_Pack(2, __pyx_n_s_c, __pyx_n_s_w); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_same_class_protected, 78, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 78, __pyx_L1_error)

  /* "python_visible.pxi":90
 * 
 * 
 * def help_protected(o: object) -> None:             # <<<<<<<<<<<<<<
 *     '''
 *     help_protected(o: object) -> None
 */
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_help_protected, 90, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 90, __pyx_L1_error)

  /* "python_visible.pxi":100
 * 
 * 
 * def contains(p: object, o: object):             # <<<<<<<<<<<<<<
 *     '''
 *     contains(p: object, o: object): bool: whether 'p' wraps 'o'
 */
  __pyx_tuple__59 = PyTuple_Pack(2, __pyx_n_s_p, __pyx_n_s_o); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_contains_2, 100, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 100, __pyx_L1_error)

  /* "python_visible.pxi":114
 * 
 * 
 * def isimmutable(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isimmutable(o: object) -> bool: 'o' is KNOWN to be immutable
 */
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_isimmutable, 114, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 114, __pyx_L1_error)

  /* "python_visible.pxi":138
 * 
 * 
 * def iswrapped(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     iswrapped(o: object) -> bool:
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_iswrapped, 138, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 138, __pyx_L1_error)

  /* "python_visible.pxi":146
 * 
 * 
 * def isfrozen(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isfrozen(o: object) -> bool: 'o' was created using freeze()
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_isfrozen, 146, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 146, __pyx_L1_error)

  /* "python_visible.pxi":155
 * 
 * 
 * def isprivate(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprivate(o: object) -> bool: 'o' was created using private()
 */
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_isprivate, 155, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 155, __pyx_L1_error)

  /* "python_visible.pxi":165
 * 
 * 
 * def isprotected(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprotected(o: object) -> bool: 'o' was created using protect()
 */
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_isprotected, 165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 165, __pyx_L1_error)

  /* "python_visible.pxi":175
 * 
 * 
 * def isreadonly(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isreadonly(o: object, a: str) -> bool:
 */
  __pyx_tuple__66 = PyTuple_Pack(2, __pyx_n_s_o, __pyx_n_s_a); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_isreadonly, 175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 175, __pyx_L1_error)

  /* "python_visible.pxi":195
 * 
 * 
 * def isvisible(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isvisible(o: object, a: str) -> bool:
 */
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_isvisible, 195, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 195, __pyx_L1_error)

  /* "python_visible.pxi":216
 * 
 * 
 * def wrap(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     wrap(o: object) -> object:
 */
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_wrap, 216, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 216, __pyx_L1_error)

  /* "python_visible.pxi":239
 * 
 * 
 * def freeze(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     freeze(o: object) -> object:
 */
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_freeze, 239, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 239, __pyx_L1_error)

  /* "python_visible.pxi":261
 * 
 * 
 * def private(o: object, frozen: bool = False) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     private(o: object, frozen: bool = False) -> object:
 */
  __pyx_tuple__71 = PyTuple_Pack(2, __pyx_n_s_o, __pyx_n_s_frozen); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_private, 261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 261, __pyx_L1_error)
  __pyx_tuple__73 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "python_visible.pxi":294
 * 
 * 
 * def protect(             # <<<<<<<<<<<<<<
 *     o: object,
 *     frozen: bool = False, dynamic: bool = True,
 */
  __pyx_tuple__74 = PyTuple_Pack(17, __pyx_n_s_o, __pyx_n_s_frozen, __pyx_n_s_dynamic, __pyx_n_s_hide_private, __pyx_n_s_ro_data, __pyx_n_s_ro_method, __pyx_n_s_ro, __pyx_n_s_rw, __pyx_n_s_hide, __pyx_n_s_kwargs, __pyx_n_s_kw1, __pyx_n_s_d, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_kw2, __pyx_n_s_rules, __pyx_n_s_want_frozen); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(9, 0, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__74, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_protect, 294, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 294, __pyx_L1_error)

  /* "python_visible.pxi":392
 * # ------------------------------------------------------------------------
 * 
 * def never_writeable():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable() -> set(str): Attributes that are never writeable
 */
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_never_writeable, 392, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 392, __pyx_L1_error)

  /* "python_visible.pxi":399
 *     return overridden_always
 * 
 * def never_writeable_private():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable_private() -> set(str): Attributes that are never
 */
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_never_writeable_private, 399, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 399, __pyx_L1_error)

  /* "python_visible.pxi":409
 *     ))
 * 
 * def hidden_pickle_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     hidden_pickle_attributes() -> set(str): Attributes that are never
 */
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_hidden_pickle_attributes, 409, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 409, __pyx_L1_error)

  /* "python_visible.pxi":416
 *     return pickle_attributes
 * 
 * def always_delegated_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     always_delegated_attributes() -> set(str): Attributes that are
 */
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_always_delegated_attributes, 416, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 416, __pyx_L1_error)

  /* "python_visible.pxi":423
 *     return always_delegated
 * 
 * def immutable_builtin_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     immutable_builtin_attributes() -> frozenset(str)
 */
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_immutable_builtin_attributes, 423, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 423, __pyx_L1_error)

  /* "python_visible.pxi":444
 * 
 * 
 * def __dir__():             # <<<<<<<<<<<<<<
 *     return __all__
 * 
 */
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_python_visible_pxi, __pyx_n_s_dir, 444, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 444, __pyx_L1_error)

  /* "global_cdefs.pxi":8
 * # PROT_ATTR_NAME is set ONLY in get_protected_attr_name()
 * cdef str PROT_ATTR_NAME = get_protected_attr_name()
 * cdef Exception frozen_error = ProtectionError('Object is read-only')             # <<<<<<<<<<<<<<
 * (
 *     immutable_types_set,
 */
  __pyx_tuple__82 = PyTuple_Pack(1, __pyx_kp_s_Object_is_read_only); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(13, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "global_cdefs.pxi":18
 *     Py_LT, Py_EQ, Py_GT, Py_LE, Py_NE, Py_GE,
 * )
 * cdef object overridden_always = frozenset([             # <<<<<<<<<<<<<<
 *     '__getattribute__', '__delattr__', '__setattr__',
 *     '__hash__',
 */
  __pyx_tuple__83 = PyTuple_Pack(4, __pyx_n_s_getattribute, __pyx_n_s_delattr, __pyx_n_s_setattr, __pyx_n_s_hash_2); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(13, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "global_cdefs.pxi":22
 *     '__hash__',
 * ])
 * cdef object pickle_attributes = frozenset([             # <<<<<<<<<<<<<<
 *     '__reduce__', '__reduce_ex__',
 *     '__getsate__', '__setstate__',
 */
  __pyx_tuple__84 = PyTuple_Pack(4, __pyx_n_s_reduce, __pyx_n_s_reduce_ex, __pyx_n_s_getsate, __pyx_n_s_setstate); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(13, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "global_cdefs.pxi":29
 *     PROT_ATTR_NAME,
 * ])
 * cdef object always_delegated = frozenset([             # <<<<<<<<<<<<<<
 *     '__doc__',
 *     '__weakref__',
 */
  __pyx_tuple__85 = PyTuple_Pack(3, __pyx_n_s_doc, __pyx_n_s_weakref, __pyx_n_s_package); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(13, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "global_cdefs.pxi":34
 *     '__package__',
 * ])
 * cdef object always_frozen = frozenset([             # <<<<<<<<<<<<<<
 *     '__dict__', '__slots__', '__class__',
 *     '__module__',
 */
  __pyx_tuple__86 = PyTuple_Pack(4, __pyx_n_s_dict_2, __pyx_n_s_slots, __pyx_n_s_class, __pyx_n_s_module); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(13, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "global_cdefs.pxi":41
 * # Use compiled regex - no function call, no str operations
 * # Python 2 str does not have isidentifier() method
 * cdef object attr_identifier = re.compile('^[_a-zA-Z][a-zA-Z0-9_]*$')             # <<<<<<<<<<<<<<
 * # ro_private_attr: Start with single _, ending n non-underscore
 * cdef object ro_private_attr = re.compile('^_[^_].*?(?<!_)$')
 */
  __pyx_tuple__87 = PyTuple_Pack(1, __pyx_kp_s_a_zA_Z_a_zA_Z0_9); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(13, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);

  /* "global_cdefs.pxi":43
 * cdef object attr_identifier = re.compile('^[_a-zA-Z][a-zA-Z0-9_]*$')
 * # ro_private_attr: Start with single _, ending n non-underscore
 * cdef object ro_private_attr = re.compile('^_[^_].*?(?<!_)$')             # <<<<<<<<<<<<<<
 * # unmangled_private_attr: Start with double _, end in non-underscore or single _
 * cdef object unmangled_private_attr = re.compile('^__[^_].*?[^_][_]{0,1}$')
 */
  __pyx_tuple__89 = PyTuple_Pack(1, __pyx_kp_s__88); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(13, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "global_cdefs.pxi":45
 * cdef object ro_private_attr = re.compile('^_[^_].*?(?<!_)$')
 * # unmangled_private_attr: Start with double _, end in non-underscore or single _
 * cdef object unmangled_private_attr = re.compile('^__[^_].*?[^_][_]{0,1}$')             # <<<<<<<<<<<<<<
 * # PY2 'Old style' CLASSES (not INSTANCES of such classes) do not have
 * # __class__ attribute, so Private wrappers around such classes will hide
 */
  __pyx_tuple__90 = PyTuple_Pack(1, __pyx_kp_s_0_1); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(13, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "global_cdefs.pxi":124
 * ])
 * # These attributes of FunctionType are writable only in PY2
 * py2_function_attrs_rw = frozenset([             # <<<<<<<<<<<<<<
 *     '__doc__', '__name__', '__module__',
 *     '__defaults__', '__code__', '__dict__',
 */
  __pyx_tuple__91 = PyTuple_Pack(6, __pyx_n_s_doc, __pyx_n_s_name, __pyx_n_s_module, __pyx_n_s_defaults, __pyx_n_s_code, __pyx_n_s_dict_2); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(13, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "ProtectionData.pxi":107
 *         raise ProtectionError('Object is read-only')
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return list(self.attributes_map.keys())
 * 
 */
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_ProtectionData_pxi, __pyx_n_s_dir, 107, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(1, 107, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__93 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_3, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___ProtectionData__set_state(self, __pyx_state)
 */
  __pyx_tuple__95 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "Proxy.pxi":31
 *         return str(self.pvt_o)
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(self.pvt_o)
 * 
 */
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_bytes_2, 31, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(2, 31, __pyx_L1_error)

  /* "Proxy.pxi":56
 * 
 *     # Representations - return immutable
 *     def __format__(self, val):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__format__(val)
 * 
 */
  __pyx_tuple__98 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_val); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(2, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_format, 56, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(2, 56, __pyx_L1_error)

  /* "Proxy.pxi":76
 *         return self.pvt_o.__len__()
 * 
 *     def __length_hint__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__length_hint__()
 * 
 */
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_length_hint, 76, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(2, 76, __pyx_L1_error)

  /* "Proxy.pxi":92
 *         return self.pvt_o.__invert__()
 * 
 *     def __complex__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__complex__()
 * 
 */
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_complex_2, 92, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(2, 92, __pyx_L1_error)

  /* "Proxy.pxi":104
 *         return self.pvt_o.__index__()
 * 
 *     def __round__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__round__()
 * 
 */
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_round, 104, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(2, 104, __pyx_L1_error)

  /* "Proxy.pxi":107
 *         return self.pvt_o.__round__()
 * 
 *     def __trunc__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__trunc__()
 * 
 */
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_trunc, 107, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(2, 107, __pyx_L1_error)

  /* "Proxy.pxi":110
 *         return self.pvt_o.__trunc__()
 * 
 *     def __floor__(self):             # <<<<<<<<<<<<<<
 *         return math.floor(self.pvt_o)
 * 
 */
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_floor_2, 110, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(2, 110, __pyx_L1_error)

  /* "Proxy.pxi":113
 *         return math.floor(self.pvt_o)
 * 
 *     def __ceil__(self):             # <<<<<<<<<<<<<<
 *         return math.ceil(self.pvt_o)
 * 
 */
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_ceil_2, 113, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(2, 113, __pyx_L1_error)

  /* "Proxy.pxi":376
 * 
 *     # Just pass to wrapped
 *     def __instancecheck__(self, inst):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 */
  __pyx_tuple__106 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_inst); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(2, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__106, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_instancecheck, 376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(2, 376, __pyx_L1_error)

  /* "Proxy.pxi":379
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 *     def __subclasscheck__(self, subclass):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__subclasscheck__(subclass)
 * 
 */
  __pyx_tuple__108 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_subclass); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_subclasscheck, 379, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(2, 379, __pyx_L1_error)

  /* "Proxy.pxi":383
 * 
 *     # Context managers - intrinsic behavior - pass to wrapped
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__enter__()
 * 
 */
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_enter, 383, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(2, 383, __pyx_L1_error)

  /* "Proxy.pxi":386
 *         return self.pvt_o.__enter__()
 * 
 *     def __exit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__exit(exc_type, exc_value, tb)
 * 
 */
  __pyx_tuple__111 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_exc_type, __pyx_n_s_exc_value, __pyx_n_s_tb); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(2, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_exit, 386, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(2, 386, __pyx_L1_error)

  /* "Proxy.pxi":390
 * 
 *     # Async context managers - intrinsic behavior - pass to wrapped
 *     def __aenter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aenter__()
 * 
 */
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_aenter, 390, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(2, 390, __pyx_L1_error)

  /* "Proxy.pxi":393
 *         return self.pvt_o.__aenter__()
 * 
 *     def __aexit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aexit(exc_type, exc_value, tb)
 * 
 */
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_aexit, 393, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(2, 393, __pyx_L1_error)

  /* "Proxy.pxi":398
 *     # Customizing positional arguments in class pattern matching
 *     # Returns tuple of strings (immutable) - pass to wrapped
 *     def __match_args__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__match_args__()
 * 
 */
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_match_args, 398, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(2, 398, __pyx_L1_error)

  /* "Proxy.pxi":402
 * 
 *     # Coroutine objects - intrinsic behavior - pass to wrapped
 *     def send(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.send(value)
 * 
 */
  __pyx_tuple__116 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_value); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(2, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_send, 402, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(2, 402, __pyx_L1_error)

  /* "Proxy.pxi":405
 *         return self.pvt_o.send(value)
 * 
 *     def throw(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.throw(value)
 * 
 */
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_throw, 405, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(2, 405, __pyx_L1_error)

  /* "Proxy.pxi":428
 *     # --------------------------------------------------------------------
 * 
 *     def clear(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_tuple__119 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(2, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_clear, 428, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(2, 428, __pyx_L1_error)

  /* "Proxy.pxi":440
 *         return self.pvt_o.clear(*args, **kwargs)
 * 
 *     def setdefault(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,CollectionsABC.MutableMapping
 */
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_setdefault, 440, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(2, 440, __pyx_L1_error)

  /* "Proxy.pxi":447
 *         return self.pvt_o.setdefault(*args, **kwargs)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_pop, 447, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(2, 447, __pyx_L1_error)

  /* "Proxy.pxi":459
 *         return self.pvt_o.pop(*args, **kwargs)
 * 
 *     def popitem(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_popitem, 459, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(2, 459, __pyx_L1_error)

  /* "Proxy.pxi":466
 *         return self.pvt_o.popitem(*args, **kwargs)
 * 
 *     def update(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_update, 466, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(2, 466, __pyx_L1_error)

  /* "Proxy.pxi":474
 * 
 * 
 *     def append(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_append, 474, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(2, 474, __pyx_L1_error)

  /* "Proxy.pxi":481
 *         return self.pvt_o.append(*args, **kwargs)
 * 
 *     def extend(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_extend, 481, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(2, 481, __pyx_L1_error)

  /* "Proxy.pxi":488
 *         return self.pvt_o.extend(*args, **kwargs)
 * 
 *     def insert(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_insert, 488, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(2, 488, __pyx_L1_error)

  /* "Proxy.pxi":495
 *         return self.pvt_o.insert(*args, **kwargs)
 * 
 *     def sort(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_sort, 495, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(2, 495, __pyx_L1_error)

  /* "Proxy.pxi":502
 *         return self.pvt_o.sort(*args, **kwargs)
 * 
 *     def add(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */
  __pyx_codeobj__129 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_add_2, 502, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__129)) __PYX_ERR(2, 502, __pyx_L1_error)

  /* "Proxy.pxi":509
 *         return self.pvt_o.add(*args, **kwargs)
 * 
 *     def discard(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_discard, 509, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(2, 509, __pyx_L1_error)

  /* "Proxy.pxi":516
 *         return self.pvt_o.discard(*args, **kwargs)
 * 
 *     def remove(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_remove, 516, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(2, 516, __pyx_L1_error)

  /* "Proxy.pxi":527
 *         return self.pvt_o.remove(*args, **kwargs)
 * 
 *     def reverse(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_codeobj__132 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Proxy_pxi, __pyx_n_s_reverse, 527, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__132)) __PYX_ERR(2, 527, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Proxy__set_state(self, __pyx_state)
 */
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "Wrapped_Frozen.pxi":394
 *         delattr(self.pvt_o, a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.wrapped_dir()
 * 
 */
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Wrapped_Frozen_pxi, __pyx_n_s_dir, 394, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(3, 394, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Wrapped__set_state(self, __pyx_state)
 */
  __pyx_codeobj__137 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__137)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Frozen__set_state(self, __pyx_state)
 */
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":99
 *         return str(ret)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         # TODO: Something strange here: Just wrapping self.pvt_o and returning
 *         # as a new FrozenPrivacyDict doesn't seem to adopt all the behavior
 */
  __pyx_tuple__140 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_d); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(4, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_copy, 99, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(4, 99, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":115
 *     # --------------------------------------------------------------------
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):
 */
  __pyx_tuple__142 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_k); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(4, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_keys, 115, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(4, 115, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":121
 *             yield k
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
  __pyx_tuple__143 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(4, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_items, 121, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(4, 121, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":126
 *             yield self.fif((k, v))
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_values, 126, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(4, 126, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":135
 *     # --------------------------------------------------------------------
 * 
 *     def keys_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields keys
 *         ret = []
 */
  __pyx_tuple__144 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_ret_2, __pyx_n_s_k); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(4, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);
  __pyx_codeobj__145 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__144, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_keys_py2, 135, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__145)) __PYX_ERR(4, 135, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":144
 *         return ret
 * 
 *     def items_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields items
 *         ret = []
 */
  __pyx_tuple__146 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_ret_2, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(4, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  __pyx_codeobj__147 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_items_py2, 144, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__147)) __PYX_ERR(4, 144, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":152
 *         return self.fif(ret)
 * 
 *     def values_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields values
 *         ret = []
 */
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_values_py2, 152, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(4, 152, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":160
 *         return self.fif(ret)
 * 
 *     def iterkeys(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.pvt_o.keys():
 */
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_iterkeys, 160, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(4, 160, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":167
 *             yield k
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_iteritems, 167, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(4, 167, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":173
 *             yield self.fif((k, v))
 * 
 *     def itervalues(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_itervalues, 173, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(4, 173, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":178
 *             yield self.fif(self.pvt_o[k])
 * 
 *     def viewitems(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.items()))
 * 
 */
  __pyx_codeobj__149 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_viewitems, 178, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__149)) __PYX_ERR(4, 178, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":181
 *         return self.fif(set(self.items()))
 * 
 *     def viewvalues(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.values()))
 * 
 */
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_viewvalues, 181, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(4, 181, __pyx_L1_error)

  /* "PrivacyDict_FrozenPrivacyDict.pxi":184
 *         return self.fif(set(self.values()))
 * 
 *     def viewkeys(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.keys()))
 * 
 */
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px, __pyx_n_s_viewkeys, 184, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(4, 184, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__152 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__152)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrivacyDict__set_state(self, __pyx_state)
 */
  __pyx_codeobj__153 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__153)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivacyDict__set_state(self, __pyx_state)
 */
  __pyx_codeobj__155 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__155)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "Private_FrozenPrivate.pxi":129
 *         self.private_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.private_dir()
 * 
 */
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Private_FrozenPrivate_pxi, __pyx_n_s_dir, 129, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(7, 129, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__157 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__157)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Private__set_state(self, __pyx_state)
 */
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__159 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__159)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivate__set_state(self, __pyx_state)
 */
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "Protected_FrozenProtected.pxi":265
 *         self.protected_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.protected_dir()
 * 
 */
  __pyx_codeobj__161 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Protected_FrozenProtected_pxi, __pyx_n_s_dir, 265, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__161)) __PYX_ERR(8, 265, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Protected__set_state(self, __pyx_state)
 */
  __pyx_codeobj__163 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__163)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__164 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__164)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenProtected__set_state(self, __pyx_state)
 */
  __pyx_codeobj__165 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__165)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "HiddenPartial.pxi":37
 *         raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['__call__']
 * 
 */
  __pyx_codeobj__166 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_HiddenPartial_pxi, __pyx_n_s_dir, 37, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__166)) __PYX_ERR(9, 37, __pyx_L1_error)

  /* "HiddenPartial.pxi":55
 *         return str(type(functools.partial))
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(type(functools.partial))
 * 
 */
  __pyx_codeobj__167 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_HiddenPartial_pxi, __pyx_n_s_bytes_2, 55, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__167)) __PYX_ERR(9, 55, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__168 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__168)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___HiddenPartial__set_state(self, __pyx_state)
 */
  __pyx_codeobj__169 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__169)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle___ProtectionData(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__170 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__170)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);
  __pyx_codeobj__171 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___ProtectionData, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__171)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__172 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Proxy, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__172)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__173 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Wrapped, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__173)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__174 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Frozen, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__174)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__175 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PrivacyDict, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__175)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FrozenPrivacyDict, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__177 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Private, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__177)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FrozenPrivate, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__179 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Protected, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__179)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FrozenProtected, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__181 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___HiddenPartial, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__181)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_get.method_name = &__pyx_n_s_get;
  __pyx_umethod_PyDict_Type_update.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_update.method_name = &__pyx_n_s_update;
  __pyx_umethod_PySet_Type_intersection.type = (PyObject*)&PySet_Type;
  __pyx_umethod_PySet_Type_intersection.method_name = &__pyx_n_s_intersection;
  __pyx_umethod_PySet_Type_union.type = (PyObject*)&PySet_Type;
  __pyx_umethod_PySet_Type_union.method_name = &__pyx_n_s_union;
  #if CYTHON_USE_MODULE_STATE
  if (__Pyx_InitString(__pyx_string_tab[0], &__pyx_n_s_) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[1], &__pyx_kp_s_0_1) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[2], &__pyx_n_s_AssertionError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[3], &__pyx_n_s_AttributeError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[4], &__pyx_n_s_C) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[5], &__pyx_kp_s_Cannot_add_attribute_s_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[6], &__pyx_kp_s_Cannot_delete_attribute_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[7], &__pyx_kp_s_Cannot_delete_attribute_s_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[8], &__pyx_kp_s_Cannot_delete_private_attribute) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[9], &__pyx_kp_s_Cannot_modify_attribute_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[10], &__pyx_kp_s_Cannot_set_attribute_s_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[11], &__pyx_kp_s_Cannot_set_private_attribute_s_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[12], &__pyx_n_s_CollectionsABC) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[13], &__pyx_kp_s_Double_wrapped) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[14], &__pyx_n_s_FrameType) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[15], &__pyx_n_s_Frozen) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[16], &__pyx_n_s_FrozenPrivacyDict) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[17], &__pyx_n_s_FrozenPrivacyDict___reduce_cytho) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[18], &__pyx_n_s_FrozenPrivacyDict___setstate_cyt) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[19], &__pyx_n_s_FrozenPrivate) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[20], &__pyx_n_s_FrozenPrivate___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[21], &__pyx_n_s_FrozenPrivate___setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[22], &__pyx_n_s_FrozenProtected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[23], &__pyx_n_s_FrozenProtected___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[24], &__pyx_n_s_FrozenProtected___setstate_cytho) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[25], &__pyx_n_s_Frozen___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[26], &__pyx_n_s_Frozen___setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[27], &__pyx_n_s_HiddenPartial) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[28], &__pyx_n_s_HiddenPartial___bytes) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[29], &__pyx_n_s_HiddenPartial___dir) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[30], &__pyx_n_s_HiddenPartial___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[31], &__pyx_n_s_HiddenPartial___setstate_cytho) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[32], &__pyx_kp_s_HiddenPartial_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[33], &__pyx_kp_s_Incompatible_checksums_0x_x_vs_0) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[34], &__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[35], &__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[36], &__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[37], &__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[38], &__pyx_n_s_KeyError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[39], &__pyx_n_s_Mapping) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[40], &__pyx_n_s_ModuleType) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[41], &__pyx_n_s_MutableMapping) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[42], &__pyx_n_s_MutableSequence) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[43], &__pyx_n_s_MutableSet) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[44], &__pyx_n_s_None) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[45], &__pyx_n_s_NotImplemented) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[46], &__pyx_kp_s_Object_Private_s_has_no_attribut) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[47], &__pyx_kp_s_Object_Protected_s_has_no_attrib) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[48], &__pyx_kp_s_Object_Wrapped_s_has_no_attribut) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[49], &__pyx_kp_s_Object___HiddenPartial_has_no_at) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[50], &__pyx_kp_s_Object_is_read_only) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[51], &__pyx_kp_s_Object_s_has_no_attribute_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[52], &__pyx_n_s_PYPY) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[53], &__pyx_n_s_PickleError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[54], &__pyx_n_s_PrivacyDict) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[55], &__pyx_kp_s_PrivacyDict_FrozenPrivacyDict_px) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[56], &__pyx_n_s_PrivacyDict___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[57], &__pyx_n_s_PrivacyDict___setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[58], &__pyx_n_s_PrivacyDict_copy) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[59], &__pyx_n_s_PrivacyDict_items) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[60], &__pyx_n_s_PrivacyDict_items_py2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[61], &__pyx_n_s_PrivacyDict_iteritems) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[62], &__pyx_n_s_PrivacyDict_iterkeys) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[63], &__pyx_n_s_PrivacyDict_itervalues) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[64], &__pyx_n_s_PrivacyDict_keys) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[65], &__pyx_n_s_PrivacyDict_keys_py2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[66], &__pyx_n_s_PrivacyDict_values) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[67], &__pyx_n_s_PrivacyDict_values_py2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[68], &__pyx_n_s_PrivacyDict_viewitems) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[69], &__pyx_n_s_PrivacyDict_viewkeys) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[70], &__pyx_n_s_PrivacyDict_viewvalues) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[71], &__pyx_n_s_Private) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[72], &__pyx_kp_s_Private_FrozenPrivate_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[73], &__pyx_n_s_Private___dir) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[74], &__pyx_n_s_Private___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[75], &__pyx_n_s_Private___setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[76], &__pyx_n_s_Protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[77], &__pyx_n_s_Protected_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[78], &__pyx_kp_s_Protected_FrozenProtected_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[79], &__pyx_n_s_Protected___dir) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[80], &__pyx_n_s_Protected___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[81], &__pyx_n_s_Protected___setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[82], &__pyx_n_s_ProtectionData) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[83], &__pyx_n_s_ProtectionData___dir) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[84], &__pyx_n_s_ProtectionData___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[85], &__pyx_n_s_ProtectionData___setstate_cyth) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[86], &__pyx_kp_s_ProtectionData_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[87], &__pyx_n_s_ProtectionError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[88], &__pyx_n_s_Proxy) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[89], &__pyx_n_s_Proxy___aenter) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[90], &__pyx_n_s_Proxy___aexit) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[91], &__pyx_n_s_Proxy___bytes) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[92], &__pyx_n_s_Proxy___ceil) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[93], &__pyx_n_s_Proxy___complex) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[94], &__pyx_n_s_Proxy___enter) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[95], &__pyx_n_s_Proxy___exit) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[96], &__pyx_n_s_Proxy___floor) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[97], &__pyx_n_s_Proxy___format) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[98], &__pyx_n_s_Proxy___instancecheck) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[99], &__pyx_n_s_Proxy___iter) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[100], &__pyx_n_s_Proxy___length_hint) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[101], &__pyx_n_s_Proxy___match_args) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[102], &__pyx_n_s_Proxy___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[103], &__pyx_n_s_Proxy___round) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[104], &__pyx_n_s_Proxy___setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[105], &__pyx_n_s_Proxy___subclasscheck) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[106], &__pyx_n_s_Proxy___trunc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[107], &__pyx_n_s_Proxy__aexit) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[108], &__pyx_n_s_Proxy__exit) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[109], &__pyx_n_s_Proxy_add) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[110], &__pyx_n_s_Proxy_append) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[111], &__pyx_n_s_Proxy_clear) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[112], &__pyx_n_s_Proxy_discard) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[113], &__pyx_n_s_Proxy_extend) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[114], &__pyx_n_s_Proxy_insert) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[115], &__pyx_n_s_Proxy_pop) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[116], &__pyx_n_s_Proxy_popitem) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[117], &__pyx_kp_s_Proxy_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[118], &__pyx_n_s_Proxy_remove) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[119], &__pyx_n_s_Proxy_reverse) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[120], &__pyx_n_s_Proxy_send) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[121], &__pyx_n_s_Proxy_setdefault) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[122], &__pyx_n_s_Proxy_sort) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[123], &__pyx_n_s_Proxy_throw) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[124], &__pyx_n_s_Proxy_update) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[125], &__pyx_n_s_PyPy) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[126], &__pyx_n_s_Pyx_CFunc_object__lParenWrappe) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[127], &__pyx_n_s_Pyx_CFunc_object__lParenWrappe_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[128], &__pyx_n_s_Pyx_CFunc_object__lParenWrappe_3) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[129], &__pyx_kp_s_Read_only_attribute_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[130], &__pyx_n_s_RecursionError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[131], &__pyx_n_s_RuntimeError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[132], &__pyx_n_s_Sequence) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[133], &__pyx_n_s_Set) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[134], &__pyx_n_s_TypeError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[135], &__pyx_n_s_Unknown_OldStyle_Class) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[136], &__pyx_n_s_Wrapped) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[137], &__pyx_kp_s_Wrapped_Frozen_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[138], &__pyx_n_s_Wrapped___dir) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[139], &__pyx_n_s_Wrapped___reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[140], &__pyx_n_s_Wrapped___setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[141], &__pyx_n_s_Wrapped_comparator_locals_pass_t) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[142], &__pyx_kp_s_Wrapped_object_cannot_be_pickled) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[143], &__pyx_kp_s__17) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[144], &__pyx_n_s__182) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[145], &__pyx_n_s__2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[146], &__pyx_kp_s__3) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[147], &__pyx_kp_u__33) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[148], &__pyx_kp_s__4) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[149], &__pyx_n_s__44) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[150], &__pyx_kp_s__88) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[151], &__pyx_n_s_a) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[152], &__pyx_kp_s_a_zA_Z_a_zA_Z0_9) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[153], &__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[154], &__pyx_n_s_abc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[155], &__pyx_n_s_abs) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[156], &__pyx_n_s_add) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[157], &__pyx_n_s_add_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[158], &__pyx_n_s_aenter) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[159], &__pyx_n_s_aexit) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[160], &__pyx_n_s_alist) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[161], &__pyx_n_s_all) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[162], &__pyx_n_s_always_delegated_attributes) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[163], &__pyx_n_s_and) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[164], &__pyx_n_s_append) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[165], &__pyx_n_s_args) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[166], &__pyx_n_s_asyncio_coroutines) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[167], &__pyx_n_s_attr_type_check) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[168], &__pyx_n_s_attribute_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[169], &__pyx_n_s_basestring) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[170], &__pyx_n_s_bool) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[171], &__pyx_n_s_bool_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[172], &__pyx_n_s_build_regex) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[173], &__pyx_n_s_builtin) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[174], &__pyx_n_s_builtins) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[175], &__pyx_n_s_bytearray) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[176], &__pyx_n_s_bytes) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[177], &__pyx_n_s_bytes_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[178], &__pyx_n_s_c) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[179], &__pyx_n_s_call) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[180], &__pyx_n_s_ceil) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[181], &__pyx_n_s_ceil_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[182], &__pyx_n_s_cfunc_to_py) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[183], &__pyx_n_s_class) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[184], &__pyx_n_s_class_getitem) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[185], &__pyx_n_s_clear) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[186], &__pyx_n_s_cline_in_traceback) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[187], &__pyx_n_s_close) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[188], &__pyx_n_s_cmp) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[189], &__pyx_n_s_cn) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[190], &__pyx_n_s_code) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[191], &__pyx_n_s_collections) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[192], &__pyx_n_s_collections_abc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[193], &__pyx_n_s_compile) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[194], &__pyx_n_s_complex) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[195], &__pyx_n_s_complex_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[196], &__pyx_n_s_contains) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[197], &__pyx_n_s_contains_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[198], &__pyx_n_s_copy) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[199], &__pyx_n_s_d) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[200], &__pyx_n_s_defaults) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[201], &__pyx_n_s_delattr) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[202], &__pyx_n_s_delete) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[203], &__pyx_n_s_delitem) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[204], &__pyx_n_s_dict) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[205], &__pyx_n_s_dict_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[206], &__pyx_n_s_dict_3) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[207], &__pyx_n_s_difference) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[208], &__pyx_n_s_dir) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[209], &__pyx_kp_u_disable) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[210], &__pyx_n_s_discard) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[211], &__pyx_n_s_divmod) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[212], &__pyx_n_s_doc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[213], &__pyx_n_s_dynamic) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[214], &__pyx_kp_u_enable) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[215], &__pyx_n_s_endswith) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[216], &__pyx_n_s_enter) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[217], &__pyx_n_s_environ) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[218], &__pyx_n_s_eq) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[219], &__pyx_n_s_exc_type) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[220], &__pyx_n_s_exc_value) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[221], &__pyx_n_s_exit) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[222], &__pyx_n_s_extend) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[223], &__pyx_n_s_file) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[224], &__pyx_n_s_float) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[225], &__pyx_n_s_float_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[226], &__pyx_n_s_floor) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[227], &__pyx_n_s_floor_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[228], &__pyx_n_s_floordiv) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[229], &__pyx_n_s_format) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[230], &__pyx_n_s_freeze) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[231], &__pyx_n_s_frozen) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[232], &__pyx_n_s_frozenset) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[233], &__pyx_n_s_functools) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[234], &__pyx_kp_u_gc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[235], &__pyx_n_s_ge) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[236], &__pyx_n_s_get) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[237], &__pyx_n_s_getattribute) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[238], &__pyx_n_s_getitem) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[239], &__pyx_n_s_getsate) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[240], &__pyx_n_s_getstate) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[241], &__pyx_kp_s_global_c_functions_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[242], &__pyx_n_s_gt) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[243], &__pyx_n_s_hash) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[244], &__pyx_n_s_hash_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[245], &__pyx_n_s_hash_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[246], &__pyx_n_s_hash_val) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[247], &__pyx_n_s_help) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[248], &__pyx_n_s_help_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[249], &__pyx_n_s_help_str) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[250], &__pyx_n_s_help_val) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[251], &__pyx_n_s_hidden_pickle_attributes) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[252], &__pyx_n_s_hide) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[253], &__pyx_n_s_hide_private) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[254], &__pyx_n_s_hide_regex) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[255], &__pyx_n_s_iadd) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[256], &__pyx_n_s_iand) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[257], &__pyx_n_s_id) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[258], &__pyx_n_s_id_class) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[259], &__pyx_n_s_id_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[260], &__pyx_n_s_id_val) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[261], &__pyx_n_s_ifloordiv) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[262], &__pyx_n_s_ilshift) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[263], &__pyx_n_s_imatmul) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[264], &__pyx_n_s_immutable_builtin_attributes) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[265], &__pyx_n_s_imod) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[266], &__pyx_n_s_import) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[267], &__pyx_n_s_imul) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[268], &__pyx_n_s_index) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[269], &__pyx_n_s_init) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[270], &__pyx_n_s_init___locals_C) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[271], &__pyx_n_s_init_subclass) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[272], &__pyx_n_s_initializing) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[273], &__pyx_n_s_insert) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[274], &__pyx_n_s_inst) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[275], &__pyx_n_s_instance_of_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[276], &__pyx_n_s_instancecheck) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[277], &__pyx_n_s_instanceof) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[278], &__pyx_n_s_int) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[279], &__pyx_n_s_int_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[280], &__pyx_n_s_intersection) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[281], &__pyx_n_s_invert) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[282], &__pyx_n_s_ior) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[283], &__pyx_n_s_ipow) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[284], &__pyx_n_s_irshift) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[285], &__pyx_n_s_is_coroutine) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[286], &__pyx_kp_u_isenabled) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[287], &__pyx_n_s_isfrozen) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[288], &__pyx_n_s_isimmutable) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[289], &__pyx_n_s_isinstance) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[290], &__pyx_n_s_isinstance_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[291], &__pyx_n_s_isinstance_val) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[292], &__pyx_n_s_isprivate) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[293], &__pyx_n_s_isprotected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[294], &__pyx_n_s_isreadonly) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[295], &__pyx_n_s_issubclass) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[296], &__pyx_n_s_issubclass_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[297], &__pyx_n_s_issubclass_val) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[298], &__pyx_n_s_isub) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[299], &__pyx_n_s_isvisible) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[300], &__pyx_n_s_iswrapped) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[301], &__pyx_n_s_items) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[302], &__pyx_n_s_items_py2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[303], &__pyx_n_s_iter) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[304], &__pyx_n_s_iteritems) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[305], &__pyx_n_s_iterkeys) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[306], &__pyx_n_s_itervalues) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[307], &__pyx_n_s_itruediv) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[308], &__pyx_n_s_ixor) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[309], &__pyx_n_s_join) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[310], &__pyx_n_s_k) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[311], &__pyx_n_s_keys) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[312], &__pyx_n_s_keys_py2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[313], &__pyx_n_s_kw1) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[314], &__pyx_n_s_kw2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[315], &__pyx_n_s_kwargs) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[316], &__pyx_n_s_le) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[317], &__pyx_n_s_len) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[318], &__pyx_n_s_length_hint) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[319], &__pyx_n_s_list) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[320], &__pyx_n_s_long) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[321], &__pyx_n_s_lshift) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[322], &__pyx_n_s_lstrip) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[323], &__pyx_n_s_lt) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[324], &__pyx_n_s_main) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[325], &__pyx_n_s_major) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[326], &__pyx_n_s_match) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[327], &__pyx_n_s_match_args) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[328], &__pyx_n_s_math) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[329], &__pyx_n_s_matmul) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[330], &__pyx_n_s_metaclass) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[331], &__pyx_n_s_minor) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[332], &__pyx_n_s_mod) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[333], &__pyx_n_s_module) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[334], &__pyx_n_s_modules) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[335], &__pyx_n_s_mro_entries) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[336], &__pyx_n_s_mul) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[337], &__pyx_n_s_multiwrapped) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[338], &__pyx_n_s_name) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[339], &__pyx_n_s_ne) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[340], &__pyx_n_s_neg) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[341], &__pyx_n_s_never_writeable) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[342], &__pyx_n_s_never_writeable_private) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[343], &__pyx_n_s_new) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[344], &__pyx_n_s_o) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[345], &__pyx_kp_s_o_Invalid_type_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[346], &__pyx_n_s_object) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[347], &__pyx_n_s_oldstyle_class) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[348], &__pyx_n_s_op) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[349], &__pyx_n_s_or) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[350], &__pyx_n_s_os) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[351], &__pyx_n_s_p) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[352], &__pyx_n_s_package) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[353], &__pyx_n_s_partial) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[354], &__pyx_n_s_pass_to_wrapped) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[355], &__pyx_n_s_pattern) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[356], &__pyx_n_s_pickle) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[357], &__pyx_n_s_platform) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[358], &__pyx_n_s_pop) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[359], &__pyx_n_s_popitem) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[360], &__pyx_n_s_pos) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[361], &__pyx_n_s_pow) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[362], &__pyx_n_s_prepare) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[363], &__pyx_n_s_private) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[364], &__pyx_n_s_protect) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[365], &__pyx_n_s_protected_rules_from_kwargs_loca) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[366], &__pyx_n_s_py2_function_attrs_rw) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[367], &__pyx_n_s_pydoc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[368], &__pyx_n_s_pyprotect_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[369], &__pyx_n_s_python_implementation) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[370], &__pyx_kp_s_python_visible_pxi) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[371], &__pyx_n_s_pyx_PickleError) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[372], &__pyx_n_s_pyx_checksum) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[373], &__pyx_n_s_pyx_result) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[374], &__pyx_n_s_pyx_state) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[375], &__pyx_n_s_pyx_type) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[376], &__pyx_n_s_pyx_unpickle_Frozen) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[377], &__pyx_n_s_pyx_unpickle_FrozenPrivacyDict) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[378], &__pyx_n_s_pyx_unpickle_FrozenPrivate) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[379], &__pyx_n_s_pyx_unpickle_FrozenProtected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[380], &__pyx_n_s_pyx_unpickle_PrivacyDict) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[381], &__pyx_n_s_pyx_unpickle_Private) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[382], &__pyx_n_s_pyx_unpickle_Protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[383], &__pyx_n_s_pyx_unpickle_Proxy) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[384], &__pyx_n_s_pyx_unpickle_Wrapped) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[385], &__pyx_n_s_pyx_unpickle___HiddenPartial) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[386], &__pyx_n_s_pyx_unpickle___ProtectionData) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[387], &__pyx_n_s_pyx_vtable) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[388], &__pyx_n_s_qualname) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[389], &__pyx_n_s_r) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[390], &__pyx_n_s_radd) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[391], &__pyx_n_s_rand) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[392], &__pyx_n_s_rdivmod) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[393], &__pyx_n_s_re) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[394], &__pyx_n_s_reduce) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[395], &__pyx_n_s_reduce_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[396], &__pyx_n_s_reduce_ex) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[397], &__pyx_n_s_remove) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[398], &__pyx_n_s_render_doc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[399], &__pyx_n_s_repr) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[400], &__pyx_n_s_ret) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[401], &__pyx_n_s_ret_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[402], &__pyx_n_s_return) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[403], &__pyx_n_s_reverse) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[404], &__pyx_n_s_rfloordiv) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[405], &__pyx_n_s_rl) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[406], &__pyx_n_s_rlshift) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[407], &__pyx_n_s_rmatmul) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[408], &__pyx_n_s_rmod) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[409], &__pyx_n_s_rmul) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[410], &__pyx_n_s_ro) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[411], &__pyx_n_s_ro_data) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[412], &__pyx_n_s_ro_method) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[413], &__pyx_n_s_ro_regex) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[414], &__pyx_n_s_ror) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[415], &__pyx_n_s_round) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[416], &__pyx_n_s_rpow) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[417], &__pyx_n_s_rrshift) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[418], &__pyx_n_s_rshift) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[419], &__pyx_n_s_rstrip) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[420], &__pyx_n_s_rsub) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[421], &__pyx_n_s_rtruediv) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[422], &__pyx_n_s_rules) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[423], &__pyx_n_s_rw) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[424], &__pyx_n_s_rw_regex) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[425], &__pyx_n_s_rxor) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[426], &__pyx_kp_s_s) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[427], &__pyx_kp_s_s_________0_1) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[428], &__pyx_n_s_same_class_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[429], &__pyx_n_s_self) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[430], &__pyx_n_s_send) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[431], &__pyx_n_s_set) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[432], &__pyx_n_s_set_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[433], &__pyx_n_s_set_name) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[434], &__pyx_n_s_setattr) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[435], &__pyx_n_s_setdefault) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[436], &__pyx_n_s_setitem) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[437], &__pyx_n_s_setstate) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[438], &__pyx_n_s_setstate_cython) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[439], &__pyx_n_s_slots) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[440], &__pyx_n_s_sort) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[441], &__pyx_n_s_spec) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[442], &__pyx_n_s_splitlines) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[443], &__pyx_n_s_startswith) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[444], &__pyx_n_s_state) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[445], &__pyx_n_s_str) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[446], &__pyx_n_s_str_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[447], &__pyx_kp_s_stringsource) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[448], &__pyx_n_s_sub) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[449], &__pyx_n_s_subclass) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[450], &__pyx_n_s_subclass_of_protected) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[451], &__pyx_n_s_subclasscheck) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[452], &__pyx_n_s_subclassof) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[453], &__pyx_n_s_super) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[454], &__pyx_n_s_sys) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[455], &__pyx_n_s_tb) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[456], &__pyx_n_s_test) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[457], &__pyx_n_s_testop) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[458], &__pyx_n_s_throw) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[459], &__pyx_n_s_truediv) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[460], &__pyx_n_s_trunc) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[461], &__pyx_n_s_tuple) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[462], &__pyx_n_s_type) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[463], &__pyx_n_s_types) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[464], &__pyx_n_s_unichr) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[465], &__pyx_n_s_unicode) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[466], &__pyx_n_s_union) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[467], &__pyx_n_s_update) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[468], &__pyx_n_s_use_setstate) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[469], &__pyx_n_s_v) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[470], &__pyx_n_s_val) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[471], &__pyx_n_s_value) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[472], &__pyx_n_s_values) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[473], &__pyx_n_s_values_py2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[474], &__pyx_n_s_version_info) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[475], &__pyx_n_s_viewitems) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[476], &__pyx_n_s_viewkeys) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[477], &__pyx_n_s_viewvalues) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[478], &__pyx_n_s_w) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[479], &__pyx_n_s_want_frozen) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[480], &__pyx_n_s_weakref) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[481], &__pyx_n_s_wrap) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[482], &__pyx_n_s_x) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[483], &__pyx_n_s_x_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[484], &__pyx_n_s_xor) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  #endif
  #if !CYTHON_USE_MODULE_STATE
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  #endif
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_6017409 = PyInt_FromLong(6017409L); if (unlikely(!__pyx_int_6017409)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_19817578 = PyInt_FromLong(19817578L); if (unlikely(!__pyx_int_19817578)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_50167005 = PyInt_FromLong(50167005L); if (unlikely(!__pyx_int_50167005)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_78131249 = PyInt_FromLong(78131249L); if (unlikely(!__pyx_int_78131249)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_115090883 = PyInt_FromLong(115090883L); if (unlikely(!__pyx_int_115090883)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_152356376 = PyInt_FromLong(152356376L); if (unlikely(!__pyx_int_152356376)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_155231502 = PyInt_FromLong(155231502L); if (unlikely(!__pyx_int_155231502)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_161740782 = PyInt_FromLong(161740782L); if (unlikely(!__pyx_int_161740782)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_180490543 = PyInt_FromLong(180490543L); if (unlikely(!__pyx_int_180490543)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_209109265 = PyInt_FromLong(209109265L); if (unlikely(!__pyx_int_209109265)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_210565405 = PyInt_FromLong(210565405L); if (unlikely(!__pyx_int_210565405)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_218598457 = PyInt_FromLong(218598457L); if (unlikely(!__pyx_int_218598457)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_247595846 = PyInt_FromLong(247595846L); if (unlikely(!__pyx_int_247595846)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_252507329 = PyInt_FromLong(252507329L); if (unlikely(!__pyx_int_252507329)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_int_262487005 = PyInt_FromLong(262487005L); if (unlikely(!__pyx_int_262487005)) __PYX_ERR(10, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  return 0;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_9pyprotect_9protected_builtin_module = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_immutable_types_set = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_builtins_ids = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_builtin_module_immutable_attributes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_frozen_error = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_overridden_always = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_pickle_attributes = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_special_attributes = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_always_delegated = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_always_frozen = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_attr_identifier = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_ro_private_attr = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_unmangled_private_attr = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_mangled_private_attr_classname_regex = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_mangled_private_attr_regex_fmt = Py_None; Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_m_block = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_m_numeric = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_m_compare = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_9pyprotect_9protected_m_safe = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___ProtectionData = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___ProtectionData_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___ProtectionData)) __PYX_ERR(1, 4, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___ProtectionData_spec, __pyx_ptype_9pyprotect_9protected___ProtectionData) < 0) __PYX_ERR(1, 4, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___ProtectionData = &__pyx_type_9pyprotect_9protected___ProtectionData;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___ProtectionData) < 0) __PYX_ERR(1, 4, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___ProtectionData->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected___ProtectionData) < 0) __PYX_ERR(1, 4, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected_Proxy = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_Proxy_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected_Proxy)) __PYX_ERR(2, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_Proxy_spec, __pyx_ptype_9pyprotect_9protected_Proxy) < 0) __PYX_ERR(2, 3, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_Proxy = &__pyx_type_9pyprotect_9protected_Proxy;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_Proxy) < 0) __PYX_ERR(2, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_Proxy->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected_Proxy->tp_dictoffset && __pyx_ptype_9pyprotect_9protected_Proxy->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected_Proxy->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(2, 3, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_5Proxy___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_5Proxy___init__.doc = __pyx_doc_9pyprotect_9protected_5Proxy___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_5Proxy___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Proxy, (PyObject *) __pyx_ptype_9pyprotect_9protected_Proxy) < 0) __PYX_ERR(2, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_Proxy) < 0) __PYX_ERR(2, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_Wrapped = &__pyx_vtable_9pyprotect_9protected_Wrapped;
  __pyx_vtable_9pyprotect_9protected_Wrapped.attr_hidden = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_attr_hidden;
  __pyx_vtable_9pyprotect_9protected_Wrapped.fif = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_fif;
  __pyx_vtable_9pyprotect_9protected_Wrapped.freeze = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_freeze;
  __pyx_vtable_9pyprotect_9protected_Wrapped.multiwrapped = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_multiwrapped;
  __pyx_vtable_9pyprotect_9protected_Wrapped.id_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_id_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.hash_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_hash_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.isinstance_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_isinstance_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.issubclass_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_issubclass_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.instanceof_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_instanceof_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.subclassof_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_subclassof_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.help_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_help_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.help_str_protected = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_help_str_protected;
  __pyx_vtable_9pyprotect_9protected_Wrapped.visible = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_visible;
  __pyx_vtable_9pyprotect_9protected_Wrapped.writeable = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_writeable;
  __pyx_vtable_9pyprotect_9protected_Wrapped.testop = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_testop;
  __pyx_vtable_9pyprotect_9protected_Wrapped.get_rules = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_get_rules;
  __pyx_vtable_9pyprotect_9protected_Wrapped.comparator = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_comparator;
  __pyx_vtable_9pyprotect_9protected_Wrapped.wrapped_getattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_getattr;
  __pyx_vtable_9pyprotect_9protected_Wrapped.wrapped_check_setattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_check_setattr;
  __pyx_vtable_9pyprotect_9protected_Wrapped.wrapped_check_delattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_check_delattr;
  __pyx_vtable_9pyprotect_9protected_Wrapped.wrapped_dir = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_7Wrapped_wrapped_dir;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_Wrapped = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_Wrapped_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_Wrapped)) __PYX_ERR(3, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_Wrapped_spec, __pyx_ptype_9pyprotect_9protected_Wrapped) < 0) __PYX_ERR(3, 3, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_Wrapped = &__pyx_type_9pyprotect_9protected_Wrapped;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_Wrapped->tp_base = __pyx_ptype_9pyprotect_9protected_Proxy;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_Wrapped) < 0) __PYX_ERR(3, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_Wrapped->tp_print = 0;
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(3, 3, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_7Wrapped___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_7Wrapped___init__.doc = __pyx_doc_9pyprotect_9protected_7Wrapped___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_7Wrapped___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_Wrapped, __pyx_vtabptr_9pyprotect_9protected_Wrapped) < 0) __PYX_ERR(3, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_Wrapped) < 0) __PYX_ERR(3, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Wrapped, (PyObject *) __pyx_ptype_9pyprotect_9protected_Wrapped) < 0) __PYX_ERR(3, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_Wrapped) < 0) __PYX_ERR(3, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_Frozen = &__pyx_vtable_9pyprotect_9protected_Frozen;
  __pyx_vtable_9pyprotect_9protected_Frozen.__pyx_base = *__pyx_vtabptr_9pyprotect_9protected_Wrapped;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_Frozen = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_Frozen_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_Frozen)) __PYX_ERR(3, 414, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_Frozen_spec, __pyx_ptype_9pyprotect_9protected_Frozen) < 0) __PYX_ERR(3, 414, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_Frozen = &__pyx_type_9pyprotect_9protected_Frozen;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_Frozen->tp_base = __pyx_ptype_9pyprotect_9protected_Wrapped;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_Frozen) < 0) __PYX_ERR(3, 414, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_Frozen->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected_Frozen->tp_dictoffset && __pyx_ptype_9pyprotect_9protected_Frozen->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected_Frozen->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_Frozen, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(3, 414, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_6Frozen___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_6Frozen___init__.doc = __pyx_doc_9pyprotect_9protected_6Frozen___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_6Frozen___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_Frozen, __pyx_vtabptr_9pyprotect_9protected_Frozen) < 0) __PYX_ERR(3, 414, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_Frozen) < 0) __PYX_ERR(3, 414, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Frozen, (PyObject *) __pyx_ptype_9pyprotect_9protected_Frozen) < 0) __PYX_ERR(3, 414, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_Frozen) < 0) __PYX_ERR(3, 414, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_PrivacyDict = &__pyx_vtable_9pyprotect_9protected_PrivacyDict;
  __pyx_vtable_9pyprotect_9protected_PrivacyDict.__pyx_base = *__pyx_vtabptr_9pyprotect_9protected_Wrapped;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_PrivacyDict = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_PrivacyDict_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_PrivacyDict)) __PYX_ERR(4, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_PrivacyDict_spec, __pyx_ptype_9pyprotect_9protected_PrivacyDict) < 0) __PYX_ERR(4, 3, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_PrivacyDict = &__pyx_type_9pyprotect_9protected_PrivacyDict;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_base = __pyx_ptype_9pyprotect_9protected_Wrapped;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_PrivacyDict) < 0) __PYX_ERR(4, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_print = 0;
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(4, 3, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_11PrivacyDict___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_11PrivacyDict___init__.doc = __pyx_doc_9pyprotect_9protected_11PrivacyDict___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_11PrivacyDict___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_PrivacyDict, __pyx_vtabptr_9pyprotect_9protected_PrivacyDict) < 0) __PYX_ERR(4, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_PrivacyDict) < 0) __PYX_ERR(4, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PrivacyDict, (PyObject *) __pyx_ptype_9pyprotect_9protected_PrivacyDict) < 0) __PYX_ERR(4, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_PrivacyDict) < 0) __PYX_ERR(4, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_FrozenPrivacyDict = &__pyx_vtable_9pyprotect_9protected_FrozenPrivacyDict;
  __pyx_vtable_9pyprotect_9protected_FrozenPrivacyDict.__pyx_base = *__pyx_vtabptr_9pyprotect_9protected_PrivacyDict;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_FrozenPrivacyDict_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict)) __PYX_ERR(4, 204, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_FrozenPrivacyDict_spec, __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict) < 0) __PYX_ERR(4, 204, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict = &__pyx_type_9pyprotect_9protected_FrozenPrivacyDict;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict->tp_base = __pyx_ptype_9pyprotect_9protected_PrivacyDict;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict) < 0) __PYX_ERR(4, 204, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict->tp_dictoffset && __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(4, 204, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_17FrozenPrivacyDict___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_17FrozenPrivacyDict___init__.doc = __pyx_doc_9pyprotect_9protected_17FrozenPrivacyDict___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_17FrozenPrivacyDict___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict, __pyx_vtabptr_9pyprotect_9protected_FrozenPrivacyDict) < 0) __PYX_ERR(4, 204, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict) < 0) __PYX_ERR(4, 204, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FrozenPrivacyDict, (PyObject *) __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict) < 0) __PYX_ERR(4, 204, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict) < 0) __PYX_ERR(4, 204, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_Private = &__pyx_vtable_9pyprotect_9protected_Private;
  __pyx_vtable_9pyprotect_9protected_Private.__pyx_base = *__pyx_vtabptr_9pyprotect_9protected_Wrapped;
  __pyx_vtable_9pyprotect_9protected_Private.__pyx_base.visible = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Private_visible;
  __pyx_vtable_9pyprotect_9protected_Private.__pyx_base.writeable = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_7Private_writeable;
  __pyx_vtable_9pyprotect_9protected_Private.private_visible = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *))__pyx_f_9pyprotect_9protected_7Private_private_visible;
  __pyx_vtable_9pyprotect_9protected_Private.private_writeable = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *))__pyx_f_9pyprotect_9protected_7Private_private_writeable;
  __pyx_vtable_9pyprotect_9protected_Private.private_getattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *))__pyx_f_9pyprotect_9protected_7Private_private_getattr;
  __pyx_vtable_9pyprotect_9protected_Private.private_check_setattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *, PyObject *))__pyx_f_9pyprotect_9protected_7Private_private_check_setattr;
  __pyx_vtable_9pyprotect_9protected_Private.private_check_delattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Private *, PyObject *))__pyx_f_9pyprotect_9protected_7Private_private_check_delattr;
  __pyx_vtable_9pyprotect_9protected_Private.private_dir = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Private *))__pyx_f_9pyprotect_9protected_7Private_private_dir;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_Private = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_Private_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_Private)) __PYX_ERR(7, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_Private_spec, __pyx_ptype_9pyprotect_9protected_Private) < 0) __PYX_ERR(7, 3, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_Private = &__pyx_type_9pyprotect_9protected_Private;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_Private->tp_base = __pyx_ptype_9pyprotect_9protected_Wrapped;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_Private) < 0) __PYX_ERR(7, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_Private->tp_print = 0;
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_Private, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(7, 3, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_7Private___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_7Private___init__.doc = __pyx_doc_9pyprotect_9protected_7Private___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_7Private___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_Private, __pyx_vtabptr_9pyprotect_9protected_Private) < 0) __PYX_ERR(7, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_Private) < 0) __PYX_ERR(7, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Private, (PyObject *) __pyx_ptype_9pyprotect_9protected_Private) < 0) __PYX_ERR(7, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_Private) < 0) __PYX_ERR(7, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_FrozenPrivate = &__pyx_vtable_9pyprotect_9protected_FrozenPrivate;
  __pyx_vtable_9pyprotect_9protected_FrozenPrivate.__pyx_base = *__pyx_vtabptr_9pyprotect_9protected_Private;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_Private); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_FrozenPrivate = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_FrozenPrivate_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_FrozenPrivate)) __PYX_ERR(7, 144, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_FrozenPrivate_spec, __pyx_ptype_9pyprotect_9protected_FrozenPrivate) < 0) __PYX_ERR(7, 144, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_FrozenPrivate = &__pyx_type_9pyprotect_9protected_FrozenPrivate;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_FrozenPrivate->tp_base = __pyx_ptype_9pyprotect_9protected_Private;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_FrozenPrivate) < 0) __PYX_ERR(7, 144, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_FrozenPrivate->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected_FrozenPrivate->tp_dictoffset && __pyx_ptype_9pyprotect_9protected_FrozenPrivate->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected_FrozenPrivate->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivate, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(7, 144, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_13FrozenPrivate___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_13FrozenPrivate___init__.doc = __pyx_doc_9pyprotect_9protected_13FrozenPrivate___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_13FrozenPrivate___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_FrozenPrivate, __pyx_vtabptr_9pyprotect_9protected_FrozenPrivate) < 0) __PYX_ERR(7, 144, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_FrozenPrivate) < 0) __PYX_ERR(7, 144, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FrozenPrivate, (PyObject *) __pyx_ptype_9pyprotect_9protected_FrozenPrivate) < 0) __PYX_ERR(7, 144, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_FrozenPrivate) < 0) __PYX_ERR(7, 144, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_Protected = &__pyx_vtable_9pyprotect_9protected_Protected;
  __pyx_vtable_9pyprotect_9protected_Protected.__pyx_base = *__pyx_vtabptr_9pyprotect_9protected_Private;
  __pyx_vtable_9pyprotect_9protected_Protected.__pyx_base.__pyx_base.visible = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_9Protected_visible;
  __pyx_vtable_9pyprotect_9protected_Protected.__pyx_base.__pyx_base.writeable = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *, PyObject *))__pyx_f_9pyprotect_9protected_9Protected_writeable;
  __pyx_vtable_9pyprotect_9protected_Protected.__pyx_base.__pyx_base.get_rules = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Wrapped *))__pyx_f_9pyprotect_9protected_9Protected_get_rules;
  __pyx_vtable_9pyprotect_9protected_Protected.process_rules = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *))__pyx_f_9pyprotect_9protected_9Protected_process_rules;
  __pyx_vtable_9pyprotect_9protected_Protected.build_cache = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *))__pyx_f_9pyprotect_9protected_9Protected_build_cache;
  __pyx_vtable_9pyprotect_9protected_Protected.check_1_op = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, PyObject *, struct __pyx_opt_args_9pyprotect_9protected_9Protected_check_1_op *__pyx_optional_args))__pyx_f_9pyprotect_9protected_9Protected_check_1_op;
  __pyx_vtable_9pyprotect_9protected_Protected.protected_visible = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_visible *__pyx_optional_args))__pyx_f_9pyprotect_9protected_9Protected_protected_visible;
  __pyx_vtable_9pyprotect_9protected_Protected.protected_writeable = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, struct __pyx_opt_args_9pyprotect_9protected_9Protected_protected_writeable *__pyx_optional_args))__pyx_f_9pyprotect_9protected_9Protected_protected_writeable;
  __pyx_vtable_9pyprotect_9protected_Protected.aclcheck = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, PyObject *))__pyx_f_9pyprotect_9protected_9Protected_aclcheck;
  __pyx_vtable_9pyprotect_9protected_Protected.protected_getattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *))__pyx_f_9pyprotect_9protected_9Protected_protected_getattr;
  __pyx_vtable_9pyprotect_9protected_Protected.protected_check_setattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *, PyObject *))__pyx_f_9pyprotect_9protected_9Protected_protected_check_setattr;
  __pyx_vtable_9pyprotect_9protected_Protected.protected_check_delattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *, PyObject *))__pyx_f_9pyprotect_9protected_9Protected_protected_check_delattr;
  __pyx_vtable_9pyprotect_9protected_Protected.protected_dir = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected_Protected *))__pyx_f_9pyprotect_9protected_9Protected_protected_dir;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_Private); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_Protected = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_Protected_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_Protected)) __PYX_ERR(8, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_Protected_spec, __pyx_ptype_9pyprotect_9protected_Protected) < 0) __PYX_ERR(8, 3, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_Protected = &__pyx_type_9pyprotect_9protected_Protected;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_Protected->tp_base = __pyx_ptype_9pyprotect_9protected_Private;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_Protected) < 0) __PYX_ERR(8, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_Protected->tp_print = 0;
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(8, 3, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_9Protected___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_9Protected___init__.doc = __pyx_doc_9pyprotect_9protected_9Protected___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_9Protected___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_Protected, __pyx_vtabptr_9pyprotect_9protected_Protected) < 0) __PYX_ERR(8, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_Protected) < 0) __PYX_ERR(8, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Protected_2, (PyObject *) __pyx_ptype_9pyprotect_9protected_Protected) < 0) __PYX_ERR(8, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_Protected) < 0) __PYX_ERR(8, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected_FrozenProtected = &__pyx_vtable_9pyprotect_9protected_FrozenProtected;
  __pyx_vtable_9pyprotect_9protected_FrozenProtected.__pyx_base = *__pyx_vtabptr_9pyprotect_9protected_Protected;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_9pyprotect_9protected_Protected); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_9pyprotect_9protected_FrozenProtected = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected_FrozenProtected_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_9pyprotect_9protected_FrozenProtected)) __PYX_ERR(8, 279, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected_FrozenProtected_spec, __pyx_ptype_9pyprotect_9protected_FrozenProtected) < 0) __PYX_ERR(8, 279, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected_FrozenProtected = &__pyx_type_9pyprotect_9protected_FrozenProtected;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_9pyprotect_9protected_FrozenProtected->tp_base = __pyx_ptype_9pyprotect_9protected_Protected;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected_FrozenProtected) < 0) __PYX_ERR(8, 279, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected_FrozenProtected->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected_FrozenProtected->tp_dictoffset && __pyx_ptype_9pyprotect_9protected_FrozenProtected->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected_FrozenProtected->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenProtected, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(8, 279, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_15FrozenProtected___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_15FrozenProtected___init__.doc = __pyx_doc_9pyprotect_9protected_15FrozenProtected___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_15FrozenProtected___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected_FrozenProtected, __pyx_vtabptr_9pyprotect_9protected_FrozenProtected) < 0) __PYX_ERR(8, 279, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected_FrozenProtected) < 0) __PYX_ERR(8, 279, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FrozenProtected, (PyObject *) __pyx_ptype_9pyprotect_9protected_FrozenProtected) < 0) __PYX_ERR(8, 279, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected_FrozenProtected) < 0) __PYX_ERR(8, 279, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9pyprotect_9protected___HiddenPartial = &__pyx_vtable_9pyprotect_9protected___HiddenPartial;
  __pyx_vtable_9pyprotect_9protected___HiddenPartial.wrapped_getattr = (PyObject *(*)(struct __pyx_obj_9pyprotect_9protected___HiddenPartial *, PyObject *))__pyx_f_9pyprotect_9protected_15__HiddenPartial_wrapped_getattr;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___HiddenPartial = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___HiddenPartial_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___HiddenPartial)) __PYX_ERR(9, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___HiddenPartial_spec, __pyx_ptype_9pyprotect_9protected___HiddenPartial) < 0) __PYX_ERR(9, 3, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___HiddenPartial = &__pyx_type_9pyprotect_9protected___HiddenPartial;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___HiddenPartial) < 0) __PYX_ERR(9, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___HiddenPartial->tp_print = 0;
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(9, 3, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_9pyprotect_9protected_15__HiddenPartial___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_9pyprotect_9protected_15__HiddenPartial___init__.doc = __pyx_doc_9pyprotect_9protected_15__HiddenPartial___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_9pyprotect_9protected_15__HiddenPartial___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9pyprotect_9protected___HiddenPartial, __pyx_vtabptr_9pyprotect_9protected___HiddenPartial) < 0) __PYX_ERR(9, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9pyprotect_9protected___HiddenPartial) < 0) __PYX_ERR(9, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_HiddenPartial, (PyObject *) __pyx_ptype_9pyprotect_9protected___HiddenPartial) < 0) __PYX_ERR(9, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9pyprotect_9protected___HiddenPartial) < 0) __PYX_ERR(9, 3, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct____iter___spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__)) __PYX_ERR(2, 49, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct____iter___spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__) < 0) __PYX_ERR(2, 49, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__ = &__pyx_type_9pyprotect_9protected___pyx_scope_struct____iter__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__) < 0) __PYX_ERR(2, 49, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct____iter__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator)) __PYX_ERR(3, 221, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator_spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator) < 0) __PYX_ERR(3, 221, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator = &__pyx_type_9pyprotect_9protected___pyx_scope_struct_1_comparator;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator) < 0) __PYX_ERR(3, 221, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_1_comparator->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys)) __PYX_ERR(4, 115, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys_spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys) < 0) __PYX_ERR(4, 115, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys = &__pyx_type_9pyprotect_9protected___pyx_scope_struct_2_keys;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys) < 0) __PYX_ERR(4, 115, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_2_keys->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items)) __PYX_ERR(4, 121, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items_spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items) < 0) __PYX_ERR(4, 121, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items = &__pyx_type_9pyprotect_9protected___pyx_scope_struct_3_items;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items) < 0) __PYX_ERR(4, 121, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_3_items->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values)) __PYX_ERR(4, 126, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values_spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values) < 0) __PYX_ERR(4, 126, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values = &__pyx_type_9pyprotect_9protected___pyx_scope_struct_4_values;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values) < 0) __PYX_ERR(4, 126, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_4_values->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys)) __PYX_ERR(4, 160, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys_spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys) < 0) __PYX_ERR(4, 160, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys = &__pyx_type_9pyprotect_9protected___pyx_scope_struct_5_iterkeys;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys) < 0) __PYX_ERR(4, 160, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_5_iterkeys->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems)) __PYX_ERR(4, 167, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems_spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems) < 0) __PYX_ERR(4, 167, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems = &__pyx_type_9pyprotect_9protected___pyx_scope_struct_6_iteritems;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems) < 0) __PYX_ERR(4, 167, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_6_iteritems->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues_spec, NULL); if (unlikely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues)) __PYX_ERR(4, 173, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues_spec, __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues) < 0) __PYX_ERR(4, 173, __pyx_L1_error)
  #else
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues = &__pyx_type_9pyprotect_9protected___pyx_scope_struct_7_itervalues;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues) < 0) __PYX_ERR(4, 173, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues->tp_dictoffset && __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9pyprotect_9protected___pyx_scope_struct_7_itervalues->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_spec, NULL); if (unlikely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self)) __PYX_ERR(6, 66, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self_spec, __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #else
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self = &__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self->tp_dictoffset && __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__rParen_to_py_4self->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_spec, NULL); if (unlikely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c)) __PYX_ERR(6, 66, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c_spec, __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #else
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c = &__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c->tp_dictoffset && __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__rParen_to_py_4self_1c->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_spec, NULL); if (unlikely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op)) __PYX_ERR(6, 66, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op_spec, __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #else
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op = &__pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op->tp_dictoffset && __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(11, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_protected(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_protected},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "protected",
      __pyx_k_Module_with_methods_to_wrap_an, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initprotected(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initprotected(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_protected(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_protected(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_protected(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'protected' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("protected", __pyx_methods, __pyx_k_Module_with_methods_to_wrap_an, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(10, 1, __pyx_L1_error)
  #elif CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely((add_module_result < 0))) __PYX_ERR(10, 1, __pyx_L1_error)
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(10, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(10, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(10, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(10, 1, __pyx_L1_error);
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_protected(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(10, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyprotect__protected) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(10, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyprotect.protected")) {
      if (unlikely((PyDict_SetItemString(modules, "pyprotect.protected", __pyx_m) < 0))) __PYX_ERR(10, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(10, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(10, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  #endif

  /* "imports.pxi":2
 * 
 * import sys             # <<<<<<<<<<<<<<
 * cdef bint PY2
 * cdef object builtin_module
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_sys, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_2) < 0) __PYX_ERR(12, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":5
 * cdef bint PY2
 * cdef object builtin_module
 * if sys.version_info.major > 2:             # <<<<<<<<<<<<<<
 *     PY2 = False
 *     builtin_module = sys.modules['builtins']
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_version_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_major); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_2, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(12, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "imports.pxi":6
 * cdef object builtin_module
 * if sys.version_info.major > 2:
 *     PY2 = False             # <<<<<<<<<<<<<<
 *     builtin_module = sys.modules['builtins']
 *     import collections.abc as CollectionsABC
 */
    __pyx_v_9pyprotect_9protected_PY2 = 0;

    /* "imports.pxi":7
 * if sys.version_info.major > 2:
 *     PY2 = False
 *     builtin_module = sys.modules['builtins']             # <<<<<<<<<<<<<<
 *     import collections.abc as CollectionsABC
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_modules); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_s_builtins); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_builtin_module);
    __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_builtin_module, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "imports.pxi":8
 *     PY2 = False
 *     builtin_module = sys.modules['builtins']
 *     import collections.abc as CollectionsABC             # <<<<<<<<<<<<<<
 * else:
 *     PY2 = True
 */
    __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_collections_abc, __pyx_tuple__45); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_CollectionsABC, __pyx_t_3) < 0) __PYX_ERR(12, 8, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "imports.pxi":5
 * cdef bint PY2
 * cdef object builtin_module
 * if sys.version_info.major > 2:             # <<<<<<<<<<<<<<
 *     PY2 = False
 *     builtin_module = sys.modules['builtins']
 */
    goto __pyx_L2;
  }

  /* "imports.pxi":10
 *     import collections.abc as CollectionsABC
 * else:
 *     PY2 = True             # <<<<<<<<<<<<<<
 *     builtin_module = sys.modules['__builtin__']
 *     import collections as CollectionsABC
 */
  /*else*/ {
    __pyx_v_9pyprotect_9protected_PY2 = 1;

    /* "imports.pxi":11
 * else:
 *     PY2 = True
 *     builtin_module = sys.modules['__builtin__']             # <<<<<<<<<<<<<<
 *     import collections as CollectionsABC
 * import platform
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_modules); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_s_builtin); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_builtin_module);
    __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_builtin_module, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "imports.pxi":12
 *     PY2 = True
 *     builtin_module = sys.modules['__builtin__']
 *     import collections as CollectionsABC             # <<<<<<<<<<<<<<
 * import platform
 * PYPY = (platform.python_implementation() == 'PyPy')
 */
    __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_collections, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_CollectionsABC, __pyx_t_3) < 0) __PYX_ERR(12, 12, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L2:;

  /* "imports.pxi":13
 *     builtin_module = sys.modules['__builtin__']
 *     import collections as CollectionsABC
 * import platform             # <<<<<<<<<<<<<<
 * PYPY = (platform.python_implementation() == 'PyPy')
 * del platform
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_platform, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_platform, __pyx_t_3) < 0) __PYX_ERR(12, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "imports.pxi":14
 *     import collections as CollectionsABC
 * import platform
 * PYPY = (platform.python_implementation() == 'PyPy')             # <<<<<<<<<<<<<<
 * del platform
 * import os
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_platform); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_python_implementation); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_n_s_PyPy, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PYPY, __pyx_t_2) < 0) __PYX_ERR(12, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":15
 * import platform
 * PYPY = (platform.python_implementation() == 'PyPy')
 * del platform             # <<<<<<<<<<<<<<
 * import os
 * import re
 */
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_platform) < 0) __PYX_ERR(12, 15, __pyx_L1_error)

  /* "imports.pxi":16
 * PYPY = (platform.python_implementation() == 'PyPy')
 * del platform
 * import os             # <<<<<<<<<<<<<<
 * import re
 * import types
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_os, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_2) < 0) __PYX_ERR(12, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":17
 * del platform
 * import os
 * import re             # <<<<<<<<<<<<<<
 * import types
 * import functools
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_re, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_2) < 0) __PYX_ERR(12, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":18
 * import os
 * import re
 * import types             # <<<<<<<<<<<<<<
 * import functools
 * import pydoc
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_types, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_types, __pyx_t_2) < 0) __PYX_ERR(12, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":19
 * import re
 * import types
 * import functools             # <<<<<<<<<<<<<<
 * import pydoc
 * import math
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_functools, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_functools, __pyx_t_2) < 0) __PYX_ERR(12, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":20
 * import types
 * import functools
 * import pydoc             # <<<<<<<<<<<<<<
 * import math
 * if PYPY and PY2:
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_pydoc, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pydoc, __pyx_t_2) < 0) __PYX_ERR(12, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":21
 * import functools
 * import pydoc
 * import math             # <<<<<<<<<<<<<<
 * if PYPY and PY2:
 *     int = long
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_math, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_2) < 0) __PYX_ERR(12, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "imports.pxi":22
 * import pydoc
 * import math
 * if PYPY and PY2:             # <<<<<<<<<<<<<<
 *     int = long
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PYPY); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(12, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __pyx_v_9pyprotect_9protected_PY2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "imports.pxi":23
 * import math
 * if PYPY and PY2:
 *     int = long             # <<<<<<<<<<<<<<
 */
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_int, ((PyObject *)(&PyLong_Type))) < 0) __PYX_ERR(12, 23, __pyx_L1_error)

    /* "imports.pxi":22
 * import pydoc
 * import math
 * if PYPY and PY2:             # <<<<<<<<<<<<<<
 *     int = long
 */
  }

  /* "python_visible.pxi":6
 * # ------------------------------------------------------------------------
 * 
 * def attribute_protected():             # <<<<<<<<<<<<<<
 *     '''
 *     attribute_protected() -> str: name of special attribute in Wrapped objects
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_1attribute_protected, 0, __pyx_n_s_attribute_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_attribute_protected, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":12
 *     return PROT_ATTR_NAME
 * 
 * def id_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     id_protected(o: object) -> int:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_3id_protected, 0, __pyx_n_s_id_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_id_protected, __pyx_t_3) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":22
 * 
 * 
 * def hash_protected(o: object) -> int:             # <<<<<<<<<<<<<<
 *     '''
 *     hash_protected(o: object) -> int:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5hash_protected, 0, __pyx_n_s_hash_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hash_protected, __pyx_t_2) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":32
 * 
 * 
 * def isinstance_protected(o: object, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isinstance_protected(o: object, c: type) -> bool:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_c, __pyx_n_s_type) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7isinstance_protected, 0, __pyx_n_s_isinstance_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isinstance_protected, __pyx_t_3) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":43
 * 
 * 
 * def issubclass_protected(o: type, c: type) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     issubclass_protected(o: object, c: type) -> bool:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_type) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_c, __pyx_n_s_type) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_9issubclass_protected, 0, __pyx_n_s_issubclass_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_issubclass_protected, __pyx_t_2) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":54
 * 
 * 
 * def instance_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     instance_of_protected(x: object, w: object) -> bool
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_x, __pyx_n_s_object) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_w, __pyx_n_s_object) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11instance_of_protected, 0, __pyx_n_s_instance_of_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_instance_of_protected, __pyx_t_3) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":66
 * 
 * 
 * def subclass_of_protected(x: object, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     subclass_of_protected(x: object, w: object) -> bool
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_x, __pyx_n_s_object) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_w, __pyx_n_s_object) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_13subclass_of_protected, 0, __pyx_n_s_subclass_of_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_subclass_of_protected, __pyx_t_2) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":78
 * 
 * 
 * def same_class_protected(c: type, w: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     same_class_protected(c: type, w: object) -> bool
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_c, __pyx_n_s_type) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_w, __pyx_n_s_object) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_15same_class_protected, 0, __pyx_n_s_same_class_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_same_class_protected, __pyx_t_3) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":90
 * 
 * 
 * def help_protected(o: object) -> None:             # <<<<<<<<<<<<<<
 *     '''
 *     help_protected(o: object) -> None
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_17help_protected, 0, __pyx_n_s_help_protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_help_protected, __pyx_t_2) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":100
 * 
 * 
 * def contains(p: object, o: object):             # <<<<<<<<<<<<<<
 *     '''
 *     contains(p: object, o: object): bool: whether 'p' wraps 'o'
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_p, __pyx_n_s_object) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_19contains, 0, __pyx_n_s_contains_2, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_contains_2, __pyx_t_3) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":114
 * 
 * 
 * def isimmutable(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isimmutable(o: object) -> bool: 'o' is KNOWN to be immutable
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_21isimmutable, 0, __pyx_n_s_isimmutable, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isimmutable, __pyx_t_2) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":138
 * 
 * 
 * def iswrapped(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     iswrapped(o: object) -> bool:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_23iswrapped, 0, __pyx_n_s_iswrapped, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_iswrapped, __pyx_t_3) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":146
 * 
 * 
 * def isfrozen(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isfrozen(o: object) -> bool: 'o' was created using freeze()
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_25isfrozen, 0, __pyx_n_s_isfrozen, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isfrozen, __pyx_t_2) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":155
 * 
 * 
 * def isprivate(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprivate(o: object) -> bool: 'o' was created using private()
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_27isprivate, 0, __pyx_n_s_isprivate, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isprivate, __pyx_t_3) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":165
 * 
 * 
 * def isprotected(o: object) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isprotected(o: object) -> bool: 'o' was created using protect()
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_29isprotected, 0, __pyx_n_s_isprotected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isprotected, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":175
 * 
 * 
 * def isreadonly(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isreadonly(o: object, a: str) -> bool:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_a, __pyx_n_s_str) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_31isreadonly, 0, __pyx_n_s_isreadonly, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isreadonly, __pyx_t_3) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":195
 * 
 * 
 * def isvisible(o: object, a: str) -> bool:             # <<<<<<<<<<<<<<
 *     '''
 *     isvisible(o: object, a: str) -> bool:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_a, __pyx_n_s_str) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_33isvisible, 0, __pyx_n_s_isvisible, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isvisible, __pyx_t_2) < 0) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":216
 * 
 * 
 * def wrap(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     wrap(o: object) -> object:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_object) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_35wrap, 0, __pyx_n_s_wrap, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_wrap, __pyx_t_3) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":239
 * 
 * 
 * def freeze(o: object) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     freeze(o: object) -> object:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_object) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_37freeze, 0, __pyx_n_s_freeze, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_freeze, __pyx_t_2) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":261
 * 
 * 
 * def private(o: object, frozen: bool = False) -> object:             # <<<<<<<<<<<<<<
 *     '''
 *     private(o: object, frozen: bool = False) -> object:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 261, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_frozen, __pyx_n_s_bool) < 0) __PYX_ERR(0, 261, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_object) < 0) __PYX_ERR(0, 261, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_39private, 0, __pyx_n_s_private, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__73);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_private, __pyx_t_3) < 0) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "python_visible.pxi":294
 * 
 * 
 * def protect(             # <<<<<<<<<<<<<<
 *     o: object,
 *     frozen: bool = False, dynamic: bool = True,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_o, __pyx_n_s_object) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_frozen, __pyx_n_s_bool) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dynamic, __pyx_n_s_bool) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_hide_private, __pyx_n_s_bool) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ro_data, __pyx_n_s_bool) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ro_method, __pyx_n_s_bool) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_41protect, 0, __pyx_n_s_protect, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults), 3)) __PYX_ERR(0, 294, __pyx_L1_error)

  /* "python_visible.pxi":299
 *     hide_private: bool = False,
 *     ro_data: bool = False, ro_method: bool = True,
 *     ro=[], rw=[], hide=[],             # <<<<<<<<<<<<<<
 * ):
 *     '''
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_2)->__pyx_arg_ro = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_2)->__pyx_arg_rw = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_2)->__pyx_arg_hide = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_9pyprotect_9protected_76__defaults__);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_protect, __pyx_t_2) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":392
 * # ------------------------------------------------------------------------
 * 
 * def never_writeable():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable() -> set(str): Attributes that are never writeable
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_43never_writeable, 0, __pyx_n_s_never_writeable, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_never_writeable, __pyx_t_2) < 0) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":399
 *     return overridden_always
 * 
 * def never_writeable_private():             # <<<<<<<<<<<<<<
 *     '''
 *     never_writeable_private() -> set(str): Attributes that are never
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_45never_writeable_private, 0, __pyx_n_s_never_writeable_private, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_never_writeable_private, __pyx_t_2) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":409
 *     ))
 * 
 * def hidden_pickle_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     hidden_pickle_attributes() -> set(str): Attributes that are never
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_47hidden_pickle_attributes, 0, __pyx_n_s_hidden_pickle_attributes, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hidden_pickle_attributes, __pyx_t_2) < 0) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":416
 *     return pickle_attributes
 * 
 * def always_delegated_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     always_delegated_attributes() -> set(str): Attributes that are
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_49always_delegated_attributes, 0, __pyx_n_s_always_delegated_attributes, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_always_delegated_attributes, __pyx_t_2) < 0) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":423
 *     return always_delegated
 * 
 * def immutable_builtin_attributes():             # <<<<<<<<<<<<<<
 *     '''
 *     immutable_builtin_attributes() -> frozenset(str)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_51immutable_builtin_attributes, 0, __pyx_n_s_immutable_builtin_attributes, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_immutable_builtin_attributes, __pyx_t_2) < 0) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":430
 *     return builtin_module_immutable_attributes
 * 
 * __all__ = [             # <<<<<<<<<<<<<<
 *     'contains', 'freeze', 'id_protected', 'immutable_builtin_attributes',
 *     'isfrozen', 'isimmutable', 'isinstance_protected', 'isprivate',
 */
  __pyx_t_2 = PyList_New(27); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_contains_2);
  __Pyx_GIVEREF(__pyx_n_s_contains_2);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_contains_2);
  __Pyx_INCREF(__pyx_n_s_freeze);
  __Pyx_GIVEREF(__pyx_n_s_freeze);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_freeze);
  __Pyx_INCREF(__pyx_n_s_id_protected);
  __Pyx_GIVEREF(__pyx_n_s_id_protected);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_id_protected);
  __Pyx_INCREF(__pyx_n_s_immutable_builtin_attributes);
  __Pyx_GIVEREF(__pyx_n_s_immutable_builtin_attributes);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_immutable_builtin_attributes);
  __Pyx_INCREF(__pyx_n_s_isfrozen);
  __Pyx_GIVEREF(__pyx_n_s_isfrozen);
  PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_isfrozen);
  __Pyx_INCREF(__pyx_n_s_isimmutable);
  __Pyx_GIVEREF(__pyx_n_s_isimmutable);
  PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_isimmutable);
  __Pyx_INCREF(__pyx_n_s_isinstance_protected);
  __Pyx_GIVEREF(__pyx_n_s_isinstance_protected);
  PyList_SET_ITEM(__pyx_t_2, 6, __pyx_n_s_isinstance_protected);
  __Pyx_INCREF(__pyx_n_s_isprivate);
  __Pyx_GIVEREF(__pyx_n_s_isprivate);
  PyList_SET_ITEM(__pyx_t_2, 7, __pyx_n_s_isprivate);
  __Pyx_INCREF(__pyx_n_s_isprotected);
  __Pyx_GIVEREF(__pyx_n_s_isprotected);
  PyList_SET_ITEM(__pyx_t_2, 8, __pyx_n_s_isprotected);
  __Pyx_INCREF(__pyx_n_s_isreadonly);
  __Pyx_GIVEREF(__pyx_n_s_isreadonly);
  PyList_SET_ITEM(__pyx_t_2, 9, __pyx_n_s_isreadonly);
  __Pyx_INCREF(__pyx_n_s_iswrapped);
  __Pyx_GIVEREF(__pyx_n_s_iswrapped);
  PyList_SET_ITEM(__pyx_t_2, 10, __pyx_n_s_iswrapped);
  __Pyx_INCREF(__pyx_n_s_private);
  __Pyx_GIVEREF(__pyx_n_s_private);
  PyList_SET_ITEM(__pyx_t_2, 11, __pyx_n_s_private);
  __Pyx_INCREF(__pyx_n_s_protect);
  __Pyx_GIVEREF(__pyx_n_s_protect);
  PyList_SET_ITEM(__pyx_t_2, 12, __pyx_n_s_protect);
  __Pyx_INCREF(__pyx_n_s_wrap);
  __Pyx_GIVEREF(__pyx_n_s_wrap);
  PyList_SET_ITEM(__pyx_t_2, 13, __pyx_n_s_wrap);
  __Pyx_INCREF(__pyx_n_s_help_protected);
  __Pyx_GIVEREF(__pyx_n_s_help_protected);
  PyList_SET_ITEM(__pyx_t_2, 14, __pyx_n_s_help_protected);
  __Pyx_INCREF(__pyx_n_s_attribute_protected);
  __Pyx_GIVEREF(__pyx_n_s_attribute_protected);
  PyList_SET_ITEM(__pyx_t_2, 15, __pyx_n_s_attribute_protected);
  __Pyx_INCREF(__pyx_n_s_isvisible);
  __Pyx_GIVEREF(__pyx_n_s_isvisible);
  PyList_SET_ITEM(__pyx_t_2, 16, __pyx_n_s_isvisible);
  __Pyx_INCREF(__pyx_n_s_file);
  __Pyx_GIVEREF(__pyx_n_s_file);
  PyList_SET_ITEM(__pyx_t_2, 17, __pyx_n_s_file);
  __Pyx_INCREF(__pyx_n_s_never_writeable);
  __Pyx_GIVEREF(__pyx_n_s_never_writeable);
  PyList_SET_ITEM(__pyx_t_2, 18, __pyx_n_s_never_writeable);
  __Pyx_INCREF(__pyx_n_s_never_writeable_private);
  __Pyx_GIVEREF(__pyx_n_s_never_writeable_private);
  PyList_SET_ITEM(__pyx_t_2, 19, __pyx_n_s_never_writeable_private);
  __Pyx_INCREF(__pyx_n_s_hidden_pickle_attributes);
  __Pyx_GIVEREF(__pyx_n_s_hidden_pickle_attributes);
  PyList_SET_ITEM(__pyx_t_2, 20, __pyx_n_s_hidden_pickle_attributes);
  __Pyx_INCREF(__pyx_n_s_always_delegated_attributes);
  __Pyx_GIVEREF(__pyx_n_s_always_delegated_attributes);
  PyList_SET_ITEM(__pyx_t_2, 21, __pyx_n_s_always_delegated_attributes);
  __Pyx_INCREF(__pyx_n_s_ProtectionError);
  __Pyx_GIVEREF(__pyx_n_s_ProtectionError);
  PyList_SET_ITEM(__pyx_t_2, 22, __pyx_n_s_ProtectionError);
  __Pyx_INCREF(__pyx_n_s_issubclass_protected);
  __Pyx_GIVEREF(__pyx_n_s_issubclass_protected);
  PyList_SET_ITEM(__pyx_t_2, 23, __pyx_n_s_issubclass_protected);
  __Pyx_INCREF(__pyx_n_s_instance_of_protected);
  __Pyx_GIVEREF(__pyx_n_s_instance_of_protected);
  PyList_SET_ITEM(__pyx_t_2, 24, __pyx_n_s_instance_of_protected);
  __Pyx_INCREF(__pyx_n_s_subclass_of_protected);
  __Pyx_GIVEREF(__pyx_n_s_subclass_of_protected);
  PyList_SET_ITEM(__pyx_t_2, 25, __pyx_n_s_subclass_of_protected);
  __Pyx_INCREF(__pyx_n_s_same_class_protected);
  __Pyx_GIVEREF(__pyx_n_s_same_class_protected);
  PyList_SET_ITEM(__pyx_t_2, 26, __pyx_n_s_same_class_protected);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_2) < 0) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":444
 * 
 * 
 * def __dir__():             # <<<<<<<<<<<<<<
 *     return __all__
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_53__dir__, 0, __pyx_n_s_dir, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dir, __pyx_t_2) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "python_visible.pxi":447
 *     return __all__
 * 
 * class ProtectionError(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  __Pyx_GIVEREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_3 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_CalculateMetaclass(NULL, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_6, __pyx_t_3, __pyx_n_s_ProtectionError, __pyx_n_s_ProtectionError, (PyObject *) NULL, __pyx_n_s_pyprotect_protected, (PyObject *) NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 447, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_6, __pyx_n_s_ProtectionError, __pyx_t_3, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ProtectionError, __pyx_t_2) < 0) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "global_cdefs.pxi":7
 * # The module only uses PROT_ATTR_NAME, never '_Protected_____'
 * # PROT_ATTR_NAME is set ONLY in get_protected_attr_name()
 * cdef str PROT_ATTR_NAME = get_protected_attr_name()             # <<<<<<<<<<<<<<
 * cdef Exception frozen_error = ProtectionError('Object is read-only')
 * (
 */
  __pyx_t_3 = __pyx_f_9pyprotect_9protected_get_protected_attr_name(); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_3))) __PYX_ERR(13, 7, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_PROT_ATTR_NAME);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_PROT_ATTR_NAME, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "global_cdefs.pxi":8
 * # PROT_ATTR_NAME is set ONLY in get_protected_attr_name()
 * cdef str PROT_ATTR_NAME = get_protected_attr_name()
 * cdef Exception frozen_error = ProtectionError('Object is read-only')             # <<<<<<<<<<<<<<
 * (
 *     immutable_types_set,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ProtectionError); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__82, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(__Pyx_PyException_Check(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("Exception", __pyx_t_6))) __PYX_ERR(13, 8, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_frozen_error);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_frozen_error, ((PyObject*)__pyx_t_6));
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "global_cdefs.pxi":13
 *     builtin_module_immutable_attributes,
 *     builtins_ids
 * ) = get_immutables()             # <<<<<<<<<<<<<<
 * cimport cython
 * from cpython.object cimport (
 */
  __pyx_t_6 = __pyx_f_9pyprotect_9protected_get_immutables(); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
    PyObject* sequence = __pyx_t_6;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(13, 10, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_8 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(13, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
    index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_7);
    index = 2; __pyx_t_2 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(13, 10, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(13, 10, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }

  /* "global_cdefs.pxi":10
 * cdef Exception frozen_error = ProtectionError('Object is read-only')
 * (
 *     immutable_types_set,             # <<<<<<<<<<<<<<
 *     builtin_module_immutable_attributes,
 *     builtins_ids
 */
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_3))) __PYX_ERR(13, 10, __pyx_L1_error)
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_7))) __PYX_ERR(13, 10, __pyx_L1_error)
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_2))) __PYX_ERR(13, 10, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_immutable_types_set);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_immutable_types_set, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_builtin_module_immutable_attributes);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_builtin_module_immutable_attributes, ((PyObject*)__pyx_t_7));
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_builtins_ids);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_builtins_ids, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":18
 *     Py_LT, Py_EQ, Py_GT, Py_LE, Py_NE, Py_GE,
 * )
 * cdef object overridden_always = frozenset([             # <<<<<<<<<<<<<<
 *     '__getattribute__', '__delattr__', '__setattr__',
 *     '__hash__',
 */
  __pyx_t_6 = __Pyx_PyFrozenSet_New(__pyx_tuple__83); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_overridden_always);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_overridden_always, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "global_cdefs.pxi":22
 *     '__hash__',
 * ])
 * cdef object pickle_attributes = frozenset([             # <<<<<<<<<<<<<<
 *     '__reduce__', '__reduce_ex__',
 *     '__getsate__', '__setstate__',
 */
  __pyx_t_6 = __Pyx_PyFrozenSet_New(__pyx_tuple__84); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_pickle_attributes);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_pickle_attributes, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "global_cdefs.pxi":26
 *     '__getsate__', '__setstate__',
 * ])
 * cdef object special_attributes = frozenset([             # <<<<<<<<<<<<<<
 *     PROT_ATTR_NAME,
 * ])
 */
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_9pyprotect_9protected_PROT_ATTR_NAME);
  __Pyx_GIVEREF(__pyx_v_9pyprotect_9protected_PROT_ATTR_NAME);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_9pyprotect_9protected_PROT_ATTR_NAME);
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_special_attributes);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_special_attributes, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":29
 *     PROT_ATTR_NAME,
 * ])
 * cdef object always_delegated = frozenset([             # <<<<<<<<<<<<<<
 *     '__doc__',
 *     '__weakref__',
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_tuple__85); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_always_delegated);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_always_delegated, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":34
 *     '__package__',
 * ])
 * cdef object always_frozen = frozenset([             # <<<<<<<<<<<<<<
 *     '__dict__', '__slots__', '__class__',
 *     '__module__',
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_tuple__86); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_always_frozen);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_always_frozen, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":41
 * # Use compiled regex - no function call, no str operations
 * # Python 2 str does not have isidentifier() method
 * cdef object attr_identifier = re.compile('^[_a-zA-Z][a-zA-Z0-9_]*$')             # <<<<<<<<<<<<<<
 * # ro_private_attr: Start with single _, ending n non-underscore
 * cdef object ro_private_attr = re.compile('^_[^_].*?(?<!_)$')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__87, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_attr_identifier);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_attr_identifier, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":43
 * cdef object attr_identifier = re.compile('^[_a-zA-Z][a-zA-Z0-9_]*$')
 * # ro_private_attr: Start with single _, ending n non-underscore
 * cdef object ro_private_attr = re.compile('^_[^_].*?(?<!_)$')             # <<<<<<<<<<<<<<
 * # unmangled_private_attr: Start with double _, end in non-underscore or single _
 * cdef object unmangled_private_attr = re.compile('^__[^_].*?[^_][_]{0,1}$')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__89, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_ro_private_attr);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_ro_private_attr, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":45
 * cdef object ro_private_attr = re.compile('^_[^_].*?(?<!_)$')
 * # unmangled_private_attr: Start with double _, end in non-underscore or single _
 * cdef object unmangled_private_attr = re.compile('^__[^_].*?[^_][_]{0,1}$')             # <<<<<<<<<<<<<<
 * # PY2 'Old style' CLASSES (not INSTANCES of such classes) do not have
 * # __class__ attribute, so Private wrappers around such classes will hide
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_6)) __PYX_ERR(13, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__90, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_unmangled_private_attr);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_unmangled_private_attr, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":49
 * # __class__ attribute, so Private wrappers around such classes will hide
 * # ALL similar looking attributes
 * cdef object mangled_private_attr_classname_regex = '[a-zA-Z][a-zA-Z0-9]*'             # <<<<<<<<<<<<<<
 * cdef object mangled_private_attr_regex_fmt = '^_%s__[^_](.*?[^_]|)[_]{0,1}$'
 * 
 */
  __Pyx_INCREF(__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2);
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_mangled_private_attr_classname_regex);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_mangled_private_attr_classname_regex, __pyx_kp_s_a_zA_Z_a_zA_Z0_9_2);
  __Pyx_GIVEREF(__pyx_kp_s_a_zA_Z_a_zA_Z0_9_2);

  /* "global_cdefs.pxi":50
 * # ALL similar looking attributes
 * cdef object mangled_private_attr_classname_regex = '[a-zA-Z][a-zA-Z0-9]*'
 * cdef object mangled_private_attr_regex_fmt = '^_%s__[^_](.*?[^_]|)[_]{0,1}$'             # <<<<<<<<<<<<<<
 * 
 * # ------------------------------------------------------------------------
 */
  __Pyx_INCREF(__pyx_kp_s_s_________0_1);
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_mangled_private_attr_regex_fmt);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_mangled_private_attr_regex_fmt, __pyx_kp_s_s_________0_1);
  __Pyx_GIVEREF(__pyx_kp_s_s_________0_1);

  /* "global_cdefs.pxi":60
 * 
 * # m_block used in Wrapped.wrapped_getattr and Protected.protected_getattr
 * cdef set m_block = set([             # <<<<<<<<<<<<<<
 *     # If MutableMapping:
 *     '__setitem__', '__delitem__',
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_s_setitem) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_delitem) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_iadd) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_imul) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_isub) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_imatmul) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_itruediv) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ifloordiv) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_imod) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ipow) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ilshift) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_irshift) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_iand) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ior) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ixor) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_set_2) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_delete) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_add_2) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_append) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_clear) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_discard) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_popitem) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_insert) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_pop) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_remove) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_reverse) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_setdefault) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_sort) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_update) < 0) __PYX_ERR(13, 60, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_m_block);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_m_block, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":75
 * #
 * 
 * cdef set m_numeric = set([             # <<<<<<<<<<<<<<
 *     # Emulating numeric types - return immutable
 *     '__add__', '__mul__', '__sub__', '__matmul__',
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_s_add) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_mul) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_sub) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_matmul) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_truediv) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_floordiv) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_mod) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_divmod) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_pow) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_lshift) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rshift) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_and) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_or) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_xor) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_radd) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rmul) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rsub) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rmatmul) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rtruediv) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rfloordiv) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rmod) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rdivmod) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rpow) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rlshift) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rrshift) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rand) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ror) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_rxor) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_neg) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_pos) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_abs) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_invert) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_complex_2) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_int_2) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_float_2) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_index) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_round) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_trunc) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_floor_2) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ceil_2) < 0) __PYX_ERR(13, 75, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_m_numeric);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_m_numeric, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":91
 * 
 * # m_compare not used anywhere
 * cdef set m_compare = set([             # <<<<<<<<<<<<<<
 *     # Comparisons - non-mutating, returning immutable bool
 *     # These are automatically implemented by Cython because we
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_s_lt) < 0) __PYX_ERR(13, 91, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_le) < 0) __PYX_ERR(13, 91, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_eq) < 0) __PYX_ERR(13, 91, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ne) < 0) __PYX_ERR(13, 91, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_gt) < 0) __PYX_ERR(13, 91, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_ge) < 0) __PYX_ERR(13, 91, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_cmp) < 0) __PYX_ERR(13, 91, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_m_compare);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_m_compare, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":102
 * # m_safe not used anywhere
 * # m_safe definitely do not mutate. If present, pass to wrapped
 * cdef set m_safe = set([             # <<<<<<<<<<<<<<
 *     # Representations - return immutable
 *     '__format__',
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_s_format) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_bool_2) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_contains) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_len) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_length_hint) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_instancecheck) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_subclasscheck) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_init_subclass) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_set_name) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_prepare) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_send) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_throw) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_close) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_enter) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_exit) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_aenter) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_aexit) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_s_match_args) < 0) __PYX_ERR(13, 102, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_9pyprotect_9protected_m_safe);
  __Pyx_DECREF_SET(__pyx_v_9pyprotect_9protected_m_safe, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "global_cdefs.pxi":124
 * ])
 * # These attributes of FunctionType are writable only in PY2
 * py2_function_attrs_rw = frozenset([             # <<<<<<<<<<<<<<
 *     '__doc__', '__name__', '__module__',
 *     '__defaults__', '__code__', '__dict__',
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_tuple__91); if (unlikely(!__pyx_t_2)) __PYX_ERR(13, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py2_function_attrs_rw, __pyx_t_2) < 0) __PYX_ERR(13, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ProtectionData.pxi":107
 *         raise ProtectionError('Object is read-only')
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return list(self.attributes_map.keys())
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_16__ProtectionData_9__dir__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProtectionData___dir, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected___ProtectionData->tp_dict, __pyx_n_s_dir, __pyx_t_2) < 0) __PYX_ERR(1, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected___ProtectionData);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_16__ProtectionData_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProtectionData___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected___ProtectionData->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected___ProtectionData);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___ProtectionData, (type(self), 0xfa53bdd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___ProtectionData__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_16__ProtectionData_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ProtectionData___setstate_cyth, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected___ProtectionData->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected___ProtectionData);

  /* "Proxy.pxi":31
 *         return str(self.pvt_o)
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(self.pvt_o)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_7__bytes__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___bytes, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_bytes_2, __pyx_t_2) < 0) __PYX_ERR(2, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":56
 * 
 *     # Representations - return immutable
 *     def __format__(self, val):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__format__(val)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_14__format__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___format, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_format, __pyx_t_2) < 0) __PYX_ERR(2, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":76
 *         return self.pvt_o.__len__()
 * 
 *     def __length_hint__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__length_hint__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_24__length_hint__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___length_hint, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_length_hint, __pyx_t_2) < 0) __PYX_ERR(2, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":92
 *         return self.pvt_o.__invert__()
 * 
 *     def __complex__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__complex__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_34__complex__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___complex, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_complex_2, __pyx_t_2) < 0) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":104
 *         return self.pvt_o.__index__()
 * 
 *     def __round__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__round__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_42__round__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___round, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_round, __pyx_t_2) < 0) __PYX_ERR(2, 104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":107
 *         return self.pvt_o.__round__()
 * 
 *     def __trunc__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__trunc__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_44__trunc__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___trunc, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_trunc, __pyx_t_2) < 0) __PYX_ERR(2, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":110
 *         return self.pvt_o.__trunc__()
 * 
 *     def __floor__(self):             # <<<<<<<<<<<<<<
 *         return math.floor(self.pvt_o)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_46__floor__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___floor, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_floor_2, __pyx_t_2) < 0) __PYX_ERR(2, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":113
 *         return math.floor(self.pvt_o)
 * 
 *     def __ceil__(self):             # <<<<<<<<<<<<<<
 *         return math.ceil(self.pvt_o)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_48__ceil__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___ceil, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_ceil_2, __pyx_t_2) < 0) __PYX_ERR(2, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":376
 * 
 *     # Just pass to wrapped
 *     def __instancecheck__(self, inst):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_136__instancecheck__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___instancecheck, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_instancecheck, __pyx_t_2) < 0) __PYX_ERR(2, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":379
 *         return self.pvt_o.__instancecheck__(inst)
 * 
 *     def __subclasscheck__(self, subclass):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__subclasscheck__(subclass)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_138__subclasscheck__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___subclasscheck, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_subclasscheck, __pyx_t_2) < 0) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":383
 * 
 *     # Context managers - intrinsic behavior - pass to wrapped
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__enter__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_140__enter__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___enter, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_enter, __pyx_t_2) < 0) __PYX_ERR(2, 383, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":386
 *         return self.pvt_o.__enter__()
 * 
 *     def __exit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__exit(exc_type, exc_value, tb)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_142__exit__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___exit, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_exit, __pyx_t_2) < 0) __PYX_ERR(2, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":390
 * 
 *     # Async context managers - intrinsic behavior - pass to wrapped
 *     def __aenter__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aenter__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_144__aenter__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___aenter, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_aenter, __pyx_t_2) < 0) __PYX_ERR(2, 390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":393
 *         return self.pvt_o.__aenter__()
 * 
 *     def __aexit__(self, exc_type, exc_value, tb):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__aexit(exc_type, exc_value, tb)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_146__aexit__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___aexit, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_aexit, __pyx_t_2) < 0) __PYX_ERR(2, 393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":398
 *     # Customizing positional arguments in class pattern matching
 *     # Returns tuple of strings (immutable) - pass to wrapped
 *     def __match_args__(self):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.__match_args__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_148__match_args__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___match_args, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__115)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_match_args, __pyx_t_2) < 0) __PYX_ERR(2, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":402
 * 
 *     # Coroutine objects - intrinsic behavior - pass to wrapped
 *     def send(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.send(value)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_150send, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_send, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__117)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_send, __pyx_t_2) < 0) __PYX_ERR(2, 402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":405
 *         return self.pvt_o.send(value)
 * 
 *     def throw(self, value):             # <<<<<<<<<<<<<<
 *         return self.pvt_o.throw(value)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_152throw, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_throw, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_throw, __pyx_t_2) < 0) __PYX_ERR(2, 405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":428
 *     # --------------------------------------------------------------------
 * 
 *     def clear(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_158clear, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_clear, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__120)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_clear, __pyx_t_2) < 0) __PYX_ERR(2, 428, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":440
 *         return self.pvt_o.clear(*args, **kwargs)
 * 
 *     def setdefault(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,CollectionsABC.MutableMapping
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_160setdefault, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_setdefault, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_setdefault, __pyx_t_2) < 0) __PYX_ERR(2, 440, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":447
 *         return self.pvt_o.setdefault(*args, **kwargs)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_162pop, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_pop, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__122)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_pop, __pyx_t_2) < 0) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":459
 *         return self.pvt_o.pop(*args, **kwargs)
 * 
 *     def popitem(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_164popitem, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_popitem, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_popitem, __pyx_t_2) < 0) __PYX_ERR(2, 459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":466
 *         return self.pvt_o.popitem(*args, **kwargs)
 * 
 *     def update(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableMapping
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_166update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_update, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__124)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_update, __pyx_t_2) < 0) __PYX_ERR(2, 466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":474
 * 
 * 
 *     def append(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_168append, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_append, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__125)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_append, __pyx_t_2) < 0) __PYX_ERR(2, 474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":481
 *         return self.pvt_o.append(*args, **kwargs)
 * 
 *     def extend(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_170extend, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_extend, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__126)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_extend, __pyx_t_2) < 0) __PYX_ERR(2, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":488
 *         return self.pvt_o.extend(*args, **kwargs)
 * 
 *     def insert(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_172insert, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_insert, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__127)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_insert, __pyx_t_2) < 0) __PYX_ERR(2, 488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":495
 *         return self.pvt_o.insert(*args, **kwargs)
 * 
 *     def sort(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSequence
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_174sort, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_sort, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__128)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_sort, __pyx_t_2) < 0) __PYX_ERR(2, 495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":502
 *         return self.pvt_o.sort(*args, **kwargs)
 * 
 *     def add(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_176add, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_add, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__129)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_add_2, __pyx_t_2) < 0) __PYX_ERR(2, 502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":509
 *         return self.pvt_o.add(*args, **kwargs)
 * 
 *     def discard(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o, CollectionsABC.MutableSet
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_178discard, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_discard, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_discard, __pyx_t_2) < 0) __PYX_ERR(2, 509, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":516
 *         return self.pvt_o.discard(*args, **kwargs)
 * 
 *     def remove(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_180remove, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_remove, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__131)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_remove, __pyx_t_2) < 0) __PYX_ERR(2, 516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Proxy.pxi":527
 *         return self.pvt_o.remove(*args, **kwargs)
 * 
 *     def reverse(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         if self.frozen and isinstance(
 *             self.pvt_o,
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_182reverse, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy_reverse, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__132)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_reverse, __pyx_t_2) < 0) __PYX_ERR(2, 527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_184__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__133)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Proxy, (type(self), 0x914c618, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Proxy__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_5Proxy_186__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Proxy___setstate_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Proxy->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Proxy);

  /* "Wrapped_Frozen.pxi":394
 *         delattr(self.pvt_o, a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.wrapped_dir()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7Wrapped_9__dir__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Wrapped___dir, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__135)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped->tp_dict, __pyx_n_s_dir, __pyx_t_2) < 0) __PYX_ERR(3, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Wrapped);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7Wrapped_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Wrapped___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Wrapped);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Wrapped, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Wrapped__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7Wrapped_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Wrapped___setstate_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__137)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Wrapped->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Wrapped);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_6Frozen_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Frozen___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__138)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Frozen->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Frozen);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Frozen, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Frozen__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_6Frozen_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Frozen___setstate_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__139)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Frozen->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Frozen);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":99
 *         return str(ret)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         # TODO: Something strange here: Just wrapping self.pvt_o and returning
 *         # as a new FrozenPrivacyDict doesn't seem to adopt all the behavior
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_15copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_copy, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__141)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_copy, __pyx_t_2) < 0) __PYX_ERR(4, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":115
 *     # --------------------------------------------------------------------
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         for k in self.pvt_o.keys():
 *             if self.attr_hidden(k):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_17keys, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_keys, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_keys, __pyx_t_2) < 0) __PYX_ERR(4, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":121
 *             yield k
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_20items, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_items, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_items, __pyx_t_2) < 0) __PYX_ERR(4, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":126
 *             yield self.fif((k, v))
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         for k in self.keys():
 *             v = self.pvt_o[k]
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_23values, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_values, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_values, __pyx_t_2) < 0) __PYX_ERR(4, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":135
 *     # --------------------------------------------------------------------
 * 
 *     def keys_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields keys
 *         ret = []
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_26keys_py2, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_keys_py2, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__145)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_keys_py2, __pyx_t_2) < 0) __PYX_ERR(4, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":144
 *         return ret
 * 
 *     def items_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields items
 *         ret = []
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_28items_py2, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_items_py2, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__147)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_items_py2, __pyx_t_2) < 0) __PYX_ERR(4, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":152
 *         return self.fif(ret)
 * 
 *     def values_py2(self):             # <<<<<<<<<<<<<<
 *         # In PY2 returns list, in PY3, yields values
 *         ret = []
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_30values_py2, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_values_py2, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_values_py2, __pyx_t_2) < 0) __PYX_ERR(4, 152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":160
 *         return self.fif(ret)
 * 
 *     def iterkeys(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.pvt_o.keys():
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_32iterkeys, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_iterkeys, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_iterkeys, __pyx_t_2) < 0) __PYX_ERR(4, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":167
 *             yield k
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_35iteritems, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_iteritems, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_iteritems, __pyx_t_2) < 0) __PYX_ERR(4, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":173
 *             yield self.fif((k, v))
 * 
 *     def itervalues(self):             # <<<<<<<<<<<<<<
 *         # PY2 only
 *         for k in self.iterkeys():
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_38itervalues, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_itervalues, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_itervalues, __pyx_t_2) < 0) __PYX_ERR(4, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":178
 *             yield self.fif(self.pvt_o[k])
 * 
 *     def viewitems(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.items()))
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_41viewitems, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_viewitems, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__149)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_viewitems, __pyx_t_2) < 0) __PYX_ERR(4, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":181
 *         return self.fif(set(self.items()))
 * 
 *     def viewvalues(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.values()))
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_43viewvalues, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_viewvalues, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_viewvalues, __pyx_t_2) < 0) __PYX_ERR(4, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "PrivacyDict_FrozenPrivacyDict.pxi":184
 *         return self.fif(set(self.values()))
 * 
 *     def viewkeys(self):             # <<<<<<<<<<<<<<
 *         return self.fif(set(self.keys()))
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_45viewkeys, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict_viewkeys, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__151)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_viewkeys, __pyx_t_2) < 0) __PYX_ERR(4, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_51__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__152)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PrivacyDict__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_11PrivacyDict_53__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PrivacyDict___setstate_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__153)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_PrivacyDict->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_PrivacyDict);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_17FrozenPrivacyDict_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FrozenPrivacyDict___reduce_cytho, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivacyDict, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivacyDict__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_17FrozenPrivacyDict_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FrozenPrivacyDict___setstate_cyt, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__155)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_FrozenPrivacyDict);

  /* "Private_FrozenPrivate.pxi":129
 *         self.private_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.private_dir()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7Private_9__dir__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Private___dir, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Private->tp_dict, __pyx_n_s_dir, __pyx_t_2) < 0) __PYX_ERR(7, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Private);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7Private_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Private___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__157)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Private->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Private);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Private, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Private__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_7Private_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Private___setstate_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Private->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Private);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_13FrozenPrivate_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FrozenPrivate___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__159)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivate->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_FrozenPrivate);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenPrivate, (type(self), 0xc76c111, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenPrivate__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_13FrozenPrivate_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FrozenPrivate___setstate_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenPrivate->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_FrozenPrivate);

  /* "Protected_FrozenProtected.pxi":265
 *         self.protected_check_delattr(a)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return self.protected_dir()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_9Protected_9__dir__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Protected___dir, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__161)); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected->tp_dict, __pyx_n_s_dir, __pyx_t_2) < 0) __PYX_ERR(8, 265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Protected);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_9Protected_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Protected___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Protected);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Protected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Protected__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_9Protected_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Protected___setstate_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__163)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_Protected->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_Protected);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_15FrozenProtected_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FrozenProtected___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__164)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenProtected->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_FrozenProtected);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FrozenProtected, (type(self), 0xd078c39, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FrozenProtected__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_15FrozenProtected_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FrozenProtected___setstate_cytho, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__165)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected_FrozenProtected->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected_FrozenProtected);

  /* "HiddenPartial.pxi":37
 *         raise ProtectionError('Cannot delete attribute: %s' % (a,))
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['__call__']
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_15__HiddenPartial_9__dir__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_HiddenPartial___dir, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__166)); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial->tp_dict, __pyx_n_s_dir, __pyx_t_2) < 0) __PYX_ERR(9, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected___HiddenPartial);

  /* "HiddenPartial.pxi":55
 *         return str(type(functools.partial))
 * 
 *     def __bytes__(self):             # <<<<<<<<<<<<<<
 *         return bytes(type(functools.partial))
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_15__HiddenPartial_17__bytes__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_HiddenPartial___bytes, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__167)); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial->tp_dict, __pyx_n_s_bytes_2, __pyx_t_2) < 0) __PYX_ERR(9, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected___HiddenPartial);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_15__HiddenPartial_21__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_HiddenPartial___reduce_cython, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__168)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial->tp_dict, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected___HiddenPartial);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___HiddenPartial, (type(self), 0x940a50e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___HiddenPartial__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_15__HiddenPartial_23__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_HiddenPartial___setstate_cytho, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__169)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_9pyprotect_9protected___HiddenPartial->tp_dict, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9pyprotect_9protected___HiddenPartial);

  /* "(tree fragment)":1
 * def __pyx_unpickle___ProtectionData(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_55__pyx_unpickle___ProtectionData, 0, __pyx_n_s_pyx_unpickle___ProtectionData, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__171)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___ProtectionData, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle___ProtectionData__set_state(<__ProtectionData> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___ProtectionData__set_state(__ProtectionData __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.attributes_map = __pyx_state[0]; __pyx_result.freeze = __pyx_state[1]; __pyx_result.hash = __pyx_state[2]; __pyx_result.help = __pyx_state[3]; __pyx_result.help_str = __pyx_state[4]; __pyx_result.id = __pyx_state[5]; __pyx_result.id_class = __pyx_state[6]; __pyx_result.instanceof = __pyx_state[7]; __pyx_result.isinstance = __pyx_state[8]; __pyx_result.issubclass = __pyx_state[9]; __pyx_result.multiwrapped = __pyx_state[10]; __pyx_result.private = __pyx_state[11]; __pyx_result.protect = __pyx_state[12]; __pyx_result.rules = __pyx_state[13]; __pyx_result.subclassof = __pyx_state[14]; __pyx_result.testop = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_57__pyx_unpickle_Proxy, 0, __pyx_n_s_pyx_unpickle_Proxy, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__172)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Proxy, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Wrapped(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_59__pyx_unpickle_Wrapped, 0, __pyx_n_s_pyx_unpickle_Wrapped, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__173)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Wrapped, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Wrapped__set_state(<Wrapped> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Wrapped__set_state(Wrapped __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_61__pyx_unpickle_Frozen, 0, __pyx_n_s_pyx_unpickle_Frozen, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__174)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Frozen, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PrivacyDict(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_63__pyx_unpickle_PrivacyDict, 0, __pyx_n_s_pyx_unpickle_PrivacyDict, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__175)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PrivacyDict, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PrivacyDict__set_state(<PrivacyDict> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PrivacyDict__set_state(PrivacyDict __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_65__pyx_unpickle_FrozenPrivacyDict, 0, __pyx_n_s_pyx_unpickle_FrozenPrivacyDict, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FrozenPrivacyDict, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Private(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_67__pyx_unpickle_Private, 0, __pyx_n_s_pyx_unpickle_Private, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__177)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Private, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Private__set_state(<Private> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Private__set_state(Private __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cn = __pyx_state[0]; __pyx_result.frozen = __pyx_state[1]; __pyx_result.hidden_private_attr = __pyx_state[2]; __pyx_result.oldstyle_class = __pyx_state[3]; __pyx_result.protected_attribute = __pyx_state[4]; __pyx_result.pvt_o = __pyx_state[5]; __pyx_result.rules = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_69__pyx_unpickle_FrozenPrivate, 0, __pyx_n_s_pyx_unpickle_FrozenPrivate, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FrozenPrivate, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Protected(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_71__pyx_unpickle_Protected, 0, __pyx_n_s_pyx_unpickle_Protected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__179)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Protected, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Protected__set_state(<Protected> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Protected__set_state(Protected __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acl_cache = __pyx_state[0]; __pyx_result.cn = __pyx_state[1]; __pyx_result.dir_out = __pyx_state[2]; __pyx_result.frozen = __pyx_state[3]; __pyx_result.hidden_private_attr = __pyx_state[4]; __pyx_result.oldstyle_class = __pyx_state[5]; __pyx_result.protected_attribute = __pyx_state[6]; __pyx_result.pvt_o = __pyx_state[7]; __pyx_result.rules = __pyx_state[8]
 *     if len(__pyx_state) > 9 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_73__pyx_unpickle_FrozenProtected, 0, __pyx_n_s_pyx_unpickle_FrozenProtected, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FrozenProtected, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___HiddenPartial(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9pyprotect_9protected_75__pyx_unpickle___HiddenPartial, 0, __pyx_n_s_pyx_unpickle___HiddenPartial, NULL, __pyx_n_s_pyprotect_protected, __pyx_d, ((PyObject *)__pyx_codeobj__181)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___HiddenPartial, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyprotect/protected.pyx":1
 * # coding=utf8             # <<<<<<<<<<<<<<
 * '''
 * Module with methods to wrap an object and additionally restrict
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(10, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op")
 * cdef object __Pyx_CFunc_object__lParenWrapped__comma_object__comma_object__rParen_to_py_4self_1a_2op(object (*f)(Wrapped, object, object) ):             # <<<<<<<<<<<<<<
 *     def wrap(Wrapped self, object a, object op):
 *         """wrap(self: 'Wrapped', a, op)"""
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyprotect.protected", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyprotect.protected");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#if _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        if (kwds_is_tuple) {
            if (pos >= PyTuple_GET_SIZE(kwds)) break;
            key = PyTuple_GET_ITEM(kwds, pos);
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]);
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
    Py_DECREF(argstuple);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
#ifdef __Pyx_CyFunction_USED
        if (__Pyx_IsCyOrPyCFunction(func))
#else
        if (PyCFunction_Check(func))
#endif
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
                return __Pyx_PyObject_CallMethO(func, NULL);
            }
        }
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (PyCFunction_Check(func))
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
                return __Pyx_PyObject_CallMethO(func, args[0]);
            }
        }
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    #if CYTHON_VECTORCALL
    vectorcallfunc f = _PyVectorcall_Function(func);
    if (f) {
        return f(func, args, (size_t)nargs, kwargs);
    }
    #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
    if (__Pyx_CyFunction_CheckExact(func)) {
        __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
        if (f) return f(func, args, (size_t)nargs, kwargs);
    }
    #endif
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg = NULL;
    return __Pyx_PyObject_FastCall(func, (&arg)+1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY ||  CYTHON_COMPILING_IN_LIMITED_API
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            if ((PY_VERSION_HEX >= 0x030700A0)) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    (void) spec;
    (void) type;
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    PyObject *abi_module = PyImport_AddModule((char*) __PYX_ABI_MODULE_NAME);
    if (unlikely(!abi_module)) return NULL;
    Py_INCREF(abi_module);
    return abi_module;
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    (void) module;
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
        PyList_SET_ITEM(fromlist, 0, marker);
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    cf->m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#ifdef _Py_TPFLAGS_HAVE_VECTORCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    (void) module;
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        if (unlikely(PyTuple_GET_SIZE(kw) == 0))
            return 1;
        if (!kw_allowed) {
            key = PyTuple_GET_ITEM(kw, 0);
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < PyTuple_GET_SIZE(kw); pos++) {
            key = PyTuple_GET_ITEM(kw, pos);
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
#endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* RaiseClosureNameError */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectLookupSpecial */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict_2);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (meth) {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                #if CYTHON_COMPILING_IN_LIMITED_API
                module = PyImport_ImportModuleLevelObject(
                    name, empty_dict, empty_dict, from_list, 1);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                #endif
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            #if CYTHON_COMPILING_IN_LIMITED_API
            module = PyImport_ImportModuleLevelObject(
                name, empty_dict, empty_dict, from_list, level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__33);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n = PyTuple_GET_SIZE(bases);
    for (i = 1; i < n; i++)
    {
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
        if (dictoffset == 0 && b->tp_dictoffset)
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "extension type '%.200s' has no __dict__ slot, "
                "but base type '" __Pyx_FMT_TYPENAME "' has: "
                "either add 'cdef dict __dict__' to the extension type "
                "or add '__slots__ = [...]' to the base type",
                type_name, b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__44;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    Py_ssize_t i, nparts;
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (likely(module))
        return module;
    return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    _PyTraceback_Add(funcname, filename, py_line);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XSETREF(name, __Pyx_NewRef(__pyx_n_s__182));
    }
    return name;
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        tmp_tb = PyException_GetTraceback(tmp_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    (void) module;
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compile time version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}
#endif

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#if _MSV_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
